Package: clock
Title: Date-Time Types and Tools
Version: 0.7.4
Authors@R: c(
    person("Davis", "Vaughan", , "davis@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Provides a comprehensive library for date-time manipulations
    using a new family of orthogonal date-time classes (durations, time
    points, zoned-times, and calendars) that partition responsibilities so
    that the complexities of time zones are only considered when they are
    really needed. Capabilities include: date-time parsing, formatting,
    arithmetic, extraction and updating of components, and rounding.
License: MIT + file LICENSE
URL: https://clock.r-lib.org, https://github.com/r-lib/clock
BugReports: https://github.com/r-lib/clock/issues
Depends: R (>= 4.0.0)
Imports: cli (>= 3.6.4), lifecycle (>= 1.0.4), rlang (>= 1.1.5), tzdb
        (>= 0.5.0), vctrs (>= 0.6.5)
Suggests: covr, knitr, magrittr, pillar, rmarkdown, slider (>= 0.3.2),
        testthat (>= 3.0.0), withr
LinkingTo: cpp11 (>= 0.5.2), tzdb (>= 0.5.0)
VignetteBuilder: knitr
Config/build/compilation-database: true
Config/Needs/website: lubridate, tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.2
NeedsCompilation: yes
Packaged: 2026-01-13 14:52:13 UTC; davis
Author: Davis Vaughan [aut, cre],
  Posit Software, PBC [cph, fnd]
Maintainer: Davis Vaughan <davis@posit.co>
Repository: CRAN
Date/Publication: 2026-01-13 16:10:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("[",clock_rcrd)
S3method("[[",clock_rcrd)
S3method("names<-",clock_rcrd)
S3method(add_days,Date)
S3method(add_days,POSIXt)
S3method(add_days,clock_calendar)
S3method(add_days,clock_duration)
S3method(add_days,clock_time_point)
S3method(add_days,clock_weekday)
S3method(add_days,clock_zoned_time)
S3method(add_days,default)
S3method(add_hours,POSIXt)
S3method(add_hours,clock_calendar)
S3method(add_hours,clock_duration)
S3method(add_hours,clock_time_point)
S3method(add_hours,clock_zoned_time)
S3method(add_hours,default)
S3method(add_microseconds,clock_calendar)
S3method(add_microseconds,clock_duration)
S3method(add_microseconds,clock_time_point)
S3method(add_microseconds,clock_zoned_time)
S3method(add_microseconds,default)
S3method(add_milliseconds,clock_calendar)
S3method(add_milliseconds,clock_duration)
S3method(add_milliseconds,clock_time_point)
S3method(add_milliseconds,clock_zoned_time)
S3method(add_milliseconds,default)
S3method(add_minutes,POSIXt)
S3method(add_minutes,clock_calendar)
S3method(add_minutes,clock_duration)
S3method(add_minutes,clock_time_point)
S3method(add_minutes,clock_zoned_time)
S3method(add_minutes,default)
S3method(add_months,Date)
S3method(add_months,POSIXt)
S3method(add_months,clock_duration)
S3method(add_months,clock_time_point)
S3method(add_months,clock_year_month_day)
S3method(add_months,clock_year_month_weekday)
S3method(add_months,clock_zoned_time)
S3method(add_months,default)
S3method(add_nanoseconds,clock_calendar)
S3method(add_nanoseconds,clock_duration)
S3method(add_nanoseconds,clock_time_point)
S3method(add_nanoseconds,clock_zoned_time)
S3method(add_nanoseconds,default)
S3method(add_quarters,Date)
S3method(add_quarters,POSIXt)
S3method(add_quarters,clock_duration)
S3method(add_quarters,clock_time_point)
S3method(add_quarters,clock_year_month_day)
S3method(add_quarters,clock_year_month_weekday)
S3method(add_quarters,clock_year_quarter_day)
S3method(add_quarters,clock_zoned_time)
S3method(add_quarters,default)
S3method(add_seconds,POSIXt)
S3method(add_seconds,clock_calendar)
S3method(add_seconds,clock_duration)
S3method(add_seconds,clock_time_point)
S3method(add_seconds,clock_zoned_time)
S3method(add_seconds,default)
S3method(add_weeks,Date)
S3method(add_weeks,POSIXt)
S3method(add_weeks,clock_calendar)
S3method(add_weeks,clock_duration)
S3method(add_weeks,clock_time_point)
S3method(add_weeks,clock_zoned_time)
S3method(add_weeks,default)
S3method(add_years,Date)
S3method(add_years,POSIXt)
S3method(add_years,clock_duration)
S3method(add_years,clock_iso_year_week_day)
S3method(add_years,clock_time_point)
S3method(add_years,clock_year_day)
S3method(add_years,clock_year_month_day)
S3method(add_years,clock_year_month_weekday)
S3method(add_years,clock_year_quarter_day)
S3method(add_years,clock_year_week_day)
S3method(add_years,clock_zoned_time)
S3method(add_years,default)
S3method(as.Date,clock_calendar)
S3method(as.Date,clock_time_point)
S3method(as.Date,clock_zoned_time)
S3method(as.POSIXct,clock_calendar)
S3method(as.POSIXct,clock_naive_time)
S3method(as.POSIXct,clock_sys_time)
S3method(as.POSIXct,clock_zoned_time)
S3method(as.POSIXlt,clock_calendar)
S3method(as.POSIXlt,clock_naive_time)
S3method(as.POSIXlt,clock_sys_time)
S3method(as.POSIXlt,clock_zoned_time)
S3method(as.character,clock_duration)
S3method(as.character,clock_iso_year_week_day)
S3method(as.character,clock_naive_time)
S3method(as.character,clock_sys_time)
S3method(as.character,clock_weekday)
S3method(as.character,clock_year_day)
S3method(as.character,clock_year_month_day)
S3method(as.character,clock_year_month_weekday)
S3method(as.character,clock_year_quarter_day)
S3method(as.character,clock_year_week_day)
S3method(as.character,clock_zoned_time)
S3method(as.double,clock_duration)
S3method(as.integer,clock_duration)
S3method(as_date,Date)
S3method(as_date,POSIXt)
S3method(as_date,clock_calendar)
S3method(as_date,clock_time_point)
S3method(as_date,clock_zoned_time)
S3method(as_date_time,Date)
S3method(as_date_time,POSIXt)
S3method(as_date_time,clock_calendar)
S3method(as_date_time,clock_naive_time)
S3method(as_date_time,clock_sys_time)
S3method(as_date_time,clock_zoned_time)
S3method(as_duration,clock_duration)
S3method(as_duration,clock_time_point)
S3method(as_iso_year_week_day,Date)
S3method(as_iso_year_week_day,POSIXt)
S3method(as_iso_year_week_day,clock_calendar)
S3method(as_iso_year_week_day,clock_iso_year_week_day)
S3method(as_iso_year_week_day,clock_time_point)
S3method(as_iso_year_week_day,default)
S3method(as_naive_time,Date)
S3method(as_naive_time,POSIXt)
S3method(as_naive_time,clock_duration)
S3method(as_naive_time,clock_iso_year_week_day)
S3method(as_naive_time,clock_naive_time)
S3method(as_naive_time,clock_sys_time)
S3method(as_naive_time,clock_year_day)
S3method(as_naive_time,clock_year_month_day)
S3method(as_naive_time,clock_year_month_weekday)
S3method(as_naive_time,clock_year_quarter_day)
S3method(as_naive_time,clock_year_week_day)
S3method(as_naive_time,clock_zoned_time)
S3method(as_naive_time,default)
S3method(as_sys_time,Date)
S3method(as_sys_time,POSIXt)
S3method(as_sys_time,clock_duration)
S3method(as_sys_time,clock_iso_year_week_day)
S3method(as_sys_time,clock_naive_time)
S3method(as_sys_time,clock_sys_time)
S3method(as_sys_time,clock_year_day)
S3method(as_sys_time,clock_year_month_day)
S3method(as_sys_time,clock_year_month_weekday)
S3method(as_sys_time,clock_year_quarter_day)
S3method(as_sys_time,clock_year_week_day)
S3method(as_sys_time,clock_zoned_time)
S3method(as_sys_time,default)
S3method(as_weekday,Date)
S3method(as_weekday,POSIXt)
S3method(as_weekday,clock_calendar)
S3method(as_weekday,clock_time_point)
S3method(as_weekday,clock_weekday)
S3method(as_year_day,Date)
S3method(as_year_day,POSIXt)
S3method(as_year_day,clock_calendar)
S3method(as_year_day,clock_time_point)
S3method(as_year_day,clock_year_day)
S3method(as_year_day,default)
S3method(as_year_month_day,Date)
S3method(as_year_month_day,POSIXt)
S3method(as_year_month_day,clock_calendar)
S3method(as_year_month_day,clock_time_point)
S3method(as_year_month_day,clock_year_month_day)
S3method(as_year_month_day,default)
S3method(as_year_month_weekday,Date)
S3method(as_year_month_weekday,POSIXt)
S3method(as_year_month_weekday,clock_calendar)
S3method(as_year_month_weekday,clock_time_point)
S3method(as_year_month_weekday,clock_year_month_weekday)
S3method(as_year_month_weekday,default)
S3method(as_year_quarter_day,Date)
S3method(as_year_quarter_day,POSIXt)
S3method(as_year_quarter_day,clock_calendar)
S3method(as_year_quarter_day,clock_time_point)
S3method(as_year_quarter_day,clock_year_quarter_day)
S3method(as_year_quarter_day,default)
S3method(as_year_week_day,Date)
S3method(as_year_week_day,POSIXt)
S3method(as_year_week_day,clock_calendar)
S3method(as_year_week_day,clock_time_point)
S3method(as_year_week_day,clock_year_week_day)
S3method(as_year_week_day,default)
S3method(as_zoned_time,Date)
S3method(as_zoned_time,POSIXt)
S3method(as_zoned_time,clock_naive_time)
S3method(as_zoned_time,clock_sys_time)
S3method(as_zoned_time,clock_zoned_time)
S3method(as_zoned_time,default)
S3method(calendar_count_between,clock_calendar)
S3method(calendar_count_between,clock_iso_year_week_day)
S3method(calendar_count_between,clock_year_day)
S3method(calendar_count_between,clock_year_month_day)
S3method(calendar_count_between,clock_year_month_weekday)
S3method(calendar_count_between,clock_year_quarter_day)
S3method(calendar_count_between,clock_year_week_day)
S3method(calendar_end,clock_calendar)
S3method(calendar_end,clock_iso_year_week_day)
S3method(calendar_end,clock_year_day)
S3method(calendar_end,clock_year_month_day)
S3method(calendar_end,clock_year_month_weekday)
S3method(calendar_end,clock_year_quarter_day)
S3method(calendar_end,clock_year_week_day)
S3method(calendar_group,clock_calendar)
S3method(calendar_group,clock_iso_year_week_day)
S3method(calendar_group,clock_year_day)
S3method(calendar_group,clock_year_month_day)
S3method(calendar_group,clock_year_month_weekday)
S3method(calendar_group,clock_year_quarter_day)
S3method(calendar_group,clock_year_week_day)
S3method(calendar_is_precision,clock_iso_year_week_day)
S3method(calendar_is_precision,clock_year_day)
S3method(calendar_is_precision,clock_year_month_day)
S3method(calendar_is_precision,clock_year_month_weekday)
S3method(calendar_is_precision,clock_year_quarter_day)
S3method(calendar_is_precision,clock_year_week_day)
S3method(calendar_leap_year,clock_calendar)
S3method(calendar_leap_year,clock_iso_year_week_day)
S3method(calendar_leap_year,clock_year_day)
S3method(calendar_leap_year,clock_year_month_day)
S3method(calendar_leap_year,clock_year_month_weekday)
S3method(calendar_leap_year,clock_year_quarter_day)
S3method(calendar_leap_year,clock_year_week_day)
S3method(calendar_month_factor,clock_calendar)
S3method(calendar_month_factor,clock_year_month_day)
S3method(calendar_month_factor,clock_year_month_weekday)
S3method(calendar_name,clock_iso_year_week_day)
S3method(calendar_name,clock_year_day)
S3method(calendar_name,clock_year_month_day)
S3method(calendar_name,clock_year_month_weekday)
S3method(calendar_name,clock_year_quarter_day)
S3method(calendar_name,clock_year_week_day)
S3method(calendar_narrow,clock_calendar)
S3method(calendar_narrow,clock_iso_year_week_day)
S3method(calendar_narrow,clock_year_day)
S3method(calendar_narrow,clock_year_month_day)
S3method(calendar_narrow,clock_year_month_weekday)
S3method(calendar_narrow,clock_year_quarter_day)
S3method(calendar_narrow,clock_year_week_day)
S3method(calendar_precision,clock_calendar)
S3method(calendar_start,clock_calendar)
S3method(calendar_start,clock_iso_year_week_day)
S3method(calendar_start,clock_year_day)
S3method(calendar_start,clock_year_month_day)
S3method(calendar_start,clock_year_month_weekday)
S3method(calendar_start,clock_year_quarter_day)
S3method(calendar_start,clock_year_week_day)
S3method(calendar_widen,clock_calendar)
S3method(calendar_widen,clock_iso_year_week_day)
S3method(calendar_widen,clock_year_day)
S3method(calendar_widen,clock_year_month_day)
S3method(calendar_widen,clock_year_month_weekday)
S3method(calendar_widen,clock_year_quarter_day)
S3method(calendar_widen,clock_year_week_day)
S3method(clock_maximum,clock_duration)
S3method(clock_maximum,clock_iso_year_week_day)
S3method(clock_maximum,clock_time_point)
S3method(clock_maximum,clock_year_day)
S3method(clock_maximum,clock_year_month_day)
S3method(clock_maximum,clock_year_month_weekday)
S3method(clock_maximum,clock_year_quarter_day)
S3method(clock_maximum,clock_year_week_day)
S3method(clock_minimum,clock_duration)
S3method(clock_minimum,clock_iso_year_week_day)
S3method(clock_minimum,clock_time_point)
S3method(clock_minimum,clock_year_day)
S3method(clock_minimum,clock_year_month_day)
S3method(clock_minimum,clock_year_month_weekday)
S3method(clock_minimum,clock_year_quarter_day)
S3method(clock_minimum,clock_year_week_day)
S3method(date_ceiling,Date)
S3method(date_ceiling,POSIXt)
S3method(date_count_between,Date)
S3method(date_count_between,POSIXt)
S3method(date_end,Date)
S3method(date_end,POSIXt)
S3method(date_floor,Date)
S3method(date_floor,POSIXt)
S3method(date_format,Date)
S3method(date_format,POSIXt)
S3method(date_group,Date)
S3method(date_group,POSIXt)
S3method(date_leap_year,Date)
S3method(date_leap_year,POSIXt)
S3method(date_round,Date)
S3method(date_round,POSIXt)
S3method(date_seq,Date)
S3method(date_seq,POSIXt)
S3method(date_set_zone,Date)
S3method(date_set_zone,POSIXt)
S3method(date_shift,Date)
S3method(date_shift,POSIXt)
S3method(date_spanning_seq,Date)
S3method(date_spanning_seq,POSIXt)
S3method(date_start,Date)
S3method(date_start,POSIXt)
S3method(date_zone,Date)
S3method(date_zone,POSIXt)
S3method(format,clock_duration)
S3method(format,clock_iso_year_week_day)
S3method(format,clock_time_point)
S3method(format,clock_weekday)
S3method(format,clock_year_day)
S3method(format,clock_year_month_day)
S3method(format,clock_year_month_weekday)
S3method(format,clock_year_quarter_day)
S3method(format,clock_year_week_day)
S3method(format,clock_zoned_time)
S3method(get_day,Date)
S3method(get_day,POSIXt)
S3method(get_day,clock_iso_year_week_day)
S3method(get_day,clock_year_day)
S3method(get_day,clock_year_month_day)
S3method(get_day,clock_year_month_weekday)
S3method(get_day,clock_year_quarter_day)
S3method(get_day,clock_year_week_day)
S3method(get_day,default)
S3method(get_hour,POSIXt)
S3method(get_hour,clock_iso_year_week_day)
S3method(get_hour,clock_year_day)
S3method(get_hour,clock_year_month_day)
S3method(get_hour,clock_year_month_weekday)
S3method(get_hour,clock_year_quarter_day)
S3method(get_hour,clock_year_week_day)
S3method(get_hour,default)
S3method(get_index,clock_year_month_weekday)
S3method(get_index,default)
S3method(get_microsecond,clock_iso_year_week_day)
S3method(get_microsecond,clock_year_day)
S3method(get_microsecond,clock_year_month_day)
S3method(get_microsecond,clock_year_month_weekday)
S3method(get_microsecond,clock_year_quarter_day)
S3method(get_microsecond,clock_year_week_day)
S3method(get_microsecond,default)
S3method(get_millisecond,clock_iso_year_week_day)
S3method(get_millisecond,clock_year_day)
S3method(get_millisecond,clock_year_month_day)
S3method(get_millisecond,clock_year_month_weekday)
S3method(get_millisecond,clock_year_quarter_day)
S3method(get_millisecond,clock_year_week_day)
S3method(get_millisecond,default)
S3method(get_minute,POSIXt)
S3method(get_minute,clock_iso_year_week_day)
S3method(get_minute,clock_year_day)
S3method(get_minute,clock_year_month_day)
S3method(get_minute,clock_year_month_weekday)
S3method(get_minute,clock_year_quarter_day)
S3method(get_minute,clock_year_week_day)
S3method(get_minute,default)
S3method(get_month,Date)
S3method(get_month,POSIXt)
S3method(get_month,clock_year_month_day)
S3method(get_month,clock_year_month_weekday)
S3method(get_month,default)
S3method(get_nanosecond,clock_iso_year_week_day)
S3method(get_nanosecond,clock_year_day)
S3method(get_nanosecond,clock_year_month_day)
S3method(get_nanosecond,clock_year_month_weekday)
S3method(get_nanosecond,clock_year_quarter_day)
S3method(get_nanosecond,clock_year_week_day)
S3method(get_nanosecond,default)
S3method(get_quarter,clock_year_quarter_day)
S3method(get_quarter,default)
S3method(get_second,POSIXt)
S3method(get_second,clock_iso_year_week_day)
S3method(get_second,clock_year_day)
S3method(get_second,clock_year_month_day)
S3method(get_second,clock_year_month_weekday)
S3method(get_second,clock_year_quarter_day)
S3method(get_second,clock_year_week_day)
S3method(get_second,default)
S3method(get_week,clock_iso_year_week_day)
S3method(get_week,clock_year_week_day)
S3method(get_week,default)
S3method(get_year,Date)
S3method(get_year,POSIXt)
S3method(get_year,clock_iso_year_week_day)
S3method(get_year,clock_year_day)
S3method(get_year,clock_year_month_day)
S3method(get_year,clock_year_month_weekday)
S3method(get_year,clock_year_quarter_day)
S3method(get_year,clock_year_week_day)
S3method(get_year,default)
S3method(invalid_any,clock_iso_year_week_day)
S3method(invalid_any,clock_year_day)
S3method(invalid_any,clock_year_month_day)
S3method(invalid_any,clock_year_month_weekday)
S3method(invalid_any,clock_year_quarter_day)
S3method(invalid_any,clock_year_week_day)
S3method(invalid_any,default)
S3method(invalid_count,clock_iso_year_week_day)
S3method(invalid_count,clock_year_day)
S3method(invalid_count,clock_year_month_day)
S3method(invalid_count,clock_year_month_weekday)
S3method(invalid_count,clock_year_quarter_day)
S3method(invalid_count,clock_year_week_day)
S3method(invalid_count,default)
S3method(invalid_detect,clock_iso_year_week_day)
S3method(invalid_detect,clock_year_day)
S3method(invalid_detect,clock_year_month_day)
S3method(invalid_detect,clock_year_month_weekday)
S3method(invalid_detect,clock_year_quarter_day)
S3method(invalid_detect,clock_year_week_day)
S3method(invalid_detect,default)
S3method(invalid_remove,clock_calendar)
S3method(invalid_remove,default)
S3method(invalid_resolve,clock_iso_year_week_day)
S3method(invalid_resolve,clock_year_day)
S3method(invalid_resolve,clock_year_month_day)
S3method(invalid_resolve,clock_year_month_weekday)
S3method(invalid_resolve,clock_year_quarter_day)
S3method(invalid_resolve,clock_year_week_day)
S3method(invalid_resolve,default)
S3method(max,clock_rcrd)
S3method(min,clock_rcrd)
S3method(names,clock_rcrd)
S3method(obj_print_data,clock_calendar)
S3method(obj_print_data,clock_duration)
S3method(obj_print_data,clock_time_point)
S3method(obj_print_data,clock_zoned_time)
S3method(obj_print_footer,clock_calendar)
S3method(obj_print_footer,clock_time_point)
S3method(obj_print_footer,clock_zoned_time)
S3method(print,clock_calendar)
S3method(print,clock_labels)
S3method(print,clock_locale)
S3method(print,clock_time_point)
S3method(print,clock_zoned_time)
S3method(range,clock_rcrd)
S3method(seq,clock_duration)
S3method(seq,clock_iso_year_week_day)
S3method(seq,clock_time_point)
S3method(seq,clock_year_day)
S3method(seq,clock_year_month_day)
S3method(seq,clock_year_month_weekday)
S3method(seq,clock_year_quarter_day)
S3method(seq,clock_year_week_day)
S3method(set_day,Date)
S3method(set_day,POSIXt)
S3method(set_day,clock_iso_year_week_day)
S3method(set_day,clock_year_day)
S3method(set_day,clock_year_month_day)
S3method(set_day,clock_year_month_weekday)
S3method(set_day,clock_year_quarter_day)
S3method(set_day,clock_year_week_day)
S3method(set_day,default)
S3method(set_hour,POSIXt)
S3method(set_hour,clock_iso_year_week_day)
S3method(set_hour,clock_year_day)
S3method(set_hour,clock_year_month_day)
S3method(set_hour,clock_year_month_weekday)
S3method(set_hour,clock_year_quarter_day)
S3method(set_hour,clock_year_week_day)
S3method(set_hour,default)
S3method(set_index,clock_year_month_weekday)
S3method(set_index,default)
S3method(set_microsecond,clock_iso_year_week_day)
S3method(set_microsecond,clock_year_day)
S3method(set_microsecond,clock_year_month_day)
S3method(set_microsecond,clock_year_month_weekday)
S3method(set_microsecond,clock_year_quarter_day)
S3method(set_microsecond,clock_year_week_day)
S3method(set_microsecond,default)
S3method(set_millisecond,clock_iso_year_week_day)
S3method(set_millisecond,clock_year_day)
S3method(set_millisecond,clock_year_month_day)
S3method(set_millisecond,clock_year_month_weekday)
S3method(set_millisecond,clock_year_quarter_day)
S3method(set_millisecond,clock_year_week_day)
S3method(set_millisecond,default)
S3method(set_minute,POSIXt)
S3method(set_minute,clock_iso_year_week_day)
S3method(set_minute,clock_year_day)
S3method(set_minute,clock_year_month_day)
S3method(set_minute,clock_year_month_weekday)
S3method(set_minute,clock_year_quarter_day)
S3method(set_minute,clock_year_week_day)
S3method(set_minute,default)
S3method(set_month,Date)
S3method(set_month,POSIXt)
S3method(set_month,clock_year_month_day)
S3method(set_month,clock_year_month_weekday)
S3method(set_month,default)
S3method(set_nanosecond,clock_iso_year_week_day)
S3method(set_nanosecond,clock_year_day)
S3method(set_nanosecond,clock_year_month_day)
S3method(set_nanosecond,clock_year_month_weekday)
S3method(set_nanosecond,clock_year_quarter_day)
S3method(set_nanosecond,clock_year_week_day)
S3method(set_nanosecond,default)
S3method(set_quarter,clock_year_quarter_day)
S3method(set_quarter,default)
S3method(set_second,POSIXt)
S3method(set_second,clock_iso_year_week_day)
S3method(set_second,clock_year_day)
S3method(set_second,clock_year_month_day)
S3method(set_second,clock_year_month_weekday)
S3method(set_second,clock_year_quarter_day)
S3method(set_second,clock_year_week_day)
S3method(set_second,default)
S3method(set_week,clock_iso_year_week_day)
S3method(set_week,clock_year_week_day)
S3method(set_week,default)
S3method(set_year,Date)
S3method(set_year,POSIXt)
S3method(set_year,clock_iso_year_week_day)
S3method(set_year,clock_year_day)
S3method(set_year,clock_year_month_day)
S3method(set_year,clock_year_month_weekday)
S3method(set_year,clock_year_quarter_day)
S3method(set_year,clock_year_week_day)
S3method(set_year,default)
S3method(vec_arith,clock_duration)
S3method(vec_arith,clock_iso_year_week_day)
S3method(vec_arith,clock_naive_time)
S3method(vec_arith,clock_sys_time)
S3method(vec_arith,clock_weekday)
S3method(vec_arith,clock_year_day)
S3method(vec_arith,clock_year_month_day)
S3method(vec_arith,clock_year_month_weekday)
S3method(vec_arith,clock_year_quarter_day)
S3method(vec_arith,clock_year_week_day)
S3method(vec_arith.Date,clock_duration)
S3method(vec_arith.POSIXct,clock_duration)
S3method(vec_arith.POSIXlt,clock_duration)
S3method(vec_arith.clock_duration,Date)
S3method(vec_arith.clock_duration,MISSING)
S3method(vec_arith.clock_duration,POSIXct)
S3method(vec_arith.clock_duration,POSIXlt)
S3method(vec_arith.clock_duration,clock_duration)
S3method(vec_arith.clock_duration,clock_iso_year_week_day)
S3method(vec_arith.clock_duration,clock_naive_time)
S3method(vec_arith.clock_duration,clock_sys_time)
S3method(vec_arith.clock_duration,clock_weekday)
S3method(vec_arith.clock_duration,clock_year_day)
S3method(vec_arith.clock_duration,clock_year_month_day)
S3method(vec_arith.clock_duration,clock_year_month_weekday)
S3method(vec_arith.clock_duration,clock_year_quarter_day)
S3method(vec_arith.clock_duration,clock_year_week_day)
S3method(vec_arith.clock_duration,numeric)
S3method(vec_arith.clock_iso_year_week_day,MISSING)
S3method(vec_arith.clock_iso_year_week_day,clock_duration)
S3method(vec_arith.clock_iso_year_week_day,clock_iso_year_week_day)
S3method(vec_arith.clock_iso_year_week_day,numeric)
S3method(vec_arith.clock_naive_time,MISSING)
S3method(vec_arith.clock_naive_time,clock_duration)
S3method(vec_arith.clock_naive_time,clock_naive_time)
S3method(vec_arith.clock_naive_time,numeric)
S3method(vec_arith.clock_sys_time,MISSING)
S3method(vec_arith.clock_sys_time,clock_duration)
S3method(vec_arith.clock_sys_time,clock_sys_time)
S3method(vec_arith.clock_sys_time,numeric)
S3method(vec_arith.clock_weekday,MISSING)
S3method(vec_arith.clock_weekday,clock_duration)
S3method(vec_arith.clock_weekday,clock_weekday)
S3method(vec_arith.clock_weekday,numeric)
S3method(vec_arith.clock_year_day,MISSING)
S3method(vec_arith.clock_year_day,clock_duration)
S3method(vec_arith.clock_year_day,clock_year_day)
S3method(vec_arith.clock_year_day,numeric)
S3method(vec_arith.clock_year_month_day,MISSING)
S3method(vec_arith.clock_year_month_day,clock_duration)
S3method(vec_arith.clock_year_month_day,clock_year_month_day)
S3method(vec_arith.clock_year_month_day,numeric)
S3method(vec_arith.clock_year_month_weekday,MISSING)
S3method(vec_arith.clock_year_month_weekday,clock_duration)
S3method(vec_arith.clock_year_month_weekday,clock_year_month_weekday)
S3method(vec_arith.clock_year_month_weekday,numeric)
S3method(vec_arith.clock_year_quarter_day,MISSING)
S3method(vec_arith.clock_year_quarter_day,clock_duration)
S3method(vec_arith.clock_year_quarter_day,clock_year_quarter_day)
S3method(vec_arith.clock_year_quarter_day,numeric)
S3method(vec_arith.clock_year_week_day,MISSING)
S3method(vec_arith.clock_year_week_day,clock_duration)
S3method(vec_arith.clock_year_week_day,clock_year_week_day)
S3method(vec_arith.clock_year_week_day,numeric)
S3method(vec_arith.numeric,clock_duration)
S3method(vec_arith.numeric,clock_iso_year_week_day)
S3method(vec_arith.numeric,clock_naive_time)
S3method(vec_arith.numeric,clock_sys_time)
S3method(vec_arith.numeric,clock_weekday)
S3method(vec_arith.numeric,clock_year_day)
S3method(vec_arith.numeric,clock_year_month_day)
S3method(vec_arith.numeric,clock_year_month_weekday)
S3method(vec_arith.numeric,clock_year_quarter_day)
S3method(vec_arith.numeric,clock_year_week_day)
S3method(vec_cast,clock_duration.clock_duration)
S3method(vec_cast,clock_iso_year_week_day.clock_iso_year_week_day)
S3method(vec_cast,clock_naive_time.clock_naive_time)
S3method(vec_cast,clock_sys_time.clock_sys_time)
S3method(vec_cast,clock_weekday.clock_weekday)
S3method(vec_cast,clock_year_day.clock_year_day)
S3method(vec_cast,clock_year_month_day.clock_year_month_day)
S3method(vec_cast,clock_year_month_weekday.clock_year_month_weekday)
S3method(vec_cast,clock_year_quarter_day.clock_year_quarter_day)
S3method(vec_cast,clock_year_week_day.clock_year_week_day)
S3method(vec_cast,clock_zoned_time.clock_zoned_time)
S3method(vec_math,clock_duration)
S3method(vec_math,clock_rcrd)
S3method(vec_math,clock_weekday)
S3method(vec_proxy,clock_duration)
S3method(vec_proxy,clock_iso_year_week_day)
S3method(vec_proxy,clock_time_point)
S3method(vec_proxy,clock_year_day)
S3method(vec_proxy,clock_year_month_day)
S3method(vec_proxy,clock_year_month_weekday)
S3method(vec_proxy,clock_year_quarter_day)
S3method(vec_proxy,clock_year_week_day)
S3method(vec_proxy,clock_zoned_time)
S3method(vec_proxy_compare,clock_weekday)
S3method(vec_proxy_compare,clock_year_month_weekday)
S3method(vec_ptype,clock_iso_year_week_day)
S3method(vec_ptype,clock_naive_time)
S3method(vec_ptype,clock_sys_time)
S3method(vec_ptype,clock_weekday)
S3method(vec_ptype,clock_year_day)
S3method(vec_ptype,clock_year_month_day)
S3method(vec_ptype,clock_year_month_weekday)
S3method(vec_ptype,clock_year_quarter_day)
S3method(vec_ptype,clock_year_week_day)
S3method(vec_ptype,clock_zoned_time)
S3method(vec_ptype2,clock_duration.clock_duration)
S3method(vec_ptype2,clock_iso_year_week_day.clock_iso_year_week_day)
S3method(vec_ptype2,clock_naive_time.clock_naive_time)
S3method(vec_ptype2,clock_sys_time.clock_sys_time)
S3method(vec_ptype2,clock_weekday.clock_weekday)
S3method(vec_ptype2,clock_year_day.clock_year_day)
S3method(vec_ptype2,clock_year_month_day.clock_year_month_day)
S3method(vec_ptype2,clock_year_month_weekday.clock_year_month_weekday)
S3method(vec_ptype2,clock_year_quarter_day.clock_year_quarter_day)
S3method(vec_ptype2,clock_year_week_day.clock_year_week_day)
S3method(vec_ptype2,clock_zoned_time.clock_zoned_time)
S3method(vec_ptype_abbr,clock_duration)
S3method(vec_ptype_abbr,clock_iso_year_week_day)
S3method(vec_ptype_abbr,clock_naive_time)
S3method(vec_ptype_abbr,clock_sys_time)
S3method(vec_ptype_abbr,clock_weekday)
S3method(vec_ptype_abbr,clock_year_day)
S3method(vec_ptype_abbr,clock_year_month_day)
S3method(vec_ptype_abbr,clock_year_month_weekday)
S3method(vec_ptype_abbr,clock_year_quarter_day)
S3method(vec_ptype_abbr,clock_year_week_day)
S3method(vec_ptype_abbr,clock_zoned_time)
S3method(vec_ptype_full,clock_duration)
S3method(vec_ptype_full,clock_iso_year_week_day)
S3method(vec_ptype_full,clock_naive_time)
S3method(vec_ptype_full,clock_sys_time)
S3method(vec_ptype_full,clock_weekday)
S3method(vec_ptype_full,clock_year_day)
S3method(vec_ptype_full,clock_year_month_day)
S3method(vec_ptype_full,clock_year_month_weekday)
S3method(vec_ptype_full,clock_year_quarter_day)
S3method(vec_ptype_full,clock_year_week_day)
S3method(vec_ptype_full,clock_zoned_time)
S3method(vec_restore,clock_duration)
S3method(vec_restore,clock_iso_year_week_day)
S3method(vec_restore,clock_time_point)
S3method(vec_restore,clock_year_day)
S3method(vec_restore,clock_year_month_day)
S3method(vec_restore,clock_year_month_weekday)
S3method(vec_restore,clock_year_quarter_day)
S3method(vec_restore,clock_year_week_day)
S3method(vec_restore,clock_zoned_time)
export(add_days)
export(add_hours)
export(add_microseconds)
export(add_milliseconds)
export(add_minutes)
export(add_months)
export(add_nanoseconds)
export(add_quarters)
export(add_seconds)
export(add_weeks)
export(add_years)
export(as_date)
export(as_date_time)
export(as_duration)
export(as_iso_year_week_day)
export(as_naive_time)
export(as_sys_time)
export(as_weekday)
export(as_year_day)
export(as_year_month_day)
export(as_year_month_weekday)
export(as_year_quarter_day)
export(as_year_week_day)
export(as_zoned_time)
export(calendar_count_between)
export(calendar_end)
export(calendar_group)
export(calendar_leap_year)
export(calendar_month_factor)
export(calendar_narrow)
export(calendar_precision)
export(calendar_spanning_seq)
export(calendar_start)
export(calendar_widen)
export(clock_labels)
export(clock_labels_languages)
export(clock_labels_lookup)
export(clock_locale)
export(date_build)
export(date_ceiling)
export(date_count_between)
export(date_end)
export(date_floor)
export(date_format)
export(date_group)
export(date_leap_year)
export(date_month_factor)
export(date_now)
export(date_parse)
export(date_round)
export(date_seq)
export(date_set_zone)
export(date_shift)
export(date_spanning_seq)
export(date_start)
export(date_time_build)
export(date_time_info)
export(date_time_parse)
export(date_time_parse_RFC_3339)
export(date_time_parse_abbrev)
export(date_time_parse_complete)
export(date_time_set_zone)
export(date_time_zone)
export(date_today)
export(date_weekday_factor)
export(date_zone)
export(duration_cast)
export(duration_ceiling)
export(duration_days)
export(duration_floor)
export(duration_hours)
export(duration_microseconds)
export(duration_milliseconds)
export(duration_minutes)
export(duration_months)
export(duration_nanoseconds)
export(duration_precision)
export(duration_quarters)
export(duration_round)
export(duration_seconds)
export(duration_spanning_seq)
export(duration_weeks)
export(duration_years)
export(get_day)
export(get_hour)
export(get_index)
export(get_microsecond)
export(get_millisecond)
export(get_minute)
export(get_month)
export(get_nanosecond)
export(get_quarter)
export(get_second)
export(get_week)
export(get_year)
export(invalid_any)
export(invalid_count)
export(invalid_detect)
export(invalid_remove)
export(invalid_resolve)
export(is_duration)
export(is_iso_year_week_day)
export(is_naive_time)
export(is_sys_time)
export(is_weekday)
export(is_year_day)
export(is_year_month_day)
export(is_year_month_weekday)
export(is_year_quarter_day)
export(is_year_week_day)
export(is_zoned_time)
export(iso_year_week_day)
export(naive_time_info)
export(naive_time_parse)
export(set_day)
export(set_hour)
export(set_index)
export(set_microsecond)
export(set_millisecond)
export(set_minute)
export(set_month)
export(set_nanosecond)
export(set_quarter)
export(set_second)
export(set_week)
export(set_year)
export(sys_time_info)
export(sys_time_now)
export(sys_time_parse)
export(sys_time_parse_RFC_3339)
export(time_point_cast)
export(time_point_ceiling)
export(time_point_count_between)
export(time_point_floor)
export(time_point_precision)
export(time_point_round)
export(time_point_shift)
export(time_point_spanning_seq)
export(tzdb_names)
export(tzdb_version)
export(weekday)
export(weekday_code)
export(weekday_factor)
export(year_day)
export(year_month_day)
export(year_month_day_parse)
export(year_month_weekday)
export(year_quarter_day)
export(year_week_day)
export(zoned_time_info)
export(zoned_time_now)
export(zoned_time_parse_abbrev)
export(zoned_time_parse_complete)
export(zoned_time_precision)
export(zoned_time_set_zone)
export(zoned_time_zone)
import(rlang)
import(vctrs)
importFrom(tzdb,tzdb_names)
importFrom(tzdb,tzdb_version)
useDynLib(clock, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# clock <a href="https://clock.r-lib.org"><img src="man/figures/logo.png" align="right" height="138"/></a>

<!-- badges: start -->

[![Codecov test
coverage](https://codecov.io/gh/r-lib/clock/graph/badge.svg)](https://app.codecov.io/gh/r-lib/clock)
[![R-CMD-check](https://github.com/r-lib/clock/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/clock/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

clock is an R package for working with date-times. It is packed with
features, including utilities for: parsing, formatting, arithmetic,
rounding, and extraction/updating of individual components. In addition
to these tools for manipulating date-times, clock provides entirely new
date-time types which are structured to reduce the agony of working with
time zones as much as possible. At a high-level, clock:

- Provides a new family of date-time classes (durations, time points,
  zoned-times, and calendars) that partition responsibilities so that
  you only have to think about time zones when you need them.

- Implements a high level API for R’s native date (Date) and date-time
  (POSIXct) classes that lets you get productive quickly without having
  to learn the details of clock’s new date-time types.

- Requires explicit handling of invalid dates (e.g. what date is one
  month after January 31st?) and nonexistent or ambiguous times (caused
  by daylight saving time issues).

- Is built on the C++ [date](https://github.com/HowardHinnant/date)
  library, which provides a correct and high-performance backend.

There are four key classes in clock, inspired by the design of the C++
date and chrono libraries. Some types are more efficient than others at
particular operations, and it is only when all 4 are taken as a whole do
you get a complete date time library.

- A *duration* counts units of time, like “5 years” or “6 nanoseconds”.
  Bigger units are defined in terms of seconds, i.e. 1 day is 86400
  seconds and 1 year is 365.2425 days. Durations are important because
  they form the backbone of clock; it’s relatively rare to use them
  directly.

- A *time point* records an instant in time, like “1:24pm January 1st
  2015”. It combines a *duration* with a “clock” that defines when to
  start counting and what exactly to count. There are two important
  types of time in clock: sys-time and naive-time. They’re equivalent
  until you start working with zoned-times.

- A *zoned-time* is a time point paired with a time zone. You can create
  them from either a sys-time or a naive-time, depending on whether you
  want to convert the time point from UTC (leaving the underlying
  duration unchanged, but changing the printed time), or declare that
  the time point is in a specific time zone (leaving the printed time
  unchanged, but changing the underlying duration). Zoned-times are
  primarily needed for communication with humans.

- A *calendar* represents a date using a combination of fields like
  year-month-day, year-month-weekday, year-quarter-day, year-day, or
  iso-year-week-day, along with hour/minute/second fields to represent
  time within a day (so they’re similar to R’s POSIXlt). Calendar
  objects are extremely efficient at arithmetic involving irregular
  periods such as months, quarters, and years and at getting and setting
  specified components. A calendar can represent invalid dates (like
  2020-02-31) which only need to be resolved when converting back to a
  time point.

## Installation

Install the released version from [CRAN](https://CRAN.R-project.org)
with:

``` r
install.packages("clock")
```

Install the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("r-lib/clock")
```

## Learning

With clock, there is a high-level API for manipulating R’s native date
and date-time types (Date, POSIXct, and POSIXlt), as well as a low-level
API for more advanced date-time manipulations. The high-level API should
allow you to quickly get productive with clock without having to learn
many of the new types. An easy way to get an overview of this is by
looking at the [High Level
API](https://clock.r-lib.org/reference/index.html#section-high-level-api)
section of the pkgdown reference page.

You’ll also want to take a look at clock’s vignettes:

- [Getting Started](https://clock.r-lib.org/articles/clock.html)

- [Motivations for
  clock](https://clock.r-lib.org/articles/articles/motivations.html)

- [Examples and Recipes](https://clock.r-lib.org/articles/recipes.html)

- [Frequently Asked
  Questions](https://clock.r-lib.org/articles/faq.html)

You can also watch this [5 minute lightning
talk](https://www.youtube.com/watch?v=DPU7TJZXWvo) about clock:

<p>

<a href="https://www.youtube.com/watch?v=DPU7TJZXWvo">
<img src="https://embed-ssl.wistia.com/deliveries/3ee43fc82b901742748feddc5d3e9796dcd7d7b9.jpg?image_play_button_size=2x&amp;image_crop_resized=960x540&amp;image_play_button=1&amp;image_play_button_color=4287c7e0" width="600" height="337" style="width: 600px; height: 337px;"/></a>

</p>

## Acknowledgments

The ideas used to build clock have come from a number of places:

- First and foremost, clock depends on and is inspired by the
  [date](https://github.com/HowardHinnant/date) library by Howard
  Hinnant, a variant of which has been voted in to C++20.

- The R “record” types that clock is built on come from
  [vctrs](https://github.com/r-lib/vctrs).

- The [nanotime](https://github.com/eddelbuettel/nanotime) package was
  the first to implement a nanosecond resolution time type for R.

- The [zoo](https://CRAN.R-project.org/package=zoo) package was the
  first to implement year-month and year-quarter types, and has
  functioned as a very successful time series infrastructure package for
  many years.

Additionally, I’d like to thank my wife for being a power user of clock,
and having a number of discussions with me around bugs and missing
features. You can thank her for `date_count_between()`.

#' Create a clock locale
#'
#' A clock locale contains the information required to format and parse dates.
#' The defaults have been chosen to match US English. A clock locale object can
#' be provided to `format()` methods or parse functions (like
#' [year_month_day_parse()]) to override the defaults.
#'
#' @param labels `[clock_labels / character(1)]`
#'
#'   Character representations of localized weekday names, month names, and
#'   AM/PM names. Either the language code as string (passed on to
#'   [clock_labels_lookup()]), or an object created by [clock_labels()].
#'
#' @param decimal_mark `[character(1)]`
#'
#'   Symbol used for the decimal place when formatting sub-second date-times.
#'   Either `","` or `"."`.
#'
#' @return A `"clock_locale"` object.
#'
#' @export
#' @examples
#' clock_locale()
#' clock_locale(labels = "fr")
clock_locale <- function(labels = "en", decimal_mark = ".") {
  if (is_character(labels)) {
    labels <- clock_labels_lookup(labels)
  }
  if (!is_clock_labels(labels)) {
    abort("`labels` must be a 'clock_labels' object.")
  }

  ok <- identical(decimal_mark, ".") || identical(decimal_mark, ",")
  if (!ok) {
    abort("`decimal_mark` must be either ',' or '.'.")
  }

  new_clock_locale(labels, decimal_mark)
}

new_clock_locale <- function(labels, decimal_mark) {
  structure(
    list(
      labels = labels,
      decimal_mark = decimal_mark
    ),
    class = "clock_locale"
  )
}

#' @export
print.clock_locale <- function (x, ...) {
  cat("<clock_locale>\n")
  cat("Decimal Mark: ", x$decimal_mark, "\n", sep = "")
  print(x$labels)
}

is_clock_locale <- function(x) {
  inherits(x, "clock_locale")
}

test_that("can create a default locale object", {
  locale <- clock_locale()
  expect_s3_class(locale, "clock_locale")
  expect_snapshot(locale)
})

test_that("can change labels with either the language name or labels object", {
  labels1 <- clock_locale("fr")
  labels2 <- clock_locale(clock_labels_lookup("fr"))
  expect_identical(labels1, labels2)
})

test_that("must use a valid clock-labels object", {
  expect_snapshot_error(clock_locale(1))
})

test_that("must use a valid decimal-mark", {
  expect_snapshot_error(clock_locale(decimal_mark = "f"))
})

test_that("can change the decimal-mark", {
  x <- clock_locale(decimal_mark = ",")
  expect_identical(x$decimal_mark, ",")
  expect_snapshot(x)
})

#' Calendar: year-month-day
#'
#' `year_month_day()` constructs the most common calendar type using the
#' Gregorian year, month, day, and time of day components.
#'
#' @details
#' Fields are recycled against each other.
#'
#' Fields are collected in order until the first `NULL` field is located. No
#' fields after the first `NULL` field are used.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param year `[integer]`
#'
#'   The year. Values `[-32767, 32767]` are generally allowed.
#'
#' @param month `[integer / NULL]`
#'
#'   The month. Values `[1, 12]` are allowed.
#'
#' @param day `[integer / "last" / NULL]`
#'
#'   The day of the month. Values `[1, 31]` are allowed.
#'
#'   If `"last"`, then the last day of the month is returned.
#'
#' @param hour `[integer / NULL]`
#'
#'   The hour. Values `[0, 23]` are allowed.
#'
#' @param minute `[integer / NULL]`
#'
#'   The minute. Values `[0, 59]` are allowed.
#'
#' @param second `[integer / NULL]`
#'
#'   The second. Values `[0, 59]` are allowed.
#'
#' @param subsecond `[integer / NULL]`
#'
#'   The subsecond. If specified, `subsecond_precision` must also be specified
#'   to determine how to interpret the `subsecond`.
#'
#'   If using milliseconds, values `[0, 999]` are allowed.
#'
#'   If using microseconds, values `[0, 999999]` are allowed.
#'
#'   If using nanoseconds, values `[0, 999999999]` are allowed.
#'
#' @param subsecond_precision `[character(1) / NULL]`
#'
#'   The precision to interpret `subsecond` as. One of: `"millisecond"`,
#'   `"microsecond"`, or `"nanosecond"`.
#'
#' @return A year-month-day calendar vector.
#'
#' @export
#' @examples
#' # Just the year
#' x <- year_month_day(2019:2025)
#'
#' # Year-month type
#' year_month_day(2020, 1:12)
#'
#' # The most common use case involves year, month, and day fields
#' x <- year_month_day(2020, clock_months$january, 1:5)
#' x
#'
#' # Precision can go all the way out to nanosecond
#' year_month_day(2019, 1, 2, 2, 40, 45, 200, subsecond_precision = "nanosecond")
year_month_day <- function(year,
                           month = NULL,
                           day = NULL,
                           hour = NULL,
                           minute = NULL,
                           second = NULL,
                           subsecond = NULL,
                           ...,
                           subsecond_precision = NULL) {
  check_dots_empty()

  # Stop on the first `NULL` argument
  if (is_null(month)) {
    precision <- PRECISION_YEAR
    fields <- list(year = year)
  } else if (is_null(day)) {
    precision <- PRECISION_MONTH
    fields <- list(year = year, month = month)
  } else if (is_null(hour)) {
    precision <- PRECISION_DAY
    fields <- list(year = year, month = month, day = day)
  } else if (is_null(minute)) {
    precision <- PRECISION_HOUR
    fields <- list(year = year, month = month, day = day, hour = hour)
  } else if (is_null(second)) {
    precision <- PRECISION_MINUTE
    fields <- list(year = year, month = month, day = day, hour = hour, minute = minute)
  } else if (is_null(subsecond)) {
    precision <- PRECISION_SECOND
    fields <- list(year = year, month = month, day = day, hour = hour, minute = minute, second = second)
  } else {
    precision <- calendar_validate_subsecond_precision(subsecond_precision)
    fields <- list(year = year, month = month, day = day, hour = hour, minute = minute, second = second, subsecond = subsecond)
  }

  if (is_last(fields$day)) {
    fields$day <- 1L
    last <- TRUE
  } else {
    last <- FALSE
  }

  fields <- vec_recycle_common(!!!fields)
  fields <- vec_cast_common(!!!fields, .to = integer())

  fields <- collect_year_month_day_fields(fields, precision)

  names <- NULL

  out <- new_year_month_day_from_fields(fields, precision, names)

  if (last) {
    out <- set_day(out, "last")
  }

  out
}

# ------------------------------------------------------------------------------

#' @export
vec_proxy.clock_year_month_day <- function(x, ...) {
  .Call(`_clock_clock_rcrd_proxy`, x)
}

#' @export
vec_restore.clock_year_month_day <- function(x, to, ...) {
  .Call(`_clock_year_month_day_restore`, x, to)
}

# ------------------------------------------------------------------------------

#' @export
format.clock_year_month_day <- function(x, ...) {
  out <- format_year_month_day_cpp(x, calendar_precision_attribute(x))
  names(out) <- names(x)
  out
}

#' @export
vec_ptype_full.clock_year_month_day <- function(x, ...) {
  calendar_ptype_full(x, "year_month_day")
}

#' @export
vec_ptype_abbr.clock_year_month_day <- function(x, ...) {
  calendar_ptype_abbr(x, "ymd")
}

# ------------------------------------------------------------------------------

#' Parsing: year-month-day
#'
#' @description
#' `year_month_day_parse()` parses strings into a year-month-day.
#'
#' The default options assume `x` should be parsed at day precision, using a
#' `format` string of `"%Y-%m-%d"`.
#'
#' If a more precise precision than day is used, then time components will also
#' be parsed. The default format separates date and time components by a `"T"`
#' and the time components by a `":"`. For example, setting the precision to
#' `"second"` will use a default format of `"%Y-%m-%dT%H:%M:%S"`. This is
#' aligned with the [format()] method for year-month-day, and with the RFC 3339
#' standard.
#'
#' @details
#' `year_month_day_parse()` completely ignores the `%z` and `%Z` commands.
#'
#' @inheritSection zoned-parsing Full Precision Parsing
#'
#' @inheritParams zoned-parsing
#'
#' @param x `[character]`
#'
#'   A character vector to parse.
#'
#' @param precision `[character(1)]`
#'
#'   A precision for the resulting year-month-day. One of:
#'
#'   - `"year"`
#'   - `"month"`
#'   - `"day"`
#'   - `"hour"`
#'   - `"minute"`
#'   - `"second"`
#'   - `"millisecond"`
#'   - `"microsecond"`
#'   - `"nanosecond"`
#'
#'   Setting the `precision` determines how much information `%S` attempts
#'   to parse.
#'
#' @param locale `[clock_locale]`
#'
#'   A locale object created by [clock_locale()].
#'
#' @return A year-month-day calendar vector. If a parsing fails, `NA` is
#'   returned.
#'
#' @export
#' @examples
#' x <- "2019-01-01"
#'
#' # Default parses at day precision
#' year_month_day_parse(x)
#'
#' # Can parse at less precise precisions too
#' year_month_day_parse(x, precision = "month")
#' year_month_day_parse(x, precision = "year")
#'
#' # Even invalid dates can be round-tripped through format<->parse calls
#' invalid <- year_month_day(2019, 2, 30)
#' year_month_day_parse(format(invalid))
#'
#' # Can parse with time of day
#' year_month_day_parse(
#'   "2019-01-30T02:30:00.123456789",
#'   precision = "nanosecond"
#' )
#'
#' # Can parse using multiple format strings, which will be tried
#' # in the order they are provided
#' x <- c("2019-01-01", "2020-01-01", "2021/2/3")
#' formats <- c("%Y-%m-%d", "%Y/%m/%d")
#' year_month_day_parse(x, format = formats)
#'
#' # Can parse using other format tokens as well
#' year_month_day_parse(
#'   "January, 2019",
#'   format = "%B, %Y",
#'   precision = "month"
#' )
#'
#' # Parsing a French year-month-day
#' year_month_day_parse(
#'   "octobre 1, 2000",
#'   format = "%B %d, %Y",
#'   locale = clock_locale("fr")
#' )
year_month_day_parse <- function(x,
                                 ...,
                                 format = NULL,
                                 precision = "day",
                                 locale = clock_locale()) {
  check_dots_empty()

  if (!is_character(x)) {
    abort("`x` must be a character vector.")
  }

  precision <- validate_precision_string(precision)
  if (!year_month_day_is_valid_precision(precision)) {
    abort("`precision` must be a valid precision for 'year_month_day'.")
  }

  if (is_null(format)) {
    format <- year_month_day_format(precision)
  }

  if (!is_clock_locale(locale)) {
    abort("`locale` must be a 'clock_locale' object.")
  }

  labels <- locale$labels
  mark <- locale$decimal_mark

  fields <- year_month_day_parse_cpp(
    x,
    format,
    precision,
    labels$month,
    labels$month_abbrev,
    labels$weekday,
    labels$weekday_abbrev,
    labels$am_pm,
    mark
  )

  new_year_month_day_from_fields(fields, precision, names(x))
}

year_month_day_format <- function(precision) {
  precision <- precision_to_string(precision)

  switch(
    precision,
    year = "%Y",
    month = "%Y-%m",
    day = "%Y-%m-%d",
    hour = "%Y-%m-%dT%H",
    minute = "%Y-%m-%dT%H:%M",
    second = ,
    millisecond = ,
    microsecond = ,
    nanosecond = "%Y-%m-%dT%H:%M:%S",
    abort("Internal error: Unknown precision.")
  )
}

# ------------------------------------------------------------------------------

#' Is `x` a year-month-day?
#'
#' Check if `x` is a year-month-day.
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @return Returns `TRUE` if `x` inherits from `"clock_year_month_day"`,
#'   otherwise returns `FALSE`.
#'
#' @export
#' @examples
#' is_year_month_day(year_month_day(2019))
is_year_month_day <- function(x) {
  inherits(x, "clock_year_month_day")
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype.clock_year_month_day <- function(x, ...) {
  switch(
    calendar_precision_attribute(x) + 1L,
    clock_empty_year_month_day_year,
    abort("Internal error: Invalid precision"),
    clock_empty_year_month_day_month,
    abort("Internal error: Invalid precision"),
    clock_empty_year_month_day_day,
    clock_empty_year_month_day_hour,
    clock_empty_year_month_day_minute,
    clock_empty_year_month_day_second,
    clock_empty_year_month_day_millisecond,
    clock_empty_year_month_day_microsecond,
    clock_empty_year_month_day_nanosecond,
    abort("Internal error: Invalid precision.")
  )
}

#' @export
vec_ptype2.clock_year_month_day.clock_year_month_day <- function(x, y, ...) {
  ptype2_calendar_and_calendar(x, y, ...)
}

#' @export
vec_cast.clock_year_month_day.clock_year_month_day <- function(x, to, ...) {
  cast_calendar_to_calendar(x, to, ...)
}

# ------------------------------------------------------------------------------

#' @export
calendar_is_valid_precision.clock_year_month_day <- function(x, precision) {
  year_month_day_is_valid_precision(precision)
}

year_month_day_is_valid_precision <- function(precision) {
  if (!is_valid_precision(precision)) {
    FALSE
  } else if (precision == PRECISION_YEAR || precision == PRECISION_MONTH) {
    TRUE
  } else if (precision >= PRECISION_DAY && precision <= PRECISION_NANOSECOND) {
    TRUE
  } else {
    FALSE
  }
}

# ------------------------------------------------------------------------------

#' @export
invalid_detect.clock_year_month_day <- function(x) {
  invalid_detect_year_month_day_cpp(x, calendar_precision_attribute(x))
}

#' @export
invalid_any.clock_year_month_day <- function(x) {
  invalid_any_year_month_day_cpp(x, calendar_precision_attribute(x))
}

#' @export
invalid_count.clock_year_month_day <- function(x) {
  invalid_count_year_month_day_cpp(x, calendar_precision_attribute(x))
}

#' @export
invalid_resolve.clock_year_month_day <- function(x, ..., invalid = NULL) {
  check_dots_empty()
  precision <- calendar_precision_attribute(x)
  invalid <- validate_invalid(invalid)
  fields <- invalid_resolve_year_month_day_cpp(x, precision, invalid)
  new_year_month_day_from_fields(fields, precision, names(x))
}

# ------------------------------------------------------------------------------

#' Getters: year-month-day
#'
#' @description
#' These are year-month-day methods for the [getter generics][clock-getters].
#'
#' - `get_year()` returns the Gregorian year.
#'
#' - `get_month()` returns the month of the year.
#'
#' - `get_day()` returns the day of the month.
#'
#' - There are sub-daily getters for extracting more precise components.
#'
#' @param x `[clock_year_month_day]`
#'
#'   A year-month-day to get the component from.
#'
#' @return The component.
#'
#' @name year-month-day-getters
#' @examples
#' x <- year_month_day(2019, 1:3, 5:7, 1, 20, 30)
#'
#' get_month(x)
#' get_day(x)
#' get_second(x)
#'
#' # Cannot extract more precise components
#' y <- year_month_day(2019, 1)
#' try(get_day(y))
#'
#' # Cannot extract components that don't exist for this calendar
#' try(get_quarter(x))
NULL

#' @rdname year-month-day-getters
#' @export
get_year.clock_year_month_day <- function(x) {
  field_year(x)
}

#' @rdname year-month-day-getters
#' @export
get_month.clock_year_month_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_MONTH, "get_month")
  field_month(x)
}

#' @rdname year-month-day-getters
#' @export
get_day.clock_year_month_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_DAY, "get_day")
  field_day(x)
}

#' @rdname year-month-day-getters
#' @export
get_hour.clock_year_month_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_HOUR, "get_hour")
  field_hour(x)
}

#' @rdname year-month-day-getters
#' @export
get_minute.clock_year_month_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_MINUTE, "get_minute")
  field_minute(x)
}

#' @rdname year-month-day-getters
#' @export
get_second.clock_year_month_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_SECOND, "get_second")
  field_second(x)
}

#' @rdname year-month-day-getters
#' @export
get_millisecond.clock_year_month_day <- function(x) {
  calendar_require_precision(x, PRECISION_MILLISECOND, "get_millisecond")
  field_subsecond(x)
}

#' @rdname year-month-day-getters
#' @export
get_microsecond.clock_year_month_day <- function(x) {
  calendar_require_precision(x, PRECISION_MICROSECOND, "get_microsecond")
  field_subsecond(x)
}

#' @rdname year-month-day-getters
#' @export
get_nanosecond.clock_year_month_day <- function(x) {
  calendar_require_precision(x, PRECISION_NANOSECOND, "get_nanosecond")
  field_subsecond(x)
}

# ------------------------------------------------------------------------------

#' Setters: year-month-day
#'
#' @description
#' These are year-month-day methods for the
#' [setter generics][clock-setters].
#'
#' - `set_year()` sets the Gregorian year.
#'
#' - `set_month()` sets the month of the year. Valid values are in the range
#'   of `[1, 12]`.
#'
#' - `set_day()` sets the day of the month. Valid values are in the range
#'   of `[1, 31]`.
#'
#' - There are sub-daily setters for setting more precise components.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[clock_year_month_day]`
#'
#'   A year-month-day vector.
#'
#' @param value `[integer / "last"]`
#'
#'   The value to set the component to.
#'
#'   For `set_day()`, this can also be `"last"` to set the day to the
#'   last day of the month.
#'
#' @return `x` with the component set.
#'
#' @name year-month-day-setters
#' @examples
#' x <- year_month_day(2019, 1:3)
#'
#' # Set the day
#' set_day(x, 12:14)
#'
#' # Set to the "last" day of the month
#' set_day(x, "last")
#'
#' # Set to an invalid day of the month
#' invalid <- set_day(x, 31)
#' invalid
#'
#' # Then resolve the invalid day by choosing the next valid day
#' invalid_resolve(invalid, invalid = "next")
#'
#' # Cannot set a component two levels more precise than where you currently are
#' try(set_hour(x, 5))
NULL

#' @rdname year-month-day-setters
#' @export
set_year.clock_year_month_day <- function(x, value, ...) {
  check_dots_empty()
  set_field_year_month_day(x, value, "year")
}

#' @rdname year-month-day-setters
#' @export
set_month.clock_year_month_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_YEAR, "set_month")
  set_field_year_month_day(x, value, "month")
}

#' @rdname year-month-day-setters
#' @export
set_day.clock_year_month_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_MONTH, "set_day")
  set_field_year_month_day(x, value, "day")
}

#' @rdname year-month-day-setters
#' @export
set_hour.clock_year_month_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_DAY, "set_hour")
  set_field_year_month_day(x, value, "hour")
}

#' @rdname year-month-day-setters
#' @export
set_minute.clock_year_month_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_HOUR, "set_minute")
  set_field_year_month_day(x, value, "minute")
}

#' @rdname year-month-day-setters
#' @export
set_second.clock_year_month_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_MINUTE, "set_second")
  set_field_year_month_day(x, value, "second")
}

#' @rdname year-month-day-setters
#' @export
set_millisecond.clock_year_month_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_MILLISECOND), "set_millisecond")
  set_field_year_month_day(x, value, "millisecond")
}

#' @rdname year-month-day-setters
#' @export
set_microsecond.clock_year_month_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_MICROSECOND), "set_microsecond")
  set_field_year_month_day(x, value, "microsecond")
}

#' @rdname year-month-day-setters
#' @export
set_nanosecond.clock_year_month_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_NANOSECOND), "set_nanosecond")
  set_field_year_month_day(x, value, "nanosecond")
}

set_field_year_month_day <- function(x, value, component) {
  if (is_last(value) && identical(component, "day")) {
    return(set_field_year_month_day_last(x))
  }

  precision_fields <- calendar_precision_attribute(x)
  precision_value <- year_month_day_component_to_precision(component)
  precision_out <- precision_common2(precision_fields, precision_value)

  value <- vec_cast(value, integer(), x_arg = "value")
  args <- vec_recycle_common(x = x, value = value)
  x <- args$x
  value <- args$value

  result <- set_field_year_month_day_cpp(x, value, precision_fields, precision_value)
  fields <- result$fields
  field <- year_month_day_component_to_field(component)
  fields[[field]] <- result$value

  new_year_month_day_from_fields(fields, precision_out, names = names(x))
}

set_field_year_month_day_last <- function(x) {
  precision_fields <- calendar_precision_attribute(x)
  precision_out <- precision_common2(precision_fields, PRECISION_DAY)

  result <- set_field_year_month_day_last_cpp(x, precision_fields)
  fields <- result$fields
  fields[["day"]] <- result$value

  new_year_month_day_from_fields(fields, precision_out, names = names(x))
}

# ------------------------------------------------------------------------------

#' @export
calendar_name.clock_year_month_day <- function(x) {
  "year_month_day"
}

# ------------------------------------------------------------------------------

year_month_day_component_to_precision <- function(component) {
  switch(
    component,
    year = PRECISION_YEAR,
    month = PRECISION_MONTH,
    day = PRECISION_DAY,
    hour = PRECISION_HOUR,
    minute = PRECISION_MINUTE,
    second = PRECISION_SECOND,
    millisecond = PRECISION_MILLISECOND,
    microsecond = PRECISION_MICROSECOND,
    nanosecond = PRECISION_NANOSECOND,
    abort("Internal error: Unknown component name.")
  )
}

year_month_day_component_to_field <- function(component) {
  switch (
    component,
    year = component,
    month = component,
    day = component,
    hour = component,
    minute = component,
    second = component,
    millisecond = "subsecond",
    microsecond = "subsecond",
    nanosecond = "subsecond",
    abort("Internal error: Unknown component name.")
  )
}

# ------------------------------------------------------------------------------

#' Support for vctrs arithmetic
#'
#' @inheritParams vctrs::vec_arith
#' @name clock-arith
#'
#' @return The result of the arithmetic operation.
#' @examples
#' vctrs::vec_arith("+", year_month_day(2019), 1)
NULL

#' @rdname clock-arith
#' @method vec_arith clock_year_month_day
#' @export
vec_arith.clock_year_month_day <- function(op, x, y, ...) {
  UseMethod("vec_arith.clock_year_month_day", y)
}

#' @method vec_arith.clock_year_month_day MISSING
#' @export
vec_arith.clock_year_month_day.MISSING <- function(op, x, y, ...) {
  arith_calendar_and_missing(op, x, y, ...)
}

#' @method vec_arith.clock_year_month_day clock_year_month_day
#' @export
vec_arith.clock_year_month_day.clock_year_month_day <- function(op, x, y, ...) {
  arith_calendar_and_calendar(op, x, y, ..., calendar_minus_calendar_fn = year_month_day_minus_year_month_day)
}

#' @method vec_arith.clock_year_month_day clock_duration
#' @export
vec_arith.clock_year_month_day.clock_duration <- function(op, x, y, ...) {
  arith_calendar_and_duration(op, x, y, ...)
}

#' @method vec_arith.clock_duration clock_year_month_day
#' @export
vec_arith.clock_duration.clock_year_month_day <- function(op, x, y, ...) {
  arith_duration_and_calendar(op, x, y, ...)
}

#' @method vec_arith.clock_year_month_day numeric
#' @export
vec_arith.clock_year_month_day.numeric <- function(op, x, y, ...) {
  arith_calendar_and_numeric(op, x, y, ...)
}

#' @method vec_arith.numeric clock_year_month_day
#' @export
vec_arith.numeric.clock_year_month_day <- function(op, x, y, ...) {
  arith_numeric_and_calendar(op, x, y, ...)
}

year_month_day_minus_year_month_day <- function(op, x, y, ...) {
  args <- vec_recycle_common(x = x, y = y)
  args <- vec_cast_common(!!!args)
  x <- args$x
  y <- args$y

  names <- names_common(x, y)

  precision <- calendar_precision_attribute(x)

  if (precision > PRECISION_MONTH) {
    stop_incompatible_op(op, x, y, ...)
  }

  fields <- year_month_day_minus_year_month_day_cpp(x, y, precision)

  new_duration_from_fields(fields, precision, names)
}

# ------------------------------------------------------------------------------

#' Arithmetic: year-month-day
#'
#' @description
#' These are year-month-day methods for the
#' [arithmetic generics][clock-arithmetic].
#'
#' - `add_years()`
#'
#' - `add_quarters()`
#'
#' - `add_months()`
#'
#' Notably, _you cannot add days to a year-month-day_. For day-based arithmetic,
#' first convert to a time point with [as_naive_time()] or [as_sys_time()].
#'
#' @details
#' Adding a single quarter with `add_quarters()` is equivalent to adding
#' 3 months.
#'
#' `x` and `n` are recycled against each other.
#'
#' @inheritParams add_years
#'
#' @param x `[clock_year_month_day]`
#'
#'   A year-month-day vector.
#'
#' @return `x` after performing the arithmetic.
#'
#' @name year-month-day-arithmetic
#'
#' @examples
#' x <- year_month_day(2019, 1, 1)
#'
#' add_years(x, 1:5)
#'
#' y <- year_month_day(2019, 1, 31)
#'
#' # Adding 1 month to `y` generates an invalid date
#' y_plus <- add_months(y, 1:2)
#' y_plus
#'
#' # Invalid dates are fine, as long as they are eventually resolved
#' # by either manually resolving, or by calling `invalid_resolve()`
#'
#' # Resolve by returning the previous / next valid moment in time
#' invalid_resolve(y_plus, invalid = "previous")
#' invalid_resolve(y_plus, invalid = "next")
#'
#' # Manually resolve by setting to the last day of the month
#' invalid <- invalid_detect(y_plus)
#' y_plus[invalid] <- set_day(y_plus[invalid], "last")
#' y_plus
NULL

#' @rdname year-month-day-arithmetic
#' @export
add_years.clock_year_month_day <- function(x, n, ...) {
  year_month_day_plus_duration(x, n, PRECISION_YEAR)
}

#' @rdname year-month-day-arithmetic
#' @export
add_quarters.clock_year_month_day <- function(x, n, ...) {
  calendar_require_minimum_precision(x, PRECISION_MONTH, "add_quarters")
  year_month_day_plus_duration(x, n, PRECISION_QUARTER)
}

#' @rdname year-month-day-arithmetic
#' @export
add_months.clock_year_month_day <- function(x, n, ...) {
  calendar_require_minimum_precision(x, PRECISION_MONTH, "add_months")
  year_month_day_plus_duration(x, n, PRECISION_MONTH)
}

year_month_day_plus_duration <- function(x, n, precision_n) {
  precision_fields <- calendar_precision_attribute(x)

  n <- duration_collect_n(n, precision_n)
  args <- vec_recycle_common(x = x, n = n)
  x <- args$x
  n <- args$n

  names <- names_common(x, n)

  fields <- year_month_day_plus_duration_cpp(x, n, precision_fields, precision_n)

  new_year_month_day_from_fields(fields, precision_fields, names = names)
}

# ------------------------------------------------------------------------------

#' Convert to year-month-day
#'
#' `as_year_month_day()` converts a vector to the year-month-day calendar.
#' Time points, Dates, POSIXct, and other calendars can all be converted to
#' year-month-day.
#'
#' @param x `[vector]`
#'
#'   A vector to convert to year-month-day.
#'
#' @return A year-month-day vector.
#' @export
#' @examples
#' # From Date
#' as_year_month_day(as.Date("2019-01-01"))
#'
#' # From POSIXct, which assumes that the naive time is what should be converted
#' as_year_month_day(as.POSIXct("2019-01-01 02:30:30", "America/New_York"))
#'
#' # From other calendars
#' as_year_month_day(year_quarter_day(2019, quarter = 2, day = 50))
as_year_month_day <- function(x)  {
  UseMethod("as_year_month_day")
}

#' @export
as_year_month_day.default <- function(x) {
  stop_clock_unsupported_conversion(x, "clock_year_month_day")
}

#' @export
as_year_month_day.clock_year_month_day <- function(x) {
  x
}

# ------------------------------------------------------------------------------

#' @export
as_sys_time.clock_year_month_day <- function(x) {
  calendar_require_all_valid(x)
  precision <- calendar_precision_attribute(x)
  fields <- as_sys_time_year_month_day_cpp(x, precision)
  new_sys_time_from_fields(fields, precision, clock_rcrd_names(x))
}

#' @export
as_naive_time.clock_year_month_day <- function(x) {
  as_naive_time(as_sys_time(x))
}

#' @export
as.character.clock_year_month_day <- function(x, ...) {
  format(x)
}

# ------------------------------------------------------------------------------

#' @export
calendar_leap_year.clock_year_month_day <- function(x) {
  x <- get_year(x)
  gregorian_leap_year_cpp(x)
}

# ------------------------------------------------------------------------------

#' @export
calendar_month_factor.clock_year_month_day <- function(x,
                                                       ...,
                                                       labels = "en",
                                                       abbreviate = FALSE) {
  calendar_month_factor_impl(x, labels, abbreviate, ...)
}

# ------------------------------------------------------------------------------

#' Grouping: year-month-day
#'
#' @description
#' This is a year-month-day method for the [calendar_group()] generic.
#'
#' Grouping for a year-month-day object can be done at any precision, as
#' long as `x` is at least as precise as `precision`.
#'
#' @inheritParams calendar_group
#'
#' @param x `[clock_year_month_day]`
#'
#'   A year-month-day vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"month"`
#'   - `"day"`
#'   - `"hour"`
#'   - `"minute"`
#'   - `"second"`
#'   - `"millisecond"`
#'   - `"microsecond"`
#'   - `"nanosecond"`
#'
#' @return `x` grouped at the specified `precision`.
#'
#' @name year-month-day-group
#'
#' @export
#' @examples
#' steps <- duration_days(seq(0, 100, by = 5))
#' x <- year_month_day(2019, 1, 1)
#' x <- as_naive_time(x) + steps
#' x <- as_year_month_day(x)
#' x
#'
#' # Group by a single month
#' calendar_group(x, "month")
#'
#' # Or multiple months
#' calendar_group(x, "month", n = 2)
#'
#' # Group 3 days of the month together
#' y <- year_month_day(2019, 1, 1:12)
#' calendar_group(y, "day", n = 3)
#'
#' # Group by 5 nanosecond of the current second
#' z <- year_month_day(
#'   2019, 1, 2, 1, 5, 20, 1:20,
#'   subsecond_precision = "nanosecond"
#' )
#' calendar_group(z, "nanosecond", n = 5)
calendar_group.clock_year_month_day <- function(x, precision, ..., n = 1L) {
  n <- validate_calendar_group_n(n)
  x <- calendar_narrow(x, precision)

  precision <- validate_precision_string(precision)

  if (precision == PRECISION_YEAR) {
    value <- get_year(x)
    value <- group_component0(value, n)
    x <- set_year(x, value)
    return(x)
  }
  if (precision == PRECISION_MONTH) {
    value <- get_month(x)
    value <- group_component1(value, n)
    x <- set_month(x, value)
    return(x)
  }
  if (precision == PRECISION_DAY) {
    value <- get_day(x)
    value <- group_component1(value, n)
    x <- set_day(x, value)
    return(x)
  }

  x <- calendar_group_time(x, n, precision)
  x
}

# ------------------------------------------------------------------------------

#' Narrow: year-month-day
#'
#' This is a year-month-day method for the [calendar_narrow()] generic. It
#' narrows a year-month-day vector to the specified `precision`.
#'
#' @inheritParams year-month-day-group
#'
#' @return `x` narrowed to the supplied `precision`.
#'
#' @name year-month-day-narrow
#'
#' @export
#' @examples
#' # Hour precision
#' x <- year_month_day(2019, 1, 3, 4)
#' x
#'
#' # Narrowed to day precision
#' calendar_narrow(x, "day")
#'
#' # Or month precision
#' calendar_narrow(x, "month")
#'
#' # Subsecond precision can be narrowed to second precision
#' milli <- calendar_widen(x, "millisecond")
#' micro <- calendar_widen(x, "microsecond")
#' milli
#' micro
#'
#' calendar_narrow(milli, "second")
#' calendar_narrow(micro, "second")
#'
#' # But once you have "locked in" a subsecond precision, it can't be
#' # narrowed to another subsecond precision
#' try(calendar_narrow(micro, "millisecond"))
calendar_narrow.clock_year_month_day <- function(x, precision) {
  precision <- validate_precision_string(precision)

  out_fields <- list()
  x_fields <- unclass(x)

  if (precision >= PRECISION_YEAR) {
    out_fields[["year"]] <- x_fields[["year"]]
  }
  if (precision >= PRECISION_MONTH) {
    out_fields[["month"]] <- x_fields[["month"]]
  }
  if (precision >= PRECISION_DAY) {
    out_fields[["day"]] <- x_fields[["day"]]
  }

  out_fields <- calendar_narrow_time(out_fields, precision, x_fields)

  new_year_month_day_from_fields(out_fields, precision, names = names(x))
}

# ------------------------------------------------------------------------------

#' Widen: year-month-day
#'
#' This is a year-month-day method for the [calendar_widen()] generic. It
#' widens a year-month-day vector to the specified `precision`.
#'
#' @inheritParams year-month-day-group
#'
#' @return `x` widened to the supplied `precision`.
#'
#' @name year-month-day-widen
#'
#' @export
#' @examples
#' # Month precision
#' x <- year_month_day(2019, 1)
#' x
#'
#' # Widen to day precision
#' calendar_widen(x, "day")
#'
#' # Or second precision
#' sec <- calendar_widen(x, "second")
#' sec
#'
#' # Second precision can be widened to subsecond precision
#' milli <- calendar_widen(sec, "millisecond")
#' micro <- calendar_widen(sec, "microsecond")
#' milli
#' micro
#'
#' # But once you have "locked in" a subsecond precision, it can't
#' # be widened again
#' try(calendar_widen(milli, "microsecond"))
calendar_widen.clock_year_month_day <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  if (precision >= PRECISION_MONTH && x_precision < PRECISION_MONTH) {
    x <- set_month(x, 1L)
  }
  if (precision >= PRECISION_DAY && x_precision < PRECISION_DAY) {
    x <- set_day(x, 1L)
  }

  x <- calendar_widen_time(x, x_precision, precision)

  x
}

# ------------------------------------------------------------------------------

#' Boundaries: year-month-day
#'
#' This is a year-month-day method for the [calendar_start()] and
#' [calendar_end()] generics. They adjust components of a calendar to the
#' start or end of a specified `precision`.
#'
#' @inheritParams year-month-day-group
#'
#' @return `x` at the same precision, but with some components altered to be
#'   at the boundary value.
#'
#' @name year-month-day-boundary
#'
#' @examples
#' # Hour precision
#' x <- year_month_day(2019, 2:4, 5, 6)
#' x
#'
#' # Compute the start of the month
#' calendar_start(x, "month")
#'
#' # Or the end of the month, notice that the hour value is adjusted as well
#' calendar_end(x, "month")
#'
#' # Compare that with just setting the day of the month to `"last"`, which
#' # doesn't adjust any other components
#' set_day(x, "last")
#'
#' # You can't compute the start / end at a more precise precision than
#' # the input is at
#' try(calendar_start(x, "second"))
NULL

#' @rdname year-month-day-boundary
#' @export
calendar_start.clock_year_month_day <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  calendar_start_end_checks(x, x_precision, precision, "start")

  if (precision <= PRECISION_YEAR && x_precision > PRECISION_YEAR) {
    x <- set_month(x, 1L)
  }
  if (precision <= PRECISION_MONTH && x_precision > PRECISION_MONTH) {
    x <- set_day(x, 1L)
  }

  x <- calendar_start_time(x, x_precision, precision)

  x
}

#' @rdname year-month-day-boundary
#' @export
calendar_end.clock_year_month_day <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  calendar_start_end_checks(x, x_precision, precision, "end")

  if (precision <= PRECISION_YEAR && x_precision > PRECISION_YEAR) {
    x <- set_month(x, 12L)
  }
  if (precision <= PRECISION_MONTH && x_precision > PRECISION_MONTH) {
    x <- set_day(x, "last")
  }

  x <- calendar_end_time(x, x_precision, precision)

  x
}

# ------------------------------------------------------------------------------

#' Counting: year-month-day
#'
#' This is a year-month-day method for the [calendar_count_between()] generic.
#' It counts the number of `precision` units between `start` and `end`
#' (i.e., the number of years or months).
#'
#' @details
#' `"quarter"` is equivalent to `"month"` precision with `n` set to `n * 3L`.
#'
#' @inheritParams calendar-count-between
#'
#' @param start,end `[clock_year_month_day]`
#'
#'   A pair of year-month-day vectors. These will be recycled to their
#'   common size.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"quarter"`
#'   - `"month"`
#'
#' @inherit calendar-count-between return
#'
#' @name year-month-day-count-between
#'
#' @export
#' @examples
#' # Compute an individual's age in years
#' x <- year_month_day(2001, 2, 4)
#' today <- year_month_day(2021, 11, 30)
#' calendar_count_between(x, today, "year")
#'
#' # Compute the number of months between two dates, taking
#' # into account the day of the month and time of day
#' x <- year_month_day(2000, 4, 2, 5)
#' y <- year_month_day(2000, 7, c(1, 2, 2), c(3, 4, 6))
#' calendar_count_between(x, y, "month")
calendar_count_between.clock_year_month_day <- function(start,
                                                        end,
                                                        precision,
                                                        ...,
                                                        n = 1L) {
  NextMethod()
}

calendar_count_between_standardize_precision_n.clock_year_month_day <- function(x,
                                                                                precision,
                                                                                n) {
  precision_int <- validate_precision_string(precision)

  allowed_precisions <- c(PRECISION_YEAR, PRECISION_QUARTER, PRECISION_MONTH)

  if (!(precision_int %in% allowed_precisions)) {
    abort("`precision` must be one of: 'year', 'quarter', 'month'.")
  }

  if (precision_int == PRECISION_QUARTER) {
    # See #270's comment for why this is well-defined
    precision <- "month"
    n <- n * 3L
  }

  list(precision = precision, n = n)
}

calendar_count_between_compute.clock_year_month_day <- function(start,
                                                                end,
                                                                precision) {
  precision <- validate_precision_string(precision)

  if (precision == PRECISION_YEAR) {
    out <- get_year(end) - get_year(start)
    return(out)
  }

  if (precision == PRECISION_MONTH) {
    out <- (get_year(end) - get_year(start)) * 12L
    out <- out + (get_month(end) - get_month(start))
    return(out)
  }

  abort("Internal error: `precision` should be 'year' or 'month' at this point.")
}

calendar_count_between_proxy_compare.clock_year_month_day <- function(start,
                                                                      end,
                                                                      precision) {
  precision <- validate_precision_string(precision)

  start <- vec_proxy_compare(start)
  end <- vec_proxy_compare(end)

  if (precision >= PRECISION_YEAR) {
    start$year <- NULL
    end$year <- NULL
  }
  if (precision >= PRECISION_MONTH) {
    start$month <- NULL
    end$month <- NULL
  }

  list(start = start, end = end)
}

# ------------------------------------------------------------------------------

#' Sequences: year-month-day
#'
#' @description
#' This is a year-month-day method for the [seq()] generic.
#'
#' Sequences can only be generated for `"year"` and `"month"` precision
#' year-month-day vectors.
#'
#' When calling `seq()`, exactly two of the following must be specified:
#' - `to`
#' - `by`
#' - Either `length.out` or `along.with`
#'
#' @inheritParams seq.clock_duration
#'
#' @param from `[clock_year_month_day(1)]`
#'
#'   A `"year"` or `"month"` precision year-month-day to start the sequence
#'   from.
#'
#'   `from` is always included in the result.
#'
#' @param to `[clock_year_month_day(1) / NULL]`
#'
#'   A `"year"` or `"month"` precision year-month-day to stop the sequence
#'   at.
#'
#'   `to` is cast to the type of `from`.
#'
#'   `to` is only included in the result if the resulting sequence divides
#'   the distance between `from` and `to` exactly.
#'
#' @return A sequence with the type of `from`.
#'
#' @export
#' @examples
#' # Monthly sequence
#' x <- seq(year_month_day(2019, 1), year_month_day(2020, 12), by = 1)
#' x
#'
#' # Which we can then set the day of to get a sequence of end-of-month values
#' set_day(x, "last")
#'
#' # Daily sequences are not allowed. Use a naive-time for this instead.
#' try(seq(year_month_day(2019, 1, 1), by = 2, length.out = 2))
#' seq(as_naive_time(year_month_day(2019, 1, 1)), by = 2, length.out = 2)
seq.clock_year_month_day <- function(from,
                                     to = NULL,
                                     by = NULL,
                                     length.out = NULL,
                                     along.with = NULL,
                                     ...) {
  precision <- calendar_precision_attribute(from)

  if (precision > PRECISION_MONTH) {
    abort("`from` must be 'year' or 'month' precision.")
  }

  seq_impl(
    from = from,
    to = to,
    by = by,
    length.out = length.out,
    along.with = along.with,
    precision = precision,
    ...
  )
}

# ------------------------------------------------------------------------------

clock_init_year_month_day_utils <- function(env) {
  year <- year_month_day(integer())

  assign("clock_empty_year_month_day_year", year, envir = env)
  assign("clock_empty_year_month_day_month", calendar_widen(year, "month"), envir = env)
  assign("clock_empty_year_month_day_day", calendar_widen(year, "day"), envir = env)
  assign("clock_empty_year_month_day_hour", calendar_widen(year, "hour"), envir = env)
  assign("clock_empty_year_month_day_minute", calendar_widen(year, "minute"), envir = env)
  assign("clock_empty_year_month_day_second", calendar_widen(year, "second"), envir = env)
  assign("clock_empty_year_month_day_millisecond", calendar_widen(year, "millisecond"), envir = env)
  assign("clock_empty_year_month_day_microsecond", calendar_widen(year, "microsecond"), envir = env)
  assign("clock_empty_year_month_day_nanosecond", calendar_widen(year, "nanosecond"), envir = env)

  invisible(NULL)
}

# ------------------------------------------------------------------------------
# invalid_remove()

test_that("can remove invalid dates from calendar types", {
  x <- year_month_day(2020, 2, 28:30)
  expect_identical(invalid_remove(x), x[1:2])

  x <- year_day(2019, 365:366)
  expect_identical(invalid_remove(x), x[1L])
})

test_that("errors on unsupported types", {
  expect_snapshot_error(invalid_remove(1))
})

#' Create or retrieve date related labels
#'
#' @description
#' When parsing and formatting dates, you often need to know how weekdays of
#' the week and months are represented as text. These functions allow you
#' to either create your own labels, or look them up from a standard set of
#' language specific labels. The standard list is derived from ICU
#' (<https://unicode-org.github.io/icu/>) via the stringi package.
#'
#' - `clock_labels_lookup()` looks up a set of labels from a
#'   given language code.
#'
#' - `clock_labels_languages()` lists the language codes that are accepted.
#'
#' - `clock_labels()` lets you create your own set of labels. Use this if the
#'   currently supported languages don't meet your needs.
#'
#' @param month,month_abbrev `[character(12)]`
#'
#'   Full and abbreviated month names. Starts with January.
#'
#' @param weekday,weekday_abbrev `[character(7)]`
#'
#'   Full and abbreviated weekday names. Starts with Sunday.
#'
#' @param am_pm `[character(2)]`
#'
#'   Names used for AM and PM.
#'
#' @return A `"clock_labels"` object.
#'
#' @export
#' @examples
#' clock_labels_lookup("en")
#' clock_labels_lookup("ko")
#' clock_labels_lookup("fr")
clock_labels <- function(month,
                         month_abbrev = month,
                         weekday,
                         weekday_abbrev = weekday,
                         am_pm) {
  if (!is_character(month, n = 12L)) {
    abort("`month` must be a character vector of length 12.")
  }
  if (!is_character(month_abbrev, n = 12L)) {
    abort("`month_abbrev` must be a character vector of length 12.")
  }
  if (!is_character(weekday, n = 7L)) {
    abort("`weekday` must be a character vector of length 7.")
  }
  if (!is_character(weekday_abbrev, n = 7L)) {
    abort("`weekday_abbrev` must be a character vector of length 7.")
  }
  if (!is_character(am_pm, n = 2L)) {
    abort("`am_pm` must be a character vector of length 2.")
  }

  structure(
    list(
      month = enc2utf8(month),
      month_abbrev = enc2utf8(month_abbrev),
      weekday = enc2utf8(weekday),
      weekday_abbrev = enc2utf8(weekday_abbrev),
      am_pm = enc2utf8(am_pm)
    ),
    class = "clock_labels"
  )
}

#' @rdname clock_labels
#' @param language `[character(1)]`
#'
#'   A BCP 47 locale, generally constructed from a two or three
#'   digit language code. See `clock_labels_languages()` for a complete list of
#'   available locales.
#' @export
clock_labels_lookup <- function(language) {
  if (!is_character(language, n = 1L)) {
    abort("`language` must be a character vector of length 1.")
  }

  labels <- clock_labels_list[[language]]

  if (is.null(labels)) {
    abort(paste0("Unknown language '", language, "'."))
  }

  labels
}

#' @export
#' @rdname clock_labels
clock_labels_languages <- function() {
  names(clock_labels_list)
}

#' @export
print.clock_labels <- function(x, ...) {
  cat("<clock_labels>\n")

  if (identical(x$weekday, x$weekday_abbrev)) {
    weekday <- paste0(x$weekday, collapse = ", ")
  } else {
    weekday <- paste0(x$weekday, " (", x$weekday_abbrev, ")", collapse = ", ")
  }

  if (identical(x$month, x$month_abbrev)) {
    month <- paste0(x$month, collapse = ", ")
  } else {
    month <- paste0(x$month, " (", x$month_abbrev, ")", collapse = ", ")
  }

  am_pm <- paste0(x$am_pm, collapse = "/")

  cat_wrap("Weekdays: ", weekday)
  cat_wrap("Months:   ", month)
  cat_wrap("AM/PM:    ", am_pm)
}

is_clock_labels <- function(x) {
  inherits(x, "clock_labels")
}

cat_wrap <- function(header, body) {
  body <- strwrap(body, exdent = nchar(header))
  cat(header, paste(body, collapse = "\n"), "\n", sep = "")
}

# ------------------------------------------------------------------------------
# clock_labels()

test_that("can make custom labels", {
  months <- month.name
  weekdays <- c("Su", "Mo", "Tu", "We", "Th", "Fr", "Sa")
  am_pm <- c("A", "P")

  labels <- clock_labels(month = month.name, weekday = weekdays, am_pm = am_pm)

  expect_s3_class(labels, "clock_labels")
  expect_snapshot(labels)
})

test_that("input is validated", {
  months <- month.name
  weekdays <- c("Su", "Mo", "Tu", "We", "Th", "Fr", "Sa")
  am_pm <- c("A", "P")

  expect_snapshot_error(clock_labels(1))
  expect_snapshot_error(clock_labels("x"))
  expect_snapshot_error(clock_labels(months, 1))
  expect_snapshot_error(clock_labels(months, "x"))
  expect_snapshot_error(clock_labels(months, months, 1))
  expect_snapshot_error(clock_labels(months, months, "x"))
  expect_snapshot_error(clock_labels(months, months, weekdays, 1))
  expect_snapshot_error(clock_labels(months, months, weekdays, "x"))
  expect_snapshot_error(clock_labels(months, months, weekdays, weekdays, 1))
  expect_snapshot_error(clock_labels(months, months, weekdays, weekdays, "x"))
})

test_that("custom labels are converted to UTF-8 upon entry", {
  labels <- clock_labels_lookup("fr")
  month <- iconv(labels$month, from = "UTF-8", to = "latin1")

  labels <- clock_labels(month, month, labels$weekday, labels$weekday, labels$am_pm)

  # French February can be marked as latin1
  before <- month[2]
  after <- labels$month[2]

  expect_identical(Encoding(before), "latin1")
  expect_identical(Encoding(after), "UTF-8")
})

# ------------------------------------------------------------------------------
# clock_labels_lookup()

test_that("can lookup a language", {
  labels <- clock_labels_lookup("fr")

  expect_s3_class(labels, "clock_labels")
  expect_snapshot(labels)
})

test_that("must be a valid language code", {
  expect_snapshot_error(clock_labels_lookup(1))
  expect_snapshot_error(clock_labels_lookup("foo"))
})

# ------------------------------------------------------------------------------
# clock_labels_languages()

test_that("can list all the languages", {
  langs <- clock_labels_languages()

  expect_length(langs, 196)
  expect_snapshot(langs)
})

# ------------------------------------------------------------------------------

#' @export
print.clock_calendar <- function(x, ..., max = NULL) {
  clock_print(x, max)
}

# - Each subclass implements a `format()` method
# - Unlike vctrs, don't use `print(quote = FALSE)` since we want to match base R
#' @export
obj_print_data.clock_calendar <- function(x, ..., max) {
  if (vec_is_empty(x)) {
    return(invisible(x))
  }

  x <- max_slice(x, max)

  out <- format(x)

  # Pass `max` to avoid base R's default footer
  print(out, max = max)

  invisible(x)
}

#' @export
obj_print_footer.clock_calendar <- function(x, ..., max) {
  clock_print_footer(x, max)
}

# Align left to match pillar_shaft.Date
# @export - lazy in .onLoad()
pillar_shaft.clock_calendar <- function(x, ...) {
  out <- format(x)
  pillar::new_pillar_shaft_simple(out, align = "left")
}

# ------------------------------------------------------------------------------

# Note: Cannot cast between calendar precisions. Casting to a more precise
# precision is undefined because we consider things like year-month to be
# a range of days over the whole month, and it would be impossible to map
# that to just one day.

ptype2_calendar_and_calendar <- function(x, y, ...) {
  if (calendar_precision_attribute(x) == calendar_precision_attribute(y)) {
    x
  } else {
    stop_incompatible_type(
      x,
      y,
      ...,
      details = "Can't combine calendars with different precisions."
    )
  }
}

cast_calendar_to_calendar <- function(x, to, ...) {
  if (calendar_precision_attribute(x) == calendar_precision_attribute(to)) {
    x
  } else {
    stop_incompatible_cast(
      x,
      to,
      ...,
      details = "Can't cast between calendars with different precisions."
    )
  }
}

# ------------------------------------------------------------------------------

#' Is the calendar year a leap year?
#'
#' @description
#' `calendar_leap_year()` detects if the calendar year is a leap year - i.e.
#' does it contain one or more extra components than other years?
#'
#' A particular year is a leap year if:
#'
#' - [year_month_day()]: February has 29 days.
#'
#' - [year_month_weekday()]: February has a weekday that occurs 5 times.
#'
#' - [year_week_day()]: There are 53 weeks in the year, resulting in 371
#'   days in the year.
#'
#' - [iso_year_week_day()]: There are 53 weeks in the year, resulting in 371
#'   days in the year.
#'
#' - [year_quarter_day()]: One of the quarters has 1 more day than normal (the
#'   quarter with an extra day depends on the `start` used, but will always be
#'   the same for a particular `start`). This aligns with Gregorian leap years
#'   for all `start`s except February, in which case the leap year is always 1
#'   year after the Gregorian leap year.
#'
#' - [year_day()]: There are 366 days in the year.
#'
#' @param x `[calendar]`
#'
#'   A calendar type to detect leap years in.
#'
#' @return A logical vector the same size as `x`. Returns `TRUE` if in a leap
#'   year, `FALSE` if not in a leap year, and `NA` if `x` is `NA`.
#'
#' @export
#' @examples
#' x <- year_month_day(c(2019:2024, NA))
#' calendar_leap_year(x)
#'
#' # For year-quarter-day, the leap year typically aligns with the Gregorian
#' # leap year, unless the `start` is February, in which case the leap year is
#' # always 1 year after the Gregorian leap year
#' x <- year_quarter_day(2020:2021, start = clock_months$january)
#' calendar_leap_year(x)
#'
#' x <- year_quarter_day(2020:2021, start = clock_months$february)
#' calendar_leap_year(x)
#'
#' # With a January start, 2020 has the extra day
#' get_day(year_quarter_day(2020, 1:4, "last", start = clock_months$january))
#' get_day(year_quarter_day(2021, 1:4, "last", start = clock_months$january))
#' get_day(year_quarter_day(2022, 1:4, "last", start = clock_months$january))
#'
#' # With a February start, 2021 has the extra day
#' get_day(year_quarter_day(2020, 1:4, "last", start = clock_months$february))
#' get_day(year_quarter_day(2021, 1:4, "last", start = clock_months$february))
#' get_day(year_quarter_day(2022, 1:4, "last", start = clock_months$february))
calendar_leap_year <- function(x) {
  check_calendar(x)
  UseMethod("calendar_leap_year")
}

#' @export
calendar_leap_year.clock_calendar <- function(x) {
  stop_clock_unsupported(x)
}

# ------------------------------------------------------------------------------

#' Convert a calendar to an ordered factor of month names
#'
#' @description
#' `calendar_month_factor()` extracts the month values from a calendar and
#' converts them to an ordered factor of month names. This can be useful in
#' combination with ggplot2, or for modeling.
#'
#' This function is only relevant for calendar types that use a month field,
#' i.e. [year_month_day()] and [year_month_weekday()]. The calendar type must
#' have at least month precision.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams clock_locale
#'
#' @param x `[calendar]`
#'
#'   A calendar vector.
#'
#' @param abbreviate `[logical(1)]`
#'
#'   If `TRUE`, the abbreviated month names from `labels` will be used.
#'
#'   If `FALSE`, the full month names from `labels` will be used.
#'
#' @return An ordered factor representing the months.
#'
#' @export
#' @examples
#' x <- year_month_day(2019, 1:12)
#'
#' calendar_month_factor(x)
#' calendar_month_factor(x, abbreviate = TRUE)
#' calendar_month_factor(x, labels = "fr")
calendar_month_factor <- function(x, ..., labels = "en", abbreviate = FALSE) {
  check_calendar(x)
  UseMethod("calendar_month_factor")
}

#' @export
calendar_month_factor.clock_calendar <- function(
  x,
  ...,
  labels = "en",
  abbreviate = FALSE
) {
  stop_clock_unsupported(x)
}

calendar_month_factor_impl <- function(
  x,
  labels,
  abbreviate,
  ...,
  error_call = caller_env()
) {
  check_dots_empty0(...)

  if (calendar_precision_attribute(x) < PRECISION_MONTH) {
    cli::cli_abort(
      "{.arg x} must have at least {.str month} precision.",
      call = error_call
    )
  }

  if (is_character(labels)) {
    labels <- clock_labels_lookup(labels)
  }
  check_clock_labels(labels, call = error_call)

  check_bool(abbreviate, call = error_call)

  if (abbreviate) {
    labels <- labels$month_abbrev
  } else {
    labels <- labels$month
  }

  x <- get_month(x)
  x <- labels[x]

  factor(x, levels = labels, ordered = TRUE)
}

# ------------------------------------------------------------------------------

#' Group calendar components
#'
#' @description
#' `calendar_group()` groups at a multiple of the specified precision. Grouping
#' alters the value of a single component (i.e. the month component
#' if grouping by month). Components that are more precise than the precision
#' being grouped at are dropped altogether (i.e. the day component is dropped
#' if grouping by month).
#'
#' Each calendar has its own help page describing the grouping process in more
#' detail:
#'
#' - [year-month-day][year-month-day-group]
#'
#' - [year-month-weekday][year-month-weekday-group]
#'
#' - [year-week-day][year-week-day-group]
#'
#' - [iso-year-week-day][iso-year-week-day-group]
#'
#' - [year-quarter-day][year-quarter-day-group]
#'
#' - [year-day][year-day-group]
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[calendar]`
#'
#'   A calendar vector.
#'
#' @param precision `[character(1)]`
#'
#'   A precision. Allowed precisions are dependent on the calendar used.
#'
#' @param n `[positive integer(1)]`
#'
#'   A single positive integer specifying a multiple of `precision` to use.
#'
#' @return `x` grouped at the specified `precision`.
#'
#' @export
#' @examples
#' # See the calendar specific help pages for more examples
#' x <- year_month_day(2019, c(1, 1, 2, 2, 3, 3, 4, 4), 1:8)
#' x
#'
#' # Group by two months
#' calendar_group(x, "month", n = 2)
#'
#' # Group by two days of the month
#' calendar_group(x, "day", n = 2)
calendar_group <- function(x, precision, ..., n = 1L) {
  check_dots_empty0(...)

  check_calendar(x)

  check_precision(precision)
  precision <- precision_to_integer(precision)
  calendar_check_precision(x, precision)

  x_precision <- calendar_precision_attribute(x)

  if (precision > x_precision) {
    precision <- precision_to_string(precision)
    x_precision <- precision_to_string(x_precision)

    message <- paste0(
      "Can't group at a precision ({.str {precision}}) ",
      "that is more precise than `x` ({.str {x_precision}})."
    )
    cli::cli_abort(message)
  }

  if (precision > PRECISION_SECOND && x_precision != precision) {
    # Grouping nanosecond precision by millisecond would be inconsistent
    # with our general philosophy that you "lock in" the subsecond precision.
    precision <- precision_to_string(precision)
    x_precision <- precision_to_string(x_precision)

    message <- paste0(
      "Can't group a subsecond precision `x` ({.str {x_precision}}) ",
      "by another subsecond precision ({.str {precision}})."
    )
    cli::cli_abort(message)
  }

  UseMethod("calendar_group")
}

#' @export
calendar_group.clock_calendar <- function(x, precision, ..., n = 1L) {
  stop_clock_unsupported(x)
}

calendar_group_time <- function(x, n, precision) {
  if (precision == PRECISION_HOUR) {
    value <- get_hour(x)
    value <- group_component0(value, n)
    x <- set_hour(x, value)
    return(x)
  }
  if (precision == PRECISION_MINUTE) {
    value <- get_minute(x)
    value <- group_component0(value, n)
    x <- set_minute(x, value)
    return(x)
  }
  if (precision == PRECISION_SECOND) {
    value <- get_second(x)
    value <- group_component0(value, n)
    x <- set_second(x, value)
    return(x)
  }

  # Generic ensures that if `x_precision > PRECISION_SECOND`,
  # then `precision == x_precision`, making this safe.
  if (precision == PRECISION_MILLISECOND) {
    value <- get_millisecond(x)
    value <- group_component0(value, n)
    x <- set_millisecond(x, value)
    return(x)
  }
  if (precision == PRECISION_MICROSECOND) {
    value <- get_microsecond(x)
    value <- group_component0(value, n)
    x <- set_microsecond(x, value)
    return(x)
  }
  if (precision == PRECISION_NANOSECOND) {
    value <- get_nanosecond(x)
    value <- group_component0(value, n)
    x <- set_nanosecond(x, value)
    return(x)
  }

  abort("Unknown precision.", .internal = TRUE)
}

group_component0 <- function(x, n) {
  (x %/% n) * n
}
group_component1 <- function(x, n) {
  ((x - 1L) %/% n) * n + 1L
}

validate_calendar_group_n <- function(n, ..., error_call = caller_env()) {
  check_number_whole(n, min = 0, call = error_call)
  n <- vec_cast(n, integer(), call = error_call)
  n
}

# ------------------------------------------------------------------------------

#' Narrow a calendar to a less precise precision
#'
#' @description
#' `calendar_narrow()` narrows `x` to the specified `precision`. It does so
#' by dropping components that represent a precision that is finer than
#' `precision`.
#'
#' Each calendar has its own help page describing the precisions that you
#' can narrow to:
#'
#' - [year-month-day][year-month-day-narrow]
#'
#' - [year-month-weekday][year-month-weekday-narrow]
#'
#' - [year-week-day][year-week-day-narrow]
#'
#' - [iso-year-week-day][iso-year-week-day-narrow]
#'
#' - [year-quarter-day][year-quarter-day-narrow]
#'
#' - [year-day][year-day-narrow]
#'
#' @details
#' A subsecond precision `x` cannot be narrowed to another subsecond precision.
#' You cannot narrow from, say, `"nanosecond"` to `"millisecond"` precision.
#' clock operates under the philosophy that once you have set the subsecond
#' precision of a calendar, it is "locked in" at that precision. If you
#' expected this to use integer division to divide the nanoseconds by 1e6 to
#' get to millisecond precision, you probably want to convert to a time point
#' first, and use [time_point_floor()].
#'
#' @inheritParams calendar_group
#'
#' @return `x` narrowed to the supplied `precision`.
#'
#' @export
#' @examples
#' # Hour precision
#' x <- year_month_day(2019, 1, 3, 4)
#' x
#'
#' # Narrowed to day precision
#' calendar_narrow(x, "day")
#'
#' # Or month precision
#' calendar_narrow(x, "month")
calendar_narrow <- function(x, precision) {
  check_calendar(x)

  check_precision(precision)
  precision <- precision_to_integer(precision)
  calendar_check_precision(x, precision)

  x_precision <- calendar_precision_attribute(x)

  if (precision > x_precision) {
    precision <- precision_to_string(precision)
    x_precision <- precision_to_string(x_precision)

    message <- paste0(
      "Can't narrow to a precision ({.str {precision}}) ",
      "that is wider than `x` ({.str {x_precision}})."
    )
    cli::cli_abort(message)
  }

  if (precision > PRECISION_SECOND && x_precision != precision) {
    # Allowing Nanosecond -> Millisecond wouldn't be consistent with us
    # disallowing `set_millisecond(<calendar<nanosecond>>)`, and is ambiguous.
    precision <- precision_to_string(precision)
    x_precision <- precision_to_string(x_precision)

    message <- paste0(
      "Can't narrow a subsecond precision `x` ({.str {x_precision}}) ",
      "to another subsecond precision ({.str {precision}})."
    )
    cli::cli_abort(message)
  }

  UseMethod("calendar_narrow")
}

#' @export
calendar_narrow.clock_calendar <- function(x, precision) {
  stop_clock_unsupported(x)
}

calendar_narrow_time <- function(out_fields, out_precision, x_fields) {
  if (out_precision >= PRECISION_HOUR) {
    out_fields[["hour"]] <- x_fields[["hour"]]
  }
  if (out_precision >= PRECISION_MINUTE) {
    out_fields[["minute"]] <- x_fields[["minute"]]
  }
  if (out_precision >= PRECISION_SECOND) {
    out_fields[["second"]] <- x_fields[["second"]]
  }
  if (out_precision > PRECISION_SECOND) {
    out_fields[["subsecond"]] <- x_fields[["subsecond"]]
  }

  out_fields
}

# ------------------------------------------------------------------------------

#' Widen a calendar to a more precise precision
#'
#' @description
#' `calendar_widen()` widens `x` to the specified `precision`. It does so
#' by setting new components to their smallest value.
#'
#' Each calendar has its own help page describing the precisions that you
#' can widen to:
#'
#' - [year-month-day][year-month-day-widen]
#'
#' - [year-month-weekday][year-month-weekday-widen]
#'
#' - [year-week-day][year-week-day-widen]
#'
#' - [iso-year-week-day][iso-year-week-day-widen]
#'
#' - [year-quarter-day][year-quarter-day-widen]
#'
#' - [year-day][year-day-widen]
#'
#' @details
#' A subsecond precision `x` cannot be widened. You cannot widen from, say,
#' `"millisecond"` to `"nanosecond"` precision. clock operates under the
#' philosophy that once you have set the subsecond precision of a calendar,
#' it is "locked in" at that precision. If you expected this to multiply
#' the milliseconds by 1e6 to get to nanosecond precision, you probably
#' want to convert to a time point first, and use [time_point_cast()].
#'
#' Generally, clock treats calendars at a specific precision as a _range_ of
#' values. For example, a month precision year-month-day is treated as a range
#' over `[yyyy-mm-01, yyyy-mm-last]`, with no assumption about the day of the
#' month. However, occasionally it is useful to quickly widen a calendar,
#' assuming that you want the beginning of this range to be used for each
#' component. This is where `calendar_widen()` can come in handy.
#'
#' @inheritParams calendar_group
#'
#' @return `x` widened to the supplied `precision`.
#'
#' @export
#' @examples
#' # Month precision
#' x <- year_month_day(2019, 1)
#' x
#'
#' # Widen to day precision
#' calendar_widen(x, "day")
#'
#' # Or second precision
#' calendar_widen(x, "second")
calendar_widen <- function(x, precision) {
  check_calendar(x)

  check_precision(precision)
  precision <- precision_to_integer(precision)
  calendar_check_precision(x, precision)

  x_precision <- calendar_precision_attribute(x)

  if (x_precision > precision) {
    precision <- precision_to_string(precision)
    x_precision <- precision_to_string(x_precision)

    message <- paste0(
      "Can't widen to a precision ({.str {precision}}) ",
      "that is narrower than `x` ({.str {x_precision}})."
    )
    cli::cli_abort(message)
  }

  if (x_precision > PRECISION_SECOND && x_precision != precision) {
    # Allowing Millisecond -> Nanosecond wouldn't be consistent with us
    # disallowing `set_nanosecond(<calendar<millisecond>>)`, and is ambiguous.
    precision <- precision_to_string(precision)
    x_precision <- precision_to_string(x_precision)

    message <- paste0(
      "Can't widen a subsecond precision `x` ({.str {x_precision}}) ",
      "to another subsecond precision ({.str {precision}})."
    )
    cli::cli_abort(message)
  }

  UseMethod("calendar_widen")
}

#' @export
calendar_widen.clock_calendar <- function(x, precision) {
  stop_clock_unsupported(x)
}

calendar_widen_time <- function(x, x_precision, precision) {
  if (precision >= PRECISION_HOUR && x_precision < PRECISION_HOUR) {
    x <- set_hour(x, 0L)
  }
  if (precision >= PRECISION_MINUTE && x_precision < PRECISION_MINUTE) {
    x <- set_minute(x, 0L)
  }
  if (precision >= PRECISION_SECOND && x_precision < PRECISION_SECOND) {
    x <- set_second(x, 0L)
  }

  # `x` is required to fulfill:
  # `x_precision < PRECISION_SECOND` or `x_precision == precision`
  if (precision == PRECISION_MILLISECOND && x_precision != precision) {
    x <- set_millisecond(x, 0L)
  }
  if (precision == PRECISION_MICROSECOND && x_precision != precision) {
    x <- set_microsecond(x, 0L)
  }
  if (precision == PRECISION_NANOSECOND && x_precision != precision) {
    x <- set_nanosecond(x, 0L)
  }

  x
}

# ------------------------------------------------------------------------------

#' Boundaries: calendars
#'
#' @description
#' - `calendar_start()` computes the start of a calendar at a particular
#'   `precision`, such as the "start of the quarter".
#'
#' - `calendar_end()` computes the end of a calendar at a particular
#'   `precision`, such as the "end of the month".
#'
#' For both `calendar_start()` and `calendar_end()`, the precision of `x` is
#' always retained.
#'
#' Each calendar has its own help page describing the precisions that you
#' can compute a boundary at:
#'
#' - [year-month-day][year-month-day-boundary]
#'
#' - [year-month-weekday][year-month-weekday-boundary]
#'
#' - [year-week-day][year-week-day-boundary]
#'
#' - [iso-year-week-day][iso-year-week-day-boundary]
#'
#' - [year-quarter-day][year-quarter-day-boundary]
#'
#' - [year-day][year-day-boundary]
#'
#' @inheritParams calendar_group
#'
#' @return `x` at the same precision, but with some components altered to be
#'   at the boundary value.
#'
#' @name calendar-boundary
#' @examples
#' # Hour precision
#' x <- year_month_day(2019, 2:4, 5, 6)
#' x
#'
#' # Compute the start of the month
#' calendar_start(x, "month")
#'
#' # Or the end of the month, notice that the hour value is adjusted as well
#' calendar_end(x, "month")
NULL

#' @rdname calendar-boundary
#' @export
calendar_start <- function(x, precision) {
  check_calendar(x)
  UseMethod("calendar_start")
}

#' @export
calendar_start.clock_calendar <- function(x, precision) {
  stop_clock_unsupported(x)
}

#' @rdname calendar-boundary
#' @export
calendar_end <- function(x, precision) {
  check_calendar(x)
  UseMethod("calendar_end")
}

#' @export
calendar_end.clock_calendar <- function(x, precision) {
  stop_clock_unsupported(x)
}

calendar_start_end_checks <- function(x, x_precision, precision, which) {
  calendar_check_precision(x, precision)

  if (x_precision < precision) {
    precision <- precision_to_string(precision)
    x_precision <- precision_to_string(x_precision)

    message <- paste0(
      "Can't compute the {which} of `x` ({.str {x_precision}}) ",
      "at a more precise precision ({.str {precision}})."
    )
    cli::cli_abort(message)
  }

  if (precision > PRECISION_SECOND && x_precision != precision) {
    # Computing the start/end of nanosecond precision at millisecond precision
    # would be inconsistent with our general philosophy that you "lock in"
    # the subsecond precision.
    precision <- precision_to_string(precision)
    x_precision <- precision_to_string(x_precision)

    message <- paste0(
      "Can't compute the {which} of a subsecond precision `x` ({.str {x_precision}}) ",
      "at another subsecond precision ({.str {precision}})."
    )
    cli::cli_abort(message)
  }

  invisible(x)
}

calendar_start_time <- function(x, x_precision, precision) {
  values <- list(
    hour = 0L,
    minute = 0L,
    second = 0L,
    millisecond = 0L,
    microsecond = 0L,
    nanosecond = 0L
  )

  calendar_start_end_time(x, x_precision, precision, values)
}

calendar_end_time <- function(x, x_precision, precision) {
  values <- list(
    hour = 23L,
    minute = 59L,
    second = 59L,
    millisecond = 999L,
    microsecond = 999999L,
    nanosecond = 999999999L
  )

  calendar_start_end_time(x, x_precision, precision, values)
}

calendar_start_end_time <- function(x, x_precision, precision, values) {
  if (precision <= PRECISION_DAY && x_precision > PRECISION_DAY) {
    x <- set_hour(x, values$hour)
  }
  if (precision <= PRECISION_HOUR && x_precision > PRECISION_HOUR) {
    x <- set_minute(x, values$minute)
  }
  if (precision <= PRECISION_MINUTE && x_precision > PRECISION_MINUTE) {
    x <- set_second(x, values$second)
  }
  if (precision <= PRECISION_SECOND && x_precision > PRECISION_SECOND) {
    if (x_precision == PRECISION_MILLISECOND) {
      x <- set_millisecond(x, values$millisecond)
    } else if (x_precision == PRECISION_MICROSECOND) {
      x <- set_microsecond(x, values$microsecond)
    } else if (x_precision == PRECISION_NANOSECOND) {
      x <- set_nanosecond(x, values$nanosecond)
    }
  }

  x
}

# ------------------------------------------------------------------------------

#' Counting: calendars
#'
#' @description
#' `calendar_count_between()` counts the number of `precision` units between
#' `start` and `end` (i.e., the number of years or months). This count
#' corresponds to the _whole number_ of units, and will never return a
#' fractional value.
#'
#' This is suitable for, say, computing the whole number of years or months
#' between two calendar dates, accounting for the day and time of day.
#'
#' Each calendar has its own help page describing the precisions that you can
#' count at:
#'
#' - [year-month-day][year-month-day-count-between]
#'
#' - [year-month-weekday][year-month-weekday-count-between]
#'
#' - [year-week-day][year-week-day-count-between]
#'
#' - [iso-year-week-day][iso-year-week-day-count-between]
#'
#' - [year-quarter-day][year-quarter-day-count-between]
#'
#' - [year-day][year-day-count-between]
#'
#' @section Comparison Direction:
#' The computed count has the property that if `start <= end`, then
#' `start + <count> <= end`. Similarly, if `start >= end`, then
#' `start + <count> >= end`. In other words, the comparison direction between
#' `start` and `end` will never change after adding the count to `start`. This
#' makes this function useful for repeated count computations at
#' increasingly fine precisions.
#'
#' @inheritParams calendar_group
#'
#' @param start,end `[clock_calendar]`
#'
#'   A pair of calendar vectors. These will be recycled to their common size.
#'
#' @return An integer representing the number of `precision` units between
#' `start` and `end`.
#'
#' @name calendar-count-between
#' @examples
#' # Number of whole years between these dates
#' x <- year_month_day(2000, 01, 05)
#' y <- year_month_day(2005, 01, 04:06)
#'
#' # Note that `2000-01-05 -> 2005-01-04` is only 4 full years
#' calendar_count_between(x, y, "year")
NULL

#' @rdname calendar-count-between
#' @export
calendar_count_between <- function(start, end, precision, ..., n = 1L) {
  check_calendar(start)
  UseMethod("calendar_count_between")
}

#' @export
calendar_count_between.clock_calendar <- function(
  start,
  end,
  precision,
  ...,
  n = 1L
) {
  check_dots_empty0(...)
  check_calendar(end)

  size <- vec_size_common(start = start, end = end)

  args <- vec_cast_common(start = start, end = end)
  args <- vec_recycle_common(!!!args, .size = size)
  start <- args[[1]]
  end <- args[[2]]

  check_number_whole(n, min = 0)
  n <- vec_cast(n, integer())

  check_precision(precision)
  precision_int <- precision_to_integer(precision)

  if (calendar_precision_attribute(start) < precision_int) {
    start_precision <- precision_to_string(calendar_precision_attribute(start))

    message <- paste0(
      "Precision of inputs ({.str {start_precision}}) must be at least as precise ",
      "as {.arg precision} ({.str {precision}})."
    )

    cli::cli_abort(message)
  }

  args <- calendar_count_between_standardize_precision_n(start, precision, n)
  precision <- args$precision
  n <- args$n

  # Core computation to get the difference (pre-adjustment).
  # Result is an integer because it represents a count of duration units.
  out <- calendar_count_between_compute(start, end, precision)

  # Comparison proxy, truncated to avoid fields already when computing `out`
  args <- calendar_count_between_proxy_compare(start, end, precision)
  start_proxy <- args$start
  end_proxy <- args$end

  if (ncol(start_proxy) == 0L) {
    # vctrs bug with vec_compare()?
    # https://github.com/r-lib/vctrs/issues/1500
    comparison <- vec_rep(0L, size)
  } else {
    comparison <- vec_compare(end_proxy, start_proxy)
  }

  # - When `start > end` and the non-year portion of `start < end`, add 1
  # - When `start < end` and the non-year portion of `start > end`, subtract 1
  adjustment <- vec_rep(-1L, size)
  adjustment[start > end] <- 1L
  adjustment[comparison != adjustment] <- 0L

  out <- out + adjustment

  if (n != 1L) {
    out <- out %/% n
  }

  out
}

# Internal generic
calendar_count_between_standardize_precision_n <- function(x, precision, n) {
  UseMethod("calendar_count_between_standardize_precision_n")
}

# Internal generic
calendar_count_between_compute <- function(start, end, precision) {
  UseMethod("calendar_count_between_compute")
}

# Internal generic
calendar_count_between_proxy_compare <- function(start, end, precision) {
  UseMethod("calendar_count_between_proxy_compare")
}

# ------------------------------------------------------------------------------

#' Spanning sequence: calendars
#'
#' @description
#' `calendar_spanning_seq()` generates a regular sequence along the span of
#' `x`, i.e. along `[min(x), max(x)]`. The sequence is generated at the
#' precision of `x`.
#'
#' Importantly, sequences can only be generated if the underlying [seq()] method
#' for the calendar in question supports a `from` and `to` value at the same
#' precision as `x`. For example, you can't compute a day precision spanning
#' sequence for a [year_month_day()] calendar (you can only compute a year
#' and month one). To create a day precision sequence, you'd have to convert to
#' a time-point first. See the individual [seq()] method documentation to learn
#' what precisions are allowed.
#'
#' @details
#' Missing values are automatically removed before the sequence is generated.
#'
#' If you need more precise sequence generation, call [range()] and [seq()]
#' directly.
#'
#' @param x `[clock_calendar]`
#'
#'   A calendar vector.
#'
#' @return A sequence along `[min(x), max(x)]`.
#'
#' @export
#' @examples
#' x <- year_month_day(c(2019, 2022, 2020), c(2, 5, 3))
#' x
#'
#' # Month precision spanning sequence
#' calendar_spanning_seq(x)
#'
#' # Quarter precision:
#' x <- year_quarter_day(c(2005, 2006, 2003), c(4, 2, 3))
#' calendar_spanning_seq(x)
#'
#' # Can't generate sequences if `seq()` doesn't allow the precision
#' x <- year_month_day(2019, c(1, 2, 1), c(20, 3, 25))
#' try(calendar_spanning_seq(x))
#'
#' # Generally this means you need to convert to a time point and use
#' # `time_point_spanning_seq()` instead
#' time_point_spanning_seq(as_sys_time(x))
calendar_spanning_seq <- function(x) {
  check_calendar(x)
  spanning_seq_impl(x)
}

# ------------------------------------------------------------------------------

#' Precision: calendar
#'
#' `calendar_precision()` extracts the precision from a calendar object. It
#' returns the precision as a single string.
#'
#' @param x `[clock_calendar]`
#'
#'   A calendar.
#'
#' @return A single string holding the precision of the calendar.
#'
#' @export
#' @examples
#' calendar_precision(year_month_day(2019))
#' calendar_precision(year_month_day(2019, 1, 1))
#' calendar_precision(year_quarter_day(2019, 3))
calendar_precision <- function(x) {
  check_calendar(x)
  UseMethod("calendar_precision")
}

#' @export
calendar_precision.clock_calendar <- function(x) {
  precision <- calendar_precision_attribute(x)
  precision <- precision_to_string(precision)
  precision
}

# ------------------------------------------------------------------------------

# Internal generic
calendar_name <- function(x) {
  UseMethod("calendar_name")
}

# ------------------------------------------------------------------------------

# Internal generic
calendar_is_precision <- function(x, precision) {
  UseMethod("calendar_is_precision")
}

calendar_check_precision <- function(
  x,
  precision,
  ...,
  error_call = caller_env()
) {
  if (calendar_is_precision(x, precision)) {
    return(invisible(NULL))
  }

  message <- paste0(
    "{.arg precision} must be a valid precision for a {.cls {calendar_name(x)}}, ",
    "not {.str {precision_to_string(precision)}}."
  )

  cli::cli_abort(message, call = error_call)
}

# ------------------------------------------------------------------------------

calendar_precision_attribute <- function(x) {
  attr(x, "precision", exact = TRUE)
}

# ------------------------------------------------------------------------------

calendar_check_minimum_precision <- function(
  x,
  precision,
  ...,
  arg = caller_arg(x),
  call = caller_env()
) {
  x_precision <- calendar_precision_attribute(x)

  if (x_precision >= precision) {
    return(invisible(NULL))
  }

  message <- c(
    "Can't perform this operation because of the precision of {.arg {arg}}.",
    i = "The precision of {.arg {arg}} must be at least {.str {precision_to_string(precision)}}.",
    i = "{.arg {arg}} has a precision of {.str {precision_to_string(x_precision)}}."
  )

  cli::cli_abort(message, call = call)
}

calendar_check_maximum_precision <- function(
  x,
  precision,
  ...,
  arg = caller_arg(x),
  call = caller_env()
) {
  x_precision <- calendar_precision_attribute(x)

  if (x_precision <= precision) {
    return(invisible(NULL))
  }

  message <- c(
    "Can't perform this operation because of the precision of {.arg {arg}}.",
    i = "The precision of {.arg {arg}} must be at most {.str {precision_to_string(precision)}}.",
    i = "{.arg {arg}} has a precision of {.str {precision_to_string(x_precision)}}."
  )

  cli::cli_abort(message, call = call)
}

calendar_check_exact_precision <- function(
  x,
  precision,
  ...,
  arg = caller_arg(x),
  call = caller_env()
) {
  x_precision <- calendar_precision_attribute(x)

  if (x_precision %in% precision) {
    return(invisible(NULL))
  }

  precision <- vapply(precision, precision_to_string, character(1))

  message <- c(
    "Can't perform this operation because of the precision of {.arg {arg}}.",
    i = "The precision of {.arg {arg}} must be {.or {.str {precision}}}.",
    i = "{.arg {arg}} has a precision of {.str {precision_to_string(x_precision)}}."
  )

  cli::cli_abort(message, call = call)
}

# For use in calendar constructor helpers
calendar_check_subsecond_precision <- function(
  subsecond_precision,
  ...,
  call = caller_env()
) {
  if (is_null(subsecond_precision)) {
    cli::cli_abort(
      "When {.arg subsecond} is provided, {.arg subsecond_precision} must also be specified.",
      call = call
    )
  }

  check_precision_subsecond(subsecond_precision, call = call)
}

# ------------------------------------------------------------------------------

calendar_check_no_invalid <- function(
  x,
  ...,
  arg = caller_arg(x),
  call = caller_env()
) {
  if (!invalid_any(x)) {
    return(invisible(NULL))
  }

  loc <- invalid_detect(x)
  loc <- which(loc)

  message <- c(
    "Can't convert {.arg {arg}} to another type because some dates are invalid.",
    i = "The following locations are invalid: {loc}.",
    i = "Resolve invalid dates with {.fn invalid_resolve}."
  )

  cli::cli_abort(message, call = call)
}

# ------------------------------------------------------------------------------

calendar_ptype_full <- function(x, class) {
  precision <- calendar_precision_attribute(x)
  precision <- precision_to_string(precision)
  paste0(class, "<", precision, ">")
}

calendar_ptype_abbr <- function(x, abbr) {
  precision <- calendar_precision_attribute(x)
  precision <- precision_to_string(precision)
  precision <- precision_abbr(precision)
  paste0(abbr, "<", precision, ">")
}

# ------------------------------------------------------------------------------

arith_calendar_and_missing <- function(op, x, y, ...) {
  switch(
    op,
    "+" = x,
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_calendar_and_calendar <- function(
  op,
  x,
  y,
  ...,
  calendar_minus_calendar_fn
) {
  switch(
    op,
    "-" = calendar_minus_calendar_fn(op, x, y, ...),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_calendar_and_duration <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(x, y),
    "-" = add_duration(x, -y),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_duration_and_calendar <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(y, x, swapped = TRUE),
    "-" = stop_incompatible_op(
      op,
      x,
      y,
      details = "Can't subtract a calendar from a duration.",
      ...
    ),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_calendar_and_numeric <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(
      x,
      duration_helper(y, calendar_precision_attribute(x), retain_names = TRUE)
    ),
    "-" = add_duration(
      x,
      duration_helper(-y, calendar_precision_attribute(x), retain_names = TRUE)
    ),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_numeric_and_calendar <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(
      y,
      duration_helper(x, calendar_precision_attribute(y), retain_names = TRUE),
      swapped = TRUE
    ),
    "-" = stop_incompatible_op(
      op,
      x,
      y,
      details = "Can't subtract a calendar from a duration.",
      ...
    ),
    stop_incompatible_op(op, x, y, ...)
  )
}

# ------------------------------------------------------------------------------

# Special support for when `lag * differences >= n`
#
# In vctrs, this forces a return value of `vec_slice(x, 0L)`, but this is
# not correct for calendar types. See also, `diff.Date()` which exists for a
# similar reason.
calendar_diff_is_empty <- function(x, lag, differences) {
  # Same errors as vctrs
  stopifnot(length(lag) == 1L, lag >= 1L)
  stopifnot(length(differences) == 1L, differences >= 1L)
  n <- vec_size(x)
  lag * differences >= n
}

# ------------------------------------------------------------------------------

#' @export
as_year_month_day.clock_calendar <- function(x, ...) {
  check_dots_empty0(...)
  as_year_month_day(as_sys_time(x))
}

#' @export
as_year_month_weekday.clock_calendar <- function(x, ...) {
  check_dots_empty0(...)
  as_year_month_weekday(as_sys_time(x))
}

#' @export
as_year_week_day.clock_calendar <- function(x, ..., start = NULL) {
  check_dots_empty0(...)
  as_year_week_day(as_sys_time(x), start = start)
}

#' @export
as_iso_year_week_day.clock_calendar <- function(x, ...) {
  check_dots_empty0(...)
  as_iso_year_week_day(as_sys_time(x))
}

#' @export
as_year_day.clock_calendar <- function(x, ...) {
  check_dots_empty0(...)
  as_year_day(as_sys_time(x))
}

#' @export
as_year_quarter_day.clock_calendar <- function(x, ..., start = NULL) {
  check_dots_empty0(...)
  as_year_quarter_day(as_sys_time(x), start = start)
}

# ------------------------------------------------------------------------------

#' @export
add_weeks.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported(x, details = advice_convert_to_time_point())
}

#' @export
add_days.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported(x, details = advice_convert_to_time_point())
}

#' @export
add_hours.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported(x, details = advice_convert_to_time_point())
}

#' @export
add_minutes.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported(x, details = advice_convert_to_time_point())
}

#' @export
add_seconds.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported(x, details = advice_convert_to_time_point())
}

#' @export
add_milliseconds.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported(x, details = advice_convert_to_time_point())
}

#' @export
add_microseconds.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported(x, details = advice_convert_to_time_point())
}

#' @export
add_nanoseconds.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported(x, details = advice_convert_to_time_point())
}

advice_convert_to_time_point <- function() {
  c(
    i = "Do you need to convert to a time point first?",
    i = cli::format_inline(
      "Use {.fn as_naive_time} or {.fn as_sys_time} to convert to a time point."
    )
  )
}

# ------------------------------------------------------------------------------

calendar_minimum <- function(precision, year) {
  out <- calendar_widen(year, precision)
  out <- calendar_start(out, "year")
  out
}

calendar_maximum <- function(precision, year) {
  out <- calendar_widen(year, precision)
  out <- calendar_end(out, "year")
  out
}

# ------------------------------------------------------------------------------

field_year <- function(x) {
  # The `year` field is the first field of every calendar type, which makes
  # it the field that names are on. When extracting the field, we don't ever
  # want the names to come with it.
  out <- field(x, "year")
  names(out) <- NULL
  out
}
field_quarter <- function(x) {
  field(x, "quarter")
}
field_month <- function(x) {
  field(x, "month")
}
field_week <- function(x) {
  field(x, "week")
}
field_day <- function(x) {
  field(x, "day")
}
field_hour <- function(x) {
  field(x, "hour")
}
field_minute <- function(x) {
  field(x, "minute")
}
field_second <- function(x) {
  field(x, "second")
}
field_subsecond <- function(x) {
  field(x, "subsecond")
}
field_index <- function(x) {
  field(x, "index")
}

is_calendar <- function(x) {
  inherits(x, "clock_calendar")
}

check_calendar <- function(x, ..., arg = caller_arg(x), call = caller_env()) {
  check_inherits(x, what = "clock_calendar", arg = arg, call = call)
}

# ------------------------------------------------------------------------------
# year_month_day()

test_that("helper can create different precisions", {
  x <- year_month_day(2019, 1:2)
  expect_identical(get_year(x), c(2019L, 2019L))
  expect_identical(get_month(x), 1:2)

  x <- year_month_day(2019, 1:2, 3)
  expect_identical(get_day(x), c(3L, 3L))
})

test_that("can create subsecond precision calendars", {
  x <- year_month_day(2019, 1, 1, 0, 0, 0, 1, subsecond_precision = "millisecond")
  expect_identical(get_millisecond(x), 1L)

  x <- year_month_day(2019, 1, 1, 0, 0, 0, 1, subsecond_precision = "microsecond")
  expect_identical(get_microsecond(x), 1L)

  x <- year_month_day(2019, 1, 1, 0, 0, 0, 1, subsecond_precision = "nanosecond")
  expect_identical(get_nanosecond(x), 1L)
})

test_that("validates value ranges", {
  expect_snapshot_error(year_month_day(50000))
  expect_snapshot_error(year_month_day(2020, 13))
  expect_snapshot_error(year_month_day(2020, 1, 32))
  expect_snapshot_error(year_month_day(2020, 1, 1, 24))
  expect_snapshot_error(year_month_day(2020, 1, 1, 1, 60))
  expect_snapshot_error(year_month_day(2020, 1, 1, 1, 1, 60))
  expect_snapshot_error(year_month_day(2020, 1, 1, 1, 1, 1, 1000, subsecond_precision = "millisecond"))
  expect_snapshot_error(year_month_day(2020, 1, 1, 1, 1, 1, 1000000, subsecond_precision = "microsecond"))
  expect_snapshot_error(year_month_day(2020, 1, 1, 1, 1, 1, 1000000000, subsecond_precision = "nanosecond"))
})

test_that("can create a date at the boundary", {
  x <- year_month_day(32767, 12, 31)
  expect_identical(get_year(x), 32767L)

  x <- year_month_day(-32767, 1, 1)
  expect_identical(get_year(x), -32767L)
})

test_that("can get the last day of the month", {
  x <- year_month_day(2019, 1:2, "last")
  expect_identical(get_day(x), c(31L, 28L))
})

test_that("ignores values past first `NULL`", {
  expect_identical(year_month_day(2019, day = 2), year_month_day(2019))
})

test_that("NA values propagate", {
  x <- year_month_day(2019, 1:3, c(NA, 2, 3), c(3, 4, NA))
  expect_identical(is.na(x), c(TRUE, FALSE, TRUE))
})

test_that("names of `year` are not retained", {
  expect_named(year_month_day(c(x = 1)), NULL)
})

# ------------------------------------------------------------------------------
# vec_ptype()

test_that("ptype is correct", {
  base <- year_month_day(1)
  ptype <- year_month_day(integer())

  for (precision in precision_names()) {
    if (precision == "quarter" || precision == "week") {
      next
    }

    x <- calendar_widen(base, precision)
    expect <- calendar_widen(ptype, precision)

    expect_identical(vec_ptype(x), expect)
  }
})

# ------------------------------------------------------------------------------
# vec_proxy() / vec_restore()

test_that("proxy is a data frame", {
  expect_identical(vec_proxy(year_month_day(2019)), data_frame(year = 2019L))
  expect_identical(vec_proxy(year_month_day(2019, 1)), data_frame(year = 2019L, month = 1L))
})

test_that("proxy has names on `year`", {
  x <- set_names(year_month_day(2019, 1, 1), "nm")
  year <- vec_proxy(x)$year
  expect_named(year, "nm")
})

test_that("restore works", {
  to <- year_month_day(2019, 1:5)
  proxy <- vec_slice(vec_proxy(to), 1:2)
  expect_identical(vec_restore(proxy, to), year_month_day(2019, 1:2))
})

# ------------------------------------------------------------------------------
# vec_ptype_full()

test_that("full ptype is correct", {
  expect_snapshot_output(vec_ptype_full(year_month_day(2019)))
  expect_snapshot_output(vec_ptype_full(year_month_day(2019, 1, 1)))
  expect_snapshot_output(vec_ptype_full(year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "nanosecond")))
  expect_snapshot_output(vec_ptype_full(year_month_day(2019, 2, 31)))
})

# ------------------------------------------------------------------------------
# vec_ptype_abbr()

test_that("abbreviated ptype is correct", {
  expect_snapshot_output(vec_ptype_abbr(year_month_day(2019)))
  expect_snapshot_output(vec_ptype_abbr(year_month_day(2019, 1, 1)))
  expect_snapshot_output(vec_ptype_abbr(year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "nanosecond")))
  expect_snapshot_output(vec_ptype_abbr(year_month_day(2019, 2, 31)))
})

# ------------------------------------------------------------------------------
# as_year_quarter_day()

test_that("invalid dates must be resolved when converting to another calendar", {
  expect_snapshot_error(as_year_quarter_day(year_month_day(2019, 2, 31)))
})

# ------------------------------------------------------------------------------
# as_sys_time()

test_that("invalid dates must be resolved when converting to a sys-time", {
  expect_snapshot_error(as_sys_time(year_month_day(2019, 2, 31)))
})

# ------------------------------------------------------------------------------
# as_naive_time()

test_that("invalid dates must be resolved when converting to a naive-time", {
  expect_snapshot_error(as_naive_time(year_month_day(2019, 2, 31)))
})

# ------------------------------------------------------------------------------
# format()

test_that("default formats are correct", {
  expect_snapshot(format(year_month_day(2019)))
  expect_snapshot(format(year_month_day(2019, 1)))
  expect_snapshot(format(year_month_day(2019, 1, 1, 1)))
  expect_snapshot(format(year_month_day(2019, 1, 1, 1, 2, 3, 50, subsecond_precision = "microsecond")))
})

# ------------------------------------------------------------------------------
# as.character()

test_that("as.character() works", {
  x <- year_month_day(2019, 1)
  y <- year_month_day(2019, 1, 2)

  expect_identical(as.character(x), format(x))
  expect_identical(as.character(y), format(y))
})

# ------------------------------------------------------------------------------
# year_month_day_parse()

test_that("default parses at day precision with ISO format", {
  x <- "2019-01-01"

  expect_identical(
    year_month_day_parse(x),
    year_month_day(2019, 1, 1)
  )
})

test_that("can parse years or year-months", {
  x <- "2019"
  y <- "2019-01"

  expect_identical(
    year_month_day_parse(x, precision = "year"),
    year_month_day(2019)
  )
  expect_identical(
    year_month_day_parse(y, precision = "month"),
    year_month_day(2019, 01)
  )
})

test_that("can parse second and subsecond precision", {
  x <- "2019-01-01T05:10:20"
  y <- "2019-01-01T05:10:20.1234"

  expect_identical(
    year_month_day_parse(x, precision = "second"),
    year_month_day(2019, 1, 1, 5, 10, 20)
  )
  expect_identical(
    year_month_day_parse(y, precision = "millisecond"),
    year_month_day(2019, 1, 1, 5, 10, 20, 123, subsecond_precision = "millisecond")
  )
  expect_identical(
    year_month_day_parse(y, precision = "microsecond"),
    year_month_day(2019, 1, 1, 5, 10, 20, 123400, subsecond_precision = "microsecond")
  )
})

test_that("can parse invalid dates", {
  x <- "2019-2-31"

  expect_identical(
    year_month_day_parse(x),
    year_month_day(2019, 2, 31)
  )
})

test_that("names are retained from the input", {
  x <- c(foo = "2019-01-01")

  expect_named(year_month_day_parse(x), "foo")
})

test_that("multiple formats can be provided", {
  x <- c("2019-01", "2020 Jan")
  formats <- c("%Y %B", "%Y-%m")

  expect_identical(
    year_month_day_parse(x, format = formats, precision = "month"),
    year_month_day(c(2019, 2020), 1)
  )
})

test_that("failure to parse results in `NA`", {
  x <- "2020-01-ohno"

  expect_warning(
    expect_identical(
      year_month_day_parse(x),
      year_month_day(NA, NA, NA)
    )
  )
})

test_that("failure to parse results in a warning", {
  expect_warning(year_month_day_parse("foo", precision = "year"), class = "clock_warning_parse_failures")
  expect_warning(year_month_day_parse("foo", precision = "month"), class = "clock_warning_parse_failures")
  expect_warning(year_month_day_parse("foo", precision = "day"), class = "clock_warning_parse_failures")
  expect_warning(year_month_day_parse("foo", precision = "hour"), class = "clock_warning_parse_failures")
  expect_warning(year_month_day_parse("foo", precision = "minute"), class = "clock_warning_parse_failures")
  expect_warning(year_month_day_parse("foo", precision = "second"), class = "clock_warning_parse_failures")
  expect_warning(year_month_day_parse("foo", precision = "millisecond"), class = "clock_warning_parse_failures")
  expect_warning(year_month_day_parse("foo", precision = "microsecond"), class = "clock_warning_parse_failures")
  expect_warning(year_month_day_parse("foo", precision = "nanosecond"), class = "clock_warning_parse_failures")

  expect_snapshot(year_month_day_parse("foo"))
})

test_that("can use a different locale", {
  x <- "janvier 05, 2020"
  y <- "2019-01-01T00:00:00,123456"

  expect_identical(
    year_month_day_parse(x, format = "%B %d, %Y", locale = clock_locale("fr")),
    year_month_day(2020, 1, 5)
  )
  expect_identical(
    year_month_day_parse(y, locale = clock_locale(decimal_mark = ","), precision = "microsecond"),
    year_month_day(2019, 1, 1, 0, 0, 0, 123456, subsecond_precision = "microsecond")
  )
})

test_that("can use a different locale with UTF-8 strings", {
  x <- c("1月 01 2019", "3月 05 2020")
  y <- "ለካቲት 01 2019"

  expect_identical(
    year_month_day_parse(x, format = "%B %d %Y", locale = clock_locale("ja")),
    year_month_day(c(2019, 2020), c(1, 3), c(1, 5))
  )
  expect_identical(
    year_month_day_parse(y, format = "%B %d %Y", locale = clock_locale("ti")),
    year_month_day(2019, 2, 1)
  )
})

test_that("`format` argument is translated to UTF-8", {
  x <- "f\u00E9v 2019-05-19"

  format <- "f\u00E9v %Y-%m-%d"
  format <- iconv(format, from = "UTF-8", to = "latin1")

  expect_identical(Encoding(x), "UTF-8")
  expect_identical(Encoding(format), "latin1")

  expect_identical(
    year_month_day_parse(x, format = format),
    year_month_day(2019, 5, 19)
  )
})

test_that("`x` is translated to UTF-8", {
  x <- "2019-f\u00E9vrier-01"
  x <- iconv(x, from = "UTF-8", to = "latin1")

  locale <- clock_locale("fr")
  format <- "%Y-%B-%d"

  expect_identical(Encoding(x), "latin1")
  expect_identical(Encoding(locale$labels$month[2]), "UTF-8")

  expect_identical(
    year_month_day_parse(x, format = format, locale = locale),
    year_month_day(2019, 2, 1)
  )
})

test_that("parsing NA returns NA", {
  expect_identical(
    year_month_day_parse(NA_character_),
    year_month_day(NA, NA, NA)
  )
  expect_identical(
    year_month_day_parse(NA_character_, precision = "month"),
    year_month_day(NA, NA)
  )
})

test_that("parsing doesn't round parsed components more precise than the resulting container (#207)", {
  # With year-month-day, only the year/month/day components are extracted at the end,
  # the hour component isn't touched
  expect_identical(
    year_month_day_parse("2019-12-31 12", format = "%Y-%m-%d %H", precision = "day"),
    year_month_day(2019, 12, 31)
  )
})

test_that("parsing rounds parsed subsecond components more precise than the resulting container (#207)", {
  # Requesting `%7S` parses the full `01.1238`, and the `1238` portion is rounded up immediately
  # after parsing the `%S` command, not at the very end
  expect_identical(
    year_month_day_parse("2019-01-01 01:01:01.1238", format = "%Y-%m-%d %H:%M:%7S", precision = "millisecond"),
    year_month_day(2019, 1, 1, 1, 1, 1, 124, subsecond_precision = "millisecond")
  )
})

# ------------------------------------------------------------------------------
# calendar_group()

test_that("works with negative years", {
  year <- c(-2, -1, 0, 1, 2)
  x <- year_month_day(year, 1, 1)

  expect_identical(calendar_group(x, "year"), year_month_day(year))
  expect_identical(calendar_group(x, "year", n = 2), year_month_day(c(-2, -2, 0, 0, 2)))
})

# ------------------------------------------------------------------------------
# calendar_narrow()

test_that("can narrow to month", {
  x_expect <- year_month_day(2019, 2)
  x <- set_day(x_expect, 1)
  expect_identical(calendar_narrow(x, "month"), x_expect)
  expect_identical(calendar_narrow(x_expect, "month"), x_expect)
})

test_that("can narrow to day", {
  x_expect <- year_month_day(2019, 2, 3)
  x <- set_hour(x_expect, 5)
  expect_identical(calendar_narrow(x, "day"), x_expect)
  expect_identical(calendar_narrow(x_expect, "day"), x_expect)
})

test_that("can narrow to hour", {
  x_expect <- year_month_day(2019, 2, 3, 4)
  x <- set_minute(x_expect, 5)
  expect_identical(calendar_narrow(x, "hour"), x_expect)
  expect_identical(calendar_narrow(x_expect, "hour"), x_expect)
})

test_that("can narrow to minute", {
  x_expect <- year_month_day(2019, 2, 3, 4, 5)
  x <- set_second(x_expect, 6)
  expect_identical(calendar_narrow(x, "minute"), x_expect)
  expect_identical(calendar_narrow(x_expect, "minute"), x_expect)
})

test_that("can narrow to second", {
  expect <- year_month_day(2019, 2, 3, 4, 5, 6)
  x <- set_millisecond(expect, 7)
  y <- set_nanosecond(expect, 7)
  expect_identical(calendar_narrow(x, "second"), expect)
  expect_identical(calendar_narrow(y, "second"), expect)
  expect_identical(calendar_narrow(expect, "second"), expect)
})

# ------------------------------------------------------------------------------
# calendar_widen()

test_that("can widen to month", {
  x <- year_month_day(2019)
  expect_identical(calendar_widen(x, "month"), set_month(x, 1))
})

test_that("can widen to day", {
  x <- year_month_day(2019)
  y <- year_month_day(2019, 02)
  expect_identical(calendar_widen(x, "day"), set_day(set_month(x, 1), 1))
  expect_identical(calendar_widen(y, "day"), set_day(y, 1))
})

test_that("can widen to hour", {
  x <- year_month_day(2019)
  y <- year_month_day(2019, 02, 02)
  expect_identical(calendar_widen(x, "hour"), set_hour(set_day(set_month(x, 1), 1), 0))
  expect_identical(calendar_widen(y, "hour"), set_hour(y, 0))
})

test_that("can widen to minute", {
  x <- year_month_day(2019)
  y <- year_month_day(2019, 02, 02, 02)
  x_expect <- year_month_day(2019, 1, 1, 0, 0)
  y_expect <- set_minute(y, 0)
  expect_identical(calendar_widen(x, "minute"), x_expect)
  expect_identical(calendar_widen(y, "minute"), y_expect)
})

test_that("can widen to second", {
  x <- year_month_day(2019)
  y <- year_month_day(2019, 02, 02, 02, 02)
  x_expect <- year_month_day(2019, 1, 1, 0, 0, 0)
  y_expect <- set_second(y, 0)
  expect_identical(calendar_widen(x, "second"), x_expect)
  expect_identical(calendar_widen(y, "second"), y_expect)
})

test_that("can widen to subsecond precision", {
  x <- year_month_day(2019)
  y <- year_month_day(2019, 02, 02, 02, 02, 02)
  x_expect <- year_month_day(2019, 1, 1, 0, 0, 0, 0, subsecond_precision = "microsecond")
  y_expect <- set_nanosecond(y, 0)
  expect_identical(calendar_widen(x, "microsecond"), x_expect)
  expect_identical(calendar_widen(y, "nanosecond"), y_expect)
})

# ------------------------------------------------------------------------------
# calendar_start()

test_that("can compute year start", {
  x <- year_month_day(2019)
  expect_identical(calendar_start(x, "year"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "millisecond")
  expect <- year_month_day(2019, 1, 1, 0, 0, 0, 0, subsecond_precision = "millisecond")
  expect_identical(calendar_start(x, "year"), expect)
})

test_that("can compute month start", {
  x <- year_month_day(2019, 2)
  expect_identical(calendar_start(x, "month"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "microsecond")
  expect <- year_month_day(2019, 2, 1, 0, 0, 0, 0, subsecond_precision = "microsecond")
  expect_identical(calendar_start(x, "month"), expect)
})

# ------------------------------------------------------------------------------
# calendar_end()

test_that("can compute year end", {
  x <- year_month_day(2019)
  expect_identical(calendar_end(x, "year"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "millisecond")
  expect <- year_month_day(2019, 12, 31, 23, 59, 59, 999L, subsecond_precision = "millisecond")
  expect_identical(calendar_end(x, "year"), expect)
})

test_that("can compute month end", {
  x <- year_month_day(2019, 2)
  expect_identical(calendar_end(x, "month"), x)

  x <- year_month_day(2019, 2:3, 2, 2, 2, 2, 2, subsecond_precision = "microsecond")
  expect <- year_month_day(2019, 2:3, c(28, 31), 23, 59, 59, 999999L, subsecond_precision = "microsecond")
  expect_identical(calendar_end(x, "month"), expect)
})

# ------------------------------------------------------------------------------
# calendar_month_factor()

test_that("can get a month factor", {
  expect_identical(
    calendar_month_factor(year_month_day(2019, 1:12)),
    factor(month.name, levels = month.name, ordered = TRUE)
  )
})

test_that("can abbreviate month names", {
  expect_identical(
    calendar_month_factor(year_month_day(2019, 1:12), abbreviate = TRUE),
    factor(month.abb, levels = month.abb, ordered = TRUE)
  )
})

test_that("can adjust labels language", {
  labels <- clock_labels_lookup("fr")$month
  expect_identical(
    calendar_month_factor(year_month_day(2019, 1:12), labels = "fr"),
    factor(labels, levels = labels, ordered = TRUE)
  )
})

test_that("requires month precision", {
  expect_snapshot_error(calendar_month_factor(year_month_day(2019)))
})

test_that("`labels` is validated", {
  expect_snapshot_error(calendar_month_factor(year_month_day(2019, 1), labels = 1))
})

test_that("`abbreviate` is validated", {
  expect_snapshot_error(calendar_month_factor(year_month_day(2019, 1), abbreviate = "foo"))
  expect_snapshot_error(calendar_month_factor(year_month_day(2019, 1), abbreviate = 1))
  expect_snapshot_error(calendar_month_factor(year_month_day(2019, 1), abbreviate = c(TRUE, FALSE)))
})

# ------------------------------------------------------------------------------
# calendar_count_between()

test_that("can compute year and month counts", {
  x <- year_month_day(2019, 1, 1)
  y <- year_month_day(2020, 3, 4)

  expect_identical(calendar_count_between(x, y, "year"), 1L)
  expect_identical(calendar_count_between(x, y, "month"), 14L)
  expect_identical(calendar_count_between(x, y, "month", n = 2), 7L)
})

test_that("can compute a quarter count", {
  x <- year_month_day(2019, 1, 2)

  y <- year_month_day(2019, 4, c(1, 3))
  expect_identical(calendar_count_between(x, y, "quarter"), c(0L, 1L))
  expect_identical(
    calendar_count_between(x, y, "quarter"),
    calendar_count_between(x, y, "month", n = 3L)
  )

  y <- year_month_day(2020, 4, c(1, 3))
  expect_identical(calendar_count_between(x, y, "quarter", n = 2L), c(2L, 2L))
  expect_identical(
    calendar_count_between(x, y, "quarter", n = 2L),
    calendar_count_between(x, y, "month", n = 6L)
  )
})

test_that("can't compute a unsupported count precision", {
  x <- year_month_day(2019, 1, 1)
  expect_snapshot((expect_error(calendar_count_between(x, x, "day"))))
})

test_that("positive / negative counts are correct", {
  start <- year_month_day(1972, 03, 04)


  end <- year_month_day(1973, 03, 03)
  expect_identical(calendar_count_between(start, end, "year"), 0L)
  expect_identical(calendar_count_between(start, end, "month"), 11L)

  end <- year_month_day(1973, 03, 04)
  expect_identical(calendar_count_between(start, end, "year"), 1L)
  expect_identical(calendar_count_between(start, end, "month"), 12L)

  end <- year_month_day(1973, 03, 05)
  expect_identical(calendar_count_between(start, end, "year"), 1L)
  expect_identical(calendar_count_between(start, end, "month"), 12L)


  end <- year_month_day(1971, 03, 03)
  expect_identical(calendar_count_between(start, end, "year"), -1L)
  expect_identical(calendar_count_between(start, end, "month"), -12L)

  end <- year_month_day(1971, 03, 04)
  expect_identical(calendar_count_between(start, end, "year"), -1L)
  expect_identical(calendar_count_between(start, end, "month"), -12L)

  end <- year_month_day(1971, 03, 05)
  expect_identical(calendar_count_between(start, end, "year"), 0L)
  expect_identical(calendar_count_between(start, end, "month"), -11L)
})

# ------------------------------------------------------------------------------
# seq()

test_that("only granular precisions are allowed", {
  expect_snapshot_error(seq(year_month_day(2019, 1, 1), by = 1, length.out = 2))
})

test_that("seq(to, by) works", {
  expect_identical(seq(year_month_day(2019, 1), to = year_month_day(2019, 6), by = 2), year_month_day(2019, c(1, 3, 5)))
  expect_identical(seq(year_month_day(2019, 1), to = year_month_day(2019, 5), by = 2), year_month_day(2019, c(1, 3, 5)))

  expect_identical(seq(year_month_day(2019, 1), to = year_month_day(2018, 9), by = -2), year_month_day(c(2019, 2018, 2018), c(1, 11, 9)))
  expect_identical(seq(year_month_day(2019, 1), to = year_month_day(2018, 8), by = -2), year_month_day(c(2019, 2018, 2018), c(1, 11, 9)))
})

test_that("seq(to, length.out) works", {
  expect_identical(seq(year_month_day(2019, 1), to = year_month_day(2019, 5), length.out = 2), year_month_day(2019, c(1, 5)))
  expect_identical(seq(year_month_day(2019, 1), to = year_month_day(2019, 5), length.out = 1), year_month_day(2019, 1))
  expect_identical(seq(year_month_day(2019, 1), to = year_month_day(2019, 5), length.out = 5), year_month_day(2019, 1:5))

  expect_identical(seq(year_month_day(2019, 1), to = year_month_day(2019, 5), along.with = 1:2), year_month_day(2019, c(1, 5)))
})

test_that("seq(by, length.out) works", {
  expect_identical(seq(year_month_day(2019, 1), by = 2, length.out = 3), year_month_day(2019, c(1, 3, 5)))
  expect_identical(seq(year_month_day(2019, 1), by = -2, length.out = 3), year_month_day(c(2019, 2018, 2018), c(1, 11, 9)))

  expect_identical(seq(year_month_day(2019, 1), by = 2, along.with = 1:3), year_month_day(2019, c(1, 3, 5)))
})

test_that("`by` can be a duration", {
  expect_identical(
    seq(year_month_day(2019, 1), to = year_month_day(2025, 5), by = duration_years(1)),
    seq(year_month_day(2019, 1), to = year_month_day(2025, 5), by = 12)
  )
  expect_identical(
    seq(year_month_day(2019, 3), by = duration_years(1), length.out = 5),
    seq(year_month_day(2019, 3), by = 12, length.out = 5)
  )
})

# ------------------------------------------------------------------------------
# invalid_resolve()

test_that("strict mode can be activated", {
  local_options(clock.strict = TRUE)
  expect_snapshot_error(invalid_resolve(year_month_day(2019, 1, 1)))
})

test_that("throws known classed error", {
  expect_snapshot_error(invalid_resolve(year_month_day(2019, 2, 31)))
  expect_error(invalid_resolve(year_month_day(2019, 2, 31)), class = "clock_error_invalid_date")
})

# ------------------------------------------------------------------------------
# vec_math()

test_that("is.nan() works", {
  x <- year_month_day(c(2019, NA))
  expect_identical(is.nan(x), c(FALSE, FALSE))
})

test_that("is.finite() works", {
  x <- year_month_day(c(2019, NA))
  expect_identical(is.finite(x), c(TRUE, FALSE))
})

test_that("is.infinite() works", {
  x <- year_month_day(c(2019, NA))
  expect_identical(is.infinite(x), c(FALSE, FALSE))
})

# ------------------------------------------------------------------------------
# print() / obj_print_data() / obj_print_footer()

test_that("normal print method works", {
  x <- year_month_day(2019, 1:5)
  expect_snapshot(x)
})

test_that("can limit with `max`", {
  x <- year_month_day(2019, 1:5)

  expect_snapshot(print(x, max = 2))
  expect_snapshot(print(x, max = 4))

  # no footer if length >= max
  expect_snapshot(print(x, max = 5))
  expect_snapshot(print(x, max = 6))
})

test_that("`max` defaults to `getOption('max.print')` but can be overridden", {
  local_options(max.print = 3)

  x <- year_month_day(2019, 1:5)

  expect_snapshot(x)
  expect_snapshot(print(x, max = 4))
  expect_snapshot(print(x, max = 5))
})

test_that("`max` is validated", {
  x <- year_month_day(2019)

  expect_snapshot_error(print(x, max = -1))
  expect_snapshot_error(print(x, max = c(1, 2)))
  expect_snapshot_error(print(x, max = NA_integer_))
  expect_snapshot_error(print(x, max = "foo"))
})

# ------------------------------------------------------------------------------
# calendar_group()

test_that("group: `precision` is validated", {
  expect_snapshot_error(calendar_group(year_month_day(2019), "foo"))
})

test_that("group: `precision` must be calendar specific", {
  expect_snapshot_error(calendar_group(year_month_day(2019), "quarter"))
})

test_that("group: `precision` can't be wider than `x`", {
  expect_snapshot_error(calendar_group(year_month_day(2019, 1, 1), "second"))
})

test_that("group: can't group a subsecond precision `x` at another subsecond precision", {
  x <- calendar_widen(year_month_day(2019), "nanosecond")
  expect_snapshot_error(calendar_group(x, "microsecond"))
})

test_that("group: can group subsecond precision at the same subsecond precision", {
  x <- year_month_day(2019, 1, 1, 1, 1, 1, 0:1, subsecond_precision = "millisecond")
  expect_identical(calendar_group(x, "millisecond", n = 2L), x[c(1, 1)])
})

# ------------------------------------------------------------------------------
# calendar_narrow()

test_that("narrow: `precision` is validated", {
  expect_snapshot_error(calendar_narrow(year_month_day(2019), "foo"))
})

test_that("narrow: `precision` must be calendar specific", {
  expect_snapshot_error(calendar_narrow(year_month_day(2019), "quarter"))
})

test_that("narrow: `precision` can't be wider than `x`", {
  expect_snapshot_error(calendar_narrow(year_month_day(2019, 1, 1), "second"))
})

test_that("narrow: can't narrow a subsecond precision `x` to another subsecond precision", {
  x <- calendar_widen(year_month_day(2019), "nanosecond")
  expect_snapshot_error(calendar_narrow(x, "microsecond"))
})

test_that("narrow: can narrow subsecond precision to same subsecond precision", {
  x <- year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "millisecond")
  expect_identical(calendar_narrow(x, "millisecond"), x)
})

# ------------------------------------------------------------------------------
# calendar_widen()

test_that("widen: `precision` is validated", {
  expect_snapshot_error(calendar_widen(year_month_day(2019), "foo"))
})

test_that("widen: `precision` must be calendar specific", {
  expect_snapshot_error(calendar_widen(year_month_day(2019), "quarter"))
})

test_that("widen: `precision` can't be narrower than `x`", {
  expect_snapshot_error(calendar_widen(year_month_day(2019, 1, 1), "month"))
})

test_that("widen: can't widen a subsecond precision `x` to another subsecond precision", {
  x <- calendar_widen(year_month_day(2019), "millisecond")
  expect_snapshot_error(calendar_widen(x, "microsecond"))
})

test_that("widen: can widen subsecond precision to the same subsecond precision", {
  x <- year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "millisecond")
  expect_identical(calendar_widen(x, "millisecond"), x)
})

# ------------------------------------------------------------------------------
# calendar_start()

test_that("start: `x` is validated", {
  expect_snapshot_error(calendar_start(1))
})

test_that("start: `precision` is validated", {
  expect_snapshot_error(calendar_start(year_month_day(2019), "foo"))
  expect_snapshot_error(calendar_start(year_month_day(2019), 1))
})

test_that("start: errors on unsupported precision", {
  expect_snapshot_error(calendar_start(year_month_day(2019, 1), "quarter"))
})

test_that("start: `precision` can't be more precise than `x`", {
  expect_snapshot_error(calendar_start(year_month_day(2019), "month"))
})

test_that("start: can't mix different subsecond precisions", {
  x <- year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "microsecond")
  expect_snapshot_error(calendar_start(x, "millisecond"))
})

test_that("start: can use same subsecond precision", {
  x <- year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "microsecond")
  expect_identical(calendar_start(x, "microsecond"), x)
})

test_that("start: can compute day start", {
  x <- year_month_day(2019, 2, 2)
  expect_identical(calendar_start(x, "day"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "nanosecond")
  expect <- year_month_day(2019, 2, 2, 0, 0, 0, 0, subsecond_precision = "nanosecond")
  expect_identical(calendar_start(x, "day"), expect)
})

test_that("start: can compute hour start", {
  x <- year_month_day(2019, 2, 2, 2)
  expect_identical(calendar_start(x, "hour"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "nanosecond")
  expect <- year_month_day(2019, 2, 2, 2, 0, 0, 0, subsecond_precision = "nanosecond")
  expect_identical(calendar_start(x, "hour"), expect)
})

test_that("start: can compute minute start", {
  x <- year_month_day(2019, 2, 2, 2, 2)
  expect_identical(calendar_start(x, "minute"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "nanosecond")
  expect <- year_month_day(2019, 2, 2, 2, 2, 0, 0, subsecond_precision = "nanosecond")
  expect_identical(calendar_start(x, "minute"), expect)
})

test_that("start: can compute second start", {
  x <- year_month_day(2019, 2, 2, 2, 2, 2)
  expect_identical(calendar_start(x, "second"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "nanosecond")
  expect <- year_month_day(2019, 2, 2, 2, 2, 2, 0, subsecond_precision = "nanosecond")
  expect_identical(calendar_start(x, "second"), expect)
})

test_that("start: invalid dates are adjusted", {
  x <- year_month_day(2019, 2, 31, 3)
  expect_identical(calendar_start(x, "year"), year_month_day(2019, 1, 1, 0))
  expect_identical(calendar_start(x, "day"), year_month_day(2019, 2, 31, 0))
})

# ------------------------------------------------------------------------------
# calendar_end()

test_that("end: `x` is validated", {
  expect_snapshot_error(calendar_end(1))
})

test_that("end: `precision` is validated", {
  expect_snapshot_error(calendar_end(year_month_day(2019), "foo"))
  expect_snapshot_error(calendar_end(year_month_day(2019), 1))
})

test_that("end: errors on unsupported precision", {
  expect_snapshot_error(calendar_end(year_month_day(2019, 1), "quarter"))
})

test_that("end: `precision` can't be more precise than `x`", {
  expect_snapshot_error(calendar_end(year_month_day(2019), "month"))
})

test_that("end: can't mix different subsecond precisions", {
  x <- year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "microsecond")
  expect_snapshot_error(calendar_end(x, "millisecond"))
})

test_that("end: can use same subsecond precision", {
  x <- year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "microsecond")
  expect_identical(calendar_end(x, "microsecond"), x)
})

test_that("end: can compute day end", {
  x <- year_month_day(2019, 2, 2)
  expect_identical(calendar_end(x, "day"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "nanosecond")
  expect <- year_month_day(2019, 2, 2, 23, 59, 59, 999999999L, subsecond_precision = "nanosecond")
  expect_identical(calendar_end(x, "day"), expect)
})

test_that("end: can compute hour end", {
  x <- year_month_day(2019, 2, 2, 2)
  expect_identical(calendar_end(x, "hour"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "nanosecond")
  expect <- year_month_day(2019, 2, 2, 2, 59, 59, 999999999L, subsecond_precision = "nanosecond")
  expect_identical(calendar_end(x, "hour"), expect)
})

test_that("end: can compute minute end", {
  x <- year_month_day(2019, 2, 2, 2, 2)
  expect_identical(calendar_end(x, "minute"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "nanosecond")
  expect <- year_month_day(2019, 2, 2, 2, 2, 59, 999999999L, subsecond_precision = "nanosecond")
  expect_identical(calendar_end(x, "minute"), expect)
})

test_that("end: can compute second end", {
  x <- year_month_day(2019, 2, 2, 2, 2, 2)
  expect_identical(calendar_end(x, "second"), x)

  x <- year_month_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "nanosecond")
  expect <- year_month_day(2019, 2, 2, 2, 2, 2, 999999999L, subsecond_precision = "nanosecond")
  expect_identical(calendar_end(x, "second"), expect)
})

test_that("end: invalid dates are adjusted", {
  x <- year_month_day(2019, 2, 31, 3)
  expect_identical(calendar_end(x, "year"), year_month_day(2019, 12, 31, 23))
  expect_identical(calendar_end(x, "month"), year_month_day(2019, 2, 28, 23))
  expect_identical(calendar_end(x, "day"), year_month_day(2019, 2, 31, 23))
})

# ------------------------------------------------------------------------------
# calendar_count_between()

test_that("`n` gets used", {
  x <- year_month_day(2019, 1)
  y <- year_month_day(2019, 7)
  expect_identical(calendar_count_between(x, y, "month", n = 2), 3L)
})

test_that("`end` must be a calendar", {
  x <- year_month_day(2019)
  expect_snapshot((expect_error(calendar_count_between(x, 1, "year"))))
})

test_that("can't count with a precision finer than the calendar precision", {
  x <- year_month_day(2019)
  expect_snapshot((expect_error(calendar_count_between(x, x, "month"))))
})

test_that("`n` is validated", {
  x <- year_month_day(2019)

  expect_snapshot({
    (expect_error(calendar_count_between(x, x, "year", n = NA_integer_)))
    (expect_error(calendar_count_between(x, x, "year", n = -1)))
    (expect_error(calendar_count_between(x, x, "year", n = 1.5)))
    (expect_error(calendar_count_between(x, x, "year", n = "x")))
    (expect_error(calendar_count_between(x, x, "year", n = c(1L, 2L))))
  })
})

# ------------------------------------------------------------------------------
# calendar_precision()

test_that("precision: can get the precision", {
  expect_identical(calendar_precision(year_month_day(2019, 1)), "month")
  expect_identical(calendar_precision(year_day(2019, 100)), "day")
  expect_identical(calendar_precision(year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "nanosecond")), "nanosecond")
})

test_that("precision: can only be called on calendars", {
  expect_snapshot_error(calendar_precision(sys_days(0)))
})

new_naive_time_from_fields <- function(fields, precision, names) {
  new_time_point_from_fields(fields, precision, CLOCK_NAIVE, names)
}

# ------------------------------------------------------------------------------

naive_days <- function(n = integer()) {
  names <- NULL
  duration <- duration_days(n)
  new_naive_time_from_fields(duration, PRECISION_DAY, names)
}

naive_seconds <- function(n = integer()) {
  names <- NULL
  duration <- duration_seconds(n)
  new_naive_time_from_fields(duration, PRECISION_SECOND, names)
}

# ------------------------------------------------------------------------------

#' Is `x` a naive-time?
#'
#' This function determines if the input is a naive-time object.
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @return `TRUE` if `x` inherits from `"clock_naive_time"`, otherwise `FALSE`.
#'
#' @export
#' @examples
#' is_naive_time(1)
#' is_naive_time(as_naive_time(duration_days(1)))
is_naive_time <- function(x) {
  inherits(x, "clock_naive_time")
}

# ------------------------------------------------------------------------------

#' Parsing: naive-time
#'
#' @description
#' `naive_time_parse()` is a parser into a naive-time.
#'
#' `naive_time_parse()` is useful when you have date-time strings like
#' `"2020-01-01T01:04:30"`. If there is no attached UTC offset or time zone
#' name, then parsing this string as a naive-time is your best option. If
#' you know that this string should be interpreted in a specific time zone,
#' parse as a naive-time, then use [as_zoned_time()].
#'
#' The default options assume that `x` should be parsed at second precision,
#' using a `format` string of `"%Y-%m-%dT%H:%M:%S"`. This matches the default
#' result from calling `format()` on a naive-time.
#'
#' _`naive_time_parse()` ignores both the `%z` and `%Z` commands._
#'
#' If your date-time strings contain a full time zone name and a UTC offset, use
#' [zoned_time_parse_complete()]. If they contain a time zone abbreviation, use
#' [zoned_time_parse_abbrev()].
#'
#' If your date-time strings contain a UTC offset, but not a full time zone
#' name, use [sys_time_parse()].
#'
#' @inheritSection zoned-parsing Full Precision Parsing
#'
#' @inheritParams sys_time_parse
#'
#' @return A naive-time.
#'
#' @export
#' @examples
#' naive_time_parse("2020-01-01T05:06:07")
#'
#' # Day precision
#' naive_time_parse("2020-01-01", precision = "day")
#'
#' # Nanosecond precision, but using a day based format
#' naive_time_parse("2020-01-01", format = "%Y-%m-%d", precision = "nanosecond")
#'
#' # Remember that the `%z` and `%Z` commands are ignored entirely!
#' naive_time_parse(
#'   "2020-01-01 -4000 America/New_York",
#'   format = "%Y-%m-%d %z %Z"
#' )
#'
#' # ---------------------------------------------------------------------------
#' # Fractional seconds and POSIXct
#'
#' # If you have a string with fractional seconds and want to convert it to
#' # a POSIXct, remember that clock treats POSIXct as a second precision type.
#' # Ideally, you'd use a clock type that can support fractional seconds, but
#' # if you really want to parse it into a POSIXct, the correct way to do so
#' # is to parse the full fractional time point with the correct `precision`,
#' # then round to seconds using whatever convention you require, and finally
#' # convert that to POSIXct.
#' x <- c("2020-01-01T00:00:00.123", "2020-01-01T00:00:00.555")
#'
#' # First, parse string with full precision
#' x <- naive_time_parse(x, precision = "millisecond")
#' x
#'
#' # Then round to second with a floor, ceiling, or round to nearest
#' time_point_floor(x, "second")
#' time_point_round(x, "second")
#'
#' # Finally, convert to POSIXct
#' as_date_time(time_point_round(x, "second"), zone = "UTC")
naive_time_parse <- function(x,
                             ...,
                             format = NULL,
                             precision = "second",
                             locale = clock_locale()) {
  precision <- validate_time_point_precision_string(precision)

  fields <- time_point_parse(
    x = x,
    ...,
    format = format,
    precision = precision,
    locale = locale,
    clock = CLOCK_NAIVE
  )

  new_naive_time_from_fields(fields, precision, names(x))
}

# ------------------------------------------------------------------------------

#' Convert to a naive-time
#'
#' @description
#' `as_naive_time()` converts `x` to a naive-time.
#'
#' You can convert to a naive-time from any calendar type, as long as it has
#' at least day precision. There also must not be any invalid dates. If invalid
#' dates exist, they must first be resolved with [invalid_resolve()].
#'
#' Converting to a naive-time from a sys-time or zoned-time retains the printed
#' time, but drops the assumption that the time should be interpreted with any
#' specific time zone.
#'
#' Converting to a naive-time from a duration just wraps the duration in a
#' naive-time object, there is no assumption about the time zone. The duration
#' must have at least day precision.
#'
#' There are convenience methods for converting to a naive-time from R's
#' native date and date-time types. Like converting from a zoned-time, these
#' retain the printed time.
#'
#' @param x `[object]`
#'
#'   An object to convert to a naive-time.
#'
#' @return A naive-time vector.
#'
#' @export
#' @examples
#' x <- as.Date("2019-01-01")
#' as_naive_time(x)
#'
#' ym <- year_month_day(2019, 02)
#'
#' # A minimum of day precision is required
#' try(as_naive_time(ym))
#'
#' ymd <- set_day(ym, 10)
#' as_naive_time(ymd)
as_naive_time <- function(x) {
  UseMethod("as_naive_time")
}

#' @export
as_naive_time.clock_naive_time <- function(x) {
  x
}

#' @export
as_naive_time.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("as_naive_time")
}

# ------------------------------------------------------------------------------

#' @export
as_sys_time.clock_naive_time <- function(x) {
  new_sys_time_from_fields(x, time_point_precision_attribute(x), clock_rcrd_names(x))
}

#' Convert to a zoned-time from a naive-time
#'
#' @description
#' This is a naive-time method for the [as_zoned_time()] generic.
#'
#' Converting to a zoned-time from a naive-time retains the printed time,
#' but changes the underlying duration, depending on the `zone` that you choose.
#'
#' Naive-times are time points with a yet-to-be-determined time zone. By
#' converting them to a zoned-time, all you are doing is specifying that
#' time zone while attempting to keep all other printed information the
#' same (if possible).
#'
#' If you want to retain the underlying duration, try converting to a zoned-time
#' [from a sys-time][as-zoned-time-sys-time], which is a time point
#' interpreted as having a UTC time zone.
#'
#' @section Daylight Saving Time:
#'
#' Converting from a naive-time to a zoned-time is not always possible due to
#' daylight saving time issues. There are two types of these issues:
#'
#' _Nonexistent_ times are the result of daylight saving time "gaps".
#' For example, in the America/New_York time zone, there was a daylight
#' saving time gap 1 second after `"2020-03-08 01:59:59"`, where the clocks
#' changed from `01:59:59 -> 03:00:00`, completely skipping the 2 o'clock hour.
#' This means that if you had a naive time of `"2020-03-08 02:30:00"`, you
#' couldn't convert that straight into a zoned-time with this time zone. To
#' resolve these issues, the `nonexistent` argument can be used to specify
#' one of many nonexistent time resolution strategies.
#'
#' _Ambiguous_ times are the result of daylight saving time "fallbacks".
#' For example, in the America/New_York time zone, there was a daylight
#' saving time fallback 1 second after `"2020-11-01 01:59:59 EDT"`, at which
#' point the clocks "fell backwards" by 1 hour, resulting in a printed time of
#' `"2020-11-01 01:00:00 EST"` (note the EDT->EST shift). This resulted in two
#' 1 o'clock hours for this day, so if you had a naive time of
#' `"2020-11-01 01:30:00"`, you wouldn't be able to convert that directly
#' into a zoned-time with this time zone, as there is no way for clock to know
#' which of the two ambiguous times you wanted. To resolve these issues,
#' the `ambiguous` argument can be used to specify one of many ambiguous
#' time resolution strategies.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[clock_naive_time]`
#'
#'   A naive-time to convert to a zoned-time.
#'
#' @param zone `[character(1)]`
#'
#'   The zone to convert to.
#'
#' @param nonexistent `[character / NULL]`
#'
#'   One of the following nonexistent time resolution strategies, allowed to be
#'   either length 1, or the same length as the input:
#'
#'   - `"roll-forward"`: The next valid instant in time.
#'
#'   - `"roll-backward"`: The previous valid instant in time.
#'
#'   - `"shift-forward"`: Shift the nonexistent time forward by the size of
#'     the daylight saving time gap.
#'
#'   - `"shift-backward`: Shift the nonexistent time backward by the size of
#'     the daylight saving time gap.
#'
#'   - `"NA"`: Replace nonexistent times with `NA`.
#'
#'   - `"error"`: Error on nonexistent times.
#'
#'   Using either `"roll-forward"` or `"roll-backward"` is generally
#'   recommended over shifting, as these two strategies maintain the
#'   _relative ordering_ between elements of the input.
#'
#'   If `NULL`, defaults to `"error"`.
#'
#'   If `getOption("clock.strict")` is `TRUE`, `nonexistent` must be supplied
#'   and cannot be `NULL`. This is a convenient way to make production code
#'   robust to nonexistent times.
#'
#' @param ambiguous `[character / zoned_time / POSIXct / list(2) / NULL]`
#'
#'   One of the following ambiguous time resolution strategies, allowed to be
#'   either length 1, or the same length as the input:
#'
#'   - `"earliest"`: Of the two possible times, choose the earliest one.
#'
#'   - `"latest"`: Of the two possible times, choose the latest one.
#'
#'   - `"NA"`: Replace ambiguous times with `NA`.
#'
#'   - `"error"`: Error on ambiguous times.
#'
#'   Alternatively, `ambiguous` is allowed to be a zoned_time (or POSIXct) that
#'   is either length 1, or the same length as the input. If an ambiguous time
#'   is encountered, the zoned_time is consulted. If the zoned_time corresponds
#'   to a naive_time that is also ambiguous _and_ uses the same daylight saving
#'   time transition point as the original ambiguous time, then the offset of
#'   the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
#'   resolved by consulting the zoned_time, then this method falls back to
#'   `NULL`.
#'
#'   Finally, `ambiguous` is allowed to be a list of size 2, where the first
#'   element of the list is a zoned_time (as described above), and the second
#'   element of the list is an ambiguous time resolution strategy to use when
#'   the ambiguous time cannot be resolved by consulting the zoned_time.
#'   Specifying a zoned_time on its own is identical to `list(<zoned_time>,
#'   NULL)`.
#'
#'   If `NULL`, defaults to `"error"`.
#'
#'   If `getOption("clock.strict")` is `TRUE`, `ambiguous` must be supplied and
#'   cannot be `NULL`. Additionally, `ambiguous` cannot be specified as a
#'   zoned_time on its own, as this implies `NULL` for ambiguous times that the
#'   zoned_time cannot resolve. Instead, it must be specified as a list
#'   alongside an ambiguous time resolution strategy as described above. This is
#'   a convenient way to make production code robust to ambiguous times.
#'
#' @return A zoned-time vector.
#'
#' @name as-zoned-time-naive-time
#' @export
#' @examples
#' library(magrittr)
#'
#' x <- as_naive_time(year_month_day(2019, 1, 1))
#'
#' # Converting a naive-time to a zoned-time generally retains the
#' # printed time, while changing the underlying duration.
#' as_zoned_time(x, "America/New_York")
#' as_zoned_time(x, "America/Los_Angeles")
#'
#' # ---------------------------------------------------------------------------
#' # Nonexistent time:
#'
#' new_york <- "America/New_York"
#'
#' # There was a daylight saving gap in the America/New_York time zone on
#' # 2020-03-08 01:59:59 -> 03:00:00, which means that one of these
#' # naive-times don't exist in that time zone. By default, attempting to
#' # convert it to a zoned time will result in an error.
#' nonexistent_time <- year_month_day(2020, 03, 08, c(02, 03), c(45, 30), 00)
#' nonexistent_time <- as_naive_time(nonexistent_time)
#' try(as_zoned_time(nonexistent_time, new_york))
#'
#' # Resolve this by specifying a nonexistent time resolution strategy
#' as_zoned_time(nonexistent_time, new_york, nonexistent = "roll-forward")
#' as_zoned_time(nonexistent_time, new_york, nonexistent = "roll-backward")
#'
#' # Note that rolling backwards will choose the last possible moment in
#' # time at the current precision of the input
#' nonexistent_nanotime <- time_point_cast(nonexistent_time, "nanosecond")
#' nonexistent_nanotime
#' as_zoned_time(nonexistent_nanotime, new_york, nonexistent = "roll-backward")
#'
#' # A word of caution - Shifting does not guarantee that the relative ordering
#' # of the input is maintained
#' shifted <- as_zoned_time(
#'   nonexistent_time,
#'   new_york,
#'   nonexistent = "shift-forward"
#' )
#' shifted
#'
#' # 02:45:00 < 03:30:00
#' nonexistent_time[1] < nonexistent_time[2]
#' # 03:45:00 > 03:30:00 (relative ordering is lost)
#' shifted[1] < shifted[2]
#'
#' # ---------------------------------------------------------------------------
#' # Ambiguous time:
#'
#' new_york <- "America/New_York"
#'
#' # There was a daylight saving time fallback in the America/New_York time
#' # zone on 2020-11-01 01:59:59 EDT -> 2020-11-01 01:00:00 EST, resulting
#' # in two 1 o'clock hours. This means that the following naive time is
#' # ambiguous since we don't know which of the two 1 o'clocks it belongs to.
#' # By default, attempting to convert it to a zoned time will result in an
#' # error.
#' ambiguous_time <- year_month_day(2020, 11, 01, 01, 30, 00)
#' ambiguous_time <- as_naive_time(ambiguous_time)
#' try(as_zoned_time(ambiguous_time, new_york))
#'
#' # Resolve this by specifying an ambiguous time resolution strategy
#' earliest <- as_zoned_time(ambiguous_time, new_york, ambiguous = "earliest")
#' latest <- as_zoned_time(ambiguous_time, new_york, ambiguous = "latest")
#' na <- as_zoned_time(ambiguous_time, new_york, ambiguous = "NA")
#' earliest
#' latest
#' na
#'
#' # Now assume that you were given the following zoned-times, i.e.,
#' # you didn't build them from scratch so you already know their otherwise
#' # ambiguous offsets
#' x <- c(earliest, latest)
#' x
#'
#' # To set the seconds to 5 in both, you might try:
#' x_naive <- x %>%
#'   as_naive_time() %>%
#'   as_year_month_day() %>%
#'   set_second(5) %>%
#'   as_naive_time()
#'
#' x_naive
#'
#' # But this fails because you've "lost" the information about which
#' # offsets these ambiguous times started in
#' try(as_zoned_time(x_naive, zoned_time_zone(x)))
#'
#' # To get around this, you can use that information by specifying
#' # `ambiguous = x`, which will use the offset from `x` to resolve the
#' # ambiguity in `x_naive` as long as `x` is also an ambiguous time with the
#' # same daylight saving time transition point as `x_naive` (i.e. here
#' # everything has a transition point of `"2020-11-01 01:00:00 EST"`).
#' as_zoned_time(x_naive, zoned_time_zone(x), ambiguous = x)
#'
#' # Say you added one more time to `x` that would not be considered ambiguous
#' # in naive-time
#' x <- c(x, as_zoned_time(as_sys_time(latest) + 3600, zoned_time_zone(latest)))
#' x
#'
#' # Imagine you want to floor this vector to a multiple of 2 hours, with
#' # an origin of 1am that day. You can do this by subtracting the origin,
#' # flooring, then adding it back
#' origin <- year_month_day(2019, 11, 01, 01, 00, 00) %>%
#'   as_naive_time() %>%
#'   as_duration()
#'
#' x_naive <- x %>%
#'   as_naive_time() %>%
#'   add_seconds(-origin) %>%
#'   time_point_floor("hour", n = 2) %>%
#'   add_seconds(origin)
#'
#' x_naive
#'
#' # You again have ambiguous naive-time points, so you might try using
#' # `ambiguous = x`. It looks like this took care of the first two problems,
#' # but we have an issue at location 3.
#' try(as_zoned_time(x_naive, zoned_time_zone(x), ambiguous = x))
#'
#' # When we floored from 02:30:00 -> 01:00:00, we went from being
#' # unambiguous -> ambiguous. In clock, this is something you must handle
#' # explicitly, and cannot be handled by using information from `x`. You can
#' # handle this while still retaining the behavior for the other two
#' # time points that were ambiguous before and after the floor by passing a
#' # list containing `x` and an ambiguous time resolution strategy to use
#' # when information from `x` can't resolve ambiguities:
#' as_zoned_time(x_naive, zoned_time_zone(x), ambiguous = list(x, "latest"))
as_zoned_time.clock_naive_time <- function(x,
                                           zone,
                                           ...,
                                           nonexistent = NULL,
                                           ambiguous = NULL) {
  check_dots_empty()

  zone <- zone_validate(zone)

  # Promote to at least seconds precision for `zoned_time`
  ptype <- vec_ptype2(x, clock_empty_naive_time_second, y_arg = "")
  x <- vec_cast(x, ptype)

  size <- vec_size(x)
  precision <- time_point_precision_attribute(x)
  names <- clock_rcrd_names(x)

  nonexistent <- validate_nonexistent(nonexistent, size)

  info <- validate_ambiguous(ambiguous, size, zone)
  method <- info$method

  if (identical(method, "string")) {
    ambiguous <- info$ambiguous
    fields <- as_zoned_sys_time_from_naive_time_cpp(x, precision, zone, nonexistent, ambiguous)
  } else if (identical(method, "reference")) {
    reference <- info$reference
    ambiguous <- info$ambiguous
    fields <- as_zoned_sys_time_from_naive_time_with_reference_cpp(x, precision, zone, nonexistent, ambiguous, reference)
  } else {
    abort("Internal error: Unknown ambiguous handling method.")
  }

  new_zoned_time_from_fields(fields, precision, zone, names)
}

validate_nonexistent <- function(nonexistent, size) {
  nonexistent <- strict_validate_nonexistent(nonexistent)

  nonexistent_size <- vec_size(nonexistent)

  if (nonexistent_size != 1L && nonexistent_size != size) {
    abort(paste0("`nonexistent` must have length 1, or ", size, "."))
  }

  if (!is_character(nonexistent)) {
    abort("`nonexistent` must be a character vector, or `NULL`.")
  }

  nonexistent
}

validate_ambiguous <- function(ambiguous, size, zone) {
  if (is_null(ambiguous)) {
    ambiguous <- strict_validate_ambiguous(ambiguous)
    return(list(method = "string", ambiguous = ambiguous))
  }

  if (is_character(ambiguous)) {
    ambiguous <- validate_ambiguous_chr(ambiguous, size)
    return(list(method = "string", ambiguous = ambiguous))
  }

  if (is_zoned_time(ambiguous) || inherits(ambiguous, "POSIXt")) {
    # Implied `NULL`, to be validated by `strict_validate_ambiguous()`
    ambiguous <- list(ambiguous, NULL)
  }

  if (is_list(ambiguous)) {
    result <- validate_ambiguous_list(ambiguous, size, zone)
    reference <- result$reference
    ambiguous <- result$ambiguous
    return(list(method = "reference", reference = reference, ambiguous = ambiguous))
  }

  abort("`ambiguous` must be a character vector, a zoned-time, a POSIXct, or a list.")
}

validate_ambiguous_chr <- function(ambiguous, size) {
  ambiguous_size <- vec_size(ambiguous)

  if (ambiguous_size != 1L && ambiguous_size != size) {
    abort(paste0("`ambiguous` must have length 1, or ", size, "."))
  }

  ambiguous
}

validate_ambiguous_zoned <- function(ambiguous, size, zone) {
  # POSIXt -> zoned_time
  reference <- as_zoned_time(ambiguous)

  reference_size <- vec_size(reference)
  reference_zone <- zoned_time_zone_attribute(reference)

  if (reference_size != 1L && reference_size != size) {
    abort(paste0("A zoned-time or POSIXct `ambiguous` must have length 1, or ", size, "."))
  }
  if (reference_zone != zone) {
    abort("A zoned-time or POSIXct `ambiguous` must have the same zone as `zone`.")
  }

  # Force seconds precision to avoid the need for C++ templating
  sys_time <- as_sys_time(reference)
  sys_time <- time_point_floor(sys_time, "second")
  reference <- as_zoned_time(sys_time, reference_zone)

  reference
}

validate_ambiguous_list <- function(ambiguous, size, zone) {
  if (length(ambiguous) != 2L) {
    abort("A list `ambiguous` must have length 2.")
  }

  reference <- ambiguous[[1]]

  if (!is_zoned_time(reference) && !inherits(reference, "POSIXt")) {
    abort("The first element of a list `ambiguous` must be a zoned-time or POSIXt.")
  }

  reference <- validate_ambiguous_zoned(reference, size, zone)

  ambiguous <- ambiguous[[2]]

  if (is_null(ambiguous)) {
    ambiguous <- strict_validate_ambiguous(ambiguous)
  }
  if (!is_character(ambiguous)) {
    abort("The second element of a list `ambiguous` must be a character vector, or `NULL`.")
  }

  ambiguous <- validate_ambiguous_chr(ambiguous, size)

  list(reference = reference, ambiguous = ambiguous)
}


#' @export
as.character.clock_naive_time <- function(x, ...) {
  format(x)
}

# ------------------------------------------------------------------------------

#' Info: naive-time
#'
#' @description
#' `naive_time_info()` retrieves a set of low-level information generally not
#' required for most date-time manipulations. It is used implicitly
#' by `as_zoned_time()` when converting from a naive-time.
#'
#' It returns a data frame with the following columns:
#'
#' - `type`: A character vector containing one of:
#'
#'   - `"unique"`: The naive-time maps uniquely to a zoned-time that can be
#'   created with `zone`.
#'
#'   - `"nonexistent"`: The naive-time does not exist as a zoned-time that can
#'   be created with `zone`.
#'
#'   - `"ambiguous"`: The naive-time exists twice as a zoned-time that can be
#'   created with `zone`.
#'
#' - `first`: A [sys_time_info()] data frame.
#'
#' - `second`: A [sys_time_info()] data frame.
#'
#' ## type == "unique"
#'
#' - `first` will be filled out with sys-info representing daylight saving time
#' information for that time point in `zone`.
#'
#' - `second` will contain only `NA` values, as there is no ambiguity to
#' represent information for.
#'
#' ## type == "nonexistent"
#'
#' - `first` will be filled out with the sys-info that ends just prior to `x`.
#'
#' - `second` will be filled out with the sys-info that begins just after `x`.
#'
#' ## type == "ambiguous"
#'
#' - `first` will be filled out with the sys-info that ends just after `x`.
#'
#' - `second` will be filled out with the sys-info that starts just before `x`.
#'
#' @details
#' If the tibble package is installed, it is recommended to convert the output
#' to a tibble with `as_tibble()`, as that will print the df-cols much nicer.
#'
#' @param x `[clock_naive_time]`
#'
#'   A naive-time.
#'
#' @param zone `[character]`
#'
#'   A valid time zone name.
#'
#'   Unlike most functions in clock, in `naive_time_info()` `zone` is vectorized
#'   and is recycled against `x`.
#'
#' @return A data frame of low level information.
#'
#' @export
#' @examples
#' library(vctrs)
#'
#' x <- year_month_day(1970, 04, 26, 02, 30, 00)
#' x <- as_naive_time(x)
#'
#' # Maps uniquely to a time in London
#' naive_time_info(x, "Europe/London")
#'
#' # This naive-time never existed in New York!
#' # A DST gap jumped the time from 01:59:59 -> 03:00:00,
#' # skipping the 2 o'clock hour
#' zone <- "America/New_York"
#' info <- naive_time_info(x, zone)
#' info
#'
#' # You can recreate various `nonexistent` strategies with this info
#' as_zoned_time(x, zone, nonexistent = "roll-forward")
#' as_zoned_time(info$first$end, zone)
#'
#' as_zoned_time(x, zone, nonexistent = "roll-backward")
#' as_zoned_time(info$first$end - 1, zone)
#'
#' as_zoned_time(x, zone, nonexistent = "shift-forward")
#' as_zoned_time(as_sys_time(x) - info$first$offset, zone)
#'
#' as_zoned_time(x, zone, nonexistent = "shift-backward")
#' as_zoned_time(as_sys_time(x) - info$second$offset, zone)
#'
#' # ---------------------------------------------------------------------------
#' # Normalizing to UTC
#'
#' # Imagine you had the following printed times, and knowledge that they
#' # are to be interpreted as in the corresponding time zones
#' df <- data_frame(
#'   x = c("2020-01-05 02:30:00", "2020-06-03 12:20:05"),
#'   zone = c("America/Los_Angeles", "Europe/London")
#' )
#'
#' # The times are assumed to be naive-times, i.e. if you lived in the `zone`
#' # at the moment the time was recorded, then you would have seen that time
#' # printed on the clock. Currently, these are strings. To convert them to
#' # a time based type, you'll have to acknowledge that R only lets you have
#' # 1 time zone in a vector of date-times at a time. So you'll need to
#' # normalize these naive-times. The easiest thing to normalize them to
#' # is UTC.
#' df$naive <- naive_time_parse(df$x)
#'
#' # Get info about the naive times using a vector of zones
#' info <- naive_time_info(df$naive, df$zone)
#' info
#'
#' # We'll assume that some system generated these naive-times with no
#' # chance of them ever being nonexistent or ambiguous. So now all we have
#' # to do is use the offset to convert the naive-time to a sys-time. The
#' # relationship used is:
#' # offset = naive_time - sys_time
#' df$sys <- as_sys_time(df$naive) - info$first$offset
#' df
#'
#' # At this point, both times are in UTC. From here, you can convert them
#' # both to either America/Los_Angeles or Europe/London as required.
#' as_zoned_time(df$sys, "America/Los_Angeles")
#' as_zoned_time(df$sys, "Europe/London")
naive_time_info <- function(x, zone) {
  if (!is_naive_time(x)) {
    abort("`x` must be a naive-time.")
  }

  precision <- time_point_precision_attribute(x)

  # Recycle `x` to the common size. `zone` is recycled internally as required,
  # which is more efficient than reloading the time zone repeatedly.
  size <- vec_size_common(x = x, zone = zone)
  x <- vec_recycle(x, size)

  fields <- naive_time_info_cpp(x, precision, zone)

  new_naive_time_info_from_fields(fields)
}

new_naive_time_info_from_fields <- function(fields) {
  fields[["first"]] <- new_sys_time_info_from_fields(fields[["first"]])
  fields[["second"]] <- new_sys_time_info_from_fields(fields[["second"]])
  new_data_frame(fields)
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype.clock_naive_time <- function(x, ...) {
  switch(
    time_point_precision_attribute(x) + 1L,
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    clock_empty_naive_time_day,
    clock_empty_naive_time_hour,
    clock_empty_naive_time_minute,
    clock_empty_naive_time_second,
    clock_empty_naive_time_millisecond,
    clock_empty_naive_time_microsecond,
    clock_empty_naive_time_nanosecond,
    abort("Internal error: Invalid precision.")
  )
}

#' @export
vec_ptype2.clock_naive_time.clock_naive_time <- function(x, y, ...) {
  ptype2_time_point_and_time_point(x, y, ...)
}

#' @export
vec_cast.clock_naive_time.clock_naive_time <- function(x, to, ...) {
  cast_time_point_to_time_point(x, to, ...)
}

# ------------------------------------------------------------------------------

#' @rdname clock-arith
#' @method vec_arith clock_naive_time
#' @export
vec_arith.clock_naive_time <- function(op, x, y, ...) {
  UseMethod("vec_arith.clock_naive_time", y)
}

#' @method vec_arith.clock_naive_time MISSING
#' @export
vec_arith.clock_naive_time.MISSING <- function(op, x, y, ...) {
  arith_time_point_and_missing(op, x, y, ...)
}

#' @method vec_arith.clock_naive_time clock_naive_time
#' @export
vec_arith.clock_naive_time.clock_naive_time <- function(op, x, y, ...) {
  arith_time_point_and_time_point(op, x, y, ...)
}

#' @method vec_arith.clock_naive_time clock_duration
#' @export
vec_arith.clock_naive_time.clock_duration <- function(op, x, y, ...) {
  arith_time_point_and_duration(op, x, y, ...)
}

#' @method vec_arith.clock_duration clock_naive_time
#' @export
vec_arith.clock_duration.clock_naive_time <- function(op, x, y, ...) {
  arith_duration_and_time_point(op, x, y, ...)
}

#' @method vec_arith.clock_naive_time numeric
#' @export
vec_arith.clock_naive_time.numeric <- function(op, x, y, ...) {
  arith_time_point_and_numeric(op, x, y, ...)
}

#' @method vec_arith.numeric clock_naive_time
#' @export
vec_arith.numeric.clock_naive_time <- function(op, x, y, ...) {
  arith_numeric_and_time_point(op, x, y, ...)
}

# ------------------------------------------------------------------------------

clock_init_naive_time_utils <- function(env) {
  day <- as_naive_time(year_month_day(integer(), integer(), integer()))

  assign("clock_empty_naive_time_day", day, envir = env)
  assign("clock_empty_naive_time_hour", time_point_cast(day, "hour"), envir = env)
  assign("clock_empty_naive_time_minute", time_point_cast(day, "minute"), envir = env)
  assign("clock_empty_naive_time_second", time_point_cast(day, "second"), envir = env)
  assign("clock_empty_naive_time_millisecond", time_point_cast(day, "millisecond"), envir = env)
  assign("clock_empty_naive_time_microsecond", time_point_cast(day, "microsecond"), envir = env)
  assign("clock_empty_naive_time_nanosecond", time_point_cast(day, "nanosecond"), envir = env)

  invisible(NULL)
}

# ------------------------------------------------------------------------------
# naive_time_info()

test_that("naive-time-info returns the right structure", {
  info <- naive_time_info(naive_days(0), "America/New_York")

  expect_type(info$type, "character")
  expect_s3_class(info$first, "data.frame")
  expect_s3_class(info$second, "data.frame")
})

test_that("unique info works", {
  info <- naive_time_info(naive_days(0), "America/New_York")
  na_sys_info <- sys_time_info(sys_days(NA), "America/New_York")

  expect_identical(info$type, "unique")
  expect_identical(info$second, na_sys_info)

  begin <- as_sys_time(year_month_day(1969, 10, 26, 06, 00, 00))
  end <- as_sys_time(year_month_day(1970, 4, 26, 7, 0, 0))
  offset <- duration_seconds(-18000)

  expect_identical(info$first$begin, begin)
  expect_identical(info$first$end, end)
  expect_identical(info$first$offset, offset)
  expect_identical(info$first$dst, FALSE)
  expect_identical(info$first$abbreviation, "EST")
})

test_that("nonexistent info works", {
  x <- as_naive_time(year_month_day(1970, 4, 26, 2, 30, 00))
  info <- naive_time_info(x, "America/New_York")

  expect_identical(info$type, "nonexistent")

  begin <- as_sys_time(year_month_day(1969, 10, 26, 06, 00, 00))
  end <- as_sys_time(year_month_day(1970, 4, 26, 7, 0, 0))
  offset <- duration_seconds(-18000)

  expect_identical(info$first$begin, begin)
  expect_identical(info$first$end, end)
  expect_identical(info$first$offset, offset)
  expect_identical(info$first$dst, FALSE)
  expect_identical(info$first$abbreviation, "EST")

  begin <- as_sys_time(year_month_day(1970, 4, 26, 7, 0, 0))
  end <- as_sys_time(year_month_day(1970, 10, 25, 6, 0, 0))
  offset <- duration_seconds(-14400)

  expect_identical(info$second$begin, begin)
  expect_identical(info$second$end, end)
  expect_identical(info$second$offset, offset)
  expect_identical(info$second$dst, TRUE)
  expect_identical(info$second$abbreviation, "EDT")
})

test_that("ambiguous info works", {
  x <- as_naive_time(year_month_day(1970, 10, 25, 1, 30, 00))
  info <- naive_time_info(x, "America/New_York")

  expect_identical(info$type, "ambiguous")

  begin <- as_sys_time(year_month_day(1970, 4, 26, 7, 0, 0))
  end <- as_sys_time(year_month_day(1970, 10, 25, 6, 0, 0))
  offset <- duration_seconds(-14400)

  expect_identical(info$first$begin, begin)
  expect_identical(info$first$end, end)
  expect_identical(info$first$offset, offset)
  expect_identical(info$first$dst, TRUE)
  expect_identical(info$first$abbreviation, "EDT")

  begin <- as_sys_time(year_month_day(1970, 10, 25, 6, 0, 0))
  end <- as_sys_time(year_month_day(1971, 4, 25, 7, 0, 0))
  offset <- duration_seconds(-18000)

  expect_identical(info$second$begin, begin)
  expect_identical(info$second$end, end)
  expect_identical(info$second$offset, offset)
  expect_identical(info$second$dst, FALSE)
  expect_identical(info$second$abbreviation, "EST")
})

test_that("all rows are NA when x is NA", {
  info <- naive_time_info(naive_days(NA), "UTC")
  na_sys_info <- sys_time_info(sys_days(NA), "UTC")
  df <- data_frame(type = NA_character_, first = na_sys_info, second = na_sys_info)
  expect_identical(info, df)
})

test_that("x must be naive", {
  expect_snapshot_error(naive_time_info(sys_days(0), "UTC"))
})

test_that("zone is vectorized and recycled against x", {
  info <- naive_time_info(naive_days(0), c("UTC", "America/New_York"))
  expect_identical(nrow(info), 2L)

  expect_snapshot_error(naive_time_info(naive_days(0:3), c("UTC", "America/New_York")))
})

# ------------------------------------------------------------------------------
# as.character()

test_that("as.character() works", {
  x <- as_naive_time(year_month_day(2019, 1, 1))
  expect_identical(as.character(x), "2019-01-01")

  x <- as_naive_time(year_month_day(2019, 1, 1, 1, 1))
  expect_identical(as.character(x), "2019-01-01T01:01")
})

# ------------------------------------------------------------------------------
# naive_time_parse()

test_that("can parse day precision", {
  x <- c("2019-01-01", "2019-01-31")

  expect_identical(
    naive_time_parse(x, precision = "day"),
    as_naive_time(year_month_day(2019, 1, c(1, 31)))
  )
})

test_that("can parse second precision", {
  x <- c("2019-01-01T00:00:05", "2019-01-31T00:00:10")

  expect_identical(
    naive_time_parse(x, precision = "second"),
    as_naive_time(year_month_day(2019, 1, c(1, 31), 00, 00, c(05, 10)))
  )
})

test_that("can parse subsecond precision", {
  x <- c("2019-01-01T00:00:05.123", "2019-01-31T00:00:10.124")
  y <- c("2019-01-01T00:00:05.12345", "2019-01-31T00:00:10.124567")
  z <- c("2019-01-01T00:00:05.12345678", "2019-01-31T00:00:10.124567899")

  sec <- year_month_day(2019, 1, c(1, 31), 00, 00, c(05, 10))

  expect_identical(
    naive_time_parse(x, precision = "millisecond"),
    as_naive_time(set_millisecond(sec, c(123, 124)))
  )
  expect_identical(
    naive_time_parse(y, precision = "microsecond"),
    as_naive_time(set_microsecond(sec, c(123450, 124567)))
  )
  expect_identical(
    naive_time_parse(z, precision = "nanosecond"),
    as_naive_time(set_nanosecond(sec, c(123456780, 124567899)))
  )
})

test_that("parsing works if `precision` uses a default that doesn't attempt to capture all the info", {
  # Uses %Y-%m-%d
  x <- "2019-01-01T01:00:00"
  expect_identical(
    naive_time_parse(x, precision = "day"),
    as_naive_time(year_month_day(2019, 1, 1))
  )

  # Uses %Y-%m-%dT%H:%M
  x <- "2019-01-01T01:00:59"
  expect_identical(
    naive_time_parse(x, precision = "minute"),
    as_naive_time(year_month_day(2019, 1, 1, 1, 0))
  )
})

test_that("parsing day components with second precision uses midnight as time", {
  x <- "2019/1/1"

  expect_identical(
    naive_time_parse(x, format = "%Y/%m/%d", precision = "second"),
    as_naive_time(year_month_day(2019, 1, 1, 0, 0, 0))
  )
})

test_that("cannot parse invalid dates", {
  x <- "2019-02-31"

  expect_warning(
    expect_identical(
      naive_time_parse(x, precision = "day"),
      naive_days(NA)
    )
  )

  expect_snapshot(naive_time_parse(x, precision = "day"))
})

test_that("can parse with multiple formats", {
  x <- c("2019-01-01", "2020/1/2", "January 05, 2019")
  formats <- c("%Y-%m-%d", "%Y/%m/%d", "%B %d, %Y")

  expect_identical(
    naive_time_parse(x, format = formats, precision = "day"),
    as_naive_time(year_month_day(c(2019, 2020, 2019), 1, c(1, 2, 5)))
  )
})

test_that("failure to parse results in NA", {
  x <- "2019-01-oh"

  expect_warning(
    expect_identical(
      naive_time_parse(x, format = "%Y-%m-%d", precision = "day"),
      naive_days(NA)
    )
  )
})

test_that("failure to parse throws a warning", {
  expect_warning(naive_time_parse("foo"), class = "clock_warning_parse_failures")
  expect_snapshot(naive_time_parse("foo"))
})

test_that("names of input are kept", {
  x <- c(foo = "2019-01-01")
  expect_named(naive_time_parse(x, precision = "day"), "foo")
})

test_that("can use a different locale", {
  x <- "janvier 01, 2019"
  y <- "2019-01-01T00:00:00,123456"

  expect_identical(
    naive_time_parse(x, format = "%B %d, %Y", precision = "day", locale = clock_locale("fr")),
    as_naive_time(year_month_day(2019, 1, 1))
  )
  expect_identical(
    naive_time_parse(y, precision = "microsecond", locale = clock_locale(decimal_mark = ",")),
    as_naive_time(year_month_day(2019, 1, 1, 0, 0, 0, 123456, subsecond_precision = "microsecond"))
  )
})

test_that("`x` is translated to UTF-8", {
  x <- "f\u00E9vrier 05, 2019"
  x <- iconv(x, from = "UTF-8", to = "latin1")

  locale <- clock_locale("fr")
  format <- "%B %d, %Y"

  expect_identical(Encoding(x[1]), "latin1")
  expect_identical(Encoding(locale$labels$month[2]), "UTF-8")

  expect_identical(
    naive_time_parse(x, format = format, precision = "day", locale = locale),
    as_naive_time(year_month_day(2019, 2, 5))
  )
})

test_that("%z is completely ignored, but is required to be parsed correctly if specified", {
  x <- "2019-01-01 00:00:00+0100"
  y <- "2019-01-01 00:00:00"

  expect_identical(
    naive_time_parse(x, format = "%Y-%m-%d %H:%M:%S%z"),
    as_naive_time(year_month_day(2019, 1, 1, 0, 0, 0))
  )
  expect_warning(
    expect_identical(
      naive_time_parse(y, format = "%Y-%m-%d %H:%M:%S%z"),
      naive_seconds(NA)
    )
  )
})

test_that("%Z is completely ignored", {
  x <- "2019-01-01 00:00:00 America/New_York"

  expect_identical(
    naive_time_parse(x, format = "%Y-%m-%d %H:%M:%S %Z"),
    as_naive_time(year_month_day(2019, 1, 1, 0, 0, 0))
  )
})

test_that("parsing rounds parsed components more precise than the resulting container (#207) (#230) (undocumented)", {
  expect_identical(
    naive_time_parse("2019-12-31 11", format = "%Y-%m-%d %H", precision = "day"),
    as_naive_time(year_month_day(2019, 12, 31))
  )
  expect_identical(
    naive_time_parse("2019-12-31 12", format = "%Y-%m-%d %H", precision = "day"),
    as_naive_time(year_month_day(2020, 1, 1))
  )

  # If you don't try and parse them, it won't round
  expect_identical(
    naive_time_parse("2019-12-31 12", format = "%Y-%m-%d", precision = "day"),
    as_naive_time(year_month_day(2019, 12, 31))
  )
})

test_that("parsing rounds parsed subsecond components more precise than the resulting container (#207) (#230) (undocumented)", {
  # Default N for milliseconds is 6, so `%6S` (2 hour seconds, 1 for decimal, 3 for subseconds)
  expect_identical(
    naive_time_parse("2019-01-01 01:01:01.1238", format = "%Y-%m-%d %H:%M:%S", precision = "millisecond"),
    as_naive_time(year_month_day(2019, 1, 1, 1, 1, 1, 123, subsecond_precision = "millisecond"))
  )

  # Requesting `%7S` parses the full `01.1238`, and the `1238` portion is rounded up
  expect_identical(
    naive_time_parse("2019-01-01 01:01:01.1238", format = "%Y-%m-%d %H:%M:%7S", precision = "millisecond"),
    as_naive_time(year_month_day(2019, 1, 1, 1, 1, 1, 124, subsecond_precision = "millisecond"))
  )
})

test_that("parsing fails when undocumented rounding behavior would result in invalid 60 second component (#230) (undocumented)", {
  expect_warning(
    expect_identical(
      naive_time_parse("2019-01-01 01:01:59.550", format = "%Y-%m-%d %H:%M:%6S", precision = "second"),
      as_naive_time(year_month_day(NA, NA, NA, NA, NA, NA))
    ),
    class = "clock_warning_parse_failures"
  )
})

# ------------------------------------------------------------------------------
# format()

test_that("default format is correct", {
  expect_snapshot(format(naive_seconds(0)))
})

test_that("`%Z` generates format warnings (#204)", {
  x <- naive_seconds(0)

  expect_warning(format(x, format = "%Z"), class = "clock_warning_format_failures")

  expect_snapshot(format(x, format = "%Z"))
  expect_snapshot(format(c(x, x), format = "%Z"))
})

test_that("`%z` generates format warnings (#204)", {
  x <- naive_seconds(0)

  expect_warning(format(x, format = "%z"), class = "clock_warning_format_failures")

  expect_snapshot(format(x, format = "%z"))
  expect_snapshot(format(c(x, x), format = "%z"))
})

# ------------------------------------------------------------------------------
# as_zoned_time()

test_that("can convert non-ambiguous/nonexistent times to zoned time", {
  zone <- "America/New_York"
  x <- as_naive_time(year_month_day(2019, 1, 1))
  expect <- as_sys_time(year_month_day(2019, 1, 1, 5))
  expect_identical(as_zoned_time(x, zone), as_zoned_time(expect, zone))
})

test_that("sub daily time point precision is retained", {
  zone <- "America/New_York"
  x <- as_naive_time(year_month_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "nanosecond"))
  expect_identical(zoned_time_precision_attribute(as_zoned_time(x, zone)), PRECISION_NANOSECOND)
})

test_that("day precision time point is promoted", {
  zone <- "America/New_York"
  x <- as_naive_time(year_month_day(2019, 1, 1))
  expect_identical(zoned_time_precision_attribute(as_zoned_time(x, zone)), PRECISION_SECOND)
})

test_that("can resolve ambiguous issues - character", {
  zone <- "America/New_York"
  x <- as_naive_time(year_month_day(1970, 10, 25, 01, 30, 00, 01, subsecond_precision = "millisecond"))
  earliest <- as_sys_time(year_month_day(1970, 10, 25, 05, 30, 00, 01, subsecond_precision = "millisecond"))
  latest <- as_sys_time(year_month_day(1970, 10, 25, 06, 30, 00, 01, subsecond_precision = "millisecond"))

  expect_snapshot_error(as_zoned_time(x, zone))
  expect_error(as_zoned_time(x, zone), class = "clock_error_ambiguous_time")

  expect_identical(
    as_zoned_time(x, zone, ambiguous = "earliest"),
    as_zoned_time(earliest, zone)
  )
  expect_identical(
    as_zoned_time(x, zone, ambiguous = "latest"),
    as_zoned_time(latest, zone)
  )
  expect_identical(
    as_zoned_time(x, zone, ambiguous = "NA"),
    as_zoned_time(as_naive_time(duration_milliseconds(NA)), zone)
  )
})

test_that("can resolve ambiguous issues - zoned-time", {
  zone <- "America/New_York"
  nt <- as_naive_time(year_month_day(1970, 10, 25, 01, 30, c(00, 00)))
  zt <- as_zoned_time(nt, zone, ambiguous = c("earliest", "latest"))

  expect_identical(
    as_zoned_time(nt, zone, ambiguous = zt),
    zt
  )

  nt <- as_naive_time(year_month_day(1970, 10, 25, c(01, 02), 30, 00))
  zt <- as_zoned_time(nt, zone, ambiguous = "earliest")

  expect_identical(
    as_zoned_time(nt, zone, ambiguous = zt),
    zt
  )

  # Issue at location 2 because zt[2] isn't ambiguous so we can't use offset
  # information from it
  nt_tweaked <- as_naive_time(set_hour(as_year_month_day(nt), 1))
  expect_snapshot_error(as_zoned_time(nt_tweaked, zone, ambiguous = zt))

  # Jump from one ambiguous time to another. Still can't use offset info,
  # because the ambiguous time transitions are different.
  ymd <- year_month_day(1969, 10, 26, 01, 30, 00)
  nt <- as_naive_time(ymd)
  zt <- as_zoned_time(nt, zone, ambiguous = "earliest")
  nt_tweaked <- nt + duration_days(364)
  expect_snapshot_error(as_zoned_time(nt_tweaked, zone, ambiguous = zt))
})

test_that("can resolve ambiguous issues - list", {
  zone <- "America/New_York"
  nt <- as_naive_time(year_month_day(1970, 10, 25, c(01, 01, 02), 30, 00))
  zt <- as_zoned_time(nt, zone, ambiguous = c("earliest", "latest", "error"))
  nt_tweaked <- as_naive_time(set_hour(as_year_month_day(nt), 1))

  # First two are resolved from consulting `zt`, otherwise resolved with "latest"
  expect_identical(
    as_zoned_time(nt_tweaked, zone, ambiguous = list(zt, "latest")),
    as_zoned_time(nt_tweaked, zone, ambiguous = c("earliest", "latest", "latest"))
  )
})

test_that("zoned-time ambiguous argument is recycled", {
  zone <- "America/New_York"
  nt <- as_naive_time(year_month_day(1970, 10, 25, 01, 30, 00))
  zt <- as_zoned_time(nt, zone, ambiguous = "earliest")
  nt_tweaked <- nt + duration_seconds(c(0, 1))

  expect_identical(
    as_zoned_time(nt_tweaked, zone, ambiguous = zt),
    as_zoned_time(nt_tweaked, zone, ambiguous = "earliest")
  )
})

test_that("can resolve nonexistent issues", {
  zone <- "America/New_York"
  x <- as_naive_time(year_month_day(1970, 04, 26, 02, 30, 00))

  expect_snapshot_error(as_zoned_time(x, zone))
  expect_error(as_zoned_time(x, zone), class = "clock_error_nonexistent_time")

  expect_identical(
    as_zoned_time(x, zone, nonexistent = "roll-forward"),
    as_zoned_time(add_minutes(x, 30), zone)
  )
  expect_identical(
    as_zoned_time(x, zone, nonexistent = "roll-backward"),
    as_zoned_time(add_seconds(add_minutes(x, -30), -1), zone)
  )
  expect_identical(
    as_zoned_time(x, zone, nonexistent = "shift-forward"),
    as_zoned_time(add_minutes(x, 60), zone)
  )
  expect_identical(
    as_zoned_time(x, zone, nonexistent = "shift-backward"),
    as_zoned_time(add_minutes(x, -60), zone)
  )
  expect_identical(
    as_zoned_time(x, zone, nonexistent = "NA"),
    as_zoned_time(naive_seconds(NA), zone)
  )
})

test_that("nonexistent can be vectorized", {
  zone <- "America/New_York"
  x <- as_naive_time(year_month_day(1970, 04, 26, 02, 00, c(00, 00)))

  expect_identical(
    as_zoned_time(x, zone, nonexistent = c("roll-forward", "roll-backward")),
    as_zoned_time(x + duration_seconds(c(3600, -1)), zone)
  )
})

test_that("roll-backward uses precision to determine last moment in time", {
  zone <- "America/New_York"
  w <- as_naive_time(year_month_day(1970, 04, 26, 02, 00, 00))
  x <- w + duration_milliseconds(0)
  y <- w + duration_microseconds(0)
  z <- w + duration_nanoseconds(0)

  expect_identical(
    as_zoned_time(w, zone, nonexistent = "roll-backward"),
    as_zoned_time(add_seconds(w, -1), zone)
  )
  expect_identical(
    as_zoned_time(x, zone, nonexistent = "roll-backward"),
    as_zoned_time(add_milliseconds(x, -1), zone)
  )
  expect_identical(
    as_zoned_time(y, zone, nonexistent = "roll-backward"),
    as_zoned_time(add_microseconds(y, -1), zone)
  )
  expect_identical(
    as_zoned_time(z, zone, nonexistent = "roll-backward"),
    as_zoned_time(add_nanoseconds(z, -1), zone)
  )
})

test_that("names are retained", {
  zone <- "America/New_York"
  x <- as_naive_time(year_month_day(2019, 1, 1))
  x <- c(foo = x)
  expect_named(as_zoned_time(x, zone), "foo")
})

test_that("NA pass through", {
  x <- as_zoned_time(naive_seconds(NA), "America/New_York")
  expect_true(is.na(x))
})

test_that("`ambiguous` is validated", {
  zone <- "America/New_York"
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, ambiguous = 1))
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, ambiguous = "foo"))
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, ambiguous = c("earliest", "latest")))

  ambiguous <- as_zoned_time(naive_seconds(), "America/Los_Angeles")
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, ambiguous = ambiguous))

  reference <- as_zoned_time(naive_seconds(), zone)
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, ambiguous = list()))
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, ambiguous = list(1, 1)))
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, ambiguous = list(reference, 1)))
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, ambiguous = list(reference, "foo")))
})

test_that("`nonexistent` is validated", {
  zone <- "America/New_York"
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, nonexistent = 1))
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, nonexistent = "foo"))
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, nonexistent = c("roll-forward", "roll-forward")))
})

test_that("zone is validated", {
  expect_snapshot_error(as_zoned_time(naive_seconds(), "foo"))
  expect_snapshot_error(as_zoned_time(naive_seconds(), 1))
  expect_snapshot_error(as_zoned_time(naive_seconds(), c("America/New_York", "EST", "EDT")))
})

test_that("strict mode can be activated - nonexistent", {
  local_options(clock.strict = TRUE)
  zone <- "America/New_York"
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, ambiguous = "earliest"))
})

test_that("strict mode can be activated - ambiguous", {
  zone <- "America/New_York"
  zt <- as_zoned_time(naive_seconds(), zone)

  local_options(clock.strict = TRUE)

  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, nonexistent = "roll-forward"))
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, nonexistent = "roll-forward", ambiguous = zt))
  expect_snapshot_error(as_zoned_time(naive_seconds(), zone, nonexistent = "roll-forward", ambiguous = list(zt, NULL)))
})

test_that("empty dots are checked", {
  expect_snapshot_error(as_zoned_time(naive_seconds(), "UTC", "roll-forward"))
})

# ------------------------------------------------------------------------------
# vec_ptype()

test_that("ptype is correct", {
  base <- naive_days(0)
  ptype <- naive_days(integer())

  for (precision in precision_names()) {
    if (validate_precision_string(precision) < PRECISION_DAY) {
      next
    }

    x <- time_point_cast(base, precision)
    expect <- time_point_cast(ptype, precision)

    expect_identical(vec_ptype(x), expect)
  }
})

# ------------------------------------------------------------------------------
# vec_math()

test_that("is.nan() works", {
  x <- naive_days(c(2019, NA))
  expect_identical(is.nan(x), c(FALSE, FALSE))
})

test_that("is.finite() works", {
  x <- naive_days(c(2019, NA))
  expect_identical(is.finite(x), c(TRUE, FALSE))
})

test_that("is.infinite() works", {
  x <- naive_days(c(2019, NA))
  expect_identical(is.infinite(x), c(FALSE, FALSE))
})

#' Invalid calendar dates
#'
#' @description
#' This family of functions is for working with _invalid_ calendar dates.
#'
#' Invalid dates represent dates made up of valid individual components, which
#' taken as a whole don't represent valid calendar dates. For example, for
#' [year_month_day()] the following component ranges are valid:
#' `year: [-32767, 32767]`, `month: [1, 12]`, `day: [1, 31]`.
#' However, the date `2019-02-31` doesn't exist even though it is made up
#' of valid components. This is an example of an invalid date.
#'
#' Invalid dates are allowed in clock, provided that they are eventually
#' resolved by using `invalid_resolve()` or by manually resolving them through
#' arithmetic or setter functions.
#'
#' @details
#' Invalid dates must be resolved before converting them to a time point.
#'
#' It is recommended to use `"previous"` or `"next"` for resolving invalid
#' dates, as these ensure that _relative ordering_ among `x` is maintained.
#' This is a often a very important property to maintain when doing time series
#' data analysis. See the examples for more information.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[calendar]`
#'
#'   A calendar vector.
#'
#' @param invalid `[character(1) / NULL]`
#'
#'   One of the following invalid date resolution strategies:
#'
#'   - `"previous"`: The previous valid instant in time.
#'
#'   - `"previous-day"`: The previous valid day in time, keeping the time of
#'     day.
#'
#'   - `"next"`: The next valid instant in time.
#'
#'   - `"next-day"`: The next valid day in time, keeping the time of day.
#'
#'   - `"overflow"`: Overflow by the number of days that the input is invalid
#'     by. Time of day is dropped.
#'
#'   - `"overflow-day"`: Overflow by the number of days that the input is
#'     invalid by. Time of day is kept.
#'
#'   - `"NA"`: Replace invalid dates with `NA`.
#'
#'   - `"error"`: Error on invalid dates.
#'
#'   Using either `"previous"` or `"next"` is generally recommended, as these
#'   two strategies maintain the _relative ordering_ between elements of the
#'   input.
#'
#'   If `NULL`, defaults to `"error"`.
#'
#'   If `getOption("clock.strict")` is `TRUE`, `invalid` must be supplied and
#'   cannot be `NULL`. This is a convenient way to make production code robust
#'   to invalid dates.
#'
#' @return
#' - `invalid_detect()`: Returns a logical vector detecting invalid dates.
#'
#' - `invalid_any()`: Returns `TRUE` if any invalid dates are detected.
#'
#' - `invalid_count()`: Returns a single integer containing the number of
#'   invalid dates.
#'
#' - `invalid_remove()`: Returns `x` with invalid dates removed.
#'
#' - `invalid_resolve()`: Returns `x` with invalid dates resolved using the
#'   `invalid` strategy.
#'
#' @name clock-invalid
#' @examples
#' # Invalid date
#' x <- year_month_day(2019, 04, 30:31, c(3, 2), 30, 00)
#' x
#'
#' invalid_detect(x)
#'
#' # Previous valid moment in time
#' x_previous <- invalid_resolve(x, invalid = "previous")
#' x_previous
#'
#' # Previous valid day, retaining time of day
#' x_previous_day <- invalid_resolve(x, invalid = "previous-day")
#' x_previous_day
#'
#' # Note that `"previous"` retains the relative ordering in `x`
#' x[1] < x[2]
#' x_previous[1] < x_previous[2]
#'
#' # But `"previous-day"` here does not!
#' x_previous_day[1] < x_previous_day[2]
#'
#' # Remove invalid dates entirely
#' invalid_remove(x)
#'
#' y <- year_quarter_day(2019, 1, 90:92)
#' y
#'
#' # Overflow rolls forward by the number of days between `y` and the previous
#' # valid date
#' invalid_resolve(y, invalid = "overflow")
NULL

# ------------------------------------------------------------------------------

#' @rdname clock-invalid
#' @export
invalid_detect <- function(x) {
  UseMethod("invalid_detect")
}

#' @export
invalid_detect.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("invalid_detect")
}

# ------------------------------------------------------------------------------

#' @rdname clock-invalid
#' @export
invalid_any <- function(x) {
  UseMethod("invalid_any")
}

#' @export
invalid_any.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("invalid_any")
}

# ------------------------------------------------------------------------------

#' @rdname clock-invalid
#' @export
invalid_count <- function(x) {
  UseMethod("invalid_count")
}

#' @export
invalid_count.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("invalid_count")
}

# ------------------------------------------------------------------------------

#' @rdname clock-invalid
#' @export
invalid_remove <- function(x) {
  UseMethod("invalid_remove")
}

#' @export
invalid_remove.clock_calendar <- function(x) {
  x[!invalid_detect(x)]
}

# ------------------------------------------------------------------------------

#' @rdname clock-invalid
#' @export
invalid_resolve <- function(x, ..., invalid = NULL) {
  UseMethod("invalid_resolve")
}

#' @export
invalid_resolve.clock_calendar <- function(x, ..., invalid = NULL) {
  stop_clock_unsupported_calendar_op("invalid_resolve")
}

validate_invalid <- function(invalid) {
  invalid <- strict_validate_invalid(invalid)

  if (!is_string(invalid)) {
    abort("`invalid` must be a character vector with length 1.")
  }

  invalid
}

#' Construct a duration
#'
#' @description
#' These helpers construct durations of the specified precision. Durations
#' represent units of time.
#'
#' Durations are separated into two categories:
#'
#' **Calendrical**
#'
#' - year
#' - quarter
#' - month
#'
#' **Chronological**
#'
#' - week
#' - day
#' - hour
#' - minute
#' - second
#' - millisecond
#' - microsecond
#' - nanosecond
#'
#' Calendrical durations are generally used when manipulating calendar types,
#' like year-month-day. Chronological durations are generally used when
#' working with time points, like sys-time or naive-time.
#'
#' @section Internal Representation:
#'
#' Durations are internally represented as an integer number of "ticks" along
#' with a ratio describing how it converts to a number of seconds. The
#' following duration ratios are used in clock:
#'
#' - `1 year == 31556952 seconds`
#'
#' - `1 quarter == 7889238 seconds`
#'
#' - `1 month == 2629746 seconds`
#'
#' - `1 week == 604800 seconds`
#'
#' - `1 day == 86400 seconds`
#'
#' - `1 hour == 3600 seconds`
#'
#' - `1 minute == 60 seconds`
#'
#' - `1 second == 1 second`
#'
#' - `1 millisecond == 1 / 1000 seconds`
#'
#' - `1 microsecond == 1 / 1000000 seconds`
#'
#' - `1 nanosecond == 1 / 1000000000 seconds`
#'
#' A duration of 1 year is defined to correspond to the
#' average length of a proleptic Gregorian year, i.e. 365.2425 days.
#'
#' A duration of 1 month is defined as exactly 1/12 of a year.
#'
#' A duration of 1 quarter is defined as exactly 1/4 of a year.
#'
#' A duration of 1 week is defined as exactly 7 days.
#'
#' These conversions come into play when doing operations like adding or
#' flooring durations. Generally, you add two calendrical durations together
#' to get a new calendrical duration, rather than adding a calendrical and
#' a chronological duration together. The one exception is [duration_cast()],
#' which can cast durations to any other precision, with a potential loss of
#' information.
#'
#' @param n `[integer]`
#'
#'   The number of units of time to use when creating the duration.
#'
#' @return A duration of the specified precision.
#'
#' @name duration-helper
#'
#' @examples
#' duration_years(1:5)
#' duration_nanoseconds(1:5)
NULL

#' @rdname duration-helper
#' @export
duration_years <- function(n = integer()) {
  duration_helper(n, PRECISION_YEAR)
}
#' @rdname duration-helper
#' @export
duration_quarters <- function(n = integer()) {
  duration_helper(n, PRECISION_QUARTER)
}
#' @rdname duration-helper
#' @export
duration_months <- function(n = integer()) {
  duration_helper(n, PRECISION_MONTH)
}
#' @rdname duration-helper
#' @export
duration_weeks <- function(n = integer()) {
  duration_helper(n, PRECISION_WEEK)
}
#' @rdname duration-helper
#' @export
duration_days <- function(n = integer()) {
  duration_helper(n, PRECISION_DAY)
}
#' @rdname duration-helper
#' @export
duration_hours <- function(n = integer()) {
  duration_helper(n, PRECISION_HOUR)
}
#' @rdname duration-helper
#' @export
duration_minutes <- function(n = integer()) {
  duration_helper(n, PRECISION_MINUTE)
}
#' @rdname duration-helper
#' @export
duration_seconds <- function(n = integer()) {
  duration_helper(n, PRECISION_SECOND)
}
#' @rdname duration-helper
#' @export
duration_milliseconds <- function(n = integer()) {
  duration_helper(n, PRECISION_MILLISECOND)
}
#' @rdname duration-helper
#' @export
duration_microseconds <- function(n = integer()) {
  duration_helper(n, PRECISION_MICROSECOND)
}
#' @rdname duration-helper
#' @export
duration_nanoseconds <- function(n = integer()) {
  duration_helper(n, PRECISION_NANOSECOND)
}

duration_helper <- function(n, precision, ..., retain_names = FALSE, n_arg = "n") {
  check_dots_empty()

  # Generally don't retain names for helpers like `duration_years()`,
  # but might need to during arithmetic
  if (retain_names) {
    names <- names(n)
  } else {
    names <- NULL
  }

  n <- vec_cast(n, integer(), x_arg = n_arg)
  fields <- duration_helper_cpp(n, precision)

  new_duration_from_fields(fields, precision, names)
}

# ------------------------------------------------------------------------------

#' @export
format.clock_duration <- function(x, ...) {
  out <- format_duration_cpp(x, duration_precision_attribute(x))
  names(out) <- names(x)
  out
}

#' @export
obj_print_data.clock_duration <- function(x, ...) {
  if (vec_size(x) == 0L) {
    return(invisible(x))
  }

  out <- format(x)
  print(out, quote = FALSE, na.print = "NA")

  invisible(x)
}

#' @export
vec_ptype_full.clock_duration <- function(x, ...) {
  precision <- duration_precision_attribute(x)
  precision <- precision_to_string(precision)
  paste0("duration<", precision, ">")
}

#' @export
vec_ptype_abbr.clock_duration <- function(x, ...) {
  precision <- duration_precision_attribute(x)
  precision <- precision_to_string(precision)
  precision <- precision_abbr(precision)
  paste0("dur<", precision, ">")
}

# ------------------------------------------------------------------------------

#' @export
vec_proxy.clock_duration <- function(x, ...) {
  .Call(`_clock_clock_rcrd_proxy`, x)
}

#' @export
vec_restore.clock_duration <- function(x, to, ...) {
  .Call(`_clock_duration_restore`, x, to)
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype2.clock_duration.clock_duration <- function(x, y, ...) {
  x_precision <- duration_precision_attribute(x)
  y_precision <- duration_precision_attribute(y)

  if (x_precision == y_precision) {
    return(x)
  }

  precision <- duration_precision_common_cpp(x_precision, y_precision)

  if (is.na(precision)) {
    stop_incompatible_type(
      x = x,
      y = y,
      ...,
      details = "Can't combine calendrical durations with chronological durations."
    )
  }

  if (precision == x_precision) {
    x
  } else {
    y
  }
}

#' @export
vec_cast.clock_duration.clock_duration <- function(x, to, ...) {
  x_precision <- duration_precision_attribute(x)
  to_precision <- duration_precision_attribute(to)

  if (x_precision == to_precision) {
    return(x)
  }

  precision <- duration_precision_common_cpp(x_precision, to_precision)

  if (is.na(precision)) {
    stop_incompatible_cast(
      x = x,
      to = to,
      ...,
      details = "Can't cast between calendrical durations and chronological durations."
    )
  }

  if (x_precision > to_precision) {
    stop_incompatible_cast(
      x = x,
      to = to,
      ...,
      details = "Can't cast to a less precise precision."
    )
  }

  fields <- duration_cast_cpp(x, x_precision, to_precision)

  new_duration_from_fields(fields, to_precision, clock_rcrd_names(x))
}

# ------------------------------------------------------------------------------

#' Convert to a duration
#'
#' @description
#' You generally convert to a duration from either a sys-time or a naive-time.
#' The precision of the input is retained in the returned duration.
#'
#' To round an existing duration to another precision, see [duration_floor()].
#'
#' @param x `[object]`
#'
#'   An object to convert to a duration.
#'
#' @return A duration with the same precision as `x`.
#'
#' @export
#' @examples
#' x <- as_sys_time(year_month_day(2019, 01, 01))
#'
#' # The number of days since 1970-01-01 UTC
#' as_duration(x)
#'
#' x <- x + duration_seconds(1)
#' x
#'
#' # The number of seconds since 1970-01-01 00:00:00 UTC
#' as_duration(x)
as_duration <- function(x) {
  UseMethod("as_duration")
}

#' @export
as_duration.clock_duration <- function(x) {
  x
}

# ------------------------------------------------------------------------------

#' @export
as_sys_time.clock_duration <- function(x) {
  names <- clock_rcrd_names(x)

  # Promote to at least day precision for sys-time
  ptype <- vec_ptype2(x, duration_days(), y_arg = "")
  x <- vec_cast(x, ptype)

  precision <- duration_precision_attribute(x)

  new_sys_time_from_fields(x, precision, names)
}

#' @export
as_naive_time.clock_duration <- function(x) {
  as_naive_time(as_sys_time(x))
}

#' @export
as.integer.clock_duration <- function(x, ...) {
  out <- duration_as_integer_cpp(x, duration_precision_attribute(x))
  names(out) <- names(x)
  out
}

#' @export
as.double.clock_duration <- function(x, ...) {
  out <- duration_as_double_cpp(x, duration_precision_attribute(x))
  names(out) <- names(x)
  out
}

# ------------------------------------------------------------------------------

#' Cast a duration between precisions
#'
#' @description
#' Casting is one way to change a duration's precision.
#'
#' Casting to a less precise precision will completely drop information that
#' is more precise than the precision that you are casting to. It does so
#' in a way that makes it round towards zero.
#'
#' Casting to a more precise precision is done through a multiplication by
#' a conversion factor between the current precision and the new precision.
#'
#' @details
#' When you want to change to a less precise precision, you often want
#' [duration_floor()] instead of `duration_cast()`, as that rounds towards
#' negative infinity, which is generally the desired behavior when working with
#' time points (especially ones pre-1970, which are stored as negative
#' durations).
#'
#' @param x `[clock_duration]`
#'
#'   A duration.
#'
#' @param precision `[character(1)]`
#'
#'   A precision. One of:
#'
#'   - `"year"`
#'
#'   - `"quarter"`
#'
#'   - `"month"`
#'
#'   - `"week"`
#'
#'   - `"day"`
#'
#'   - `"hour"`
#'
#'   - `"minute"`
#'
#'   - `"second"`
#'
#'   - `"millisecond"`
#'
#'   - `"microsecond"`
#'
#'   - `"nanosecond"`
#'
#' @return `x` cast to the new `precision`.
#'
#' @export
#' @examples
#' x <- duration_seconds(c(86401, -86401))
#'
#' # Casting rounds towards 0
#' cast <- duration_cast(x, "day")
#' cast
#'
#' # Flooring rounds towards negative infinity
#' floor <- duration_floor(x, "day")
#' floor
#'
#' # Flooring is generally more useful when working with time points,
#' # note that the cast ends up rounding the pre-1970 date up to the next
#' # day, while the post-1970 date is rounded down.
#' as_sys_time(x)
#' as_sys_time(cast)
#' as_sys_time(floor)
#'
#' # Casting to a more precise precision
#' duration_cast(x, "millisecond")
duration_cast <- function(x, precision) {
  if (!is_duration(x)) {
    abort("`x` must be a duration.")
  }
  x_precision <- duration_precision_attribute(x)
  precision <- validate_precision_string(precision)
  fields <- duration_cast_cpp(x, x_precision, precision)
  new_duration_from_fields(fields, precision, clock_rcrd_names(x))
}

# ------------------------------------------------------------------------------

#' Duration rounding
#'
#' @description
#' - `duration_floor()` rounds a duration down to a multiple of the specified
#'   `precision`.
#'
#' - `duration_ceiling()` rounds a duration up to a multiple of the specified
#'   `precision`.
#'
#' - `duration_round()` rounds up or down depending on what is closer,
#'   rounding up on ties.
#'
#' @details
#' You can floor calendrical durations to other calendrical durations, and
#' chronological durations to other chronological durations, but you can't
#' floor a chronological duration to a calendrical duration (such as flooring
#' from day to month). For more information, see the documentation on the
#' [duration helper][duration-helper] page.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams duration_cast
#'
#' @param n `[positive integer(1)]`
#'
#'   A positive integer specifying the multiple of `precision` to use.
#'
#' @return `x` rounded to the `precision`.
#'
#' @name duration-rounding
#'
#' @examples
#' x <- duration_seconds(c(86399, 86401))
#'
#' duration_floor(x, "day")
#' duration_ceiling(x, "day")
#'
#' # Can't floor from a chronological duration (seconds)
#' # to a calendrical duration (months)
#' try(duration_floor(x, "month"))
#'
#' # Every 2 days, using an origin of day 0
#' y <- duration_seconds(c(0, 86400, 86400 * 2, 86400 * 3))
#' duration_floor(y, "day", n = 2)
#'
#' # Shifting the origin to be day 1
#' origin <- duration_days(1)
#' duration_floor(y - origin, "day", n = 2) + origin
#'
#' # Rounding will round ties up
#' half_day <- 86400 / 2
#' half_day_durations <- duration_seconds(c(half_day - 1, half_day, half_day + 1))
#' duration_round(half_day_durations, "day")
#'
#' # With larger units
#' x <- duration_months(c(0, 15, 24))
#' duration_floor(x, "year")
#' duration_floor(x, "quarter")
NULL

#' @rdname duration-rounding
#' @export
duration_floor <- function(x, precision, ..., n = 1L) {
  duration_rounder(x, precision, n, duration_floor_cpp, "floor", ...)
}

#' @rdname duration-rounding
#' @export
duration_ceiling <- function(x, precision, ..., n = 1L) {
  duration_rounder(x, precision, n, duration_ceiling_cpp, "ceiling", ...)
}

#' @rdname duration-rounding
#' @export
duration_round <- function(x, precision, ..., n = 1L) {
  duration_rounder(x, precision, n, duration_round_cpp, "round", ...)
}

duration_rounder <- function(x, precision, n, rounder, verb, ...) {
  check_dots_empty()

  if (!is_duration(x)) {
    abort("`x` must be a duration object.")
  }

  n <- vec_cast(n, integer(), x_arg = "n")
  if (!is_number(n)) {
    abort("`n` must be a single number.")
  }
  if (n <= 0L) {
    abort("`n` must be a positive number.")
  }

  precision <- validate_precision_string(precision)
  x_precision <- duration_precision_attribute(x)

  if (x_precision < precision) {
    abort(paste0("Can't ", verb, " to a more precise precision."))
  }
  if (!duration_has_common_precision_cpp(x_precision, precision)) {
    precision <- precision_to_string(precision)
    x_precision <- precision_to_string(x_precision)
    message <- paste0(
      "Can't ", verb, " from ",
      "a chronological precision (", x_precision, ") to ",
      "a calendrical precision (", precision, ")."
    )
    abort(message)
  }

  fields <- rounder(x, x_precision, precision, n)

  new_duration_from_fields(fields, precision, clock_rcrd_names(x))
}

# ------------------------------------------------------------------------------

#' Sequences: duration
#'
#' @description
#' This is a duration method for the [seq()] generic.
#'
#' Using `seq()` on duration objects always retains the type of `from`.
#'
#' When calling `seq()`, exactly two of the following must be specified:
#' - `to`
#' - `by`
#' - Either `length.out` or `along.with`
#'
#' @details
#' If `from > to` and `by > 0`, then the result will be length 0. This matches
#' the behavior of [rlang::seq2()], and results in nicer theoretical
#' properties when compared with throwing an error. Similarly, if `from < to`
#' and `by < 0`, then the result will also be length 0.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param from `[clock_duration(1)]`
#'
#'   A duration to start the sequence from.
#'
#' @param to `[clock_duration(1) / NULL]`
#'
#'   A duration to stop the sequence at.
#'
#'   `to` is cast to the type of `from`.
#'
#'   `to` is only included in the result if the resulting sequence divides
#'   the distance between `from` and `to` exactly.
#'
#' @param by `[integer(1) / clock_duration(1) / NULL]`
#'
#'   The unit to increment the sequence by.
#'
#'   If `by` is an integer, it is transformed into a duration with the
#'   precision of `from`.
#'
#'   If `by` is a duration, it is cast to the type of `from`.
#'
#' @param length.out `[positive integer(1) / NULL]`
#'
#'   The length of the resulting sequence.
#'
#'   If specified, `along.with` must be `NULL`.
#'
#' @param along.with `[vector / NULL]`
#'
#'   A vector who's length determines the length of the resulting sequence.
#'
#'   Equivalent to `length.out = vec_size(along.with)`.
#'
#'   If specified, `length.out` must be `NULL`.
#'
#' @return A sequence with the type of `from`.
#'
#' @export
#' @examples
#' seq(duration_days(0), duration_days(100), by = 5)
#'
#' # Using a duration `by`. Note that `by` is cast to the type of `from`.
#' seq(duration_days(0), duration_days(100), by = duration_weeks(1))
#'
#' # `to` is cast from 5 years to 60 months
#' # `by` is cast from 1 quarter to 4 months
#' seq(duration_months(0), duration_years(5), by = duration_quarters(1))
#'
#' seq(duration_days(20), by = 2, length.out = 5)
seq.clock_duration <- function(from,
                               to = NULL,
                               by = NULL,
                               length.out = NULL,
                               along.with = NULL,
                               ...) {
  check_dots_empty()

  vec_assert(from, size = 1L)
  if (is.na(from)) {
    abort("`from` can't be `NA`.")
  }

  has_to <- !is_null(to)
  has_by <- !is_null(by)
  has_lo <- !is_null(length.out)
  has_aw <- !is_null(along.with)

  if (has_aw) {
    if (has_lo) {
      abort("Can only specify one of `length.out` and `along.with`.")
    } else {
      has_lo <- TRUE
      length.out <- vec_size(along.with)
    }
  }

  n_has <- sum(has_to, has_by, has_lo)

  if (n_has != 2L) {
    message <- paste0(
      "Must specify exactly two of:\n",
      "- `to`\n",
      "- `by`\n",
      "- Either `length.out` or `along.with`"
    )
    abort(message)
  }

  if (has_to) {
    to <- vec_cast(to, from, x_arg = "to", to_arg = "from")

    vec_assert(to, size = 1L, arg = "to")
    if (is.na(to)) {
      abort("`to` can't be `NA`.")
    }
  }

  if (has_by) {
    precision <- duration_precision_attribute(from)
    by <- duration_collect_by(by, precision)

    vec_assert(by, size = 1L, arg = "by")
    if (is.na(by)) {
      abort("`by` can't be `NA`.")
    }
    if (by == duration_helper(0L, precision)) {
      abort("`by` can't be `0`.")
    }
  }

  if (has_lo) {
    length.out <- check_length_out(length.out, arg = "length.out")
  }

  if (has_to) {
    if (has_by) {
      duration_seq_to_by(from, to, by)
    } else {
      duration_seq_to_lo(from, to, length.out)
    }
  } else {
    duration_seq_by_lo(from, by, length.out)
  }
}

duration_seq_to_by <- function(from, to, by) {
  names <- NULL
  precision <- duration_precision_attribute(from)
  fields <- duration_seq_to_by_cpp(from, precision, to, by)
  new_duration_from_fields(fields, precision, names)
}

duration_seq_to_lo <- function(from, to, length.out) {
  names <- NULL
  precision <- duration_precision_attribute(from)
  fields <- duration_seq_to_lo_cpp(from, precision, to, length.out)
  new_duration_from_fields(fields, precision, names)
}

duration_seq_by_lo <- function(from, by, length.out) {
  names <- NULL
  precision <- duration_precision_attribute(from)
  fields <- duration_seq_by_lo_cpp(from, precision, by, length.out)
  new_duration_from_fields(fields, precision, names)
}

duration_collect_by <- function(by, precision) {
  if (is_duration(by)) {
    to <- duration_helper(integer(), precision)
    vec_cast(by, to, x_arg = "by")
  } else {
    duration_helper(by, precision, n_arg = "by")
  }
}

check_length_out <- function(length.out, arg) {
  length.out <- vec_cast(length.out, integer(), x_arg = arg)

  vec_assert(length.out, size = 1L, arg = arg)

  if (is.na(length.out)) {
    abort(paste0("`", arg, "` can't be `NA`."))
  }

  if (length.out < 0) {
    abort(paste0("`", arg, "` can't be negative."))
  }

  length.out
}

# Used by other methods that eventually call the duration seq() method
seq_impl <- function(from, to, by, length.out, along.with, precision, ...) {
  if (!is_null(to)) {
    to <- vec_cast(to, from, x_arg = "to", to_arg = "from")
    to <- to - from
  }

  start <- from
  from <- duration_helper(0L, precision)

  steps <- seq(
    from = from,
    to = to,
    by = by,
    length.out = length.out,
    along.with = along.with,
    ...
  )

  start + steps
}

# ------------------------------------------------------------------------------

#' @export
vec_math.clock_duration <- function(.fn, .x, ...) {
  switch(
    .fn,
    abs = duration_abs(.x),
    sign = duration_sign(.x),
    # Pass on to `vec_math.clock_rcrd()`
    NextMethod()
  )
}

duration_abs <- function(x) {
  precision <- duration_precision_attribute(x)
  fields <- duration_abs_cpp(x, precision)
  new_duration_from_fields(fields, precision, clock_rcrd_names(x))
}

duration_sign <- function(x) {
  out <- duration_sign_cpp(x, duration_precision_attribute(x))
  names(out) <- names(x)
  out
}

# ------------------------------------------------------------------------------

#' @export
#' @method vec_arith clock_duration
vec_arith.clock_duration <- function(op, x, y, ...) {
  UseMethod("vec_arith.clock_duration", y)
}

#' @export
#' @method vec_arith.clock_duration MISSING
vec_arith.clock_duration.MISSING <- function(op, x, y, ...) {
  switch(
    op,
    "+" = x,
    "-" = duration_unary_minus(x),
  )
}

#' @export
#' @method vec_arith.clock_duration clock_duration
vec_arith.clock_duration.clock_duration <- function(op, x, y, ...) {
  switch(
    op,
    "+" = duration_plus(x, y, names_common(x, y)),
    "-" = duration_minus(x, y, names_common(x, y)),
    "/" = stop_incompatible_op(op, x, y, details = "Durations only support integer division. Did you want `%/%`?"),
    "%%" = duration_modulus(x, y, names_common(x, y)),
    "%/%" = duration_integer_divide(x, y, names_common(x, y)),
    stop_incompatible_op(op, x, y)
  )
}

#' @export
#' @method vec_arith.clock_duration numeric
vec_arith.clock_duration.numeric <- function(op, x, y, ...) {
  switch(
    op,
    "*" = duration_scalar_multiply(x, y, names_common(x, y)),
    "/" = stop_incompatible_op(op, x, y, details = "Durations only support integer division. Did you want `%/%`?"),
    "%%" = duration_scalar_modulus(x, y, names_common(x, y)),
    "%/%" = duration_scalar_divide(x, y, names_common(x, y)),
    stop_incompatible_op(op, x, y)
  )
}

#' @export
#' @method vec_arith.numeric clock_duration
vec_arith.numeric.clock_duration <- function(op, x, y, ...) {
  switch(
    op,
    "*" = duration_scalar_multiply(y, x, names_common(x, y)),
    "/" = stop_incompatible_op(op, x, y, details = "Durations only support integer division. Did you want `%/%`?"),
    "%/%" = stop_incompatible_op(op, x, y, details = "Cannot divide a numeric by a duration."),
    stop_incompatible_op(op, x, y)
  )
}

# ------------------------------------------------------------------------------

#' Arithmetic: duration
#'
#' @description
#' These are duration methods for the
#' [arithmetic generics][clock-arithmetic].
#'
#' - `add_years()`
#'
#' - `add_quarters()`
#'
#' - `add_months()`
#'
#' - `add_weeks()`
#'
#' - `add_days()`
#'
#' - `add_hours()`
#'
#' - `add_minutes()`
#'
#' - `add_seconds()`
#'
#' - `add_milliseconds()`
#'
#' - `add_microseconds()`
#'
#' - `add_nanoseconds()`
#'
#' When adding to a duration using one of these functions, a second duration
#' is created based on the function name and `n`. The two durations are then
#' added together, and the precision of the result is determined as the
#' _more precise precision_ of the two durations.
#'
#' @details
#' You can add calendrical durations to other calendrical durations,
#' and chronological durations to other chronological durations, but you can't
#' add a chronological duration to a calendrical duration
#' (such as adding days and months). For more information, see the
#' documentation on the [duration helper][duration-helper] page.
#'
#' `x` and `n` are recycled against each other.
#'
#' @inheritParams add_years
#'
#' @param x `[clock_duration]`
#'
#'   A duration vector.
#'
#' @return `x` after performing the arithmetic, possibly with a more precise
#'   precision.
#'
#' @name duration-arithmetic
#'
#' @examples
#' x <- duration_seconds(5)
#'
#' # Addition in the same precision
#' add_seconds(x, 1:10)
#'
#' # Addition with days, defined as 86400 seconds
#' add_days(x, 1)
#'
#' # Similarly, if you start with days and add seconds, you get the common
#' # precision of the two back, which is seconds
#' y <- duration_days(1)
#' add_seconds(y, 5)
#'
#' # But you can't add a chronological duration (days) and
#' # a calendrical duration (months)
#' try(add_months(y, 1))
#'
#' # You can add years to a duration of months, which adds
#' # an additional 12 months / year
#' z <- duration_months(5)
#' add_years(z, 1)
NULL

#' @rdname duration-arithmetic
#' @export
add_years.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_YEAR)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_quarters.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_QUARTER)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_months.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_MONTH)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_weeks.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_WEEK)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_days.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_DAY)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_hours.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_HOUR)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_minutes.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_MINUTE)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_seconds.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_SECOND)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_milliseconds.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_MILLISECOND)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_microseconds.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_MICROSECOND)
  duration_plus(x, n, names_common(x, n))
}
#' @rdname duration-arithmetic
#' @export
add_nanoseconds.clock_duration <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_NANOSECOND)
  duration_plus(x, n, names_common(x, n))
}

duration_collect_n <- function(n, precision) {
  if (is_duration_with_precision(n, precision)) {
    return(n)
  }

  if (is_duration(n)) {
    precision_string <- precision_to_string(precision)
    abort(paste0("`n` must have '", precision_string, "' precision."))
  }

  duration_helper(n, precision, retain_names = TRUE)
}

# ------------------------------------------------------------------------------

duration_plus <- function(x, y, names) {
  duration_arith(x, y, names, duration_plus_cpp)
}
duration_minus <- function(x, y, names) {
  duration_arith(x, y, names, duration_minus_cpp)
}
duration_modulus <- function(x, y, names) {
  duration_arith(x, y, names, duration_modulus_cpp)
}

duration_arith <- function(x, y, names, fn) {
  if (!is_duration(x) || !is_duration(y)) {
    abort("`x` and `y` must both be 'duration' objects.")
  }

  args <- vec_cast_common(x = x, y = y)
  args <- vec_recycle_common(!!!args, names = names)
  x <- args$x
  y <- args$y
  names <- args$names

  precision <- duration_precision_attribute(x)

  fields <- fn(x, y, precision)

  new_duration_from_fields(fields, precision, names)
}

duration_integer_divide <- function(x, y, names) {
  duration_arith_integer(x, y, names, duration_integer_divide_cpp)
}

duration_arith_integer <- function(x, y, names, fn) {
  if (!is_duration(x) || !is_duration(y)) {
    abort("`x` and `y` must both be 'duration' objects.")
  }

  args <- vec_cast_common(x = x, y = y)
  args <- vec_recycle_common(!!!args, names = names)
  x <- args$x
  y <- args$y
  names <- args$names

  precision <- duration_precision_attribute(x)

  out <- fn(x, y, precision)

  names(out) <- names

  out
}

duration_scalar_multiply <- function(x, y, names) {
  duration_scalar_arith(x, y, names, duration_scalar_multiply_cpp)
}
duration_scalar_modulus <- function(x, y, names) {
  duration_scalar_arith(x, y, names, duration_scalar_modulus_cpp)
}
duration_scalar_divide <- function(x, y, names) {
  duration_scalar_arith(x, y, names, duration_scalar_divide_cpp)
}

duration_scalar_arith <- function(x, y, names, fn) {
  if (!is_duration(x)) {
    abort("`x` must be a 'duration' object.")
  }

  precision <- duration_precision_attribute(x)

  y <- vec_cast(y, integer(), x_arg = "y")

  args <- vec_recycle_common(x = x, y = y, names = names)
  x <- args$x
  y <- args$y
  names <- args$names

  fields <- fn(x, y, precision)

  new_duration_from_fields(fields, precision, names)
}

names_common <- function(x, y) {
  names <- names(x)
  if (is_null(names)) {
    names <- names(y)
  }
  names
}

duration_unary_minus <- function(x) {
  precision <- duration_precision_attribute(x)
  fields <- duration_unary_minus_cpp(x, precision)
  new_duration_from_fields(fields, precision, names(x))
}

# ------------------------------------------------------------------------------

#' Is `x` a duration?
#'
#' This function determines if the input is a duration object.
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @return `TRUE` if `x` inherits from `"clock_duration"`, otherwise `FALSE`.
#'
#' @export
#' @examples
#' is_duration(1)
#' is_duration(duration_days(1))
is_duration <- function(x) {
  inherits(x, "clock_duration")
}

is_duration_with_precision <- function(x, precision) {
  is_duration(x) && duration_precision_attribute(x) == precision
}

# ------------------------------------------------------------------------------

#' Precision: duration
#'
#' `duration_precision()` extracts the precision from a duration object. It
#' returns the precision as a single string.
#'
#' @param x `[clock_duration]`
#'
#'   A duration.
#'
#' @return A single string holding the precision of the duration.
#'
#' @export
#' @examples
#' duration_precision(duration_seconds(1))
#' duration_precision(duration_nanoseconds(2))
#' duration_precision(duration_quarters(1:5))
duration_precision <- function(x) {
  if (!is_duration(x)) {
    abort("`x` must be a 'clock_duration'.")
  }
  precision <- duration_precision_attribute(x)
  precision <- precision_to_string(precision)
  precision
}

# ------------------------------------------------------------------------------

duration_precision_attribute <- function(x) {
  attr(x, "precision", exact = TRUE)
}

# ------------------------------------------------------------------------------

field_ticks <- function(x) {
  # The `ticks` field is the first field of every
  # duration / time point / zoned time type, which makes it the field that
  # names are on. When extracting the field, we don't ever
  # want the names to come with it.
  out <- field(x, "ticks")
  names(out) <- NULL
  out
}
field_ticks_of_day <- function(x) {
  field(x, "ticks_of_day")
}
field_ticks_of_second <- function(x) {
  field(x, "ticks_of_second")
}

# ------------------------------------------------------------------------------
# duration_precision_common_cpp()

test_that("correctly computes common duration precision", {
  granular <- c(
    PRECISION_YEAR,
    PRECISION_QUARTER,
    PRECISION_MONTH
  )

  precise <- c(
    PRECISION_WEEK,
    PRECISION_DAY,
    PRECISION_HOUR,
    PRECISION_MINUTE,
    PRECISION_SECOND,
    PRECISION_MILLISECOND,
    PRECISION_MICROSECOND,
    PRECISION_NANOSECOND
  )

  for (p1 in granular) {
    for (p2 in granular) {
      expect_identical(duration_precision_common_cpp(p1, p2), max(p1, p2))
    }
  }

  for (p1 in precise) {
    for (p2 in precise) {
      expect_identical(duration_precision_common_cpp(p1, p2), max(p1, p2))
    }
  }

  for (p1 in granular) {
    for (p2 in precise) {
      expect_identical(duration_precision_common_cpp(p1, p2), NA_integer_)
    }
  }
})

# ------------------------------------------------------------------------------
# duration_floor() / _ceiling() / _round()

test_that("floor rounds down", {
  x <- duration_days(2) + duration_seconds(-1:1)
  x <- c(-x, x)

  expect2 <- duration_seconds(c(-172800, -172800, -172802, 172798, 172800, 172800))
  expect3 <- duration_days(c(-2, -2, -3, 1, 2, 2))
  expect4 <- duration_days(c(-2, -2, -4, 0, 2, 2))

  expect_identical(duration_floor(x, "second"), x)
  expect_identical(duration_floor(x, "second", n = 2), expect2)
  expect_identical(duration_floor(x, "day"), expect3)
  expect_identical(duration_floor(x, "day", n = 2), expect4)
})

test_that("ceiling rounds up", {
  x <- duration_days(2) + duration_seconds(-1:1)
  x <- c(-x, x)

  expect2 <- duration_seconds(c(-172798, -172800, -172800, 172800, 172800, 172802))
  expect3 <- duration_days(c(-1, -2, -2, 2, 2, 3))
  expect4 <- duration_days(c(0, -2, -2, 2, 2, 4))

  expect_identical(duration_ceiling(x, "second"), x)
  expect_identical(duration_ceiling(x, "second", n = 2), expect2)
  expect_identical(duration_ceiling(x, "day"), expect3)
  expect_identical(duration_ceiling(x, "day", n = 2), expect4)
})

test_that("round rounds to nearest, ties round up", {
  x <- duration_days(2) + duration_seconds(-1:3)
  x <- c(-x, x)

  expect2 <- duration_seconds(c(-172800, -172800, -172800, -172800, -172804, 172800, 172800, 172800, 172804, 172804))
  expect3 <- duration_days(c(-2, -2, -2, -2, -2, 2, 2, 2, 2, 2))
  expect4 <- duration_days(c(0, 0, -4, -4, -4, 0, 4, 4, 4, 4))

  expect_identical(duration_round(x, "second"), x)
  expect_identical(duration_round(x, "second", n = 4), expect2)
  expect_identical(duration_round(x, "day"), expect3)
  expect_identical(duration_round(x, "day", n = 4), expect4)
})

test_that("can't round to more precise precision", {
  expect_error(duration_floor(duration_seconds(1), "millisecond"), "more precise")
})

test_that("can't round across common precision boundary", {
  expect_snapshot_error(duration_ceiling(duration_weeks(), "month"))
  expect_snapshot_error(duration_floor(duration_seconds(), "year"))
})

test_that("input is validated", {
  expect_error(duration_floor(1, "year"), "must be a duration object")
  expect_error(duration_floor(duration_seconds(1), "foo"), "not recognized")
  expect_error(duration_floor(duration_seconds(1), "day", n = -1), "positive number")
})

# ------------------------------------------------------------------------------
# seq()

test_that("seq() validates from", {
  expect_snapshot_error(seq(duration_years(1:2)), class = "vctrs_error_assert_size")
  expect_snapshot_error(seq(duration_years(NA_integer_)))
})

test_that("seq() validates length.out / along.with exclusiveness", {
  expect_snapshot_error(seq(duration_years(1L), length.out = 1, along.with = 2))
})

test_that("seq() only takes two optional args", {
  x <- duration_years(1L)
  expect_snapshot_error(seq(x, to = duration_years(1), by = 1, length.out = 1))
  expect_snapshot_error(seq(x, to = duration_years(1), by = 1, along.with = 1))
})

test_that("seq() requires two optional args", {
  x <- duration_years(1L)
  expect_snapshot_error(seq(x, to = duration_years(1)))
  expect_snapshot_error(seq(x, by = 1))
  expect_snapshot_error(seq(x, length.out = 1))
  expect_snapshot_error(seq(x, along.with = 1))
})

test_that("seq() validates `to`", {
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1:2), by = 1), class = "vctrs_error_assert_size")
  expect_snapshot_error(seq(duration_years(1L), to = 1, by = 1))
  expect_snapshot_error(seq(duration_years(1L), to = duration_days(1), by = 1))
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(NA_integer_), by = 1))
})

test_that("seq() validates `by`", {
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1L), by = 1:2), class = "vctrs_error_assert_size")
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1L), by = NA_integer_))
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1L), by = 0))
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1L), by = duration_years(0)))
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1L), by = "x"), class = "vctrs_error_incompatible_type")
})

test_that("`by` must be castable to the type of `from`", {
  expect_snapshot_error(seq(duration_years(0), to = duration_years(1), by = duration_months(1)))
  expect_snapshot_error(seq(duration_years(0), to = duration_years(1), by = duration_days(1)))
  expect_snapshot_error(seq(duration_days(0), to = duration_days(1), by = duration_years(1)))
})

test_that("seq() validates `length.out`", {
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1L), length.out = 1:2), class = "vctrs_error_assert_size")
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1L), length.out = NA_integer_))
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1L), length.out = -1))
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(1L), length.out = "x"), class = "vctrs_error_incompatible_type")
})

test_that("seq() validates dots", {
  expect_snapshot_error(seq(duration_years(1), duration_years(1), 1, 1, 1, 1))
})

test_that("seq() enforces non-fractional results", {
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(2L), length.out = 3))
  expect_snapshot_error(seq(duration_years(1L), to = duration_years(2L), along.with = 1:3))
})

test_that("seq() works when from and to are identical", {
  expect_identical(seq(duration_years(1L), to = duration_years(1L), by = 1), duration_years(1L))
  expect_identical(seq(duration_years(1L), to = duration_years(1L), by = -1), duration_years(1L))
})

test_that("seq() with `from > to && by > 0` or `from < to && by > 0` results in length 0 output (#282)", {
  expect_identical(seq(duration_years(2L), to = duration_years(1L), by = 1), duration_years())
  expect_identical(seq(duration_years(5L), to = duration_years(1L), by = 1), duration_years())

  expect_identical(seq(duration_years(1L), to = duration_years(2L), by = -1), duration_years())
  expect_identical(seq(duration_years(1L), to = duration_years(5L), by = -1), duration_years())
})

test_that("seq(to, by) works", {
  expect_identical(seq(duration_years(0L), to = duration_years(4L), by = 2), duration_years(c(0L, 2L, 4L)))
  expect_identical(seq(duration_years(0L), to = duration_years(5L), by = 2), duration_years(c(0L, 2L, 4L)))

  expect_identical(seq(duration_years(0L), to = duration_years(-4L), by = -2), duration_years(c(0L, -2L, -4L)))
  expect_identical(seq(duration_years(0L), to = duration_years(-5L), by = -2), duration_years(c(0L, -2L, -4L)))

  expect_identical(seq(duration_years(4L), to = duration_years(0L), by = -2), duration_years(c(4L, 2L, 0L)))
  expect_identical(seq(duration_years(4L), to = duration_years(-1L), by = -2), duration_years(c(4L, 2L, 0L)))
})

test_that("seq(to, by = <duration>) works", {
  expect_identical(
    seq(duration_years(0), to = duration_years(4), by = duration_years(1)),
    seq(duration_years(0), to = duration_years(4), by = 1)
  )
  expect_identical(
    seq(duration_months(0), to = duration_months(20), by = duration_years(1)),
    seq(duration_months(0), to = duration_months(20), by = 12)
  )
  expect_identical(
    seq(duration_seconds(0), to = duration_seconds(1000), by = duration_minutes(2)),
    seq(duration_seconds(0), to = duration_seconds(1000), by = 120)
  )
  expect_identical(
    seq(duration_nanoseconds(0), by = duration_days(2), length.out = 5),
    duration_nanoseconds(0) + duration_days(c(0, 2, 4, 6, 8))
  )
  expect_identical(
    seq(duration_nanoseconds(0), to = duration_days(100000), by = duration_days(10000)),
    duration_nanoseconds(0) + duration_days(seq(0L, 100000L, by = 10000L))
  )
  expect_identical(
    seq(duration_nanoseconds(0), to = -duration_days(100000), by = -duration_days(10000)),
    duration_nanoseconds(0) - duration_days(seq(0L, 100000L, by = 10000L))
  )
})

test_that("seq(to, length.out) works", {
  expect_identical(seq(duration_years(0L), to = duration_years(4L), length.out = 2), duration_years(c(0L, 4L)))
  expect_identical(seq(duration_years(0L), to = duration_years(4L), length.out = 1), duration_years(c(0L)))
  expect_identical(seq(duration_years(0L), to = duration_years(4L), length.out = 5), duration_years(c(0:4)))

  expect_identical(seq(duration_years(0L), to = duration_years(-4L), length.out = 2), duration_years(c(0L, -4L)))
  expect_identical(seq(duration_years(0L), to = duration_years(-6L), length.out = 3), duration_years(c(0L, -3L, -6L)))

  expect_identical(seq(duration_years(0L), to = duration_years(4L), along.with = 1:2), duration_years(c(0L, 4L)))
})

test_that("seq(to, length.out = 1) is special cased to return `from`", {
  expect_identical(
    seq(duration_years(1), duration_years(5), length.out = 1),
    duration_years(1)
  )
})

test_that("seq(by, length.out) works", {
  expect_identical(seq(duration_years(0L), by = 2, length.out = 3), duration_years(c(0L, 2L, 4L)))
  expect_identical(seq(duration_years(0L), by = -2, length.out = 3), duration_years(c(0L, -2L, -4L)))

  expect_identical(seq(duration_years(0L), by = 2, along.with = 1:3), duration_years(c(0L, 2L, 4L)))
})

test_that("`to` is always cast to `from`", {
  expect_identical(
    seq(duration_months(0), to = duration_years(1), by = 2),
    seq(duration_months(0), to = duration_months(12), by = 2)
  )

  expect_snapshot_error(seq(duration_days(0), to = duration_years(5), by = 2))
  expect_snapshot_error(seq(duration_years(0), to = duration_months(5), by = 2))
})

test_that("special test to ensure we never lose precision (i.e. by trying to convert to double)", {
  expect_identical(
    seq(duration_nanoseconds(0), duration_cast(duration_years(10), "nanosecond"), length.out = 3),
    duration_nanoseconds(0) + duration_cast(duration_years(c(0, 5, 10)), "nanosecond")
  )
})

# ------------------------------------------------------------------------------
# add_*()

test_that("can't add chronological and calendrical durations", {
  expect_snapshot_error(add_seconds(duration_years(1), 1))
  expect_snapshot_error(add_years(duration_seconds(1), 1))
})

# ------------------------------------------------------------------------------
# as_sys_time() / as_naive_time()

test_that("can convert week precision duration to time point", {
  expect_identical(as_sys_time(duration_weeks(c(0, 1))), sys_days(c(0, 7)))
  expect_identical(as_naive_time(duration_weeks(c(0, 1))), naive_days(c(0, 7)))
})

test_that("can't convert calendrical duration to time point", {
  expect_snapshot_error(as_sys_time(duration_years(0)))
  expect_snapshot_error(as_naive_time(duration_years(0)))
})

# ------------------------------------------------------------------------------
# duration_precision()

test_that("precision: can get the precision", {
  expect_identical(duration_precision(duration_months(2:5)), "month")
  expect_identical(duration_precision(duration_days(1)), "day")
  expect_identical(duration_precision(duration_nanoseconds(5:6)), "nanosecond")
})

test_that("precision: can only be called on durations", {
  expect_snapshot_error(duration_precision(sys_days(0)))
})

# ------------------------------------------------------------------------------
# vec_arith()

test_that("`<duration> / <duration>` is not allowed", {
  expect_snapshot(
    (expect_error(duration_years(1) / duration_years(2)))
  )
})

test_that("`<duration> %/% <duration>` works", {
  expect_identical(duration_years(5) %/% duration_years(2:3), c(2L, 1L))
  expect_identical(duration_days(10) %/% duration_hours(7), 34L)
})

test_that("`<duration> %/% <duration>` propagates NA", {
  expect_identical(duration_hours(NA) %/% duration_hours(1), NA_integer_)
  expect_identical(duration_hours(1) %/% duration_hours(NA), NA_integer_)
})

test_that("`<duration> %/% <duration>` propagates names", {
  expect_named(c(x = duration_hours(1)) %/% duration_hours(1:2), c("x", "x"))
  expect_named(c(x = duration_hours(1)) %/% c(y = duration_hours(1)), "x")
  expect_named(duration_hours(1) %/% c(y = duration_hours(1)), "y")
})

test_that("`<duration> %/% <duration>` results in NA for OOB values", {
  skip_on_cran()

  one <- duration_hours(1)
  numerator <- duration_hours(.Machine$integer.max)
  denominator <- duration_hours(1)

  expect_identical(numerator %/% denominator, .Machine$integer.max)
  expect_identical(-numerator %/% denominator, -.Machine$integer.max)

  expect_snapshot(out <- (numerator + one) %/% denominator)
  expect_identical(out, NA_integer_)

  expect_snapshot(out <- (-numerator - one) %/% denominator)
  expect_identical(out, NA_integer_)
})

test_that("`<duration> %% <numeric>` works (#273)", {
  expect_identical(duration_hours(7) %% 4, duration_hours(3))
})

test_that("`<duration> %% <numeric>` propagates `NA`", {
  expect_identical(duration_hours(7) %% NA_integer_, duration_hours(NA))
  expect_identical(duration_hours(NA) %% 4, duration_hours(NA))
})

test_that("`<duration> %% <numeric>` casts the numeric to integer", {
  expect_snapshot((expect_error(duration_hours(5) %% 2.5)))
})

# ------------------------------------------------------------------------------
# vec_math()

test_that("is.nan() works", {
  x <- duration_years(c(1, NA))
  expect_identical(is.nan(x), c(FALSE, FALSE))
})

test_that("is.finite() works", {
  x <- duration_years(c(1, NA))
  expect_identical(is.finite(x), c(TRUE, FALSE))
})

test_that("is.infinite() works", {
  x <- duration_years(c(1, NA))
  expect_identical(is.infinite(x), c(FALSE, FALSE))
})

test_that("abs() works", {
  x <- duration_hours(c(-2, -1, 0, 1, 2, NA))
  expect <- duration_hours(c(2, 1, 0, 1, 2, NA))
  expect_identical(abs(x), expect)
})

test_that("abs() propagates names", {
  x <- set_names(duration_years(1:2), c("a", "b"))
  expect_named(abs(x), c("a", "b"))
})

test_that("sign() works", {
  x <- duration_hours(c(-2, -1, 0, 1, 2, NA))
  expect <- c(-1L, -1L, 0L, 1L, 1L, NA)
  expect_identical(sign(x), expect)
})

test_that("sign() propagates names", {
  x <- set_names(duration_years(1:2), c("a", "b"))
  expect_named(sign(x), c("a", "b"))
})

# ------------------------------------------------------------------------------
# print() / obj_print_data() / obj_print_footer()

test_that("normal print method works", {
  x <- as_sys_time(year_month_day(2019, 1:5, 1))
  expect_snapshot(x)
})

test_that("can limit with `max`", {
  x <- as_sys_time(year_month_day(2019, 1:5, 1))

  expect_snapshot(print(x, max = 2))
  expect_snapshot(print(x, max = 4))

  # no footer if length >= max
  expect_snapshot(print(x, max = 5))
  expect_snapshot(print(x, max = 6))
})

test_that("`max` defaults to `getOption('max.print')` but can be overridden", {
  local_options(max.print = 3)

  x <- as_naive_time(year_month_day(2019, 1:5, 1))

  expect_snapshot(x)
  expect_snapshot(print(x, max = 4))
  expect_snapshot(print(x, max = 5))
})

# ------------------------------------------------------------------------------
# time_point_floor() / _ceiling() / _round()

test_that("can round to less precise precision", {
  x <- naive_seconds(c(-86401, -86400, -86399, 0, 86399, 86400, 86401))

  floor <- naive_days(c(-2, -1, -1, 0, 0, 1, 1))
  ceiling <- naive_days(c(-1, -1, 0, 0, 1, 1, 2))
  round <- naive_days(c(-1, -1, -1, 0, 1, 1, 1))

  expect_identical(time_point_floor(x, "day"), floor)
  expect_identical(time_point_ceiling(x, "day"), ceiling)
  expect_identical(time_point_round(x, "day"), round)

  floor <- naive_days(c(-2, -2, -2, 0, 0, 0, 0))
  ceiling <- naive_days(c(0, 0, 0, 0, 2, 2, 2))
  round <- naive_days(c(-2, 0, 0, 0, 0, 2, 2))

  expect_identical(time_point_floor(x, "day", n = 2), floor)
  expect_identical(time_point_ceiling(x, "day", n = 2), ceiling)
  expect_identical(time_point_round(x, "day", n = 2), round)
})

test_that("can round with `origin` altering starting point", {
  x <- sys_seconds(c(-86401, -86400, -86399, 0, 86399, 86400, 86401))

  origin <- sys_days(-1)

  floor <- sys_days(c(-3, -1, -1, -1, -1, 1, 1))
  ceiling <- sys_days(c(-1, -1, 1, 1, 1, 1, 3))
  round <- sys_days(c(-1, -1, -1, 1, 1, 1, 1))

  expect_identical(time_point_floor(x, "day", origin = origin, n = 2), floor)
  expect_identical(time_point_ceiling(x, "day", origin = origin, n = 2), ceiling)
  expect_identical(time_point_round(x, "day", origin = origin, n = 2), round)
})

test_that("cannot floor to more precise precision", {
  expect_snapshot(error = TRUE, time_point_floor(naive_days(), "second"))
})

test_that("rounding with `origin` requires same clock", {
  origin <- sys_days(0)
  x <- naive_days(0)
  expect_snapshot(error = TRUE, time_point_floor(x, "day", origin = origin))
})

test_that("`origin` can be cast to a more precise `precision`, but not to a less precise one", {
  origin1 <- as_naive_time(duration_days(1))
  origin2 <- as_naive_time(duration_milliseconds(0))
  x <- naive_seconds(0)

  expect_identical(
    time_point_floor(x, "hour", origin = origin1, n = 5),
    time_point_floor(x - as_duration(origin1), "hour", n = 5) + as_duration(origin1)
  )

  expect_snapshot(error = TRUE, time_point_floor(x, "hour", origin = origin2))
})

test_that("`origin` must be size 1", {
  origin <- naive_days(0:1)
  x <- naive_days(0)
  expect_snapshot(error = TRUE, time_point_floor(x, "day", origin = origin))
})

test_that("`origin` must not be `NA`", {
  origin <- naive_days(NA)
  x <- naive_days(0)
  expect_snapshot(error = TRUE, time_point_floor(x, "day", origin = origin))
})

test_that("`origin` can't be Date or POSIXt", {
  origin1 <- new_date(0)
  origin2 <- new_datetime(0, "America/New_York")
  x <- naive_days(0)
  expect_snapshot(error = TRUE, time_point_floor(x, "day", origin = origin1))
  expect_snapshot(error = TRUE, time_point_floor(x, "day", origin = origin2))
})

# ------------------------------------------------------------------------------
# time_point_shift()

test_that("can shift to next weekday", {
  expect_identical(
    time_point_shift(
      naive_days(0:1),
      weekday(clock_weekdays$sunday)
    ),
    naive_days(c(3, 3))
  )
})

test_that("can shift to next if on the boundary", {
  naive_sunday <- naive_days(3)
  sunday <- weekday(clock_weekdays$sunday)

  expect_identical(
    time_point_shift(naive_sunday, sunday),
    naive_sunday
  )
  expect_identical(
    time_point_shift(naive_sunday, sunday, boundary = "advance"),
    naive_sunday + 7
  )
})

test_that("can shift to previous weekday", {
  expect_identical(
    time_point_shift(
      naive_days(0:1),
      weekday(clock_weekdays$sunday),
      which = "previous"
    ),
    naive_days(c(-4, -4))
  )
})

test_that("can shift to previous weekday if on boundary", {
  naive_sunday <- naive_days(3)
  sunday <- weekday(clock_weekdays$sunday)

  expect_identical(
    time_point_shift(naive_sunday, sunday, which = "previous"),
    naive_sunday
  )
  expect_identical(
    time_point_shift(naive_sunday, sunday, which = "previous", boundary = "advance"),
    naive_sunday - 7
  )
})

test_that("`target` is recycled to size of `x`", {
  expect_identical(
    time_point_shift(
      sys_days(0:1),
      weekday(1:2)
    ),
    sys_days(3:4)
  )

  expect_snapshot(error = TRUE, time_point_shift(sys_days(0), weekday(1:2)))
})

test_that("`x` is validated", {
  expect_snapshot(error = TRUE, time_point_shift(1))
})

test_that("`target` is validated", {
  expect_snapshot(error = TRUE, time_point_shift(sys_days(0), 1))
})

test_that("`which` is validated", {
  expect_snapshot(error = TRUE, time_point_shift(sys_days(), weekday(), which = 1))
  expect_snapshot(error = TRUE, time_point_shift(sys_days(), weekday(), which = "foo"))
  expect_snapshot(
    error = TRUE,
    time_point_shift(sys_days(), weekday(), which = c("next", "previous"))
  )
})

test_that("`boundary` is validated", {
  expect_snapshot(error = TRUE, time_point_shift(sys_days(), weekday(), boundary = 1))
  expect_snapshot(
    error = TRUE,
    time_point_shift(sys_days(), weekday(), boundary = "foo")
  )
  expect_snapshot(
    error = TRUE,
    time_point_shift(sys_days(), weekday(), boundary = c("keep", "advance"))
  )
})

# ------------------------------------------------------------------------------
# time_point_count_between()

test_that("can count units between", {
  x <- as_naive_time(year_month_day(1990, 02, 03, 04))
  y <- as_naive_time(year_month_day(1995, 04, 05, 03))

  expect_identical(time_point_count_between(x, y, "day"), 1886L)
  expect_identical(time_point_count_between(x, y, "hour"), 45287L)
})

test_that("'week' is an allowed precision", {
  x <- sys_days(0)
  y <- sys_days(13:15)

  expect_identical(time_point_count_between(x, y, "week"), c(1L, 2L, 2L))
})

test_that("`n` affects the result", {
  x <- sys_days(0)
  y <- sys_days(10)

  expect_identical(time_point_count_between(x, y, "day", n = 2L), 5L)
  expect_identical(time_point_count_between(x, y, "day", n = 3L), 3L)
})

test_that("negative vs positive differences are handled correctly", {
  one_hour <- duration_hours(1)

  x <- sys_days(0)
  y <- sys_days(1)
  z <- sys_days(-1)

  expect_identical(time_point_count_between(x, y - one_hour, "day"), 0L)
  expect_identical(time_point_count_between(x, y, "day"), 1L)
  expect_identical(time_point_count_between(x, y + one_hour, "day"), 1L)

  expect_identical(time_point_count_between(x, z - one_hour, "day"), -1L)
  expect_identical(time_point_count_between(x, z, "day"), -1L)
  expect_identical(time_point_count_between(x, z + one_hour, "day"), 0L)
})

test_that("common precision of inputs and `precision` is taken", {
  expect_identical(
    time_point_count_between(sys_days(0), sys_days(2) + duration_hours(1), "second"),
    176400L
  )
  expect_identical(
    time_point_count_between(sys_seconds(0), sys_seconds(86401), "day"),
    1L
  )
})

test_that("OOB results return a warning and NA", {
  expect_snapshot({
    out <- time_point_count_between(sys_days(0), sys_days(1000), "nanosecond")
  })
  expect_identical(out, NA_integer_)
})

test_that("both inputs must be time points", {
  expect_snapshot({
    (expect_error(time_point_count_between(sys_days(1), 1)))
    (expect_error(time_point_count_between(1, sys_days(1))))
  })
})

test_that("both inputs must be compatible", {
  x <- sys_days(1)
  y <- naive_days(1)

  expect_snapshot(
    (expect_error(
      time_point_count_between(x, y)
    ))
  )
})

test_that("`n` is validated", {
  x <- sys_days(1)

  expect_snapshot({
    (expect_error(time_point_count_between(x, x, "day", n = NA_integer_)))
    (expect_error(time_point_count_between(x, x, "day", n = -1)))
    (expect_error(time_point_count_between(x, x, "day", n = 1.5)))
    (expect_error(time_point_count_between(x, x, "day", n = "x")))
    (expect_error(time_point_count_between(x, x, "day", n = c(1L, 2L))))
  })
})

test_that("`precision` must be a time point precision", {
  x <- sys_days(1)

  expect_snapshot(
    (expect_error(
      time_point_count_between(x, x, "year")
    ))
  )
})

# ------------------------------------------------------------------------------
# seq()

test_that("seq(to, by) works", {
  expect_identical(
    seq(sys_days(0L), to = sys_days(4L), by = 2),
    sys_days(c(0L, 2L, 4L))
  )
  expect_identical(
    seq(sys_days(0L), to = sys_days(5L), by = 2),
    sys_days(c(0L, 2L, 4L))
  )

  expect_identical(
    seq(sys_seconds(0L), to = sys_seconds(-4L), by = -2),
    sys_seconds(c(0L, -2L, -4L))
  )
  expect_identical(
    seq(sys_seconds(0L), to = sys_seconds(-5L), by = -2),
    sys_seconds(c(0L, -2L, -4L))
  )
})

test_that("seq(to, length.out) works", {
  expect_identical(
    seq(naive_days(0L), to = naive_days(4L), length.out = 2),
    naive_days(c(0L, 4L))
  )
  expect_identical(
    seq(naive_days(0L), to = naive_days(4L), length.out = 1),
    naive_days(c(0L))
  )
  expect_identical(
    seq(naive_days(0L), to = naive_days(4L), length.out = 5),
    naive_days(c(0:4))
  )

  expect_identical(
    seq(naive_seconds(0L), to = naive_seconds(4L), along.with = 1:2),
    naive_seconds(c(0L, 4L))
  )
})

test_that("seq(by, length.out) works", {
  expect_identical(
    seq(naive_seconds(0L), by = 2, length.out = 3),
    naive_seconds(c(0L, 2L, 4L))
  )
  expect_identical(
    seq(naive_seconds(0L), by = -2, length.out = 3),
    naive_seconds(c(0L, -2L, -4L))
  )

  expect_identical(
    seq(naive_seconds(0L), by = 2, along.with = 1:3),
    naive_seconds(c(0L, 2L, 4L))
  )
})

test_that("seq() with `from > to && by > 0` or `from < to && by < 0` results in length 0 output (#282)", {
  expect_identical(seq(naive_days(2L), to = naive_days(1L), by = 1), naive_days())
  expect_identical(seq(naive_days(5L), to = naive_days(1L), by = 1), naive_days())

  expect_identical(seq(naive_days(1L), to = naive_days(2L), by = -1), naive_days())
  expect_identical(seq(naive_days(1L), to = naive_days(5L), by = -1), naive_days())

  # In particular, handles the case where subtraction of distant `from` and `to` would overflow
  x <- as_naive_time(duration_cast(duration_years(200), "nanosecond"))
  y <- as_naive_time(duration_cast(duration_years(-200), "nanosecond"))
  expect_identical(seq(x, y, by = 1), as_naive_time(duration_nanoseconds()))
  expect_identical(seq(y, x, by = -1), as_naive_time(duration_nanoseconds()))
})

test_that("`by` can be a duration", {
  expect_identical(
    seq(naive_seconds(0), to = naive_seconds(1000), by = duration_minutes(1)),
    seq(naive_seconds(0), to = naive_seconds(1000), by = 60)
  )
  expect_identical(
    seq(
      as_naive_time(duration_nanoseconds(0)),
      to = as_naive_time(duration_nanoseconds(2e9)),
      by = duration_seconds(1)
    ),
    seq(
      as_naive_time(duration_nanoseconds(0)),
      to = as_naive_time(duration_nanoseconds(2e9)),
      by = 1e9
    )
  )
})

test_that("can't mix chronological time points and calendrical durations", {
  expect_snapshot(
    error = TRUE,
    seq(naive_seconds(0), by = duration_years(1), length.out = 2)
  )
})

test_that("can't mix clocks in seq()", {
  expect_snapshot(error = TRUE, seq(sys_seconds(0), to = naive_seconds(5), by = 1))
})

test_that("`to` is always cast to `from`", {
  expect_identical(
    seq(naive_seconds(0), to = naive_days(12), by = duration_days(2)),
    seq(naive_seconds(0), to = naive_seconds(12 * 86400), by = 86400 * 2)
  )

  expect_snapshot(error = TRUE, seq(naive_days(0), to = naive_seconds(5), by = 2))
})

test_that("can make nanosecond precision seqs", {
  x <- as_naive_time(duration_nanoseconds(0))
  y <- as_naive_time(duration_nanoseconds(10))

  expect_identical(seq(x, by = 2, length.out = 5), x + c(0, 2, 4, 6, 8))
  expect_identical(seq(x, y, by = 3), x + c(0, 3, 6, 9))
})

# ------------------------------------------------------------------------------
# time_point_spanning_seq()

test_that("generates the regular sequence along the full span", {
  x <- naive_days(c(-5, 5, 6, 0))
  expect_identical(time_point_spanning_seq(x), naive_days(-5:6))
})

test_that("missing values are removed", {
  x <- naive_days(c(1, NA, 0, 2))
  expect_identical(time_point_spanning_seq(x), naive_days(0:2))

  x <- naive_days(c(NA, NA))
  expect_identical(time_point_spanning_seq(x), naive_days())
})

test_that("works with empty vectors", {
  x <- naive_days()
  expect_identical(time_point_spanning_seq(x), x)
})

test_that("validates the input", {
  expect_snapshot(error = TRUE, {
    time_point_spanning_seq(1)
  })
})

# ------------------------------------------------------------------------------
# vec_arith()

test_that("duration to add to a time-point must have at least week precision (#120)", {
  expect_snapshot(error = TRUE, naive_seconds(0) + duration_years(1))
})

# ------------------------------------------------------------------------------
# time_point_precision()

test_that("precision: can get the precision", {
  expect_identical(time_point_precision(as_naive_time(duration_days(2:5))), "day")
  expect_identical(
    time_point_precision(as_naive_time(duration_nanoseconds(2:5))),
    "nanosecond"
  )
})

test_that("precision: can only be called on time points", {
  expect_snapshot(error = TRUE, time_point_precision(duration_days()))
})

# ------------------------------------------------------------------------------
# add_*()

test_that("unsupported time point addition throws good error", {
  x <- naive_seconds()

  expect_snapshot(error = TRUE, {
    add_years(x, 1)
  })
  expect_snapshot(error = TRUE, {
    add_quarters(x, 1)
  })
  expect_snapshot(error = TRUE, {
    add_months(x, 1)
  })
})

# ------------------------------------------------------------------------------
# diff()

test_that("works with `lag` and `differences`", {
  x <- naive_seconds(1:8)
  expect_identical(diff(x), rep(duration_seconds(1), 7))
  expect_identical(diff(x, lag = 2), rep(duration_seconds(2), 6))
  expect_identical(diff(x, differences = 2), rep(duration_seconds(0), 6))
  expect_identical(
    diff(x, lag = 3, differences = 2),
    rep(duration_seconds(0), 2)
  )

  x <- sys_seconds(1:8)
  expect_identical(diff(x), rep(duration_seconds(1), 7))
  expect_identical(diff(x, lag = 2), rep(duration_seconds(2), 6))
  expect_identical(diff(x, differences = 2), rep(duration_seconds(0), 6))
  expect_identical(
    diff(x, lag = 3, differences = 2),
    rep(duration_seconds(0), 2)
  )
})

test_that("works with `lag` and `differences` that force an empty result (#364)", {
  expect_identical(diff(naive_seconds()), duration_seconds())
  expect_identical(diff(naive_seconds(1)), duration_seconds())
  expect_identical(diff(naive_seconds(1:2)), duration_seconds(1))
  expect_identical(
    diff(naive_seconds(1:8), lag = 4, differences = 3),
    duration_seconds()
  )
})

# ------------------------------------------------------------------------------
# clock_minimum() / clock_maximum()

test_that("minimums are right", {
  skip_if_not_on_os("mac")

  # Known that the only time point that prints the limits right is nanosecond
  # due to how the print method goes through year-month-day and the year there
  # isn't high enough
  expect_snapshot({
    # clock_minimum(as_sys_time(duration_days()))
    # clock_minimum(as_sys_time(duration_hours()))
    # clock_minimum(as_sys_time(duration_minutes()))
    # clock_minimum(as_sys_time(duration_seconds()))
    # clock_minimum(as_sys_time(duration_milliseconds()))
    # clock_minimum(as_sys_time(duration_microseconds()))
    clock_minimum(as_sys_time(duration_nanoseconds()))

    # clock_minimum(as_naive_time(duration_days()))
    # clock_minimum(as_naive_time(duration_hours()))
    # clock_minimum(as_naive_time(duration_minutes()))
    # clock_minimum(as_naive_time(duration_seconds()))
    # clock_minimum(as_naive_time(duration_milliseconds()))
    # clock_minimum(as_naive_time(duration_microseconds()))
    clock_minimum(as_naive_time(duration_nanoseconds()))
  })

  for (precision in precision_names()) {
    precision <- precision_to_integer(precision)

    if (precision < PRECISION_DAY) {
      next
    }

    x <- duration_helper(0L, precision)

    expect_identical(as_duration(clock_minimum(as_sys_time(x))), clock_minimum(x))
    expect_identical(as_duration(clock_minimum(as_naive_time(x))), clock_minimum(x))
  }
})

test_that("maximums are right", {
  skip_if_not_on_os("mac")

  # Known that the only time point that prints the limits right is nanosecond
  # due to how the print method goes through year-month-day and the year there
  # isn't high enough
  expect_snapshot({
    # clock_maximum(as_sys_time(duration_days()))
    # clock_maximum(as_sys_time(duration_hours()))
    # clock_maximum(as_sys_time(duration_minutes()))
    # clock_maximum(as_sys_time(duration_seconds()))
    # clock_maximum(as_sys_time(duration_milliseconds()))
    # clock_maximum(as_sys_time(duration_microseconds()))
    clock_maximum(as_sys_time(duration_nanoseconds()))

    # clock_maximum(as_naive_time(duration_days()))
    # clock_maximum(as_naive_time(duration_hours()))
    # clock_maximum(as_naive_time(duration_minutes()))
    # clock_maximum(as_naive_time(duration_seconds()))
    # clock_maximum(as_naive_time(duration_milliseconds()))
    # clock_maximum(as_naive_time(duration_microseconds()))
    clock_maximum(as_naive_time(duration_nanoseconds()))
  })

  for (precision in precision_names()) {
    precision <- precision_to_integer(precision)

    if (precision < PRECISION_DAY) {
      next
    }

    x <- duration_helper(0L, precision)

    expect_identical(as_duration(clock_maximum(as_sys_time(x))), clock_maximum(x))
    expect_identical(as_duration(clock_maximum(as_naive_time(x))), clock_maximum(x))
  }
})

# ------------------------------------------------------------------------------
# min() / max() / range()

test_that("min() / max() / range() works", {
  x <- naive_days(c(1, 3, 2, 1, -1))

  expect_identical(min(x), naive_days(-1))
  expect_identical(max(x), naive_days(3))
  expect_identical(range(x), naive_days(c(-1, 3)))
})

test_that("min() / max() / range() works with `NA`", {
  x <- naive_days(c(1, NA, 2, 0))

  expect_identical(min(x), naive_days(NA))
  expect_identical(max(x), naive_days(NA))
  expect_identical(range(x), naive_days(c(NA, NA)))

  expect_identical(min(x, na.rm = TRUE), naive_days(0))
  expect_identical(max(x, na.rm = TRUE), naive_days(2))
  expect_identical(range(x, na.rm = TRUE), naive_days(c(0, 2)))
})

test_that("min() / max() / range() works when empty", {
  x <- naive_days()

  expect_identical(min(x), clock_maximum(x))
  expect_identical(max(x), clock_minimum(x))
  expect_identical(range(x), c(clock_maximum(x), clock_minimum(x)))

  x <- naive_days(c(NA, NA))

  expect_identical(min(x, na.rm = TRUE), clock_maximum(x))
  expect_identical(max(x, na.rm = TRUE), clock_minimum(x))
  expect_identical(range(x, na.rm = TRUE), c(clock_maximum(x), clock_minimum(x)))
})

# nocov start

.onLoad <- function(libname, pkgname) {
  tzdb::tzdb_initialize()

  clock_init_utils()

  clock_ns <- topenv(environment())

  # Initializers must run after initializing C++ utils and setting tzdata
  clock_init_duration_utils(clock_ns)
  clock_init_limits_init(clock_ns)
  clock_init_year_month_day_utils(clock_ns)
  clock_init_year_month_weekday_utils(clock_ns)
  clock_init_iso_year_week_day_utils(clock_ns)
  clock_init_year_day_utils(clock_ns)
  clock_init_sys_time_utils(clock_ns)
  clock_init_naive_time_utils(clock_ns)
  clock_init_zoned_time_utils(clock_ns)
  clock_init_weekday_utils(clock_ns)

  vctrs::s3_register(
    "pillar::pillar_shaft",
    "clock_calendar",
    pillar_shaft.clock_calendar
  )
  vctrs::s3_register(
    "pillar::pillar_shaft",
    "clock_time_point",
    pillar_shaft.clock_time_point
  )
  vctrs::s3_register(
    "pillar::pillar_shaft",
    "clock_zoned_time",
    pillar_shaft.clock_zoned_time
  )

  vctrs::s3_register(
    "slider::slider_plus",
    "Date.clock_duration",
    slider_plus.Date.clock_duration
  )
  vctrs::s3_register(
    "slider::slider_plus",
    "POSIXct.clock_duration",
    slider_plus.POSIXct.clock_duration
  )
  vctrs::s3_register(
    "slider::slider_plus",
    "POSIXlt.clock_duration",
    slider_plus.POSIXlt.clock_duration
  )

  vctrs::s3_register(
    "slider::slider_minus",
    "Date.clock_duration",
    slider_minus.Date.clock_duration
  )
  vctrs::s3_register(
    "slider::slider_minus",
    "POSIXct.clock_duration",
    slider_minus.POSIXct.clock_duration
  )
  vctrs::s3_register(
    "slider::slider_minus",
    "POSIXlt.clock_duration",
    slider_minus.POSIXlt.clock_duration
  )
}

# nocov end

#' Calendar: year-quarter-day
#'
#' `year_quarter_day()` constructs a calendar from the fiscal year, fiscal
#' quarter, and day of the quarter, along with a value determining which
#' month the fiscal year `start`s in.
#'
#' @details
#' Fields are recycled against each other.
#'
#' Fields are collected in order until the first `NULL` field is located. No
#' fields after the first `NULL` field are used.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams year_month_day
#'
#' @param year `[integer]`
#'
#'   The fiscal year. Values `[-32767, 32767]` are generally allowed.
#'
#' @param quarter `[integer / NULL]`
#'
#'   The fiscal quarter. Values `[1, 4]` are allowed.
#'
#' @param day `[integer / "last" / NULL]`
#'
#'   The day of the quarter. Values `[1, 92]` are allowed.
#'
#'   If `"last"`, the last day of the quarter is returned.
#'
#' @param start `[integer(1) / NULL]`
#'
#'   The month to start the fiscal year in. 1 = January and 12 = December.
#'
#'   If `NULL`, a `start` of January will be used.
#'
#' @return A year-quarter-day calendar vector.
#'
#' @export
#' @examples
#' # Year-quarter type
#' x <- year_quarter_day(2019, 1:4)
#' x
#'
#' add_quarters(x, 2)
#'
#' # Set the day to the last day of the quarter
#' x <- set_day(x, "last")
#' x
#'
#' # Start the fiscal year in June
#' june <- 6L
#' y <- year_quarter_day(2019, 1:4, "last", start = june)
#'
#' # Compare the year-month-day values that result from having different
#' # fiscal year start months
#' as_year_month_day(x)
#' as_year_month_day(y)
year_quarter_day <- function(year,
                             quarter = NULL,
                             day = NULL,
                             hour = NULL,
                             minute = NULL,
                             second = NULL,
                             subsecond = NULL,
                             ...,
                             start = NULL,
                             subsecond_precision = NULL) {
  check_dots_empty()

  start <- quarterly_validate_start(start)

  # Stop on the first `NULL` argument
  if (is_null(quarter)) {
    precision <- PRECISION_YEAR
    fields <- list(year = year)
  } else if (is_null(day)) {
    precision <- PRECISION_QUARTER
    fields <- list(year = year, quarter = quarter)
  } else if (is_null(hour)) {
    precision <- PRECISION_DAY
    fields <- list(year = year, quarter = quarter, day = day)
  } else if (is_null(minute)) {
    precision <- PRECISION_HOUR
    fields <- list(year = year, quarter = quarter, day = day, hour = hour)
  } else if (is_null(second)) {
    precision <- PRECISION_MINUTE
    fields <- list(year = year, quarter = quarter, day = day, hour = hour, minute = minute)
  } else if (is_null(subsecond)) {
    precision <- PRECISION_SECOND
    fields <- list(year = year, quarter = quarter, day = day, hour = hour, minute = minute, second = second)
  } else {
    precision <- calendar_validate_subsecond_precision(subsecond_precision)
    fields <- list(year = year, quarter = quarter, day = day, hour = hour, minute = minute, second = second, subsecond = subsecond)
  }

  if (is_last(fields$day)) {
    fields$day <- 1L
    last <- TRUE
  } else {
    last <- FALSE
  }

  fields <- vec_recycle_common(!!!fields)
  fields <- vec_cast_common(!!!fields, .to = integer())

  fields <- collect_year_quarter_day_fields(fields, precision, start)

  names <- NULL

  out <- new_year_quarter_day_from_fields(fields, precision, start, names)

  if (last) {
    out <- set_day(out, "last")
  }

  out
}

# ------------------------------------------------------------------------------

#' @export
vec_proxy.clock_year_quarter_day <- function(x, ...) {
  .Call(`_clock_clock_rcrd_proxy`, x)
}

#' @export
vec_restore.clock_year_quarter_day <- function(x, to, ...) {
  .Call(`_clock_year_quarter_day_restore`, x, to)
}

# ------------------------------------------------------------------------------

#' @export
format.clock_year_quarter_day <- function(x, ...) {
  out <- format_year_quarter_day_cpp(x, calendar_precision_attribute(x), quarterly_start(x))
  names(out) <- names(x)
  out
}

#' @export
vec_ptype_full.clock_year_quarter_day <- function(x, ...) {
  start <- quarterly_start(x)
  start <- quarterly_start_prettify(start, abbreviate = FALSE)
  class <- paste0("year_quarter_day<", start, ">")
  calendar_ptype_full(x, class)
}

#' @export
vec_ptype_abbr.clock_year_quarter_day <- function(x, ...) {
  start <- quarterly_start(x)
  start <- quarterly_start_prettify(start, abbreviate = TRUE)
  class <- paste0("yqd<", start, ">")
  calendar_ptype_abbr(x, class)
}

# ------------------------------------------------------------------------------

#' Is `x` a year-quarter-day?
#'
#' Check if `x` is a year-quarter-day.
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @return Returns `TRUE` if `x` inherits from `"clock_year_quarter_day"`,
#'   otherwise returns `FALSE`.
#'
#' @export
#' @examples
#' is_year_quarter_day(year_quarter_day(2019))
is_year_quarter_day <- function(x) {
  inherits(x, "clock_year_quarter_day")
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype.clock_year_quarter_day <- function(x, ...) {
  names <- NULL
  precision <- calendar_precision_attribute(x)
  start <- quarterly_start(x)

  f <- integer()

  fields <- switch(
    precision + 1L,
    list(year = f),
    list(year = f, quarter = f),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    list(year = f, quarter = f, day = f),
    list(year = f, quarter = f, day = f, hour = f),
    list(year = f, quarter = f, day = f, hour = f, minute = f),
    list(year = f, quarter = f, day = f, hour = f, minute = f, second = f),
    list(year = f, quarter = f, day = f, hour = f, minute = f, second = f, subsecond = f),
    list(year = f, quarter = f, day = f, hour = f, minute = f, second = f, subsecond = f),
    list(year = f, quarter = f, day = f, hour = f, minute = f, second = f, subsecond = f),
    abort("Internal error: Invalid precision.")
  )

  new_year_quarter_day_from_fields(fields, precision, start, names)
}

#' @export
vec_ptype2.clock_year_quarter_day.clock_year_quarter_day <- function(x, y, ...) {
  if (quarterly_start(x) != quarterly_start(y)) {
    stop_incompatible_type(x, y, ..., details = "Can't combine quarterly types with different `start`s.")
  }

  ptype2_calendar_and_calendar(x, y, ...)
}

#' @export
vec_cast.clock_year_quarter_day.clock_year_quarter_day <- function(x, to, ...) {
  if (quarterly_start(x) != quarterly_start(to)) {
    stop_incompatible_cast(x, to, ..., details = "Can't cast between quarterly types with different `start`s.")
  }

  cast_calendar_to_calendar(x, to, ...)
}

# ------------------------------------------------------------------------------

#' @export
calendar_is_valid_precision.clock_year_quarter_day <- function(x, precision) {
  year_quarter_day_is_valid_precision(precision)
}

year_quarter_day_is_valid_precision <- function(precision) {
  if (!is_valid_precision(precision)) {
    FALSE
  } else if (precision == PRECISION_YEAR || precision == PRECISION_QUARTER) {
    TRUE
  } else if (precision >= PRECISION_DAY && precision <= PRECISION_NANOSECOND) {
    TRUE
  } else {
    FALSE
  }
}

# ------------------------------------------------------------------------------

#' @export
invalid_detect.clock_year_quarter_day <- function(x) {
  invalid_detect_year_quarter_day_cpp(x, calendar_precision_attribute(x), quarterly_start(x))
}

#' @export
invalid_any.clock_year_quarter_day <- function(x) {
  invalid_any_year_quarter_day_cpp(x, calendar_precision_attribute(x), quarterly_start(x))
}

#' @export
invalid_count.clock_year_quarter_day <- function(x) {
  invalid_count_year_quarter_day_cpp(x, calendar_precision_attribute(x), quarterly_start(x))
}

#' @export
invalid_resolve.clock_year_quarter_day <- function(x, ..., invalid = NULL) {
  check_dots_empty()
  precision <- calendar_precision_attribute(x)
  start <- quarterly_start(x)
  invalid <- validate_invalid(invalid)
  fields <- invalid_resolve_year_quarter_day_cpp(x, precision, start, invalid)
  new_year_quarter_day_from_fields(fields, precision, start, names = names(x))
}

# ------------------------------------------------------------------------------

#' Getters: year-quarter-day
#'
#' @description
#' These are year-quarter-day methods for the
#' [getter generics][clock-getters].
#'
#' - `get_year()` returns the fiscal year. Note that this can differ from the
#'   Gregorian year if `start != 1L`.
#'
#' - `get_quarter()` returns the fiscal quarter as a value between 1-4.
#'
#' - `get_day()` returns the day of the fiscal quarter as a value between 1-92.
#'
#' - There are sub-daily getters for extracting more precise components.
#'
#' @param x `[clock_year_quarter_day]`
#'
#'   A year-quarter-day to get the component from.
#'
#' @return The component.
#'
#' @name year-quarter-day-getters
#' @examples
#' x <- year_quarter_day(2020, 1:4)
#'
#' get_quarter(x)
#'
#' # Set and then get the last day of the quarter
#' x <- set_day(x, "last")
#' get_day(x)
#'
#' # Start the fiscal year in November and choose the 50th day in
#' # each quarter of 2020
#' november <- 11
#' y <- year_quarter_day(2020, 1:4, 50, start = 11)
#' y
#'
#' get_day(y)
#'
#' # What does that map to in year-month-day?
#' as_year_month_day(y)
NULL

#' @rdname year-quarter-day-getters
#' @export
get_year.clock_year_quarter_day <- function(x) {
  field_year(x)
}

#' @rdname year-quarter-day-getters
#' @export
get_quarter.clock_year_quarter_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_QUARTER, "get_quarter")
  field_quarter(x)
}

#' @rdname year-quarter-day-getters
#' @export
get_day.clock_year_quarter_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_DAY, "get_day")
  field_day(x)
}

#' @rdname year-quarter-day-getters
#' @export
get_hour.clock_year_quarter_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_HOUR, "get_hour")
  field_hour(x)
}

#' @rdname year-quarter-day-getters
#' @export
get_minute.clock_year_quarter_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_MINUTE, "get_minute")
  field_minute(x)
}

#' @rdname year-quarter-day-getters
#' @export
get_second.clock_year_quarter_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_SECOND, "get_second")
  field_second(x)
}

#' @rdname year-quarter-day-getters
#' @export
get_millisecond.clock_year_quarter_day <- function(x) {
  calendar_require_precision(x, PRECISION_MILLISECOND, "get_millisecond")
  field_subsecond(x)
}

#' @rdname year-quarter-day-getters
#' @export
get_microsecond.clock_year_quarter_day <- function(x) {
  calendar_require_precision(x, PRECISION_MICROSECOND, "get_microsecond")
  field_subsecond(x)
}

#' @rdname year-quarter-day-getters
#' @export
get_nanosecond.clock_year_quarter_day <- function(x) {
  calendar_require_precision(x, PRECISION_NANOSECOND, "get_nanosecond")
  field_subsecond(x)
}

# ------------------------------------------------------------------------------

#' Setters: year-quarter-day
#'
#' @description
#' These are year-quarter-day methods for the
#' [setter generics][clock-setters].
#'
#' - `set_year()` sets the fiscal year.
#'
#' - `set_quarter()` sets the fiscal quarter of the year.
#'   Valid values are in the range of `[1, 4]`.
#'
#' - `set_day()` sets the day of the fiscal quarter.
#'   Valid values are in the range of `[1, 92]`.
#'
#' - There are sub-daily setters for setting more precise components.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[clock_year_quarter_day]`
#'
#'   A year-quarter-day vector.
#'
#' @param value `[integer / "last"]`
#'
#'   The value to set the component to.
#'
#'   For `set_day()`, this can also be `"last"` to adjust to the last
#'   day of the current fiscal quarter.
#'
#' @return `x` with the component set.
#'
#' @name year-quarter-day-setters
#' @examples
#' library(magrittr)
#'
#' # Quarter precision vector
#' x <- year_quarter_day(2019, 1:4)
#' x
#'
#' # Promote to day precision by setting the day
#' x <- set_day(x, 1)
#' x
#'
#' # Or set to the last day of the quarter
#' x <- set_day(x, "last")
#' x
#'
#' # What year-month-day is this?
#' as_year_month_day(x)
#'
#' # Set to an invalid day of the quarter
#' # (not all quarters have 92 days)
#' invalid <- set_day(x, 92)
#' invalid
#'
#' # Here are the invalid ones
#' invalid[invalid_detect(invalid)]
#'
#' # Resolve the invalid dates by choosing the previous/next valid moment
#' invalid_resolve(invalid, invalid = "previous")
#' invalid_resolve(invalid, invalid = "next")
#'
#' # Or resolve by "overflowing" by the number of days that you have
#' # gone past the last valid day
#' invalid_resolve(invalid, invalid = "overflow")
#'
#' # This is similar to
#' days <- get_day(invalid) - 1L
#' invalid %>%
#'   set_day(1) %>%
#'   as_naive_time() %>%
#'   add_days(days) %>%
#'   as_year_quarter_day()
NULL

#' @rdname year-quarter-day-setters
#' @export
set_year.clock_year_quarter_day <- function(x, value, ...) {
  check_dots_empty()
  set_field_year_quarter_day(x, value, "year")
}

#' @rdname year-quarter-day-setters
#' @export
set_quarter.clock_year_quarter_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_YEAR, "set_quarter")
  set_field_year_quarter_day(x, value, "quarter")
}

#' @rdname year-quarter-day-setters
#' @export
set_day.clock_year_quarter_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_QUARTER, "set_day")
  set_field_year_quarter_day(x, value, "day")
}

#' @rdname year-quarter-day-setters
#' @export
set_hour.clock_year_quarter_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_DAY, "set_hour")
  set_field_year_quarter_day(x, value, "hour")
}

#' @rdname year-quarter-day-setters
#' @export
set_minute.clock_year_quarter_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_HOUR, "set_minute")
  set_field_year_quarter_day(x, value, "minute")
}

#' @rdname year-quarter-day-setters
#' @export
set_second.clock_year_quarter_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_MINUTE, "set_second")
  set_field_year_quarter_day(x, value, "second")
}

#' @rdname year-quarter-day-setters
#' @export
set_millisecond.clock_year_quarter_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_MILLISECOND), "set_millisecond")
  set_field_year_quarter_day(x, value, "millisecond")
}

#' @rdname year-quarter-day-setters
#' @export
set_microsecond.clock_year_quarter_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_MICROSECOND), "set_microsecond")
  set_field_year_quarter_day(x, value, "microsecond")
}

#' @rdname year-quarter-day-setters
#' @export
set_nanosecond.clock_year_quarter_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_NANOSECOND), "set_nanosecond")
  set_field_year_quarter_day(x, value, "nanosecond")
}

set_field_year_quarter_day <- function(x, value, component) {
  if (is_last(value) && identical(component, "day")) {
    return(set_field_year_quarter_day_last(x))
  }

  start <- quarterly_start(x)

  precision_fields <- calendar_precision_attribute(x)
  precision_value <- year_quarter_day_component_to_precision(component)
  precision_out <- precision_common2(precision_fields, precision_value)

  value <- vec_cast(value, integer(), x_arg = "value")
  args <- vec_recycle_common(x = x, value = value)
  x <- args$x
  value <- args$value

  result <- set_field_year_quarter_day_cpp(x, value, precision_fields, precision_value, start)
  fields <- result$fields
  field <- year_quarter_day_component_to_field(component)
  fields[[field]] <- result$value

  new_year_quarter_day_from_fields(fields, precision_out, start, names = names(x))
}

set_field_year_quarter_day_last <- function(x) {
  start <- quarterly_start(x)

  precision_fields <- calendar_precision_attribute(x)
  precision_out <- precision_common2(precision_fields, PRECISION_DAY)

  result <- set_field_year_quarter_day_last_cpp(x, precision_fields, start)
  fields <- result$fields
  fields[["day"]] <- result$value

  new_year_quarter_day_from_fields(fields, precision_out, start, names = names(x))
}

# ------------------------------------------------------------------------------

#' @export
calendar_name.clock_year_quarter_day <- function(x) {
  "year_quarter_day"
}

# ------------------------------------------------------------------------------

year_quarter_day_component_to_precision <- function(component) {
  switch(
    component,
    year = PRECISION_YEAR,
    quarter = PRECISION_QUARTER,
    day = PRECISION_DAY,
    hour = PRECISION_HOUR,
    minute = PRECISION_MINUTE,
    second = PRECISION_SECOND,
    millisecond = PRECISION_MILLISECOND,
    microsecond = PRECISION_MICROSECOND,
    nanosecond = PRECISION_NANOSECOND,
    abort("Internal error: Unknown component name.")
  )
}

year_quarter_day_component_to_field <- function(component) {
  switch (
    component,
    year = component,
    quarter = component,
    day = component,
    hour = component,
    minute = component,
    second = component,
    millisecond = "subsecond",
    microsecond = "subsecond",
    nanosecond = "subsecond",
    abort("Internal error: Unknown component name.")
  )
}

# ------------------------------------------------------------------------------

#' @rdname clock-arith
#' @method vec_arith clock_year_quarter_day
#' @export
vec_arith.clock_year_quarter_day <- function(op, x, y, ...) {
  UseMethod("vec_arith.clock_year_quarter_day", y)
}

#' @method vec_arith.clock_year_quarter_day MISSING
#' @export
vec_arith.clock_year_quarter_day.MISSING <- function(op, x, y, ...) {
  arith_calendar_and_missing(op, x, y, ...)
}

#' @method vec_arith.clock_year_quarter_day clock_year_quarter_day
#' @export
vec_arith.clock_year_quarter_day.clock_year_quarter_day <- function(op, x, y, ...) {
  arith_calendar_and_calendar(op, x, y, ..., calendar_minus_calendar_fn = year_quarter_day_minus_year_quarter_day)
}

#' @method vec_arith.clock_year_quarter_day clock_duration
#' @export
vec_arith.clock_year_quarter_day.clock_duration <- function(op, x, y, ...) {
  arith_calendar_and_duration(op, x, y, ...)
}

#' @method vec_arith.clock_duration clock_year_quarter_day
#' @export
vec_arith.clock_duration.clock_year_quarter_day <- function(op, x, y, ...) {
  arith_duration_and_calendar(op, x, y, ...)
}

#' @method vec_arith.clock_year_quarter_day numeric
#' @export
vec_arith.clock_year_quarter_day.numeric <- function(op, x, y, ...) {
  arith_calendar_and_numeric(op, x, y, ...)
}

#' @method vec_arith.numeric clock_year_quarter_day
#' @export
vec_arith.numeric.clock_year_quarter_day <- function(op, x, y, ...) {
  arith_numeric_and_calendar(op, x, y, ...)
}

year_quarter_day_minus_year_quarter_day <- function(op, x, y, ...) {
  args <- vec_recycle_common(x = x, y = y)
  args <- vec_cast_common(!!!args)
  x <- args$x
  y <- args$y

  names <- names_common(x, y)

  start <- quarterly_start(x)
  precision <- calendar_precision_attribute(x)

  if (precision > PRECISION_QUARTER) {
    stop_incompatible_op(op, x, y, ...)
  }

  fields <- year_quarter_day_minus_year_quarter_day_cpp(x, y, precision, start)

  new_duration_from_fields(fields, precision, names = names)
}

# ------------------------------------------------------------------------------

#' Arithmetic: year-quarter-day
#'
#' @description
#' These are year-quarter-day methods for the
#' [arithmetic generics][clock-arithmetic].
#'
#' - `add_years()`
#'
#' - `add_quarters()`
#'
#' Notably, _you cannot add days to a year-quarter-day_. For day-based
#' arithmetic, first convert to a time point with [as_naive_time()] or
#' [as_sys_time()].
#'
#' @details
#' `x` and `n` are recycled against each other.
#'
#' @inheritParams add_years
#'
#' @param x `[clock_year_quarter_day]`
#'
#'   A year-quarter-day vector.
#'
#' @return `x` after performing the arithmetic.
#'
#' @name year-quarter-day-arithmetic
#'
#' @examples
#' x <- year_quarter_day(2019, 1:3)
#' x
#'
#' add_quarters(x, 2)
#'
#' # Make the fiscal year start in March
#' y <- year_quarter_day(2019, 1:2, 1, start = 3)
#' y
#'
#' add_quarters(y, 1)
#'
#' # What year-month-day does this correspond to?
#' # Note that the fiscal year doesn't necessarily align with the Gregorian
#' # year!
#' as_year_month_day(add_quarters(y, 1))
NULL

#' @rdname year-quarter-day-arithmetic
#' @export
add_years.clock_year_quarter_day <- function(x, n, ...) {
  year_quarter_day_plus_duration(x, n, PRECISION_YEAR)
}

#' @rdname year-quarter-day-arithmetic
#' @export
add_quarters.clock_year_quarter_day <- function(x, n, ...) {
  calendar_require_minimum_precision(x, PRECISION_QUARTER, "add_quarters")
  year_quarter_day_plus_duration(x, n, PRECISION_QUARTER)
}

year_quarter_day_plus_duration <- function(x, n, precision_n) {
  start <- quarterly_start(x)
  precision_fields <- calendar_precision_attribute(x)

  n <- duration_collect_n(n, precision_n)
  args <- vec_recycle_common(x = x, n = n)
  x <- args$x
  n <- args$n

  names <- names_common(x, n)

  fields <- year_quarter_day_plus_duration_cpp(x, n, precision_fields, precision_n, start)

  new_year_quarter_day_from_fields(fields, precision_fields, start, names = names)
}

# ------------------------------------------------------------------------------

#' Convert to year-quarter-day
#'
#' @description
#' `as_year_quarter_day()` converts a vector to the year-quarter-day
#' calendar. Time points, Dates, POSIXct, and other calendars can all be
#' converted to year-quarter-day.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[vector]`
#'
#'   A vector to convert to year-quarter-day.
#'
#' @param start `[integer(1) / NULL]`
#'
#'   The month to start the fiscal year in. 1 = January and 12 = December.
#'
#'   If `NULL`:
#'
#'   - If `x` is a year-quarter-day, it will be returned as is.
#'
#'   - Otherwise, a `start` of January will be used.
#'
#' @return A year-quarter-day vector.
#' @export
#' @examples
#' # From Date
#' as_year_quarter_day(as.Date("2019-01-01"))
#' as_year_quarter_day(as.Date("2019-01-01"), start = 3)
#'
#' # From POSIXct, which assumes that the naive time is what should be converted
#' as_year_quarter_day(as.POSIXct("2019-01-01 02:30:30", "America/New_York"))
#'
#' # From other calendars
#' tuesday <- 3
#' as_year_quarter_day(year_month_weekday(2019, 2, tuesday, 2))
#'
#' # Converting between `start`s
#' x <- year_quarter_day(2019, 01, 01, start = 2)
#' x
#'
#' # Default keeps the same start
#' as_year_quarter_day(x)
#'
#' # But you can change it
#' as_year_quarter_day(x, start = 1)
as_year_quarter_day <- function(x, ..., start = NULL)  {
  UseMethod("as_year_quarter_day")
}

#' @export
as_year_quarter_day.default <- function(x, ..., start = NULL) {
  stop_clock_unsupported_conversion(x, "clock_year_quarter_day")
}

#' @export
as_year_quarter_day.clock_year_quarter_day <- function(x, ..., start = NULL) {
  check_dots_empty()

  if (is_null(start)) {
    return(x)
  }

  start <- quarterly_validate_start(start)
  x_start <- quarterly_start(x)

  if (x_start == start) {
    return(x)
  }

  as_year_quarter_day(as_sys_time(x), start = start)
}

# ------------------------------------------------------------------------------

#' @export
as_sys_time.clock_year_quarter_day <- function(x) {
  calendar_require_all_valid(x)
  start <- quarterly_start(x)
  precision <- calendar_precision_attribute(x)
  fields <- as_sys_time_year_quarter_day_cpp(x, precision, start)
  new_sys_time_from_fields(fields, precision, clock_rcrd_names(x))
}

#' @export
as_naive_time.clock_year_quarter_day <- function(x) {
  as_naive_time(as_sys_time(x))
}

#' @export
as.character.clock_year_quarter_day <- function(x, ...) {
  format(x)
}

# ------------------------------------------------------------------------------

#' Grouping: year-quarter-day
#'
#' @description
#' This is a year-quarter-day method for the [calendar_group()] generic.
#'
#' Grouping for a year-quarter-day object can be done at any precision, as
#' long as `x` is at least as precise as `precision`.
#'
#' @inheritParams calendar_group
#'
#' @param x `[clock_year_quarter_day]`
#'
#'   A year-quarter-day vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"quarter"`
#'   - `"day"`
#'   - `"hour"`
#'   - `"minute"`
#'   - `"second"`
#'   - `"millisecond"`
#'   - `"microsecond"`
#'   - `"nanosecond"`
#'
#' @return `x` grouped at the specified `precision`.
#'
#' @name year-quarter-day-group
#'
#' @export
#' @examples
#' x <- year_quarter_day(2019, 1:4)
#' x <- c(x, set_year(x, 2020))
#'
#' # Group by 3 quarters
#' # Note that this is a grouping of 3 quarters of the current year
#' # (i.e. the count resets at the beginning of the next year)
#' calendar_group(x, "quarter", n = 3)
#'
#' # Group by 5 days of the current quarter
#' y <- year_quarter_day(2019, 1, 1:90)
#' calendar_group(y, "day", n = 5)
calendar_group.clock_year_quarter_day <- function(x, precision, ..., n = 1L) {
  n <- validate_calendar_group_n(n)
  x <- calendar_narrow(x, precision)

  precision <- validate_precision_string(precision)

  if (precision == PRECISION_YEAR) {
    value <- get_year(x)
    value <- group_component0(value, n)
    x <- set_year(x, value)
    return(x)
  }
  if (precision == PRECISION_QUARTER) {
    value <- get_quarter(x)
    value <- group_component1(value, n)
    x <- set_quarter(x, value)
    return(x)
  }
  if (precision == PRECISION_DAY) {
    value <- get_day(x)
    value <- group_component1(value, n)
    x <- set_day(x, value)
    return(x)
  }

  x <- calendar_group_time(x, n, precision)
  x
}

# ------------------------------------------------------------------------------

#' Narrow: year-quarter-day
#'
#' This is a year-quarter-day method for the [calendar_narrow()] generic. It
#' narrows a year-quarter-day vector to the specified `precision`.
#'
#' @inheritParams year-quarter-day-group
#'
#' @return `x` narrowed to the supplied `precision`.
#'
#' @name year-quarter-day-narrow
#'
#' @export
#' @examples
#' # Day precision
#' x <- year_quarter_day(2019, 1, 5)
#' x
#'
#' # Narrow to quarter precision
#' calendar_narrow(x, "quarter")
calendar_narrow.clock_year_quarter_day <- function(x, precision) {
  precision <- validate_precision_string(precision)

  start <- quarterly_start(x)

  out_fields <- list()
  x_fields <- unclass(x)

  if (precision >= PRECISION_YEAR) {
    out_fields[["year"]] <- x_fields[["year"]]
  }
  if (precision >= PRECISION_QUARTER) {
    out_fields[["quarter"]] <- x_fields[["quarter"]]
  }
  if (precision >= PRECISION_DAY) {
    out_fields[["day"]] <- x_fields[["day"]]
  }

  out_fields <- calendar_narrow_time(out_fields, precision, x_fields)

  new_year_quarter_day_from_fields(out_fields, precision, start, names = names(x))
}

# ------------------------------------------------------------------------------

#' Widen: year-quarter-day
#'
#' This is a year-quarter-day method for the [calendar_widen()] generic. It
#' widens a year-quarter-day vector to the specified `precision`.
#'
#' @inheritParams year-quarter-day-group
#'
#' @return `x` widened to the supplied `precision`.
#'
#' @name year-quarter-day-widen
#'
#' @export
#' @examples
#' # Quarter precision
#' x <- year_quarter_day(2019, 1)
#' x
#'
#' # Widen to day precision
#' calendar_widen(x, "day")
#'
#' # Or second precision
#' sec <- calendar_widen(x, "second")
#' sec
calendar_widen.clock_year_quarter_day <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  if (precision >= PRECISION_QUARTER && x_precision < PRECISION_QUARTER) {
    x <- set_quarter(x, 1L)
  }
  if (precision >= PRECISION_DAY && x_precision < PRECISION_DAY) {
    x <- set_day(x, 1L)
  }

  x <- calendar_widen_time(x, x_precision, precision)

  x
}

# ------------------------------------------------------------------------------

#' Boundaries: year-quarter-day
#'
#' This is a year-quarter-day method for the [calendar_start()] and
#' [calendar_end()] generics. They adjust components of a calendar to the
#' start or end of a specified `precision`.
#'
#' @inheritParams year-quarter-day-group
#'
#' @return `x` at the same precision, but with some components altered to be
#'   at the boundary value.
#'
#' @name year-quarter-day-boundary
#'
#' @examples
#' x <- year_quarter_day(2019:2020, 2:3, 5, 6, 7, 8, start = clock_months$march)
#' x
#'
#' # Compute the last moment of the fiscal quarter
#' calendar_end(x, "quarter")
#'
#' # Compare that to just setting the day to `"last"`,
#' # which doesn't affect the other components
#' set_day(x, "last")
#'
#' # Compute the start of the fiscal year
#' calendar_start(x, "year")
#'
#' as_date(calendar_start(x, "year"))
NULL

#' @rdname year-quarter-day-boundary
#' @export
calendar_start.clock_year_quarter_day <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  calendar_start_end_checks(x, x_precision, precision, "start")

  if (precision <= PRECISION_YEAR && x_precision > PRECISION_YEAR) {
    x <- set_quarter(x, 1L)
  }
  if (precision <= PRECISION_QUARTER && x_precision > PRECISION_QUARTER) {
    x <- set_day(x, 1L)
  }

  x <- calendar_start_time(x, x_precision, precision)

  x
}

#' @rdname year-quarter-day-boundary
#' @export
calendar_end.clock_year_quarter_day <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  calendar_start_end_checks(x, x_precision, precision, "end")

  if (precision <= PRECISION_YEAR && x_precision > PRECISION_YEAR) {
    x <- set_quarter(x, 4L)
  }
  if (precision <= PRECISION_QUARTER && x_precision > PRECISION_QUARTER) {
    x <- set_day(x, "last")
  }

  x <- calendar_end_time(x, x_precision, precision)

  x
}

# ------------------------------------------------------------------------------

#' Counting: year-quarter-day
#'
#' This is a year-quarter-day method for the [calendar_count_between()] generic.
#' It counts the number of `precision` units between `start` and `end` (i.e.,
#' the number of years or quarters).
#'
#' @inheritParams calendar-count-between
#'
#' @param start,end `[clock_year_quarter_day]`
#'
#'   A pair of year-quarter-day vectors. These will be recycled to their
#'   common size.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"quarter"`
#'
#' @inherit calendar-count-between return
#'
#' @name year-quarter-day-count-between
#'
#' @export
#' @examples
#' # Compute the number of whole quarters between two dates
#' x <- year_quarter_day(2020, 3, 91)
#' y <- year_quarter_day(2025, 4, c(90, 92))
#' calendar_count_between(x, y, "quarter")
#'
#' # Note that this is not always the same as the number of whole 3 month
#' # periods between two dates
#' x <- as_year_month_day(x)
#' y <- as_year_month_day(y)
#' calendar_count_between(x, y, "month", n = 3)
calendar_count_between.clock_year_quarter_day <- function(start,
                                                          end,
                                                          precision,
                                                          ...,
                                                          n = 1L) {
  NextMethod()
}

calendar_count_between_standardize_precision_n.clock_year_quarter_day <- function(x,
                                                                                  precision,
                                                                                  n) {
  precision_int <- validate_precision_string(precision)

  allowed_precisions <- c(PRECISION_YEAR, PRECISION_QUARTER)

  if (!(precision_int %in% allowed_precisions)) {
    abort("`precision` must be one of: 'year', 'quarter'.")
  }

  list(precision = precision, n = n)
}

calendar_count_between_compute.clock_year_quarter_day <- function(start,
                                                                  end,
                                                                  precision) {
  precision <- validate_precision_string(precision)

  if (precision == PRECISION_YEAR) {
    out <- get_year(end) - get_year(start)
    return(out)
  }

  if (precision == PRECISION_QUARTER) {
    out <- (get_year(end) - get_year(start)) * 4L
    out <- out + (get_quarter(end) - get_quarter(start))
    return(out)
  }

  abort("Internal error: `precision` should be 'year' or 'quarter' at this point.")
}

calendar_count_between_proxy_compare.clock_year_quarter_day <- function(start,
                                                                        end,
                                                                        precision) {
  precision <- validate_precision_string(precision)

  start <- vec_proxy_compare(start)
  end <- vec_proxy_compare(end)

  if (precision >= PRECISION_YEAR) {
    start$year <- NULL
    end$year <- NULL
  }
  if (precision >= PRECISION_QUARTER) {
    start$quarter <- NULL
    end$quarter <- NULL
  }

  list(start = start, end = end)
}

# ------------------------------------------------------------------------------

#' Sequences: year-quarter-day
#'
#' @description
#' This is a year-quarter-day method for the [seq()] generic.
#'
#' Sequences can only be generated for `"year"` and `"quarter"` precision
#' year-quarter-day vectors.
#'
#' When calling `seq()`, exactly two of the following must be specified:
#' - `to`
#' - `by`
#' - Either `length.out` or `along.with`
#'
#' @inheritParams seq.clock_duration
#'
#' @param from `[clock_year_quarter_day(1)]`
#'
#'   A `"year"` or `"quarter"` precision year-quarter-day to start the sequence
#'   from.
#'
#'   `from` is always included in the result.
#'
#' @param to `[clock_year_quarter_day(1) / NULL]`
#'
#'   A `"year"` or `"quarter"` precision year-quarter-day to stop the sequence
#'   at.
#'
#'   `to` is cast to the type of `from`.
#'
#'   `to` is only included in the result if the resulting sequence divides
#'   the distance between `from` and `to` exactly.
#'
#' @return A sequence with the type of `from`.
#'
#' @export
#' @examples
#' # Quarterly sequence
#' x <- seq(year_quarter_day(2020, 1), year_quarter_day(2026, 3), by = 2)
#' x
#'
#' # Which we can then set the day of the quarter of
#' set_day(x, "last")
seq.clock_year_quarter_day <- function(from,
                                       to = NULL,
                                       by = NULL,
                                       length.out = NULL,
                                       along.with = NULL,
                                       ...) {
  precision <- calendar_precision_attribute(from)

  if (precision > PRECISION_QUARTER) {
    abort("`from` must be 'year' or 'quarter' precision.")
  }

  seq_impl(
    from = from,
    to = to,
    by = by,
    length.out = length.out,
    along.with = along.with,
    precision = precision,
    ...
  )
}

# ------------------------------------------------------------------------------

quarterly_start <- function(x) {
  attr(x, "start", exact = TRUE)
}

quarterly_start_prettify <- function(start, ..., abbreviate = FALSE) {
  check_dots_empty()
  if (abbreviate) {
    month.abb[start]
  } else {
    month.name[start]
  }
}

quarterly_validate_start <- function(start) {
  if (is_null(start)) {
    return(1L)
  }

  start <- vec_cast(start, integer(), x_arg = "start")

  if (!is_number(start)) {
    abort("`start` must be a single number.")
  }

  if (start < 1L || start > 12L) {
    abort("`start` must be a number between [1, 12].")
  }

  start
}

#' Integer codes
#'
#' @description
#'
#' Objects with useful mappings from month names and weekday names
#' to integer codes.
#'
#' ## Month codes (`clock_months`)
#'
#' - `january == 1`
#' - `february == 2`
#' - `march == 3`
#' - `april == 4`
#' - `may == 5`
#' - `june == 6`
#' - `july == 7`
#' - `august == 8`
#' - `september == 9`
#' - `october == 10`
#' - `november == 11`
#' - `december == 12`
#'
#' ## Weekday codes (`clock_weekdays`)
#'
#' - `sunday == 1`
#' - `monday == 2`
#' - `tuesday == 3`
#' - `wednesday == 4`
#' - `thursday == 5`
#' - `friday == 6`
#' - `saturday == 7`
#'
#' ## ISO weekday codes (`clock_iso_weekdays`)
#'
#' - `monday == 1`
#' - `tuesday == 2`
#' - `wednesday == 3`
#' - `thursday == 4`
#' - `friday == 5`
#' - `saturday == 6`
#' - `sunday == 7`
#'
#' @name clock-codes
#'
#' @examples
#' weekday(clock_weekdays$wednesday)
#'
#' year_month_weekday(2019, clock_months$april, clock_weekdays$monday, 1:4)
#'
#' iso_year_week_day(2020, 52, clock_iso_weekdays$thursday)
NULL

#' @format - `clock_months`: An environment containing month codes.
#' @rdname clock-codes
"clock_months"

#' @format - `clock_weekdays`: An environment containing weekday codes.
#' @rdname clock-codes
"clock_weekdays"

#' @format - `clock_iso_weekdays`: An environment containing ISO weekday codes.
#' @rdname clock-codes
"clock_iso_weekdays"

#' Calendar: iso-year-week-day
#'
#' `iso_year_week_day()` constructs a calendar from the ISO year, week number,
#' and week day.
#'
#' @details
#' Fields are recycled against each other.
#'
#' Fields are collected in order until the first `NULL` field is located. No
#' fields after the first `NULL` field are used.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams year_month_day
#'
#' @param year `[integer]`
#'
#'   The ISO year. Values `[-32767, 32767]` are generally allowed.
#'
#' @param week `[integer / "last" / NULL]`
#'
#'   The ISO week. Values `[1, 53]` are allowed.
#'
#'   If `"last"`, then the last week of the ISO year is returned.
#'
#' @param day `[integer / NULL]`
#'
#'   The day of the week. Values `[1, 7]` are allowed, with 1 = Monday
#'   and 7 = Sunday, in accordance with the ISO specifications.
#'
#' @return A iso-year-week-day calendar vector.
#'
#' @export
#' @examples
#' # Year-week
#' x <- iso_year_week_day(2019:2025, 1)
#' x
#'
#' # 2nd day of the first ISO week in multiple years
#' iso_days <- set_day(x, clock_iso_weekdays$tuesday)
#' iso_days
#'
#' # What year-month-day is this?
#' as_year_month_day(iso_days)
iso_year_week_day <- function(year,
                              week = NULL,
                              day = NULL,
                              hour = NULL,
                              minute = NULL,
                              second = NULL,
                              subsecond = NULL,
                              ...,
                              subsecond_precision = NULL) {
  check_dots_empty()

  # Stop on the first `NULL` argument
  if (is_null(week)) {
    precision <- PRECISION_YEAR
    fields <- list(year = year)
  } else if (is_null(day)) {
    precision <- PRECISION_WEEK
    fields <- list(year = year, week = week)
  } else if (is_null(hour)) {
    precision <- PRECISION_DAY
    fields <- list(year = year, week = week, day = day)
  } else if (is_null(minute)) {
    precision <- PRECISION_HOUR
    fields <- list(year = year, week = week, day = day, hour = hour)
  } else if (is_null(second)) {
    precision <- PRECISION_MINUTE
    fields <- list(year = year, week = week, day = day, hour = hour, minute = minute)
  } else if (is_null(subsecond)) {
    precision <- PRECISION_SECOND
    fields <- list(year = year, week = week, day = day, hour = hour, minute = minute, second = second)
  } else {
    precision <- calendar_validate_subsecond_precision(subsecond_precision)
    fields <- list(year = year, week = week, day = day, hour = hour, minute = minute, second = second, subsecond = subsecond)
  }

  if (is_last(fields$week)) {
    fields$week <- 1L
    last <- TRUE
  } else {
    last <- FALSE
  }

  fields <- vec_recycle_common(!!!fields)
  fields <- vec_cast_common(!!!fields, .to = integer())

  fields <- collect_iso_year_week_day_fields(fields, precision)

  names <- NULL

  out <- new_iso_year_week_day_from_fields(fields, precision, names)

  if (last) {
    out <- set_week(out, "last")
  }

  out
}

# ------------------------------------------------------------------------------

#' @export
vec_proxy.clock_iso_year_week_day <- function(x, ...) {
  .Call(`_clock_clock_rcrd_proxy`, x)
}

#' @export
vec_restore.clock_iso_year_week_day <- function(x, to, ...) {
  .Call(`_clock_iso_year_week_day_restore`, x, to)
}

# ------------------------------------------------------------------------------

#' @export
format.clock_iso_year_week_day <- function(x, ...) {
  out <- format_iso_year_week_day_cpp(x, calendar_precision_attribute(x))
  names(out) <- names(x)
  out
}

#' @export
vec_ptype_full.clock_iso_year_week_day <- function(x, ...) {
  calendar_ptype_full(x, "iso_year_week_day")
}

#' @export
vec_ptype_abbr.clock_iso_year_week_day <- function(x, ...) {
  calendar_ptype_abbr(x, "iso_ywd")
}

# ------------------------------------------------------------------------------

#' Is `x` a iso-year-week-day?
#'
#' Check if `x` is a iso-year-week-day.
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @return Returns `TRUE` if `x` inherits from `"clock_iso_year_week_day"`,
#'   otherwise returns `FALSE`.
#'
#' @export
#' @examples
#' is_iso_year_week_day(iso_year_week_day(2019))
is_iso_year_week_day <- function(x) {
  inherits(x, "clock_iso_year_week_day")
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype.clock_iso_year_week_day <- function(x, ...) {
  switch(
    calendar_precision_attribute(x) + 1L,
    clock_empty_iso_year_week_day_year,
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    clock_empty_iso_year_week_day_week,
    clock_empty_iso_year_week_day_day,
    clock_empty_iso_year_week_day_hour,
    clock_empty_iso_year_week_day_minute,
    clock_empty_iso_year_week_day_second,
    clock_empty_iso_year_week_day_millisecond,
    clock_empty_iso_year_week_day_microsecond,
    clock_empty_iso_year_week_day_nanosecond,
    abort("Internal error: Invalid precision.")
  )
}

#' @export
vec_ptype2.clock_iso_year_week_day.clock_iso_year_week_day <- function(x, y, ...) {
  ptype2_calendar_and_calendar(x, y, ...)
}

#' @export
vec_cast.clock_iso_year_week_day.clock_iso_year_week_day <- function(x, to, ...) {
  cast_calendar_to_calendar(x, to, ...)
}

# ------------------------------------------------------------------------------

#' @export
calendar_is_valid_precision.clock_iso_year_week_day <- function(x, precision) {
  iso_year_week_day_is_valid_precision(precision)
}

iso_year_week_day_is_valid_precision <- function(precision) {
  if (!is_valid_precision(precision)) {
    FALSE
  } else if (precision == PRECISION_YEAR || precision == PRECISION_WEEK) {
    TRUE
  } else if (precision >= PRECISION_DAY && precision <= PRECISION_NANOSECOND) {
    TRUE
  } else {
    FALSE
  }
}

# ------------------------------------------------------------------------------

#' @export
invalid_detect.clock_iso_year_week_day <- function(x) {
  invalid_detect_iso_year_week_day_cpp(x, calendar_precision_attribute(x))
}

#' @export
invalid_any.clock_iso_year_week_day <- function(x) {
  invalid_any_iso_year_week_day_cpp(x, calendar_precision_attribute(x))
}

#' @export
invalid_count.clock_iso_year_week_day <- function(x) {
  invalid_count_iso_year_week_day_cpp(x, calendar_precision_attribute(x))
}

#' @export
invalid_resolve.clock_iso_year_week_day <- function(x, ..., invalid = NULL) {
  check_dots_empty()
  precision <- calendar_precision_attribute(x)
  invalid <- validate_invalid(invalid)
  fields <- invalid_resolve_iso_year_week_day_cpp(x, precision, invalid)
  new_iso_year_week_day_from_fields(fields, precision, names = names(x))
}

# ------------------------------------------------------------------------------

#' Getters: iso-year-week-day
#'
#' @description
#' These are iso-year-week-day methods for the
#' [getter generics][clock-getters].
#'
#' - `get_year()` returns the ISO year. Note that this can differ from the
#'   Gregorian year.
#'
#' - `get_week()` returns the ISO week of the current ISO year.
#'
#' - `get_day()` returns a value between 1-7 indicating the weekday of the
#'   current ISO week, where 1 = Monday and 7 = Sunday, in line with the
#'   ISO standard.
#'
#' - There are sub-daily getters for extracting more precise components.
#'
#' @param x `[clock_iso_year_week_day]`
#'
#'   A iso-year-week-day to get the component from.
#'
#' @return The component.
#'
#' @name iso-year-week-day-getters
#' @examples
#' x <- iso_year_week_day(2019, 50:52, 1:3)
#' x
#'
#' # Get the ISO week
#' get_week(x)
#'
#' # Gets the weekday, 1 = Monday, 7 = Sunday
#' get_day(x)
#'
#' # Note that the ISO year can differ from the Gregorian year
#' iso <- iso_year_week_day(2019, 1, 1)
#' ymd <- as_year_month_day(iso)
#'
#' get_year(iso)
#' get_year(ymd)
NULL

#' @rdname iso-year-week-day-getters
#' @export
get_year.clock_iso_year_week_day <- function(x) {
  field_year(x)
}

#' @rdname iso-year-week-day-getters
#' @export
get_week.clock_iso_year_week_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_WEEK, "get_week")
  field_week(x)
}

#' @rdname iso-year-week-day-getters
#' @export
get_day.clock_iso_year_week_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_DAY, "get_day")
  field_day(x)
}

#' @rdname iso-year-week-day-getters
#' @export
get_hour.clock_iso_year_week_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_HOUR, "get_hour")
  field_hour(x)
}

#' @rdname iso-year-week-day-getters
#' @export
get_minute.clock_iso_year_week_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_MINUTE, "get_minute")
  field_minute(x)
}

#' @rdname iso-year-week-day-getters
#' @export
get_second.clock_iso_year_week_day <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_SECOND, "get_second")
  field_second(x)
}

#' @rdname iso-year-week-day-getters
#' @export
get_millisecond.clock_iso_year_week_day <- function(x) {
  calendar_require_precision(x, PRECISION_MILLISECOND, "get_millisecond")
  field_subsecond(x)
}

#' @rdname iso-year-week-day-getters
#' @export
get_microsecond.clock_iso_year_week_day <- function(x) {
  calendar_require_precision(x, PRECISION_MICROSECOND, "get_microsecond")
  field_subsecond(x)
}

#' @rdname iso-year-week-day-getters
#' @export
get_nanosecond.clock_iso_year_week_day <- function(x) {
  calendar_require_precision(x, PRECISION_NANOSECOND, "get_nanosecond")
  field_subsecond(x)
}

# ------------------------------------------------------------------------------

#' Setters: iso-year-week-day
#'
#' @description
#' These are iso-year-week-day methods for the
#' [setter generics][clock-setters].
#'
#' - `set_year()` sets the ISO year.
#'
#' - `set_week()` sets the ISO week of the year. Valid values are in the range
#'   of `[1, 53]`.
#'
#' - `set_day()` sets the day of the week. Valid values are in the range of
#'   `[1, 7]`, with 1 = Monday, and 7 = Sunday.
#'
#' - There are sub-daily setters for setting more precise components.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[clock_iso_year_week_day]`
#'
#'   A iso-year-week-day vector.
#'
#' @param value `[integer / "last"]`
#'
#'   The value to set the component to.
#'
#'   For `set_week()`, this can also be `"last"` to adjust to the last
#'   week of the current ISO year.
#'
#' @return `x` with the component set.
#'
#' @name iso-year-week-day-setters
#' @examples
#' # Year precision vector
#' x <- iso_year_week_day(2019:2023)
#'
#' # Promote to week precision by setting the week
#' # (Note that some ISO weeks have 52 weeks, and others have 53)
#' x <- set_week(x, "last")
#' x
#'
#' # Set to an invalid week
#' invalid <- set_week(x, 53)
#' invalid
#'
#' # Here are the invalid ones (they only have 52 weeks)
#' invalid[invalid_detect(invalid)]
#'
#' # Resolve the invalid dates by choosing the previous/next valid moment
#' invalid_resolve(invalid, invalid = "previous")
#' invalid_resolve(invalid, invalid = "next")
NULL

#' @rdname iso-year-week-day-setters
#' @export
set_year.clock_iso_year_week_day <- function(x, value, ...) {
  check_dots_empty()
  set_field_iso_year_week_day(x, value, "year")
}

#' @rdname iso-year-week-day-setters
#' @export
set_week.clock_iso_year_week_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_YEAR, "set_week")
  set_field_iso_year_week_day(x, value, "week")
}

#' @rdname iso-year-week-day-setters
#' @export
set_day.clock_iso_year_week_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_WEEK, "set_day")
  set_field_iso_year_week_day(x, value, "day")
}

#' @rdname iso-year-week-day-setters
#' @export
set_hour.clock_iso_year_week_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_DAY, "set_hour")
  set_field_iso_year_week_day(x, value, "hour")
}

#' @rdname iso-year-week-day-setters
#' @export
set_minute.clock_iso_year_week_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_HOUR, "set_minute")
  set_field_iso_year_week_day(x, value, "minute")
}

#' @rdname iso-year-week-day-setters
#' @export
set_second.clock_iso_year_week_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_MINUTE, "set_second")
  set_field_iso_year_week_day(x, value, "second")
}

#' @rdname iso-year-week-day-setters
#' @export
set_millisecond.clock_iso_year_week_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_MILLISECOND), "set_millisecond")
  set_field_iso_year_week_day(x, value, "millisecond")
}

#' @rdname iso-year-week-day-setters
#' @export
set_microsecond.clock_iso_year_week_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_MICROSECOND), "set_microsecond")
  set_field_iso_year_week_day(x, value, "microsecond")
}

#' @rdname iso-year-week-day-setters
#' @export
set_nanosecond.clock_iso_year_week_day <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_NANOSECOND), "set_nanosecond")
  set_field_iso_year_week_day(x, value, "nanosecond")
}

set_field_iso_year_week_day <- function(x, value, component) {
  if (is_last(value) && identical(component, "week")) {
    return(set_field_iso_year_week_day_last(x))
  }

  precision_fields <- calendar_precision_attribute(x)
  precision_value <- iso_year_week_day_component_to_precision(component)
  precision_out <- precision_common2(precision_fields, precision_value)

  value <- vec_cast(value, integer(), x_arg = "value")
  args <- vec_recycle_common(x = x, value = value)
  x <- args$x
  value <- args$value

  result <- set_field_iso_year_week_day_cpp(x, value, precision_fields, precision_value)
  fields <- result$fields
  field <- iso_year_week_day_component_to_field(component)
  fields[[field]] <- result$value

  new_iso_year_week_day_from_fields(fields, precision_out, names = names(x))
}

set_field_iso_year_week_day_last <- function(x) {
  precision_fields <- calendar_precision_attribute(x)
  precision_out <- precision_common2(precision_fields, PRECISION_WEEK)

  result <- set_field_iso_year_week_day_last_cpp(x, precision_fields)
  fields <- result$fields
  fields[["week"]] <- result$value

  new_iso_year_week_day_from_fields(fields, precision_out, names = names(x))
}

# ------------------------------------------------------------------------------

#' @export
calendar_name.clock_iso_year_week_day <- function(x) {
  "iso_year_week_day"
}

# ------------------------------------------------------------------------------

iso_year_week_day_component_to_precision <- function(component) {
  switch(
    component,
    year = PRECISION_YEAR,
    week = PRECISION_WEEK,
    day = PRECISION_DAY,
    hour = PRECISION_HOUR,
    minute = PRECISION_MINUTE,
    second = PRECISION_SECOND,
    millisecond = PRECISION_MILLISECOND,
    microsecond = PRECISION_MICROSECOND,
    nanosecond = PRECISION_NANOSECOND,
    abort("Internal error: Unknown component name.")
  )
}

iso_year_week_day_component_to_field <- function(component) {
  switch (
    component,
    year = component,
    week = component,
    day = component,
    hour = component,
    minute = component,
    second = component,
    millisecond = "subsecond",
    microsecond = "subsecond",
    nanosecond = "subsecond",
    abort("Internal error: Unknown component name.")
  )
}

# ------------------------------------------------------------------------------

#' @rdname clock-arith
#' @method vec_arith clock_iso_year_week_day
#' @export
vec_arith.clock_iso_year_week_day <- function(op, x, y, ...) {
  UseMethod("vec_arith.clock_iso_year_week_day", y)
}

#' @method vec_arith.clock_iso_year_week_day MISSING
#' @export
vec_arith.clock_iso_year_week_day.MISSING <- function(op, x, y, ...) {
  arith_calendar_and_missing(op, x, y, ...)
}

#' @method vec_arith.clock_iso_year_week_day clock_iso_year_week_day
#' @export
vec_arith.clock_iso_year_week_day.clock_iso_year_week_day <- function(op, x, y, ...) {
  arith_calendar_and_calendar(op, x, y, ..., calendar_minus_calendar_fn = iso_year_week_day_minus_iso_year_week_day)
}

#' @method vec_arith.clock_iso_year_week_day clock_duration
#' @export
vec_arith.clock_iso_year_week_day.clock_duration <- function(op, x, y, ...) {
  arith_calendar_and_duration(op, x, y, ...)
}

#' @method vec_arith.clock_duration clock_iso_year_week_day
#' @export
vec_arith.clock_duration.clock_iso_year_week_day <- function(op, x, y, ...) {
  arith_duration_and_calendar(op, x, y, ...)
}

#' @method vec_arith.clock_iso_year_week_day numeric
#' @export
vec_arith.clock_iso_year_week_day.numeric <- function(op, x, y, ...) {
  arith_calendar_and_numeric(op, x, y, ...)
}

#' @method vec_arith.numeric clock_iso_year_week_day
#' @export
vec_arith.numeric.clock_iso_year_week_day <- function(op, x, y, ...) {
  arith_numeric_and_calendar(op, x, y, ...)
}

iso_year_week_day_minus_iso_year_week_day <- function(op, x, y, ...) {
  args <- vec_recycle_common(x = x, y = y)
  args <- vec_cast_common(!!!args)
  x <- args$x
  y <- args$y

  names <- names_common(x, y)

  precision <- calendar_precision_attribute(x)

  if (precision > PRECISION_YEAR) {
    stop_incompatible_op(op, x, y, ...)
  }

  fields <- iso_year_week_day_minus_iso_year_week_day_cpp(x, y, precision)

  new_duration_from_fields(fields, precision, names = names)
}

# ------------------------------------------------------------------------------

#' Arithmetic: iso-year-week-day
#'
#' @description
#' These are iso-year-week-day methods for the
#' [arithmetic generics][clock-arithmetic].
#'
#' - `add_years()`
#'
#' You cannot add weeks or days to an iso-year-week-day calendar. Adding
#' days is much more efficiently done by converting to a time point first
#' by using [as_naive_time()] or [as_sys_time()]. Adding weeks is equally
#' as efficient as adding 7 days. Additionally, adding weeks to an invalid
#' iso-year-week object containing `iso_year_week_day(2019, 53)` would be
#' undefined, as the 53rd ISO week of 2019 doesn't exist to begin with.
#'
#' @details
#' `x` and `n` are recycled against each other.
#'
#' @inheritParams add_years
#'
#' @param x `[clock_iso_year_week_day]`
#'
#'   A iso-year-week-day vector.
#'
#' @return `x` after performing the arithmetic.
#'
#' @name iso-year-week-day-arithmetic
#'
#' @examples
#' x <- iso_year_week_day(2019, 1, 1)
#' add_years(x, 1:2)
NULL

#' @rdname iso-year-week-day-arithmetic
#' @export
add_years.clock_iso_year_week_day <- function(x, n, ...) {
  iso_year_week_day_plus_duration(x, n, PRECISION_YEAR)
}

iso_year_week_day_plus_duration <- function(x, n, precision_n) {
  precision_fields <- calendar_precision_attribute(x)

  n <- duration_collect_n(n, precision_n)
  args <- vec_recycle_common(x = x, n = n)
  x <- args$x
  n <- args$n

  names <- names_common(x, n)

  fields <- iso_year_week_day_plus_duration_cpp(x, n, precision_fields, precision_n)

  new_iso_year_week_day_from_fields(fields, precision_fields, names = names)
}

# ------------------------------------------------------------------------------

#' Convert to iso-year-week-day
#'
#' `as_iso_year_week_day()` converts a vector to the iso-year-week-day
#' calendar. Time points, Dates, POSIXct, and other calendars can all be
#' converted to iso-year-week-day.
#'
#' @param x `[vector]`
#'
#'   A vector to convert to iso-year-week-day.
#'
#' @return A iso-year-week-day vector.
#' @export
#' @examples
#' # From Date
#' as_iso_year_week_day(as.Date("2019-01-01"))
#'
#' # From POSIXct, which assumes that the naive time is what should be converted
#' as_iso_year_week_day(as.POSIXct("2019-01-01 02:30:30", "America/New_York"))
#'
#' # From other calendars
#' as_iso_year_week_day(year_quarter_day(2019, quarter = 2, day = 50))
as_iso_year_week_day <- function(x)  {
  UseMethod("as_iso_year_week_day")
}

#' @export
as_iso_year_week_day.default <- function(x) {
  stop_clock_unsupported_conversion(x, "clock_iso_year_week_day")
}

#' @export
as_iso_year_week_day.clock_iso_year_week_day <- function(x) {
  x
}

# ------------------------------------------------------------------------------

#' @export
as_sys_time.clock_iso_year_week_day <- function(x) {
  calendar_require_all_valid(x)
  precision <- calendar_precision_attribute(x)
  fields <- as_sys_time_iso_year_week_day_cpp(x, precision)
  new_sys_time_from_fields(fields, precision, clock_rcrd_names(x))
}

#' @export
as_naive_time.clock_iso_year_week_day <- function(x) {
  as_naive_time(as_sys_time(x))
}

#' @export
as.character.clock_iso_year_week_day <- function(x, ...) {
  format(x)
}

# ------------------------------------------------------------------------------

#' Grouping: iso-year-week-day
#'
#' @description
#' This is a iso-year-week-day method for the [calendar_group()] generic.
#'
#' Grouping for a iso-year-week-day object can be done at any precision, as
#' long as `x` is at least as precise as `precision`.
#'
#' @inheritParams calendar_group
#'
#' @param x `[clock_iso_year_week_day]`
#'
#'   A iso-year-week-day vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"week"`
#'   - `"day"`
#'   - `"hour"`
#'   - `"minute"`
#'   - `"second"`
#'   - `"millisecond"`
#'   - `"microsecond"`
#'   - `"nanosecond"`
#'
#' @return `x` grouped at the specified `precision`.
#'
#' @name iso-year-week-day-group
#'
#' @export
#' @examples
#' x <- iso_year_week_day(2019, 1:52)
#'
#' # Group by 3 ISO weeks
#' calendar_group(x, "week", n = 3)
#'
#' y <- iso_year_week_day(2000:2020, 1, 1)
#'
#' # Group by 2 ISO years
#' calendar_group(y, "year", n = 2)
calendar_group.clock_iso_year_week_day <- function(x, precision, ..., n = 1L) {
  n <- validate_calendar_group_n(n)
  x <- calendar_narrow(x, precision)

  precision <- validate_precision_string(precision)

  if (precision == PRECISION_YEAR) {
    value <- get_year(x)
    value <- group_component0(value, n)
    x <- set_year(x, value)
    return(x)
  }
  if (precision == PRECISION_WEEK) {
    value <- get_week(x)
    value <- group_component1(value, n)
    x <- set_week(x, value)
    return(x)
  }
  if (precision == PRECISION_DAY) {
    value <- get_day(x)
    value <- group_component1(value, n)
    x <- set_day(x, value)
    return(x)
  }

  x <- calendar_group_time(x, n, precision)
  x
}

# ------------------------------------------------------------------------------

#' Narrow: iso-year-week-day
#'
#' This is a iso-year-week-day method for the [calendar_narrow()] generic. It
#' narrows a iso-year-week-day vector to the specified `precision`.
#'
#' @inheritParams iso-year-week-day-group
#'
#' @return `x` narrowed to the supplied `precision`.
#'
#' @name iso-year-week-day-narrow
#'
#' @export
#' @examples
#' # Day precision
#' x <- iso_year_week_day(2019, 1, 5)
#' x
#'
#' # Narrowed to week precision
#' calendar_narrow(x, "week")
calendar_narrow.clock_iso_year_week_day <- function(x, precision) {
  precision <- validate_precision_string(precision)

  out_fields <- list()
  x_fields <- unclass(x)

  if (precision >= PRECISION_YEAR) {
    out_fields[["year"]] <- x_fields[["year"]]
  }
  if (precision >= PRECISION_WEEK) {
    out_fields[["week"]] <- x_fields[["week"]]
  }
  if (precision >= PRECISION_DAY) {
    out_fields[["day"]] <- x_fields[["day"]]
  }

  out_fields <- calendar_narrow_time(out_fields, precision, x_fields)

  new_iso_year_week_day_from_fields(out_fields, precision, names = names(x))
}

# ------------------------------------------------------------------------------

#' Widen: iso-year-week-day
#'
#' This is a iso-year-week-day method for the [calendar_widen()] generic. It
#' widens a iso-year-week-day vector to the specified `precision`.
#'
#' @inheritParams iso-year-week-day-group
#'
#' @return `x` widened to the supplied `precision`.
#'
#' @name iso-year-week-day-widen
#'
#' @export
#' @examples
#' # Week precision
#' x <- iso_year_week_day(2019, 1)
#' x
#'
#' # Widen to day precision
#' # In the ISO calendar, the first day of the week is a Monday
#' calendar_widen(x, "day")
#'
#' # Or second precision
#' sec <- calendar_widen(x, "second")
#' sec
calendar_widen.clock_iso_year_week_day <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  if (precision >= PRECISION_WEEK && x_precision < PRECISION_WEEK) {
    x <- set_week(x, 1L)
  }
  if (precision >= PRECISION_DAY && x_precision < PRECISION_DAY) {
    x <- set_day(x, 1L)
  }

  x <- calendar_widen_time(x, x_precision, precision)

  x
}

# ------------------------------------------------------------------------------

#' Boundaries: iso-year-week-day
#'
#' This is an iso-year-week-day method for the [calendar_start()] and
#' [calendar_end()] generics. They adjust components of a calendar to the
#' start or end of a specified `precision`.
#'
#' @inheritParams iso-year-week-day-group
#'
#' @return `x` at the same precision, but with some components altered to be
#'   at the boundary value.
#'
#' @name iso-year-week-day-boundary
#'
#' @examples
#' x <- iso_year_week_day(2019:2020, 5, 6, 10)
#' x
#'
#' # Compute the last moment of the last iso week of the year
#' calendar_end(x, "year")
#'
#' # Compare that to just setting the week to `"last"`,
#' # which doesn't affect the other components
#' set_week(x, "last")
NULL

#' @rdname iso-year-week-day-boundary
#' @export
calendar_start.clock_iso_year_week_day <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  calendar_start_end_checks(x, x_precision, precision, "start")

  if (precision <= PRECISION_YEAR && x_precision > PRECISION_YEAR) {
    x <- set_week(x, 1L)
  }
  if (precision <= PRECISION_WEEK && x_precision > PRECISION_WEEK) {
    x <- set_day(x, 1L)
  }

  x <- calendar_start_time(x, x_precision, precision)

  x
}

#' @rdname iso-year-week-day-boundary
#' @export
calendar_end.clock_iso_year_week_day <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  calendar_start_end_checks(x, x_precision, precision, "end")

  if (precision <= PRECISION_YEAR && x_precision > PRECISION_YEAR) {
    x <- set_week(x, "last")
  }
  if (precision <= PRECISION_WEEK && x_precision > PRECISION_WEEK) {
    x <- set_day(x, 7L)
  }

  x <- calendar_end_time(x, x_precision, precision)

  x
}

# ------------------------------------------------------------------------------

#' Counting: iso-year-week-day
#'
#' This is an iso-year-week-day method for the [calendar_count_between()]
#' generic. It counts the number of `precision` units between `start` and `end`
#' (i.e., the number of ISO years).
#'
#' @inheritParams calendar-count-between
#'
#' @param start,end `[clock_iso_year_week_day]`
#'
#'   A pair of iso-year-week-day vectors. These will be recycled to their
#'   common size.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'
#' @inherit calendar-count-between return
#'
#' @name iso-year-week-day-count-between
#'
#' @export
#' @examples
#' # Compute the number of whole ISO years between two dates
#' x <- iso_year_week_day(2001, 1, 2)
#' y <- iso_year_week_day(2021, 1, c(1, 3))
#' calendar_count_between(x, y, "year")
calendar_count_between.clock_iso_year_week_day <- function(start,
                                                           end,
                                                           precision,
                                                           ...,
                                                           n = 1L) {
  NextMethod()
}

calendar_count_between_standardize_precision_n.clock_iso_year_week_day <- function(x,
                                                                                   precision,
                                                                                   n) {
  precision_int <- validate_precision_string(precision)

  allowed_precisions <- c(PRECISION_YEAR)

  if (!(precision_int %in% allowed_precisions)) {
    abort("`precision` must be one of: 'year'.")
  }

  list(precision = precision, n = n)
}

calendar_count_between_compute.clock_iso_year_week_day <- function(start,
                                                                   end,
                                                                   precision) {
  precision <- validate_precision_string(precision)

  if (precision == PRECISION_YEAR) {
    out <- get_year(end) - get_year(start)
    return(out)
  }

  abort("Internal error: `precision` should be 'year' at this point.")
}

calendar_count_between_proxy_compare.clock_iso_year_week_day <- function(start,
                                                                         end,
                                                                         precision) {
  precision <- validate_precision_string(precision)

  start <- vec_proxy_compare(start)
  end <- vec_proxy_compare(end)

  if (precision >= PRECISION_YEAR) {
    start$year <- NULL
    end$year <- NULL
  }

  list(start = start, end = end)
}

# ------------------------------------------------------------------------------

#' Sequences: iso-year-week-day
#'
#' @description
#' This is a iso-year-week-day method for the [seq()] generic.
#'
#' Sequences can only be generated for `"year"` precision
#' iso-year-week-day vectors. If you need to generate week-based sequences,
#' you'll have to convert to a time point first.
#'
#' When calling `seq()`, exactly two of the following must be specified:
#' - `to`
#' - `by`
#' - Either `length.out` or `along.with`
#'
#' @inheritParams seq.clock_duration
#'
#' @param from `[clock_iso_year_week_day(1)]`
#'
#'   A `"year"` precision iso-year-week-day to start the sequence
#'   from.
#'
#'   `from` is always included in the result.
#'
#' @param to `[clock_iso_year_week_day(1) / NULL]`
#'
#'   A `"year"` precision iso-year-week-day to stop the sequence
#'   at.
#'
#'   `to` is cast to the type of `from`.
#'
#'   `to` is only included in the result if the resulting sequence divides
#'   the distance between `from` and `to` exactly.
#'
#' @return A sequence with the type of `from`.
#'
#' @export
#' @examples
#' # Yearly sequence
#' x <- seq(iso_year_week_day(2020), iso_year_week_day(2026), by = 2)
#' x
#'
#' # Which we can then set the week of.
#' # Some years have 53 ISO weeks, some have 52.
#' set_week(x, "last")
seq.clock_iso_year_week_day <- function(from,
                                        to = NULL,
                                        by = NULL,
                                        length.out = NULL,
                                        along.with = NULL,
                                        ...) {
  precision <- calendar_precision_attribute(from)

  if (precision > PRECISION_YEAR) {
    abort("`from` must be 'year' precision.")
  }

  seq_impl(
    from = from,
    to = to,
    by = by,
    length.out = length.out,
    along.with = along.with,
    precision = precision,
    ...
  )
}

# ------------------------------------------------------------------------------

clock_init_iso_year_week_day_utils <- function(env) {
  year <- iso_year_week_day(integer())

  assign("clock_empty_iso_year_week_day_year", year, envir = env)
  assign("clock_empty_iso_year_week_day_week", calendar_widen(year, "week"), envir = env)
  assign("clock_empty_iso_year_week_day_day", calendar_widen(year, "day"), envir = env)
  assign("clock_empty_iso_year_week_day_hour", calendar_widen(year, "hour"), envir = env)
  assign("clock_empty_iso_year_week_day_minute", calendar_widen(year, "minute"), envir = env)
  assign("clock_empty_iso_year_week_day_second", calendar_widen(year, "second"), envir = env)
  assign("clock_empty_iso_year_week_day_millisecond", calendar_widen(year, "millisecond"), envir = env)
  assign("clock_empty_iso_year_week_day_microsecond", calendar_widen(year, "microsecond"), envir = env)
  assign("clock_empty_iso_year_week_day_nanosecond", calendar_widen(year, "nanosecond"), envir = env)

  invisible(NULL)
}

#' Is `x` a zoned-time?
#'
#' This function determines if the input is a zoned-time object.
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @return `TRUE` if `x` inherits from `"clock_zoned_time"`, otherwise `FALSE`.
#'
#' @export
#' @examples
#' is_zoned_time(1)
#' is_zoned_time(zoned_time_now("America/New_York"))
is_zoned_time <- function(x) {
  inherits(x, "clock_zoned_time")
}

# ------------------------------------------------------------------------------

zoned_time_zone_attribute <- function(x) {
  attr(x, "zone", exact = TRUE)
}
zoned_time_set_zone_attribute <- function(x, zone) {
  attr(x, "zone") <- zone
  x
}

zoned_time_precision_attribute <- function(x) {
  attr(x, "precision", exact = TRUE)
}

# ------------------------------------------------------------------------------

#' Formatting: zoned-time
#'
#' @description
#' This is a zoned-time method for the [format()] generic.
#'
#' This function allows you to format a zoned-time using a flexible `format`
#' string.
#'
#' If `format` is `NULL`, a default format of `"%Y-%m-%dT%H:%M:%S%Ez[%Z]"` is
#' used. This matches the default format that [zoned_time_parse_complete()]
#' parses. Additionally, this format matches the de-facto standard extension to
#' RFC 3339 for creating completely unambiguous date-times.
#'
#' @param x `[clock_zoned_time]`
#'
#'   A zoned-time.
#'
#' @param ... `[dots]`
#'
#'   Not used, but no error will be thrown if not empty to remain compatible
#'   with usage of the `format()` generic.
#'
#' @param format `[character(1) / NULL]`
#'
#'   If `NULL`, a default format is used, which depends on the type of the
#'   input.
#'
#'   Otherwise, a format string which is a combination of:
#'
#'   **Year**
#'
#'   - `%C`: The year divided by 100 using floored division. If the result
#'   is a single decimal digit, it is prefixed with `0`.
#'
#'   - `%y`: The last two decimal digits of the year. If the result is a single
#'   digit it is prefixed by `0`.
#'
#'   - `%Y`: The year as a decimal number. If the result is less than four
#'   digits it is left-padded with `0` to four digits.
#'
#'   **Month**
#'
#'   - `%b`, `%h`: The `locale`'s abbreviated month name.
#'
#'   - `%B`: The `locale`'s full month name.
#'
#'   - `%m`: The month as a decimal number. January is `01`. If the result is a
#'   single digit, it is prefixed with `0`.
#'
#'   **Day**
#'
#'   - `%d`: The day of month as a decimal number. If the result is a single
#'   decimal digit, it is prefixed with `0`.
#'
#'   **Day of the week**
#'
#'   - `%a`: The `locale`'s abbreviated weekday name.
#'
#'   - `%A`: The `locale`'s full weekday name.
#'
#'   - `%w`: The weekday as a decimal number (`0-6`), where Sunday is `0`.
#'
#'   **ISO 8601 week-based year**
#'
#'   - `%g`: The last two decimal digits of the ISO week-based year. If the
#'   result is a single digit it is prefixed by `0`.
#'
#'   - `%G`: The ISO week-based year as a decimal number. If the result is less
#'   than four digits it is left-padded with `0` to four digits.
#'
#'   - `%V`: The ISO week-based week number as a decimal number. If the result
#'   is a single digit, it is prefixed with `0`.
#'
#'   - `%u`: The ISO weekday as a decimal number (`1-7`), where Monday is `1`.
#'
#'   **Week of the year**
#'
#'   - `%U`: The week number of the year as a decimal number. The first Sunday
#'   of the year is the first day of week `01`. Days of the same year prior to
#'   that are in week `00`. If the result is a single digit, it is prefixed with
#'   `0`.
#'
#'   - `%W`: The week number of the year as a decimal number. The first Monday
#'   of the year is the first day of week `01`. Days of the same year prior to
#'   that are in week `00`. If the result is a single digit, it is prefixed with
#'   `0`.
#'
#'   **Day of the year**
#'
#'   - `%j`: The day of the year as a decimal number. January 1 is `001`. If the
#'   result is less than three digits, it is left-padded with `0` to three
#'   digits.
#'
#'   **Date**
#'
#'   - `%D`, `%x`: Equivalent to `%m/%d/%y`.
#'
#'   - `%F`: Equivalent to `%Y-%m-%d`.
#'
#'   **Time of day**
#'
#'   - `%H`: The hour (24-hour clock) as a decimal number. If the result is a
#'   single digit, it is prefixed with `0`.
#'
#'   - `%I`: The hour (12-hour clock) as a decimal number. If the result is a
#'   single digit, it is prefixed with `0`.
#'
#'   - `%M`: The minute as a decimal number. If the result is a single digit, it
#'   is prefixed with `0`.
#'
#'   - `%S`: Seconds as a decimal number. Fractional seconds are printed at the
#'   precision of the input. The character for the decimal point is localized
#'   according to `locale`.
#'
#'   - `%p`: The `locale`'s equivalent of the AM/PM designations associated with
#'   a 12-hour clock.
#'
#'   - `%R`: Equivalent to `%H:%M`.
#'
#'   - `%T`, `%X`: Equivalent to `%H:%M:%S`.
#'
#'   - `%r`: Nearly equivalent to `%I:%M:%S %p`, but seconds are always printed
#'   at second precision.
#'
#'   **Time zone**
#'
#'   - `%z`: The offset from UTC in the ISO 8601 format. For example `-0430`
#'   refers to 4 hours 30 minutes behind UTC. If the offset is zero, `+0000` is
#'   used. The modified command `%Ez` inserts a `:` between the hour and
#'   minutes, like `-04:30`.
#'
#'   - `%Z`: The full time zone name. If `abbreviate_zone` is `TRUE`, the time
#'   zone abbreviation.
#'
#'   **Miscellaneous**
#'
#'   - `%c`: A date and time representation. Similar to, but not exactly the
#'   same as, `%a %b %d %H:%M:%S %Y`.
#'
#'   - `%%`: A `%` character.
#'
#'   - `%n`: A newline character.
#'
#'   - `%t`: A horizontal-tab character.
#'
#' @param locale `[clock_locale]`
#'
#'   A locale object created from [clock_locale()].
#'
#' @param abbreviate_zone `[logical(1)]`
#'
#'   If `TRUE`, `%Z` returns an abbreviated time zone name.
#'
#'   If `FALSE`, `%Z` returns the full time zone name.
#'
#' @return A character vector of the formatted input.
#'
#' @export
#' @examples
#' x <- year_month_day(2019, 1, 1)
#' x <- as_zoned_time(as_naive_time(x), "America/New_York")
#'
#' format(x)
#' format(x, format = "%B %d, %Y")
#' format(x, format = "%B %d, %Y", locale = clock_locale("fr"))
format.clock_zoned_time <- function(x,
                                    ...,
                                    format = NULL,
                                    locale = clock_locale(),
                                    abbreviate_zone = FALSE) {
  if (!is_clock_locale(locale)) {
    abort("`locale` must be a 'clock_locale' object.")
  }

  zone <- zoned_time_zone_attribute(x)
  precision <- zoned_time_precision_attribute(x)

  if (is_null(format)) {
    # Collect internal option
    print_zone_name <- zoned_time_print_zone_name(...)
    format <- zoned_time_format(print_zone_name)
  }

  labels <- locale$labels
  decimal_mark <- locale$decimal_mark

  out <- format_zoned_time_cpp(
    fields = x,
    zone = zone,
    abbreviate_zone = abbreviate_zone,
    format = format,
    precision_int = precision,
    month = labels$month,
    month_abbrev = labels$month_abbrev,
    weekday = labels$weekday,
    weekday_abbrev = labels$weekday_abbrev,
    am_pm = labels$am_pm,
    decimal_mark = decimal_mark
  )

  names(out) <- clock_rcrd_names(x)

  out
}

zoned_time_print_zone_name <- function(..., print_zone_name = TRUE) {
  print_zone_name
}
zoned_time_format <- function(print_zone_name) {
  if (print_zone_name) {
    "%Y-%m-%dT%H:%M:%S%Ez[%Z]"
  } else {
    "%Y-%m-%dT%H:%M:%S%Ez"
  }
}

# ------------------------------------------------------------------------------

#' Parsing: zoned-time
#'
#' @description
#' There are two parsers into a zoned-time, `zoned_time_parse_complete()` and
#' `zoned_time_parse_abbrev()`.
#'
#' ## zoned_time_parse_complete()
#'
#' `zoned_time_parse_complete()` is a parser for _complete_ date-time strings,
#' like `"2019-01-01T00:00:00-05:00[America/New_York]"`. A complete date-time
#' string has both the time zone offset and full time zone name in the string,
#' which is the only way for the string itself to contain all of the information
#' required to construct a zoned-time. Because of this,
#' `zoned_time_parse_complete()` requires both the `%z` and `%Z` commands to be
#' supplied in the `format` string.
#'
#' The default options assume that `x` should be parsed at second precision,
#' using a `format` string of `"%Y-%m-%dT%H:%M:%S%Ez[%Z]"`. This matches the
#' default result from calling `format()` on a zoned-time. Additionally, this
#' format matches the de-facto standard extension to RFC 3339 for creating
#' completely unambiguous date-times.
#'
#' ## zoned_time_parse_abbrev()
#'
#' `zoned_time_parse_abbrev()` is a parser for date-time strings containing only
#' a time zone abbreviation, like `"2019-01-01 00:00:00 EST"`. The time zone
#' abbreviation is not enough to identify the full time zone name that the
#' date-time belongs to, so the full time zone name must be supplied as the
#' `zone` argument. However, the time zone abbreviation can help with resolving
#' ambiguity around daylight saving time fallbacks.
#'
#' For `zoned_time_parse_abbrev()`, `%Z` must be supplied and is interpreted as
#' the time zone abbreviation rather than the full time zone name.
#'
#' If used, the `%z` command must parse correctly, but its value will be
#' completely ignored.
#'
#' The default options assume that `x` should be parsed at second precision,
#' using a `format` string of `"%Y-%m-%d %H:%M:%S %Z"`. This matches the default
#' result from calling `print()` or `format(usetz = TRUE)` on a POSIXct
#' date-time.
#'
#' @details
#' If `zoned_time_parse_complete()` is given input that is length zero, all
#' `NA`s, or completely fails to parse, then no time zone will be able to be
#' determined. In that case, the result will use `"UTC"`.
#'
#' If your date-time strings contain time zone offsets (like `-04:00`), but
#' not the full time zone name, you might need [sys_time_parse()].
#'
#' If your date-time strings don't contain time zone offsets or the full time
#' zone name, you might need to use [naive_time_parse()]. From there, if you
#' know the time zone that the date-times are supposed to be in, you can convert
#' to a zoned-time with [as_zoned_time()].
#'
#' @section Full Precision Parsing:
#'
#' It is highly recommended to parse all of the information in the date-time
#' string into a type at least as precise as the string. For example, if your
#' string has fractional seconds, but you only require seconds, specify a
#' sub-second `precision`, then round to seconds manually using whatever
#' convention is appropriate for your use case. Parsing such a string directly
#' into a second precision result is ambiguous and undefined, and is unlikely to
#' work as you might expect.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[character]`
#'
#'   A character vector to parse.
#'
#' @param zone `[character(1)]`
#'
#'   A full time zone name.
#'
#' @param format `[character / NULL]`
#'
#'   A format string. A combination of the following commands, or `NULL`,
#'   in which case a default format string is used.
#'
#'   A vector of multiple format strings can be supplied. They will be tried in
#'   the order they are provided.
#'
#'   **Year**
#'
#'   - `%C`: The century as a decimal number. The modified command `%NC` where
#'   `N` is a positive decimal integer specifies the maximum number of
#'   characters to read. If not specified, the default is `2`. Leading zeroes
#'   are permitted but not required.
#'
#'   - `%y`: The last two decimal digits of the year. If the century is not
#'   otherwise specified (e.g. with `%C`), values in the range `[69 - 99]` are
#'   presumed to refer to the years `[1969 - 1999]`, and values in the range
#'   `[00 - 68]` are presumed to refer to the years `[2000 - 2068]`. The
#'   modified command `%Ny`, where `N` is a positive decimal integer, specifies
#'   the maximum number of characters to read. If not specified, the default is
#'   `2`. Leading zeroes are permitted but not required.
#'
#'   - `%Y`: The year as a decimal number. The modified command `%NY` where `N`
#'   is a positive decimal integer specifies the maximum number of characters to
#'   read. If not specified, the default is `4`. Leading zeroes are permitted
#'   but not required.
#'
#'   **Month**
#'
#'   - `%b`, `%B`, `%h`: The `locale`'s full or abbreviated case-insensitive
#'   month name.
#'
#'   - `%m`: The month as a decimal number. January is `1`. The modified command
#'   `%Nm` where `N` is a positive decimal integer specifies the maximum number
#'   of characters to read. If not specified, the default is `2`. Leading zeroes
#'   are permitted but not required.
#'
#'   **Day**
#'
#'   - `%d`, `%e`: The day of the month as a decimal number. The modified
#'   command `%Nd` where `N` is a positive decimal integer specifies the maximum
#'   number of characters to read. If not specified, the default is `2`. Leading
#'   zeroes are permitted but not required.
#'
#'   **Day of the week**
#'
#'   - `%a`, `%A`: The `locale`'s full or abbreviated case-insensitive weekday
#'   name.
#'
#'   - `%w`: The weekday as a decimal number (`0-6`), where Sunday is `0`. The
#'   modified command `%Nw` where `N` is a positive decimal integer specifies
#'   the maximum number of characters to read. If not specified, the default is
#'   `1`. Leading zeroes are permitted but not required.
#'
#'   **ISO 8601 week-based year**
#'
#'   - `%g`: The last two decimal digits of the ISO week-based year. The
#'   modified command `%Ng` where `N` is a positive decimal integer specifies
#'   the maximum number of characters to read. If not specified, the default is
#'   `2`. Leading zeroes are permitted but not required.
#'
#'   - `%G`: The ISO week-based year as a decimal number. The modified command
#'   `%NG` where `N` is a positive decimal integer specifies the maximum number
#'   of characters to read. If not specified, the default is `4`. Leading zeroes
#'   are permitted but not required.
#'
#'   - `%V`: The ISO week-based week number as a decimal number. The modified
#'   command `%NV` where `N` is a positive decimal integer specifies the maximum
#'   number of characters to read. If not specified, the default is `2`. Leading
#'   zeroes are permitted but not required.
#'
#'   - `%u`: The ISO weekday as a decimal number (`1-7`), where Monday is `1`.
#'   The modified command `%Nu` where `N` is a positive decimal integer
#'   specifies the maximum number of characters to read. If not specified, the
#'   default is `1`. Leading zeroes are permitted but not required.
#'
#'   **Week of the year**
#'
#'   - `%U`: The week number of the year as a decimal number. The first Sunday
#'   of the year is the first day of week `01`. Days of the same year prior to
#'   that are in week `00`. The modified command `%NU` where `N` is a positive
#'   decimal integer specifies the maximum number of characters to read. If not
#'   specified, the default is `2`. Leading zeroes are permitted but not
#'   required.
#'
#'   - `%W`: The week number of the year as a decimal number. The first Monday
#'   of the year is the first day of week `01`. Days of the same year prior to
#'   that are in week `00`. The modified command `%NW` where `N` is a positive
#'   decimal integer specifies the maximum number of characters to read. If not
#'   specified, the default is `2`. Leading zeroes are permitted but not
#'   required.
#'
#'   **Day of the year**
#'
#'   - `%j`: The day of the year as a decimal number. January 1 is `1`. The
#'   modified command `%Nj` where `N` is a positive decimal integer specifies
#'   the maximum number of characters to read. If not specified, the default is
#'   `3`. Leading zeroes are permitted but not required.
#'
#'   **Date**
#'
#'   - `%D`, `%x`: Equivalent to `%m/%d/%y`.
#'
#'   - `%F`: Equivalent to `%Y-%m-%d`. If modified with a width (like `%NF`),
#'   the width is applied to only `%Y`.
#'
#'   **Time of day**
#'
#'   - `%H`: The hour (24-hour clock) as a decimal number. The modified command
#'   `%NH` where `N` is a positive decimal integer specifies the maximum number
#'   of characters to read. If not specified, the default is `2`. Leading zeroes
#'   are permitted but not required.
#'
#'   - `%I`: The hour (12-hour clock) as a decimal number. The modified command
#'   `%NI` where `N` is a positive decimal integer specifies the maximum number
#'   of characters to read. If not specified, the default is `2`. Leading zeroes
#'   are permitted but not required.
#'
#'   - `%M`: The minutes as a decimal number. The modified command `%NM` where
#'   `N` is a positive decimal integer specifies the maximum number of
#'   characters to read. If not specified, the default is `2`. Leading zeroes
#'   are permitted but not required.
#'
#'   - `%S`: The seconds as a decimal number. Leading zeroes are permitted but
#'   not required. If encountered, the `locale` determines the decimal point
#'   character. Generally, the maximum number of characters to read is
#'   determined by the precision that you are parsing at. For example, a
#'   precision of `"second"` would read a maximum of 2 characters, while a
#'   precision of `"millisecond"` would read a maximum of 6 (2 for the values
#'   before the decimal point, 1 for the decimal point, and 3 for the values
#'   after it). The modified command `%NS`, where `N` is a positive decimal
#'   integer, can be used to exactly specify the maximum number of characters to
#'   read. This is only useful if you happen to have seconds with more than 1
#'   leading zero.
#'
#'   - `%p`: The `locale`'s equivalent of the AM/PM designations associated with
#'   a 12-hour clock. The command `%I` must precede `%p` in the format string.
#'
#'   - `%R`: Equivalent to `%H:%M`.
#'
#'   - `%T`, `%X`: Equivalent to `%H:%M:%S`.
#'
#'   - `%r`: Equivalent to `%I:%M:%S %p`.
#'
#'   **Time zone**
#'
#'   - `%z`: The offset from UTC in the format `[+|-]hh[mm]`. For example
#'   `-0430` refers to 4 hours 30 minutes behind UTC. And `04` refers to 4 hours
#'   ahead of UTC. The modified command `%Ez` parses a `:` between the hours and
#'   minutes and leading zeroes on the hour field are optional:
#'   `[+|-]h[h][:mm]`. For example `-04:30` refers to 4 hours 30 minutes behind
#'   UTC. And `4` refers to 4 hours ahead of UTC.
#'
#'   - `%Z`: The full time zone name or the time zone abbreviation, depending on
#'   the function being used. A single word is parsed. This word can only
#'   contain characters that are alphanumeric, or one of `'_'`, `'/'`, `'-'` or
#'   `'+'`.
#'
#'   **Miscellaneous**
#'
#'   - `%c`: A date and time representation. Equivalent to
#'   `%a %b %d %H:%M:%S %Y`.
#'
#'   - `%%`: A `%` character.
#'
#'   - `%n`: Matches one white space character. `%n`, `%t`, and a space can be
#'   combined to match a wide range of white-space patterns. For example `"%n "`
#'   matches one or more white space characters, and `"%n%t%t"` matches one to
#'   three white space characters.
#'
#'   - `%t`: Matches zero or one white space characters.
#'
#' @param precision `[character(1)]`
#'
#'   A precision for the resulting zoned-time. One of:
#'
#'   - `"second"`
#'
#'   - `"millisecond"`
#'
#'   - `"microsecond"`
#'
#'   - `"nanosecond"`
#'
#'   Setting the `precision` determines how much information `%S` attempts
#'   to parse.
#'
#' @param locale `[clock_locale]`
#'
#'   A locale object created from [clock_locale()].
#'
#' @return A zoned-time.
#'
#' @name zoned-parsing
#'
#' @examples
#' library(magrittr)
#'
#' zoned_time_parse_complete("2019-01-01T01:02:03-05:00[America/New_York]")
#'
#' zoned_time_parse_complete(
#'   "January 21, 2019 -0500 America/New_York",
#'   format = "%B %d, %Y %z %Z"
#' )
#'
#' # Nanosecond precision
#' x <- "2019/12/31 01:05:05.123456700-05:00[America/New_York]"
#' zoned_time_parse_complete(
#'   x,
#'   format = "%Y/%m/%d %H:%M:%S%Ez[%Z]",
#'   precision = "nanosecond"
#' )
#'
#' # The `%z` offset must correspond to the true offset that would be used
#' # if the input was parsed as a naive-time and then converted to a zoned-time
#' # with `as_zoned_time()`. For example, the time that was parsed above used an
#' # offset of `-05:00`. We can confirm that this is correct with:
#' year_month_day(2019, 1, 1, 1, 2, 3) %>%
#'   as_naive_time() %>%
#'   as_zoned_time("America/New_York")
#'
#' # So the following would not parse correctly
#' zoned_time_parse_complete("2019-01-01T01:02:03-04:00[America/New_York]")
#'
#' # `%z` is useful for breaking ties in otherwise ambiguous times. For example,
#' # these two times are on either side of a daylight saving time fallback.
#' # Without the `%z` offset, you wouldn't be able to tell them apart!
#' x <- c(
#'   "1970-10-25T01:30:00-04:00[America/New_York]",
#'   "1970-10-25T01:30:00-05:00[America/New_York]"
#' )
#'
#' zoned_time_parse_complete(x)
#'
#' # If you have date-time strings with time zone abbreviations,
#' # `zoned_time_parse_abbrev()` should be able to help. The `zone` must be
#' # provided, because multiple countries may use the same time zone
#' # abbreviation. For example:
#' x <- "1970-01-01 02:30:30 IST"
#'
#' # IST = India Standard Time
#' zoned_time_parse_abbrev(x, "Asia/Kolkata")
#'
#' # IST = Israel Standard Time
#' zoned_time_parse_abbrev(x, "Asia/Jerusalem")
#'
#' # The time zone abbreviation is mainly useful for resolving ambiguity
#' # around daylight saving time fallbacks. Without the abbreviation, these
#' # date-times would be ambiguous.
#' x <- c(
#'   "1970-10-25 01:30:00 EDT",
#'   "1970-10-25 01:30:00 EST"
#' )
#' zoned_time_parse_abbrev(x, "America/New_York")
NULL

#' @rdname zoned-parsing
#' @export
zoned_time_parse_complete <- function(x,
                                      ...,
                                      format = NULL,
                                      precision = "second",
                                      locale = clock_locale()) {
  check_dots_empty()

  precision <- validate_zoned_time_precision_string(precision)

  if (!is_clock_locale(locale)) {
    abort("`locale` must be a 'clock_locale' object.")
  }

  if (is_null(format)) {
    # Use both %z and %Z
    format <- zoned_time_format(print_zone_name = TRUE)
  }

  labels <- locale$labels
  mark <- locale$decimal_mark

  result <- zoned_time_parse_complete_cpp(
    x,
    format,
    precision,
    labels$month,
    labels$month_abbrev,
    labels$weekday,
    labels$weekday_abbrev,
    labels$am_pm,
    mark
  )

  new_zoned_time_from_fields(result$fields, precision, result$zone, names(x))
}

#' @rdname zoned-parsing
#' @export
zoned_time_parse_abbrev <- function(x,
                                    zone,
                                    ...,
                                    format = NULL,
                                    precision = "second",
                                    locale = clock_locale()) {
  check_dots_empty()

  precision <- validate_zoned_time_precision_string(precision)

  if (!is_clock_locale(locale)) {
    abort("`locale` must be a 'clock_locale' object.")
  }

  if (is_null(format)) {
    # Like what R POSIXct prints
    format <- "%Y-%m-%d %H:%M:%S %Z"
  }

  labels <- locale$labels
  mark <- locale$decimal_mark

  fields <- zoned_time_parse_abbrev_cpp(
    x,
    zone,
    format,
    precision,
    labels$month,
    labels$month_abbrev,
    labels$weekday,
    labels$weekday_abbrev,
    labels$am_pm,
    mark
  )

  new_zoned_time_from_fields(fields, precision, zone, names(x))
}

# ------------------------------------------------------------------------------

# ptype2 / cast will prevent zoned times with different zones from being
# compared, so the equality proxy doesn't need to worry about the `zone`.

#' @export
vec_proxy.clock_zoned_time <- function(x, ...) {
  .Call(`_clock_clock_rcrd_proxy`, x)
}

#' @export
vec_restore.clock_zoned_time <- function(x, to, ...) {
  .Call(`_clock_zoned_time_restore`, x, to)
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype_full.clock_zoned_time <- function(x, ...) {
  zone <- zone_pretty(zoned_time_zone_attribute(x))
  precision <- zoned_time_precision_attribute(x)
  precision <- precision_to_string(precision)
  paste0("zoned_time<", precision, "><", zone, ">")
}

#' @export
vec_ptype_abbr.clock_zoned_time <- function(x, ...) {
  zone <- zone_pretty(zoned_time_zone_attribute(x))
  precision <- zoned_time_precision_attribute(x)
  precision <- precision_to_string(precision)
  precision <- precision_abbr(precision)
  paste0("zt<", precision, "><", zone, ">")
}

zone_pretty <- function(zone) {
  if (identical(zone, "")) {
    zone <- zone_current()
    zone <- paste0(zone, " (current)")
  }
  zone
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype.clock_zoned_time <- function(x, ...) {
  zone <- zoned_time_zone_attribute(x)

  ptype_utc <- switch(
    zoned_time_precision_attribute(x) + 1L,
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    clock_empty_zoned_time_utc_second,
    clock_empty_zoned_time_utc_millisecond,
    clock_empty_zoned_time_utc_microsecond,
    clock_empty_zoned_time_utc_nanosecond,
    abort("Internal error: Invalid precision.")
  )

  ptype <- zoned_time_set_zone_attribute(ptype_utc, zone)

  ptype
}

#' @export
vec_ptype2.clock_zoned_time.clock_zoned_time <- function(x, y, ...) {
  x_zone <- zoned_time_zone_attribute(x)
  y_zone <- zoned_time_zone_attribute(y)

  if (x_zone != y_zone) {
    stop_incompatible_type(x, y, ..., details = "Zones can't differ.")
  }

  x_precision <- zoned_time_precision_attribute(x)
  y_precision <- zoned_time_precision_attribute(y)

  if (x_precision >= y_precision) {
    x
  } else {
    y
  }
}

#' @export
vec_cast.clock_zoned_time.clock_zoned_time <- function(x, to, ...) {
  x_zone <- zoned_time_zone_attribute(x)
  to_zone <- zoned_time_zone_attribute(to)

  if (x_zone != to_zone) {
    stop_incompatible_cast(x, to, ..., details = "Zones can't differ.")
  }

  x_precision <- zoned_time_precision_attribute(x)
  to_precision <- zoned_time_precision_attribute(to)

  if (x_precision == to_precision) {
    return(x)
  }

  if (x_precision > to_precision) {
    stop_incompatible_cast(x, to, ..., details = "Precision would be lost.")
  }

  fields <- duration_cast_cpp(x, x_precision, to_precision)

  names <- clock_rcrd_names(x)

  new_zoned_time_from_fields(fields, to_precision, x_zone, names)
}

# ------------------------------------------------------------------------------

#' @export
print.clock_zoned_time <- function(x, ..., max = NULL) {
  clock_print(x, max)
}

# - Pass through internal option to not print zone name
# - Unlike vctrs, don't use `print(quote = FALSE)` since we want to match base R
#' @export
obj_print_data.clock_zoned_time <- function(x, ..., max) {
  if (vec_is_empty(x)) {
    return(invisible(x))
  }

  x <- max_slice(x, max)

  out <- format(x, print_zone_name = FALSE)

  # Pass `max` to avoid base R's default footer
  print(out, max = max)

  invisible(x)
}

#' @export
obj_print_footer.clock_zoned_time <- function(x, ..., max) {
  clock_print_footer(x, max)
}

# Align left to match pillar_shaft.Date
# @export - lazy in .onLoad()
pillar_shaft.clock_zoned_time <- function(x, ...) {
  out <- format(x, print_zone_name = FALSE)
  pillar::new_pillar_shaft_simple(out, align = "left")
}

# ------------------------------------------------------------------------------

#' Convert to a zoned-time
#'
#' @description
#' `as_zoned_time()` converts `x` to a zoned-time. You generally convert
#' to a zoned time from either a sys-time or a naive time. Each are documented
#' on their own page:
#'
#' - [sys-time][as-zoned-time-sys-time]
#'
#' - [naive-time][as-zoned-time-naive-time]
#'
#' There are also convenience methods for converting to a zoned time from
#' native R date and date-time types:
#'
#' - [dates (Date)][as-zoned-time-Date]
#'
#' - [date-times (POSIXct / POSIXlt)][as-zoned-time-posixt]
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[object]`
#'
#'   An object to convert to a zoned-time.
#'
#' @return A zoned-time vector.
#'
#' @export
#' @examples
#' x <- as.Date("2019-01-01")
#' as_zoned_time(x, "Europe/London")
#'
#' y <- as_naive_time(year_month_day(2019, 2, 1))
#' as_zoned_time(y, zone = "America/New_York")
as_zoned_time <- function(x, ...) {
  UseMethod("as_zoned_time")
}

#' @export
as_zoned_time.default <- function(x, ...) {
  stop_clock_unsupported_conversion(x, "clock_zoned_time")
}

#' @export
as_zoned_time.clock_zoned_time <- function(x, ...) {
  x
}

# ------------------------------------------------------------------------------

#' @export
as_sys_time.clock_zoned_time <- function(x) {
  names <- clock_rcrd_names(x)
  precision <- zoned_time_precision_attribute(x)
  new_sys_time_from_fields(x, precision, names)
}

#' @export
as_naive_time.clock_zoned_time <- function(x) {
  names <- clock_rcrd_names(x)
  zone <- zoned_time_zone_attribute(x)
  precision <- zoned_time_precision_attribute(x)
  fields <- get_naive_time_cpp(x, precision, zone)
  new_naive_time_from_fields(fields, precision, names)
}

#' @export
as.character.clock_zoned_time <- function(x, ...) {
  format(x)
}

# ------------------------------------------------------------------------------

#' What is the current zoned-time?
#'
#' @description
#' `zoned_time_now()` returns the current time in the corresponding `zone`. It
#' is a wrapper around [sys_time_now()] that attaches the time zone.
#'
#' @details
#' The time is returned with a nanosecond precision, but the actual amount
#' of data returned is OS dependent. Usually, information at at least the
#' microsecond level is returned, with some platforms returning nanosecond
#' information.
#'
#' @param zone `[character(1)]`
#'
#'   A time zone to get the current time for.
#'
#' @return A zoned-time of the current time.
#'
#' @export
#' @examples
#' x <- zoned_time_now("America/New_York")
zoned_time_now <- function(zone) {
  names <- NULL
  sys_time <- sys_time_now()
  precision <- time_point_precision_attribute(sys_time)
  zone <- zone_validate(zone)
  new_zoned_time_from_fields(sys_time, precision, zone, names)
}

# ------------------------------------------------------------------------------

#' Get or set the time zone
#'
#' @description
#' `zoned_time_zone()` gets the time zone.
#'
#' `zoned_time_set_zone()` sets the time zone _without changing the
#' underlying instant_. This means that the result will represent the equivalent
#' time in the new time zone.
#'
#' @param x `[zoned_time / Date / POSIXt]`
#'
#'   A zoned time to get or set the time zone of.
#'
#' @param zone `[character(1)]`
#'
#'   A valid time zone to switch to.
#'
#' @return
#' `zoned_time_zone()` returns a string containing the time zone.
#'
#' `zoned_time_set_zone()` returns `x` with an altered time zone attribute. The
#' underlying instant is _not_ changed.
#'
#' @name zoned-zone
#'
#' @examples
#' x <- year_month_day(2019, 1, 1)
#' x <- as_zoned_time(as_naive_time(x), "America/New_York")
#' x
#'
#' zoned_time_zone(x)
#'
#' # Equivalent UTC time
#' zoned_time_set_zone(x, "UTC")
#'
#' # To force a new time zone with the same printed time,
#' # convert to a naive time that has no implied time zone,
#' # then convert back to a zoned time in the new time zone.
#' nt <- as_naive_time(x)
#' nt
#' as_zoned_time(nt, "UTC")
NULL

#' @rdname zoned-zone
#' @export
zoned_time_zone <- function(x) {
  UseMethod("zoned_time_zone")
}

#' @export
zoned_time_zone.clock_zoned_time <- function(x) {
  zoned_time_zone_attribute(x)
}

#' @rdname zoned-zone
#' @export
zoned_time_set_zone <- function(x, zone) {
  UseMethod("zoned_time_set_zone")
}

#' @export
zoned_time_set_zone.clock_zoned_time <- function(x, zone) {
  zone <- zone_validate(zone)
  zoned_time_set_zone_attribute(x, zone)
}

# ------------------------------------------------------------------------------

#' Precision: zoned-time
#'
#' `zoned_time_precision()` extracts the precision from a zoned-time. It
#' returns the precision as a single string.
#'
#' @param x `[clock_zoned_time]`
#'
#'   A zoned-time.
#'
#' @return A single string holding the precision of the zoned-time.
#'
#' @export
#' @examples
#' zoned_time_precision(zoned_time_now("America/New_York"))
zoned_time_precision <- function(x) {
  if (!is_zoned_time(x)) {
    abort("`x` must be a 'clock_zoned_time'.")
  }
  precision <- zoned_time_precision_attribute(x)
  precision <- precision_to_string(precision)
  precision
}

# ------------------------------------------------------------------------------

#' @export
add_years.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_years")
}

#' @export
add_quarters.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_quarters")
}

#' @export
add_months.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_months")
}

#' @export
add_weeks.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_weeks")
}

#' @export
add_days.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_days")
}

#' @export
add_hours.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_hours")
}

#' @export
add_minutes.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_minutes")
}

#' @export
add_seconds.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_seconds")
}

#' @export
add_milliseconds.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_milliseconds")
}

#' @export
add_microseconds.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_microseconds")
}

#' @export
add_nanoseconds.clock_zoned_time <- function(x, n, ...) {
  stop_clock_unsupported_zoned_time_op("add_nanoseconds")
}

# ------------------------------------------------------------------------------

zone_validate <- function(zone) {
  if (!is_string(zone)) {
    abort("`zone` must be a single string.")
  }

  if (!zone_is_valid(zone)) {
    message <- paste0("'", zone, "' is not a known time zone.")
    abort(message)
  }

  zone
}

# ------------------------------------------------------------------------------

validate_zoned_time_precision_string <- function(precision) {
  precision <- validate_precision_string(precision)

  if (!is_valid_zoned_time_precision(precision)) {
    abort("`precision` must be at least 'second' precision.")
  }

  precision
}

is_valid_zoned_time_precision <- function(precision) {
  precision >= PRECISION_SECOND
}

# ------------------------------------------------------------------------------

clock_init_zoned_time_utils <- function(env) {
  assign("clock_empty_zoned_time_utc_second", as_zoned_time(as_sys_time(duration_seconds()), "UTC"), envir = env)
  assign("clock_empty_zoned_time_utc_millisecond", as_zoned_time(as_sys_time(duration_milliseconds()), "UTC"), envir = env)
  assign("clock_empty_zoned_time_utc_microsecond", as_zoned_time(as_sys_time(duration_microseconds()), "UTC"), envir = env)
  assign("clock_empty_zoned_time_utc_nanosecond", as_zoned_time(as_sys_time(duration_nanoseconds()), "UTC"), envir = env)

  invisible(NULL)
}

#' Calendar: year-month-weekday
#'
#' `year_month_weekday()` constructs a calendar vector from the Gregorian
#' year, month, weekday, and index specifying that this is the n-th weekday
#' of the month.
#'
#' @details
#' Fields are recycled against each other.
#'
#' Fields are collected in order until the first `NULL` field is located. No
#' fields after the first `NULL` field are used.
#'
#' @inheritParams year_month_day
#'
#' @param day `[integer / NULL]`
#'
#'   The weekday of the month. Values `[1, 7]` are allowed, where `1` is
#'   Sunday and `7` is Saturday.
#'
#' @param index `[integer / "last" / NULL]`
#'
#'   The index specifying that `day` is the n-th weekday of the month.
#'   Values `[1, 5]` are allowed.
#'
#'   If `"last"`, then the last instance of `day` in the current month
#'   is returned.
#'
#' @return A year-month-weekday calendar vector.
#'
#' @export
#' @examples
#' # All Fridays in January, 2019
#' # Note that there was no 5th Friday in January
#' x <- year_month_weekday(
#'   2019,
#'   clock_months$january,
#'   clock_weekdays$friday,
#'   1:5
#' )
#' x
#'
#' invalid_detect(x)
#'
#' # Resolve this invalid date by using the previous valid date
#' invalid_resolve(x, invalid = "previous")
year_month_weekday <- function(year,
                               month = NULL,
                               day = NULL,
                               index = NULL,
                               hour = NULL,
                               minute = NULL,
                               second = NULL,
                               subsecond = NULL,
                               ...,
                               subsecond_precision = NULL) {
  if (xor(is_null(day), is_null(index))) {
    abort("If either `day` or `index` is specified, both must be specified.")
  }

  # Stop on the first `NULL` argument
  if (is_null(month)) {
    precision <- PRECISION_YEAR
    fields <- list(year = year)
  } else if (is_null(index)) {
    precision <- PRECISION_MONTH
    fields <- list(year = year, month = month)
  } else if (is_null(hour)) {
    precision <- PRECISION_DAY
    fields <- list(year = year, month = month, day = day, index = index)
  } else if (is_null(minute)) {
    precision <- PRECISION_HOUR
    fields <- list(year = year, month = month, day = day, index = index, hour = hour)
  } else if (is_null(second)) {
    precision <- PRECISION_MINUTE
    fields <- list(year = year, month = month, day = day, index = index, hour = hour, minute = minute)
  } else if (is_null(subsecond)) {
    precision <- PRECISION_SECOND
    fields <- list(year = year, month = month, day = day, index = index, hour = hour, minute = minute, second = second)
  } else {
    precision <- calendar_validate_subsecond_precision(subsecond_precision)
    fields <- list(year = year, month = month, day = day, index = index, hour = hour, minute = minute, second = second, subsecond = subsecond)
  }

  if (is_last(fields$index)) {
    fields$index <- 1L
    last <- TRUE
  } else {
    last <- FALSE
  }

  fields <- vec_recycle_common(!!!fields)
  fields <- vec_cast_common(!!!fields, .to = integer())

  fields <- collect_year_month_weekday_fields(fields, precision)

  names <- NULL

  out <- new_year_month_weekday_from_fields(fields, precision, names)

  if (last) {
    out <- set_index(out, "last")
  }

  out
}

# ------------------------------------------------------------------------------

#' @export
vec_proxy.clock_year_month_weekday <- function(x, ...) {
  .Call(`_clock_clock_rcrd_proxy`, x)
}

#' @export
vec_restore.clock_year_month_weekday <- function(x, to, ...) {
  .Call(`_clock_year_month_weekday_restore`, x, to)
}

#' @export
vec_proxy_compare.clock_year_month_weekday <- function(x, ...) {
  precision <- calendar_precision_attribute(x)

  if (precision >= PRECISION_DAY) {
    # See issue #32
    message <- paste0(
      "'year_month_weekday' types with a precision of >= 'day' cannot be ",
      "trivially compared or ordered. ",
      "Convert to 'year_month_day' to compare using day-of-month values."
    )
    abort(message)
  }

  # Year / month year-month-weekday precision can be compared without ambiguity
  vec_proxy(x)
}

# ------------------------------------------------------------------------------

#' @export
format.clock_year_month_weekday <- function(x, ...) {
  out <- format_year_month_weekday_cpp(x, calendar_precision_attribute(x))
  names(out) <- names(x)
  out
}

#' @export
vec_ptype_full.clock_year_month_weekday <- function(x, ...) {
  calendar_ptype_full(x, "year_month_weekday")
}

#' @export
vec_ptype_abbr.clock_year_month_weekday <- function(x, ...) {
  calendar_ptype_abbr(x, "ymw")
}

# ------------------------------------------------------------------------------

#' Is `x` a year-month-weekday?
#'
#' Check if `x` is a year-month-weekday.
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @return Returns `TRUE` if `x` inherits from `"clock_year_month_weekday"`,
#'   otherwise returns `FALSE`.
#'
#' @export
#' @examples
#' is_year_month_weekday(year_month_weekday(2019))
is_year_month_weekday <- function(x) {
  inherits(x, "clock_year_month_weekday")
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype.clock_year_month_weekday <- function(x, ...) {
  switch(
    calendar_precision_attribute(x) + 1L,
    clock_empty_year_month_weekday_year,
    abort("Internal error: Invalid precision"),
    clock_empty_year_month_weekday_month,
    abort("Internal error: Invalid precision"),
    clock_empty_year_month_weekday_day,
    clock_empty_year_month_weekday_hour,
    clock_empty_year_month_weekday_minute,
    clock_empty_year_month_weekday_second,
    clock_empty_year_month_weekday_millisecond,
    clock_empty_year_month_weekday_microsecond,
    clock_empty_year_month_weekday_nanosecond,
    abort("Internal error: Invalid precision.")
  )
}

#' @export
vec_ptype2.clock_year_month_weekday.clock_year_month_weekday <- function(x, y, ...) {
  ptype2_calendar_and_calendar(x, y, ...)
}

#' @export
vec_cast.clock_year_month_weekday.clock_year_month_weekday <- function(x, to, ...) {
  cast_calendar_to_calendar(x, to, ...)
}

# ------------------------------------------------------------------------------

#' @export
calendar_is_valid_precision.clock_year_month_weekday <- function(x, precision) {
  year_month_weekday_is_valid_precision(precision)
}

year_month_weekday_is_valid_precision <- function(precision) {
  if (!is_valid_precision(precision)) {
    FALSE
  } else if (precision == PRECISION_YEAR || precision == PRECISION_MONTH) {
    TRUE
  } else if (precision >= PRECISION_DAY && precision <= PRECISION_NANOSECOND) {
    TRUE
  } else {
    FALSE
  }
}

# ------------------------------------------------------------------------------

#' @export
invalid_detect.clock_year_month_weekday <- function(x) {
  invalid_detect_year_month_weekday_cpp(x, calendar_precision_attribute(x))
}

#' @export
invalid_any.clock_year_month_weekday <- function(x) {
  invalid_any_year_month_weekday_cpp(x, calendar_precision_attribute(x))
}

#' @export
invalid_count.clock_year_month_weekday <- function(x) {
  invalid_count_year_month_weekday_cpp(x, calendar_precision_attribute(x))
}

#' @export
invalid_resolve.clock_year_month_weekday <- function(x, ..., invalid = NULL) {
  check_dots_empty()
  precision <- calendar_precision_attribute(x)
  invalid <- validate_invalid(invalid)
  fields <- invalid_resolve_year_month_weekday_cpp(x, precision, invalid)
  new_year_month_weekday_from_fields(fields, precision, names = names(x))
}

# ------------------------------------------------------------------------------

#' Getters: year-month-weekday
#'
#' @description
#' These are year-month-weekday methods for the
#' [getter generics][clock-getters].
#'
#' - `get_year()` returns the Gregorian year.
#'
#' - `get_month()` returns the month of the year.
#'
#' - `get_day()` returns the day of the week encoded from 1-7, where 1 = Sunday
#'   and 7 = Saturday.
#'
#' - `get_index()` returns a value from 1-5 indicating that the corresponding
#'   weekday is the n-th instance of that weekday in the current month.
#'
#' - There are sub-daily getters for extracting more precise components.
#'
#' @param x `[clock_year_month_weekday]`
#'
#'   A year-month-weekday to get the component from.
#'
#' @return The component.
#'
#' @name year-month-weekday-getters
#' @examples
#' monday <- clock_weekdays$monday
#' thursday <- clock_weekdays$thursday
#'
#' x <- year_month_weekday(2019, 1, monday:thursday, 1:4)
#' x
#'
#' # Gets the weekday, 1 = Sunday, 7 = Saturday
#' get_day(x)
#'
#' # Gets the index indicating which instance of that particular weekday
#' # it is in the current month (i.e. the "1st Monday of January, 2019")
#' get_index(x)
NULL

#' @rdname year-month-weekday-getters
#' @export
get_year.clock_year_month_weekday <- function(x) {
  field_year(x)
}

#' @rdname year-month-weekday-getters
#' @export
get_month.clock_year_month_weekday <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_MONTH, "get_month")
  field_month(x)
}

#' @rdname year-month-weekday-getters
#' @export
get_day.clock_year_month_weekday <- function(x) {
  # [Sunday, Saturday] -> [1, 7]
  calendar_require_minimum_precision(x, PRECISION_DAY, "get_day")
  field_day(x)
}

#' @rdname year-month-weekday-getters
#' @export
get_index.clock_year_month_weekday <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_DAY, "get_index")
  field_index(x)
}

#' @rdname year-month-weekday-getters
#' @export
get_hour.clock_year_month_weekday <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_HOUR, "get_hour")
  field_hour(x)
}

#' @rdname year-month-weekday-getters
#' @export
get_minute.clock_year_month_weekday <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_MINUTE, "get_minute")
  field_minute(x)
}

#' @rdname year-month-weekday-getters
#' @export
get_second.clock_year_month_weekday <- function(x) {
  calendar_require_minimum_precision(x, PRECISION_SECOND, "get_second")
  field_second(x)
}

#' @rdname year-month-weekday-getters
#' @export
get_millisecond.clock_year_month_weekday <- function(x) {
  calendar_require_precision(x, PRECISION_MILLISECOND, "get_millisecond")
  field_subsecond(x)
}

#' @rdname year-month-weekday-getters
#' @export
get_microsecond.clock_year_month_weekday <- function(x) {
  calendar_require_precision(x, PRECISION_MICROSECOND, "get_microsecond")
  field_subsecond(x)
}

#' @rdname year-month-weekday-getters
#' @export
get_nanosecond.clock_year_month_weekday <- function(x) {
  calendar_require_precision(x, PRECISION_NANOSECOND, "get_nanosecond")
  field_subsecond(x)
}

# ------------------------------------------------------------------------------

#' Setters: year-month-weekday
#'
#' @description
#' These are year-month-weekday methods for the
#' [setter generics][clock-setters].
#'
#' - `set_year()` sets the Gregorian year.
#'
#' - `set_month()` sets the month of the year. Valid values are in the range
#'   of `[1, 12]`.
#'
#' - `set_day()` sets the day of the week. Valid values are in the range of
#'   `[1, 7]`, with 1 = Sunday, and 7 = Saturday.
#'
#' - `set_index()` sets the index indicating that the corresponding
#'   weekday is the n-th instance of that weekday in the current month. Valid
#'   values are in the range of `[1, 5]`.
#'
#' - There are sub-daily setters for setting more precise components.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[clock_year_month_weekday]`
#'
#'   A year-month-weekday vector.
#'
#' @param value `[integer / "last"]`
#'
#'   The value to set the component to.
#'
#'   For `set_index()`, this can also be `"last"` to adjust to the last
#'   instance of the corresponding weekday in that month.
#'
#' @param index `[NULL / integer / "last"]`
#'
#'   This argument is only used with `set_day()`, and allows you to set the
#'   index while also setting the weekday.
#'
#'   If `x` is a month precision year-month-weekday, `index` is required to
#'   be set, as you must specify the weekday and the index simultaneously to
#'   promote from month to day precision.
#'
#' @return `x` with the component set.
#'
#' @name year-month-weekday-setters
#' @examples
#' x <- year_month_weekday(2019, 1:3)
#'
#' set_year(x, 2020:2022)
#'
#' # Setting the weekday on a month precision year-month-weekday requires
#' # also setting the `index` to fully specify the day information
#' x <- set_day(x, clock_weekdays$sunday, index = 1)
#' x
#'
#' # Once you have at least day precision, you can set the weekday and
#' # the index separately
#' set_day(x, clock_weekdays$monday)
#' set_index(x, 3)
#'
#' # Set to the "last" instance of the corresponding weekday in this month
#' # (Note that some months have 4 Sundays, and others have 5)
#' set_index(x, "last")
#'
#' # Set to an invalid index
#' # January and February of 2019 don't have 5 Sundays!
#' invalid <- set_index(x, 5)
#' invalid
#'
#' # Resolve the invalid dates by choosing the previous/next valid moment
#' invalid_resolve(invalid, invalid = "previous")
#' invalid_resolve(invalid, invalid = "next")
#'
#' # You can also "overflow" the index. This keeps the weekday, but resets
#' # the index to 1 and increments the month value by 1.
#' invalid_resolve(invalid, invalid = "overflow")
NULL

#' @rdname year-month-weekday-setters
#' @export
set_year.clock_year_month_weekday <- function(x, value, ...) {
  check_dots_empty()
  set_field_year_month_weekday(x, value, "year")
}

#' @rdname year-month-weekday-setters
#' @export
set_month.clock_year_month_weekday <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_YEAR, "set_month")
  set_field_year_month_weekday(x, value, "month")
}

#' @rdname year-month-weekday-setters
#' @export
set_day.clock_year_month_weekday <- function(x, value, ..., index = NULL) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_MONTH, "set_day")

  has_index <- !is_null(index)
  precision <- calendar_precision_attribute(x)

  if (precision == PRECISION_MONTH) {
    if (!has_index) {
      abort("For 'month' precision 'year_month_weekday', both the day and index must be set simultaneously.")
    }

    # Promote up to day precision so we can assign to fields individually
    ones <- ones_along(x, na_propagate = TRUE)
    fields <- list(year = get_year(x), month = get_month(x), day = ones, index = ones)
    x <- new_year_month_weekday_from_fields(fields, PRECISION_DAY, names(x))
  }

  out <- set_field_year_month_weekday(x, value, "day")

  if (has_index) {
    out <- set_field_year_month_weekday(out, index, "index")
  }

  out
}

#' @rdname year-month-weekday-setters
#' @export
set_index.clock_year_month_weekday <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_DAY, "set_index")
  set_field_year_month_weekday(x, value, "index")
}

#' @rdname year-month-weekday-setters
#' @export
set_hour.clock_year_month_weekday <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_DAY, "set_hour")
  set_field_year_month_weekday(x, value, "hour")
}

#' @rdname year-month-weekday-setters
#' @export
set_minute.clock_year_month_weekday <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_HOUR, "set_minute")
  set_field_year_month_weekday(x, value, "minute")
}

#' @rdname year-month-weekday-setters
#' @export
set_second.clock_year_month_weekday <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_minimum_precision(x, PRECISION_MINUTE, "set_second")
  set_field_year_month_weekday(x, value, "second")
}

#' @rdname year-month-weekday-setters
#' @export
set_millisecond.clock_year_month_weekday <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_MILLISECOND), "set_millisecond")
  set_field_year_month_weekday(x, value, "millisecond")
}

#' @rdname year-month-weekday-setters
#' @export
set_microsecond.clock_year_month_weekday <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_MICROSECOND), "set_microsecond")
  set_field_year_month_weekday(x, value, "microsecond")
}

#' @rdname year-month-weekday-setters
#' @export
set_nanosecond.clock_year_month_weekday <- function(x, value, ...) {
  check_dots_empty()
  calendar_require_any_of_precisions(x, c(PRECISION_SECOND, PRECISION_NANOSECOND), "set_nanosecond")
  set_field_year_month_weekday(x, value, "nanosecond")
}

set_field_year_month_weekday <- function(x, value, component) {
  if (is_last(value) && identical(component, "index")) {
    return(set_field_year_month_weekday_last(x))
  }

  precision_fields <- calendar_precision_attribute(x)
  precision_value <- year_month_weekday_component_to_precision(component)
  precision_out <- precision_common2(precision_fields, precision_value)

  value <- vec_cast(value, integer(), x_arg = "value")
  args <- vec_recycle_common(x = x, value = value)
  x <- args$x
  value <- args$value

  result <- set_field_year_month_weekday_cpp(x, value, precision_fields, component)
  fields <- result$fields
  field <- year_month_weekday_component_to_field(component)
  fields[[field]] <- result$value

  new_year_month_weekday_from_fields(fields, precision_out, names = names(x))
}

set_field_year_month_weekday_last <- function(x) {
  # We require 'day' precision to set the `index` at all, so no
  # need to find a common precision here
  precision_fields <- calendar_precision_attribute(x)

  result <- set_field_year_month_weekday_last_cpp(x, precision_fields)
  fields <- result$fields
  fields[["index"]] <- result$value

  new_year_month_weekday_from_fields(fields, precision_fields, names = names(x))
}

# ------------------------------------------------------------------------------

#' @export
calendar_name.clock_year_month_weekday <- function(x) {
  "year_month_weekday"
}

# ------------------------------------------------------------------------------

year_month_weekday_component_to_precision <- function(component) {
  switch(
    component,
    year = PRECISION_YEAR,
    month = PRECISION_MONTH,
    day = PRECISION_DAY,
    index = PRECISION_DAY,
    hour = PRECISION_HOUR,
    minute = PRECISION_MINUTE,
    second = PRECISION_SECOND,
    millisecond = PRECISION_MILLISECOND,
    microsecond = PRECISION_MICROSECOND,
    nanosecond = PRECISION_NANOSECOND,
    abort("Internal error: Unknown component name.")
  )
}

year_month_weekday_component_to_field <- function(component) {
  switch (
    component,
    year = component,
    month = component,
    day = component,
    index = component,
    hour = component,
    minute = component,
    second = component,
    millisecond = "subsecond",
    microsecond = "subsecond",
    nanosecond = "subsecond",
    abort("Internal error: Unknown component name.")
  )
}

# ------------------------------------------------------------------------------

#' @rdname clock-arith
#' @method vec_arith clock_year_month_weekday
#' @export
vec_arith.clock_year_month_weekday <- function(op, x, y, ...) {
  UseMethod("vec_arith.clock_year_month_weekday", y)
}

#' @method vec_arith.clock_year_month_weekday MISSING
#' @export
vec_arith.clock_year_month_weekday.MISSING <- function(op, x, y, ...) {
  arith_calendar_and_missing(op, x, y, ...)
}

#' @method vec_arith.clock_year_month_weekday clock_year_month_weekday
#' @export
vec_arith.clock_year_month_weekday.clock_year_month_weekday <- function(op, x, y, ...) {
  arith_calendar_and_calendar(op, x, y, ..., calendar_minus_calendar_fn = year_month_weekday_minus_year_month_weekday)
}

#' @method vec_arith.clock_year_month_weekday clock_duration
#' @export
vec_arith.clock_year_month_weekday.clock_duration <- function(op, x, y, ...) {
  arith_calendar_and_duration(op, x, y, ...)
}

#' @method vec_arith.clock_duration clock_year_month_weekday
#' @export
vec_arith.clock_duration.clock_year_month_weekday <- function(op, x, y, ...) {
  arith_duration_and_calendar(op, x, y, ...)
}

#' @method vec_arith.clock_year_month_weekday numeric
#' @export
vec_arith.clock_year_month_weekday.numeric <- function(op, x, y, ...) {
  arith_calendar_and_numeric(op, x, y, ...)
}

#' @method vec_arith.numeric clock_year_month_weekday
#' @export
vec_arith.numeric.clock_year_month_weekday <- function(op, x, y, ...) {
  arith_numeric_and_calendar(op, x, y, ...)
}

year_month_weekday_minus_year_month_weekday <- function(op, x, y, ...) {
  args <- vec_recycle_common(x = x, y = y)
  args <- vec_cast_common(!!!args)
  x <- args$x
  y <- args$y

  names <- names_common(x, y)

  precision <- calendar_precision_attribute(x)

  if (precision > PRECISION_MONTH) {
    stop_incompatible_op(op, x, y, ...)
  }

  fields <- year_month_weekday_minus_year_month_weekday_cpp(x, y, precision)

  new_duration_from_fields(fields, precision, names)
}

# ------------------------------------------------------------------------------

#' Arithmetic: year-month-weekday
#'
#' @description
#' These are year-month-weekday methods for the
#' [arithmetic generics][clock-arithmetic].
#'
#' - `add_years()`
#'
#' - `add_quarters()`
#'
#' - `add_months()`
#'
#' Notably, _you cannot add days to a year-month-weekday_. For day-based
#' arithmetic, first convert to a time point with [as_naive_time()] or
#' [as_sys_time()].
#'
#' @details
#' Adding a single quarter with `add_quarters()` is equivalent to adding
#' 3 months.
#'
#' `x` and `n` are recycled against each other.
#'
#' @inheritParams add_years
#'
#' @param x `[clock_year_month_weekday]`
#'
#'   A year-month-weekday vector.
#'
#' @return `x` after performing the arithmetic.
#'
#' @name year-month-weekday-arithmetic
#'
#' @examples
#' # 2nd Friday in January, 2019
#' x <- year_month_weekday(2019, 1, clock_weekdays$friday, 2)
#' x
#'
#' add_months(x, 1:5)
#'
#' # These don't necessarily correspond to the same day of the month
#' as_year_month_day(add_months(x, 1:5))
NULL

#' @rdname year-month-weekday-arithmetic
#' @export
add_years.clock_year_month_weekday <- function(x, n, ...) {
  year_month_weekday_plus_duration(x, n, PRECISION_YEAR)
}

#' @rdname year-month-weekday-arithmetic
#' @export
add_quarters.clock_year_month_weekday <- function(x, n, ...) {
  calendar_require_minimum_precision(x, PRECISION_MONTH, "add_quarters")
  year_month_weekday_plus_duration(x, n, PRECISION_QUARTER)
}

#' @rdname year-month-weekday-arithmetic
#' @export
add_months.clock_year_month_weekday <- function(x, n, ...) {
  calendar_require_minimum_precision(x, PRECISION_MONTH, "add_months")
  year_month_weekday_plus_duration(x, n, PRECISION_MONTH)
}

year_month_weekday_plus_duration <- function(x, n, precision_n) {
  precision_fields <- calendar_precision_attribute(x)

  n <- duration_collect_n(n, precision_n)
  args <- vec_recycle_common(x = x, n = n)
  x <- args$x
  n <- args$n

  names <- names_common(x, n)

  fields <- year_month_weekday_plus_duration_cpp(x, n, precision_fields, precision_n)

  new_year_month_weekday_from_fields(fields, precision_fields, names = names)
}

# ------------------------------------------------------------------------------

#' Convert to year-month-weekday
#'
#' `as_year_month_weekday()` converts a vector to the year-month-weekday
#' calendar. Time points, Dates, POSIXct, and other calendars can all be
#' converted to year-month-weekday.
#'
#' @param x `[vector]`
#'
#'   A vector to convert to year-month-weekday.
#'
#' @return A year-month-weekday vector.
#' @export
#' @examples
#' # From Date
#' as_year_month_weekday(as.Date("2019-01-01"))
#'
#' # From POSIXct, which assumes that the naive time is what should be converted
#' as_year_month_weekday(as.POSIXct("2019-01-01 02:30:30", "America/New_York"))
#'
#' # From other calendars
#' as_year_month_weekday(year_quarter_day(2019, quarter = 2, day = 50))
as_year_month_weekday <- function(x)  {
  UseMethod("as_year_month_weekday")
}

#' @export
as_year_month_weekday.default <- function(x) {
  stop_clock_unsupported_conversion(x, "clock_year_month_weekday")
}

#' @export
as_year_month_weekday.clock_year_month_weekday <- function(x) {
  x
}

# ------------------------------------------------------------------------------

#' @export
as_sys_time.clock_year_month_weekday <- function(x) {
  calendar_require_all_valid(x)
  precision <- calendar_precision_attribute(x)
  fields <- as_sys_time_year_month_weekday_cpp(x, precision)
  new_sys_time_from_fields(fields, precision, clock_rcrd_names(x))
}

#' @export
as_naive_time.clock_year_month_weekday <- function(x) {
  as_naive_time(as_sys_time(x))
}

#' @export
as.character.clock_year_month_weekday <- function(x, ...) {
  format(x)
}

# ------------------------------------------------------------------------------

#' @export
calendar_leap_year.clock_year_month_weekday <- function(x) {
  x <- get_year(x)
  gregorian_leap_year_cpp(x)
}

# ------------------------------------------------------------------------------

#' @export
calendar_month_factor.clock_year_month_weekday <- function(x,
                                                           ...,
                                                           labels = "en",
                                                           abbreviate = FALSE) {
  calendar_month_factor_impl(x, labels, abbreviate, ...)
}

# ------------------------------------------------------------------------------

#' Grouping: year-month-weekday
#'
#' @description
#' This is a year-month-weekday method for the [calendar_group()] generic.
#'
#' Grouping for a year-month-weekday object can be done at any precision except
#' for `"day"`, as long as `x` is at least as precise as `precision`.
#'
#' @details
#' Grouping by `"day"` is undefined for a year-month-weekday because there are
#' two day fields, the weekday and the index, and there is no clear way to
#' define how to group by that.
#'
#' @inheritParams calendar_group
#'
#' @param x `[clock_year_month_weekday]`
#'
#'   A year-month-weekday vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"month"`
#'   - `"hour"`
#'   - `"minute"`
#'   - `"second"`
#'   - `"millisecond"`
#'   - `"microsecond"`
#'   - `"nanosecond"`
#'
#' @return `x` grouped at the specified `precision`.
#'
#' @name year-month-weekday-group
#'
#' @export
#' @examples
#' x <- year_month_weekday(2019, 1:12, clock_weekdays$sunday, 1, 00, 05, 05)
#' x
#'
#' # Group by 3 months - drops more precise components!
#' calendar_group(x, "month", n = 3)
calendar_group.clock_year_month_weekday <- function(x, precision, ..., n = 1L) {
  n <- validate_calendar_group_n(n)
  x <- calendar_narrow(x, precision)

  precision <- validate_precision_string(precision)

  if (precision == PRECISION_YEAR) {
    value <- get_year(x)
    value <- group_component0(value, n)
    x <- set_year(x, value)
    return(x)
  }
  if (precision == PRECISION_MONTH) {
    value <- get_month(x)
    value <- group_component1(value, n)
    x <- set_month(x, value)
    return(x)
  }
  if (precision == PRECISION_DAY) {
    message <- paste0(
      "Grouping 'year_month_weekday' by 'day' precision is undefined. ",
      "Convert to 'year_month_day' to group by day of month."
    )
    abort(message)
  }

  x <- calendar_group_time(x, n, precision)
  x
}

# ------------------------------------------------------------------------------

#' Narrow: year-month-weekday
#'
#' This is a year-month-weekday method for the [calendar_narrow()] generic. It
#' narrows a year-month-weekday vector to the specified `precision`.
#'
#' @inheritParams year-month-weekday-widen
#'
#' @return `x` narrowed to the supplied `precision`.
#'
#' @name year-month-weekday-narrow
#'
#' @export
#' @examples
#' # Day precision
#' x <- year_month_weekday(2019, 1, 1, 2)
#' x
#'
#' # Narrowed to month precision
#' calendar_narrow(x, "month")
calendar_narrow.clock_year_month_weekday <- function(x, precision) {
  precision <- validate_precision_string(precision)

  out_fields <- list()
  x_fields <- unclass(x)

  if (precision >= PRECISION_YEAR) {
    out_fields[["year"]] <- x_fields[["year"]]
  }
  if (precision >= PRECISION_MONTH) {
    out_fields[["month"]] <- x_fields[["month"]]
  }
  if (precision >= PRECISION_DAY) {
    out_fields[["day"]] <- x_fields[["day"]]
    out_fields[["index"]] <- x_fields[["index"]]
  }

  out_fields <- calendar_narrow_time(out_fields, precision, x_fields)

  new_year_month_weekday_from_fields(out_fields, precision, names = names(x))
}

# ------------------------------------------------------------------------------

#' Widen: year-month-weekday
#'
#' This is a year-month-weekday method for the [calendar_widen()] generic. It
#' widens a year-month-weekday vector to the specified `precision`.
#'
#' @details
#' Widening a month precision year-month-weekday to day precision will set
#' the day and the index to `1`. This sets the weekday components to the
#' first Sunday of the month.
#'
#' @inheritParams year-month-weekday-group
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"month"`
#'   - `"day"`
#'   - `"hour"`
#'   - `"minute"`
#'   - `"second"`
#'   - `"millisecond"`
#'   - `"microsecond"`
#'   - `"nanosecond"`
#'
#' @return `x` widened to the supplied `precision`.
#'
#' @name year-month-weekday-widen
#'
#' @export
#' @examples
#' # Month precision
#' x <- year_month_weekday(2019, 1)
#' x
#'
#' # Widen to day precision
#' # Note that this sets both the day and index to 1,
#' # i.e. the first Sunday of the month.
#' calendar_widen(x, "day")
#'
#' # Or second precision
#' sec <- calendar_widen(x, "second")
#' sec
calendar_widen.clock_year_month_weekday <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  if (precision >= PRECISION_MONTH && x_precision < PRECISION_MONTH) {
    x <- set_month(x, 1L)
  }
  if (precision >= PRECISION_DAY && x_precision < PRECISION_DAY) {
    x <- set_day(x, 1L, index = 1L)
  }

  x <- calendar_widen_time(x, x_precision, precision)

  x
}

# ------------------------------------------------------------------------------

#' Boundaries: year-month-weekday
#'
#' @description
#' This is a year-month-weekday method for the [calendar_start()] and
#' [calendar_end()] generics. They adjust components of a calendar to the
#' start or end of a specified `precision`.
#'
#' This method is restricted to only `"year"` and `"month"` `precision`s, and
#' `x` can't be more precise than month precision. Computing the "start" of
#' a day precision year-month-weekday object isn't defined because
#' a year-month-weekday with `day = 1, index = 1` doesn't necessarily occur
#' earlier (chronologically) than `day = 2, index = 1`. Because of these
#' restrictions, this method isn't particularly useful, but is included for
#' completeness.
#'
#' @inheritParams year-month-weekday-group
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"month"`
#'
#' @return `x` at the same precision, but with some components altered to be
#'   at the boundary value.
#'
#' @name year-month-weekday-boundary
#'
#' @examples
#' # Month precision
#' x <- year_month_weekday(2019, 1:5)
#' x
#'
#' # Compute the last month of the year
#' calendar_end(x, "year")
NULL

#' @rdname year-month-weekday-boundary
#' @export
calendar_start.clock_year_month_weekday <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  calendar_start_end_checks(x, x_precision, precision, "start")

  if (x_precision >= PRECISION_DAY) {
    message <- paste0(
      "Computing the start of a 'year_month_weekday' with a precision equal to ",
      "or more precise than 'day' is undefined."
    )
    abort(message)
  }

  if (precision <= PRECISION_YEAR && x_precision > PRECISION_YEAR) {
    x <- set_month(x, 1L)
  }

  x
}

#' @rdname year-month-weekday-boundary
#' @export
calendar_end.clock_year_month_weekday <- function(x, precision) {
  x_precision <- calendar_precision_attribute(x)
  precision <- validate_precision_string(precision)

  calendar_start_end_checks(x, x_precision, precision, "end")

  if (x_precision >= PRECISION_DAY) {
    message <- paste0(
      "Computing the end of a 'year_month_weekday' with a precision equal to ",
      "or more precise than 'day' is undefined."
    )
    abort(message)
  }

  if (precision <= PRECISION_YEAR && x_precision > PRECISION_YEAR) {
    x <- set_month(x, 12L)
  }

  x
}

# ------------------------------------------------------------------------------

#' Counting: year-month-weekday
#'
#' This is a year-month-weekday method for the [calendar_count_between()]
#' generic. It counts the number of `precision` units between `start` and `end`
#' (i.e., the number of years or months).
#'
#' @details
#' Remember that year-month-weekday is not comparable when it is `"day"`
#' precision or finer, so this method is only defined for `"year"` and
#' `"month"` precision year-month-weekday objects.
#'
#' `"quarter"` is equivalent to `"month"` precision with `n` set to `n * 3L`.
#'
#' @inheritParams calendar-count-between
#'
#' @param start,end `[clock_year_month_weekday]`
#'
#'   A pair of year-month-weekday vectors. These will be recycled to their
#'   common size.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"quarter"`
#'   - `"month"`
#'
#' @inherit calendar-count-between return
#'
#' @name year-month-weekday-count-between
#'
#' @export
#' @examples
#' # Compute the number of months between two dates
#' x <- year_month_weekday(2001, 2)
#' y <- year_month_weekday(2021, c(1, 3))
#'
#' calendar_count_between(x, y, "month")
#'
#' # Remember that day precision or finer year-month-weekday objects
#' # are not comparable, so this won't work
#' x <- year_month_weekday(2001, 2, 1, 1)
#' try(calendar_count_between(x, x, "month"))
calendar_count_between.clock_year_month_weekday <- function(start,
                                                            end,
                                                            precision,
                                                            ...,
                                                            n = 1L) {
  NextMethod()
}

calendar_count_between_standardize_precision_n.clock_year_month_weekday <- function(x,
                                                                                    precision,
                                                                                    n) {
  calendar_count_between_standardize_precision_n.clock_year_month_day(x, precision, n)
}

calendar_count_between_compute.clock_year_month_weekday <- function(start,
                                                                    end,
                                                                    precision) {
  calendar_count_between_compute.clock_year_month_day(start, end, precision)
}

calendar_count_between_proxy_compare.clock_year_month_weekday <- function(start,
                                                                          end,
                                                                          precision) {
  calendar_count_between_proxy_compare.clock_year_month_day(start, end, precision)
}

# ------------------------------------------------------------------------------

#' Sequences: year-month-weekday
#'
#' @description
#' This is a year-month-weekday method for the [seq()] generic.
#'
#' Sequences can only be generated for `"year"` and `"month"` precision
#' year-month-weekday vectors.
#'
#' When calling `seq()`, exactly two of the following must be specified:
#' - `to`
#' - `by`
#' - Either `length.out` or `along.with`
#'
#' @inheritParams seq.clock_duration
#'
#' @param from `[clock_year_month_weekday(1)]`
#'
#'   A `"year"` or `"month"` precision year-month-weekday to start the sequence
#'   from.
#'
#'   `from` is always included in the result.
#'
#' @param to `[clock_year_month_weekday(1) / NULL]`
#'
#'   A `"year"` or `"month"` precision year-month-weekday to stop the sequence
#'   at.
#'
#'   `to` is cast to the type of `from`.
#'
#'   `to` is only included in the result if the resulting sequence divides
#'   the distance between `from` and `to` exactly.
#'
#' @return A sequence with the type of `from`.
#'
#' @export
#' @examples
#' # Monthly sequence
#' x <- seq(year_month_weekday(2019, 1), year_month_weekday(2020, 12), by = 1)
#' x
#'
#' # Which we can then set the indexed weekday of
#' set_day(x, clock_weekdays$sunday, index = "last")
seq.clock_year_month_weekday <- function(from,
                                         to = NULL,
                                         by = NULL,
                                         length.out = NULL,
                                         along.with = NULL,
                                         ...) {
  seq.clock_year_month_day(from, to, by, length.out, along.with, ...)
}

# ------------------------------------------------------------------------------

clock_init_year_month_weekday_utils <- function(env) {
  year <- year_month_weekday(integer())

  assign("clock_empty_year_month_weekday_year", year, envir = env)
  assign("clock_empty_year_month_weekday_month", calendar_widen(year, "month"), envir = env)
  assign("clock_empty_year_month_weekday_day", calendar_widen(year, "day"), envir = env)
  assign("clock_empty_year_month_weekday_hour", calendar_widen(year, "hour"), envir = env)
  assign("clock_empty_year_month_weekday_minute", calendar_widen(year, "minute"), envir = env)
  assign("clock_empty_year_month_weekday_second", calendar_widen(year, "second"), envir = env)
  assign("clock_empty_year_month_weekday_millisecond", calendar_widen(year, "millisecond"), envir = env)
  assign("clock_empty_year_month_weekday_microsecond", calendar_widen(year, "microsecond"), envir = env)
  assign("clock_empty_year_month_weekday_nanosecond", calendar_widen(year, "nanosecond"), envir = env)

  invisible(NULL)
}

# ------------------------------------------------------------------------------

#' @export
names.clock_rcrd <- function(x) {
  .Call(`_clock_clock_rcrd_names`, x)
}

#' @export
`names<-.clock_rcrd` <- function(x, value) {
  .Call(`_clock_clock_rcrd_set_names`, x, value)
}

# ------------------------------------------------------------------------------

# - `[.vctrs_rcrd` accidentally allows subsetting fields through `...`
#   https://github.com/r-lib/vctrs/issues/1295

#' @export
`[.clock_rcrd` <- function(x, i) {
  i <- maybe_missing(i, default = TRUE)
  vec_slice(x, i)
}

# - `[[.vctrs_rcrd` doesn't drop names because names aren't supported for rcrds
# - `[[.vctrs_rcrd` allows selections of size >1
#   https://github.com/r-lib/vctrs/issues/1294

#' @export
`[[.clock_rcrd` <- function(x, i) {
  size <- vec_size(x)
  names <- names(x)

  i <- vec_as_location2(i, n = size, names = names, arg = "i")

  # Unname - `[[` never returns input with names
  x <- unname(x)

  vec_slice(x, i)
}

# ------------------------------------------------------------------------------

#' @export
vec_math.clock_rcrd <- function(.fn, .x, ...) {
  switch(
    .fn,
    is.nan = clock_rcrd_is_nan(.x),
    is.finite = clock_rcrd_is_finite(.x),
    is.infinite = clock_rcrd_is_infinite(.x),
    NextMethod()
  )
}

clock_rcrd_is_nan <- function(x) {
  vec_rep(FALSE, vec_size(x))
}

clock_rcrd_is_finite <- function(x) {
  !vec_equal_na(x)
}

clock_rcrd_is_infinite <- function(x) {
  vec_rep(FALSE, vec_size(x))
}

new_sys_time_from_fields <- function(fields, precision, names) {
  new_time_point_from_fields(fields, precision, CLOCK_SYS, names)
}

# ------------------------------------------------------------------------------

sys_days <- function(n = integer()) {
  names <- NULL
  duration <- duration_days(n)
  new_sys_time_from_fields(duration, PRECISION_DAY, names)
}

sys_seconds <- function(n = integer()) {
  names <- NULL
  duration <- duration_seconds(n)
  new_sys_time_from_fields(duration, PRECISION_SECOND, names)
}

# ------------------------------------------------------------------------------

#' Is `x` a sys-time?
#'
#' This function determines if the input is a sys-time object.
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @return `TRUE` if `x` inherits from `"clock_sys_time"`, otherwise `FALSE`.
#'
#' @export
#' @examples
#' is_sys_time(1)
#' is_sys_time(as_sys_time(duration_days(1)))
is_sys_time <- function(x) {
  inherits(x, "clock_sys_time")
}

# ------------------------------------------------------------------------------

#' Parsing: sys-time
#'
#' @description
#' There are two parsers into a sys-time, `sys_time_parse()` and
#' `sys_time_parse_RFC_3339()`.
#'
#' ## sys_time_parse()
#'
#' `sys_time_parse()` is useful when you have date-time strings like
#' `"2020-01-01T01:04:30"` that you know should be interpreted as UTC, or like
#' `"2020-01-01T01:04:30-04:00"` with a UTC offset but no zone name. If you find
#' yourself in the latter situation, then parsing this string as a sys-time
#' using the `%Ez` command to capture the offset is probably your best option.
#' If you know that this string should be interpreted in a specific time zone,
#' parse as a sys-time to get the UTC equivalent, then use [as_zoned_time()].
#'
#' The default options assume that `x` should be parsed at second precision,
#' using a `format` string of `"%Y-%m-%dT%H:%M:%S"`. This matches the default
#' result from calling `format()` on a sys-time.
#'
#' `sys_time_parse()` is nearly equivalent to [naive_time_parse()], except for
#' the fact that the `%z` command is actually used. Using `%z` assumes that the
#' rest of the date-time string should be interpreted as a naive-time, which is
#' then shifted by the UTC offset found in `%z`. The returned time can then be
#' validly interpreted as UTC.
#'
#' _`sys_time_parse()` ignores the `%Z` command._
#'
#' ## sys_time_parse_RFC_3339()
#'
#' `sys_time_parse_RFC_3339()` is a wrapper around `sys_time_parse()` that is
#' intended to parse the extremely common date-time format outlined by
#' [RFC 3339](https://datatracker.ietf.org/doc/html/rfc3339). This document
#' outlines a profile of the ISO 8601 format that is even more restrictive.
#'
#' In particular, this function is intended to parse the following three
#' formats:
#'
#' ```
#' 2019-01-01T00:00:00Z
#' 2019-01-01T00:00:00+0430
#' 2019-01-01T00:00:00+04:30
#' ```
#'
#' This function defaults to parsing the first of these formats by using
#' a format string of `"%Y-%m-%dT%H:%M:%SZ"`.
#'
#' If your date-time strings use offsets from UTC rather than `"Z"`, then set
#' `offset` to one of the following:
#'
#' - `"%z"` if the offset is of the form `"+0430"`.
#' - `"%Ez"` if the offset is of the form `"+04:30"`.
#'
#' The RFC 3339 standard allows for replacing the `"T"` with a `"t"` or a space
#' (`" "`). Set `separator` to adjust this as needed.
#'
#' For this function, the `precision` must be at least `"second"`.
#'
#' @details
#' If your date-time strings contain a full time zone name and a UTC offset, use
#' [zoned_time_parse_complete()]. If they contain a time zone abbreviation, use
#' [zoned_time_parse_abbrev()].
#'
#' If your date-time strings don't contain an offset from UTC and you aren't
#' sure if they should be treated as UTC or not, you might consider using
#' [naive_time_parse()], since the resulting naive-time doesn't come with an
#' assumption of a UTC time zone.
#'
#' @inheritSection zoned-parsing Full Precision Parsing
#'
#' @inheritParams zoned-parsing
#'
#' @param precision `[character(1)]`
#'
#'   A precision for the resulting time point. One of:
#'
#'   - `"day"`
#'
#'   - `"hour"`
#'
#'   - `"minute"`
#'
#'   - `"second"`
#'
#'   - `"millisecond"`
#'
#'   - `"microsecond"`
#'
#'   - `"nanosecond"`
#'
#'   Setting the `precision` determines how much information `%S` attempts
#'   to parse.
#'
#' @param separator `[character(1)]`
#'
#'   The separator between the date and time components of the string. One of:
#'
#'   - `"T"`
#'
#'   - `"t"`
#'
#'   - `" "`
#'
#' @param offset `[character(1)]`
#'
#'   The format of the offset from UTC contained in the string. One of:
#'
#'   - `"Z"`
#'
#'   - `"z"`
#'
#'   - `"%z"` to parse a numeric offset of the form `"+0430"`
#'
#'   - `"%Ez"` to parse a numeric offset of the form `"+04:30"`
#'
#' @return A sys-time.
#'
#' @name sys-parsing
#'
#' @examples
#' sys_time_parse("2020-01-01T05:06:07")
#'
#' # Day precision
#' sys_time_parse("2020-01-01", precision = "day")
#'
#' # Nanosecond precision, but using a day based format
#' sys_time_parse("2020-01-01", format = "%Y-%m-%d", precision = "nanosecond")
#'
#' # Multiple format strings are allowed for heterogeneous times
#' sys_time_parse(
#'   c("2019-01-01", "2019/1/1"),
#'   format = c("%Y/%m/%d", "%Y-%m-%d"),
#'   precision = "day"
#' )
#'
#' # The `%z` command shifts the date-time by subtracting the UTC offset so
#' # that the returned sys-time can be interpreted as UTC
#' sys_time_parse(
#'   "2020-01-01 02:00:00 -0400",
#'   format = "%Y-%m-%d %H:%M:%S %z"
#' )
#'
#' # Remember that the `%Z` command is ignored entirely!
#' sys_time_parse("2020-01-01 America/New_York", format = "%Y-%m-%d %Z")
#'
#' # ---------------------------------------------------------------------------
#' # RFC 3339
#'
#' # Typical UTC format
#' x <- "2019-01-01T00:01:02Z"
#' sys_time_parse_RFC_3339(x)
#'
#' # With a UTC offset containing a `:`
#' x <- "2019-01-01T00:01:02+02:30"
#' sys_time_parse_RFC_3339(x, offset = "%Ez")
#'
#' # With a space between the date and time and no `:` in the offset
#' x <- "2019-01-01 00:01:02+0230"
#' sys_time_parse_RFC_3339(x, separator = " ", offset = "%z")
NULL

#' @rdname sys-parsing
#' @export
sys_time_parse <- function(x,
                           ...,
                           format = NULL,
                           precision = "second",
                           locale = clock_locale()) {
  precision <- validate_time_point_precision_string(precision)

  fields <- time_point_parse(
    x = x,
    ...,
    format = format,
    precision = precision,
    locale = locale,
    clock = CLOCK_SYS
  )

  new_sys_time_from_fields(fields, precision, names(x))
}

#' @rdname sys-parsing
#' @export
sys_time_parse_RFC_3339 <- function(x,
                                    ...,
                                    separator = "T",
                                    offset = "Z",
                                    precision = "second") {
  separator <- arg_match0(separator, values = c("T", "t", " "), arg_nm = "separator")
  offset <- arg_match0(offset, values = c("Z", "z", "%z", "%Ez"), arg_nm = "offset")

  format <- paste0("%Y-%m-%d", separator, "%H:%M:%S", offset)

  # Only used for error checking
  validate_RFC_3339_precision_string(precision)

  sys_time_parse(x, ..., format = format, precision = precision)
}

validate_RFC_3339_precision_string <- function(precision) {
  precision <- validate_precision_string(precision)

  if (!is_valid_RFC_3339_precision(precision)) {
    abort("`precision` must be at least 'second' precision.")
  }

  precision
}
is_valid_RFC_3339_precision <- function(precision) {
  precision >= PRECISION_SECOND
}

# ------------------------------------------------------------------------------

#' Convert to a sys-time
#'
#' @description
#' `as_sys_time()` converts `x` to a sys-time.
#'
#' You can convert to a sys-time from any calendar type, as long as it has
#' at least day precision. There also must not be any invalid dates. If invalid
#' dates exist, they must first be resolved with [invalid_resolve()].
#'
#' Converting to a sys-time from a naive-time retains the printed time,
#' but adds an assumption that the time should be interpreted in the UTC time
#' zone.
#'
#' Converting to a sys-time from a zoned-time retains the underlying duration,
#' but the printed time is the equivalent UTC time to whatever the zoned-time's
#' zone happened to be.
#'
#' Converting to a sys-time from a duration just wraps the duration in a
#' sys-time object, adding the assumption that the time should be interpreted
#' in the UTC time zone. The duration must have at least day precision.
#'
#' There are convenience methods for converting to a sys-time from R's
#' native date and date-time types. Like converting from a zoned-time, these
#' retain the underlying duration, but will change the printed time if the
#' zone was not already UTC.
#'
#' @param x `[object]`
#'
#'   An object to convert to a sys-time.
#'
#' @return A sys-time vector.
#'
#' @export
#' @examples
#' x <- as.Date("2019-01-01")
#'
#' # Dates are assumed to be naive, so the printed time is the same whether
#' # we convert it to sys-time or naive-time
#' as_sys_time(x)
#' as_naive_time(x)
#'
#' y <- as.POSIXct("2019-01-01 01:00:00", tz = "America/New_York")
#'
#' # The sys time displays the equivalent time in UTC (5 hours ahead of
#' # America/New_York at this point in the year)
#' as_sys_time(y)
#'
#' ym <- year_month_day(2019, 02)
#'
#' # A minimum of day precision is required
#' try(as_sys_time(ym))
#'
#' ymd <- set_day(ym, 10)
#' as_sys_time(ymd)
as_sys_time <- function(x) {
  UseMethod("as_sys_time")
}

#' @export
as_sys_time.clock_sys_time <- function(x) {
  x
}

#' @export
as_sys_time.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("as_sys_time")
}

# ------------------------------------------------------------------------------

#' @export
as_naive_time.clock_sys_time <- function(x) {
  new_naive_time_from_fields(x, time_point_precision_attribute(x), clock_rcrd_names(x))
}

#' Convert to a zoned-time from a sys-time
#'
#' @description
#' This is a sys-time method for the [as_zoned_time()] generic.
#'
#' Converting to a zoned-time from a sys-time retains the underlying duration,
#' but changes the printed time, depending on the `zone` that you choose.
#' Remember that sys-times are interpreted as UTC.
#'
#' If you want to retain the printed time, try converting to a zoned-time
#' [from a naive-time][as-zoned-time-naive-time], which is a time point
#' with a yet-to-be-determined time zone.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[clock_sys_time]`
#'
#'   A sys-time to convert to a zoned-time.
#'
#' @param zone `[character(1)]`
#'
#'   The zone to convert to.
#'
#' @return A zoned-time vector.
#'
#' @name as-zoned-time-sys-time
#' @export
#' @examples
#' x <- as_sys_time(year_month_day(2019, 02, 01, 02, 30, 00))
#' x
#'
#' # Since sys-time is interpreted as UTC, converting to a zoned-time with
#' # a zone of UTC retains the printed time
#' x_utc <- as_zoned_time(x, "UTC")
#' x_utc
#'
#' # Converting to a different zone results in a different printed time,
#' # which corresponds to the exact same point in time, just in a different
#' # part of the work
#' x_ny <- as_zoned_time(x, "America/New_York")
#' x_ny
as_zoned_time.clock_sys_time <- function(x, zone, ...) {
  check_dots_empty()

  zone <- zone_validate(zone)

  # Promote to at least seconds precision for `zoned_time`
  x <- vec_cast(x, vec_ptype2(x, clock_empty_sys_time_second))

  precision <- time_point_precision_attribute(x)
  names <- clock_rcrd_names(x)

  new_zoned_time_from_fields(x, precision, zone, names)
}

#' @export
as.character.clock_sys_time <- function(x, ...) {
  format(x)
}

# ------------------------------------------------------------------------------

#' What is the current sys-time?
#'
#' @description
#' `sys_time_now()` returns the current time in UTC.
#'
#' @details
#' The time is returned with a nanosecond precision, but the actual amount
#' of data returned is OS dependent. Usually, information at at least the
#' microsecond level is returned, with some platforms returning nanosecond
#' information.
#'
#' @return A sys-time of the current time in UTC.
#'
#' @export
#' @examples
#' x <- sys_time_now()
sys_time_now <- function() {
  names <- NULL
  fields <- sys_time_now_cpp()
  new_sys_time_from_fields(fields, PRECISION_NANOSECOND, names)
}

# ------------------------------------------------------------------------------

#' Info: sys-time
#'
#' @description
#' `sys_time_info()` retrieves a set of low-level information generally not
#' required for most date-time manipulations. It returns a data frame with the
#' following columns:
#'
#' - `begin`, `end`: Second precision sys-times specifying the range of the
#' current daylight saving time rule. The range is a half-open interval of
#' `[begin, end)`.
#'
#' - `offset`: A second precision `duration` specifying the offset from UTC.
#'
#' - `dst`: A logical vector specifying if daylight saving time is currently
#' active.
#'
#' - `abbreviation`: The time zone abbreviation in use throughout this `begin`
#' to `end` range.
#'
#' @details
#' If there have never been any daylight saving time transitions, the minimum
#' supported year value is returned for `begin` (typically, a year value of
#' `-32767`).
#'
#' If daylight saving time is no longer used in a time zone, the maximum
#' supported year value is returned for `end` (typically, a year value of
#' `32767`).
#'
#' The `offset` is the bridge between sys-time and naive-time for the `zone`
#' being used. The relationship of the three values is:
#'
#' ```
#' offset = naive_time - sys_time
#' ```
#'
#' @param x `[clock_sys_time]`
#'
#'   A sys-time.
#'
#' @param zone `[character]`
#'
#'   A valid time zone name.
#'
#'   Unlike most functions in clock, in `sys_time_info()` `zone` is vectorized
#'   and is recycled against `x`.
#'
#' @return A data frame of low level information.
#'
#' @export
#' @examples
#' library(vctrs)
#'
#' x <- year_month_day(2021, 03, 14, c(01, 03), c(59, 00), c(59, 00))
#' x <- as_naive_time(x)
#' x <- as_zoned_time(x, "America/New_York")
#'
#' # x[1] is in EST, x[2] is in EDT
#' x
#'
#' x_sys <- as_sys_time(x)
#'
#' info <- sys_time_info(x_sys, zoned_time_zone(x))
#' info
#'
#' # Convert `begin` and `end` to zoned-times to see the previous and
#' # next daylight saving time transitions
#' data_frame(
#'   x = x,
#'   begin = as_zoned_time(info$begin, zoned_time_zone(x)),
#'   end = as_zoned_time(info$end, zoned_time_zone(x))
#' )
#'
#' # `end` can be used to iterate through daylight saving time transitions
#' # by repeatedly calling `sys_time_info()`
#' sys_time_info(info$end, zoned_time_zone(x))
#'
#' # Multiple `zone`s can be supplied to look up daylight saving time
#' # information in different time zones
#' zones <- c("America/New_York", "America/Los_Angeles")
#'
#' info2 <- sys_time_info(x_sys[1], zones)
#' info2
#'
#' # The offset can be used to display the naive-time (i.e. the printed time)
#' # in both of those time zones
#' data_frame(
#'   zone = zones,
#'   naive_time = x_sys[1] + info2$offset
#' )
sys_time_info <- function(x, zone) {
  if (!is_sys_time(x)) {
    abort("`x` must be a sys-time.")
  }

  precision <- time_point_precision_attribute(x)

  # Recycle `x` to the common size. `zone` is recycled internally as required,
  # which is more efficient than reloading the time zone repeatedly.
  size <- vec_size_common(x = x, zone = zone)
  x <- vec_recycle(x, size)

  fields <- sys_time_info_cpp(x, precision, zone)

  new_sys_time_info_from_fields(fields)
}

new_sys_time_info_from_fields <- function(fields) {
  names <- NULL

  fields[["begin"]] <- new_sys_time_from_fields(fields[["begin"]], PRECISION_SECOND, names)
  fields[["end"]] <- new_sys_time_from_fields(fields[["end"]], PRECISION_SECOND, names)
  fields[["offset"]] <- new_duration_from_fields(fields[["offset"]], PRECISION_SECOND, names)

  new_data_frame(fields)
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype.clock_sys_time <- function(x, ...) {
  switch(
    time_point_precision_attribute(x) + 1L,
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    abort("Internal error: Invalid precision"),
    clock_empty_sys_time_day,
    clock_empty_sys_time_hour,
    clock_empty_sys_time_minute,
    clock_empty_sys_time_second,
    clock_empty_sys_time_millisecond,
    clock_empty_sys_time_microsecond,
    clock_empty_sys_time_nanosecond,
    abort("Internal error: Invalid precision.")
  )
}

#' @export
vec_ptype2.clock_sys_time.clock_sys_time <- function(x, y, ...) {
  ptype2_time_point_and_time_point(x, y, ...)
}

#' @export
vec_cast.clock_sys_time.clock_sys_time <- function(x, to, ...) {
  cast_time_point_to_time_point(x, to, ...)
}

# ------------------------------------------------------------------------------

#' @rdname clock-arith
#' @method vec_arith clock_sys_time
#' @export
vec_arith.clock_sys_time <- function(op, x, y, ...) {
  UseMethod("vec_arith.clock_sys_time", y)
}

#' @method vec_arith.clock_sys_time MISSING
#' @export
vec_arith.clock_sys_time.MISSING <- function(op, x, y, ...) {
  arith_time_point_and_missing(op, x, y, ...)
}

#' @method vec_arith.clock_sys_time clock_sys_time
#' @export
vec_arith.clock_sys_time.clock_sys_time <- function(op, x, y, ...) {
  arith_time_point_and_time_point(op, x, y, ...)
}

#' @method vec_arith.clock_sys_time clock_duration
#' @export
vec_arith.clock_sys_time.clock_duration <- function(op, x, y, ...) {
  arith_time_point_and_duration(op, x, y, ...)
}

#' @method vec_arith.clock_duration clock_sys_time
#' @export
vec_arith.clock_duration.clock_sys_time <- function(op, x, y, ...) {
  arith_duration_and_time_point(op, x, y, ...)
}

#' @method vec_arith.clock_sys_time numeric
#' @export
vec_arith.clock_sys_time.numeric <- function(op, x, y, ...) {
  arith_time_point_and_numeric(op, x, y, ...)
}

#' @method vec_arith.numeric clock_sys_time
#' @export
vec_arith.numeric.clock_sys_time <- function(op, x, y, ...) {
  arith_numeric_and_time_point(op, x, y, ...)
}

# ------------------------------------------------------------------------------

clock_init_sys_time_utils <- function(env) {
  day <- as_sys_time(year_month_day(integer(), integer(), integer()))

  assign("clock_empty_sys_time_day", day, envir = env)
  assign("clock_empty_sys_time_hour", time_point_cast(day, "hour"), envir = env)
  assign("clock_empty_sys_time_minute", time_point_cast(day, "minute"), envir = env)
  assign("clock_empty_sys_time_second", time_point_cast(day, "second"), envir = env)
  assign("clock_empty_sys_time_millisecond", time_point_cast(day, "millisecond"), envir = env)
  assign("clock_empty_sys_time_microsecond", time_point_cast(day, "microsecond"), envir = env)
  assign("clock_empty_sys_time_nanosecond", time_point_cast(day, "nanosecond"), envir = env)

  invisible(NULL)
}

# ------------------------------------------------------------------------------
# sys_time_now()

test_that("returns nanosecond precision", {
  x <- sys_time_now()
  expect_identical(time_point_precision_attribute(x), PRECISION_NANOSECOND)
})

test_that("returns a single sys-time", {
  x <- sys_time_now()
  expect_length(x, 1L)
  expect_s3_class(x, "clock_sys_time")
})

# ------------------------------------------------------------------------------
# sys_time_info()

test_that("can lookup sys-info", {
  # One in EST, one in EDT
  x <- year_month_day(2021, 03, 14, c(01, 03), c(59, 00), c(59, 00))
  x <- as_naive_time(x)
  x <- as_zoned_time(x, "America/New_York")

  info <- sys_time_info(as_sys_time(x), zoned_time_zone(x))

  beginend1 <- as_sys_time(c(
    year_month_day(2020, 11, 1, 6, 0, 0),
    year_month_day(2021, 03, 14, 7, 0, 0)
  ))
  beginend2 <- as_sys_time(c(
    year_month_day(2021, 03, 14, 7, 0, 0),
    year_month_day(2021, 11, 7, 6, 0, 0)
  ))

  expect_identical(info$begin, c(beginend1[1], beginend2[1]))
  expect_identical(info$end, c(beginend1[2], beginend2[2]))

  expect_identical(info$offset, duration_seconds(c(-18000, -14400)))
  expect_identical(info$dst, c(FALSE, TRUE))
  expect_identical(info$abbreviation, c("EST", "EDT"))
})

test_that("`zone` is vectorized and recycled against `x`", {
  zones <- c("America/New_York", "Australia/Lord_Howe")
  x <- as_sys_time(year_month_day(2019, 1, 1))

  info <- sys_time_info(x, zones)

  naive_times <- c(
    as_naive_time(as_zoned_time(x, zones[1])),
    as_naive_time(as_zoned_time(x, zones[2]))
  )

  expect_identical(as_naive_time(x + info$offset), naive_times)

  # DST is dependent on the time zone
  expect_identical(info$dst, c(FALSE, TRUE))

  expect_identical(info$abbreviation, c("EST", "+11"))
})

test_that("very old times are looked up correctly", {
  x <- year_month_day(1800, 01, 01)
  x <- as_sys_time(x)

  info <- sys_time_info(x, "America/New_York")

  end <- as_sys_time(year_month_day(1883, 11, 18, 17, 00, 00))
  offset <- duration_seconds(-17762)
  dst <- FALSE
  abbreviation <- "LMT"

  expect_identical(info$end, end)
  expect_identical(info$offset, offset)
  expect_identical(info$dst, dst)
  expect_identical(info$abbreviation, abbreviation)
})

# ------------------------------------------------------------------------------
# as.character()

test_that("as.character() works", {
  x <- as_sys_time(year_month_day(2019, 1, 1))
  expect_identical(as.character(x), "2019-01-01")

  x <- as_sys_time(year_month_day(2019, 1, 1, 1, 1))
  expect_identical(as.character(x), "2019-01-01T01:01")
})

# ------------------------------------------------------------------------------
# sys_time_parse()

# Note: Most tests are in `naive_time_parse()`. They share an implementation.

test_that("can parse day precision", {
  x <- c("2019-01-01", "2019-01-31")

  expect_identical(
    sys_time_parse(x, precision = "day"),
    as_sys_time(year_month_day(2019, 1, c(1, 31)))
  )
})

test_that("%z shifts the result by the offset", {
  x <- "2019-01-01T00:00:00+0100"
  y <- "2019-01-01T00:00:00-0100"

  expect_identical(
    sys_time_parse(x, format = "%Y-%m-%dT%H:%M:%S%z"),
    as_sys_time(year_month_day(2018, 12, 31, 23, 0, 0))
  )
  expect_identical(
    sys_time_parse(y, format = "%Y-%m-%dT%H:%M:%S%z"),
    as_sys_time(year_month_day(2019, 1, 1, 1, 0, 0))
  )
})

test_that("failure to parse throws a warning", {
  expect_warning(sys_time_parse("foo"), class = "clock_warning_parse_failures")
  expect_snapshot(sys_time_parse("foo"))
})

# ------------------------------------------------------------------------------
# sys_time_parse_RFC_3339()

test_that("can parse default RFC 3339 format", {
  x <- "2019-01-01T00:00:00Z"
  expect_identical(
    sys_time_parse_RFC_3339(x),
    as_sys_time(year_month_day(2019, 1, 1, 0, 0, 0))
  )
})

test_that("can parse with fractional seconds", {
  x <- "2019-01-01T00:00:00.123Z"
  expect_identical(
    sys_time_parse_RFC_3339(x, precision = "millisecond"),
    as_sys_time(year_month_day(2019, 1, 1, 0, 0, 0, 123, subsecond_precision = "millisecond"))
  )
})

test_that("can parse with alternative separator", {
  x <- "2019-01-01t00:00:00Z"
  expect_identical(
    sys_time_parse_RFC_3339(x, separator = "t"),
    as_sys_time(year_month_day(2019, 1, 1, 0, 0, 0))
  )

  x <- "2019-01-01 00:00:00Z"
  expect_identical(
    sys_time_parse_RFC_3339(x, separator = " "),
    as_sys_time(year_month_day(2019, 1, 1, 0, 0, 0))
  )
})

test_that("can parse with alternative offset", {
  x <- "2019-01-01T00:00:00z"
  expect_identical(
    sys_time_parse_RFC_3339(x, offset = "z"),
    as_sys_time(year_month_day(2019, 1, 1, 0, 0, 0))
  )

  x <- "2019-01-01T00:00:00-0130"
  expect_identical(
    sys_time_parse_RFC_3339(x, offset = "%z"),
    as_sys_time(year_month_day(2019, 1, 1, 1, 30, 0))
  )

  x <- "2019-01-01T00:00:00-01:30"
  expect_identical(
    sys_time_parse_RFC_3339(x, offset = "%Ez"),
    as_sys_time(year_month_day(2019, 1, 1, 1, 30, 0))
  )
})

test_that("`precision` must be at least second", {
  x <- "2019-01-01T00:00:00Z"
  expect_snapshot(error = TRUE, sys_time_parse_RFC_3339(x, precision = "day"))
})

test_that("`separator` is validated", {
  x <- "2019-01-01T00:00:00Z"
  expect_snapshot(error = TRUE, sys_time_parse_RFC_3339(x, separator = 1))
  expect_snapshot(error = TRUE, sys_time_parse_RFC_3339(x, separator = "TT"))
})

test_that("`offset` is validated", {
  x <- "2019-01-01T00:00:00Z"
  expect_snapshot(error = TRUE, sys_time_parse_RFC_3339(x, offset = 1))
  expect_snapshot(error = TRUE, sys_time_parse_RFC_3339(x, offset = "ZZ"))
})

test_that("sys-time-parse-RFC-3339: empty dots are checked", {
  x <- "2019-01-01T00:00:00Z"
  expect_snapshot(error = TRUE, sys_time_parse_RFC_3339(x, 1))
})

# ------------------------------------------------------------------------------
# format()

test_that("default format is correct", {
  expect_snapshot(format(sys_seconds(0)))
})

test_that("allows `%z` and `%Z`", {
  x <- sys_seconds(0)
  expect_identical(format(x, format = "%z"), "+0000")
  expect_identical(format(x, format = "%Ez"), "+00:00")
  expect_identical(format(x, format = "%Z"), "UTC")
})

# ------------------------------------------------------------------------------
# as_zoned_time()

test_that("empty dots are checked", {
  expect_snapshot_error(as_zoned_time(sys_seconds(), "UTC", 123))
})

# ------------------------------------------------------------------------------
# vec_ptype()

test_that("ptype is correct", {
  base <- sys_days(0)
  ptype <- sys_days(integer())

  for (precision in precision_names()) {
    if (validate_precision_string(precision) < PRECISION_DAY) {
      next
    }

    x <- time_point_cast(base, precision)
    expect <- time_point_cast(ptype, precision)

    expect_identical(vec_ptype(x), expect)
  }
})

# ------------------------------------------------------------------------------
# vec_math()

test_that("is.nan() works", {
  x <- sys_days(c(2019, NA))
  expect_identical(is.nan(x), c(FALSE, FALSE))
})

test_that("is.finite() works", {
  x <- sys_days(c(2019, NA))
  expect_identical(is.finite(x), c(TRUE, FALSE))
})

test_that("is.infinite() works", {
  x <- sys_days(c(2019, NA))
  expect_identical(is.infinite(x), c(FALSE, FALSE))
})

#' Calendar setters
#'
#' @description
#' This family of functions sets fields in a calendar vector. Each
#' calendar has its own set of supported setters, which are documented on
#' their own help page:
#'
#' - [year-month-day][year-month-day-setters]
#'
#' - [year-month-weekday][year-month-weekday-setters]
#'
#' - [iso-year-week-day][iso-year-week-day-setters]
#'
#' - [year-quarter-day][year-quarter-day-setters]
#'
#' - [year-day][year-day-setters]
#'
#' There are also convenience methods for setting certain components
#' directly on R's native date and date-time types.
#'
#' - [dates (Date)][Date-setters]
#'
#' - [date-times (POSIXct / POSIXlt)][posixt-setters]
#'
#' Some general rules about setting components on calendar types:
#'
#' - You can only set components that are relevant to the calendar type that
#'   you are working with. For example, you can't set the quarter of a
#'   year-month-day type. You'd have to convert to year-quarter-day first.
#'
#' - You can set a component that is at the current precision, or one level
#'   of precision more precise than the current precision. For example,
#'   you can set the day field of a month precision year-month-day type,
#'   but not the hour field.
#'
#' - Setting a component can result in an _invalid date_, such as
#'   `set_day(year_month_day(2019, 02), 31)`, as long as it is eventually
#'   resolved either manually or with a strategy from [invalid_resolve()].
#'
#' - With sub-second precisions, you can only set the component corresponding
#'   to the precision that you are at. For example, you can set the nanoseconds
#'   of the second while at nanosecond precision, but not milliseconds.
#'
#' @details
#' You cannot set components directly on a time point type, such as
#' sys-time or naive-time. Convert it to a calendar type first. Similarly,
#' a zoned-time must be converted to either a sys-time or naive-time, and
#' then to a calendar type, to be able to set components on it.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[object]`
#'
#'   An object to set the component for.
#'
#' @param value `[integer]`
#'
#'   The value to set the component to.
#'
#' @return `x` with the component set.
#'
#' @name clock-setters
#' @examples
#' x <- year_month_day(2019, 1:3)
#'
#' # Set the day
#' set_day(x, 12:14)
#'
#' # Set to the "last" day of the month
#' set_day(x, "last")
NULL

#' @rdname clock-setters
#' @export
set_year <- function(x, value, ...) {
  UseMethod("set_year")
}

#' @rdname clock-setters
#' @export
set_quarter <- function(x, value, ...) {
  UseMethod("set_quarter")
}

#' @rdname clock-setters
#' @export
set_month <- function(x, value, ...) {
  UseMethod("set_month")
}

#' @rdname clock-setters
#' @export
set_week <- function(x, value, ...) {
  UseMethod("set_week")
}

#' @rdname clock-setters
#' @export
set_day <- function(x, value, ...) {
  UseMethod("set_day")
}

#' @rdname clock-setters
#' @export
set_hour <- function(x, value, ...) {
  UseMethod("set_hour")
}

#' @rdname clock-setters
#' @export
set_minute <- function(x, value, ...) {
  UseMethod("set_minute")
}

#' @rdname clock-setters
#' @export
set_second <- function(x, value, ...) {
  UseMethod("set_second")
}

#' @rdname clock-setters
#' @export
set_millisecond <- function(x, value, ...) {
  UseMethod("set_millisecond")
}

#' @rdname clock-setters
#' @export
set_microsecond <- function(x, value, ...) {
  UseMethod("set_microsecond")
}

#' @rdname clock-setters
#' @export
set_nanosecond <- function(x, value, ...) {
  UseMethod("set_nanosecond")
}

#' @rdname clock-setters
#' @export
set_index <- function(x, value, ...) {
  UseMethod("set_index")
}

# ------------------------------------------------------------------------------
# year_month_weekday()

test_that("helper can create different precisions", {
  x <- year_month_weekday(2019, 1:2)
  expect_identical(get_year(x), c(2019L, 2019L))
  expect_identical(get_month(x), 1:2)

  x <- year_month_weekday(2019, 1:2, clock_weekdays$monday, 2)
  expect_identical(get_day(x), c(clock_weekdays$monday, clock_weekdays$monday))
  expect_identical(get_index(x), c(2L, 2L))
})

test_that("can create subsecond precision calendars", {
  x <- year_month_weekday(
    2019,
    1,
    1,
    1,
    0,
    0,
    0,
    1,
    subsecond_precision = "millisecond"
  )
  expect_identical(get_millisecond(x), 1L)

  x <- year_month_weekday(
    2019,
    1,
    1,
    1,
    0,
    0,
    0,
    1,
    subsecond_precision = "microsecond"
  )
  expect_identical(get_microsecond(x), 1L)

  x <- year_month_weekday(2019, 1, 1, 1, 0, 0, 0, 1, subsecond_precision = "nanosecond")
  expect_identical(get_nanosecond(x), 1L)
})

test_that("both day and index must be specified", {
  expect_snapshot(error = TRUE, year_month_weekday(2020, 1, 1))
})

test_that("validates value ranges", {
  expect_snapshot(error = TRUE, year_month_weekday(50000))
  expect_snapshot(error = TRUE, year_month_weekday(2020, 13))
  expect_snapshot(error = TRUE, year_month_weekday(2020, 1, 32, 1))
  expect_snapshot(error = TRUE, year_month_weekday(2020, 1, 1, 6))
  expect_snapshot(error = TRUE, year_month_weekday(2020, 1, 1, 1, 24))
  expect_snapshot(error = TRUE, year_month_weekday(2020, 1, 1, 1, 1, 60))
  expect_snapshot(error = TRUE, year_month_weekday(2020, 1, 1, 1, 1, 1, 60))
  expect_snapshot(
    error = TRUE,
    year_month_weekday(
      2020,
      1,
      1,
      1,
      1,
      1,
      1,
      1000,
      subsecond_precision = "millisecond"
    )
  )
  expect_snapshot(
    error = TRUE,
    year_month_weekday(
      2020,
      1,
      1,
      1,
      1,
      1,
      1,
      1000000,
      subsecond_precision = "microsecond"
    )
  )
  expect_snapshot(
    error = TRUE,
    year_month_weekday(
      2020,
      1,
      1,
      1,
      1,
      1,
      1,
      1000000000,
      subsecond_precision = "nanosecond"
    )
  )
})

test_that("can get the last indexed weekday of the month", {
  x <- year_month_weekday(2019, 1:4, clock_weekdays$monday, "last")
  expect_identical(get_index(x), c(4L, 4L, 4L, 5L))
})

test_that("`NA` propagates through 'last'", {
  x <- year_month_weekday(2019, c(1, NA), clock_weekdays$monday, 1)
  x <- set_index(x, "last")
  expect_identical(get_index(x), c(4L, NA))
})

test_that("ignores values past first `NULL`", {
  expect_identical(year_month_weekday(2019, minute = 2), year_month_weekday(2019))
})

test_that("NA values propagate", {
  x <- year_month_weekday(2019, 1:3, c(NA, 2, 3), c(3, 4, NA))
  expect_identical(is.na(x), c(TRUE, FALSE, TRUE))
})

# ------------------------------------------------------------------------------
# vec_ptype()

test_that("ptype is correct", {
  base <- year_month_weekday(1)
  ptype <- year_month_weekday(integer())

  for (precision in precision_names()) {
    if (precision == "quarter" || precision == "week") {
      next
    }

    x <- calendar_widen(base, precision)
    expect <- calendar_widen(ptype, precision)

    expect_identical(vec_ptype(x), expect)
  }
})

# ------------------------------------------------------------------------------
# vec_proxy() / vec_restore()

test_that("proxy is a data frame", {
  expect_identical(vec_proxy(year_month_weekday(2019)), data_frame(year = 2019L))
  expect_identical(
    vec_proxy(year_month_weekday(2019, 1)),
    data_frame(year = 2019L, month = 1L)
  )
})

test_that("proxy has names on `year`", {
  x <- set_names(year_month_weekday(2019, 1), "nm")
  year <- vec_proxy(x)$year
  expect_named(year, "nm")
})

test_that("restore works", {
  to <- year_month_weekday(2019, 1:5)
  proxy <- vec_slice(vec_proxy(to), 1:2)
  expect_identical(vec_restore(proxy, to), year_month_weekday(2019, 1:2))
})

# ------------------------------------------------------------------------------
# vec_proxy_compare()

test_that("can compare with year / month precision", {
  expect_identical(
    year_month_weekday(2019) > year_month_weekday(2018:2020),
    c(TRUE, FALSE, FALSE)
  )
  expect_identical(
    year_month_weekday(2019, 2) > year_month_weekday(2019, 1:3),
    c(TRUE, FALSE, FALSE)
  )
})

test_that("cannot compare / sort with day precision or finer", {
  x <- year_month_weekday(2019, 1, 1, 1)

  expect_snapshot(error = TRUE, x > x)
  expect_snapshot(error = TRUE, vec_order(x))
})

# ------------------------------------------------------------------------------
# vec_ptype_full()

test_that("full ptype is correct", {
  expect_snapshot_output(vec_ptype_full(year_month_weekday(2019)))
  expect_snapshot_output(vec_ptype_full(year_month_weekday(2019, 1, 1, 2)))
  expect_snapshot_output(
    vec_ptype_full(
      year_month_weekday(2019, 1, 1, 2, 1, 1, 1, 1, subsecond_precision = "nanosecond")
    )
  )
  expect_snapshot_output(vec_ptype_full(year_month_weekday(2019, 1, 1, 5)))
})

# ------------------------------------------------------------------------------
# vec_ptype_abbr()

test_that("abbreviated ptype is correct", {
  expect_snapshot_output(vec_ptype_abbr(year_month_weekday(2019)))
  expect_snapshot_output(vec_ptype_abbr(year_month_weekday(2019, 1, 1, 2)))
  expect_snapshot_output(
    vec_ptype_abbr(
      year_month_weekday(2019, 1, 1, 2, 1, 1, 1, 1, subsecond_precision = "nanosecond")
    )
  )
  expect_snapshot_output(vec_ptype_abbr(year_month_weekday(2019, 1, 1, 5)))
})

# ------------------------------------------------------------------------------
# set_*()

test_that("setters work", {
  x <- year_month_weekday(1L)

  x <- set_year(x, 2L)
  expect_identical(get_year(x), 2L)

  x <- set_month(x, 1L)
  expect_identical(get_month(x), 1L)

  x <- set_day(x, 2L, index = 1L)
  expect_identical(get_day(x), 2L)
  expect_identical(get_index(x), 1L)

  x <- set_index(x, 3L)
  expect_identical(get_index(x), 3L)

  x <- set_hour(x, 3L)
  expect_identical(get_hour(x), 3L)

  x <- set_minute(x, 4L)
  expect_identical(get_minute(x), 4L)

  x <- set_second(x, 5L)
  expect_identical(get_second(x), 5L)

  ms <- set_millisecond(x, 6L)
  expect_identical(get_millisecond(ms), 6L)

  us <- set_microsecond(x, 7L)
  expect_identical(get_microsecond(us), 7L)

  ns <- set_nanosecond(x, 8L)
  expect_identical(get_nanosecond(ns), 8L)
})

test_that("setters propagate all missings", {
  x <- year_month_weekday(2019, c(1, NA, 3, 4))
  x <- set_day(x, c(NA, 2, 4, 5), index = c(1, 2, 3, NA))
  expect_identical(vec_detect_missing(x), c(TRUE, TRUE, FALSE, TRUE))
})

test_that("setters recycling works both ways", {
  x <- year_month_weekday(2019)

  x <- set_month(x, 1:2)
  expect_identical(x, year_month_weekday(2019, 1:2))

  x <- set_day(x, 1, index = 1)
  expect_identical(x, year_month_weekday(2019, 1:2, 1, 1))

  expect_snapshot(error = TRUE, {
    x <- year_month_weekday(1:2)
    y <- 1:3
    set_month(x, y)
  })
})

test_that("setters require integer `value`", {
  x <- year_month_weekday(2019, 1, 2, 3, 4, 5, 6)

  expect_snapshot(error = TRUE, {
    set_year(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_month(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_day(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_index(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_hour(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_minute(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_second(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_millisecond(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_microsecond(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_nanosecond(x, 1.5)
  })
})

test_that("setters check `value` range", {
  x <- year_month_weekday(2019, 1, 2, 3, 4, 5, 6)

  expect_snapshot(error = TRUE, {
    set_year(x, 100000)
  })
  expect_snapshot(error = TRUE, {
    set_month(x, 13)
  })
  expect_snapshot(error = TRUE, {
    set_day(x, 8)
  })
  expect_snapshot(error = TRUE, {
    set_index(x, 6)
  })
  expect_snapshot(error = TRUE, {
    set_hour(x, 24)
  })
  expect_snapshot(error = TRUE, {
    set_minute(x, 60)
  })
  expect_snapshot(error = TRUE, {
    set_second(x, 60)
  })
  expect_snapshot(error = TRUE, {
    set_millisecond(x, -1)
  })
  expect_snapshot(error = TRUE, {
    set_microsecond(x, -1)
  })
  expect_snapshot(error = TRUE, {
    set_nanosecond(x, -1)
  })
})

test_that("setters require minimum precision", {
  expect_snapshot(error = TRUE, {
    set_day(year_month_weekday(year = 1), 1, index = 1)
  })
  expect_snapshot(error = TRUE, {
    set_hour(year_month_weekday(year = 1, month = 2), 1)
  })
  expect_snapshot(error = TRUE, {
    set_minute(year_month_weekday(year = 1, month = 2, day = 3, index = 1), 1)
  })
  expect_snapshot(error = TRUE, {
    set_second(year_month_weekday(year = 1, month = 2, day = 3, index = 1, hour = 4), 1)
  })
  expect_snapshot(error = TRUE, {
    set_millisecond(
      year_month_weekday(year = 1, month = 2, day = 3, index = 1, hour = 4, minute = 5),
      1
    )
  })
  expect_snapshot(error = TRUE, {
    set_microsecond(
      year_month_weekday(year = 1, month = 2, day = 3, index = 1, hour = 4, minute = 5),
      1
    )
  })
  expect_snapshot(error = TRUE, {
    set_nanosecond(
      year_month_weekday(year = 1, month = 2, day = 3, index = 1, hour = 4, minute = 5),
      1
    )
  })
})

test_that("setters require correct subsecond precision", {
  expect_snapshot(error = TRUE, {
    set_millisecond(
      year_month_weekday(
        year = 1,
        month = 2,
        day = 3,
        index = 1,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "microsecond"
      ),
      1
    )
  })
  expect_snapshot(error = TRUE, {
    set_millisecond(
      year_month_weekday(
        year = 1,
        month = 2,
        day = 3,
        index = 1,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "nanosecond"
      ),
      1
    )
  })

  expect_snapshot(error = TRUE, {
    set_microsecond(
      year_month_weekday(
        year = 1,
        month = 2,
        day = 3,
        index = 1,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "millisecond"
      ),
      1
    )
  })
  expect_snapshot(error = TRUE, {
    set_microsecond(
      year_month_weekday(
        year = 1,
        month = 2,
        day = 3,
        index = 1,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "nanosecond"
      ),
      1
    )
  })

  expect_snapshot(error = TRUE, {
    set_nanosecond(
      year_month_weekday(
        year = 1,
        month = 2,
        day = 3,
        index = 1,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "millisecond"
      ),
      1
    )
  })
  expect_snapshot(error = TRUE, {
    set_nanosecond(
      year_month_weekday(
        year = 1,
        month = 2,
        day = 3,
        index = 1,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "microsecond"
      ),
      1
    )
  })
})

test_that("setters retain names", {
  x <- year_month_weekday(2019)
  x <- set_names(x, "foo")
  expect_named(set_month(x, 2), "foo")
})

test_that("setting with named `value` strips its names", {
  x <- year_month_weekday(2019)
  x <- set_month(x, set_names(1L, "x"))
  expect_named(field(x, "month"), NULL)
})

# ------------------------------------------------------------------------------
# format()

test_that("default formats are correct", {
  expect_snapshot(format(year_month_weekday(2019)))
  expect_snapshot(format(year_month_weekday(2019, 1)))
  expect_snapshot(format(year_month_weekday(2019, 1, 1, 2, 1)))
  expect_snapshot(
    format(
      year_month_weekday(
        2019,
        1,
        1,
        2,
        1,
        2,
        3,
        50,
        subsecond_precision = "microsecond"
      )
    )
  )
})

# ------------------------------------------------------------------------------
# as.character()

test_that("as.character() works", {
  x <- year_month_weekday(2019, 1)
  y <- year_month_weekday(2019, 1, 2, 1)

  expect_identical(as.character(x), format(x))
  expect_identical(as.character(y), format(y))
})

# ------------------------------------------------------------------------------
# calendar_narrow()

test_that("can narrow to month", {
  x_expect <- year_month_weekday(2019, 2)
  x <- set_day(x_expect, 1, index = 2)
  expect_identical(calendar_narrow(x, "month"), x_expect)
  expect_identical(calendar_narrow(x_expect, "month"), x_expect)
})

test_that("can narrow to day", {
  x_expect <- year_month_weekday(2019, 2, 3, 2)
  x <- set_hour(x_expect, 5)
  expect_identical(calendar_narrow(x, "day"), x_expect)
  expect_identical(calendar_narrow(x_expect, "day"), x_expect)
})

# ------------------------------------------------------------------------------
# calendar_widen()

test_that("can widen to month", {
  x <- year_month_weekday(2019)
  expect_identical(calendar_widen(x, "month"), set_month(x, 1))
})

test_that("can widen to day", {
  x <- year_month_weekday(2019)
  y <- year_month_weekday(2019, 02)
  expect_identical(calendar_widen(x, "day"), set_day(set_month(x, 1), 1, index = 1))
  expect_identical(calendar_widen(y, "day"), set_day(y, 1, index = 1))
})

# ------------------------------------------------------------------------------
# calendar_start()

test_that("can compute year start", {
  x <- year_month_weekday(2019)
  expect_identical(calendar_start(x, "year"), x)

  x <- year_month_weekday(2019, 2)
  expect_identical(calendar_start(x, "year"), year_month_weekday(2019, 1))
})

test_that("can compute month start", {
  x <- year_month_weekday(2019, 2)
  expect_identical(calendar_start(x, "month"), x)
})

test_that("can't compute start with a year_month_weekday at day precision or greater", {
  expect_snapshot(
    error = TRUE,
    calendar_start(year_month_weekday(2019, 2, 2, 2), "day")
  )
  expect_snapshot(
    error = TRUE,
    calendar_start(year_month_weekday(2019, 2, 2, 2), "month")
  )
})

# ------------------------------------------------------------------------------
# calendar_end()

test_that("can compute year end", {
  x <- year_month_weekday(2019)
  expect_identical(calendar_end(x, "year"), x)

  x <- year_month_weekday(2019, 2)
  expect_identical(calendar_end(x, "year"), year_month_weekday(2019, 12))
})

test_that("can compute month end", {
  x <- year_month_weekday(2019, 2)
  expect_identical(calendar_end(x, "month"), x)
})

test_that("can't compute end with a year_month_weekday at day precision or greater", {
  expect_snapshot(error = TRUE, calendar_end(year_month_weekday(2019, 2, 2, 2), "day"))
  expect_snapshot(
    error = TRUE,
    calendar_end(year_month_weekday(2019, 2, 2, 2), "month")
  )
})

# ------------------------------------------------------------------------------
# calendar_month_factor()

test_that("can get a month factor", {
  expect_identical(
    calendar_month_factor(year_month_weekday(2019, 1:12)),
    factor(month.name, levels = month.name, ordered = TRUE)
  )
})

# ------------------------------------------------------------------------------
# calendar_count_between()

test_that("can compute year and month counts", {
  x <- year_month_weekday(2019, 1)
  y <- year_month_weekday(2020, 3)

  expect_identical(calendar_count_between(x, y, "year"), 1L)
  expect_identical(calendar_count_between(x, y, "month"), 14L)
  expect_identical(calendar_count_between(x, y, "month", n = 2), 7L)
})

test_that("can compute a quarter count", {
  x <- year_month_weekday(2019, 1)

  y <- year_month_weekday(2019, c(3, 5))
  expect_identical(calendar_count_between(x, y, "quarter"), c(0L, 1L))
  expect_identical(
    calendar_count_between(x, y, "quarter"),
    calendar_count_between(x, y, "month", n = 3L)
  )

  y <- year_month_weekday(2020, c(3, 5))
  expect_identical(calendar_count_between(x, y, "quarter", n = 2L), c(2L, 2L))
  expect_identical(
    calendar_count_between(x, y, "quarter", n = 2L),
    calendar_count_between(x, y, "month", n = 6L)
  )
})

test_that("positive / negative counts are correct", {
  start <- year_month_weekday(1972, 04)

  end <- year_month_weekday(1973, 03)
  expect_identical(calendar_count_between(start, end, "year"), 0L)
  expect_identical(calendar_count_between(start, end, "month"), 11L)

  end <- year_month_weekday(1973, 04)
  expect_identical(calendar_count_between(start, end, "year"), 1L)
  expect_identical(calendar_count_between(start, end, "month"), 12L)

  end <- year_month_weekday(1973, 05)
  expect_identical(calendar_count_between(start, end, "year"), 1L)
  expect_identical(calendar_count_between(start, end, "month"), 13L)

  end <- year_month_weekday(1971, 03)
  expect_identical(calendar_count_between(start, end, "year"), -1L)
  expect_identical(calendar_count_between(start, end, "month"), -13L)

  end <- year_month_weekday(1971, 04)
  expect_identical(calendar_count_between(start, end, "year"), -1L)
  expect_identical(calendar_count_between(start, end, "month"), -12L)

  end <- year_month_weekday(1971, 05)
  expect_identical(calendar_count_between(start, end, "year"), 0L)
  expect_identical(calendar_count_between(start, end, "month"), -11L)
})

test_that("can't compare a 'year_month_weekday' with day precision!", {
  x <- year_month_weekday(2019, 1, 1, 1)
  expect_snapshot((expect_error(calendar_count_between(x, x, "month"))))
})

# ------------------------------------------------------------------------------
# seq()

test_that("only granular precisions are allowed", {
  expect_snapshot(
    error = TRUE,
    seq(year_month_weekday(2019, 1, 1, 1), by = 1, length.out = 2)
  )
})

# NOTE: Most tests are done by `year_month_day()` since they share an implementation
test_that("can generate a sequence", {
  expect_identical(
    seq(year_month_day(2019, 1), to = year_month_day(2019, 6), by = 2),
    year_month_day(2019, c(1, 3, 5))
  )
})

# ------------------------------------------------------------------------------
# invalid_detect()

test_that("`invalid_detect()` works", {
  # Not possible to be invalid
  x <- year_month_weekday(2019:2020, 1:2)
  expect_identical(invalid_detect(x), c(FALSE, FALSE))

  # Now possible
  x <- year_month_weekday(2019, 1, 1, c(1, 2, 5, NA))
  expect_identical(invalid_detect(x), c(FALSE, FALSE, TRUE, FALSE))

  # Possible after that too
  x <- year_month_weekday(2019, 1, 1, c(1, 2, 5, NA), 3)
  expect_identical(invalid_detect(x), c(FALSE, FALSE, TRUE, FALSE))
})

# ------------------------------------------------------------------------------
# invalid_any()

test_that("`invalid_any()` works", {
  # Not possible to be invalid
  x <- year_month_weekday(2019:2020, 1:2)
  expect_false(invalid_any(x))

  # Now possible
  x <- year_month_weekday(2019, 1, 1, c(1, 2, 5, NA))
  expect_true(invalid_any(x))

  # Possible after that too
  x <- year_month_weekday(2019, 1, 1, c(1, 2, 5, NA), 3)
  expect_true(invalid_any(x))
})

# ------------------------------------------------------------------------------
# invalid_count()

test_that("`invalid_count()` works", {
  # Not possible to be invalid
  x <- year_month_weekday(2019:2020, 1:2)
  expect_identical(invalid_count(x), 0L)

  # Now possible
  x <- year_month_weekday(2019, 1, 1, c(1, 2, 5, NA))
  expect_identical(invalid_count(x), 1L)

  # Possible after that too
  x <- year_month_weekday(2019, 1, 1, c(1, 2, 5, NA), 3)
  expect_identical(invalid_count(x), 1L)
})

# ------------------------------------------------------------------------------
# invalid_resolve()

test_that("strict mode can be activated", {
  local_options(clock.strict = TRUE)
  expect_snapshot(error = TRUE, invalid_resolve(year_month_weekday(2019, 1, 1, 1)))
})

test_that("can resolve correctly", {
  x <- year_month_weekday(
    2019,
    1,
    clock_weekdays$friday,
    5,
    2,
    3,
    4,
    5,
    subsecond_precision = "millisecond"
  )

  expect_identical(
    invalid_resolve(x, invalid = "previous"),
    year_month_weekday(
      2019,
      1,
      clock_weekdays$thursday,
      5,
      23,
      59,
      59,
      999,
      subsecond_precision = "millisecond"
    )
  )
  expect_identical(
    invalid_resolve(x, invalid = "previous-day"),
    year_month_weekday(
      2019,
      1,
      clock_weekdays$thursday,
      5,
      2,
      3,
      4,
      5,
      subsecond_precision = "millisecond"
    )
  )
  expect_identical(
    invalid_resolve(x, invalid = "next"),
    year_month_weekday(
      2019,
      2,
      clock_weekdays$friday,
      1,
      0,
      0,
      0,
      0,
      subsecond_precision = "millisecond"
    )
  )
  expect_identical(
    invalid_resolve(x, invalid = "next-day"),
    year_month_weekday(
      2019,
      2,
      clock_weekdays$friday,
      1,
      2,
      3,
      4,
      5,
      subsecond_precision = "millisecond"
    )
  )
  expect_identical(
    invalid_resolve(x, invalid = "overflow"),
    year_month_weekday(
      2019,
      2,
      clock_weekdays$friday,
      1,
      0,
      0,
      0,
      0,
      subsecond_precision = "millisecond"
    )
  )
  expect_identical(
    invalid_resolve(x, invalid = "overflow-day"),
    year_month_weekday(
      2019,
      2,
      clock_weekdays$friday,
      1,
      2,
      3,
      4,
      5,
      subsecond_precision = "millisecond"
    )
  )
  expect_identical(
    invalid_resolve(x, invalid = "NA"),
    year_month_weekday(
      NA,
      NA,
      NA,
      NA,
      NA,
      NA,
      NA,
      NA,
      subsecond_precision = "millisecond"
    )
  )
})

test_that("throws known classed error", {
  expect_snapshot(error = TRUE, invalid_resolve(year_month_weekday(2019, 1, 1, 5)))
  expect_error(
    invalid_resolve(year_month_weekday(2019, 1, 1, 5)),
    class = "clock_error_invalid_date"
  )
})

test_that("works with always valid precisions", {
  x <- year_month_weekday(2019:2020, 1:2)
  expect_identical(invalid_resolve(x), x)
})

# ------------------------------------------------------------------------------
# vec_math()

test_that("is.nan() works", {
  x <- year_month_weekday(c(2019, NA))
  expect_identical(is.nan(x), c(FALSE, FALSE))
})

test_that("is.finite() works", {
  x <- year_month_weekday(c(2019, NA))
  expect_identical(is.finite(x), c(TRUE, FALSE))
})

test_that("is.infinite() works", {
  x <- year_month_weekday(c(2019, NA))
  expect_identical(is.infinite(x), c(FALSE, FALSE))
})

# ------------------------------------------------------------------------------
# diff()

test_that("works with `lag` and `differences`", {
  x <- year_month_weekday(2019, 1:8)
  expect_identical(diff(x), rep(duration_months(1), 7))
  expect_identical(diff(x, lag = 2), rep(duration_months(2), 6))
  expect_identical(diff(x, differences = 2), rep(duration_months(0), 6))
  expect_identical(
    diff(x, lag = 3, differences = 2),
    rep(duration_months(0), 2)
  )
})

test_that("works with `lag` and `differences` that force an empty result (#364)", {
  expect_identical(diff(year_month_weekday(integer())), duration_years())
  expect_identical(diff(year_month_weekday(1)), duration_years())
  expect_identical(
    diff(year_month_weekday(1:8), lag = 4, differences = 3),
    duration_years()
  )
})

test_that("errors on invalid precisions", {
  expect_snapshot(error = TRUE, {
    diff(year_month_weekday(2019, 1, 1, index = 2))
  })
})

test_that("errors on invalid lag/differences", {
  # These are base R errors we don't control
  expect_error(diff(year_month_weekday(2019), lag = 1:2))
  expect_error(diff(year_month_weekday(2019), differences = 1:2))
})

# ------------------------------------------------------------------------------
# clock_minimum() / clock_maximum()

test_that("minimums are right", {
  expect_snapshot({
    clock_minimum(clock_empty_year_month_weekday_year)
    clock_minimum(clock_empty_year_month_weekday_month)
  })

  # Not defined at or past day precision
  expect_snapshot(error = TRUE, {
    clock_minimum(year_month_weekday(1, 1, 1, 1))
  })
})

test_that("maximums are right", {
  expect_snapshot({
    clock_maximum(clock_empty_year_month_weekday_year)
    clock_maximum(clock_empty_year_month_weekday_month)
  })

  # Not defined at or past day precision
  expect_snapshot(error = TRUE, {
    clock_maximum(year_month_weekday(1, 1, 1, 1))
  })
})

# ------------------------------------------------------------------------------
# min() / max() / range()

test_that("min() / max() / range() works", {
  x <- year_month_weekday(c(1, 3, 2, 1, -1))

  expect_identical(min(x), year_month_weekday(-1))
  expect_identical(max(x), year_month_weekday(3))
  expect_identical(range(x), year_month_weekday(c(-1, 3)))
})

test_that("min() / max() / range() works with `NA`", {
  x <- year_month_weekday(c(1, NA, 2, 0))

  expect_identical(min(x), year_month_weekday(NA))
  expect_identical(max(x), year_month_weekday(NA))
  expect_identical(range(x), year_month_weekday(c(NA, NA)))

  expect_identical(min(x, na.rm = TRUE), year_month_weekday(0))
  expect_identical(max(x, na.rm = TRUE), year_month_weekday(2))
  expect_identical(range(x, na.rm = TRUE), year_month_weekday(c(0, 2)))
})

test_that("min() / max() / range() works when empty", {
  x <- year_month_weekday(integer())

  expect_identical(min(x), clock_maximum(x))
  expect_identical(max(x), clock_minimum(x))
  expect_identical(range(x), c(clock_maximum(x), clock_minimum(x)))

  x <- year_month_weekday(c(NA, NA))

  expect_identical(min(x, na.rm = TRUE), clock_maximum(x))
  expect_identical(max(x, na.rm = TRUE), clock_minimum(x))
  expect_identical(range(x, na.rm = TRUE), c(clock_maximum(x), clock_minimum(x)))
})

test_that("min() / max() / range() aren't defined at or past day precision", {
  x <- year_month_weekday(1, 1, 1, 1)

  expect_snapshot(error = TRUE, min(x))
  expect_snapshot(error = TRUE, max(x))
  expect_snapshot(error = TRUE, range(x))

  x <- year_month_weekday(integer(), integer(), integer(), integer())

  expect_snapshot(error = TRUE, min(x))
  expect_snapshot(error = TRUE, max(x))
  expect_snapshot(error = TRUE, range(x))
})

# ------------------------------------------------------------------------------
# add_*()

test_that("add_years() works", {
  x <- year_month_weekday(2019, 1, 2, 3:4, 5)

  expect_identical(
    add_years(x, 1:2),
    year_month_weekday(c(2020, 2021), 1, 2, 3:4, 5)
  )
  expect_identical(
    add_years(x, NA),
    vec_init(x, 2L)
  )
})

test_that("add_months() works", {
  x <- year_month_weekday(2019, 1, 2, 3:4, 5)

  expect_identical(
    add_months(x, 1:2),
    year_month_weekday(2019, 2:3, 2, 3:4, 5)
  )
  expect_identical(
    add_months(x, NA),
    vec_init(x, 2L)
  )
})

test_that("add_quarters() works (special)", {
  x <- year_month_weekday(2019, 1, 2, 3:4, 5)

  expect_identical(
    add_quarters(x, 1:2),
    add_months(x, (1:2) * 3)
  )
  expect_identical(
    add_quarters(x, NA),
    vec_init(x, 2L)
  )
})

test_that("add_*() respect recycling rules", {
  expect_length(add_years(year_month_weekday(1), 1:2), 2L)
  expect_length(add_years(year_month_weekday(1:2), 1), 2L)

  expect_length(add_years(year_month_weekday(1), integer()), 0L)
  expect_length(add_years(year_month_weekday(integer()), 1), 0L)

  expect_snapshot(error = TRUE, {
    add_years(year_month_weekday(1:2), 1:3)
  })
})

test_that("add_*() retains names", {
  x <- set_names(year_month_weekday(1), "x")
  y <- year_month_weekday(1)

  n <- set_names(1, "n")

  expect_named(add_years(x, n), "x")
  expect_named(add_years(y, n), "n")
})

# ------------------------------------------------------------------------------
# year_day()

test_that("helper can create different precisions", {
  x <- year_day(2019, 1:2)
  expect_identical(get_year(x), c(2019L, 2019L))
  expect_identical(get_day(x), 1:2)

  x <- year_day(2019)
  expect_identical(get_year(x), 2019L)
})

test_that("can create subsecond precision calendars", {
  x <- year_day(2019, 1, 0, 0, 0, 1, subsecond_precision = "millisecond")
  expect_identical(get_millisecond(x), 1L)

  x <- year_day(2019, 1, 0, 0, 0, 1, subsecond_precision = "microsecond")
  expect_identical(get_microsecond(x), 1L)

  x <- year_day(2019, 1, 0, 0, 0, 1, subsecond_precision = "nanosecond")
  expect_identical(get_nanosecond(x), 1L)
})

test_that("validates value ranges", {
  expect_snapshot(error = TRUE, year_day(50000))
  expect_snapshot(error = TRUE, year_day(2020, 367))
  expect_snapshot(error = TRUE, year_day(2020, 1, 24))
  expect_snapshot(error = TRUE, year_day(2020, 1, 1, 60))
  expect_snapshot(error = TRUE, year_day(2020, 1, 1, 1, 60))
  expect_snapshot(error = TRUE, {
    year_day(2020, 1, 1, 1, 1, 1000, subsecond_precision = "millisecond")
  })
  expect_snapshot(error = TRUE, {
    year_day(2020, 1, 1, 1, 1, 1000000, subsecond_precision = "microsecond")
  })
  expect_snapshot(error = TRUE, {
    year_day(2020, 1, 1, 1, 1, 1000000000, subsecond_precision = "nanosecond")
  })
})

test_that("can create a date at the boundary", {
  x <- year_day(32767, 365)
  expect_identical(get_year(x), 32767L)

  x <- year_day(-32767, 1)
  expect_identical(get_year(x), -32767L)
})

test_that("can get the last day of the year", {
  x <- year_day(2019:2020, "last")
  expect_identical(get_day(x), c(365L, 366L))
})

test_that("`NA` propagates through 'last'", {
  x <- year_day(2019, c(1, NA))
  x <- set_day(x, "last")
  expect_identical(get_day(x), c(365L, NA))
})

test_that("ignores values past first `NULL`", {
  expect_identical(year_day(2019, hour = 2), year_day(2019))
})

test_that("NA values propagate", {
  x <- year_day(2019, 1:3, c(NA, 2, 3), c(3, 4, NA))
  expect_identical(is.na(x), c(TRUE, FALSE, TRUE))
})

test_that("names of `year` are not retained", {
  expect_named(year_day(c(x = 1)), NULL)
})

# ------------------------------------------------------------------------------
# vec_ptype()

test_that("ptype is correct", {
  base <- year_day(1)
  ptype <- year_day(integer())

  for (precision in precision_names()) {
    if (precision == "quarter" || precision == "month" || precision == "week") {
      next
    }

    x <- calendar_widen(base, precision)
    expect <- calendar_widen(ptype, precision)

    expect_identical(vec_ptype(x), expect)
  }
})

# ------------------------------------------------------------------------------
# vec_proxy() / vec_restore()

test_that("proxy is a data frame", {
  expect_identical(vec_proxy(year_day(2019)), data_frame(year = 2019L))
  expect_identical(vec_proxy(year_day(2019, 1)), data_frame(year = 2019L, day = 1L))
})

test_that("proxy has names on `year`", {
  x <- set_names(year_day(2019, 1), "nm")
  year <- vec_proxy(x)$year
  expect_named(year, "nm")
})

test_that("restore works", {
  to <- year_day(2019, 1:5)
  proxy <- vec_slice(vec_proxy(to), 1:2)
  expect_identical(vec_restore(proxy, to), year_day(2019, 1:2))
})

# ------------------------------------------------------------------------------
# vec_ptype_full()

test_that("full ptype is correct", {
  expect_snapshot_output(vec_ptype_full(year_day(2019)))
  expect_snapshot_output(vec_ptype_full(year_day(2019, 1, 1)))
  expect_snapshot_output(
    vec_ptype_full(year_day(2019, 1, 1, 1, 1, 1, subsecond_precision = "nanosecond"))
  )
  expect_snapshot_output(vec_ptype_full(year_day(2019, 366)))
})

# ------------------------------------------------------------------------------
# vec_ptype_abbr()

test_that("abbreviated ptype is correct", {
  expect_snapshot_output(vec_ptype_abbr(year_day(2019)))
  expect_snapshot_output(vec_ptype_abbr(year_day(2019, 1, 1)))
  expect_snapshot_output(
    vec_ptype_abbr(year_day(2019, 1, 1, 1, 1, 1, subsecond_precision = "nanosecond"))
  )
  expect_snapshot_output(vec_ptype_abbr(year_day(2019, 366)))
})

# ------------------------------------------------------------------------------
# set_*()

test_that("setters work", {
  x <- year_day(1L)

  x <- set_year(x, 2L)
  expect_identical(get_year(x), 2L)

  x <- set_day(x, 2L)
  expect_identical(get_day(x), 2L)

  x <- set_hour(x, 3L)
  expect_identical(get_hour(x), 3L)

  x <- set_minute(x, 4L)
  expect_identical(get_minute(x), 4L)

  x <- set_second(x, 5L)
  expect_identical(get_second(x), 5L)

  ms <- set_millisecond(x, 6L)
  expect_identical(get_millisecond(ms), 6L)

  us <- set_microsecond(x, 7L)
  expect_identical(get_microsecond(us), 7L)

  ns <- set_nanosecond(x, 8L)
  expect_identical(get_nanosecond(ns), 8L)
})

test_that("setters propagate all missings", {
  x <- year_day(2019, c(1, NA, 3))
  x <- set_hour(x, c(NA, 2, 4))
  expect_identical(vec_detect_missing(x), c(TRUE, TRUE, FALSE))
})

test_that("setters recycling works both ways", {
  x <- year_day(2019)

  x <- set_day(x, 1:2)
  expect_identical(x, year_day(2019, 1:2))

  x <- set_hour(x, 1)
  expect_identical(x, year_day(2019, 1:2, 1))

  expect_snapshot(error = TRUE, {
    x <- year_day(1:2)
    y <- 1:3
    set_day(x, y)
  })
})

test_that("setters require integer `value`", {
  x <- year_day(2019, 1, 2, 3, 4)

  expect_snapshot(error = TRUE, {
    set_year(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_day(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_hour(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_minute(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_second(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_millisecond(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_microsecond(x, 1.5)
  })
  expect_snapshot(error = TRUE, {
    set_nanosecond(x, 1.5)
  })
})

test_that("setters check `value` range", {
  x <- year_day(2019, 1, 2, 3, 4)

  expect_snapshot(error = TRUE, {
    set_year(x, 100000)
  })
  expect_snapshot(error = TRUE, {
    set_day(x, 367)
  })
  expect_snapshot(error = TRUE, {
    set_hour(x, 24)
  })
  expect_snapshot(error = TRUE, {
    set_minute(x, 60)
  })
  expect_snapshot(error = TRUE, {
    set_second(x, 60)
  })
  expect_snapshot(error = TRUE, {
    set_millisecond(x, -1)
  })
  expect_snapshot(error = TRUE, {
    set_microsecond(x, -1)
  })
  expect_snapshot(error = TRUE, {
    set_nanosecond(x, -1)
  })
})

test_that("setters require minimum precision", {
  expect_snapshot(error = TRUE, {
    set_hour(year_day(year = 1), 1)
  })
  expect_snapshot(error = TRUE, {
    set_minute(year_day(year = 1, day = 1), 1)
  })
  expect_snapshot(error = TRUE, {
    set_second(year_day(year = 1, day = 1, hour = 1), 1)
  })
  expect_snapshot(error = TRUE, {
    set_millisecond(year_day(year = 1, day = 1, hour = 1, minute = 1), 1)
  })
  expect_snapshot(error = TRUE, {
    set_microsecond(year_day(year = 1, day = 1, hour = 1, minute = 1), 1)
  })
  expect_snapshot(error = TRUE, {
    set_nanosecond(year_day(year = 1, day = 1, hour = 1, minute = 1), 1)
  })
})

test_that("setters require correct subsecond precision", {
  expect_snapshot(error = TRUE, {
    set_millisecond(
      year_day(
        year = 1,
        day = 3,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "microsecond"
      ),
      1
    )
  })
  expect_snapshot(error = TRUE, {
    set_millisecond(
      year_day(
        year = 1,
        day = 3,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "nanosecond"
      ),
      1
    )
  })

  expect_snapshot(error = TRUE, {
    set_microsecond(
      year_day(
        year = 1,
        day = 3,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "millisecond"
      ),
      1
    )
  })
  expect_snapshot(error = TRUE, {
    set_microsecond(
      year_day(
        year = 1,
        day = 3,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "nanosecond"
      ),
      1
    )
  })

  expect_snapshot(error = TRUE, {
    set_nanosecond(
      year_day(
        year = 1,
        day = 3,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "millisecond"
      ),
      1
    )
  })
  expect_snapshot(error = TRUE, {
    set_nanosecond(
      year_day(
        year = 1,
        day = 3,
        hour = 4,
        minute = 5,
        second = 6,
        subsecond = 7,
        subsecond_precision = "microsecond"
      ),
      1
    )
  })
})

test_that("setters retain names", {
  x <- year_day(2019)
  x <- set_names(x, "foo")
  expect_named(set_day(x, 2), "foo")
})

test_that("setting with named `value` strips its names", {
  x <- year_day(2019)
  x <- set_day(x, set_names(1L, "x"))
  expect_named(field(x, "day"), NULL)
})

# ------------------------------------------------------------------------------
# as_year_quarter_day()

test_that("invalid dates must be resolved when converting to another calendar", {
  expect_snapshot(error = TRUE, as_year_quarter_day(year_day(2019, 366)))
})

# ------------------------------------------------------------------------------
# as_sys_time()

test_that("invalid dates must be resolved when converting to a sys-time", {
  expect_snapshot(error = TRUE, as_sys_time(year_day(2019, 366)))
})

# ------------------------------------------------------------------------------
# as_naive_time()

test_that("invalid dates must be resolved when converting to a naive-time", {
  expect_snapshot(error = TRUE, as_naive_time(year_day(2019, 366)))
})

# ------------------------------------------------------------------------------
# format()

test_that("default formats are correct", {
  expect_snapshot(format(year_day(2019)))
  expect_snapshot(format(year_day(2019, 1)))
  expect_snapshot(format(year_day(2019, 1, 1)))
  expect_snapshot(
    format(year_day(2019, 1, 1, 2, 3, 50, subsecond_precision = "microsecond"))
  )
})

# ------------------------------------------------------------------------------
# as.character()

test_that("as.character() works", {
  x <- year_day(2019, 1)
  y <- year_day(2019, 1, 2)

  expect_identical(as.character(x), format(x))
  expect_identical(as.character(y), format(y))
})

# ------------------------------------------------------------------------------
# calendar_group()

test_that("can group by day of the year", {
  x <- year_day(2019, 1, c(3, 4))
  expect_identical(calendar_group(x, "day"), calendar_narrow(x, "day"))
})

test_that("can group by multiple days, resetting at the new year", {
  x <- c(year_day(2019, 364:365), year_day(2020, 1:3))

  expect_identical(
    calendar_group(x, "day", n = 2),
    c(year_day(2019, c(363, 365)), year_day(2020, c(1, 1, 3)))
  )
})

# ------------------------------------------------------------------------------
# calendar_narrow()

test_that("can narrow to year", {
  x_expect <- year_day(2019)
  x <- set_day(x_expect, 1)
  expect_identical(calendar_narrow(x, "year"), x_expect)
  expect_identical(calendar_narrow(x_expect, "year"), x_expect)
})

test_that("can narrow to day", {
  x_expect <- year_day(2019, 2)
  x <- set_hour(x_expect, 5)
  expect_identical(calendar_narrow(x, "day"), x_expect)
  expect_identical(calendar_narrow(x_expect, "day"), x_expect)
})

test_that("can narrow to hour", {
  x_expect <- year_day(2019, 3, 4)
  x <- set_minute(x_expect, 5)
  expect_identical(calendar_narrow(x, "hour"), x_expect)
  expect_identical(calendar_narrow(x_expect, "hour"), x_expect)
})

test_that("can narrow to minute", {
  x_expect <- year_day(2019, 3, 4, 5)
  x <- set_second(x_expect, 6)
  expect_identical(calendar_narrow(x, "minute"), x_expect)
  expect_identical(calendar_narrow(x_expect, "minute"), x_expect)
})

test_that("can narrow to second", {
  expect <- year_day(2019, 3, 4, 5, 6)
  x <- set_millisecond(expect, 7)
  y <- set_nanosecond(expect, 7)
  expect_identical(calendar_narrow(x, "second"), expect)
  expect_identical(calendar_narrow(y, "second"), expect)
  expect_identical(calendar_narrow(expect, "second"), expect)
})

# ------------------------------------------------------------------------------
# calendar_widen()

test_that("can widen to year", {
  x <- year_day(2019)
  expect_identical(calendar_widen(x, "year"), x)
})

test_that("can widen to day", {
  x <- year_day(2019)
  expect_identical(calendar_widen(x, "day"), set_day(x, 1))
})

test_that("can widen to hour", {
  x <- year_day(2019)
  y <- year_day(2019, 02)
  expect_identical(calendar_widen(x, "hour"), set_hour(set_day(x, 1), 0))
  expect_identical(calendar_widen(y, "hour"), set_hour(y, 0))
})

test_that("can widen to minute", {
  x <- year_day(2019)
  y <- year_day(2019, 02, 02)
  x_expect <- year_day(2019, 1, 0, 0)
  y_expect <- set_minute(y, 0)
  expect_identical(calendar_widen(x, "minute"), x_expect)
  expect_identical(calendar_widen(y, "minute"), y_expect)
})

test_that("can widen to second", {
  x <- year_day(2019)
  y <- year_day(2019, 02, 02, 02)
  x_expect <- year_day(2019, 1, 0, 0, 0)
  y_expect <- set_second(y, 0)
  expect_identical(calendar_widen(x, "second"), x_expect)
  expect_identical(calendar_widen(y, "second"), y_expect)
})

test_that("can widen to subsecond precision", {
  x <- year_day(2019)
  y <- year_day(2019, 02, 02, 02, 02)
  x_expect <- year_day(2019, 1, 0, 0, 0, 0, subsecond_precision = "microsecond")
  y_expect <- set_nanosecond(y, 0)
  expect_identical(calendar_widen(x, "microsecond"), x_expect)
  expect_identical(calendar_widen(y, "nanosecond"), y_expect)
})

# ------------------------------------------------------------------------------
# calendar_start()

test_that("can compute year start", {
  x <- year_day(2019)
  expect_identical(calendar_start(x, "year"), x)

  x <- year_day(2019, 2, 2, 2, 2, 2, subsecond_precision = "millisecond")
  expect <- year_day(2019, 1, 0, 0, 0, 0, subsecond_precision = "millisecond")
  expect_identical(calendar_start(x, "year"), expect)
})

# ------------------------------------------------------------------------------
# calendar_end()

test_that("can compute year end", {
  x <- year_day(2019)
  expect_identical(calendar_end(x, "year"), x)

  x <- year_day(2019:2020, 2, 2, 2, 2, 2, subsecond_precision = "millisecond")
  expect <- year_day(
    2019:2020,
    365:366,
    23,
    59,
    59,
    999L,
    subsecond_precision = "millisecond"
  )
  expect_identical(calendar_end(x, "year"), expect)
})

# ------------------------------------------------------------------------------
# calendar_leap_year()

test_that("can detect leap years", {
  x <- year_day(c(2019, 2020, NA), 1)
  expect_identical(calendar_leap_year(x), c(FALSE, TRUE, NA))
})

# ------------------------------------------------------------------------------
# calendar_count_between()

test_that("can compute year counts", {
  x <- year_day(2019, 1)
  y <- year_day(2020, 3)

  expect_identical(calendar_count_between(x, y, "year"), 1L)
})

test_that("can't compute a unsupported difference precision", {
  x <- year_day(2019, 1)
  expect_snapshot((expect_error(calendar_count_between(x, x, "day"))))
})

test_that("positive / negative differences are correct", {
  start <- year_day(1972, 04)

  end <- year_day(1973, 03)
  expect_identical(calendar_count_between(start, end, "year"), 0L)

  end <- year_day(1973, 04)
  expect_identical(calendar_count_between(start, end, "year"), 1L)

  end <- year_day(1973, 05)
  expect_identical(calendar_count_between(start, end, "year"), 1L)

  end <- year_day(1971, 03)
  expect_identical(calendar_count_between(start, end, "year"), -1L)

  end <- year_day(1971, 04)
  expect_identical(calendar_count_between(start, end, "year"), -1L)

  end <- year_day(1971, 05)
  expect_identical(calendar_count_between(start, end, "year"), 0L)
})

# ------------------------------------------------------------------------------
# seq()

test_that("only granular precisions are allowed", {
  expect_snapshot(error = TRUE, seq(year_day(2019, 1), by = 1, length.out = 2))
})

test_that("seq(to, by) works", {
  expect_identical(
    seq(year_day(2019), to = year_day(2024), by = 2),
    year_day(c(2019, 2021, 2023))
  )
  expect_identical(
    seq(year_day(2019), to = year_day(2023), by = 2),
    year_day(c(2019, 2021, 2023))
  )
})

test_that("seq(to, length.out) works", {
  expect_identical(
    seq(year_day(2019), to = year_day(2024), length.out = 2),
    year_day(c(2019, 2024))
  )
  expect_identical(
    seq(year_day(2019), to = year_day(2024), length.out = 6),
    year_day(2019:2024)
  )

  expect_identical(
    seq(year_day(2019), to = year_day(2024), along.with = 1:2),
    year_day(c(2019, 2024))
  )
})

test_that("seq(by, length.out) works", {
  expect_identical(
    seq(year_day(2019), by = 2, length.out = 3),
    year_day(c(2019, 2021, 2023))
  )

  expect_identical(
    seq(year_day(2019), by = 2, along.with = 1:3),
    year_day(c(2019, 2021, 2023))
  )
})

# ------------------------------------------------------------------------------
# miscellaneous

test_that("can roundtrip to naive-time", {
  x <- seq(
    as_naive_time(year_month_day(-9999, 1, 1)),
    as_naive_time(year_month_day(9999, 12, 31)),
    by = 1
  )

  expect_identical(x, as_naive_time(as_year_day(x)))
})

# ------------------------------------------------------------------------------
# invalid_detect()

test_that("`invalid_detect()` works", {
  # Not possible to be invalid
  x <- year_day(2019:2020)
  expect_identical(invalid_detect(x), c(FALSE, FALSE))

  # Now possible
  x <- year_day(2019, c(1, 365, 366, NA))
  expect_identical(invalid_detect(x), c(FALSE, FALSE, TRUE, FALSE))

  # Possible after that too
  x <- year_day(2019, c(1, 365, 366, NA), 3)
  expect_identical(invalid_detect(x), c(FALSE, FALSE, TRUE, FALSE))
})

# ------------------------------------------------------------------------------
# invalid_any()

test_that("`invalid_any()` works", {
  # Not possible to be invalid
  x <- year_day(2019:2020)
  expect_false(invalid_any(x))

  # Now possible
  x <- year_day(2019, c(1, 365, 366, NA))
  expect_true(invalid_any(x))

  # Possible after that too
  x <- year_day(2019, c(1, 365, 366, NA), 3)
  expect_true(invalid_any(x))
})

# ------------------------------------------------------------------------------
# invalid_count()

test_that("`invalid_count()` works", {
  # Not possible to be invalid
  x <- year_day(2019:2020)
  expect_identical(invalid_count(x), 0L)

  # Now possible
  x <- year_day(2019, c(1, 365, 366, NA))
  expect_identical(invalid_count(x), 1L)

  # Possible after that too
  x <- year_day(2019, c(1, 365, 366, NA), 3)
  expect_identical(invalid_count(x), 1L)
})

# ------------------------------------------------------------------------------
# invalid_resolve()

test_that("strict mode can be activated", {
  local_options(clock.strict = TRUE)
  expect_snapshot(error = TRUE, invalid_resolve(year_day(2019, 1)))
})

test_that("can resolve correctly", {
  x <- year_day(2019, 366, 2, 3, 4, 5, subsecond_precision = "millisecond")

  expect_identical(
    invalid_resolve(x, invalid = "previous"),
    year_day(2019, 365, 23, 59, 59, 999, subsecond_precision = "millisecond")
  )
  expect_identical(
    invalid_resolve(x, invalid = "previous-day"),
    year_day(2019, 365, 2, 3, 4, 5, subsecond_precision = "millisecond")
  )
  expect_identical(
    invalid_resolve(x, invalid = "next"),
    year_day(2020, 1, 0, 0, 0, 0, subsecond_precision = "millisecond")
  )
  expect_identical(
    invalid_resolve(x, invalid = "next-day"),
    year_day(2020, 1, 2, 3, 4, 5, subsecond_precision = "millisecond")
  )
  expect_identical(
    invalid_resolve(x, invalid = "overflow"),
    year_day(2020, 1, 0, 0, 0, 0, subsecond_precision = "millisecond")
  )
  expect_identical(
    invalid_resolve(x, invalid = "overflow-day"),
    year_day(2020, 1, 2, 3, 4, 5, subsecond_precision = "millisecond")
  )
  expect_identical(
    invalid_resolve(x, invalid = "NA"),
    year_day(NA, NA, NA, NA, NA, NA, subsecond_precision = "millisecond")
  )
})

test_that("throws known classed error", {
  expect_snapshot(error = TRUE, invalid_resolve(year_day(2019, 366)))
  expect_error(invalid_resolve(year_day(2019, 366)), class = "clock_error_invalid_date")
})

test_that("works with always valid precisions", {
  x <- year_day(2019:2020)
  expect_identical(invalid_resolve(x), x)
})

# ------------------------------------------------------------------------------
# vec_math()

test_that("is.nan() works", {
  x <- year_day(c(2019, NA))
  expect_identical(is.nan(x), c(FALSE, FALSE))
})

test_that("is.finite() works", {
  x <- year_day(c(2019, NA))
  expect_identical(is.finite(x), c(TRUE, FALSE))
})

test_that("is.infinite() works", {
  x <- year_day(c(2019, NA))
  expect_identical(is.infinite(x), c(FALSE, FALSE))
})

# ------------------------------------------------------------------------------
# diff()

test_that("works with `lag` and `differences`", {
  x <- year_day(2019:2026)
  expect_identical(diff(x), rep(duration_years(1), 7))
  expect_identical(diff(x, lag = 2), rep(duration_years(2), 6))
  expect_identical(diff(x, differences = 2), rep(duration_years(0), 6))
  expect_identical(
    diff(x, lag = 3, differences = 2),
    rep(duration_years(0), 2)
  )
})

test_that("works with `lag` and `differences` that force an empty result (#364)", {
  expect_identical(diff(year_day(integer())), duration_years())
  expect_identical(diff(year_day(1)), duration_years())
  expect_identical(
    diff(year_day(1:8), lag = 4, differences = 3),
    duration_years()
  )
})

test_that("errors on invalid precisions", {
  expect_snapshot(error = TRUE, {
    diff(year_day(2019, 1))
  })
})

test_that("errors on invalid lag/differences", {
  # These are base R errors we don't control
  expect_error(diff(year_day(2019), lag = 1:2))
  expect_error(diff(year_day(2019), differences = 1:2))
})

# ------------------------------------------------------------------------------
# clock_minimum() / clock_maximum()

test_that("minimums are right", {
  expect_snapshot({
    clock_minimum(clock_empty_year_day_year)
    clock_minimum(clock_empty_year_day_day)
    clock_minimum(clock_empty_year_day_hour)
    clock_minimum(clock_empty_year_day_minute)
    clock_minimum(clock_empty_year_day_second)
    clock_minimum(clock_empty_year_day_millisecond)
    clock_minimum(clock_empty_year_day_microsecond)
    clock_minimum(clock_empty_year_day_nanosecond)
  })
})

test_that("maximums are right", {
  expect_snapshot({
    clock_maximum(clock_empty_year_day_year)
    clock_maximum(clock_empty_year_day_day)
    clock_maximum(clock_empty_year_day_hour)
    clock_maximum(clock_empty_year_day_minute)
    clock_maximum(clock_empty_year_day_second)
    clock_maximum(clock_empty_year_day_millisecond)
    clock_maximum(clock_empty_year_day_microsecond)
    clock_maximum(clock_empty_year_day_nanosecond)
  })
})

# ------------------------------------------------------------------------------
# min() / max() / range()

test_that("min() / max() / range() works", {
  x <- year_day(c(1, 3, 2, 1, -1))

  expect_identical(min(x), year_day(-1))
  expect_identical(max(x), year_day(3))
  expect_identical(range(x), year_day(c(-1, 3)))
})

test_that("min() / max() / range() works with `NA`", {
  x <- year_day(c(1, NA, 2, 0))

  expect_identical(min(x), year_day(NA))
  expect_identical(max(x), year_day(NA))
  expect_identical(range(x), year_day(c(NA, NA)))

  expect_identical(min(x, na.rm = TRUE), year_day(0))
  expect_identical(max(x, na.rm = TRUE), year_day(2))
  expect_identical(range(x, na.rm = TRUE), year_day(c(0, 2)))
})

test_that("min() / max() / range() works when empty", {
  x <- year_day(integer())

  expect_identical(min(x), clock_maximum(x))
  expect_identical(max(x), clock_minimum(x))
  expect_identical(range(x), c(clock_maximum(x), clock_minimum(x)))

  x <- year_day(c(NA, NA))

  expect_identical(min(x, na.rm = TRUE), clock_maximum(x))
  expect_identical(max(x, na.rm = TRUE), clock_minimum(x))
  expect_identical(range(x, na.rm = TRUE), c(clock_maximum(x), clock_minimum(x)))
})

# ------------------------------------------------------------------------------
# add_*()

test_that("add_years() works", {
  x <- year_day(2019, 1, 2, 3:4)

  expect_identical(
    add_years(x, 1:2),
    year_day(c(2020, 2021), 1, 2, 3:4)
  )
  expect_identical(
    add_years(x, NA),
    vec_init(x, 2L)
  )
})

test_that("add_*() respect recycling rules", {
  expect_length(add_years(year_day(1), 1:2), 2L)
  expect_length(add_years(year_day(1:2), 1), 2L)

  expect_length(add_years(year_day(1), integer()), 0L)
  expect_length(add_years(year_day(integer()), 1), 0L)

  expect_snapshot(error = TRUE, {
    add_years(year_day(1:2), 1:3)
  })
})

test_that("add_*() retains names", {
  x <- set_names(year_day(1), "x")
  y <- year_day(1)

  n <- set_names(1, "n")

  expect_named(add_years(x, n), "x")
  expect_named(add_years(y, n), "n")
})

#' Construct a weekday vector
#'
#' @description
#' A `weekday` is a simple type that represents a day of the week.
#'
#' The most interesting thing about the weekday type is that it implements
#' _circular arithmetic_, which makes determining the "next Monday" or
#' "previous Tuesday" from a sys-time or naive-time easy to compute.
#' See the examples.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param code `[integer]`
#'
#'   Integer codes between `[1, 7]` representing days of the week. The
#'   interpretation of these values depends on `encoding`.
#'
#' @param encoding `[character(1)]`
#'
#'   One of:
#'
#'   - `"western"`: Encode weekdays assuming `1 == Sunday` and `7 == Saturday`.
#'
#'   - `"iso"`: Encode weekdays assuming `1 == Monday` and `7 == Sunday`. This
#'   is in line with the ISO standard.
#'
#' @return A weekday vector.
#'
#' @export
#' @examples
#' x <- as_naive_time(year_month_day(2019, 01, 05))
#'
#' # This is a Saturday!
#' as_weekday(x)
#'
#' # Adjust to the next Wednesday
#' wednesday <- weekday(clock_weekdays$wednesday)
#'
#' # This returns the number of days until the next Wednesday using
#' # circular arithmetic
#' # "Wednesday - Saturday = 4 days until next Wednesday"
#' wednesday - as_weekday(x)
#'
#' # Advance to the next Wednesday
#' x_next_wednesday <- x + (wednesday - as_weekday(x))
#'
#' as_weekday(x_next_wednesday)
#'
#' # What about the previous Tuesday?
#' tuesday <- weekday(clock_weekdays$tuesday)
#' x - (as_weekday(x) - tuesday)
#'
#' # What about the next Saturday?
#' # With an additional condition that if today is a Saturday,
#' # then advance to the next one.
#' saturday <- weekday(clock_weekdays$saturday)
#' x + 1L + (saturday - as_weekday(x + 1L))
#'
#' # You can supply an ISO coding for `code` as well, where 1 == Monday.
#' weekday(1:7, encoding = "western")
#' weekday(1:7, encoding = "iso")
weekday <- function(code = integer(), ..., encoding = "western") {
  check_dots_empty()

  # No other helpers retain names, so we don't here either
  code <- unname(code)
  code <- vec_cast(code, integer(), x_arg = "code")

  oob <- (code > 7L | code < 1L) & (!is.na(code))
  if (any(oob)) {
    message <- paste0("`code` must be in range of [1, 7].")
    abort(message)
  }

  encoding <- validate_encoding(encoding)

  # Store as western encoding
  if (is_iso_encoding(encoding)) {
    code <- reencode_iso_to_western(code)
  }

  new_weekday(code)
}

new_weekday <- function(code = integer(), ..., class = NULL) {
  if (!is_integer(code)) {
    abort("`code` must be an integer vector.")
  }

  new_vctr(
    code,
    ...,
    class = c(class, "clock_weekday"),
    inherit_base_type = FALSE
  )
}

# ------------------------------------------------------------------------------

#' Is `x` a weekday?
#'
#' This function determines if the input is a weekday object.
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @return `TRUE` if `x` inherits from `"clock_weekday"`, otherwise `FALSE`.
#'
#' @export
#' @examples
#' is_weekday(1)
#' is_weekday(weekday(1))
is_weekday <- function(x) {
  inherits(x, "clock_weekday")
}

# ------------------------------------------------------------------------------

#' @export
format.clock_weekday <- function(x, ..., labels = "en", abbreviate = TRUE) {
  if (is_character(labels)) {
    labels <- clock_labels_lookup(labels)
  }
  if (!is_clock_labels(labels)) {
    abort("`labels` must be a 'clock_labels' object.")
  }

  if (!is_bool(abbreviate)) {
    abort("`abbreviate` must be `TRUE` or `FALSE`.")
  }

  if (abbreviate) {
    labels <- labels$weekday_abbrev
  } else {
    labels <- labels$weekday
  }

  out <- format_weekday_cpp(x, labels)

  names(out) <- names(x)

  out
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype_full.clock_weekday <- function(x, ...) {
  "weekday"
}

#' @export
vec_ptype_abbr.clock_weekday <- function(x, ...) {
  "weekday"
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype.clock_weekday <- function(x, ...) {
  clock_empty_weekday
}

#' @export
vec_ptype2.clock_weekday.clock_weekday <- function(x, y, ...) {
  x
}

#' @export
vec_cast.clock_weekday.clock_weekday <- function(x, to, ...) {
  x
}

# ------------------------------------------------------------------------------

#' @export
vec_proxy_compare.clock_weekday <- function(x, ...) {
  abort(paste0(
    "Can't compare or order values of the 'clock_weekday' type, ",
    "as this type does not specify a 'first' day of the week."
  ))
}

# ------------------------------------------------------------------------------

#' Convert to a weekday
#'
#' `as_weekday()` converts to a weekday type. This is normally useful for
#' converting to a weekday from a sys-time or naive-time. You can use this
#' function along with the _circular arithmetic_ that weekday implements to
#' easily get to the "next Monday" or "previous Sunday".
#'
#' @param x `[object]`
#'
#'   An object to convert to a weekday. Usually a sys-time or naive-time.
#'
#' @return A weekday.
#'
#' @export
#' @examples
#' x <- as_naive_time(year_month_day(2019, 01, 05))
#'
#' # This is a Saturday!
#' as_weekday(x)
#'
#' # See the examples in `?weekday` for more usage.
as_weekday <- function(x) {
  UseMethod("as_weekday")
}

#' @export
as_weekday.clock_weekday <- function(x) {
  x
}

#' @export
as_weekday.clock_calendar <- function(x) {
  abort("Can't extract the weekday from a calendar. Convert to a time point first.")
}

# ------------------------------------------------------------------------------

#' @export
as.character.clock_weekday <- function(x, ...) {
  format(x)
}

# ------------------------------------------------------------------------------

#' Extract underlying weekday codes
#'
#' `weekday_code()` extracts out the integer code for the weekday.
#'
#' @inheritParams weekday
#'
#' @param x `[weekday]`
#'
#'   A weekday vector.
#'
#' @return An integer vector of codes.
#'
#' @export
#' @examples
#' # Here we supply a western encoding to `weekday()`
#' x <- weekday(1:7)
#' x
#'
#' # We can extract out the codes using different encodings
#' weekday_code(x, encoding = "western")
#' weekday_code(x, encoding = "iso")
weekday_code <- function(x, ..., encoding = "western") {
  check_dots_empty()

  if (!is_weekday(x)) {
    abort("`x` must be a 'clock_weekday'.")
  }

  encoding <- validate_encoding(encoding)

  x <- unstructure(x)

  if (is_iso_encoding(encoding)) {
    x <- reencode_western_to_iso(x)
  }

  x
}

# ------------------------------------------------------------------------------

#' Convert a weekday to an ordered factor
#'
#' `weekday_factor()` converts a weekday object to an ordered factor. This
#' can be useful in combination with ggplot2, or for modeling.
#'
#' @inheritParams weekday_code
#' @inheritParams clock_locale
#'
#' @param abbreviate `[logical(1)]`
#'
#'   If `TRUE`, the abbreviated weekday names from `labels` will be used.
#'
#'   If `FALSE`, the full weekday names from `labels` will be used.
#'
#' @param encoding `[character(1)]`
#'
#'   One of:
#'
#'   - `"western"`: Encode the weekdays as an ordered factor with levels from
#'     Sunday -> Saturday.
#'
#'   - `"iso"`: Encode the weekdays as an ordered factor with levels from
#'     Monday -> Sunday.
#'
#' @return An ordered factor representing the weekdays.
#'
#' @export
#' @examples
#' x <- weekday(1:7)
#'
#' # Default to Sunday -> Saturday
#' weekday_factor(x)
#'
#' # ISO encoding is Monday -> Sunday
#' weekday_factor(x, encoding = "iso")
#'
#' # With full names
#' weekday_factor(x, abbreviate = FALSE)
#'
#' # Or a different language
#' weekday_factor(x, labels = "fr")
weekday_factor <- function(x,
                           ...,
                           labels = "en",
                           abbreviate = TRUE,
                           encoding = "western") {
  check_dots_empty()

  if (!is_weekday(x)) {
    abort("`x` must be a 'clock_weekday' object.")
  }

  if (is_character(labels)) {
    labels <- clock_labels_lookup(labels)
  }
  if (!is_clock_labels(labels)) {
    abort("`labels` must be a 'clock_labels' object.")
  }

  if (!is_bool(abbreviate)) {
    abort("`abbreviate` must be `TRUE` or `FALSE`.")
  }

  if (abbreviate) {
    labels <- labels$weekday_abbrev
  } else {
    labels <- labels$weekday
  }

  encoding <- validate_encoding(encoding)
  x <- weekday_code(x, encoding = encoding)

  if (is_iso_encoding(encoding)) {
    labels <- c(labels[2:7], labels[1L])
  }

  x <- labels[x]

  factor(x, levels = labels, ordered = TRUE)
}

# ------------------------------------------------------------------------------

#' @rdname clock-arith
#' @method vec_arith clock_weekday
#' @export
vec_arith.clock_weekday <- function(op, x, y, ...) {
  UseMethod("vec_arith.clock_weekday", y)
}

#' @method vec_arith.clock_weekday MISSING
#' @export
vec_arith.clock_weekday.MISSING <- function(op, x, y, ...) {
  switch(
    op,
    "+" = x,
    stop_incompatible_op(op, x, y, ...)
  )
}

#' @method vec_arith.clock_weekday clock_weekday
#' @export
vec_arith.clock_weekday.clock_weekday <- function(op, x, y, ...) {
  switch(
    op,
    "-" = weekday_minus_weekday(op, x, y, ...),
    stop_incompatible_op(op, x, y, ...)
  )
}

#' @method vec_arith.clock_weekday clock_duration
#' @export
vec_arith.clock_weekday.clock_duration <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(x, y),
    "-" = add_duration(x, -y),
    stop_incompatible_op(op, x, y, ...)
  )
}

#' @method vec_arith.clock_duration clock_weekday
#' @export
vec_arith.clock_duration.clock_weekday <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(y, x, swapped = TRUE),
    "-" = stop_incompatible_op(op, x, y, details = "Can't subtract a weekday from a duration.", ...),
    stop_incompatible_op(op, x, y, ...)
  )
}

#' @method vec_arith.clock_weekday numeric
#' @export
vec_arith.clock_weekday.numeric <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(x, duration_helper(y, PRECISION_DAY, retain_names = TRUE)),
    "-" = add_duration(x, duration_helper(-y, PRECISION_DAY, retain_names = TRUE)),
    stop_incompatible_op(op, x, y, ...)
  )
}

#' @method vec_arith.numeric clock_weekday
#' @export
vec_arith.numeric.clock_weekday <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(y, duration_helper(x, PRECISION_DAY, retain_names = TRUE), swapped = TRUE),
    "-" = stop_incompatible_op(op, x, y, details = "Can't subtract a weekday from a duration.", ...),
    stop_incompatible_op(op, x, y, ...)
  )
}

# Subtraction between weekdays tells days to next weekday
weekday_minus_weekday <- function(op, x, y, ...) {
  args <- vec_recycle_common(x = x, y = y)
  x <- args$x
  y <- args$y

  names <- names_common(x, y)

  fields <- weekday_minus_weekday_cpp(x, y)

  new_duration_from_fields(fields, PRECISION_DAY, names)
}

# ------------------------------------------------------------------------------

#' Arithmetic: weekday
#'
#' @description
#' These are weekday methods for the
#' [arithmetic generics][clock-arithmetic].
#'
#' - `add_days()`
#'
#' Also check out the examples on the [weekday()] page for more advanced
#' usage.
#'
#' @details
#' `x` and `n` are recycled against each other.
#'
#' @inheritParams add_years
#'
#' @param x `[clock_weekday]`
#'
#'   A weekday vector.
#'
#' @return `x` after performing the arithmetic.
#'
#' @name weekday-arithmetic
#'
#' @examples
#' saturday <- weekday(clock_weekdays$saturday)
#' saturday
#'
#' add_days(saturday, 1)
#' add_days(saturday, 2)
NULL

#' @rdname weekday-arithmetic
#' @export
add_days.clock_weekday <- function(x, n, ...) {
  n <- duration_collect_n(n, PRECISION_DAY)

  args <- vec_recycle_common(x = x, n = n)
  x <- args$x
  n <- args$n

  names <- names_common(x, n)

  code <- weekday_add_days_cpp(x, n)
  names(code) <- names

  new_weekday(code)
}

# ------------------------------------------------------------------------------

#' @export
vec_math.clock_weekday <- function(.fn, .x, ...) {
  switch(
    .fn,
    is.nan = weekday_is_nan(.x),
    is.finite = weekday_is_finite(.x),
    is.infinite = weekday_is_infinite(.x),
    NextMethod()
  )
}

weekday_is_nan <- function(x) {
  vec_rep(FALSE, vec_size(x))
}

weekday_is_finite <- function(x) {
  !vec_equal_na(x)
}

weekday_is_infinite <- function(x) {
  vec_rep(FALSE, vec_size(x))
}

# ------------------------------------------------------------------------------

validate_encoding <- function(encoding) {
  if (!is_string(encoding, string = c("western", "iso"))) {
    abort("`encoding` must be one of \"western\" or \"iso\".")
  }
  encoding
}

is_iso_encoding <- function(encoding) {
  identical(encoding, "iso")
}

reencode_iso_to_western <- function(code) {
  code <- code + 1L
  code[code == 8L] <- 1L
  code
}

reencode_western_to_iso <- function(code) {
  code <- code - 1L
  code[code == 0L] <- 7L
  code
}

# ------------------------------------------------------------------------------

clock_init_weekday_utils <- function(env) {
  assign("clock_empty_weekday", weekday(integer()), envir = env)

  invisible(NULL)
}

# ------------------------------------------------------------------------------
# format()

test_that("format - can format() a weekday", {
  expect_snapshot(format(weekday(1:7)))
})

test_that("format - can use full names", {
  expect_snapshot(format(weekday(1:7), abbreviate = FALSE))
})

test_that("format - can use a different locale", {
  expect_snapshot(format(weekday(1:7), labels = "fr", abbreviate = FALSE))
})

test_that("format - `labels` is validated", {
  expect_snapshot_error(format(weekday(1), labels = 1))
})

test_that("format - `abbreviate` is validated", {
  expect_snapshot_error(format(weekday(1), abbreviate = "foo"))
  expect_snapshot_error(format(weekday(1), abbreviate = 1))
  expect_snapshot_error(format(weekday(1), abbreviate = c(TRUE, FALSE)))
})

# ------------------------------------------------------------------------------
# as.character()

test_that("as.character() works", {
  x <- weekday(1:7)
  expect_identical(as.character(x), format(x))
})

test_that("as.character() works with NA", {
  expect_identical(as.character(weekday(NA)), NA_character_)
})

# ------------------------------------------------------------------------------
# weekday_code()

test_that("can get the western code", {
  expect_identical(weekday_code(weekday(1:7)), 1:7)
})

test_that("can get the ISO code", {
  expect_identical(weekday_code(weekday(1:7), encoding = "iso"), c(7L, 1:6))
})

test_that("names are not retained", {
  expect_named(weekday_code(c(foo = weekday(1))), NULL)
})

test_that("NA passes through", {
  expect_identical(weekday_code(weekday(NA)), NA_integer_)
})

test_that("validates `x`", {
  expect_snapshot_error(weekday_code(1))
})

test_that("weekday_code - `encoding` is validated", {
  expect_snapshot_error(weekday_code(weekday(1), encoding = "foo"))
  expect_snapshot_error(weekday_code(weekday(1), encoding = 1))
  expect_snapshot_error(weekday_code(weekday(1), encoding = c("western", "iso")))
})

# ------------------------------------------------------------------------------
# weekday_factor()

test_that("can make a weekday factor sunday->saturday", {
  order <- 7:1
  x <- weekday(order)
  x <- weekday_factor(x)

  levels <- clock_labels_lookup("en")$weekday_abbrev
  data <- levels[order]

  expect_s3_class(x, "ordered")
  expect_identical(as.character(x), data)
  expect_identical(levels(x), levels)
})

test_that("can make a weekday factor monday->sunday", {
  order <- 7:1
  x <- weekday(order)
  x <- weekday_factor(x, encoding = "iso")

  levels <- clock_labels_lookup("en")$weekday_abbrev
  data <- levels[order]
  levels <- levels[c(2:7, 1L)]

  expect_s3_class(x, "ordered")
  expect_identical(as.character(x), data)
  expect_identical(levels(x), levels)
})

test_that("can make a weekday factor with full names", {
  x <- weekday(7:1)
  x <- weekday_factor(x, abbreviate = FALSE)

  levels <- clock_labels_lookup("en")$weekday

  expect_identical(levels(x), levels)
})

test_that("can make a weekday factor with alternate labels", {
  x <- weekday(7:1)
  x <- weekday_factor(x, labels = "fr")

  levels <- clock_labels_lookup("fr")$weekday_abbrev

  expect_identical(levels(x), levels)
})

test_that("`x` is validated", {
  expect_snapshot_error(weekday_factor(1))
})

test_that("`labels` is validated", {
  expect_snapshot_error(weekday_factor(weekday(1), labels = 1))
})

test_that("`encoding` is validated", {
  expect_snapshot_error(weekday_factor(weekday(1), encoding = "foo"))
  expect_snapshot_error(weekday_factor(weekday(1), encoding = 1))
  expect_snapshot_error(weekday_factor(weekday(1), encoding = c("western", "iso")))
})

test_that("`abbreviate` is validated", {
  expect_snapshot_error(weekday_factor(weekday(1), abbreviate = "foo"))
  expect_snapshot_error(weekday_factor(weekday(1), abbreviate = 1))
  expect_snapshot_error(weekday_factor(weekday(1), abbreviate = c(TRUE, FALSE)))
})

# ------------------------------------------------------------------------------
# add_days()

test_that("can add days with circular arithmetic", {
  expect_identical(add_days(weekday(6:7), 1), weekday(c(7,1)))
})

test_that("can add days with NA", {
  expect_identical(
    add_days(weekday(c(1, 2, NA)), c(NA, 1, 2)),
    weekday(c(NA, 3, NA))
  )
})

# ------------------------------------------------------------------------------
# vec_arith()

test_that("can subtract two weekdays with circular arithmetic", {
  expect_identical(
    weekday(3) - weekday(c(1, 7)),
    duration_days(c(2, 3))
  )
})

test_that("subtraction respects NA", {
  expect_identical(
    weekday(c(1, NA)) - weekday(c(NA, 2)),
    duration_days(c(NA, NA))
  )
})

# ------------------------------------------------------------------------------
# vec_proxy_compare()

test_that("can't compare or order weekdays (#153)", {
  expect_snapshot_error(weekday(1) < weekday(2))
  expect_snapshot_error(min(weekday(1)))

  expect_snapshot_error(xtfrm(weekday(1:2)))
  expect_snapshot_error(vec_order(weekday(1:2)))
})

# ------------------------------------------------------------------------------
# vec_ptype()

test_that("ptype is correct", {
  expect_identical(vec_ptype(weekday(1:7)), weekday(integer()))
})

# ------------------------------------------------------------------------------
# vec_math()

test_that("is.nan() works", {
  x <- weekday(c(1, NA))
  expect_identical(is.nan(x), c(FALSE, FALSE))
})

test_that("is.finite() works", {
  x <- weekday(c(1, NA))
  expect_identical(is.finite(x), c(TRUE, FALSE))
})

test_that("is.infinite() works", {
  x <- weekday(c(1, NA))
  expect_identical(is.infinite(x), c(FALSE, FALSE))
})

#' Calendar getters
#'
#' @description
#' This family of functions extract fields from a calendar vector. Each
#' calendar has its own set of supported getters, which are documented on
#' their own help page:
#'
#' - [year-month-day][year-month-day-getters]
#'
#' - [year-month-weekday][year-month-weekday-getters]
#'
#' - [iso-year-week-day][iso-year-week-day-getters]
#'
#' - [year-quarter-day][year-quarter-day-getters]
#'
#' - [year-day][year-day-getters]
#'
#' There are also convenience methods for extracting certain components
#' directly from R's native date and date-time types.
#'
#' - [dates (Date)][Date-getters]
#'
#' - [date-times (POSIXct / POSIXlt)][posixt-getters]
#'
#' @details
#' You cannot extract components directly from a time point type, such as
#' sys-time or naive-time. Convert it to a calendar type first. Similarly,
#' a zoned-time must be converted to either a sys-time or naive-time, and
#' then to a calendar type, to be able to extract components from it.
#'
#' @param x `[object]`
#'
#'   An object to get the component from.
#'
#' @return The component.
#'
#' @name clock-getters
#' @examples
#' x <- year_month_day(2019, 1:3, 5:7, 1, 20, 30)
#' get_month(x)
#' get_day(x)
#' get_second(x)
NULL

#' @rdname clock-getters
#' @export
get_year <- function(x) {
  UseMethod("get_year")
}

#' @export
get_year.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_year")
}

#' @rdname clock-getters
#' @export
get_quarter <- function(x) {
  UseMethod("get_quarter")
}

#' @export
get_quarter.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_quarter")
}

#' @rdname clock-getters
#' @export
get_month <- function(x) {
  UseMethod("get_month")
}

#' @export
get_month.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_month")
}

#' @rdname clock-getters
#' @export
get_week <- function(x) {
  UseMethod("get_week")
}

#' @export
get_week.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_week")
}

#' @rdname clock-getters
#' @export
get_day <- function(x) {
  UseMethod("get_day")
}

#' @export
get_day.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_day")
}

#' @rdname clock-getters
#' @export
get_hour <- function(x) {
  UseMethod("get_hour")
}

#' @export
get_hour.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_hour")
}

#' @rdname clock-getters
#' @export
get_minute <- function(x) {
  UseMethod("get_minute")
}

#' @export
get_minute.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_minute")
}

#' @rdname clock-getters
#' @export
get_second <- function(x) {
  UseMethod("get_second")
}

#' @export
get_second.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_second")
}

#' @rdname clock-getters
#' @export
get_millisecond <- function(x) {
  UseMethod("get_millisecond")
}

#' @export
get_millisecond.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_millisecond")
}

#' @rdname clock-getters
#' @export
get_microsecond <- function(x) {
  UseMethod("get_microsecond")
}

#' @export
get_microsecond.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_microsecond")
}

#' @rdname clock-getters
#' @export
get_nanosecond <- function(x) {
  UseMethod("get_nanosecond")
}

#' @export
get_nanosecond.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_nanosecond")
}

#' @rdname clock-getters
#' @export
get_index <- function(x) {
  UseMethod("get_index")
}

#' @export
get_index.clock_calendar <- function(x) {
  stop_clock_unsupported_calendar_op("get_index")
}

# ------------------------------------------------------------------------------
# year_quarter_day()

test_that("helper can create different precisions", {
  x <- year_quarter_day(2019, 1:2)
  expect_identical(get_year(x), c(2019L, 2019L))
  expect_identical(get_quarter(x), 1:2)

  x <- year_quarter_day(2019, 1:2, 3)
  expect_identical(get_day(x), c(3L, 3L))
})

test_that("can create subsecond precision calendars", {
  x <- year_quarter_day(2019, 1, 1, 0, 0, 0, 1, subsecond_precision = "millisecond")
  expect_identical(get_millisecond(x), 1L)

  x <- year_quarter_day(2019, 1, 1, 0, 0, 0, 1, subsecond_precision = "microsecond")
  expect_identical(get_microsecond(x), 1L)

  x <- year_quarter_day(2019, 1, 1, 0, 0, 0, 1, subsecond_precision = "nanosecond")
  expect_identical(get_nanosecond(x), 1L)
})

test_that("validates value ranges", {
  expect_snapshot_error(year_quarter_day(50000))
  expect_snapshot_error(year_quarter_day(2020, 5))
  expect_snapshot_error(year_quarter_day(2020, 1, 93))
  expect_snapshot_error(year_quarter_day(2020, 1, 1, 24))
  expect_snapshot_error(year_quarter_day(2020, 1, 1, 1, 60))
  expect_snapshot_error(year_quarter_day(2020, 1, 1, 1, 1, 60))
  expect_snapshot_error(year_quarter_day(2020, 1, 1, 1, 1, 1, 1000, subsecond_precision = "millisecond"))
  expect_snapshot_error(year_quarter_day(2020, 1, 1, 1, 1, 1, 1000000, subsecond_precision = "microsecond"))
  expect_snapshot_error(year_quarter_day(2020, 1, 1, 1, 1, 1, 1000000000, subsecond_precision = "nanosecond"))
})

test_that("can get the last day of the quarter", {
  x <- year_quarter_day(2019, 1:2, "last")
  expect_identical(get_day(x), c(90L, 91L))
})

test_that("ignores values past first `NULL`", {
  expect_identical(year_quarter_day(2019, day = 2), year_quarter_day(2019))
})

test_that("NA values propagate", {
  x <- year_quarter_day(2019, 1:3, c(NA, 2, 3), c(3, 4, NA))
  expect_identical(is.na(x), c(TRUE, FALSE, TRUE))
})

# ------------------------------------------------------------------------------
# vec_ptype()

test_that("ptype is correct", {
  base <- year_quarter_day(1)
  ptype <- year_quarter_day(integer())

  for (precision in precision_names()) {
    if (precision == "month" || precision == "week") {
      next
    }

    x <- calendar_widen(base, precision)
    expect <- calendar_widen(ptype, precision)

    expect_identical(vec_ptype(x), expect)
  }
})

# ------------------------------------------------------------------------------
# vec_proxy() / vec_restore()

test_that("proxy is a data frame", {
  expect_identical(vec_proxy(year_quarter_day(2019)), data_frame(year = 2019L))
  expect_identical(vec_proxy(year_quarter_day(2019, 1)), data_frame(year = 2019L, quarter = 1L))
})

test_that("proxy has names on `year`", {
  x <- set_names(year_quarter_day(2019, 1), "nm")
  year <- vec_proxy(x)$year
  expect_named(year, "nm")
})

test_that("restore works", {
  to <- year_quarter_day(2019, 1:4)
  proxy <- vec_slice(vec_proxy(to), 1:2)
  expect_identical(vec_restore(proxy, to), year_quarter_day(2019, 1:2))
})

# ------------------------------------------------------------------------------
# vec_ptype_full()

test_that("full ptype is correct", {
  expect_snapshot_output(vec_ptype_full(year_quarter_day(2019)))
  expect_snapshot_output(vec_ptype_full(year_quarter_day(2019, start = 2)))
  expect_snapshot_output(vec_ptype_full(year_quarter_day(2019, 1, 1)))
  expect_snapshot_output(vec_ptype_full(year_quarter_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "nanosecond")))
  expect_snapshot_output(vec_ptype_full(year_quarter_day(2019, 1, 92)))
})

# ------------------------------------------------------------------------------
# vec_ptype_abbr()

test_that("abbreviated ptype is correct", {
  expect_snapshot_output(vec_ptype_abbr(year_quarter_day(2019)))
  expect_snapshot_output(vec_ptype_abbr(year_quarter_day(2019, start = 2)))
  expect_snapshot_output(vec_ptype_abbr(year_quarter_day(2019, 1, 1)))
  expect_snapshot_output(vec_ptype_abbr(year_quarter_day(2019, 1, 1, 1, 1, 1, 1, subsecond_precision = "nanosecond")))
  expect_snapshot_output(vec_ptype_abbr(year_quarter_day(2019, 1, 92)))
})

# ------------------------------------------------------------------------------
# format()

test_that("default formats are correct", {
  expect_snapshot(format(year_quarter_day(2019)))
  expect_snapshot(format(year_quarter_day(2019, 1)))
  expect_snapshot(format(year_quarter_day(2019, 1, 1, 1)))
  expect_snapshot(format(year_quarter_day(2019, 1, 1, 1, 2, 3, 50, subsecond_precision = "microsecond")))
})

# ------------------------------------------------------------------------------
# as.character()

test_that("as.character() works", {
  x <- year_quarter_day(2019, 1)
  y <- year_quarter_day(2019, 1, 2)

  expect_identical(as.character(x), format(x))
  expect_identical(as.character(y), format(y))
})

# ------------------------------------------------------------------------------
# calendar_narrow()

test_that("can narrow to week", {
  x_expect <- year_quarter_day(2019, 2)
  x <- set_day(x_expect, 1)
  expect_identical(calendar_narrow(x, "quarter"), x_expect)
  expect_identical(calendar_narrow(x_expect, "quarter"), x_expect)
})

test_that("can narrow to day", {
  x_expect <- year_quarter_day(2019, 2, 3)
  x <- set_hour(x_expect, 5)
  expect_identical(calendar_narrow(x, "day"), x_expect)
  expect_identical(calendar_narrow(x_expect, "day"), x_expect)
})

# ------------------------------------------------------------------------------
# calendar_widen()

test_that("can widen to quarter", {
  x <- year_quarter_day(2019)
  expect_identical(calendar_widen(x, "quarter"), set_quarter(x, 1))
})

test_that("can widen to day", {
  x <- year_quarter_day(2019)
  y <- year_quarter_day(2019, 02)
  expect_identical(calendar_widen(x, "day"), set_day(set_quarter(x, 1), 1))
  expect_identical(calendar_widen(y, "day"), set_day(y, 1))
})

# ------------------------------------------------------------------------------
# calendar_start()

test_that("can compute year start", {
  x <- year_quarter_day(2019)
  expect_identical(calendar_start(x, "year"), x)

  x <- year_quarter_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "millisecond")
  expect <- year_quarter_day(2019, 1, 1, 0, 0, 0, 0, subsecond_precision = "millisecond")
  expect_identical(calendar_start(x, "year"), expect)
})

test_that("can compute quarter start", {
  x <- year_quarter_day(2019, 2)
  expect_identical(calendar_start(x, "quarter"), x)

  x <- year_quarter_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "millisecond")
  expect <- year_quarter_day(2019, 2, 1, 0, 0, 0, 0, subsecond_precision = "millisecond")
  expect_identical(calendar_start(x, "quarter"), expect)
})

# ------------------------------------------------------------------------------
# calendar_end()

test_that("can compute year end", {
  x <- year_quarter_day(2019)
  expect_identical(calendar_end(x, "year"), x)

  x <- year_quarter_day(2019, 2, 2, 2, 2, 2, 2, subsecond_precision = "millisecond")
  expect <- year_quarter_day(2019, 4, 92, 23, 59, 59, 999L, subsecond_precision = "millisecond")
  expect_identical(calendar_end(x, "year"), expect)
})

test_that("can compute quarter end", {
  x <- year_quarter_day(2019, 2)
  expect_identical(calendar_end(x, "quarter"), x)

  x <- year_quarter_day(2019, 2:3, 2, 2, 2, 2, 2, subsecond_precision = "millisecond")
  expect <- year_quarter_day(2019, 2:3, 91:92, 23, 59, 59, 999L, subsecond_precision = "millisecond")
  expect_identical(calendar_end(x, "quarter"), expect)
})

# ------------------------------------------------------------------------------
# calendar_count_between()

test_that("can compute year and month counts", {
  x <- year_quarter_day(2019, 1, 1)
  y <- year_quarter_day(2020, 3, 4)

  expect_identical(calendar_count_between(x, y, "year"), 1L)
  expect_identical(calendar_count_between(x, y, "quarter"), 6L)
  expect_identical(calendar_count_between(x, y, "quarter", n = 2), 3L)
})

test_that("works with different quarter start months", {
  x <- year_quarter_day(2019, 1, 1, start = clock_months$march)
  y <- year_quarter_day(2020, 3, 4, start = clock_months$march)

  expect_identical(calendar_count_between(x, y, "year"), 1L)
  expect_identical(calendar_count_between(x, y, "quarter"), 6L)
  expect_identical(calendar_count_between(x, y, "quarter", n = 2), 3L)
})

test_that("can't compute a unsupported count precision", {
  x <- year_quarter_day(2019, 1, 1)
  expect_snapshot((expect_error(calendar_count_between(x, x, "day"))))
})

test_that("positive / negative counts are correct", {
  start <- year_quarter_day(1972, 03, 04)


  end <- year_quarter_day(1973, 03, 03)
  expect_identical(calendar_count_between(start, end, "year"), 0L)
  expect_identical(calendar_count_between(start, end, "quarter"), 3L)

  end <- year_quarter_day(1973, 03, 04)
  expect_identical(calendar_count_between(start, end, "year"), 1L)
  expect_identical(calendar_count_between(start, end, "quarter"), 4L)

  end <- year_quarter_day(1973, 03, 05)
  expect_identical(calendar_count_between(start, end, "year"), 1L)
  expect_identical(calendar_count_between(start, end, "quarter"), 4L)


  end <- year_quarter_day(1971, 03, 03)
  expect_identical(calendar_count_between(start, end, "year"), -1L)
  expect_identical(calendar_count_between(start, end, "quarter"), -4L)

  end <- year_quarter_day(1971, 03, 04)
  expect_identical(calendar_count_between(start, end, "year"), -1L)
  expect_identical(calendar_count_between(start, end, "quarter"), -4L)

  end <- year_quarter_day(1971, 03, 05)
  expect_identical(calendar_count_between(start, end, "year"), 0L)
  expect_identical(calendar_count_between(start, end, "quarter"), -3L)
})

# ------------------------------------------------------------------------------
# seq()

test_that("only granular precisions are allowed", {
  expect_snapshot_error(seq(year_quarter_day(2019, 1, 1), by = 1, length.out = 2))
})

test_that("seq(to, by) works", {
  expect_identical(seq(year_quarter_day(2019, 1), to = year_quarter_day(2020, 2), by = 2), year_quarter_day(c(2019, 2019, 2020), c(1, 3, 1)))
  expect_identical(seq(year_quarter_day(2019, 1), to = year_quarter_day(2020, 1), by = 2), year_quarter_day(c(2019, 2019, 2020), c(1, 3, 1)))
})

test_that("seq(to, length.out) works", {
  expect_identical(seq(year_quarter_day(2019, 1), to = year_quarter_day(2020, 2), length.out = 2), year_quarter_day(c(2019, 2020), c(1, 2)))
  expect_identical(seq(year_quarter_day(2019, 1), to = year_quarter_day(2020, 2), length.out = 6), year_quarter_day(2019, 1) + 0:5)

  expect_identical(seq(year_quarter_day(2019, 1), to = year_quarter_day(2020, 2), along.with = 1:2), year_quarter_day(c(2019, 2020), c(1, 2)))
})

test_that("seq(by, length.out) works", {
  expect_identical(seq(year_quarter_day(2019, 1), by = 2, length.out = 3), year_quarter_day(2019, 1) + c(0, 2, 4))
  expect_identical(seq(year_quarter_day(2019, 1), by = -2, length.out = 3), year_quarter_day(2019, 1) + c(0, -2, -4))

  expect_identical(seq(year_quarter_day(2019, 1), by = 2, along.with = 1:3), year_quarter_day(2019, 1) + c(0, 2, 4))
})

# ------------------------------------------------------------------------------
# add_*()

test_that("can add quarters", {
  expect_identical(year_quarter_day(2019, 1, 1) + duration_quarters(1), year_quarter_day(2019, 2, 1))
  expect_identical(year_quarter_day(2019, 1, 1) + duration_quarters(5), year_quarter_day(2020, 2, 1))

  expect_identical(year_quarter_day(2019, 1, 1, start = 2) + duration_quarters(5), year_quarter_day(2020, 2, 1, start = 2))
})

# ------------------------------------------------------------------------------
# miscellaneous

test_that("can roundtrip to naive-time with any `start`", {
  x <- seq(
    as_naive_time(year_month_day(-9999, 1, 1)),
    as_naive_time(year_month_day(9999, 12, 31)),
    by = 1
  )

  for (start in seq_len(12)) {
    expect_identical(x, as_naive_time(as_year_quarter_day(x, start = start)))
  }
})

test_that("can generate correct last days of the quarter with any `start`", {
  start <- 1L

  expect_identical(
    as.Date(year_quarter_day(2019, 1:4, "last", start = 1)),
    as.Date(year_month_day(2019, c(3, 6, 9, 12), "last"))
  )
  expect_identical(
    as.Date(year_quarter_day(2019, 1:4, "last", start = 2)),
    as.Date(year_month_day(c(2018, 2018, 2018, 2019), c(4, 7, 10, 1), "last"))
  )
  expect_identical(
    as.Date(year_quarter_day(2019, 1:4, "last", start = 3)),
    as.Date(year_month_day(c(2018, 2018, 2018, 2019), c(5, 8, 11, 2), "last"))
  )
  # ..., 4, 5, 6, 7, 8, 9, ...
  expect_identical(
    as.Date(year_quarter_day(2019, 1:4, "last", start = 10)),
    as.Date(year_month_day(c(2018, 2019, 2019, 2019), c(12, 3, 6, 9), "last"))
  )
  expect_identical(
    as.Date(year_quarter_day(2019, 1:4, "last", start = 11)),
    as.Date(year_month_day(2019, c(1, 4, 7, 10), "last"))
  )
  expect_identical(
    as.Date(year_quarter_day(2019, 1:4, "last", start = 12)),
    as.Date(year_month_day(2019, c(2, 5, 8, 11), "last"))
  )

  # Leap year!
  expect_identical(
    as.Date(year_quarter_day(2020, 1:4, "last", start = 1)),
    as.Date(year_month_day(2020, c(3, 6, 9, 12), "last"))
  )
  expect_identical(
    as.Date(year_quarter_day(2020, 1:4, "last", start = 2)),
    as.Date(year_month_day(c(2019, 2019, 2019, 2020), c(4, 7, 10, 1), "last"))
  )
  expect_identical(
    as.Date(year_quarter_day(2020, 1:4, "last", start = 3)),
    as.Date(year_month_day(c(2019, 2019, 2019, 2020), c(5, 8, 11, 2), "last"))
  )
})

# ------------------------------------------------------------------------------
# invalid_resolve()

test_that("strict mode can be activated", {
  local_options(clock.strict = TRUE)
  expect_snapshot_error(invalid_resolve(year_quarter_day(2019, 1, 1)))
})

# ------------------------------------------------------------------------------
# vec_math()

test_that("is.nan() works", {
  x <- year_quarter_day(c(2019, NA))
  expect_identical(is.nan(x), c(FALSE, FALSE))
})

test_that("is.finite() works", {
  x <- year_quarter_day(c(2019, NA))
  expect_identical(is.finite(x), c(TRUE, FALSE))
})

test_that("is.infinite() works", {
  x <- year_quarter_day(c(2019, NA))
  expect_identical(is.infinite(x), c(FALSE, FALSE))
})

check_between_year <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(
    x,
    lower = clock_calendar_year_minimum,
    upper = clock_calendar_year_maximum,
    call = call,
    arg = arg
  )
}
check_between_quarter <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 1L, upper = 4L, call = call, arg = arg)
}
check_between_month <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 1L, upper = 12L, call = call, arg = arg)
}
check_between_week <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 1L, upper = 53L, call = call, arg = arg)
}
check_between_day_of_year <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 1L, upper = 366L, call = call, arg = arg)
}
check_between_day_of_quarter <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 1L, upper = 92L, call = call, arg = arg)
}
check_between_day_of_month <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 1L, upper = 31L, call = call, arg = arg)
}
check_between_day_of_week <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 1L, upper = 7L, call = call, arg = arg)
}
check_between_index_of_week <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 1L, upper = 5L, call = call, arg = arg)
}
check_between_hour <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 0L, upper = 23L, call = call, arg = arg)
}
check_between_minute <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 0L, upper = 59L, call = call, arg = arg)
}
check_between_second <- function(
  x,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  check_between(x, lower = 0L, upper = 59L, call = call, arg = arg)
}
check_between_subsecond <- function(
  x,
  precision,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  upper <- switch(
    precision_to_string(precision),
    millisecond = 999L,
    microsecond = 999999L,
    nanosecond = 999999999L,
    abort("Invalid precision.", .internal = TRUE)
  )
  check_between(x, lower = 0L, upper = upper, call = call, arg = arg)
}

check_between <- function(
  x,
  lower,
  upper,
  ...,
  call = caller_env(),
  arg = caller_arg(x)
) {
  ok <- (x >= lower & x <= upper) | vec_detect_missing(x)

  if (all(ok)) {
    return(invisible(NULL))
  }

  loc <- which(!ok)

  message <- c(
    "{.arg {arg}} must be between [{lower}, {upper}].",
    i = "Invalid results at locations: {loc}."
  )

  cli::cli_abort(message, call = call)
}

to_posixct <- function(x) {
  if (is_POSIXct(x)) {
    posixct_standardize(x)
  } else if (is_POSIXlt(x)) {
    to_posixct_from_posixlt(x)
  } else {
    abort("Should either be POSIXct/POSIXlt.", .internal = TRUE)
  }
}

is_POSIXt <- function(x) {
  inherits(x, "POSIXt")
}
is_POSIXct <- function(x) {
  inherits(x, "POSIXct")
}
is_POSIXlt <- function(x) {
  inherits(x, "POSIXlt")
}

check_posixt <- function(x, ..., arg = caller_arg(x), call = caller_env()) {
  check_inherits(x, what = "POSIXt", arg = arg, call = call)
}

posixct_standardize <- function(x) {
  if (identical(typeof(x), "double")) {
    return(x)
  }

  # Convert rare integer POSIXct to double
  # Preserves names
  storage.mode(x) <- "double"

  x
}
to_posixct_from_posixlt <- function(x) {
  as.POSIXct.POSIXlt(x)
}

date_standardize <- function(x) {
  if (identical(typeof(x), "double")) {
    return(x)
  }

  # Convert somewhat rare integer Date to double.
  # Preserves names.
  storage.mode(x) <- "double"

  x
}

# ------------------------------------------------------------------------------

ones_along <- function(x, na_propagate = FALSE) {
  out <- rep(1L, vec_size(x))

  if (!na_propagate) {
    return(out)
  }

  na <- vec_detect_missing(x)
  if (any(na)) {
    out[na] <- NA_integer_
  }

  out
}
zeros_along <- function(x, na_propagate = FALSE) {
  out <- vector("integer", length = vec_size(x))

  if (!na_propagate) {
    return(out)
  }

  na <- vec_detect_missing(x)
  if (any(na)) {
    out[na] <- NA_integer_
  }

  out
}

# ------------------------------------------------------------------------------

unstructure <- function(x) {
  set_attributes(x, NULL)
}

vec_unstructure <- function(x) {
  # Must unclass first because `names()` might not be the same length before
  # and after unclassing
  x <- unclass(x)
  out <- unstructure(x)
  names(out) <- names(x)
  out
}

set_attributes <- function(x, attributes) {
  attributes(x) <- attributes
  x
}

# ------------------------------------------------------------------------------

posixt_tzone <- function(x) {
  tzone <- attr(x, "tzone", exact = TRUE)
  posixt_tzone_standardize(tzone)
}
posixt_set_tzone <- function(x, tzone) {
  attr(x, "tzone") <- tzone
  x
}

# Standardize an R time zone attribute
#
# This is slightly different from the lubridate version. For POSIXlt objects,
# the time zone attribute might be a character vector of length 3. If the first
# element is `""` (which happens on a Mac with `as.POSIXlt(Sys.time())`), then
# lubridate will look to the second element and will use that as the time zone.
# I think this is incorrect, because those are always time zone abbreviations,
# and will fail to load because they aren't true time zone names. I think that
# is the reason Vitalie opened the issue noted below, and the reason for the
# time zone map in lubridate. This function works more like
# `lubridate:::tz.POSIXt()` which just takes the first element of the tzone
# attribute. https://github.com/google/cctz/issues/46
# https://github.com/tidyverse/lubridate/blob
posixt_tzone_standardize <- function(tzone) {
  if (is_null(tzone)) {
    # Like `Sys.time()`
    return("")
  }

  if (!is_character(tzone)) {
    abort("A POSIXt time zone should either be a character vector or `NULL`.")
  }

  n <- length(tzone)

  if (n == 0L) {
    warning(
      paste0(
        "POSIXt input had a corrupt time zone attribute of `character(0)`. ",
        "Defaulting to the current zone by assuming the zone is `\"\"`."
      )
    )
    return("")
  }

  if (n == 1L) {
    return(tzone)
  }

  # Otherwise `n > 1`, likely `n == 3` for a POSIXt with time zone
  # abbreviations. The first element is either a full time zone name, or `""`,
  # so we use that
  tzone <- tzone[[1]]

  tzone
}

# ------------------------------------------------------------------------------

stop_clock <- function(message, ..., call = caller_env(), class = character()) {
  rlang::abort(message, ..., call = call, class = c(class, "clock_error"))
}

stop_clock_unsupported <- function(
  x,
  ...,
  details = NULL,
  call = caller_env()
) {
  class <- class(x)[[1L]]
  message <- cli::format_inline(
    "Can't perform this operation on a {.cls {class}}."
  )
  message <- c(message, details)
  stop_clock(message, ..., call = call, class = "clock_error_unsupported")
}

stop_clock_unsupported_conversion <- function(
  x,
  to_arg,
  ...,
  call = caller_env()
) {
  class <- class(x)[[1L]]
  message <- cli::format_inline(
    "Can't convert {.cls {class}} to {.cls {to_arg}}."
  )
  stop_clock(
    message,
    ...,
    call = call,
    class = "clock_error_unsupported_conversion"
  )
}

# Thrown from C++
stop_clock_invalid_date <- function(i, call) {
  message <- c(
    cli::format_inline("Invalid date found at location {i}."),
    i = cli::format_inline(
      "Resolve invalid date issues by specifying the {.arg invalid} argument."
    )
  )
  stop_clock(message, call = call, class = "clock_error_invalid_date")
}

# Thrown from C++
stop_clock_nonexistent_time <- function(i, call) {
  message <- c(
    cli::format_inline(
      "Nonexistent time due to daylight saving time at location {i}."
    ),
    i = cli::format_inline(
      "Resolve nonexistent time issues by specifying the {.arg nonexistent} argument."
    )
  )
  stop_clock(message, call = call, class = "clock_error_nonexistent_time")
}

# Thrown from C++
stop_clock_ambiguous_time <- function(i, call) {
  message <- c(
    cli::format_inline(
      "Ambiguous time due to daylight saving time at location {i}."
    ),
    i = cli::format_inline(
      "Resolve ambiguous time issues by specifying the {.arg ambiguous} argument."
    )
  )
  stop_clock(message, call = call, class = "clock_error_ambiguous_time")
}

# ------------------------------------------------------------------------------

warn_clock <- function(message, class = character()) {
  rlang::warn(message, class = c(class, "clock_warning"))
}

# Thrown from C++
warn_clock_parse_failures <- function(n, first) {
  if (n == 0) {
    abort("Internal error: warning thrown with zero failures.")
  } else if (n == 1) {
    message <- paste0(
      "Failed to parse 1 string at location ",
      first,
      ". ",
      "Returning `NA` at that location."
    )
  } else {
    message <- paste0(
      "Failed to parse ",
      n,
      " strings, beginning at location ",
      first,
      ". ",
      "Returning `NA` at the locations where there were parse failures."
    )
  }

  warn_clock(message, "clock_warning_parse_failures")
}

# Thrown from C++
warn_clock_format_failures <- function(n, first) {
  if (n == 0) {
    abort("Internal error: warning thrown with zero failures.")
  } else if (n == 1) {
    message <- paste0(
      "Failed to format 1 string at location ",
      first,
      ". ",
      "Returning `NA` at that location."
    )
  } else {
    message <- paste0(
      "Failed to format ",
      n,
      " strings, beginning at location ",
      first,
      ". ",
      "Returning `NA` at the locations where there were format failures."
    )
  }

  warn_clock(message, "clock_warning_format_failures")
}

# ------------------------------------------------------------------------------

max_collect <- function(max, ..., error_call = caller_env()) {
  if (is_null(max)) {
    max <- getOption("max.print", default = 1000L)
  }

  check_number_whole(max, min = 0, call = error_call)
  max <- vec_cast(max, integer(), call = error_call)

  max
}

max_slice <- function(x, max) {
  if (max < vec_size(x)) {
    vec_slice(x, seq_len(max))
  } else {
    x
  }
}

clock_print <- function(x, max, ..., error_call = caller_env()) {
  max <- max_collect(max, error_call = error_call)
  obj_print(x, max = max)
  invisible(x)
}

clock_print_footer <- function(x, max) {
  size <- vec_size(x)

  if (max >= size) {
    return(invisible(x))
  }

  n_omitted <- size - max

  if (n_omitted == 1L) {
    value <- " value."
  } else {
    value <- " values."
  }

  cat(
    "Reached `max` or `getOption('max.print')`. ",
    "Omitted ",
    n_omitted,
    value,
    "\n",
    sep = ""
  )

  invisible(x)
}

# ------------------------------------------------------------------------------

df_list_propagate_missing <- function(x, ..., size = NULL) {
  check_dots_empty0(...)

  x <- new_data_frame(x, n = size)

  complete <- vec_detect_complete(x)
  if (all(complete)) {
    return(vec_unstructure(x))
  }

  incomplete <- !complete
  missing <- vec_detect_missing(x)

  aligned <- missing == incomplete
  if (all(aligned)) {
    # Already fully missing where incomplete
    return(vec_unstructure(x))
  }

  # Propagate missings
  x <- vec_assign(x, incomplete, NA)

  vec_unstructure(x)
}

# ------------------------------------------------------------------------------

check_inherits <- function(
  x,
  what,
  ...,
  allow_null = FALSE,
  arg = caller_arg(x),
  call = caller_env()
) {
  if (!missing(x)) {
    if (inherits(x, what)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x = x,
    what = cli::format_inline("a <{what}>"),
    arg = arg,
    call = call
  )
}

check_no_missing <- function(x, ..., arg = caller_arg(x), call = caller_env()) {
  if (!vec_any_missing(x)) {
    return(invisible(NULL))
  }

  loc <- vec_detect_missing(x)
  loc <- which(loc)

  message <- c(
    "{.arg {arg}} can't contain missing values.",
    i = "The following locations are missing: {loc}."
  )

  cli::cli_abort(message, call = call)
}

# ------------------------------------------------------------------------------

vec_drop_infinite <- function(x) {
  infinite <- is.infinite(x)

  if (any(infinite)) {
    x <- vec_slice(x, !infinite)
  }

  x
}

# ------------------------------------------------------------------------------

is_last <- function(x) {
  identical(x, "last")
}

test_that("POSIXt time zones are standardized as expected", {
  expect_identical(posixt_tzone_standardize(NULL), "")
  expect_identical(posixt_tzone_standardize(""), "")
  expect_identical(posixt_tzone_standardize("America/New_York"), "America/New_York")

  # POSIXlt can return something like this
  expect_identical(posixt_tzone_standardize(c("", "EST", "EDT")), "")

  # Unknown if this can happen, but we handle it anyways
  expect_snapshot(posixt_tzone_standardize(character()))
  expect_warning(
    expect_identical(posixt_tzone_standardize(character()), "")
  )

  expect_snapshot_error(posixt_tzone_standardize(1))
})

#' @export
as_sys_time.POSIXt <- function(x) {
  # The sys-time that would give the equivalent zoned-time when a tz is attached
  as_sys_time(as_zoned_time(x))
}

#' @export
as_naive_time.POSIXt <- function(x) {
  as_naive_time(as_zoned_time(x))
}

#' Convert to a zoned-time from a date-time
#'
#' @description
#' This is a POSIXct/POSIXlt method for the [as_zoned_time()] generic.
#'
#' Converting from one of R's native date-time classes (POSIXct or POSIXlt)
#' will retain the time zone of that object. There is no `zone` argument.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time.
#'
#' @return A zoned-time.
#'
#' @name as-zoned-time-posixt
#' @export
#' @examples
#' x <- as.POSIXct("2019-01-01", tz = "America/New_York")
#' as_zoned_time(x)
as_zoned_time.POSIXt <- function(x, ...) {
  check_dots_empty()

  x <- to_posixct(x)

  names <- names(x)
  zone <- posixt_tzone(x)

  fields <- to_sys_duration_fields_from_sys_seconds_cpp(x)

  new_zoned_time_from_fields(fields, PRECISION_SECOND, zone, names)
}

#' @export
as_year_month_day.POSIXt <- function(x) {
  # Assumes zoned -> naive -> calendar is what the user expects
  x <- as_naive_time(x)
  as_year_month_day(x)
}

#' @export
as_year_month_weekday.POSIXt <- function(x) {
  # Assumes zoned -> naive -> calendar is what the user expects
  x <- as_naive_time(x)
  as_year_month_weekday(x)
}

#' @export
as_year_quarter_day.POSIXt <- function(x, ..., start = NULL) {
  # Assumes zoned -> naive -> calendar is what the user expects
  x <- as_naive_time(x)
  as_year_quarter_day(x, ..., start = start)
}

#' @export
as_iso_year_week_day.POSIXt <- function(x) {
  # Assumes zoned -> naive -> calendar is what the user expects
  x <- as_naive_time(x)
  as_iso_year_week_day(x)
}

#' @export
as_year_day.POSIXt <- function(x) {
  # Assumes zoned -> naive -> calendar is what the user expects
  x <- as_naive_time(x)
  as_year_day(x)
}

#' @export
as_weekday.POSIXt <- function(x) {
  # Assumes zoned -> naive is what the user expects
  x <- as_naive_time(x)
  as_weekday(x)
}

# ------------------------------------------------------------------------------

# Not using `check_dots_empty()` because that might
# be too aggressive with base generics. Also not passing `...` on to methods
# that do check empty dots.

# Using `tz = ""` to be compatible with the generic of `as.POSIXct()`

#' @export
as.POSIXct.clock_calendar <- function(x,
                                      tz = "",
                                      ...,
                                      nonexistent = NULL,
                                      ambiguous = NULL) {
  x <- as_naive_time(x)
  as.POSIXct(x, tz = tz, nonexistent = nonexistent, ambiguous = ambiguous)
}

#' @export
as.POSIXct.clock_sys_time <- function(x, tz = "", ...) {
  x <- as_zoned_time(x, zone = tz)
  as.POSIXct(x)
}

#' @export
as.POSIXct.clock_naive_time <- function(x,
                                        tz = "",
                                        ...,
                                        nonexistent = NULL,
                                        ambiguous = NULL) {
  x <- as_zoned_time(x, zone = tz, nonexistent = nonexistent, ambiguous = ambiguous)
  as.POSIXct(x)
}

#' @export
as.POSIXct.clock_zoned_time <- function(x, ...) {
  zone <- zoned_time_zone_attribute(x)

  x <- as_sys_time(x)
  x <- time_point_floor(x, "second")

  seconds <- to_sys_seconds_from_sys_duration_fields_cpp(x)

  names(seconds) <- clock_rcrd_names(x)

  new_datetime(seconds, zone)
}

# ------------------------------------------------------------------------------

#' @export
as.POSIXlt.clock_calendar <- function(x,
                                      tz = "",
                                      ...,
                                      nonexistent = NULL,
                                      ambiguous = NULL) {
  x <- as.POSIXct(x, tz = tz, ..., nonexistent = nonexistent, ambiguous = ambiguous)
  as.POSIXlt(x)
}

#' @export
as.POSIXlt.clock_sys_time <- function(x, tz = "", ...) {
  x <- as.POSIXct(x, tz = tz, ...)
  as.POSIXlt(x)
}

#' @export
as.POSIXlt.clock_naive_time <- as.POSIXlt.clock_calendar

#' @export
as.POSIXlt.clock_zoned_time <- function(x, ...) {
  x <- as.POSIXct(x, ...)
  as.POSIXlt(x)
}

# ------------------------------------------------------------------------------

#' Convert to a date-time
#'
#' @description
#' `as_date_time()` is a generic function that converts its input to a date-time
#' (POSIXct).
#'
#' There are methods for converting dates (Date), calendars, time points, and
#' zoned-times to date-times.
#'
#' For converting to a date, see [as_date()].
#'
#' @details
#' Note that clock always assumes that R's Date class is naive, so converting
#' a Date to a POSIXct will always attempt to retain the printed year, month,
#' and day. Where possible, the resulting time will be at midnight (`00:00:00`),
#' but in some rare cases this is not possible due to daylight saving time. If
#' that issue ever arises, an error will be thrown, which can be resolved by
#' explicitly supplying `nonexistent` or `ambiguous`.
#'
#' This is not a drop-in replacement for `as.POSIXct()`, as it only converts a
#' limited set of types to POSIXct. For parsing characters as date-times, see
#' [date_time_parse()]. For converting numerics to date-times, see
#' [vctrs::new_datetime()] or continue to use `as.POSIXct()`.
#'
#' @inheritParams as-zoned-time-naive-time
#'
#' @param x `[vector]`
#'
#'   A vector.
#'
#' @return A date-time with the same length as `x`.
#'
#' @export
#' @examples
#' x <- as.Date("2019-01-01")
#'
#' # `as.POSIXct()` will always treat Date as UTC, but will show the result
#' # of the conversion in your system time zone, which can be somewhat confusing
#' if (rlang::is_installed("withr")) {
#'   withr::with_timezone("UTC", print(as.POSIXct(x)))
#'   withr::with_timezone("Europe/Paris", print(as.POSIXct(x)))
#'   withr::with_timezone("America/New_York", print(as.POSIXct(x)))
#' }
#'
#' # `as_date_time()` will treat Date as naive, which means that the original
#' # printed date will attempt to be kept wherever possible, no matter the
#' # time zone. The time will be set to midnight.
#' as_date_time(x, "UTC")
#' as_date_time(x, "Europe/Paris")
#' as_date_time(x, "America/New_York")
#'
#' # In some rare cases, this is not possible.
#' # For example, in Asia/Beirut, there was a DST gap from
#' # 2021-03-27 23:59:59 -> 2021-03-28 01:00:00,
#' # skipping the 0th hour entirely.
#' x <- as.Date("2021-03-28")
#' try(as_date_time(x, "Asia/Beirut"))
#'
#' # To resolve this, set a `nonexistent` time resolution strategy
#' as_date_time(x, "Asia/Beirut", nonexistent = "roll-forward")
#'
#'
#' # You can also convert to date-time from other clock types
#' as_date_time(year_month_day(2019, 2, 3, 03), "America/New_York")
as_date_time <- function(x, ...) {
  UseMethod("as_date_time")
}

#' @rdname as_date_time
#' @export
as_date_time.POSIXt <- function(x, ...) {
  check_dots_empty()
  to_posixct(x)
}

#' @rdname as_date_time
#' @export
as_date_time.Date <- function(x, zone, ..., nonexistent = NULL, ambiguous = NULL) {
  check_dots_empty()
  as.POSIXct(as_naive_time(x), tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

#' @rdname as_date_time
#' @export
as_date_time.clock_calendar <- function(x, zone, ..., nonexistent = NULL, ambiguous = NULL) {
  check_dots_empty()
  as.POSIXct(x, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

#' @rdname as_date_time
#' @export
as_date_time.clock_sys_time <- function(x, zone, ...) {
  check_dots_empty()
  as.POSIXct(x, tz = zone)
}

#' @rdname as_date_time
#' @export
as_date_time.clock_naive_time <- function(x, zone, ..., nonexistent = NULL, ambiguous = NULL) {
  check_dots_empty()
  as.POSIXct(x, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

#' @rdname as_date_time
#' @export
as_date_time.clock_zoned_time <- function(x, ...) {
  check_dots_empty()
  as.POSIXct(x)
}

# ------------------------------------------------------------------------------

#' Getters: date-time
#'
#' @description
#' These are POSIXct/POSIXlt methods for the [getter generics][clock-getters].
#'
#' - `get_year()` returns the Gregorian year.
#'
#' - `get_month()` returns the month of the year.
#'
#' - `get_day()` returns the day of the month.
#'
#' - There are sub-daily getters for extracting more precise components, up to
#'   a precision of seconds.
#'
#' For more advanced component extraction, convert to the calendar type
#' that you are interested in.
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time type to get the component from.
#'
#' @return The component.
#'
#' @name posixt-getters
#' @examples
#' x <- as.POSIXct("2019-01-01", tz = "America/New_York")
#'
#' x <- add_days(x, 0:5)
#' x <- set_second(x, 10:15)
#'
#' get_day(x)
#' get_second(x)
NULL

#' @rdname posixt-getters
#' @export
get_year.POSIXt <- function(x) {
  get_posixt_field_year_month_day(x, get_year)
}
#' @rdname posixt-getters
#' @export
get_month.POSIXt <- function(x) {
  get_posixt_field_year_month_day(x, get_month)
}
#' @rdname posixt-getters
#' @export
get_day.POSIXt <- function(x) {
  get_posixt_field_year_month_day(x, get_day)
}
#' @rdname posixt-getters
#' @export
get_hour.POSIXt <- function(x) {
  get_posixt_field_year_month_day(x, get_hour)
}
#' @rdname posixt-getters
#' @export
get_minute.POSIXt <- function(x) {
  get_posixt_field_year_month_day(x, get_minute)
}
#' @rdname posixt-getters
#' @export
get_second.POSIXt <- function(x) {
  get_posixt_field_year_month_day(x, get_second)
}
get_posixt_field_year_month_day <- function(x, get_fn) {
  x <- as_year_month_day(x)
  get_fn(x)
}

# ------------------------------------------------------------------------------

#' Setters: date-time
#'
#' @description
#' These are POSIXct/POSIXlt methods for the [setter generics][clock-setters].
#'
#' - `set_year()` sets the year.
#'
#' - `set_month()` sets the month of the year. Valid values are in the range
#'   of `[1, 12]`.
#'
#' - `set_day()` sets the day of the month. Valid values are in the range
#'   of `[1, 31]`.
#'
#' - There are sub-daily setters for setting more precise components, up to
#'   a precision of seconds.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams invalid_resolve
#' @inheritParams as-zoned-time-naive-time
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time vector.
#'
#' @param value `[integer / "last"]`
#'
#'   The value to set the component to.
#'
#'   For `set_day()`, this can also be `"last"` to set the day to the
#'   last day of the month.
#'
#' @return `x` with the component set.
#'
#' @name posixt-setters
#' @examples
#' x <- as.POSIXct("2019-02-01", tz = "America/New_York")
#'
#' # Set the day
#' set_day(x, 12:14)
#'
#' # Set to the "last" day of the month
#' set_day(x, "last")
#'
#' # You cannot set a date-time to an invalid date like you can with
#' # a year-month-day. Instead, the default strategy is to error.
#' try(set_day(x, 31))
#' set_day(as_year_month_day(x), 31)
#'
#' # You can resolve these issues while setting the day by specifying
#' # an invalid date resolution strategy with `invalid`
#' set_day(x, 31, invalid = "previous")
#'
#' y <- as.POSIXct("2020-03-08 01:30:00", tz = "America/New_York")
#'
#' # Nonexistent and ambiguous times must be resolved immediately when
#' # working with R's native date-time types. An error is thrown by default.
#' try(set_hour(y, 2))
#' set_hour(y, 2, nonexistent = "roll-forward")
#' set_hour(y, 2, nonexistent = "roll-backward")
NULL

#' @rdname posixt-setters
#' @export
set_year.POSIXt <- function(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x) {
  set_posixt_field_year_month_day(x, value, invalid, nonexistent, ambiguous, set_year, ...)
}
#' @rdname posixt-setters
#' @export
set_month.POSIXt <- function(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x) {
  set_posixt_field_year_month_day(x, value, invalid, nonexistent, ambiguous, set_month, ...)
}
#' @rdname posixt-setters
#' @export
set_day.POSIXt <- function(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x) {
  set_posixt_field_year_month_day(x, value, invalid, nonexistent, ambiguous, set_day, ...)
}
#' @rdname posixt-setters
#' @export
set_hour.POSIXt <- function(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x) {
  set_posixt_field_year_month_day(x, value, invalid, nonexistent, ambiguous, set_hour, ...)
}
#' @rdname posixt-setters
#' @export
set_minute.POSIXt <- function(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x) {
  set_posixt_field_year_month_day(x, value, invalid, nonexistent, ambiguous, set_minute, ...)
}
#' @rdname posixt-setters
#' @export
set_second.POSIXt <- function(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x) {
  set_posixt_field_year_month_day(x, value, invalid, nonexistent, ambiguous, set_second, ...)
}
set_posixt_field_year_month_day <- function(x, value, invalid, nonexistent, ambiguous, set_fn, ...) {
  check_dots_empty()
  zone <- posixt_tzone(x)
  x <- as_year_month_day(x)
  x <- set_fn(x, value)
  x <- invalid_resolve(x, invalid = invalid)
  as.POSIXct(x, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

# ------------------------------------------------------------------------------

#' @method vec_arith.POSIXct clock_duration
#' @export
vec_arith.POSIXct.clock_duration <- function(op, x, y, ...) {
  arith_posixt_and_duration(op, x, y, ...)
}

#' @method vec_arith.clock_duration POSIXct
#' @export
vec_arith.clock_duration.POSIXct <- function(op, x, y, ...) {
  arith_duration_and_posixt(op, x, y, ...)
}

#' @method vec_arith.POSIXlt clock_duration
#' @export
vec_arith.POSIXlt.clock_duration <- function(op, x, y, ...) {
  arith_posixt_and_duration(op, x, y, ...)
}

#' @method vec_arith.clock_duration POSIXlt
#' @export
vec_arith.clock_duration.POSIXlt <- function(op, x, y, ...) {
  arith_duration_and_posixt(op, x, y, ...)
}

arith_posixt_and_duration <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(x, y),
    "-" = add_duration(x, -y),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_duration_and_posixt <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(y, x, swapped = TRUE),
    "-" = stop_incompatible_op(op, x, y, details = "Can't subtract a POSIXct/POSIXlt from a duration.", ...),
    stop_incompatible_op(op, x, y, ...)
  )
}

# ------------------------------------------------------------------------------

#' Arithmetic: date-time
#'
#' @description
#' These are POSIXct/POSIXlt methods for the
#' [arithmetic generics][clock-arithmetic].
#'
#' Calendrical based arithmetic:
#'
#' These functions convert to a naive-time, then to a year-month-day, perform
#' the arithmetic, then convert back to a date-time.
#'
#' - `add_years()`
#'
#' - `add_quarters()`
#'
#' - `add_months()`
#'
#' Naive-time based arithmetic:
#'
#' These functions convert to a naive-time, perform the arithmetic, then
#' convert back to a date-time.
#'
#' - `add_weeks()`
#'
#' - `add_days()`
#'
#' Sys-time based arithmetic:
#'
#' These functions convert to a sys-time, perform the arithmetic, then
#' convert back to a date-time.
#'
#' - `add_hours()`
#'
#' - `add_minutes()`
#'
#' - `add_seconds()`
#'
#' @details
#' Adding a single quarter with `add_quarters()` is equivalent to adding
#' 3 months.
#'
#' `x` and `n` are recycled against each other.
#'
#' Calendrical based arithmetic has the potential to generate invalid dates
#' (like the 31st of February), nonexistent times (due to daylight saving
#' time gaps), and ambiguous times (due to daylight saving time fallbacks).
#'
#' Naive-time based arithmetic will never generate an invalid date, but
#' may generate a nonexistent or ambiguous time (i.e. you added 1 day and
#' landed in a daylight saving time gap).
#'
#' Sys-time based arithmetic operates in the UTC time zone, which means
#' that it will never generate any invalid dates or nonexistent / ambiguous
#' times.
#'
#' The conversion from POSIXct/POSIXlt to the corresponding clock type uses
#' a "best guess" about whether you want to do the arithmetic using a naive-time
#' or a sys-time. For example, when adding months, you probably want to
#' retain the printed time when converting to a year-month-day to perform
#' the arithmetic, so the conversion goes through naive-time. However,
#' when adding smaller units like seconds, you probably want
#' `"2020-03-08 01:59:59" + 1 second` in the America/New_York time zone to
#' return `"2020-03-08 03:00:00"`, taking into account the fact that there
#' was a daylight saving time gap. This requires doing the arithmetic in
#' sys-time, so that is what clock converts to. If you disagree with this
#' heuristic for any reason, you can take control and perform the conversions
#' yourself. For example, you could convert the previous example to a
#' naive-time instead of a sys-time manually with [as_naive_time()], add
#' 1 second giving `"2020-03-08 02:00:00"`, then convert back to a
#' POSIXct/POSIXlt, dealing with the nonexistent time that gets created by
#' using the `nonexistent` argument of `as.POSIXct()`.
#'
#' @inheritParams add_years
#' @inheritParams invalid_resolve
#' @inheritParams as-zoned-time-naive-time
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time vector.
#'
#' @return `x` after performing the arithmetic.
#'
#' @name posixt-arithmetic
#'
#' @examples
#' x <- as.POSIXct("2019-01-01", tz = "America/New_York")
#'
#' add_years(x, 1:5)
#'
#' y <- as.POSIXct("2019-01-31 00:30:00", tz = "America/New_York")
#'
#' # Adding 1 month to `y` generates an invalid date. Unlike year-month-day
#' # types, R's native date-time types cannot handle invalid dates, so you must
#' # resolve them immediately. If you don't you get an error:
#' try(add_months(y, 1:2))
#' add_months(as_year_month_day(y), 1:2)
#'
#' # Resolve invalid dates by specifying an invalid date resolution strategy
#' # with the `invalid` argument. Using `"previous"` here sets the date-time to
#' # the previous valid moment in time - i.e. the end of the month. The
#' # time is set to the last moment in the day to retain the relative ordering
#' # within your input. If you are okay with potentially losing this, and
#' # want to retain your time of day, you can use `"previous-day"` to set the
#' # date-time to the previous valid day, while keeping the time of day.
#' add_months(y, 1:2, invalid = "previous")
#' add_months(y, 1:2, invalid = "previous-day")
NULL

#' @rdname posixt-arithmetic
#' @export
add_years.POSIXt <- function(x, n, ..., invalid = NULL, nonexistent = NULL, ambiguous = x) {
  add_posixt_duration_year_month_day(x, n, invalid, nonexistent, ambiguous, add_years, ...)
}
#' @rdname posixt-arithmetic
#' @export
add_quarters.POSIXt <- function(x, n, ..., invalid = NULL, nonexistent = NULL, ambiguous = x) {
  add_posixt_duration_year_month_day(x, n, invalid, nonexistent, ambiguous, add_quarters, ...)
}
#' @rdname posixt-arithmetic
#' @export
add_months.POSIXt <- function(x, n, ..., invalid = NULL, nonexistent = NULL, ambiguous = x) {
  add_posixt_duration_year_month_day(x, n, invalid, nonexistent, ambiguous, add_months, ...)
}
add_posixt_duration_year_month_day <- function(x, n, invalid, nonexistent, ambiguous, add_fn, ...) {
  check_dots_empty()
  zone <- posixt_tzone(x)
  x <- as_year_month_day(x)
  x <- add_fn(x, n)
  x <- invalid_resolve(x, invalid = invalid)
  as.POSIXct(x, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

#' @rdname posixt-arithmetic
#' @export
add_weeks.POSIXt <- function(x, n, ..., nonexistent = NULL, ambiguous = x) {
  add_posixt_duration_naive_time_point(x, n, nonexistent, ambiguous, add_weeks, ...)
}
#' @rdname posixt-arithmetic
#' @export
add_days.POSIXt <- function(x, n, ..., nonexistent = NULL, ambiguous = x) {
  add_posixt_duration_naive_time_point(x, n, nonexistent, ambiguous, add_days, ...)
}
add_posixt_duration_naive_time_point <- function(x, n, nonexistent, ambiguous, add_fn, ...) {
  check_dots_empty()
  zone <- posixt_tzone(x)
  x <- as_naive_time(x)
  x <- add_fn(x, n)
  as.POSIXct(x, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

#' @rdname posixt-arithmetic
#' @export
add_hours.POSIXt <- function(x, n, ...) {
  add_posixt_duration_sys_time_point(x, n, add_hours, ...)
}
#' @rdname posixt-arithmetic
#' @export
add_minutes.POSIXt <- function(x, n, ...) {
  add_posixt_duration_sys_time_point(x, n, add_minutes, ...)
}
#' @rdname posixt-arithmetic
#' @export
add_seconds.POSIXt <- function(x, n, ...) {
  add_posixt_duration_sys_time_point(x, n, add_seconds, ...)
}
add_posixt_duration_sys_time_point <- function(x, n, add_fn, ...) {
  check_dots_empty()
  zone <- posixt_tzone(x)
  x <- as_sys_time(x)
  x <- add_fn(x, n)
  as.POSIXct(x, tz = zone)
}

# ------------------------------------------------------------------------------

#' Group date-time components
#'
#' @description
#' This is a POSIXct/POSIXlt method for the [date_group()] generic.
#'
#' `date_group()` groups by a single component of a date-time, such as month
#' of the year, day of the month, or hour of the day.
#'
#' If you need to group by more complex components, like ISO weeks, or quarters,
#' convert to a calendar type that contains the component you are interested
#' in grouping by.
#'
#' @inheritParams date_group
#' @inheritParams invalid_resolve
#' @inheritParams as-zoned-time-naive-time
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'
#'   - `"month"`
#'
#'   - `"day"`
#'
#'   - `"hour"`
#'
#'   - `"minute"`
#'
#'   - `"second"`
#'
#' @return `x`, grouped at `precision`.
#'
#' @name posixt-group
#'
#' @export
#' @examples
#' x <- as.POSIXct("2019-01-01", "America/New_York")
#' x <- add_days(x, -3:5)
#'
#' # Group by 2 days of the current month.
#' # Note that this resets at the beginning of the month, creating day groups
#' # of [29, 30] [31] [01, 02] [03, 04].
#' date_group(x, "day", n = 2)
#'
#' # Group by month
#' date_group(x, "month")
#'
#' # Group by hour of the day
#' y <- as.POSIXct("2019-12-30", "America/New_York")
#' y <- add_hours(y, 0:12)
#' y
#'
#' date_group(y, "hour", n = 3)
date_group.POSIXt <- function(x,
                              precision,
                              ...,
                              n = 1L,
                              invalid = NULL,
                              nonexistent = NULL,
                              ambiguous = x) {
  force(ambiguous)
  zone <- date_zone(x)
  x <- as_year_month_day(x)
  x <- calendar_group(x, precision, ..., n = n)
  x <- calendar_widen(x, "second")
  as.POSIXct(x, zone, invalid = invalid, nonexistent = nonexistent, ambiguous = ambiguous)
}

# ------------------------------------------------------------------------------

#' @export
date_leap_year.POSIXt <- function(x) {
  x <- as_year_month_day(x)
  calendar_leap_year(x)
}

# ------------------------------------------------------------------------------

#' Rounding: date-time
#'
#' @description
#' These are POSIXct/POSIXlt methods for the
#' [rounding generics][date-and-date-time-rounding].
#'
#' - `date_floor()` rounds a date-time down to a multiple of
#'   the specified `precision`.
#'
#' - `date_ceiling()` rounds a date-time up to a multiple of
#'   the specified `precision`.
#'
#' - `date_round()` rounds up or down depending on what is closer,
#'   rounding up on ties.
#'
#' You can group by irregular periods such as `"month"` or `"year"` by using
#' [date_group()].
#'
#' @details
#' When rounding by `"week"`, remember that the `origin` determines the "week
#' start". By default, 1970-01-01 is the implicit origin, which is a
#' Thursday. If you would like to round by weeks with a different week start,
#' just supply an origin on the weekday you are interested in.
#'
#' @inheritParams date_floor
#' @inheritParams as-zoned-time-naive-time
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"week"`
#'
#'   - `"day"`
#'
#'   - `"hour"`
#'
#'   - `"minute"`
#'
#'   - `"second"`
#'
#'   `"week"` is an alias for `"day"` with `n * 7`.
#'
#' @param origin `[POSIXct(1) / POSIXlt(1) / NULL]`
#'
#'   An origin to start counting from.
#'
#'   `origin` must have exactly the same time zone as `x`.
#'
#'   `origin` will be floored to `precision`. If information is lost when
#'   flooring, a warning will be thrown.
#'
#'   If `NULL`, defaults to midnight on 1970-01-01 in the time zone of `x`.
#'
#' @return `x` rounded to the specified `precision`.
#'
#' @name posixt-rounding
#'
#' @examples
#' x <- as.POSIXct("2019-03-31", "America/New_York")
#' x <- add_days(x, 0:5)
#'
#' # Flooring by 2 days, note that this is not tied to the current month,
#' # and instead counts from the specified `origin`, so groups can cross
#' # the month boundary
#' date_floor(x, "day", n = 2)
#'
#' # Compare to `date_group()`, which groups by the day of the month
#' date_group(x, "day", n = 2)
#'
#' # Note that daylight saving time gaps can throw off rounding
#' x <- as.POSIXct("1970-04-26 01:59:59", "America/New_York") + c(0, 1)
#' x
#'
#' # Rounding is done in naive-time, which means that rounding by 2 hours
#' # will attempt to generate a time of 1970-04-26 02:00:00, which doesn't
#' # exist in this time zone
#' try(date_floor(x, "hour", n = 2))
#'
#' # You can handle this by specifying a nonexistent time resolution strategy
#' date_floor(x, "hour", n = 2, nonexistent = "roll-forward")
NULL

#' @rdname posixt-rounding
#' @export
date_floor.POSIXt <- function(x,
                              precision,
                              ...,
                              n = 1L,
                              origin = NULL,
                              nonexistent = NULL,
                              ambiguous = x) {
  date_time_rounder(x, precision, n, origin, nonexistent, ambiguous, time_point_floor, ...)
}

#' @rdname posixt-rounding
#' @export
date_ceiling.POSIXt <- function(x,
                                precision,
                                ...,
                                n = 1L,
                                origin = NULL,
                                nonexistent = NULL,
                                ambiguous = x) {
  date_time_rounder(x, precision, n, origin, nonexistent, ambiguous, time_point_ceiling, ...)
}

#' @rdname posixt-rounding
#' @export
date_round.POSIXt <- function(x,
                              precision,
                              ...,
                              n = 1L,
                              origin = NULL,
                              nonexistent = NULL,
                              ambiguous = x) {
  date_time_rounder(x, precision, n, origin, nonexistent, ambiguous, time_point_round, ...)
}

date_time_rounder <- function(x,
                              precision,
                              n,
                              origin,
                              nonexistent,
                              ambiguous,
                              time_point_rounder,
                              ...) {
  result <- tweak_date_rounder_precision(precision, n)
  precision <- result$precision
  n <- result$n

  zone <- date_zone(x)

  x <- as_naive_time(x)

  if (!is_null(origin)) {
    origin <- collect_date_time_rounder_origin(origin, zone, precision)
  }

  x <- time_point_rounder(x, precision, ..., n = n, origin = origin)

  as.POSIXct(x, zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

collect_date_time_rounder_origin <- function(origin, zone, precision) {
  if (!inherits(origin, "POSIXt")) {
    abort("`origin` must be a 'POSIXt'.")
  }

  origin <- to_posixct(origin)

  if (length(origin) != 1L) {
    abort("`origin` must have length 1.")
  }
  if (!is.finite(origin)) {
    abort("`origin` must not be `NA` or an infinite date.")
  }

  if (!identical(date_zone(origin), zone)) {
    abort("`origin` must have the same time zone as `x`.")
  }

  origin <- as_naive_time(origin)

  # Floor to match the precision of `precision`
  origin_old <- origin
  origin <- time_point_floor(origin, precision)

  # If we lost information while flooring, let the user know
  if (origin_old != time_point_cast(origin, "second")) {
    warn_clock_invalid_rounding_origin(precision)
  }

  origin
}

warn_clock_invalid_rounding_origin <- function(precision) {
  message <- paste0(
    "`origin` has been floored from 'second' precision to '", precision, "' ",
    "precision to match `precision`. This floor has lost information."
  )

  rlang::warn(message, class = "clock_warning_invalid_rounding_origin")
}

# ------------------------------------------------------------------------------

#' @export
date_weekday_factor.POSIXt <- function(x,
                                       ...,
                                       labels = "en",
                                       abbreviate = TRUE,
                                       encoding = "western") {
  x <- as_weekday(x)
  weekday_factor(x, ..., labels = labels, abbreviate = abbreviate, encoding = encoding)
}

# ------------------------------------------------------------------------------

#' @export
date_month_factor.POSIXt <- function(x,
                                     ...,
                                     labels = "en",
                                     abbreviate = FALSE) {
  x <- as_year_month_day(x)
  calendar_month_factor(x, ..., labels = labels, abbreviate = abbreviate)
}

# ------------------------------------------------------------------------------

#' Formatting: date-time
#'
#' @description
#' This is a POSIXct method for the [date_format()] generic.
#'
#' `date_format()` formats a date-time (POSIXct) using a `format` string.
#'
#' If `format` is `NULL`, a default format of `"%Y-%m-%dT%H:%M:%S%Ez[%Z]"` is
#' used. This matches the default format that [date_time_parse_complete()]
#' parses. Additionally, this format matches the de-facto standard extension to
#' RFC 3339 for creating completely unambiguous date-times.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams format.clock_zoned_time
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time vector.
#'
#' @return A character vector of the formatted input.
#'
#' @name posixt-formatting
#'
#' @export
#' @examples
#' x <- date_time_parse(
#'   c("1970-04-26 01:30:00", "1970-04-26 03:30:00"),
#'   zone = "America/New_York"
#' )
#'
#' # Default
#' date_format(x)
#'
#' # Which is parseable by `date_time_parse_complete()`
#' date_time_parse_complete(date_format(x))
#'
#' date_format(x, format = "%B %d, %Y %H:%M:%S")
#'
#' # By default, `%Z` uses the full zone name, but you can switch to the
#' # abbreviated name
#' date_format(x, format = "%z %Z")
#' date_format(x, format = "%z %Z", abbreviate_zone = TRUE)
date_format.POSIXt <- function(x,
                               ...,
                               format = NULL,
                               locale = clock_locale(),
                               abbreviate_zone = FALSE) {
  check_dots_empty()
  x <- as_zoned_time(x)
  format(x, format = format, locale = locale, abbreviate_zone = abbreviate_zone)
}

# ------------------------------------------------------------------------------

#' Get or set the time zone
#'
#' @description
#' - `date_zone()` gets the time zone.
#'
#' - `date_set_zone()` sets the time zone. This retains the _underlying
#' duration_, but changes the _printed time_ depending on the zone that is
#' chosen.
#'
#' @details
#' This function is only valid for date-times, as clock treats R's Date class as
#' a _naive_ type, which always has a yet-to-be-specified time zone.
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time vector.
#'
#' @param zone `[character(1)]`
#'
#'   A valid time zone to switch to.
#'
#' @return
#' - `date_zone()` returns a string containing the time zone.
#'
#' - `date_set_zone()` returns `x` with an altered printed time. The
#' underlying duration is not changed.
#'
#' @name date-zone
#'
#' @examples
#' library(magrittr)
#'
#' # Cannot set or get the zone of Date.
#' # clock assumes that Dates are naive types, like naive-time.
#' x <- as.Date("2019-01-01")
#' try(date_zone(x))
#' try(date_set_zone(x, "America/New_York"))
#'
#' x <- as.POSIXct("2019-01-02 01:30:00", tz = "America/New_York")
#' x
#'
#' date_zone(x)
#'
#' # If it is 1:30am in New York, what time is it in Los Angeles?
#' # Same underlying duration, new printed time
#' date_set_zone(x, "America/Los_Angeles")
#'
#' # If you want to retain the printed time, but change the underlying duration,
#' # convert to a naive-time to drop the time zone, then convert back to a
#' # date-time. Be aware that this requires that you handle daylight saving time
#' # irregularities with the `nonexistent` and `ambiguous` arguments to
#' # `as.POSIXct()`!
#' x %>%
#'   as_naive_time() %>%
#'   as.POSIXct("America/Los_Angeles")
#'
#' y <- as.POSIXct("2021-03-28 03:30:00", "America/New_York")
#' y
#'
#' y_nt <- as_naive_time(y)
#' y_nt
#'
#' # Helsinki had a daylight saving time gap where they jumped from
#' # 02:59:59 -> 04:00:00
#' try(as.POSIXct(y_nt, "Europe/Helsinki"))
#'
#' as.POSIXct(y_nt, "Europe/Helsinki", nonexistent = "roll-forward")
#' as.POSIXct(y_nt, "Europe/Helsinki", nonexistent = "roll-backward")
NULL

#' @rdname date-zone
#' @export
date_zone <- function(x) {
  UseMethod("date_zone")
}

#' @export
date_zone.POSIXt <- function(x) {
  posixt_tzone(x)
}

#' @rdname date-zone
#' @export
date_set_zone <- function(x, zone) {
  UseMethod("date_set_zone")
}

#' @export
date_set_zone.POSIXt <- function(x, zone) {
  x <- to_posixct(x)
  zone <- zone_validate(zone)
  posixt_set_tzone(x, zone)
}

# ------------------------------------------------------------------------------

#' Parsing: date-time
#'
#' @description
#' There are four parsers for parsing strings into POSIXct date-times,
#' `date_time_parse()`, `date_time_parse_complete()`,
#' `date_time_parse_abbrev()`, and `date_time_parse_RFC_3339()`.
#'
#' ## date_time_parse()
#'
#' `date_time_parse()` is useful for strings like `"2019-01-01 00:00:00"`, where
#' the UTC offset and full time zone name are not present in the string. The
#' string is first parsed as a naive-time without any time zone assumptions, and
#' is then converted to a POSIXct with the supplied `zone`.
#'
#' Because converting from naive-time to POSIXct may result in nonexistent or
#' ambiguous times due to daylight saving time, these must be resolved
#' explicitly with the `nonexistent` and `ambiguous` arguments.
#'
#' `date_time_parse()` completely ignores the `%z` and `%Z` commands. The only
#' time zone specific information that is used is the `zone`.
#'
#' The default `format` used is `"%Y-%m-%d %H:%M:%S"`. This matches the default
#' result from calling `format()` on a POSIXct date-time.
#'
#' ## date_time_parse_complete()
#'
#' `date_time_parse_complete()` is a parser for _complete_ date-time strings,
#' like `"2019-01-01T00:00:00-05:00[America/New_York]"`. A complete date-time
#' string has both the time zone offset and full time zone name in the string,
#' which is the only way for the string itself to contain all of the information
#' required to unambiguously construct a zoned-time. Because of this,
#' `date_time_parse_complete()` requires both the `%z` and `%Z` commands to be
#' supplied in the `format` string.
#'
#' The default `format` used is `"%Y-%m-%dT%H:%M:%S%Ez[%Z]"`. This matches the
#' default result from calling `date_format()` on a POSIXct date-time.
#' Additionally, this format matches the de-facto standard extension to RFC 3339
#' for creating completely unambiguous date-times.
#'
#' ## date_time_parse_abbrev()
#'
#' `date_time_parse_abbrev()` is a parser for date-time strings containing only
#' a time zone abbreviation, like `"2019-01-01 00:00:00 EST"`. The time zone
#' abbreviation is not enough to identify the full time zone name that the
#' date-time belongs to, so the full time zone name must be supplied as the
#' `zone` argument. However, the time zone abbreviation can help with resolving
#' ambiguity around daylight saving time fallbacks.
#'
#' For `date_time_parse_abbrev()`, `%Z` must be supplied and is interpreted as
#' the time zone abbreviation rather than the full time zone name.
#'
#' If used, the `%z` command must parse correctly, but its value will be
#' completely ignored.
#'
#' The default `format` used is `"%Y-%m-%d %H:%M:%S %Z"`. This matches the
#' default result from calling `print()` or `format(usetz = TRUE)` on a POSIXct
#' date-time.
#'
#' ## date_time_parse_RFC_3339()
#'
#' `date_time_parse_RFC_3339()` is a parser for date-time strings in the
#' extremely common date-time format outlined by [RFC
#' 3339](https://datatracker.ietf.org/doc/html/rfc3339). This document outlines
#' a profile of the ISO 8601 format that is even more restrictive, but
#' corresponds to the most common formats that are likely to be used in
#' internet protocols (i.e. through APIs).
#'
#' In particular, this function is intended to parse the following three
#' formats:
#'
#' ```
#' 2019-01-01T00:00:00Z
#' 2019-01-01T00:00:00+0430
#' 2019-01-01T00:00:00+04:30
#' ```
#'
#' This function defaults to parsing the first of these formats by using
#' a format string of `"%Y-%m-%dT%H:%M:%SZ"`.
#'
#' If your date-time strings use offsets from UTC rather than `"Z"`, then set
#' `offset` to one of the following:
#'
#' - `"%z"` if the offset is of the form `"+0430"`.
#' - `"%Ez"` if the offset is of the form `"+04:30"`.
#'
#' The RFC 3339 standard allows for replacing the `"T"` with a `"t"` or a space
#' (`" "`). Set `separator` to adjust this as needed.
#'
#' The date-times returned by this function will always be in the UTC time zone.
#'
#' @details
#' If `date_time_parse_complete()` is given input that is length zero, all
#' `NA`s, or completely fails to parse, then no time zone will be able to be
#' determined. In that case, the result will use `"UTC"`.
#'
#' If you have strings with sub-second components, then these date-time parsers
#' are not appropriate for you. Remember that clock treats POSIXct as a second
#' precision type, so parsing a string with fractional seconds directly into a
#' POSIXct is ambiguous and undefined. Instead, fully parse the string,
#' including its fractional seconds, into a clock type that can handle it, such
#' as a naive-time with [naive_time_parse()], then round to seconds with
#' whatever rounding convention is appropriate for your use case, such as
#' [time_point_floor()], and finally convert that to POSIXct with
#' [as_date_time()]. This gives you complete control over how the fractional
#' seconds are handled when converting to POSIXct.
#'
#' @inheritParams zoned-parsing
#' @inheritParams as-zoned-time-naive-time
#' @inheritParams sys-parsing
#'
#' @return A POSIXct.
#'
#' @name date-time-parse
#'
#' @examples
#' # Parse with a known `zone`, even though that information isn't in the string
#' date_time_parse("2020-01-01 05:06:07", "America/New_York")
#'
#' # Same time as above, except this is a completely unambiguous parse that
#' # doesn't require a `zone` argument, because the zone name and offset are
#' # both present in the string
#' date_time_parse_complete("2020-01-01T05:06:07-05:00[America/New_York]")
#'
#' # Only day components
#' date_time_parse("2020-01-01", "America/New_York", format = "%Y-%m-%d")
#'
#' # `date_time_parse()` may have issues with ambiguous times due to daylight
#' # saving time fallbacks. For example, there were two 1'oclock hours here:
#' x <- date_time_parse("1970-10-25 00:59:59", "America/New_York")
#'
#' # First (earliest) 1'oclock hour
#' add_seconds(x, 1)
#' # Second (latest) 1'oclock hour
#' add_seconds(x, 3601)
#'
#' # If you try to parse this ambiguous time directly, you'll get an error:
#' ambiguous_time <- "1970-10-25 01:00:00"
#' try(date_time_parse(ambiguous_time, "America/New_York"))
#'
#' # Resolve it by specifying whether you'd like to use the
#' # `earliest` or `latest` of the two possible times
#' date_time_parse(ambiguous_time, "America/New_York", ambiguous = "earliest")
#' date_time_parse(ambiguous_time, "America/New_York", ambiguous = "latest")
#'
#' # `date_time_parse_complete()` doesn't have these issues, as it requires
#' # that the offset and zone name are both in the string, which resolves
#' # the ambiguity
#' complete_times <- c(
#'   "1970-10-25T01:00:00-04:00[America/New_York]",
#'   "1970-10-25T01:00:00-05:00[America/New_York]"
#' )
#' date_time_parse_complete(complete_times)
#'
#' # `date_time_parse_abbrev()` also doesn't have these issues, since it
#' # uses the time zone abbreviation name to resolve the ambiguity
#' abbrev_times <- c(
#'   "1970-10-25 01:00:00 EDT",
#'   "1970-10-25 01:00:00 EST"
#' )
#' date_time_parse_abbrev(abbrev_times, "America/New_York")
#'
#' # ---------------------------------------------------------------------------
#' # RFC 3339
#'
#' # Typical UTC format
#' x <- "2019-01-01T00:01:02Z"
#' date_time_parse_RFC_3339(x)
#'
#' # With a UTC offset containing a `:`
#' x <- "2019-01-01T00:01:02+02:30"
#' date_time_parse_RFC_3339(x, offset = "%Ez")
#'
#' # With a space between the date and time and no `:` in the offset
#' x <- "2019-01-01 00:01:02+0230"
#' date_time_parse_RFC_3339(x, separator = " ", offset = "%z")
#'
#' # ---------------------------------------------------------------------------
#' # Sub-second components
#'
#' # If you have a string with sub-second components, but only require up to
#' # seconds, first parse them into a clock type that can handle sub-seconds to
#' # fully capture that information, then round using whatever convention is
#' # required for your use case before converting to a date-time.
#' x <- c("2019-01-01T00:00:01.1", "2019-01-01T00:00:01.78")
#'
#' x <- naive_time_parse(x, precision = "millisecond")
#' x
#'
#' time_point_floor(x, "second")
#' time_point_round(x, "second")
#'
#' as_date_time(time_point_round(x, "second"), "America/New_York")
NULL

#' @rdname date-time-parse
#' @export
date_time_parse <- function(x,
                            zone,
                            ...,
                            format = NULL,
                            locale = clock_locale(),
                            nonexistent = NULL,
                            ambiguous = NULL) {
  if (is_null(format)) {
    # Default format for `date_time_parse()` doesn't have the `T`, unlike
    # default format for `naive_time_parse()`. This is intended to parse the
    # format returned by `format(<POSIXct>)` by default.
    format <- "%Y-%m-%d %H:%M:%S"
  }

  x <- naive_time_parse(x, ..., format = format, precision = "second", locale = locale)
  as.POSIXct(x, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

#' @rdname date-time-parse
#' @export
date_time_parse_complete <- function(x, ..., format = NULL, locale = clock_locale()) {
  x <- zoned_time_parse_complete(x, ..., format = format, precision = "second", locale = locale)
  as.POSIXct(x)
}

#' @rdname date-time-parse
#' @export
date_time_parse_abbrev <- function(x, zone, ..., format = NULL, locale = clock_locale()) {
  x <- zoned_time_parse_abbrev(x, zone, ..., format = format, precision = "second", locale = locale)
  as.POSIXct(x)
}

#' @rdname date-time-parse
#' @export
date_time_parse_RFC_3339 <- function(x,
                                     ...,
                                     separator = "T",
                                     offset = "Z") {
  x <- sys_time_parse_RFC_3339(x, ..., separator = separator, offset = offset, precision = "second")
  # Hard-code UTC because we don't allow optional `zone` arguments that have a
  # default anywhere else in the package. It seems like it would be bad practice
  # to have one here, since parsed RFC 3339 strings should probably always be
  # interpreted as UTC.
  as.POSIXct(x, tz = "UTC")
}

# ------------------------------------------------------------------------------

#' Shifting: date and date-time
#'
#' @description
#' `date_shift()` shifts `x` to the `target` weekday. You can shift to the next
#' or previous weekday. If `x` is currently on the `target` weekday, you can
#' choose to leave it alone or advance it to the next instance of the `target`.
#'
#' Shifting with date-times retains the time of day where possible. Be aware
#' that you can run into daylight saving time issues if you shift into a
#' daylight saving time gap or fallback period.
#'
#' @inheritParams time_point_shift
#' @inheritParams as-zoned-time-naive-time
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time vector.
#'
#' @return `x` shifted to the `target` weekday.
#'
#' @name posixt-shifting
#'
#' @export
#' @examples
#' tuesday <- weekday(clock_weekdays$tuesday)
#'
#' x <- as.POSIXct("1970-04-22 02:30:00", "America/New_York")
#'
#' # Shift to the next Tuesday
#' date_shift(x, tuesday)
#'
#' # Be aware that you can run into daylight saving time issues!
#' # Here we shift directly into a daylight saving time gap
#' # from 01:59:59 -> 03:00:00
#' sunday <- weekday(clock_weekdays$sunday)
#' try(date_shift(x, sunday))
#'
#' # You can resolve this with the `nonexistent` argument
#' date_shift(x, sunday, nonexistent = "roll-forward")
date_shift.POSIXt <- function(x,
                              target,
                              ...,
                              which = "next",
                              boundary = "keep",
                              nonexistent = NULL,
                              ambiguous = x) {
  force(ambiguous)
  zone <- date_zone(x)
  x <- as_naive_time(x)
  x <- time_point_shift(x, target, ..., which = which, boundary = boundary)
  as.POSIXct(x, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

# ------------------------------------------------------------------------------

#' Building: date-time
#'
#' @description
#' `date_time_build()` builds a POSIXct from it's individual components.
#'
#' To build a POSIXct, it is required that you specify the `zone`.
#'
#' @details
#' Components are recycled against each other.
#'
#' @inheritParams invalid_resolve
#' @inheritParams as-zoned-time-naive-time
#'
#' @param year `[integer]`
#'
#'   The year. Values `[-32767, 32767]` are generally allowed.
#'
#' @param month `[integer]`
#'
#'   The month. Values `[1, 12]` are allowed.
#'
#' @param day `[integer / "last"]`
#'
#'   The day of the month. Values `[1, 31]` are allowed.
#'
#'   If `"last"`, then the last day of the month is returned.
#'
#' @param hour `[integer]`
#'
#'   The hour. Values `[0, 23]` are allowed.
#'
#' @param minute `[integer]`
#'
#'   The minute. Values `[0, 59]` are allowed.
#'
#' @param second `[integer]`
#'
#'   The second. Values `[0, 59]` are allowed.
#'
#' @param zone `[character(1)]`
#'
#'   A valid time zone name.
#'
#'   This argument is required, and must be specified by name.
#'
#' @return A POSIXct.
#'
#' @export
#' @examples
#' # The zone argument is required!
#' # clock always requires you to be explicit about your choice of `zone`.
#' try(date_time_build(2020))
#'
#' date_time_build(2020, zone = "America/New_York")
#'
#' # Nonexistent time due to daylight saving time gap from 01:59:59 -> 03:00:00
#' try(date_time_build(1970, 4, 26, 1:12, 30, zone = "America/New_York"))
#'
#' # Resolve with a nonexistent time resolution strategy
#' date_time_build(
#'   1970, 4, 26, 1:12, 30,
#'   zone = "America/New_York",
#'   nonexistent = "roll-forward"
#' )
date_time_build <- function(year,
                            month = 1L,
                            day = 1L,
                            hour = 0L,
                            minute = 0L,
                            second = 0L,
                            ...,
                            zone,
                            invalid = NULL,
                            nonexistent = NULL,
                            ambiguous = NULL) {
  check_dots_empty()

  if (is_missing(zone)) {
    abort("`zone` is a required argument to `date_time_build()`.")
  }

  x <- year_month_day(year, month, day, hour, minute, second)
  x <- invalid_resolve(x, invalid = invalid)
  as.POSIXct(x, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
}

# ------------------------------------------------------------------------------

#' @rdname date-today
#' @export
date_now <- function(zone) {
  as.POSIXct(zoned_time_now(zone))
}

# ------------------------------------------------------------------------------

#' Boundaries: date-time
#'
#' @description
#' This is a POSIXct/POSIXlt method for the [date_start()] and [date_end()]
#' generics.
#'
#' @inheritParams date_group
#' @inheritParams invalid_resolve
#' @inheritParams as-zoned-time-naive-time
#'
#' @param x `[POSIXct / POSIXlt]`
#'
#'   A date-time vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'
#'   - `"month"`
#'
#'   - `"day"`
#'
#'   - `"hour"`
#'
#'   - `"minute"`
#'
#'   - `"second"`
#'
#' @return `x` but with some components altered to be at the boundary value.
#'
#' @name posixt-boundary
#'
#' @examples
#' x <- date_time_build(2019:2021, 2:4, 3:5, 4, 5, 6, zone = "America/New_York")
#' x
#'
#' # Last moment of the month
#' date_end(x, "month")
#'
#' # Notice that this is different from just setting the day to `"last"`
#' set_day(x, "last")
#'
#' # Last moment of the year
#' date_end(x, "year")
#'
#' # First moment of the hour
#' date_start(x, "hour")
NULL

#' @rdname posixt-boundary
#' @export
date_start.POSIXt <- function(x,
                              precision,
                              ...,
                              invalid = NULL,
                              nonexistent = NULL,
                              ambiguous = x) {
  check_dots_empty()
  force(ambiguous)
  zone <- date_zone(x)
  x <- as_year_month_day(x)
  x <- calendar_start(x, precision)
  as.POSIXct(x, zone, invalid = invalid, nonexistent = nonexistent, ambiguous = ambiguous)
}

#' @rdname posixt-boundary
#' @export
date_end.POSIXt <- function(x,
                            precision,
                            ...,
                            invalid = NULL,
                            nonexistent = NULL,
                            ambiguous = x) {
  check_dots_empty()
  force(ambiguous)
  zone <- date_zone(x)
  x <- as_year_month_day(x)
  x <- calendar_end(x, precision)
  as.POSIXct(x, zone, invalid = invalid, nonexistent = nonexistent, ambiguous = ambiguous)
}

# ------------------------------------------------------------------------------

#' Sequences: date-time
#'
#' @description
#' This is a POSIXct method for the [date_seq()] generic.
#'
#' `date_seq()` generates a date-time (POSIXct) sequence.
#'
#' When calling `date_seq()`, exactly two of the following must be specified:
#' - `to`
#' - `by`
#' - `total_size`
#'
#' @section Sequence Generation:
#'
#' Different methods are used to generate the sequences, depending on the
#' precision implied by `by`. They are intended to generate the most intuitive
#' sequences, especially around daylight saving time gaps and fallbacks.
#'
#' See the examples for more details.
#'
#' ## Calendrical based sequences:
#'
#' These convert to a naive-time, then to a year-month-day, generate
#' the sequence, then convert back to a date-time.
#'
#' - `by = duration_years()`
#'
#' - `by = duration_quarters()`
#'
#' - `by = duration_months()`
#'
#' ## Naive-time based sequences:
#'
#' These convert to a naive-time, generate the sequence, then
#' convert back to a date-time.
#'
#' - `by = duration_weeks()`
#'
#' - `by = duration_days()`
#'
#' ## Sys-time based sequences:
#'
#' These convert to a sys-time, generate the sequence, then
#' convert back to a date-time.
#'
#' - `by = duration_hours()`
#'
#' - `by = duration_minutes()`
#'
#' - `by = duration_seconds()`
#'
#' @inheritParams date_seq
#' @inheritParams invalid_resolve
#' @inheritParams as-zoned-time-naive-time
#'
#' @param from `[POSIXct(1) / POSIXlt(1)]`
#'
#'   A date-time to start the sequence from.
#'
#' @param to `[POSIXct(1) / POSIXlt(1) / NULL]`
#'
#'   A date-time to stop the sequence at.
#'
#'   `to` is only included in the result if the resulting sequence divides
#'   the distance between `from` and `to` exactly.
#'
#'   If `to` is supplied along with `by`, all components of `to` more precise
#'   than the precision of `by` must match `from` exactly. For example, if `by =
#'   duration_months(1)`, the day, hour, minute, and second components of `to`
#'   must match the corresponding components of `from`. This ensures that the
#'   generated sequence is, at a minimum, a weakly monotonic sequence of
#'   date-times.
#'
#'   The time zone of `to` must match the time zone of `from` exactly.
#'
#' @param by `[integer(1) / clock_duration(1) / NULL]`
#'
#'   The unit to increment the sequence by.
#'
#'   If `by` is an integer, it is equivalent to `duration_seconds(by)`.
#'
#'   If `by` is a duration, it is allowed to have a precision of:
#'   - year
#'   - quarter
#'   - month
#'   - week
#'   - day
#'   - hour
#'   - minute
#'   - second
#'
#' @return A date-time vector.
#'
#' @name posixt-sequence
#'
#' @export
#' @examples
#' zone <- "America/New_York"
#'
#' from <- date_time_build(2019, 1, zone = zone)
#' to <- date_time_build(2019, 1, second = 50, zone = zone)
#'
#' # Defaults to second precision sequence
#' date_seq(from, to = to, by = 7)
#'
#' to <- date_time_build(2019, 1, 5, zone = zone)
#'
#' # Use durations to change to alternative precisions
#' date_seq(from, to = to, by = duration_days(1))
#' date_seq(from, to = to, by = duration_hours(10))
#' date_seq(from, by = duration_minutes(-2), total_size = 3)
#'
#' # Note that components of `to` more precise than the precision of `by`
#' # must match `from` exactly. For example, this is not well defined:
#' from <- date_time_build(2019, 1, 1, 0, 1, 30, zone = zone)
#' to <- date_time_build(2019, 1, 1, 5, 2, 20, zone = zone)
#' try(date_seq(from, to = to, by = duration_hours(1)))
#'
#' # The minute and second components of `to` must match `from`
#' to <- date_time_build(2019, 1, 1, 5, 1, 30, zone = zone)
#' date_seq(from, to = to, by = duration_hours(1))
#'
#' # ---------------------------------------------------------------------------
#'
#' # Invalid dates must be resolved with the `invalid` argument
#' from <- date_time_build(2019, 1, 31, zone = zone)
#' to <- date_time_build(2019, 12, 31, zone = zone)
#'
#' try(date_seq(from, to = to, by = duration_months(1)))
#' date_seq(from, to = to, by = duration_months(1), invalid = "previous-day")
#'
#' # Compare this to the base R result, which is often a source of confusion
#' seq(from, to = to, by = "1 month")
#'
#' # This is equivalent to the overflow invalid resolution strategy
#' date_seq(from, to = to, by = duration_months(1), invalid = "overflow")
#'
#' # ---------------------------------------------------------------------------
#'
#' # This date-time is 2 days before a daylight saving time gap that occurred
#' # on 2021-03-14 between 01:59:59 -> 03:00:00
#' from <- as.POSIXct("2021-03-12 02:30:00", "America/New_York")
#'
#' # So creating a daily sequence lands us in that daylight saving time gap,
#' # creating a nonexistent time
#' try(date_seq(from, by = duration_days(1), total_size = 5))
#'
#' # Resolve the nonexistent time with `nonexistent`. Note that this importantly
#' # allows times after the gap to retain the `02:30:00` time.
#' date_seq(from, by = duration_days(1), total_size = 5, nonexistent = "roll-forward")
#'
#' # Compare this to the base R behavior, where the hour is adjusted from 2->3
#' # as you cross the daylight saving time gap, and is never restored. This is
#' # equivalent to always using sys-time (rather than naive-time, like clock
#' # uses for daily sequences).
#' seq(from, by = "1 day", length.out = 5)
#'
#' # You can replicate this behavior by generating a second precision sequence
#' # of 86,400 seconds. Seconds always add in sys-time.
#' date_seq(from, by = duration_seconds(86400), total_size = 5)
#'
#' # ---------------------------------------------------------------------------
#'
#' # Usage of `to` and `total_size` must generate a non-fractional sequence
#' # between `from` and `to`
#' from <- date_time_build(2019, 1, 1, 0, 0, 0, zone = "America/New_York")
#' to <- date_time_build(2019, 1, 1, 0, 0, 3, zone = "America/New_York")
#'
#' # These are fine
#' date_seq(from, to = to, total_size = 2)
#' date_seq(from, to = to, total_size = 4)
#'
#' # But this is not!
#' try(date_seq(from, to = to, total_size = 3))
date_seq.POSIXt <- function(from,
                            ...,
                            to = NULL,
                            by = NULL,
                            total_size = NULL,
                            invalid = NULL,
                            nonexistent = NULL,
                            ambiguous = NULL) {
  check_dots_empty()

  check_number_of_supplied_optional_arguments(to, by, total_size)

  from <- to_posixct(from)
  zone <- date_zone(from)

  if (!is_null(to)) {
    if (!is_POSIXt(to)) {
      abort("If supplied, `to` must be a <POSIXct> or <POSIXlt>.")
    }

    to <- to_posixct(to)

    if (!identical(zone, date_zone(to))) {
      abort("`from` and `to` must have identical time zones.")
    }
  }

  if (!is_null(total_size)) {
    total_size <- check_length_out(total_size, arg = "total_size")
  }

  if (is_null(by)) {
    precision <- "second"
  } else if (is_duration(by)) {
    precision <- duration_precision(by)
  } else {
    precision <- "second"
    by <- duration_helper(by, PRECISION_SECOND, n_arg = "by")
  }

  precision_int <- validate_precision_string(precision)

  if (precision_int == PRECISION_QUARTER) {
    by <- duration_cast(by, "month")
    precision <- "month"
    precision_int <- PRECISION_MONTH
  }

  if (precision_int == PRECISION_WEEK) {
    by <- duration_cast(by, "day")
    precision <- "day"
    precision_int <- PRECISION_DAY
  }

  if (precision_int %in% c(PRECISION_YEAR, PRECISION_MONTH)) {
    out <- date_seq_year_month(from, to, by, total_size, precision)
    out <- invalid_resolve(out, invalid = invalid)
    out <- as.POSIXct(out, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
    return(out)
  }

  if (precision_int == PRECISION_DAY) {
    out <- date_seq_day(from, to, by, total_size, precision)
    out <- as.POSIXct(out, tz = zone, nonexistent = nonexistent, ambiguous = ambiguous)
    return(out)
  }

  if (precision_int %in% c(PRECISION_HOUR, PRECISION_MINUTE, PRECISION_SECOND)) {
    out <- date_seq_hour_minute_second(from, to, by, total_size, precision)
    out <- as.POSIXct(out, tz = zone)
    return(out)
  }

  abort("`by` must have a precision of 'year', 'quarter', 'month', 'week', 'day', 'hour', 'minute', or 'second'.")
}

# ------------------------------------------------------------------------------

#' Counting: date-times
#'
#' @description
#' This is a POSIXct/POSIXlt method for the [date_count_between()] generic.
#'
#' `date_count_between()` counts the number of `precision` units between
#' `start` and `end` (i.e., the number of years or months). This count
#' corresponds to the _whole number_ of units, and will never return a
#' fractional value.
#'
#' This is suitable for, say, computing the whole number of years or months
#' between two dates, accounting for the day of the month and the time of day.
#'
#' Internally, the date-time is converted to one of the following three clock
#' types, and the counting is done directly on that type. The choice of type is
#' based on the most common interpretation of each precision, but is ultimately
#' a heuristic. See the examples for more information.
#'
#' _Calendrical based counting:_
#'
#' These precisions convert to a year-month-day calendar and count while in that
#' type.
#'
#' - `"year"`
#'
#' - `"quarter"`
#'
#' - `"month"`
#'
#' _Naive-time based counting:_
#'
#' These precisions convert to a naive-time and count while in that type.
#'
#' - `"week"`
#'
#' - `"day"`
#'
#' _Sys-time based counting:_
#'
#' These precisions convert to a sys-time and count while in that type.
#'
#' - `"hour"`
#'
#' - `"minute"`
#'
#' - `"second"`
#'
#' @details
#' `"quarter"` is equivalent to `"month"` precision with `n` set to `n * 3L`.
#'
#' @inheritSection calendar_count_between Comparison Direction
#'
#' @inheritParams date_count_between
#'
#' @param start,end `[POSIXct / POSIXlt]`
#'
#'   A pair of date-time vectors. These will be recycled to their common
#'   size.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"quarter"`
#'   - `"month"`
#'   - `"week"`
#'   - `"day"`
#'   - `"hour"`
#'   - `"minute"`
#'   - `"second"`
#'
#' @inherit date_count_between return
#'
#' @name posixt-count-between
#'
#' @export
#' @examples
#' start <- date_time_parse("2000-05-05 02:00:00", zone = "America/New_York")
#' end <- date_time_parse(
#'   c("2020-05-05 01:00:00", "2020-05-05 03:00:00"),
#'   zone = "America/New_York"
#' )
#'
#' # Age in years
#' date_count_between(start, end, "year")
#'
#' # Number of "whole" months between these dates. i.e.
#' # `2000-05-05 02:00:00 -> 2020-04-05 02:00:00` is 239 months
#' # `2000-05-05 02:00:00 -> 2020-05-05 02:00:00` is 240 months
#' # Since `2020-05-05 01:00:00` occurs before the 2nd hour,
#' # it gets a count of 239
#' date_count_between(start, end, "month")
#'
#' # Number of seconds between
#' date_count_between(start, end, "second")
#'
#' # ---------------------------------------------------------------------------
#' # Naive-time VS Sys-time interpretation
#'
#' # The difference between whether `start` and `end` are converted to a
#' # naive-time vs a sys-time comes into play when dealing with daylight
#' # savings.
#'
#' # Here are two times around a 1 hour DST gap where clocks jumped from
#' # 01:59:59 -> 03:00:00
#' x <- date_time_build(1970, 4, 26, 1, 50, 00, zone = "America/New_York")
#' y <- date_time_build(1970, 4, 26, 3, 00, 00, zone = "America/New_York")
#'
#' # When treated like sys-times, these are considered to be 10 minutes apart,
#' # which is the amount of time that would have elapsed if you were watching
#' # a clock as it changed between these two times.
#' date_count_between(x, y, "minute")
#'
#' # Lets add a 3rd date that is ~1 day ahead of these
#' z <- date_time_build(1970, 4, 27, 1, 55, 00, zone = "America/New_York")
#'
#' # When treated like naive-times, `z` is considered to be at least 1 day ahead
#' # of `x`, because `01:55:00` is after `01:50:00`. This is probably what you
#' # expected.
#' date_count_between(x, z, "day")
#'
#' # If these were interpreted like sys-times, then `z` would not be considered
#' # to be 1 day ahead. That would look something like this:
#' date_count_between(x, z, "second")
#' trunc(date_count_between(x, z, "second") / 86400)
#'
#' # This is because there have only been 83,100 elapsed seconds since `x`,
#' # which isn't a full day's worth (86,400 seconds). But we'd generally
#' # consider `z` to be 1 day ahead of `x` (and ignore the DST gap), so that is
#' # how it is implemented.
#'
#' # You can override this by converting directly to sys-time, then using
#' # `time_point_count_between()`
#' x_st <- as_sys_time(x)
#' x_st
#'
#' z_st <- as_sys_time(z)
#' z_st
#'
#' time_point_count_between(x_st, z_st, "day")
date_count_between.POSIXt <- function(start, end, precision, ..., n = 1L) {
  check_dots_empty()

  if (!is_POSIXt(end)) {
    abort("`end` must be a <POSIXt>.")
  }

  start <- to_posixct(start)
  end <- to_posixct(end)

  start_zone <- date_zone(start)
  end_zone <- date_zone(end)

  if (!identical(start_zone, end_zone)) {
    start_zone <- zone_pretty(start_zone)
    end_zone <- zone_pretty(end_zone)

    abort(paste0(
      "`start` (", start_zone, ") and `end` (", end_zone, ") ",
      "must have identical time zones."
    ))
  }

  precision_int <- validate_precision_string(precision)

  # Designed to match `add_*()` functions to guarantee that
  # if `start <= end`, then `start + <count> <= end`
  allowed_precisions_calendar <- c(
    PRECISION_YEAR, PRECISION_QUARTER, PRECISION_MONTH
  )
  allowed_precisions_naive_time <- c(
    PRECISION_WEEK, PRECISION_DAY
  )
  allowed_precisions_sys_time <- c(
    PRECISION_HOUR, PRECISION_MINUTE, PRECISION_SECOND
  )

  date_count_between_impl(
    start = start,
    end = end,
    precision = precision,
    n = n,
    allowed_precisions_calendar = allowed_precisions_calendar,
    allowed_precisions_naive_time = allowed_precisions_naive_time,
    allowed_precisions_sys_time = allowed_precisions_sys_time
  )
}

#' Clock arithmetic
#'
#' @description
#' This is the landing page for all clock arithmetic functions. There are
#' specific sub-pages describing how arithmetic works for different calendars
#' and time points, which is where you should look for more information.
#'
#' Calendars are efficient at arithmetic with irregular units of time, such as
#' month, quarters, or years.
#'
#' - [year-month-day][year-month-day-arithmetic]
#'
#' - [year-month-weekday][year-month-weekday-arithmetic]
#'
#' - [year-quarter-day][year-quarter-day-arithmetic]
#'
#' - [iso-year-week-day][iso-year-week-day-arithmetic]
#'
#' - [year-day][year-day-arithmetic]
#'
#' Time points, such as naive-times and sys-times, are efficient at arithmetic
#' with regular, well-defined units of time, such as days, hours, seconds,
#' or nanoseconds.
#'
#' - [time-point][time-point-arithmetic]
#'
#' Durations can use any of these arithmetic functions, and return a new
#' duration with a precision corresponding to the common type of the
#' input and the function used.
#'
#' - [duration][duration-arithmetic]
#'
#' Weekdays can perform day-based circular arithmetic.
#'
#' - [weekday][weekday-arithmetic]
#'
#' There are also convenience methods for doing arithmetic directly on a
#' native R date or date-time type:
#'
#' - [dates (Date)][Date-arithmetic]
#'
#' - [date-times (POSIXct / POSIXlt)][posixt-arithmetic]
#'
#' @details
#' Months and years are considered "irregular" because some months have more
#' days then others (28, 29, 30, or 31), and some years have more days than
#' others (365 or 366).
#'
#' Days are considered "regular" because they are defined as 86,400 seconds.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[object]`
#'
#'   An object.
#'
#' @param n `[integer / clock_duration]`
#'
#'   An integer vector to be converted to a duration, or a duration
#'   corresponding to the arithmetic function being used. This corresponds
#'   to the number of duration units to add. `n` may be negative to subtract
#'   units of duration.
#'
#' @return `x` after performing the arithmetic.
#'
#' @name clock-arithmetic
#'
#' @examples
#' # See each sub-page for more specific examples
#' x <- year_month_day(2019, 2, 1)
#' add_months(x, 1)
NULL

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_years <- function(x, n, ...) {
  UseMethod("add_years")
}

#' @export
add_years.clock_time_point <- function(x, n, ...) {
  stop_clock_unsupported_time_point_op("add_years")
}

#' @export
add_years.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_years")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_quarters <- function(x, n, ...) {
  UseMethod("add_quarters")
}

#' @export
add_quarters.clock_time_point <- function(x, n, ...) {
  stop_clock_unsupported_time_point_op("add_quarters")
}

#' @export
add_quarters.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_quarters")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_months <- function(x, n, ...) {
  UseMethod("add_months")
}

#' @export
add_months.clock_time_point <- function(x, n, ...) {
  stop_clock_unsupported_time_point_op("add_months")
}

#' @export
add_months.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_months")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_weeks <- function(x, n, ...) {
  UseMethod("add_weeks")
}

#' @export
add_weeks.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_weeks")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_days <- function(x, n, ...) {
  UseMethod("add_days")
}

#' @export
add_days.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_days")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_hours <- function(x, n, ...) {
  UseMethod("add_hours")
}

#' @export
add_hours.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_hours")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_minutes <- function(x, n, ...) {
  UseMethod("add_minutes")
}

#' @export
add_minutes.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_minutes")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_seconds <- function(x, n, ...) {
  UseMethod("add_seconds")
}

#' @export
add_seconds.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_seconds")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_milliseconds <- function(x, n, ...) {
  UseMethod("add_milliseconds")
}

#' @export
add_milliseconds.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_milliseconds")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_microseconds <- function(x, n, ...) {
  UseMethod("add_microseconds")
}

#' @export
add_microseconds.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_microseconds")
}

# ------------------------------------------------------------------------------

#' @rdname clock-arithmetic
#' @export
add_nanoseconds <- function(x, n, ...) {
  UseMethod("add_nanoseconds")
}

#' @export
add_nanoseconds.clock_calendar <- function(x, n, ...) {
  stop_clock_unsupported_calendar_op("add_nanoseconds")
}

# ------------------------------------------------------------------------------

add_duration <- function(x, duration, ..., swapped = FALSE) {
  check_dots_empty()

  if (swapped) {
    # `duration` was LHS, so use names from it if applicable, making sure
    # that we only recycle everything once
    args <- vec_recycle_common(x = x, duration = duration)
    x <- args$x
    duration <- args$duration
    names(x) <- names_common(duration, x)
  }

  precision <- duration_precision_attribute(duration)
  precision <- precision_to_string(precision)

  switch (
    precision,
    year = add_years(x, duration),
    quarter = add_quarters(x, duration),
    month = add_months(x, duration),
    week = add_weeks(x, duration),
    day = add_days(x, duration),
    hour = add_hours(x, duration),
    minute = add_minutes(x, duration),
    second = add_seconds(x, duration),
    millisecond = add_milliseconds(x, duration),
    microsecond = add_microseconds(x, duration),
    nanosecond = add_nanoseconds(x, duration)
  )
}

# ------------------------------------------------------------------------------
# vec_proxy_compare()

# Really testing `vec_proxy()` and `vec_ptype2()` / `vec_cast()`

test_that("can't compare zoned-times with different zones", {
  x <- as_zoned_time(sys_days(0), "America/New_York")
  y <- as_zoned_time(sys_days(0), "UTC")

  expect_snapshot_error(x > y)
})

test_that("can compare zoned-times with same zone", {
  x <- as_zoned_time(sys_days(0:1), "America/New_York")

  expect_false(x[1] > x[2])
  expect_true(x[1] < x[2])
})

# ------------------------------------------------------------------------------
# print() / obj_print_data() / obj_print_footer()

test_that("normal print method works", {
  x <- as_zoned_time(as_sys_time(year_month_day(2019, 1:5, 1)), "America/New_York")
  expect_snapshot(x)
})

test_that("can limit with `max`", {
  x <- as_zoned_time(as_sys_time(year_month_day(2019, 1:5, 1)), "America/New_York")

  expect_snapshot(print(x, max = 2))
  expect_snapshot(print(x, max = 4))

  # no footer if length >= max
  expect_snapshot(print(x, max = 5))
  expect_snapshot(print(x, max = 6))
})

test_that("`max` defaults to `getOption('max.print')` but can be overridden", {
  local_options(max.print = 3)

  x <- as_zoned_time(as_sys_time(year_month_day(2019, 1:5, 1)), "America/New_York")

  expect_snapshot(x)
  expect_snapshot(print(x, max = 4))
  expect_snapshot(print(x, max = 5))
})

# ------------------------------------------------------------------------------
# as.character()

test_that("as.character() works", {
  expect <- "2019-01-01T01:02:03-05:00[America/New_York]"
  x <- zoned_time_parse_complete(expect)
  expect_identical(as.character(x), expect)
})

# ------------------------------------------------------------------------------
# zoned_time_parse_complete()

test_that("can parse what we format with seconds precision zoned time", {
  zone <- "America/New_York"

  x <- as_zoned_time(as_naive_time(year_month_day(2019, 1, 1)), zone)

  expect_identical(
    zoned_time_parse_complete(format(x)),
    x
  )
})

test_that("can parse subsecond zoned time", {
  zone <- "America/New_York"

  x <- "2019-01-01T01:02:03.123-05:00[America/New_York]"
  y <- "2019-01-01T01:02:03.1234-05:00[America/New_York]"
  z <- "2019-01-01T01:02:03.123456789-05:00[America/New_York]"

  expect_identical(
    zoned_time_parse_complete(x, precision = "millisecond"),
    as_zoned_time(as_naive_time(year_month_day(2019, 1, 1, 1, 2, 3, 123, subsecond_precision = "millisecond")), zone)
  )
  expect_identical(
    zoned_time_parse_complete(y, precision = "microsecond"),
    as_zoned_time(as_naive_time(year_month_day(2019, 1, 1, 1, 2, 3, 123400, subsecond_precision = "microsecond")), zone)
  )
  expect_identical(
    zoned_time_parse_complete(z, precision = "nanosecond"),
    as_zoned_time(as_naive_time(year_month_day(2019, 1, 1, 1, 2, 3, 123456789, subsecond_precision = "nanosecond")), zone)
  )
})

test_that("multiple formats can be used", {
  zone <- "America/New_York"

  x <- c(
    "1970-10-25 05:30:00-05:00[America/New_York]",
    "1970/10/25 05:30:00-05:00[America/New_York]"
  )

  formats <- c(
    "%Y-%m-%d %H:%M:%S%Ez[%Z]",
    "%Y/%m/%d %H:%M:%S%Ez[%Z]"
  )

  expect_identical(
    zoned_time_parse_complete(x, format = formats),
    as_zoned_time(
      as_naive_time(year_month_day(1970, 10, 25, 05, 30, c(00, 00))),
      zone
    )
  )
})

test_that("cannot parse nonexistent time", {
  zone <- "America/New_York"

  x <- "1970-04-26T02:30:00-05:00[America/New_York]"

  expect_warning(
    expect_identical(
      zoned_time_parse_complete(x),
      as_zoned_time(naive_seconds(NA), zone)
    )
  )

  expect_snapshot(zoned_time_parse_complete(x))
})

test_that("ambiguous times are resolved by the offset", {
  zone <- "America/New_York"

  x <- c(
    "1970-10-25T01:30:00-04:00[America/New_York]",
    "1970-10-25T01:30:00-05:00[America/New_York]"
  )

  expect_identical(
    zoned_time_parse_complete(x),
    as_zoned_time(
      as_naive_time(year_month_day(1970, 10, 25, 01, 30, c(00, 00))),
      zone,
      ambiguous = c("earliest", "latest")
    )
  )
})

test_that("offset must align with unique offset", {
  zone <- "America/New_York"

  # Should be `-05:00`
  x <- "2019-01-01T01:02:03-03:00[America/New_York]"

  expect_warning(
    expect_identical(
      zoned_time_parse_complete(x),
      as_zoned_time(naive_seconds(NA), zone)
    )
  )

  expect_snapshot(zoned_time_parse_complete(x))
})

test_that("offset must align with one of two possible ambiguous offsets", {
  zone <- "America/New_York"

  # Should be `-04:00` or `-05:00`
  x <- c(
    "1970-10-25T01:30:00-03:00[America/New_York]",
    "1970-10-25T01:30:00-06:00[America/New_York]"
  )

  expect_warning(
    expect_identical(
      zoned_time_parse_complete(x),
      as_zoned_time(naive_seconds(c(NA, NA)), zone)
    )
  )

  expect_snapshot(zoned_time_parse_complete(x))
})

test_that("cannot have differing zone names", {
  x <- c(
    "2019-01-01T01:02:03-05:00[America/New_York]",
    "2019-01-01T01:02:03-08:00[America/Los_Angeles]"
  )

  expect_snapshot_error(zoned_time_parse_complete(x))
})

test_that("zone name must be valid", {
  x <- "2019-01-01T01:02:03-05:00[America/New_Yor]"

  expect_snapshot_error(zoned_time_parse_complete(x))
})

test_that("empty input uses UTC time zone (#162)", {
  expect_identical(
    zoned_time_parse_complete(character()),
    as_zoned_time(naive_seconds(), "UTC")
  )
  expect_identical(
    zoned_time_parse_complete(character(), precision = "nanosecond"),
    as_zoned_time(as_naive_time(duration_nanoseconds()), "UTC")
  )
})

test_that("all `NA`s uses UTC time zone (#162)", {
  expect_identical(
    zoned_time_parse_complete(c(NA_character_, NA_character_)),
    as_zoned_time(naive_seconds(c(NA, NA)), "UTC")
  )
})

test_that("all failures uses UTC time zone (#162)", {
  expect_warning(
    expect_identical(
      zoned_time_parse_complete(c("foo", "bar")),
      as_zoned_time(naive_seconds(c(NA, NA)), "UTC")
    )
  )
})

test_that("`x` is translated to UTF-8", {
  x <- "2019-f\u00E9vrier-01 01:02:03-05:00[America/New_York]"
  x <- iconv(x, from = "UTF-8", to = "latin1")

  locale <- clock_locale("fr")
  format <- "%Y-%B-%d %H:%M:%S%Ez[%Z]"

  expect_identical(Encoding(x), "latin1")
  expect_identical(Encoding(locale$labels$month[2]), "UTF-8")

  expect_identical(
    zoned_time_parse_complete(x, format = format, locale = locale),
    as_zoned_time(as_naive_time(year_month_day(2019, 2, 1, 1, 2, 3)), "America/New_York")
  )
})

test_that("leftover subseconds result in a parse failure", {
  x <- "2019-01-01T01:01:01.1238-05:00[America/New_York]"

  # This is fine
  expect_identical(
    zoned_time_parse_complete(x, precision = "microsecond"),
    as_zoned_time(as_naive_time(year_month_day(2019, 1, 1, 1, 1, 1, 123800, subsecond_precision = "microsecond")), "America/New_York")
  )

  # This defaults to `%6S`, which parses `01.123` then stops,
  # leaving a `8` for `%z` to parse, resulting in a failure. Because everything
  # fails, we get a UTC time zone.
  expect_warning(
    expect_identical(
      zoned_time_parse_complete(x, precision = "millisecond"),
      as_zoned_time(naive_seconds(NA) + duration_milliseconds(NA), zone = "UTC")
    ),
    class = "clock_warning_parse_failures"
  )
})

test_that("parsing rounds parsed subsecond components more precise than the resulting container (#207) (#230) (undocumented)", {
  x <- "2019-01-01 01:01:01.1238-05:00[America/New_York]"

  # Requesting `%7S` parses the full `01.1238`, and the `1238` portion is rounded up
  expect_identical(
    zoned_time_parse_complete(x, precision = "millisecond", format = "%Y-%m-%d %H:%M:%7S%Ez[%Z]"),
    as_zoned_time(as_naive_time(year_month_day(2019, 1, 1, 1, 1, 1, 124, subsecond_precision = "millisecond")), "America/New_York")
  )
})

test_that("parsing fails when undocumented rounding behavior would result in invalid 60 second component (#230) (undocumented)", {
  x <- "2019-01-01 01:01:59.550-05:00[America/New_York]"

  # Requesting `%6S` parses the full `59.550`, which is immediately rounded to `60` which looks invalid.
  # The correct way to do this is to parse the milliseconds, then round.
  expect_warning(
    expect_identical(
      zoned_time_parse_complete(x, precision = "second", format = "%Y-%m-%d %H:%M:%6S%Ez[%Z]"),
      as_zoned_time(as_naive_time(year_month_day(NA, NA, NA, NA, NA, NA)), zone = "UTC")
    ),
    class = "clock_warning_parse_failures"
  )
})

# ------------------------------------------------------------------------------
# zoned_time_parse_abbrev()

test_that("can parse with abbreviation and zone name", {
  expect_identical(
    zoned_time_parse_abbrev("2019-01-01 01:02:03 EST", "America/New_York"),
    zoned_time_parse_complete("2019-01-01T01:02:03-05:00[America/New_York]")
  )
})

test_that("can parse when abbreviation is an offset", {
  expect_identical(
    zoned_time_parse_abbrev("2019-01-01 01:02:03 +11", "Australia/Lord_Howe"),
    zoned_time_parse_complete("2019-01-01T01:02:03+11:00[Australia/Lord_Howe]")
  )
  expect_identical(
    zoned_time_parse_abbrev("2019-10-01 01:02:03 +1030", "Australia/Lord_Howe"),
    zoned_time_parse_complete("2019-10-01T01:02:03+10:30[Australia/Lord_Howe]")
  )
})

test_that("can parse at more precise precisions", {
  expect_identical(
    zoned_time_parse_abbrev("2019-01-01 01:02:03.123 EST", "America/New_York", precision = "millisecond"),
    as_zoned_time(as_naive_time(year_month_day(2019, 1, 1, 1, 2, 3, 123, subsecond_precision = "millisecond")), "America/New_York")
  )
  expect_identical(
    zoned_time_parse_abbrev("2019-01-01 01:02:03.123456 EST", "America/New_York", precision = "nanosecond"),
    as_zoned_time(as_naive_time(year_month_day(2019, 1, 1, 1, 2, 3, 123456000, subsecond_precision = "nanosecond")), "America/New_York")
  )
})

test_that("abbreviation is used to resolve ambiguity", {
  x <- c(
    "1970-10-25 01:30:00 EDT",
    "1970-10-25 01:30:00 EST"
  )

  expect <- c(
    "1970-10-25T01:30:00-04:00[America/New_York]",
    "1970-10-25T01:30:00-05:00[America/New_York]"
  )

  expect_identical(
    zoned_time_parse_abbrev(x, "America/New_York"),
    zoned_time_parse_complete(expect)
  )
})

test_that("nonexistent times are NAs", {
  expect_warning(
    expect_identical(
      zoned_time_parse_abbrev("1970-04-26 02:30:00 EST", "America/New_York"),
      as_zoned_time(sys_seconds(NA), "America/New_York")
    )
  )
})

test_that("abbreviation must match the one implied from naive + time zone name lookup", {
  x <- "1970-01-01 00:00:00 FOOBAR"

  expect_warning(
    expect_identical(
      zoned_time_parse_abbrev(x, "America/New_York"),
      as_zoned_time(sys_days(NA), "America/New_York")
    )
  )

  # Should be EST
  x <- "1970-01-01 00:00:00 EDT"

  expect_warning(
    expect_identical(
      zoned_time_parse_abbrev(x, "America/New_York"),
      as_zoned_time(sys_days(NA), "America/New_York")
    )
  )

  expect_snapshot(zoned_time_parse_abbrev(x, "America/New_York"))
})

test_that("%Z must be used", {
  x <- "1970-01-01"

  expect_snapshot_error(
    zoned_time_parse_abbrev(x, "America/New_York", format = "%Y-%m-%d")
  )
})

test_that("%z can be parsed (but is ignored really)", {
  expect <- zoned_time_parse_complete("1970-01-01T00:00:00-05:00[America/New_York]")
  x <- "1970-01-01 00:00:00-05:00 EST"

  expect_identical(
    zoned_time_parse_abbrev(x, "America/New_York", format = "%Y-%m-%d %H:%M:%S%Ez %Z"),
    expect
  )
})

test_that("%z that is incorrect technically slips through unnoticed", {
  expect <- zoned_time_parse_complete("1970-01-01T00:00:00-05:00[America/New_York]")
  x <- "1970-01-01 00:00:00-02:00 EST"

  expect_identical(
    zoned_time_parse_abbrev(x, "America/New_York", format = "%Y-%m-%d %H:%M:%S%Ez %Z"),
    expect
  )
})

test_that("%z must parse correctly if included", {
  expect <- as_zoned_time(sys_days(NA), "America/New_York")
  x <- "1970-01-01 00:00:00-0a:00 EST"

  expect_warning(
    result <- zoned_time_parse_abbrev(x, "America/New_York", format = "%Y-%m-%d %H:%M:%S%Ez %Z")
  )

  expect_identical(result, expect)
})

test_that("multiple formats can be attempted", {
  x <- c("1970-01-01 EST", "1970-01-01 05:06:07 EST")
  formats <- c("%Y-%m-%d %H:%M:%S %Z", "%Y-%m-%d %Z")

  expect <- as_zoned_time(
    as_naive_time(year_month_day(1970, 1, 1, c(0, 5), c(0, 6), c(0, 7))),
    "America/New_York"
  )

  expect_identical(
    zoned_time_parse_abbrev(x, "America/New_York", format = formats),
    expect
  )
})

test_that("NA parses correctly", {
  expect_identical(
    zoned_time_parse_abbrev(NA_character_, "America/New_York"),
    as_zoned_time(sys_seconds(NA), "America/New_York")
  )
  expect_identical(
    zoned_time_parse_abbrev(NA_character_, "America/New_York", precision = "nanosecond"),
    as_zoned_time(as_sys_time(duration_nanoseconds(NA)), "America/New_York")
  )
})

test_that("`x` is translated to UTF-8", {
  x <- "2019-f\u00E9vrier-01 01:02:03-05:00[EST]"
  x <- iconv(x, from = "UTF-8", to = "latin1")

  locale <- clock_locale("fr")
  format <- "%Y-%B-%d %H:%M:%S%Ez[%Z]"

  expect_identical(Encoding(x), "latin1")
  expect_identical(Encoding(locale$labels$month[2]), "UTF-8")

  expect_identical(
    zoned_time_parse_abbrev(x, "America/New_York", format = format, locale = locale),
    as_zoned_time(as_naive_time(year_month_day(2019, 2, 1, 1, 2, 3)), "America/New_York")
  )
})

# ------------------------------------------------------------------------------
# add_*()

test_that("zoned-times don't support arithmetic", {
  x <- as_zoned_time(as_naive_time(year_month_day(2019, 1, 1)), "America/New_York")

  expect_snapshot_error(add_years(x, 1))
  expect_snapshot_error(add_quarters(x, 1))
  expect_snapshot_error(add_months(x, 1))
  expect_snapshot_error(add_weeks(x, 1))
  expect_snapshot_error(add_days(x, 1))
  expect_snapshot_error(add_hours(x, 1))
  expect_snapshot_error(add_minutes(x, 1))
  expect_snapshot_error(add_seconds(x, 1))
  expect_snapshot_error(add_milliseconds(x, 1))
  expect_snapshot_error(add_microseconds(x, 1))
  expect_snapshot_error(add_nanoseconds(x, 1))
})

# ------------------------------------------------------------------------------
# vec_ptype()

test_that("ptype is correct", {
  zones <- c("UTC", "America/New_York", "")

  for (zone in zones) {
    for (precision in precision_names()) {
      precision <- validate_precision_string(precision)

      if (precision < PRECISION_SECOND) {
        next
      }

      x <- duration_helper(0L, precision)
      x <- as_zoned_time(as_naive_time(x), zone)

      ptype <- duration_helper(integer(), precision)
      ptype <- as_zoned_time(as_naive_time(ptype), zone)

      expect_identical(vec_ptype(x), ptype)
    }
  }
})

# ------------------------------------------------------------------------------
# vec_math()

test_that("is.nan() works", {
  zone <- "America/New_York"
  x <- as_zoned_time(naive_days(c(2019, NA)), zone)
  expect_identical(is.nan(x), c(FALSE, FALSE))
})

test_that("is.finite() works", {
  zone <- "America/New_York"
  x <- as_zoned_time(naive_days(c(2019, NA)), zone)
  expect_identical(is.finite(x), c(TRUE, FALSE))
})

test_that("is.infinite() works", {
  zone <- "America/New_York"
  x <- as_zoned_time(naive_days(c(2019, NA)), zone)
  expect_identical(is.infinite(x), c(FALSE, FALSE))
})

# ------------------------------------------------------------------------------
# zoned_time_precision()

test_that("precision: can get the precision", {
  zone <- "America/New_York"
  expect_identical(zoned_time_precision(as_zoned_time(as_naive_time(duration_seconds(2:5)), zone)), "second")
  expect_identical(zoned_time_precision(as_zoned_time(as_naive_time(duration_nanoseconds(2:5)), zone)), "nanosecond")
})

test_that("precision: can only be called on zoned-times", {
  expect_snapshot_error(zoned_time_precision(duration_days()))
})

is_time_point <- function(x) {
  inherits(x, "clock_time_point")
}

# ------------------------------------------------------------------------------

time_point_clock_attribute <- function(x) {
  attr(x, "clock", exact = TRUE)
}

time_point_precision_attribute <- function(x) {
  attr(x, "precision", exact = TRUE)
}

time_point_duration <- function(x, retain_names = FALSE) {
  if (retain_names) {
    names <- clock_rcrd_names(x)
  } else {
    names <- NULL
  }
  precision <- time_point_precision_attribute(x)
  new_duration_from_fields(x, precision, names)
}

# ------------------------------------------------------------------------------

#' @export
format.clock_time_point <- function(x,
                                    ...,
                                    format = NULL,
                                    locale = clock_locale()) {
  if (!is_clock_locale(locale)) {
    abort("`locale` must be a 'clock_locale' object.")
  }

  clock <- time_point_clock_attribute(x)
  precision <- time_point_precision_attribute(x)

  if (is_null(format)) {
    format <- time_point_precision_format(precision)
  }

  labels <- locale$labels
  decimal_mark <- locale$decimal_mark

  out <- format_time_point_cpp(
    fields = x,
    clock = clock,
    format = format,
    precision_int = precision,
    month = labels$month,
    month_abbrev = labels$month_abbrev,
    weekday = labels$weekday,
    weekday_abbrev = labels$weekday_abbrev,
    am_pm = labels$am_pm,
    decimal_mark = decimal_mark
  )

  names(out) <- clock_rcrd_names(x)

  out
}

time_point_precision_format <- function(precision) {
  precision <- precision_to_string(precision)

  switch(
    precision,
    day = "%Y-%m-%d",
    hour = "%Y-%m-%dT%H",
    minute = "%Y-%m-%dT%H:%M",
    second = "%Y-%m-%dT%H:%M:%S",
    millisecond = "%Y-%m-%dT%H:%M:%S",
    microsecond = "%Y-%m-%dT%H:%M:%S",
    nanosecond = "%Y-%m-%dT%H:%M:%S",
    abort("Unknown precision.")
  )
}

# ------------------------------------------------------------------------------

time_point_parse <- function(x,
                             ...,
                             format,
                             precision,
                             locale,
                             clock) {
  check_dots_empty()

  if (is_null(format)) {
    format <- time_point_precision_format(precision)
  }

  if (!is_clock_locale(locale)) {
    abort("`locale` must be a 'clock_locale' object.")
  }

  labels <- locale$labels
  mark <- locale$decimal_mark

  time_point_parse_cpp(
    x,
    format,
    precision,
    clock,
    labels$month,
    labels$month_abbrev,
    labels$weekday,
    labels$weekday_abbrev,
    labels$am_pm,
    mark
  )
}

# ------------------------------------------------------------------------------

#' @export
print.clock_time_point <- function(x, ..., max = NULL) {
  clock_print(x, max)
}

# - Each subclass implements a `format()` method
# - Unlike vctrs, don't use `print(quote = FALSE)` since we want to match base R
#' @export
obj_print_data.clock_time_point <- function(x, ..., max) {
  if (vec_is_empty(x)) {
    return(invisible(x))
  }

  x <- max_slice(x, max)

  out <- format(x)

  # Pass `max` to avoid base R's default footer
  print(out, max = max)

  invisible(x)
}

#' @export
obj_print_footer.clock_time_point <- function(x, ..., max) {
  clock_print_footer(x, max)
}

# Align left to match pillar_shaft.Date
# @export - lazy in .onLoad()
pillar_shaft.clock_time_point <- function(x, ...) {
  out <- format(x)
  pillar::new_pillar_shaft_simple(out, align = "left")
}

# ------------------------------------------------------------------------------

#' @export
vec_proxy.clock_time_point <- function(x, ...) {
  .Call(`_clock_clock_rcrd_proxy`, x)
}

#' @export
vec_restore.clock_time_point <- function(x, to, ...) {
  .Call(`_clock_time_point_restore`, x, to)
}

# ------------------------------------------------------------------------------

#' @export
vec_ptype_full.clock_time_point <- function(x, ...) {
  clock <- time_point_clock_attribute(x)
  clock <- clock_to_string(clock)
  precision <- time_point_precision_attribute(x)
  precision <- precision_to_string(precision)
  paste0("time_point<", clock, "><", precision, ">")
}

#' @export
vec_ptype_abbr.clock_time_point <- function(x, ...) {
  clock <- time_point_clock_attribute(x)
  clock <- clock_to_string(clock)
  precision <- time_point_precision_attribute(x)
  precision <- precision_to_string(precision)
  precision <- precision_abbr(precision)
  paste0("tp<", clock, "><", precision, ">")
}

# ------------------------------------------------------------------------------

# Caller guarantees that clocks are identical
ptype2_time_point_and_time_point <- function(x, y, ...) {
  if (time_point_precision_attribute(x) >= time_point_precision_attribute(y)) {
    x
  } else {
    y
  }
}

# Caller guarantees that clocks are identical
cast_time_point_to_time_point <- function(x, to, ...) {
  x_precision <- time_point_precision_attribute(x)
  to_precision <- time_point_precision_attribute(to)

  if (x_precision == to_precision) {
    return(x)
  }

  if (x_precision > to_precision) {
    stop_incompatible_cast(x, to, ..., details = "Can't cast to a less precise precision.")
  }

  fields <- duration_cast_cpp(x, x_precision, to_precision)

  names <- clock_rcrd_names(x)
  clock <- time_point_clock_attribute(x)

  new_time_point_from_fields(fields, to_precision, clock, names)
}

# ------------------------------------------------------------------------------

arith_time_point_and_missing <- function(op, x, y, ...) {
  switch (
    op,
    "+" = x,
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_time_point_and_time_point <- function(op, x, y, ...) {
  switch (
    op,
    "-" = time_point_minus_time_point(x, y, names_common(x, y)),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_time_point_and_duration <- function(op, x, y, ...) {
  switch (
    op,
    "+" = time_point_plus_duration(x, y, duration_precision_attribute(y), names_common(x, y)),
    "-" = time_point_minus_duration(x, y, duration_precision_attribute(y), names_common(x, y)),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_duration_and_time_point <- function(op, x, y, ...) {
  switch (
    op,
    "+" = time_point_plus_duration(y, x, duration_precision_attribute(x), names_common(x, y)),
    "-" = stop_incompatible_op(op, x, y, details = "Can't subtract a time point from a duration.", ...),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_time_point_and_numeric <- function(op, x, y, ...) {
  precision <- time_point_precision_attribute(x)

  switch (
    op,
    "+" = time_point_plus_duration(x, y, precision, names_common(x, y)),
    "-" = time_point_minus_duration(x, y, precision, names_common(x, y)),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_numeric_and_time_point <- function(op, x, y, ...) {
  precision <- time_point_precision_attribute(y)

  switch (
    op,
    "+" = time_point_plus_duration(y, x, precision, names_common(x, y)),
    "-" = stop_incompatible_op(op, x, y, details = "Can't subtract a time point from a duration.", ...),
    stop_incompatible_op(op, x, y, ...)
  )
}

# ------------------------------------------------------------------------------

#' Arithmetic: Time points
#'
#' @description
#' These are naive-time and sys-time methods for the
#' [arithmetic generics][clock-arithmetic].
#'
#' - `add_weeks()`
#'
#' - `add_days()`
#'
#' - `add_hours()`
#'
#' - `add_minutes()`
#'
#' - `add_seconds()`
#'
#' - `add_milliseconds()`
#'
#' - `add_microseconds()`
#'
#' - `add_nanoseconds()`
#'
#' When working with zoned times, generally you convert to either sys-time
#' or naive-time, add the duration, then convert back to zoned time. Typically,
#' _weeks and days_ are added in _naive-time_, and _hours, minutes, seconds,
#' and subseconds_ are added in _sys-time_.
#'
#' If you aren't using zoned times, arithmetic on sys-times and naive-time
#' is equivalent.
#'
#' If you need to add larger irregular units of time, such as months, quarters,
#' or years, convert to a calendar type with a converter like
#' [as_year_month_day()].
#'
#' @details
#' `x` and `n` are recycled against each other.
#'
#' @inheritParams add_years
#'
#' @param x `[clock_sys_time / clock_naive_time]`
#'
#'   A time point vector.
#'
#' @return `x` after performing the arithmetic.
#'
#' @name time-point-arithmetic
#'
#' @examples
#' library(magrittr)
#'
#' # Say you started with this zoned time, and you want to add 1 day to it
#' x <- as_naive_time(year_month_day(1970, 04, 25, 02, 30, 00))
#' x <- as_zoned_time(x, "America/New_York")
#' x
#'
#' # Note that there was a daylight saving time gap on 1970-04-26 where
#' # we jumped from 01:59:59 -> 03:00:00.
#'
#' # You can choose to add 1 day in "system time", by first converting to
#' # sys-time (the equivalent UTC time), adding the day, then converting back to
#' # zoned time. If you sat still for exactly 86,400 seconds, this is the
#' # time that you would see after daylight saving time adjusted the clock
#' # (note that the hour field is shifted forward by the size of the gap)
#' as_sys_time(x)
#'
#' x %>%
#'   as_sys_time() %>%
#'   add_days(1) %>%
#'   as_zoned_time(zoned_time_zone(x))
#'
#' # Alternatively, you can add 1 day in "naive time". Naive time represents
#' # a clock time with a yet-to-be-specified time zone. It tries to maintain
#' # smaller units where possible, so adding 1 day would attempt to return
#' # "1970-04-26T02:30:00" in the America/New_York time zone, but...
#' as_naive_time(x)
#'
#' try({
#' x %>%
#'   as_naive_time() %>%
#'   add_days(1) %>%
#'   as_zoned_time(zoned_time_zone(x))
#' })
#'
#' # ...this time doesn't exist in that time zone! It is "nonexistent".
#' # You can resolve nonexistent times by setting the `nonexistent` argument
#' # when converting to zoned time. Let's roll forward to the next available
#' # moment in time.
#' x %>%
#'   as_naive_time() %>%
#'   add_days(1) %>%
#'   as_zoned_time(zoned_time_zone(x), nonexistent = "roll-forward")
NULL

#' @rdname time-point-arithmetic
#' @export
add_weeks.clock_time_point <- function(x, n, ...) {
  time_point_plus_duration(x, n, PRECISION_WEEK, names_common(x, n))
}

#' @rdname time-point-arithmetic
#' @export
add_days.clock_time_point <- function(x, n, ...) {
  time_point_plus_duration(x, n, PRECISION_DAY, names_common(x, n))
}

#' @rdname time-point-arithmetic
#' @export
add_hours.clock_time_point <- function(x, n, ...) {
  time_point_plus_duration(x, n, PRECISION_HOUR, names_common(x, n))
}

#' @rdname time-point-arithmetic
#' @export
add_minutes.clock_time_point <- function(x, n, ...) {
  time_point_plus_duration(x, n, PRECISION_MINUTE, names_common(x, n))
}

#' @rdname time-point-arithmetic
#' @export
add_seconds.clock_time_point <- function(x, n, ...) {
  time_point_plus_duration(x, n, PRECISION_SECOND, names_common(x, n))
}

#' @rdname time-point-arithmetic
#' @export
add_milliseconds.clock_time_point <- function(x, n, ...) {
  time_point_plus_duration(x, n, PRECISION_MILLISECOND, names_common(x, n))
}

#' @rdname time-point-arithmetic
#' @export
add_microseconds.clock_time_point <- function(x, n, ...) {
  time_point_plus_duration(x, n, PRECISION_MICROSECOND, names_common(x, n))
}

#' @rdname time-point-arithmetic
#' @export
add_nanoseconds.clock_time_point <- function(x, n, ...) {
  time_point_plus_duration(x, n, PRECISION_NANOSECOND, names_common(x, n))
}

time_point_plus_duration <- function(x, n, precision_n, names) {
  time_point_arith_duration(x, n, precision_n, names, duration_plus)
}
time_point_minus_duration <- function(x, n, precision_n, names) {
  time_point_arith_duration(x, n, precision_n, names, duration_minus)
}

time_point_arith_duration <- function(x, n, precision_n, names, duration_fn) {
  clock <- time_point_clock_attribute(x)
  x <- time_point_duration(x)

  n <- duration_collect_n(n, precision_n)

  # Handles recycling and casting
  duration <- duration_fn(x = x, y = n, names = names)

  names <- clock_rcrd_names(duration)
  precision <- duration_precision_attribute(duration)

  new_time_point_from_fields(duration, precision, clock, names)
}

time_point_minus_time_point <- function(x, y, names) {
  args <- vec_recycle_common(x = x, y = y, names = names)
  x <- args$x
  y <- args$y
  names <- args$names

  x_duration <- time_point_duration(x)
  y_duration <- time_point_duration(y)

  duration_minus(x = x_duration, y = y_duration, names = names)
}

# ------------------------------------------------------------------------------

#' @export
as_duration.clock_time_point <- function(x) {
  time_point_duration(x, retain_names = TRUE)
}

#' @export
as_year_month_day.clock_time_point <- function(x) {
  precision <- time_point_precision_attribute(x)
  fields <- as_year_month_day_from_sys_time_cpp(x, precision)
  new_year_month_day_from_fields(fields, precision, names = names(x))
}

#' @export
as_year_month_weekday.clock_time_point <- function(x) {
  precision <- time_point_precision_attribute(x)
  fields <- as_year_month_weekday_from_sys_time_cpp(x, precision)
  new_year_month_weekday_from_fields(fields, precision, names = names(x))
}

#' @export
as_year_quarter_day.clock_time_point <- function(x, ..., start = NULL) {
  check_dots_empty()
  precision <- time_point_precision_attribute(x)
  start <- quarterly_validate_start(start)
  fields <- as_year_quarter_day_from_sys_time_cpp(x, precision, start)
  new_year_quarter_day_from_fields(fields, precision, start, names = names(x))
}

#' @export
as_iso_year_week_day.clock_time_point <- function(x) {
  precision <- time_point_precision_attribute(x)
  fields <- as_iso_year_week_day_from_sys_time_cpp(x, precision)
  new_iso_year_week_day_from_fields(fields, precision, names = names(x))
}

#' @export
as_year_day.clock_time_point <- function(x) {
  precision <- time_point_precision_attribute(x)
  fields <- as_year_day_from_sys_time_cpp(x, precision)
  new_year_day_from_fields(fields, precision, names = names(x))
}

#' @export
as_weekday.clock_time_point <- function(x) {
  x <- time_point_cast(x, "day")
  day <- weekday_from_time_point_cpp(x)
  names(day) <- clock_rcrd_names(x)
  new_weekday(day)
}

# ------------------------------------------------------------------------------

#' Cast a time point between precisions
#'
#' @description
#' Casting is one way to change a time point's precision.
#'
#' Casting to a less precise precision will completely drop information that
#' is more precise than the precision that you are casting to. It does so
#' in a way that makes it round towards zero. When converting time points
#' to a less precise precision, you often want [time_point_floor()] instead
#' of `time_point_cast()`, as that handles pre-1970 dates (which are
#' stored as negative durations) in a more intuitive manner.
#'
#' Casting to a more precise precision is done through a multiplication by
#' a conversion factor between the current precision and the new precision.
#'
#' @param x `[clock_sys_time / clock_naive_time]`
#'
#'   A sys-time or naive-time.
#'
#' @param precision `[character(1)]`
#'
#'   A time point precision. One of:
#'
#'   - `"day"`
#'
#'   - `"hour"`
#'
#'   - `"minute"`
#'
#'   - `"second"`
#'
#'   - `"millisecond"`
#'
#'   - `"microsecond"`
#'
#'   - `"nanosecond"`
#'
#' @return `x` cast to the new `precision`.
#'
#' @export
#' @examples
#' # Hour precision time points
#' # One is pre-1970, one is post-1970
#' x <- duration_hours(c(25, -25))
#' x <- as_naive_time(x)
#' x
#'
#' # Casting rounds the underlying duration towards 0
#' cast <- time_point_cast(x, "day")
#' cast
#'
#' # Flooring rounds the underlying duration towards negative infinity,
#' # which is often more intuitive for time points.
#' # Note that the cast ends up rounding the pre-1970 date up to the next
#' # day, while the post-1970 date is rounded down.
#' floor <- time_point_floor(x, "day")
#' floor
#'
#' # Casting to a more precise precision, hour->millisecond
#' time_point_cast(x, "millisecond")
time_point_cast <- function(x, precision) {
  if (!is_time_point(x)) {
    abort("`x` must be a 'time_point'.")
  }

  x_precision <- time_point_precision_attribute(x)
  precision <- validate_time_point_precision_string(precision)

  fields <- duration_cast_cpp(x, x_precision, precision)

  names <- clock_rcrd_names(x)
  clock <- time_point_clock_attribute(x)

  new_time_point_from_fields(fields, precision, clock, names)
}

#' Time point rounding
#'
#' @description
#' - `time_point_floor()` rounds a sys-time or naive-time down to a multiple of
#'   the specified `precision`.
#'
#' - `time_point_ceiling()` rounds a sys-time or naive-time up to a multiple of
#'   the specified `precision`.
#'
#' - `time_point_round()` rounds up or down depending on what is closer,
#'   rounding up on ties.
#'
#' Rounding time points is mainly useful for rounding sub-daily time points
#' up to daily time points.
#'
#' It can also be useful for flooring by a set number of days (like 20) with
#' respect to some origin. By default, the origin is 1970-01-01 00:00:00.
#'
#' If you want to group by components, such as "day of the month", rather than
#' by "n days", see [calendar_group()].
#'
#' @section Boundary Handling:
#'
#' To understand how flooring and ceiling work, you need to know how they
#' create their intervals for rounding.
#'
#' - `time_point_floor()` constructs intervals of \code{[lower, upper)} that
#'   bound each element of `x`, then always chooses the _left-hand side_.
#'
#' - `time_point_ceiling()` constructs intervals of \code{(lower, upper]} that
#'   bound each element of `x`, then always chooses the _right-hand side_.
#'
#' As an easy example, consider 2020-01-02 00:00:05.
#'
#' To floor this to the nearest day, the following interval is constructed,
#' and the left-hand side is returned at day precision:
#'
#' \code{[2020-01-02 00:00:00, 2020-01-03 00:00:00)}
#'
#' To ceiling this to the nearest day, the following interval
#' is constructed, and the right-hand side is returned at day precision:
#'
#' \code{(2020-01-02 00:00:00, 2020-01-03 00:00:00]}
#'
#' Here is another example, this time with a time point on a boundary,
#' 2020-01-02 00:00:00.
#'
#' To floor this to the nearest day, the following interval is constructed,
#' and the left-hand side is returned at day precision:
#'
#' \code{[2020-01-02 00:00:00, 2020-01-03 00:00:00)}
#'
#' To ceiling this to the nearest day, the following interval
#' is constructed, and the right-hand side is returned at day precision:
#'
#' \code{(2020-01-01 00:00:00, 2020-01-02 00:00:00]}
#'
#' Notice that, regardless of whether you are doing a floor or ceiling, if
#' the input falls on a boundary then it will be returned as is.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams time_point_cast
#'
#' @param n `[positive integer(1)]`
#'
#'   A positive integer specifying the multiple of `precision` to use.
#'
#' @param origin `[clock_sys_time(1) / clock_naive_time(1) / NULL]`
#'
#'   An origin to begin counting from. Mostly useful when `n > 1` and you
#'   want to control how the rounding groups are created.
#'
#'   If `x` is a sys-time, `origin` must be a sys-time.
#'
#'   If `x` is a naive-time, `origin` must be a naive-time.
#'
#'   The precision of `origin` must be equally precise as or less
#'   precise than `precision`.
#'
#'   If `NULL`, a default origin of midnight on 1970-01-01 is used.
#'
#' @return `x` rounded to the new `precision`.
#'
#' @name time-point-rounding
#'
#' @examples
#' library(magrittr)
#'
#' x <- as_naive_time(year_month_day(2019, 01, 01))
#' x <- add_days(x, 0:40)
#' head(x)
#'
#' # Floor by sets of 20 days
#' # The implicit origin to start the 20 day counter is 1970-01-01
#' time_point_floor(x, "day", n = 20)
#'
#' # You can easily customize the origin by creating a duration out of the
#' # origin date of interest...
#' origin <- year_month_day(2019, 01, 01) %>%
#'   as_naive_time() %>%
#'   as_duration()
#'
#' # Which you can subtract from, floor, and then add to your input
#' time_point_floor(x - origin, "day", n = 20) + origin
#'
#' # For times on the boundary, floor and ceiling both return the input
#' # at the new precision. Notice how the first element is on the boundary,
#' # and the second is 1 second after the boundary.
#' y <- as_naive_time(year_month_day(2020, 01, 02, 00, 00, c(00, 01)))
#' time_point_floor(y, "day")
#' time_point_ceiling(y, "day")
NULL

#' @rdname time-point-rounding
#' @export
time_point_floor <- function(x, precision, ..., n = 1L, origin = NULL) {
  time_point_rounder(x, precision, n, origin, duration_floor, ...)
}
#' @rdname time-point-rounding
#' @export
time_point_ceiling <- function(x, precision, ..., n = 1L, origin = NULL) {
  time_point_rounder(x, precision, n, origin, duration_ceiling, ...)
}
#' @rdname time-point-rounding
#' @export
time_point_round <- function(x, precision, ..., n = 1L, origin = NULL) {
  time_point_rounder(x, precision, n, origin, duration_round, ...)
}

time_point_rounder <- function(x, precision, n, origin, duration_rounder, ...) {
  check_dots_empty()

  if (!is_time_point(x)) {
    abort("`x` must be a 'time_point'.")
  }

  precision_string <- precision
  precision <- validate_time_point_precision_string(precision)

  duration <- time_point_duration(x)

  has_origin <- !is_null(origin)

  if (has_origin) {
    origin <- collect_time_point_rounder_origin(origin, x, precision)
    duration <- duration - origin
  }

  duration <- duration_rounder(duration, precision_string, n = n)

  if (has_origin) {
    duration <- duration + origin
  }

  names <- clock_rcrd_names(x)
  clock <- time_point_clock_attribute(x)

  new_time_point_from_fields(duration, precision, clock, names)
}

collect_time_point_rounder_origin <- function(origin, x, precision) {
  # Cast `origin` to a time point with the same clock as `x`,
  # but with a precision of `precision`
  to_names <- NULL
  to <- duration_helper(integer(), precision)
  to <- new_time_point_from_fields(to, precision, time_point_clock_attribute(x), to_names)

  origin <- vec_cast(origin, to, x_arg = "origin")

  if (vec_size(origin) != 1L) {
    abort("`origin` must have length 1.")
  }
  if (is.na(origin)) {
    abort("`origin` must not be `NA`.")
  }

  origin <- as_duration(origin)

  origin
}

# ------------------------------------------------------------------------------

#' Shifting: time point
#'
#' @description
#' `time_point_shift()` shifts `x` to the `target` weekday. You can
#' shift to the next or previous weekday. If `x` is currently on the `target`
#' weekday, you can choose to leave it alone or advance it to the next instance
#' of the `target`.
#'
#' Weekday shifting is one of the easiest ways to floor by week while
#' controlling what is considered the first day of the week. You can also
#' accomplish this with the `origin` argument of [time_point_floor()], but
#' this is slightly easier.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[clock_time_point]`
#'
#'   A time point.
#'
#' @param target `[weekday]`
#'
#'   A weekday created from [weekday()] to target.
#'
#'   Generally this is length 1, but can also be the same length as `x`.
#'
#' @param which `[character(1)]`
#'
#'   One of:
#'
#'   - `"next"`: Shift to the next instance of the `target` weekday.
#'
#'   - `"previous`: Shift to the previous instance of the `target` weekday.
#'
#' @param boundary `[character(1)]`
#'
#'   One of:
#'
#'   - `"keep"`: If `x` is currently on the `target` weekday, return it.
#'
#'   - `"advance"`: If `x` is currently on the `target` weekday, advance it
#'   anyways.
#'
#' @return `x` shifted to the `target` weekday.
#'
#' @export
#' @examples
#' x <- as_naive_time(year_month_day(2019, 1, 1:2))
#'
#' # A Tuesday and Wednesday
#' as_weekday(x)
#'
#' monday <- weekday(clock_weekdays$monday)
#'
#' # Shift to the next Monday
#' time_point_shift(x, monday)
#'
#' # Shift to the previous Monday
#' # This is an easy way to "floor by week" with a target weekday in mind
#' time_point_shift(x, monday, which = "previous")
#'
#' # What about Tuesday?
#' tuesday <- weekday(clock_weekdays$tuesday)
#'
#' # Notice that the day that was currently on a Tuesday was not shifted
#' time_point_shift(x, tuesday)
#'
#' # You can force it to `"advance"`
#' time_point_shift(x, tuesday, boundary = "advance")
time_point_shift <- function(x,
                             target,
                             ...,
                             which = "next",
                             boundary = "keep") {
  check_dots_empty()

  if (!is_time_point(x)) {
    abort("`x` must be a 'clock_time_point'.")
  }
  if (!is_weekday(target)) {
    abort("`target` must be a 'clock_weekday'.")
  }

  target <- vec_recycle(target, vec_size(x), x_arg = "target")

  which <- validate_shift_which(which)
  boundary <- validate_shift_boundary(boundary)

  if (is_next(which)) {
    if (is_advance(boundary)) {
      x <- x + duration_days(1L)
    }
    x <- x + (target - as_weekday(x))
  } else {
    if (is_advance(boundary)) {
      x <- x - duration_days(1L)
    }
    x <- x - (as_weekday(x) - target)
  }

  x
}

validate_shift_which <- function(which) {
  if (!is_string(which, string = c("next", "previous"))) {
    abort("`which` must be either \"next\" or \"previous\".")
  }
  which
}
validate_shift_boundary <- function(boundary) {
  if (!is_string(boundary, string = c("keep", "advance"))) {
    abort("`boundary` must be either \"keep\" or \"advance\".")
  }
  boundary
}

is_next <- function(x) {
  identical(x, "next")
}
is_advance <- function(x) {
  identical(x, "advance")
}

# ------------------------------------------------------------------------------

#' Counting: time point
#'
#' @description
#' `time_point_count_between()` counts the number of `precision` units
#' between `start` and `end` (i.e., the number of days or hours). This count
#' corresponds to the _whole number_ of units, and will never return a
#' fractional value.
#'
#' This is suitable for, say, computing the whole number of days between two
#' time points, accounting for the time of day.
#'
#' @details
#' Remember that `time_point_count_between()` returns an integer vector.
#' With extremely fine precisions, such as nanoseconds, the count can quickly
#' exceed the maximum value that is allowed in an integer. In this case, an
#' `NA` will be returned with a warning.
#'
#' @inheritSection calendar_count_between Comparison Direction
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param start,end `[clock_time_point]`
#'
#'   A pair of time points. These will be recycled to their common size.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"week"`
#'   - `"day"`
#'   - `"hour"`
#'   - `"minute"`
#'   - `"second"`
#'   - `"millisecond"`
#'   - `"microsecond"`
#'   - `"nanosecond"`
#'
#' @param n `[positive integer(1)]`
#'
#'   A single positive integer specifying a multiple of `precision` to use.
#'
#' @return An integer representing the number of `precision` units between
#' `start` and `end`.
#'
#' @export
#' @examples
#' x <- as_naive_time(year_month_day(2019, 2, 3))
#' y <- as_naive_time(year_month_day(2019, 2, 10))
#'
#' # Whole number of days or hours between two time points
#' time_point_count_between(x, y, "day")
#' time_point_count_between(x, y, "hour")
#'
#' # Whole number of 2-day units
#' time_point_count_between(x, y, "day", n = 2)
#'
#' # Leap years are taken into account
#' x <- as_naive_time(year_month_day(c(2020, 2021), 2, 28))
#' y <- as_naive_time(year_month_day(c(2020, 2021), 3, 01))
#' time_point_count_between(x, y, "day")
#'
#' # Time of day is taken into account.
#' # `2020-02-02T04 -> 2020-02-03T03` is not a whole day (because of the hour)
#' # `2020-02-02T04 -> 2020-02-03T05` is a whole day
#' x <- as_naive_time(year_month_day(2020, 2, 2, 4))
#' y <- as_naive_time(year_month_day(2020, 2, 3, c(3, 5)))
#' time_point_count_between(x, y, "day")
#' time_point_count_between(x, y, "hour")
#'
#' # Can compute negative counts (using the same example from above)
#' time_point_count_between(y, x, "day")
#' time_point_count_between(y, x, "hour")
#'
#' # Repeated computation at increasingly fine precisions
#' x <- as_naive_time(year_month_day(
#'   2020, 2, 2, 4, 5, 6, 200,
#'   subsecond_precision = "microsecond"
#' ))
#' y <- as_naive_time(year_month_day(
#'   2020, 3, 1, 8, 9, 10, 100,
#'   subsecond_precision = "microsecond"
#' ))
#'
#' days <- time_point_count_between(x, y, "day")
#' x <- x + duration_days(days)
#'
#' hours <- time_point_count_between(x, y, "hour")
#' x <- x + duration_hours(hours)
#'
#' minutes <- time_point_count_between(x, y, "minute")
#' x <- x + duration_minutes(minutes)
#'
#' seconds <- time_point_count_between(x, y, "second")
#' x <- x + duration_seconds(seconds)
#'
#' microseconds <- time_point_count_between(x, y, "microsecond")
#' x <- x + duration_microseconds(microseconds)
#'
#' data.frame(
#'   days = days,
#'   hours = hours,
#'   minutes = minutes,
#'   seconds = seconds,
#'   microseconds = microseconds
#' )
time_point_count_between <- function(start, end, precision, ..., n = 1L) {
  check_dots_empty()

  if (!is_time_point(start)) {
    abort("`start` must be a <clock_time_point>.")
  }
  if (!is_time_point(end)) {
    abort("`end` must be a <clock_time_point>.")
  }

  args <- vec_cast_common(start = start, end = end)
  args <- vec_recycle_common(!!!args)
  start <- args[[1]]
  end <- args[[2]]

  precision_int <- validate_precision_string(precision)
  if (precision_int < PRECISION_WEEK) {
    abort("`precision` must be at least 'week' precision.")
  }

  n <- vec_cast(n, integer(), x_arg = "n")
  if (!is_number(n) || n <= 0L) {
    abort("`n` must be a single positive integer.")
  }

  out <- end - start
  out <- duration_cast(out, precision)

  if (n != 1L) {
    out <- out %/% n
  }

  as.integer(out)
}

# ------------------------------------------------------------------------------

#' Sequences: time points
#'
#' @description
#' This is a time point method for the [seq()] generic. It works for sys-time
#' and naive-time vectors.
#'
#' Sequences can be generated for all valid time point precisions (daily through
#' nanosecond).
#'
#' When calling `seq()`, exactly two of the following must be specified:
#' - `to`
#' - `by`
#' - Either `length.out` or `along.with`
#'
#' @inheritParams seq.clock_duration
#'
#' @param from `[clock_sys_time(1) / clock_naive_time(1)]`
#'
#'   A time point to start the sequence from.
#'
#'   `from` is always included in the result.
#'
#' @param to `[clock_sys_time(1) / clock_naive_time(1) / NULL]`
#'
#'   A time point to stop the sequence at.
#'
#'   `to` is cast to the type of `from`.
#'
#'   `to` is only included in the result if the resulting sequence divides
#'   the distance between `from` and `to` exactly.
#'
#' @return A sequence with the type of `from`.
#'
#' @export
#' @examples
#' # Daily sequence
#' seq(
#'   as_naive_time(year_month_day(2019, 1, 1)),
#'   as_naive_time(year_month_day(2019, 2, 4)),
#'   by = 5
#' )
#'
#' # Minutely sequence using minute precision naive-time
#' x <- as_naive_time(year_month_day(2019, 1, 2, 3, 3))
#' x
#'
#' seq(x, by = 4, length.out = 10)
#'
#' # You can use larger step sizes by using a duration-based `by`
#' seq(x, by = duration_days(1), length.out = 5)
#'
#' # Nanosecond sequence
#' from <- as_naive_time(year_month_day(2019, 1, 1))
#' from <- time_point_cast(from, "nanosecond")
#' to <- from + 100
#' seq(from, to, by = 10)
seq.clock_time_point <- function(from,
                                 to = NULL,
                                 by = NULL,
                                 length.out = NULL,
                                 along.with = NULL,
                                 ...) {
  precision <- time_point_precision_attribute(from)

  seq_impl(
    from = from,
    to = to,
    by = by,
    length.out = length.out,
    along.with = along.with,
    precision = precision,
    ...
  )
}

# ------------------------------------------------------------------------------

#' Precision: time point
#'
#' `time_point_precision()` extracts the precision from a time point, such
#' as a sys-time or naive-time. It returns the precision as a single string.
#'
#' @param x `[clock_time_point]`
#'
#'   A time point.
#'
#' @return A single string holding the precision of the time point.
#'
#' @export
#' @examples
#' time_point_precision(sys_time_now())
#' time_point_precision(as_naive_time(duration_days(1)))
time_point_precision <- function(x) {
  if (!is_time_point(x)) {
    abort("`x` must be a 'clock_time_point'.")
  }
  precision <- time_point_precision_attribute(x)
  precision <- precision_to_string(precision)
  precision
}

# ------------------------------------------------------------------------------

validate_time_point_precision_string <- function(precision) {
  precision <- validate_precision_string(precision)

  if (!is_valid_time_point_precision(precision)) {
    abort("`precision` must be at least 'day' precision.")
  }

  precision
}

is_valid_time_point_precision <- function(precision) {
  precision >= PRECISION_DAY
}

PRECISION_YEAR = 0L
PRECISION_QUARTER = 1L
PRECISION_MONTH = 2L
PRECISION_WEEK = 3L
PRECISION_DAY = 4L
PRECISION_HOUR = 5L
PRECISION_MINUTE = 6L
PRECISION_SECOND = 7L
PRECISION_MILLISECOND = 8L
PRECISION_MICROSECOND = 9L
PRECISION_NANOSECOND = 10L

# ------------------------------------------------------------------------------

check_precision <- function(
  x,
  ...,
  values = NULL,
  arg = caller_arg(x),
  call = caller_env()
) {
  values <- values %||% precision_names()
  check_string(x, allow_empty = FALSE, arg = arg, call = call)
  arg_match0(x, values = values, arg_nm = arg, error_call = call)
}

check_precision_subsecond <- function(
  x,
  ...,
  arg = caller_arg(x),
  call = caller_env()
) {
  check_precision(
    x = x,
    values = c("millisecond", "microsecond", "nanosecond"),
    arg = arg,
    call = call
  )
}

precision_to_integer <- function(x) {
  check_string(x, .internal = TRUE)

  switch(
    x,
    year = PRECISION_YEAR,
    quarter = PRECISION_QUARTER,
    month = PRECISION_MONTH,
    week = PRECISION_WEEK,
    day = PRECISION_DAY,
    hour = PRECISION_HOUR,
    minute = PRECISION_MINUTE,
    second = PRECISION_SECOND,
    millisecond = PRECISION_MILLISECOND,
    microsecond = PRECISION_MICROSECOND,
    nanosecond = PRECISION_NANOSECOND,
    abort("`x` not recognized.", .internal = TRUE)
  )
}

precision_abbr <- function(precision_string) {
  switch(
    precision_string,
    millisecond = "milli",
    microsecond = "micro",
    nanosecond = "nano",
    precision_string # fallthrough
  )
}

precision_common2 <- function(x, y) {
  if (x >= y) {
    x
  } else {
    y
  }
}

precision_names <- function() {
  c(
    "year",
    "quarter",
    "month",
    "week",
    "day",
    "hour",
    "minute",
    "second",
    "millisecond",
    "microsecond",
    "nanosecond"
  )
}

precision_time_names <- function() {
  c(
    "hour",
    "minute",
    "second",
    "millisecond",
    "microsecond",
    "nanosecond"
  )
}

# Internal generic
clock_minimum <- function(x) {
  UseMethod("clock_minimum")
}

# Internal generic
clock_maximum <- function(x) {
  UseMethod("clock_maximum")
}

clock_init_limits_init <- function(env) {
  assign(
    "clock_calendar_year_maximum",
    clock_get_calendar_year_maximum(),
    envir = env
  )
  assign(
    "clock_calendar_year_minimum",
    clock_get_calendar_year_minimum(),
    envir = env
  )
}

#' @export
as_sys_time.Date <- function(x) {
  names <- names(x)
  x <- unstructure(x)
  if (is.double(x)) {
    x <- floor(x)
  }
  x <- duration_days(x)
  new_sys_time_from_fields(x, PRECISION_DAY, names)
}

#' @export
as_naive_time.Date <- function(x) {
  as_naive_time(as_sys_time(x))
}

#' Convert to a zoned-time from a date
#'
#' @description
#' This is a Date method for the [as_zoned_time()] generic.
#'
#' clock assumes that Dates are _naive_ date-time types. Like naive-times, they
#' have a yet-to-be-specified time zone. This method allows you to specify that
#' time zone, keeping the printed time. If possible, the time will be set to
#' midnight (see Details for the rare case in which this is not possible).
#'
#' @details
#' In the rare instance that the specified time zone does not contain a
#' date-time at midnight due to daylight saving time, `nonexistent` can be used
#' to resolve the issue. Similarly, if there are two possible midnight times due
#' to a daylight saving time fallback, `ambiguous` can be used.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams as-zoned-time-naive-time
#'
#' @param x `[Date]`
#'
#'   A Date.
#'
#' @return A zoned-time.
#'
#' @name as-zoned-time-Date
#' @export
#' @examples
#' x <- as.Date("2019-01-01")
#'
#' # The resulting zoned-times have the same printed time, but are in
#' # different time zones
#' as_zoned_time(x, "UTC")
#' as_zoned_time(x, "America/New_York")
#'
#' # Converting Date -> zoned-time is the same as naive-time -> zoned-time
#' x <- as_naive_time(year_month_day(2019, 1, 1))
#' as_zoned_time(x, "America/New_York")
#'
#' # In Asia/Beirut, there was a DST gap from
#' # 2021-03-27 23:59:59 -> 2021-03-28 01:00:00,
#' # skipping the 0th hour entirely. This means there is no midnight value.
#' x <- as.Date("2021-03-28")
#' try(as_zoned_time(x, "Asia/Beirut"))
#'
#' # To resolve this, set a `nonexistent` time resolution strategy
#' as_zoned_time(x, "Asia/Beirut", nonexistent = "roll-forward")
as_zoned_time.Date <- function(x,
                               zone,
                               ...,
                               nonexistent = NULL,
                               ambiguous = NULL) {
  x <- as_naive_time(x)
  as_zoned_time(x, zone = zone, ..., nonexistent = nonexistent, ambiguous = ambiguous)
}

#' @export
as_year_month_day.Date <- function(x) {
  as_year_month_day(as_naive_time(x))
}

#' @export
as_year_month_weekday.Date <- function(x) {
  as_year_month_weekday(as_naive_time(x))
}

#' @export
as_year_quarter_day.Date <- function(x, ..., start = NULL) {
  as_year_quarter_day(as_naive_time(x), ..., start = start)
}

#' @export
as_iso_year_week_day.Date <- function(x) {
  as_iso_year_week_day(as_naive_time(x))
}

#' @export
as_year_day.Date <- function(x) {
  as_year_day(as_naive_time(x))
}

#' @export
as_weekday.Date <- function(x) {
  as_weekday(as_naive_time(x))
}

# ------------------------------------------------------------------------------

# Not using `check_dots_empty()` because that might
# be too aggressive with base generics

#' @export
as.Date.clock_calendar <- function(x, ...) {
  as.Date(as_naive_time(x))
}

#' @export
as.Date.clock_time_point <- function(x, ...) {
  names <- clock_rcrd_names(x)
  x <- time_point_floor(x, "day")
  x <- field_ticks(x)
  x <- as.double(x)
  names(x) <- names
  new_date(x)
}

#' @export
as.Date.clock_zoned_time <- function(x, ...) {
  as.Date(as_naive_time(x))
}

# ------------------------------------------------------------------------------

#' Convert to a date
#'
#' @description
#' `as_date()` is a generic function that converts its input to a date (Date).
#'
#' There are methods for converting date-times (POSIXct), calendars,
#' time points, and zoned-times to dates.
#'
#' For converting to a date-time, see [as_date_time()].
#'
#' @details
#' Note that clock always assumes that R's Date class is naive, so converting
#' a POSIXct to a Date will always retain the printed year, month, and day
#' value.
#'
#' This is not a drop-in replacement for `as.Date()`, as it only converts a
#' limited set of types to Date. For parsing characters as dates, see
#' [date_parse()]. For converting numerics to dates, see [vctrs::new_date()] or
#' continue to use `as.Date()`.
#'
#' @param x `[vector]`
#'
#'   A vector.
#'
#' @return A date with the same length as `x`.
#'
#' @export
#' @examples
#' x <- date_time_parse("2019-01-01 23:02:03", "America/New_York")
#'
#' # R's `as.Date.POSIXct()` method defaults to changing the printed time
#' # to UTC before converting, which can result in odd conversions like this:
#' as.Date(x)
#'
#' # `as_date()` will never change the printed time before converting
#' as_date(x)
#'
#' # Can also convert from other clock types
#' as_date(year_month_day(2019, 2, 5))
as_date <- function(x) {
  UseMethod("as_date")
}

#' @rdname as_date
#' @export
as_date.Date <- function(x) {
  date_standardize(x)
}

#' @rdname as_date
#' @export
as_date.POSIXt <- function(x) {
  as.Date(as_naive_time(x))
}

#' @rdname as_date
#' @export
as_date.clock_calendar <- function(x) {
  as.Date(x)
}

#' @rdname as_date
#' @export
as_date.clock_time_point <- function(x) {
  as.Date(x)
}

#' @rdname as_date
#' @export
as_date.clock_zoned_time <- function(x) {
  as.Date(x)
}

# ------------------------------------------------------------------------------

#' Getters: date
#'
#' @description
#' These are Date methods for the [getter generics][clock-getters].
#'
#' - `get_year()` returns the Gregorian year.
#'
#' - `get_month()` returns the month of the year.
#'
#' - `get_day()` returns the day of the month.
#'
#' For more advanced component extraction, convert to the calendar type
#' that you are interested in.
#'
#' @param x `[Date]`
#'
#'   A Date to get the component from.
#'
#' @return The component.
#'
#' @name Date-getters
#' @examples
#' x <- as.Date("2019-01-01") + 0:5
#' get_day(x)
NULL

#' @rdname Date-getters
#' @export
get_year.Date <- function(x) {
  get_date_field_year_month_day(x, get_year)
}
#' @rdname Date-getters
#' @export
get_month.Date <- function(x) {
  get_date_field_year_month_day(x, get_month)
}
#' @rdname Date-getters
#' @export
get_day.Date <- function(x) {
  get_date_field_year_month_day(x, get_day)
}
get_date_field_year_month_day <- function(x, get_fn) {
  x <- as_year_month_day(x)
  get_fn(x)
}

# ------------------------------------------------------------------------------

#' Setters: date
#'
#' @description
#' These are Date methods for the [setter generics][clock-setters].
#'
#' - `set_year()` sets the year.
#'
#' - `set_month()` sets the month of the year. Valid values are in the range
#'   of `[1, 12]`.
#'
#' - `set_day()` sets the day of the month. Valid values are in the range
#'   of `[1, 31]`.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams invalid_resolve
#'
#' @param x `[Date]`
#'
#'   A Date vector.
#'
#' @param value `[integer / "last"]`
#'
#'   The value to set the component to.
#'
#'   For `set_day()`, this can also be `"last"` to set the day to the
#'   last day of the month.
#'
#' @return `x` with the component set.
#'
#' @name Date-setters
#' @examples
#' x <- as.Date("2019-02-01")
#'
#' # Set the day
#' set_day(x, 12:14)
#'
#' # Set to the "last" day of the month
#' set_day(x, "last")
#'
#' # You cannot set a Date to an invalid day like you can with
#' # a year-month-day. Instead, the default strategy is to error.
#' try(set_day(x, 31))
#' set_day(as_year_month_day(x), 31)
#'
#' # You can resolve these issues while setting the day by specifying
#' # an invalid date resolution strategy with `invalid`
#' set_day(x, 31, invalid = "previous")
NULL

#' @rdname Date-setters
#' @export
set_year.Date <- function(x, value, ..., invalid = NULL) {
  set_date_field_year_month_day(x, value, invalid, set_year, ...)
}
#' @rdname Date-setters
#' @export
set_month.Date <- function(x, value, ..., invalid = NULL) {
  set_date_field_year_month_day(x, value, invalid, set_month, ...)
}
#' @rdname Date-setters
#' @export
set_day.Date <- function(x, value, ..., invalid = NULL) {
  set_date_field_year_month_day(x, value, invalid, set_day, ...)
}
set_date_field_year_month_day <- function(x, value, invalid, set_fn, ...) {
  check_dots_empty()
  x <- as_year_month_day(x)
  x <- set_fn(x, value)
  x <- invalid_resolve(x, invalid = invalid)
  as.Date(x)
}

# ------------------------------------------------------------------------------

#' @method vec_arith.Date clock_duration
#' @export
vec_arith.Date.clock_duration <- function(op, x, y, ...) {
  arith_date_and_duration(op, x, y, ...)
}

#' @method vec_arith.clock_duration Date
#' @export
vec_arith.clock_duration.Date <- function(op, x, y, ...) {
  arith_duration_and_date(op, x, y, ...)
}

arith_date_and_duration <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(x, y),
    "-" = add_duration(x, -y),
    stop_incompatible_op(op, x, y, ...)
  )
}

arith_duration_and_date <- function(op, x, y, ...) {
  switch(
    op,
    "+" = add_duration(y, x, swapped = TRUE),
    "-" = stop_incompatible_op(op, x, y, details = "Can't subtract a Date from a duration.", ...),
    stop_incompatible_op(op, x, y, ...)
  )
}

# ------------------------------------------------------------------------------

#' Arithmetic: date
#'
#' @description
#' These are Date methods for the
#' [arithmetic generics][clock-arithmetic].
#'
#' Calendrical based arithmetic:
#'
#' These functions convert to a year-month-day calendar, perform
#' the arithmetic, then convert back to a Date.
#'
#' - `add_years()`
#'
#' - `add_quarters()`
#'
#' - `add_months()`
#'
#' Time point based arithmetic:
#'
#' These functions convert to a time point, perform the arithmetic, then
#' convert back to a Date.
#'
#' - `add_weeks()`
#'
#' - `add_days()`
#'
#' @details
#' Adding a single quarter with `add_quarters()` is equivalent to adding
#' 3 months.
#'
#' `x` and `n` are recycled against each other.
#'
#' Only calendrical based arithmetic has the potential to generate invalid
#' dates. Time point based arithmetic, like adding days, will always generate
#' a valid date.
#'
#' @inheritParams add_years
#' @inheritParams invalid_resolve
#'
#' @param x `[Date]`
#'
#'   A Date vector.
#'
#' @return `x` after performing the arithmetic.
#'
#' @name Date-arithmetic
#'
#' @examples
#' x <- as.Date("2019-01-01")
#'
#' add_years(x, 1:5)
#'
#' y <- as.Date("2019-01-31")
#'
#' # Adding 1 month to `y` generates an invalid date. Unlike year-month-day
#' # types, R's native Date type cannot handle invalid dates, so you must
#' # resolve them immediately. If you don't you get an error:
#' try(add_months(y, 1:2))
#' add_months(as_year_month_day(y), 1:2)
#'
#' # Resolve invalid dates by specifying an invalid date resolution strategy
#' # with the `invalid` argument. Using `"previous"` here sets the date to
#' # the previous valid date - i.e. the end of the month.
#' add_months(y, 1:2, invalid = "previous")
NULL

#' @rdname Date-arithmetic
#' @export
add_years.Date <- function(x, n, ..., invalid = NULL) {
  add_date_duration_year_month_day(x, n, invalid, add_years, ...)
}
#' @rdname Date-arithmetic
#' @export
add_quarters.Date <- function(x, n, ..., invalid = NULL) {
  add_date_duration_year_month_day(x, n, invalid, add_quarters, ...)
}
#' @rdname Date-arithmetic
#' @export
add_months.Date <- function(x, n, ..., invalid = NULL) {
  add_date_duration_year_month_day(x, n, invalid, add_months, ...)
}
add_date_duration_year_month_day <- function(x, n, invalid, add_fn, ...) {
  check_dots_empty()
  x <- as_year_month_day(x)
  x <- add_fn(x, n)
  x <- invalid_resolve(x, invalid = invalid)
  as.Date(x)
}

#' @rdname Date-arithmetic
#' @export
add_weeks.Date <- function(x, n, ...) {
  add_date_duration_time_point(x, n, add_weeks, ...)
}
#' @rdname Date-arithmetic
#' @export
add_days.Date <- function(x, n, ...) {
  add_date_duration_time_point(x, n, add_days, ...)
}
add_date_duration_time_point <- function(x, n, add_fn, ...) {
  check_dots_empty()
  x <- as_naive_time(x)
  x <- add_fn(x, n)
  as.Date(x)
}

# ------------------------------------------------------------------------------

#' Group date and date-time components
#'
#' @description
#' `date_group()` groups by a single component of a date-time, such as month
#' of the year, or day of the month.
#'
#' There are separate help pages for grouping dates and date-times:
#'
#' - [dates (Date)][date-group]
#'
#' - [date-times (POSIXct/POSIXlt)][posixt-group]
#'
#' @inheritParams calendar_group
#'
#' @param x `[Date / POSIXct / POSIXlt]`
#'
#'   A date or date-time vector.
#'
#' @param precision `[character(1)]`
#'
#'   A precision. Allowed precisions are dependent on the input used.
#'
#' @return `x`, grouped at `precision`.
#'
#' @export
#' @examples
#' # See type specific documentation for more examples
#' date_group(as.Date("2019-01-01") + 0:5, "day", n = 2)
date_group <- function(x, precision, ..., n = 1L) {
  UseMethod("date_group")
}

#' Group date components
#'
#' @description
#' This is a Date method for the [date_group()] generic.
#'
#' `date_group()` groups by a single component of a Date, such as month
#' of the year, or day of the month.
#'
#' If you need to group by more complex components, like ISO weeks, or quarters,
#' convert to a calendar type that contains the component you are interested
#' in grouping by.
#'
#' @inheritParams date_group
#' @inheritParams invalid_resolve
#'
#' @param x `[Date]`
#'
#'   A date vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'
#'   - `"month"`
#'
#'   - `"day"`
#'
#' @return `x`, grouped at `precision`.
#'
#' @name date-group
#'
#' @export
#' @examples
#' x <- as.Date("2019-01-01") + -3:5
#' x
#'
#' # Group by 2 days of the current month.
#' # Note that this resets at the beginning of the month, creating day groups
#' # of [29, 30] [31] [01, 02] [03, 04].
#' date_group(x, "day", n = 2)
#'
#' # Group by month
#' date_group(x, "month")
date_group.Date <- function(x, precision, ..., n = 1L, invalid = NULL) {
  x <- as_year_month_day(x)
  x <- calendar_group(x, precision, ..., n = n)
  x <- calendar_widen(x, "day")
  as.Date(x, invalid = invalid)
}

# ------------------------------------------------------------------------------

#' Is the year a leap year?
#'
#' `date_leap_year()` detects if the year is a leap year.
#'
#' @param x `[Date / POSIXct / POSIXlt]`
#'
#'   A date or date-time to detect leap years in.
#'
#' @return A logical vector the same size as `x`. Returns `TRUE` if in a leap
#'   year, `FALSE` if not in a leap year, and `NA` if `x` is `NA`.
#'
#' @examples
#' x <- as.Date("2019-01-01")
#' x <- add_years(x, 0:5)
#' date_leap_year(x)
#'
#' y <- as.POSIXct("2019-01-01", "America/New_York")
#' y <- add_years(y, 0:5)
#' date_leap_year(y)
#' @export
date_leap_year <- function(x) {
  UseMethod("date_leap_year")
}

#' @export
date_leap_year.Date <- function(x) {
  x <- as_year_month_day(x)
  calendar_leap_year(x)
}

# ------------------------------------------------------------------------------

#' Date and date-time rounding
#'
#' @description
#' - `date_floor()` rounds a date or date-time down to a multiple of
#'   the specified `precision`.
#'
#' - `date_ceiling()` rounds a date or date-time up to a multiple of
#'   the specified `precision`.
#'
#' - `date_round()` rounds up or down depending on what is closer,
#'   rounding up on ties.
#'
#' There are separate help pages for rounding dates and date-times:
#'
#' - [dates (Date)][date-rounding]
#'
#' - [date-times (POSIXct/POSIXlt)][posixt-rounding]
#'
#' These functions round the underlying duration itself, relative to an
#' `origin`. For example, rounding to 15 hours will construct groups of
#' 15 hours, starting from `origin`, which defaults to a naive time of
#' 1970-01-01 00:00:00.
#'
#' If you want to group by components, such as "day of the month", see
#' [date_group()].
#'
#' @inheritParams date_group
#'
#' @param origin `[Date(1) / POSIXct(1) / POSIXlt(1) / NULL]`
#'
#'   An origin to start counting from. The default `origin` is
#'   midnight on 1970-01-01 in the time zone of `x`.
#'
#' @return `x` rounded to the specified `precision`.
#'
#' @name date-and-date-time-rounding
#'
#' @examples
#' # See the type specific documentation for more examples
#'
#' x <- as.Date("2019-03-31") + 0:5
#' x
#'
#' # Flooring by 2 days, note that this is not tied to the current month,
#' # and instead counts from the specified `origin`.
#' date_floor(x, "day", n = 2)
NULL

#' @rdname date-and-date-time-rounding
#' @export
date_floor <- function(x, precision, ..., n = 1L, origin = NULL) {
  UseMethod("date_floor")
}

#' @rdname date-and-date-time-rounding
#' @export
date_ceiling <- function(x, precision, ..., n = 1L, origin = NULL) {
  UseMethod("date_ceiling")
}

#' @rdname date-and-date-time-rounding
#' @export
date_round <- function(x, precision, ..., n = 1L, origin = NULL) {
  UseMethod("date_round")
}

#' Rounding: date
#'
#' @description
#' These are Date methods for the
#' [rounding generics][date-and-date-time-rounding].
#'
#' - `date_floor()` rounds a date down to a multiple of
#'   the specified `precision`.
#'
#' - `date_ceiling()` rounds a date up to a multiple of
#'   the specified `precision`.
#'
#' - `date_round()` rounds up or down depending on what is closer,
#'   rounding up on ties.
#'
#' The only supported rounding `precision`s for Dates are `"day"` and `"week"`.
#' You can group by irregular periods such as `"month"` or `"year"` by using
#' [date_group()].
#'
#' @details
#' When rounding by `"week"`, remember that the `origin` determines the "week
#' start". By default, 1970-01-01 is the implicit origin, which is a
#' Thursday. If you would like to round by weeks with a different week start,
#' just supply an origin on the weekday you are interested in.
#'
#' @inheritParams date_floor
#'
#' @param x `[Date]`
#'
#'   A date vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"week"`
#'
#'   - `"day"`
#'
#'   `"week"` is an alias for `"day"` with `n * 7`.
#'
#' @param origin `[Date(1) / NULL]`
#'
#'   An origin to start counting from. The default `origin` is
#'   1970-01-01.
#'
#' @return `x` rounded to the specified `precision`.
#'
#' @name date-rounding
#'
#' @examples
#' x <- as.Date("2019-03-31") + 0:5
#' x
#'
#' # Flooring by 2 days, note that this is not tied to the current month,
#' # and instead counts from the specified `origin`, so groups can cross
#' # the month boundary
#' date_floor(x, "day", n = 2)
#'
#' # Compare to `date_group()`, which groups by the day of the month
#' date_group(x, "day", n = 2)
#'
#' y <- as.Date("2019-01-01") + 0:20
#' y
#'
#' # Flooring by week uses an implicit `origin` of 1970-01-01, which
#' # is a Thursday
#' date_floor(y, "week")
#' as_weekday(date_floor(y, "week"))
#'
#' # If you want to round by weeks with a different week start, supply an
#' # `origin` that falls on the weekday you care about. This uses a Monday.
#' origin <- as.Date("1970-01-05")
#' as_weekday(origin)
#'
#' date_floor(y, "week", origin = origin)
#' as_weekday(date_floor(y, "week", origin = origin))
NULL

#' @rdname date-rounding
#' @export
date_floor.Date <- function(x, precision, ..., n = 1L, origin = NULL) {
  date_rounder(x, precision, n, origin, time_point_floor, ...)
}

#' @rdname date-rounding
#' @export
date_ceiling.Date <- function(x, precision, ..., n = 1L, origin = NULL) {
  date_rounder(x, precision, n, origin, time_point_ceiling, ...)
}

#' @rdname date-rounding
#' @export
date_round.Date <- function(x, precision, ..., n = 1L, origin = NULL) {
  date_rounder(x, precision, n, origin, time_point_round, ...)
}

date_rounder <- function(x, precision, n, origin, time_point_rounder, ...) {
  result <- tweak_date_rounder_precision(precision, n)
  precision <- result$precision
  n <- result$n

  x <- as_naive_time(x)

  if (!is_null(origin)) {
    origin <- collect_date_rounder_origin(origin)
  }

  x <- time_point_rounder(x, precision, ..., n = n, origin = origin)

  as.Date(x)
}

# Note:
# For Date and POSIXct, which are always day and second precision, we can
# allow a special "week" precision for the rounding functions. This isn't
# normally allowed for time points, as there is no week precision time point,
# and instead you'd do `day`,` `n = n * 7`. This makes that a little easier.
tweak_date_rounder_precision <- function(precision, n) {
  if (identical(precision, "week")) {
    precision <- "day"
    n <- n * 7L
  }

  list(precision = precision, n = n)
}

collect_date_rounder_origin <- function(origin) {
  if (!inherits(origin, "Date")) {
    abort("`origin` must be a 'Date'.")
  }
  if (length(origin) != 1L) {
    abort("`origin` must have length 1.")
  }
  if (!is.finite(origin)) {
    abort("`origin` must not be `NA` or an infinite date.")
  }

  origin <- as_naive_time(origin)

  origin
}

# ------------------------------------------------------------------------------

#' Convert a date or date-time to a weekday factor
#'
#' `date_weekday_factor()` converts a date or date-time to an ordered factor
#' with levels representing the weekday. This can be useful in combination with
#' ggplot2, or for modeling.
#'
#' @inheritParams weekday_factor
#'
#' @param x `[Date / POSIXct / POSIXlt]`
#'
#'   A date or date-time vector.
#'
#' @return An ordered factor representing the weekdays.
#'
#' @export
#' @examples
#' x <- as.Date("2019-01-01") + 0:6
#'
#' # Default to Sunday -> Saturday
#' date_weekday_factor(x)
#'
#' # ISO encoding is Monday -> Sunday
#' date_weekday_factor(x, encoding = "iso")
#'
#' # With full names
#' date_weekday_factor(x, abbreviate = FALSE)
#'
#' # Or a different language
#' date_weekday_factor(x, labels = "fr")
date_weekday_factor <- function(x,
                                ...,
                                labels = "en",
                                abbreviate = TRUE,
                                encoding = "western") {
  UseMethod("date_weekday_factor")
}

#' @export
date_weekday_factor.Date <- function(x,
                                     ...,
                                     labels = "en",
                                     abbreviate = TRUE,
                                     encoding = "western") {
  x <- as_weekday(x)
  weekday_factor(x, ..., labels = labels, abbreviate = abbreviate, encoding = encoding)
}

# ------------------------------------------------------------------------------

#' Convert a date or date-time to an ordered factor of month names
#'
#' @description
#' `date_month_factor()` extracts the month values from a date or date-time and
#' converts them to an ordered factor of month names. This can be useful in
#' combination with ggplot2, or for modeling.
#'
#' @inheritParams calendar_month_factor
#'
#' @param x `[Date / POSIXct / POSIXlt]`
#'
#'   A date or date-time vector.
#'
#' @return An ordered factor representing the months.
#'
#' @export
#' @examples
#' x <- add_months(as.Date("2019-01-01"), 0:11)
#'
#' date_month_factor(x)
#' date_month_factor(x, abbreviate = TRUE)
#' date_month_factor(x, labels = "fr")
date_month_factor <- function(x,
                              ...,
                              labels = "en",
                              abbreviate = FALSE) {
  UseMethod("date_month_factor")
}

#' @export
date_month_factor.Date <- function(x,
                                   ...,
                                   labels = "en",
                                   abbreviate = FALSE) {
  x <- as_year_month_day(x)
  calendar_month_factor(x, ..., labels = labels, abbreviate = abbreviate)
}

# ------------------------------------------------------------------------------

#' Formatting: date and date-time
#'
#' @description
#' `date_format()` formats a date (Date) or date-time (POSIXct/POSIXlt) using
#' a `format` string.
#'
#' There are separate help pages for formatting dates and date-times:
#'
#' - [dates (Date)][date-formatting]
#'
#' - [date-times (POSIXct/POSIXlt)][posixt-formatting]
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param x `[Date / POSIXct / POSIXlt]`
#'
#'   A date or date-time vector.
#'
#' @return A character vector of the formatted input.
#'
#' @export
#' @examples
#' # See method specific documentation for more examples
#'
#' x <- as.Date("2019-01-01")
#' date_format(x, format = "year: %Y, month: %m, day: %d")
date_format <- function(x, ...) {
  UseMethod("date_format")
}

#' Formatting: date
#'
#' @description
#' This is a Date method for the [date_format()] generic.
#'
#' `date_format()` formats a date (Date) using a `format` string.
#'
#' If `format` is `NULL`, a default format of `"%Y-%m-%d"` is used.
#'
#' @details
#' Because a Date is considered to be a _naive_ type in clock, meaning that
#' it currently has no implied time zone, using the `%z` or `%Z` format commands
#' is not allowed and will result in `NA`.
#'
#' @inheritParams rlang::args_dots_empty
#' @inheritParams format.clock_zoned_time
#'
#' @param x `[Date]`
#'
#'   A date vector.
#'
#' @return A character vector of the formatted input.
#'
#' @name date-formatting
#'
#' @export
#' @examples
#' x <- as.Date("2019-01-01")
#'
#' # Default
#' date_format(x)
#'
#' date_format(x, format = "year: %Y, month: %m, day: %d")
#'
#' # With different locales
#' date_format(x, format = "%A, %B %d, %Y")
#' date_format(x, format = "%A, %B %d, %Y", locale = clock_locale("fr"))
date_format.Date <- function(x,
                             ...,
                             format = NULL,
                             locale = clock_locale()) {
  check_dots_empty()
  x <- as_naive_time(x)
  format(x, format = format, locale = locale)
}

# ------------------------------------------------------------------------------

#' @export
date_zone.Date <- function(x) {
  abort("Can't get the zone of a 'Date'.")
}

#' @export
date_set_zone.Date <- function(x, zone) {
  abort("Can't set the zone of a 'Date'.")
}

# ------------------------------------------------------------------------------

#' Parsing: date
#'
#' @description
#' `date_parse()` parses strings into a Date.
#'
#' The default `format` used is `"%Y-%m-%d"`. This matches the default
#' result from calling `print()` or `format()` on a Date.
#'
#' @details
#' _`date_parse()` ignores both the `%z` and `%Z` commands,_ as clock treats
#' Date as a _naive_ type, with a yet-to-be-specified time zone.
#'
#' Parsing strings with sub-daily components, such as hours, minutes, or
#' seconds, should generally be done with [date_time_parse()]. If you only
#' need the date components from a string with sub-daily components, choose
#' one of the following:
#'
#' - If the date components are at the front of the string, and you don't want
#'   the time components to affect the date in any way, you can use
#'   [date_parse()] to parse only the date components. For example,
#'   `date_parse("2019-01-05 00:01:02", format = "%Y-%m-%d")` will parse
#'   through `05` and then stop.
#'
#' - If you want the time components to influence the date, then parse the full
#'   string with [date_time_parse()], round to day precision with a
#'   rounding function like [date_round()], and cast to date with [as_date()].
#'
#' Attempting to directly parse all components of a sub-daily string into a
#' Date is ambiguous and undefined, and is unlikely to work as you might expect.
#' For example, `date_parse("2019-01-05 00:01:02", format =
#' "%Y-%m-%d %H:%M:%S")` is not officially supported, even if it works in
#' some cases.
#'
#' @inheritParams zoned-parsing
#'
#' @return A Date.
#'
#' @export
#' @examples
#' date_parse("2020-01-01")
#'
#' date_parse(
#'   "January 5, 2020",
#'   format = "%B %d, %Y"
#' )
#'
#' # With a different locale
#' date_parse(
#'   "janvier 5, 2020",
#'   format = "%B %d, %Y",
#'   locale = clock_locale("fr")
#' )
#'
#' # A neat feature of `date_parse()` is the ability to parse
#' # the ISO year-week-day format
#' date_parse("2020-W01-2", format = "%G-W%V-%u")
#'
#' # ---------------------------------------------------------------------------
#' # Sub-daily components
#'
#' # If you have a string with sub-daily components, but only require the date,
#' # first parse them as date-times to fully parse the sub-daily components,
#' # then round using whatever convention is required for your use case before
#' # converting to date.
#' x <- c("2019-01-01 11", "2019-01-01 12")
#'
#' x <- date_time_parse(x, zone = "UTC", format = "%Y-%m-%d %H")
#' x
#'
#' date_floor(x, "day")
#' date_round(x, "day")
#'
#' as_date(date_round(x, "day"))
date_parse <- function(x, ..., format = NULL, locale = clock_locale()) {
  x <- naive_time_parse(x, ..., format = format, precision = "day", locale = locale)
  as.Date(x)
}

# ------------------------------------------------------------------------------

#' Shifting: date and date-time
#'
#' @description
#' `date_shift()` shifts `x` to the `target` weekday. You can shift to the next
#' or previous weekday. If `x` is currently on the `target` weekday, you can
#' choose to leave it alone or advance it to the next instance of the `target`.
#'
#' There are separate help pages for shifting dates and date-times:
#'
#' - [dates (Date)][date-shifting]
#'
#' - [date-times (POSIXct/POSIXlt)][posixt-shifting]
#'
#' @inheritParams time_point_shift
#'
#' @param x `[Date / POSIXct / POSIXlt]`
#'
#'   A date or date-time vector.
#'
#' @return `x` shifted to the `target` weekday.
#'
#' @name date-and-date-time-shifting
#'
#' @export
#' @examples
#' # See the type specific documentation for more examples
#'
#' x <- as.Date("2019-01-01") + 0:1
#'
#' # A Tuesday and Wednesday
#' as_weekday(x)
#'
#' monday <- weekday(clock_weekdays$monday)
#'
#' # Shift to the next Monday
#' date_shift(x, monday)
date_shift <- function(x,
                       target,
                       ...,
                       which = "next",
                       boundary = "keep") {
  UseMethod("date_shift")
}

#' Shifting: date
#'
#' @description
#' `date_shift()` shifts `x` to the `target` weekday. You can shift to the next
#' or previous weekday. If `x` is currently on the `target` weekday, you can
#' choose to leave it alone or advance it to the next instance of the `target`.
#'
#' Weekday shifting is one of the easiest ways to floor by week while
#' controlling what is considered the first day of the week. You can also
#' accomplish this with the `origin` argument of [date_floor()], but this is
#' slightly easier.
#'
#' @inheritParams time_point_shift
#'
#' @param x `[Date]`
#'
#'   A date vector.
#'
#' @return `x` shifted to the `target` weekday.
#'
#' @name date-shifting
#'
#' @export
#' @examples
#' x <- as.Date("2019-01-01") + 0:1
#'
#' # A Tuesday and Wednesday
#' as_weekday(x)
#'
#' monday <- weekday(clock_weekdays$monday)
#'
#' # Shift to the next Monday
#' date_shift(x, monday)
#'
#' # Shift to the previous Monday
#' # This is an easy way to "floor by week" with a target weekday in mind
#' date_shift(x, monday, which = "previous")
#'
#' # What about Tuesday?
#' tuesday <- weekday(clock_weekdays$tuesday)
#'
#' # Notice that the day that was currently on a Tuesday was not shifted
#' date_shift(x, tuesday)
#'
#' # You can force it to `"advance"`
#' date_shift(x, tuesday, boundary = "advance")
date_shift.Date <- function(x,
                            target,
                            ...,
                            which = "next",
                            boundary = "keep") {
  x <- as_naive_time(x)
  x <- time_point_shift(x, target, ..., which = which, boundary = boundary)
  as.Date(x)
}

# ------------------------------------------------------------------------------

#' Building: date
#'
#' @description
#' `date_build()` builds a Date from it's individual components.
#'
#' @details
#' Components are recycled against each other.
#'
#' @inheritParams invalid_resolve
#'
#' @param year `[integer]`
#'
#'   The year. Values `[-32767, 32767]` are generally allowed.
#'
#' @param month `[integer]`
#'
#'   The month. Values `[1, 12]` are allowed.
#'
#' @param day `[integer / "last"]`
#'
#'   The day of the month. Values `[1, 31]` are allowed.
#'
#'   If `"last"`, then the last day of the month is returned.
#'
#' @return A Date.
#'
#' @export
#' @examples
#' date_build(2019)
#' date_build(2019, 1:3)
#'
#' # Generating invalid dates will trigger an error
#' try(date_build(2019, 1:12, 31))
#'
#' # You can resolve this with `invalid`
#' date_build(2019, 1:12, 31, invalid = "previous")
#'
#' # But this particular case (the last day of the month) is better
#' # specified as:
#' date_build(2019, 1:12, "last")
date_build <- function(year, month = 1L, day = 1L, ..., invalid = NULL) {
  check_dots_empty()
  x <- year_month_day(year, month, day)
  x <- invalid_resolve(x, invalid = invalid)
  as.Date(x)
}

# ------------------------------------------------------------------------------

#' Current date and date-time
#'
#' @description
#' - `date_today()` returns the current date in the specified `zone` as a Date.
#'
#' - `date_now()` returns the current date-time in the specified `zone` as a
#' POSIXct.
#'
#' @details
#' clock assumes that Date is a _naive_ type, like naive-time. This means that
#' `date_today()` first looks up the current date-time in the specified `zone`,
#' then converts that to a Date, retaining the printed time while dropping any
#' information about that time zone.
#'
#' @inheritParams zoned_time_now
#'
#' @return
#' - `date_today()` a single Date.
#'
#' - `date_now()` a single POSIXct.
#'
#' @name date-today
#'
#' @examples
#' # Current date in the local time zone
#' date_today("")
#'
#' # Current date in a specified time zone
#' date_today("Europe/London")
#'
#' # Current date-time in that same time zone
#' date_now("Europe/London")
NULL

#' @rdname date-today
#' @export
date_today <- function(zone) {
  as.Date(zoned_time_now(zone))
}

# ------------------------------------------------------------------------------

#' Boundaries: date and date-time
#'
#' @description
#' - `date_start()` computes the date at the start of a particular
#'   `precision`, such as the "start of the year".
#'
#' - `date_end()` computes the date at the end of a particular
#'   `precision`, such as the "end of the month".
#'
#' There are separate help pages for computing boundaries for dates and
#' date-times:
#'
#' - [dates (Date)][date-boundary]
#'
#' - [date-times (POSIXct/POSIXlt)][posixt-boundary]
#'
#' @inheritParams date_group
#'
#' @param x `[Date / POSIXct / POSIXlt]`
#'
#'   A date or date-time vector.
#'
#' @param precision `[character(1)]`
#'
#'   A precision. Allowed precisions are dependent on the input used.
#'
#' @return `x` but with some components altered to be at the boundary value.
#'
#' @name date-and-date-time-boundary
#'
#' @examples
#' # See type specific documentation for more examples
#'
#' x <- date_build(2019, 2:4)
#'
#' date_end(x, "month")
#'
#' x <- date_time_build(2019, 2:4, 3:5, 4, 5, zone = "America/New_York")
#'
#' # Note that the hour, minute, and second components are also adjusted
#' date_end(x, "month")
NULL

#' @rdname date-and-date-time-boundary
#' @export
date_start <- function(x, precision, ...) {
  UseMethod("date_start")
}

#' @rdname date-and-date-time-boundary
#' @export
date_end <- function(x, precision, ...) {
  UseMethod("date_end")
}


#' Boundaries: date
#'
#' @description
#' This is a Date method for the [date_start()] and [date_end()] generics.
#'
#' @inheritParams date_group
#' @inheritParams invalid_resolve
#'
#' @param x `[Date]`
#'
#'   A date vector.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'
#'   - `"month"`
#'
#'   - `"day"`
#'
#' @return `x` but with some components altered to be at the boundary value.
#'
#' @name date-boundary
#'
#' @examples
#' x <- date_build(2019:2021, 2:4, 3:5)
#' x
#'
#' # Last day of the month
#' date_end(x, "month")
#'
#' # Last day of the year
#' date_end(x, "year")
#'
#' # First day of the year
#' date_start(x, "year")
NULL

#' @rdname date-boundary
#' @export
date_start.Date <- function(x, precision, ..., invalid = NULL) {
  check_dots_empty()
  x <- as_year_month_day(x)
  x <- calendar_start(x, precision)
  as.Date(x, invalid = invalid)
}

#' @rdname date-boundary
#' @export
date_end.Date <- function(x, precision, ..., invalid = NULL) {
  check_dots_empty()
  x <- as_year_month_day(x)
  x <- calendar_end(x, precision)
  as.Date(x, invalid = invalid)
}

# ------------------------------------------------------------------------------

#' Sequences: date and date-time
#'
#' @description
#' `date_seq()` generates a date (Date) or date-time (POSIXct/POSIXlt) sequence.
#'
#' There are separate help pages for generating sequences for dates and
#' date-times:
#'
#' - [dates (Date)][date-sequence]
#'
#' - [date-times (POSIXct/POSIXlt)][posixt-sequence]
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param from `[Date(1) / POSIXct(1) / POSIXlt(1)]`
#'
#'   A date or date-time to start the sequence from.
#'
#' @param to `[Date(1) / POSIXct(1) / POSIXlt(1) / NULL]`
#'
#'   A date or date-time to stop the sequence at.
#'
#'   `to` is only included in the result if the resulting sequence divides
#'   the distance between `from` and `to` exactly.
#'
#' @param by `[integer(1) / clock_duration(1) / NULL]`
#'
#'   The unit to increment the sequence by.
#'
#' @param total_size `[positive integer(1) / NULL]`
#'
#'   The size of the resulting sequence.
#'
#'   If specified alongside `to`, this must generate a non-fractional sequence
#'   between `from` and `to`.
#'
#' @return A date or date-time vector.
#'
#' @export
#' @examples
#' # See method specific documentation for more examples
#'
#' x <- as.Date("2019-01-01")
#' date_seq(x, by = duration_months(2), total_size = 20)
date_seq <- function(from,
                     ...,
                     to = NULL,
                     by = NULL,
                     total_size = NULL) {
  UseMethod("date_seq")
}

#' Sequences: date
#'
#' @description
#' This is a Date method for the [date_seq()] generic.
#'
#' `date_seq()` generates a date (Date) sequence.
#'
#' When calling `date_seq()`, exactly two of the following must be specified:
#' - `to`
#' - `by`
#' - `total_size`
#'
#' @inheritParams date_seq
#' @inheritParams invalid_resolve
#'
#' @param from `[Date(1)]`
#'
#'   A date to start the sequence from.
#'
#' @param to `[Date(1) / NULL]`
#'
#'   A date to stop the sequence at.
#'
#'   `to` is only included in the result if the resulting sequence divides
#'   the distance between `from` and `to` exactly.
#'
#'   If `to` is supplied along with `by`, all components of `to` more precise
#'   than the precision of `by` must match `from` exactly. For example, if `by =
#'   duration_months(1)`, the day component of `to` must match the day component
#'   of `from`. This ensures that the generated sequence is, at a minimum, a
#'   weakly monotonic sequence of dates.
#'
#' @param by `[integer(1) / clock_duration(1) / NULL]`
#'
#'   The unit to increment the sequence by.
#'
#'   If `by` is an integer, it is equivalent to `duration_days(by)`.
#'
#'   If `by` is a duration, it is allowed to have a precision of:
#'   - year
#'   - quarter
#'   - month
#'   - week
#'   - day
#'
#' @return A date vector.
#'
#' @name date-sequence
#'
#' @export
#' @examples
#' from <- date_build(2019, 1)
#' to <- date_build(2019, 4)
#'
#' # Defaults to daily sequence
#' date_seq(from, to = to, by = 7)
#'
#' # Use durations to change to monthly or yearly sequences
#' date_seq(from, to = to, by = duration_months(1))
#' date_seq(from, by = duration_years(-2), total_size = 3)
#'
#' # Note that components of `to` more precise than the precision of `by`
#' # must match `from` exactly. For example, this is not well defined:
#' from <- date_build(2019, 5, 2)
#' to <- date_build(2025, 7, 5)
#' try(date_seq(from, to = to, by = duration_years(1)))
#'
#' # The month and day components of `to` must match `from`
#' to <- date_build(2025, 5, 2)
#' date_seq(from, to = to, by = duration_years(1))
#'
#' # ---------------------------------------------------------------------------
#'
#' # Invalid dates must be resolved with the `invalid` argument
#' from <- date_build(2019, 1, 31)
#' to <- date_build(2019, 12, 31)
#'
#' try(date_seq(from, to = to, by = duration_months(1)))
#' date_seq(from, to = to, by = duration_months(1), invalid = "previous")
#'
#' # Compare this to the base R result, which is often a source of confusion
#' seq(from, to = to, by = "1 month")
#'
#' # This is equivalent to the overflow invalid resolution strategy
#' date_seq(from, to = to, by = duration_months(1), invalid = "overflow")
#'
#' # ---------------------------------------------------------------------------
#'
#' # Usage of `to` and `total_size` must generate a non-fractional sequence
#' # between `from` and `to`
#' from <- date_build(2019, 1, 1)
#' to <- date_build(2019, 1, 4)
#'
#' # These are fine
#' date_seq(from, to = to, total_size = 2)
#' date_seq(from, to = to, total_size = 4)
#'
#' # But this is not!
#' try(date_seq(from, to = to, total_size = 3))
date_seq.Date <- function(from,
                          ...,
                          to = NULL,
                          by = NULL,
                          total_size = NULL,
                          invalid = NULL) {
  check_dots_empty()

  check_number_of_supplied_optional_arguments(to, by, total_size)

  if (!is_null(to) && !is_Date(to)) {
    abort("If supplied, `to` must be a <Date>.")
  }

  if (!is_null(total_size)) {
    total_size <- check_length_out(total_size, arg = "total_size")
  }

  if (is_null(by)) {
    precision <- "day"
  } else if (is_duration(by)) {
    precision <- duration_precision(by)
  } else {
    precision <- "day"
    by <- duration_helper(by, PRECISION_DAY, n_arg = "by")
  }

  precision_int <- validate_precision_string(precision)

  if (precision_int == PRECISION_QUARTER) {
    by <- duration_cast(by, "month")
    precision <- "month"
    precision_int <- PRECISION_MONTH
  }

  if (precision_int == PRECISION_WEEK) {
    by <- duration_cast(by, "day")
    precision <- "day"
    precision_int <- PRECISION_DAY
  }

  if (precision_int %in% c(PRECISION_YEAR, PRECISION_MONTH)) {
    out <- date_seq_year_month(from, to, by, total_size, precision)
    out <- invalid_resolve(out, invalid = invalid)
    out <- as.Date(out)
    return(out)
  }

  if (precision_int == PRECISION_DAY) {
    out <- date_seq_day(from, to, by, total_size, precision)
    out <- as.Date(out)
    return(out)
  }

  abort("`by` must have a precision of 'year', 'quarter', 'month', 'week', or 'day'.")
}

date_seq_year_month <- function(from, to, by, total_size, precision) {
  has_time <- is_POSIXt(from)

  from <- as_year_month_day(from)
  original_from <- from
  from <- calendar_narrow(from, precision)

  if (!is_null(to)) {
    to <- as_year_month_day(to)
    check_from_to_component_equivalence(original_from, to, precision, has_time)
    to <- calendar_narrow(to, precision)
  }

  out <- seq(from, to = to, by = by, length.out = total_size)
  out <- reset_original_components(out, original_from, precision, has_time)

  out
}

date_seq_day <- function(from, to, by, total_size, precision) {
  date_seq_day_hour_minute_second(from, to, by, total_size, precision, as_naive_time)
}
date_seq_hour_minute_second <- function(from, to, by, total_size, precision) {
  date_seq_day_hour_minute_second(from, to, by, total_size, precision, as_sys_time)
}
date_seq_day_hour_minute_second <- function(from, to, by, total_size, precision, as_time_point_fn) {
  has_time <- is_POSIXt(from)

  from <- as_time_point_fn(from)
  original_from <- from
  from <- time_point_floor(from, precision)

  if (!is_null(to)) {
    to <- as_time_point_fn(to)

    check_from_to_component_equivalence(
      from = as_year_month_day(original_from),
      to = as_year_month_day(to),
      precision = precision,
      has_time = has_time
    )

    to <- time_point_floor(to, precision)
  }

  out <- seq(from, to = to, by = by, length.out = total_size)

  original_time <- original_from - from
  out <- out + original_time

  out
}

check_from_to_component_equivalence <- function(from, to, precision, has_time) {
  ok <- TRUE
  precision_int <- validate_precision_string(precision)

  if (precision_int < PRECISION_MONTH) {
    ok <- ok && is_true(get_month(from) == get_month(to))
  }
  if (precision_int < PRECISION_DAY) {
    ok <- ok && is_true(get_day(from) == get_day(to))
  }

  if (has_time) {
    if (precision_int < PRECISION_HOUR) {
      ok <- ok && is_true(get_hour(from) == get_hour(to))
    }
    if (precision_int < PRECISION_MINUTE) {
      ok <- ok && is_true(get_minute(from) == get_minute(to))
    }
    if (precision_int < PRECISION_SECOND) {
      ok <- ok && is_true(get_second(from) == get_second(to))
    }
  }

  if (!ok) {
    message <- paste0(
      "All components of `from` and `to` more precise than ",
      "'", precision, "' ",
      "must match."
    )
    abort(message)
  }

  invisible()
}

reset_original_components <- function(out, from, precision, has_time) {
  precision_int <- validate_precision_string(precision)

  if (precision_int < PRECISION_MONTH) {
    out <- set_month(out, get_month(from))
  }
  if (precision_int < PRECISION_DAY) {
    out <- set_day(out, get_day(from))
  }

  if (has_time) {
    if (precision_int < PRECISION_HOUR) {
      out <- set_hour(out, get_hour(from))
    }
    if (precision_int < PRECISION_MINUTE) {
      out <- set_minute(out, get_minute(from))
    }
    if (precision_int < PRECISION_SECOND) {
      out <- set_second(out, get_second(from))
    }
  }

  out
}

check_number_of_supplied_optional_arguments <- function(to, by, total_size) {
  has_to <- !is_null(to)
  has_by <- !is_null(by)
  has_ts <- !is_null(total_size)

  n_has <- sum(has_to, has_by, has_ts)

  if (n_has != 2L) {
    message <- paste0(
      "Must specify exactly two of:\n",
      "- `to`\n",
      "- `by`\n",
      "- `total_size`"
    )
    abort(message)
  }

  invisible()
}

# ------------------------------------------------------------------------------

#' Counting: date and date-time
#'
#' @description
#' `date_count_between()` counts the number of `precision` units between
#' `start` and `end` (i.e., the number of years or months or hours). This count
#' corresponds to the _whole number_ of units, and will never return a
#' fractional value.
#'
#' This is suitable for, say, computing the whole number of years or months
#' between two dates, accounting for the day and time of day.
#'
#' There are separate help pages for counting for dates and date-times:
#'
#' - [dates (Date)][date-count-between]
#'
#' - [date-times (POSIXct/POSIXlt)][posixt-count-between]
#'
#' @inheritSection calendar_count_between Comparison Direction
#'
#' @inheritParams calendar_count_between
#'
#' @param start,end `[Date / POSIXct / POSIXlt]`
#'
#'   A pair of date or date-time vectors. These will be recycled to their common
#'   size.
#'
#' @inherit calendar_count_between return
#'
#' @export
#' @examples
#' # See method specific documentation for more examples
#'
#' start <- date_parse("2000-05-05")
#' end <- date_parse(c("2020-05-04", "2020-05-06"))
#'
#' # Age in years
#' date_count_between(start, end, "year")
#'
#' # Number of "whole" months between these dates
#' date_count_between(start, end, "month")
date_count_between <- function(start, end, precision, ..., n = 1L) {
  UseMethod("date_count_between")
}

#' Counting: date
#'
#' @description
#' This is a Date method for the [date_count_between()] generic.
#'
#' `date_count_between()` counts the number of `precision` units between
#' `start` and `end` (i.e., the number of years or months). This count
#' corresponds to the _whole number_ of units, and will never return a
#' fractional value.
#'
#' This is suitable for, say, computing the whole number of years or months
#' between two dates, accounting for the day of the month.
#'
#' _Calendrical based counting:_
#'
#' These precisions convert to a year-month-day calendar and count while in that
#' type.
#'
#' - `"year"`
#'
#' - `"quarter"`
#'
#' - `"month"`
#'
#' _Time point based counting:_
#'
#' These precisions convert to a time point and count while in that type.
#'
#' - `"week"`
#'
#' - `"day"`
#'
#' For dates, whether a calendar or time point is used is not all that
#' important, but is is fairly important for date-times.
#'
#' @details
#' `"quarter"` is equivalent to `"month"` precision with `n` set to `n * 3L`.
#'
#' @inheritSection calendar_count_between Comparison Direction
#'
#' @inheritParams date_count_between
#'
#' @param start,end `[Date]`
#'
#'   A pair of date vectors. These will be recycled to their common
#'   size.
#'
#' @param precision `[character(1)]`
#'
#'   One of:
#'
#'   - `"year"`
#'   - `"quarter"`
#'   - `"month"`
#'   - `"week"`
#'   - `"day"`
#'
#' @inherit date_count_between return
#'
#' @name date-count-between
#'
#' @export
#' @examples
#' start <- date_parse("2000-05-05")
#' end <- date_parse(c("2020-05-04", "2020-05-06"))
#'
#' # Age in years
#' date_count_between(start, end, "year")
#'
#' # Number of "whole" months between these dates. i.e.
#' # `2000-05-05 -> 2020-04-05` is 239 months
#' # `2000-05-05 -> 2020-05-05` is 240 months
#' # Since 2020-05-04 occurs before the 5th of that month,
#' # it gets a count of 239
#' date_count_between(start, end, "month")
#'
#' # Number of "whole" quarters between (same as `"month"` with `n * 3`)
#' date_count_between(start, end, "quarter")
#' date_count_between(start, end, "month", n = 3)
#'
#' # Number of days between
#' date_count_between(start, end, "day")
#'
#' # Number of full 3 day periods between these two dates
#' date_count_between(start, end, "day", n = 3)
#'
#' # Essentially the truncated value of this
#' date_count_between(start, end, "day") / 3
#'
#' # ---------------------------------------------------------------------------
#'
#' # Breakdown into full years, months, and days between
#' x <- start
#'
#' years <- date_count_between(x, end, "year")
#' x <- add_years(x, years)
#'
#' months <- date_count_between(x, end, "month")
#' x <- add_months(x, months)
#'
#' days <- date_count_between(x, end, "day")
#' x <- add_days(x, days)
#'
#' data.frame(
#'   start = start,
#'   end = end,
#'   years = years,
#'   months = months,
#'   days = days
#' )
#'
#' # Note that when breaking down a date like that, you may need to
#' # set `invalid` during intermediate calculations
#' start <- date_build(2019, c(3, 3, 4), c(30, 31, 1))
#' end <- date_build(2019, 5, 05)
#'
#' # These are 1 month apart (plus a few days)
#' months <- date_count_between(start, end, "month")
#'
#' # But adding that 1 month to `start` results in an invalid date
#' try(add_months(start, months))
#'
#' # You can choose various ways to resolve this
#' start_previous <- add_months(start, months, invalid = "previous")
#' start_next <- add_months(start, months, invalid = "next")
#'
#' days_previous <- date_count_between(start_previous, end, "day")
#' days_next <- date_count_between(start_next, end, "day")
#'
#' # Resulting in slightly different day values.
#' # No result is "perfect". Choosing "previous" or "next" both result
#' # in multiple `start` dates having the same month/day breakdown values.
#' data.frame(
#'   start = start,
#'   end = end,
#'   months = months,
#'   days_previous = days_previous,
#'   days_next = days_next
#' )
date_count_between.Date <- function(start, end, precision, ..., n = 1L) {
  check_dots_empty()

  if (!is_Date(end)) {
    abort("`end` must be a <Date>.")
  }

  # Designed to match `add_*()` functions to guarantee that
  # if `start <= end`, then `start + <count> <= end`
  allowed_precisions_calendar <- c(
    PRECISION_YEAR, PRECISION_QUARTER, PRECISION_MONTH
  )
  allowed_precisions_naive_time <- c(
    PRECISION_WEEK, PRECISION_DAY
  )
  allowed_precisions_sys_time <- c(
  )

  date_count_between_impl(
    start = start,
    end = end,
    precision = precision,
    n = n,
    allowed_precisions_calendar = allowed_precisions_calendar,
    allowed_precisions_naive_time = allowed_precisions_naive_time,
    allowed_precisions_sys_time = allowed_precisions_sys_time
  )
}

date_count_between_impl <- function(start,
                                    end,
                                    precision,
                                    n,
                                    allowed_precisions_calendar,
                                    allowed_precisions_naive_time,
                                    allowed_precisions_sys_time) {
  precision_int <- validate_precision_string(precision)

  if (precision_int %in% allowed_precisions_calendar) {
    start <- as_year_month_day(start)
    end <- as_year_month_day(end)
    out <- calendar_count_between(start, end, precision, n = n)
    return(out)
  }

  if (precision_int %in% allowed_precisions_naive_time) {
    start <- as_naive_time(start)
    end <- as_naive_time(end)
    out <- time_point_count_between(start, end, precision, n = n)
    return(out)
  }

  if (precision_int %in% allowed_precisions_sys_time) {
    start <- as_sys_time(start)
    end <- as_sys_time(end)
    out <- time_point_count_between(start, end, precision, n = n)
    return(out)
  }

  precisions <- c(
    allowed_precisions_calendar,
    allowed_precisions_naive_time,
    allowed_precisions_sys_time
  )
  precisions <- vapply(precisions, precision_to_string, character(1))
  precisions <- encodeString(precisions, quote = "'")
  precisions <- paste0(precisions, collapse = ", ")

  abort(paste0("`precision` must be one of: ", precisions, "."))
}

# ------------------------------------------------------------------------------
# as_date()

test_that("can convert Date -> Date", {
  expect_identical(as_date(new_date(0)), new_date(0))
})

test_that("integer Dates are normalized to double", {
  # seq.Date() can make these
  x <- 0L
  class(x) <- "Date"

  expect_identical(as_date(x), new_date(0))
})

test_that("can convert POSIXct / POSIXlt -> Date (Date assumed to be naive)", {
  ct <- date_time_parse("2019-01-01 23:00:00", "America/New_York")
  lt <- as.POSIXlt(ct)

  expect <- date_parse("2019-01-01")

  expect_identical(as_date(ct), expect)
  expect_identical(as_date(lt), expect)
})

test_that("can convert calendar -> Date", {
  expect_identical(as_date(year_month_day(1970, 1, 2)), new_date(1))
  expect_identical(as_date(year_quarter_day(1970, 1, 2)), new_date(1))
})

test_that("can convert sys-time -> Date", {
  expect_identical(as_date(sys_seconds(0)), new_date(0))
})

test_that("can convert naive-time -> Date", {
  expect_identical(as_date(naive_seconds(0)), new_date(0))
})

test_that("can convert zoned-time -> Date", {
  x <- as_zoned_time(naive_time_parse("2019-01-01T23:02:03"), "America/New_York")
  expect <- date_parse("2019-01-01")
  expect_identical(as_date(x), expect)
})

# ------------------------------------------------------------------------------
# as.Date()

test_that("invalid dates must be resolved when converting to a Date", {
  expect_snapshot_error(as.Date(year_month_day(2019, 2, 31)))
})

test_that("conversion from zoned-time uses naive-time as an intermediate", {
  x <- as_naive_time(year_month_day(2019, 12, 31, 23, 30, 00))
  x <- as_zoned_time(x, "America/New_York")

  expect_identical(as.Date(x), as.Date("2019-12-31"))
})

test_that("casting to Date floors components more precise than days (#205)", {
  x <- naive_time_parse("1969-01-01T13", precision = "hour")

  expect_identical(
    as.Date(x),
    date_build(1969, 1, 1)
  )
})

# ------------------------------------------------------------------------------
# as_sys_time()

test_that("converting to sys-time floors fractional dates (#191)", {
  x <- new_date(c(-0.5, 1.5))
  y <- new_date(c(-1, 1))

  expect_identical(as_sys_time(x), as_sys_time(y))
  expect_identical(as.Date(as_sys_time(x)), y)
})

test_that("converting to sys-time works with integer storage dates", {
  # These can occur from `seq.Date(from, to, length.out)`
  x <- structure(1L, class = "Date")
  y <- new_date(1)

  expect_identical(as_sys_time(x), as_sys_time(y))
  expect_identical(as.Date(as_sys_time(x)), y)
})

# ------------------------------------------------------------------------------
# as_zoned_time()

test_that("Dates are assumed to be naive", {
  x <- new_date(0)
  nt <- as_naive_time(year_month_day(1970, 1, 1))

  expect_identical(as_zoned_time(x, "UTC"), as_zoned_time(nt, "UTC"))
  expect_identical(as_zoned_time(x, "America/New_York"), as_zoned_time(nt, "America/New_York"))
})

test_that("can resolve nonexistent midnight issues", {
  # In Asia/Beirut, DST gap from 2021-03-27 23:59:59 -> 2021-03-28 01:00:00
  zone <- "Asia/Beirut"
  x <- as.Date("2021-03-28")

  expect_snapshot_error(as_zoned_time(x, zone), class = "clock_error_nonexistent_time")

  expect_identical(
    as_zoned_time(x, zone, nonexistent = "roll-forward"),
    as_zoned_time(as_naive_time(year_month_day(2021, 03, 28, 1)), zone)
  )
})

test_that("can resolve ambiguous midnight issues", {
  # In Asia/Amman, DST fallback from 2021-10-29 00:59:59 -> 2021-10-29 00:00:00
  zone <- "Asia/Amman"
  x <- as.Date("2021-10-29")

  expect_snapshot_error(as_zoned_time(x, zone), class = "clock_error_ambiguous_time")

  expect_identical(
    as_zoned_time(x, zone, ambiguous = "earliest"),
    zoned_time_parse_complete("2021-10-29T00:00:00+03:00[Asia/Amman]")
  )
  expect_identical(
    as_zoned_time(x, zone, ambiguous = "latest"),
    zoned_time_parse_complete("2021-10-29T00:00:00+02:00[Asia/Amman]")
  )
})

# ------------------------------------------------------------------------------
# as_weekday()

test_that("can convert to weekday", {
  x <- as.Date("2019-01-01")
  expect_identical(as_weekday(x), weekday(3L))
})

test_that("converting to weekday retains names", {
  x <- c(x = as.Date("2019-01-01"))
  expect_named(as_weekday(x), names(x))
})

# ------------------------------------------------------------------------------
# date_group()

test_that("can group by year/month/day", {
  x <- as.Date("2019-01-01") + c(0, 1, 2, 3)

  expect_identical(date_group(x, "day"), x)
  expect_identical(date_group(x, "day", n = 2), x[c(1, 1, 3, 3)])

  y <- as.Date(c("2019-01-01", "2019-02-05", "2019-03-10"))
  expect <- set_day(y, 1)

  expect_identical(date_group(y, "month"), expect)
  expect_identical(date_group(y, "month", n = 2), expect[c(1, 1, 3)])

  # Using year 0 as the starting point
  z <- as.Date(c("0000-01-05", "0001-02-10", "0002-03-12", "0003-01-15", "0004-12-10"))
  expect <- set_month(set_day(z, 1), 1)

  expect_identical(date_group(z, "year"), expect)
  expect_identical(date_group(z, "year", n = 3), expect[c(1, 1, 1, 4, 4)])
})

test_that("can't group by finer precisions", {
  x <- as.Date("2019-01-01")
  expect_snapshot_error(date_group(x, "hour"))
  expect_snapshot_error(date_group(x, "nanosecond"))
})

test_that("can't group by non-year-month-day precisions", {
  x <- as.Date("2019-01-01")
  expect_snapshot_error(date_group(x, "quarter"))
})

# ------------------------------------------------------------------------------
# date_leap_year()

test_that("can detect leap years", {
  x <- c("2019-01-01", "2020-01-01", NA)
  x <- as.Date(x)
  expect_identical(date_leap_year(x), c(FALSE, TRUE, NA))
})

# ------------------------------------------------------------------------------
# date_floor()

test_that("can floor by weeks", {
  x <- c("1970-01-01", "1970-01-07", "1970-01-08")
  x <- as.Date(x)
  expect_identical(date_floor(x, "week"), x[c(1, 1, 3)])
  expect_identical(date_floor(x, "week", n = 2), x[c(1, 1, 1)])
})

test_that("can only floor by week/day", {
  expect_snapshot_error(date_floor(as.Date("2019-01-01"), "hour"))
  expect_snapshot_error(date_floor(as.Date("2019-01-01"), "month"))
})

test_that("`origin` can be used", {
  origin <- as.Date("1970-01-02")
  x <- as.Date(c("1970-01-01", "1970-01-02", "1970-01-03"))
  expect <- as.Date(c("1969-12-31", "1970-01-02", "1970-01-02"))
  expect_identical(date_floor(x, "day", n = 2, origin = origin), expect)
})

test_that("`origin` is validated", {
  x <- as.Date("2019-01-01")
  expect_snapshot_error(date_floor(x, "day", origin = 1))
  expect_snapshot_error(date_floor(x, "day", origin = new_date(NA_real_)))
  expect_snapshot_error(date_floor(x, "day", origin = new_date(Inf)))
  expect_snapshot_error(date_floor(x, "day", origin = new_date(c(0, 1))))
})

# ------------------------------------------------------------------------------
# date_ceiling()

test_that("can ceiling", {
  x <- c("1970-01-01", "1970-01-07", "1970-01-08")
  x <- as.Date(x)
  expect_identical(date_ceiling(x, "week"), x[c(1, 3, 3)])
  expect_identical(date_ceiling(x, "week", n = 2), as.Date(c("1970-01-01", "1970-01-15", "1970-01-15")))
})

# ------------------------------------------------------------------------------
# date_round()

test_that("can round", {
  x <- c("1970-01-01", "1970-01-03", "1970-01-05", "1970-01-08")
  x <- as.Date(x)
  expect_identical(date_round(x, "week"), x[c(1, 1, 4, 4)])
  expect_identical(date_round(x, "week", n = 2), as.Date(c("1970-01-01", "1970-01-01", "1970-01-01", "1970-01-15")))
})

# ------------------------------------------------------------------------------
# date_weekday_factor()

test_that("can convert to a weekday factor", {
  x <- as.Date("2019-01-01") + 0:6

  expect_identical(
    date_weekday_factor(x),
    weekday_factor(as_weekday(x)),
  )
  expect_identical(
    date_weekday_factor(x, labels = "fr", abbreviate = FALSE, encoding = "iso"),
    weekday_factor(as_weekday(x), labels = "fr", abbreviate = FALSE, encoding = "iso"),
  )
})

# ------------------------------------------------------------------------------
# date_month_factor()

test_that("can convert to a month factor", {
  x <- add_months(as.Date("2019-01-01"), 0:11)

  expect_identical(
    date_month_factor(x, labels = "fr", abbreviate = TRUE),
    calendar_month_factor(as_year_month_day(x), labels = "fr", abbreviate = TRUE),
  )
})

# ------------------------------------------------------------------------------
# date_format()

test_that("can format dates", {
  x <- as.Date("2018-12-31")
  formats <- test_all_formats(zone = FALSE)

  expect_snapshot(
    vapply(
      X = formats,
      FUN = function(format) date_format(x, format = format),
      FUN.VALUE = character(1)
    )
  )
  expect_snapshot(
    vapply(
      X = formats,
      FUN = function(format) date_format(x, format = format, locale = clock_locale("fr")),
      FUN.VALUE = character(1)
    )
  )
})

test_that("formatting Dates with `%z` or `%Z` returns NA with a warning", {
  x <- as.Date("2018-01-01")

  expect_warning(date_format(x, format = "%z"), class = "clock_warning_format_failures")
  expect_snapshot(date_format(x, format = "%z"))

  expect_warning(date_format(x, format = "%Z"), class = "clock_warning_format_failures")
  expect_snapshot(date_format(x, format = "%Z"))
})

# ------------------------------------------------------------------------------
# date_parse()

test_that("`%z` and `%Z` commands are ignored", {
  expect_identical(
    date_parse("2019-12-31 -0500", format = "%Y-%m-%d %z"),
    as.Date("2019-12-31")
  )
  expect_identical(
    date_parse("2019-12-31 America/New_York", format = "%Y-%m-%d %Z"),
    as.Date("2019-12-31")
  )
})

test_that("parsing into a date if you requested to parse time components rounds the time (#207) (#230) (undocumented)", {
  expect_identical(
    date_parse("2019-12-31 11:59:59", format = "%Y-%m-%d %H:%M:%S"),
    as.Date("2019-12-31")
  )
  expect_identical(
    date_parse("2019-12-31 12:00:00", format = "%Y-%m-%d %H:%M:%S"),
    as.Date("2020-01-01")
  )
})

test_that("parsing fails when undocumented rounding behavior would result in invalid 60 second component (#230) (undocumented)", {
  expect_warning(
    expect_identical(
      date_parse("2019-01-01 01:01:59.550", format = "%Y-%m-%d %H:%M:%6S"),
      new_date(NA_real_)
    ),
    class = "clock_warning_parse_failures"
  )
})

test_that("failure to parse throws a warning", {
  expect_warning(date_parse("foo"), class = "clock_warning_parse_failures")
  expect_snapshot(date_parse("foo"))
})

# ------------------------------------------------------------------------------
# date_shift()

test_that("can shift dates", {
  x <- date_parse("2019-01-01")
  x <- x + c(0, 1)

  monday <- weekday(clock_weekdays$monday)
  tuesday <- weekday(clock_weekdays$tuesday)

  expect_identical(
    date_shift(x, monday),
    add_days(x, c(6, 5))
  )
  expect_identical(
    date_shift(x, monday, which = "previous"),
    add_days(x, c(-1, -2))
  )
  expect_identical(
    date_shift(x, tuesday, boundary = "advance"),
    add_days(x, c(7, 6))
  )
})

# ------------------------------------------------------------------------------
# date_build()

test_that("can build a Date", {
  expect_identical(date_build(2019), as.Date("2019-01-01"))
  expect_identical(date_build(2020, 2, 3), as.Date("2020-02-03"))
})

test_that("can handle invalid dates", {
  expect_snapshot_error(date_build(2019, 1:12, 31))

  expect_identical(
    date_build(2019, 1:12, 31, invalid = "previous"),
    date_build(2019, 1:12, "last")
  )
})

# ------------------------------------------------------------------------------
# date_today()

test_that("can get the current date", {
  x <- date_today("America/New_York")
  expect_length(x, 1)
  expect_s3_class(x, "Date")
})

# ------------------------------------------------------------------------------
# date_start()

test_that("can get the start", {
  x <- date_build(2019, 2, 2)
  expect_identical(date_start(x, "day"), x)
  expect_identical(date_start(x, "month"), date_build(2019, 2, 1))
  expect_identical(date_start(x, "year"), date_build(2019, 1, 1))
})

test_that("start: can't use invalid precisions", {
  expect_snapshot_error(date_start(date_build(2019), "quarter"))
})

# ------------------------------------------------------------------------------
# date_end()

test_that("can get the end", {
  x <- date_build(2019:2020, 2, 2)
  expect_identical(date_end(x, "day"), x)
  expect_identical(date_end(x, "month"), date_build(2019:2020, 2, 28:29))
  expect_identical(date_end(x, "year"), date_build(2019:2020, 12, 31))
})

test_that("end: can't use invalid precisions", {
  expect_snapshot_error(date_end(date_build(2019), "quarter"))
})

# ------------------------------------------------------------------------------
# date_seq()

test_that("integer `by` means day precision", {
  expect_identical(
    date_seq(new_date(0), to = new_date(5), by = 1),
    date_seq(new_date(0), to = new_date(5), by = duration_days(1))
  )
})

test_that("day precision seq works", {
  expect_identical(
    date_seq(new_date(0), to = new_date(5), by = duration_days(2)),
    new_date(c(0, 2, 4))
  )
})

test_that("granular `by` are allowed", {
  expect_identical(
    date_seq(date_build(2019, 1, 1), to = date_build(2019, 6, 1), by = duration_months(2)),
    date_build(2019, c(1, 3, 5), 1)
  )
  expect_identical(
    date_seq(date_build(2019, 1, 1), to = date_build(2025, 1, 1), by = duration_years(2)),
    date_build(c(2019, 2021, 2023, 2025), 1, 1)
  )
})

test_that("combining `by` with `total_size` works", {
  expect_identical(
    date_seq(date_build(2019, 1, 1), by = 2, total_size = 3),
    date_build(2019, 1, c(1, 3, 5))
  )
})

test_that("combining `to` with `total_size` works", {
  expect_identical(
    date_seq(date_build(2019, 1, 1), to = date_build(2019, 1, 5), total_size = 3),
    date_build(2019, 1, c(1, 3, 5))
  )
})

test_that("can resolve invalid dates", {
  from <- date_build(2019, 1, 31)
  to <- date_build(2019, 5, 31)

  expect_snapshot_error(date_seq(from, to = to, by = duration_months(1)))

  expect_identical(
    date_seq(from, to = to, by = duration_months(1), invalid = "previous"),
    date_build(2019, 1:5, "last")
  )
})

test_that("quarterly `by` works", {
  expect_identical(
    date_seq(date_build(2019, 1, 2), by = duration_quarters(1), total_size = 3),
    date_seq(date_build(2019, 1, 2), by = duration_months(3), total_size = 3)
  )
})

test_that("weekly `by` works", {
  expect_identical(
    date_seq(date_build(2019, 1, 2), by = duration_weeks(1), total_size = 3),
    date_seq(date_build(2019, 1, 2), by = duration_days(7), total_size = 3)
  )
})

test_that("components of `from` more precise than `by` are restored", {
  expect_identical(
    date_seq(date_build(2019, 2, 3), by = duration_months(1), total_size = 2),
    date_build(2019, 2:3, 3)
  )
  expect_identical(
    date_seq(date_build(2019, 2, 3), by = duration_years(1), total_size = 2),
    date_build(2019:2020, 2, 3)
  )
})

test_that("seq() with `from > to && by > 0` or `from < to && by > 0` results in length 0 output (#282)", {
  expect_identical(
    date_seq(date_build(2019, 1, 2), to = date_build(2019, 1, 1), by = 1),
    date_build(integer())
  )

  expect_identical(
    date_seq(date_build(2019), to = date_build(2020), by = -1),
    date_build(integer())
  )
})

test_that("components of `to` more precise than `by` must match `from`", {
  expect_snapshot_error(date_seq(date_build(2019, 1, 1), to = date_build(2019, 2, 2), by = duration_months(1)))
  expect_snapshot_error(date_seq(date_build(2019, 1, 1), to = date_build(2019, 3, 1), by = duration_years(1)))
})

test_that("validates integerish `by`", {
  expect_snapshot_error(date_seq(new_date(1), by = 1.5, total_size = 1))
})

test_that("validates `total_size` early", {
  expect_snapshot_error(date_seq(new_date(1), by = 1, total_size = 1.5))
  expect_snapshot_error(date_seq(new_date(1), by = 1, total_size = NA))
  expect_snapshot_error(date_seq(new_date(1), by = 1, total_size = -1))
})

test_that("`to` and `total_size` must not generate a non-fractional sequence", {
  expect_snapshot_error(date_seq(new_date(0), to = new_date(3), total_size = 3))
})

test_that("requires exactly two optional arguments", {
  expect_snapshot_error(date_seq(new_date(1), by = 1))
  expect_snapshot_error(date_seq(new_date(1), total_size = 1))
  expect_snapshot_error(date_seq(new_date(1), to = new_date(1)))
})

test_that("requires `to` to be Date", {
  expect_snapshot_error(date_seq(new_date(1), to = 1, by = 1))
})

test_that("requires year, month, or day precision", {
  expect_snapshot_error(date_seq(new_date(1), to = new_date(2), by = duration_nanoseconds(1)))
})

test_that("checks empty dots", {
  expect_snapshot_error(date_seq(new_date(1), new_date(2)))
})

test_that("golden test: ensure that we never allow components of `to` to differ with `from` (#224)", {
  from <- date_build(1970, 01, 31)
  to <- date_build(1970, 03, 01)

  expect_error(date_seq(from, to = to, by = duration_months(1), invalid = "overflow"))

  # Could theoretically generate this, where the second element is past `to`
  #> "1970-01-31" "1970-03-03"
})

# ------------------------------------------------------------------------------
# date_count_between()

test_that("can compute precisions at year / quarter / month / week / day", {
  x <- date_build(2019, 1, 5)

  y <- date_build(2025, 1, c(4, 6))
  expect_identical(date_count_between(x, y, "year"), c(5L, 6L))
  expect_identical(date_count_between(x, y, "quarter"), c(23L, 24L))
  expect_identical(date_count_between(x, y, "month"), c(71L, 72L))

  y <- date_build(2019, 1, c(25, 26))
  expect_identical(date_count_between(x, y, "week"), c(2L, 3L))
  expect_identical(date_count_between(x, y, "day"), c(20L, 21L))
})

test_that("must use a valid Date precision", {
  x <- date_build(2019)
  expect_snapshot((expect_error(date_count_between(x, x, "hour"))))
})

test_that("can't count between a Date and a POSIXt", {
  x <- date_build(2019)
  y <- date_time_build(2019, zone = "UTC")
  expect_snapshot((expect_error(date_count_between(x, y, "year"))))
})

# ------------------------------------------------------------------------------
# date_zone()

test_that("cannot get the zone of a Date", {
  expect_snapshot_error(date_zone(new_date(0)))
})

# ------------------------------------------------------------------------------
# date_set_zone()

test_that("cannot set the zone of a Date", {
  expect_snapshot_error(date_set_zone(new_date(0), "UTC"))
})

# ------------------------------------------------------------------------------
# vec_arith()

test_that("<date> op <duration>", {
  expect_identical(vec_arith("+", new_date(0), duration_years(1)), new_date(365))
  expect_identical(vec_arith("-", new_date(0), duration_years(1)), new_date(-365))

  expect_snapshot_error(vec_arith("+", new_date(0), duration_hours(1)))
  expect_snapshot_error(vec_arith("*", new_date(0), duration_years(1)))
})

test_that("<duration> op <date>", {
  expect_identical(vec_arith("+", duration_years(1), new_date(0)), new_date(365))

  expect_snapshot_error(vec_arith("-", duration_years(1), new_date(0)))
  expect_snapshot_error(vec_arith("+", duration_hours(1), new_date(0)))
  expect_snapshot_error(vec_arith("*", duration_years(1), new_date(0)))
})

# ------------------------------------------------------------------------------
# as_date_time()

test_that("can convert POSIXct -> POSIXct", {
  x <- new_datetime(0, tzone = "America/New_York")
  expect_identical(as_date_time(x), x)
})

test_that("can convert POSIXlt -> POSIXct", {
  expect <- new_datetime(0, tzone = "America/New_York")
  x <- as.POSIXlt(expect)
  expect_identical(as_date_time(x), expect)
})

test_that("integer POSIXct are normalized to double", {
  x <- 0L
  class(x) <- c("POSIXct", "POSIXt")
  attr(x, "tzone") <- "America/New_York"

  expect <- new_datetime(0, tzone = "America/New_York")

  expect_identical(as_date_time(x), expect)
})

test_that("can't accidentally supply `zone` to reinterpret date-time in new zone", {
  expect_snapshot_error(as_date_time(new_datetime(0), zone = "America/New_York"))
})

test_that("can convert Date -> POSIXct (Date assumed to be naive)", {
  x <- date_parse("2019-01-01")

  expect_identical(as_date_time(x, "America/New_York"), date_time_parse("2019-01-01 00:00:00", "America/New_York"))
  expect_identical(as_date_time(x, "Europe/London"), date_time_parse("2019-01-01 00:00:00", "Europe/London"))
})

test_that("can convert calendar -> POSIXct", {
  zone <- "America/New_York"
  expect_identical(as_date_time(year_month_day(1970, 1, 2), zone), new_datetime(104400, zone))
  expect_identical(as_date_time(year_quarter_day(1970, 1, 2), zone), new_datetime(104400, zone))
})

test_that("can convert sys-time -> POSIXct", {
  expect_identical(as_date_time(sys_seconds(0), "UTC"), new_datetime(0, "UTC"))
  expect_identical(as_date_time(sys_seconds(0), "America/New_York"), new_datetime(0, "America/New_York"))
})

test_that("can convert naive-time -> POSIXct", {
  expect_identical(as_date_time(naive_seconds(0), "UTC"), new_datetime(0, "UTC"))
  expect_identical(as_date_time(naive_seconds(0), "America/New_York"), new_datetime(18000, "America/New_York"))
})

test_that("can convert zoned-time -> POSIXct", {
  x <- as_zoned_time(naive_time_parse("2019-01-01T23:02:03"), "America/New_York")
  expect <- date_time_parse("2019-01-01 23:02:03", "America/New_York")
  expect_identical(as_date_time(x), expect)
})

test_that("can resolve nonexistent midnight issues for Date -> POSIXct", {
  # In Asia/Beirut, DST gap from 2021-03-27 23:59:59 -> 2021-03-28 01:00:00
  zone <- "Asia/Beirut"
  x <- as.Date("2021-03-28")

  expect_snapshot_error(as_date_time(x, zone), class = "clock_error_nonexistent_time")

  expect_identical(
    as_date_time(x, zone, nonexistent = "roll-forward"),
    as_date_time(as_naive_time(year_month_day(2021, 03, 28, 1)), zone)
  )
})

test_that("can resolve ambiguous midnight issues for Date -> POSIXct", {
  # In Asia/Amman, DST fallback from 2021-10-29 00:59:59 -> 2021-10-29 00:00:00
  zone <- "Asia/Amman"
  x <- as.Date("2021-10-29")

  expect_snapshot_error(as_date_time(x, zone), class = "clock_error_ambiguous_time")

  expect_identical(
    as_date_time(x, zone, ambiguous = "earliest"),
    date_time_parse_complete("2021-10-29T00:00:00+03:00[Asia/Amman]")
  )
  expect_identical(
    as_date_time(x, zone, ambiguous = "latest"),
    date_time_parse_complete("2021-10-29T00:00:00+02:00[Asia/Amman]")
  )
})

# ------------------------------------------------------------------------------
# as.POSIXct()

test_that("casting to POSIXct floors components more precise than seconds (#205)", {
  x <- naive_time_parse("1969-01-01T01:01:01.678", precision = "millisecond")

  expect_identical(
    as.POSIXct(x, "America/New_York"),
    date_time_parse("1969-01-01 01:01:01", "America/New_York")
  )
})

test_that("casting to POSIXct with time point less precise than seconds works (#278)", {
  x <- as_naive_time(year_month_day(2019, 1, 1, 1))

  expect_identical(
    as.POSIXct(x, "America/New_York"),
    date_time_parse("2019-01-01 01:00:00", "America/New_York")
  )

  x <- as_sys_time(year_month_day(2019, 1, 1, 1))

  expect_identical(
    as.POSIXct(x, "America/New_York"),
    date_time_parse("2018-12-31 20:00:00", "America/New_York")
  )
})

# ------------------------------------------------------------------------------
# as_weekday()

test_that("can convert to weekday", {
  x <- as.POSIXct("2019-01-01", tz = "America/New_York")
  expect_identical(as_weekday(x), weekday(3L))
  expect_identical(as_weekday(as.POSIXlt(x)), weekday(3L))
})

test_that("converting to weekday retains names", {
  x <- c(x = as.POSIXct("2019-01-01", tz = "America/New_York"))
  expect_named(as_weekday(x), names(x))
  expect_named(as_weekday(as.POSIXlt(x)), names(x))
})

# ------------------------------------------------------------------------------
# date_group()

test_that("can group by year/month/day/hour/minute/second", {
  x <- as.POSIXct("2019-01-01", "America/New_York")
  x <- add_days(x, 0:3)

  expect_identical(date_group(x, "day"), x)
  expect_identical(date_group(x, "day", n = 2), x[c(1, 1, 3, 3)])

  y <- as.POSIXct(c("2019-01-01", "2019-02-05", "2019-03-10"), "America/New_York")
  expect <- set_day(y, 1)

  expect_identical(date_group(y, "month"), expect)
  expect_identical(date_group(y, "month", n = 2), expect[c(1, 1, 3)])

  # Using year 0 as the starting point
  z <- as.POSIXct(c("0000-01-05", "0001-02-10", "0002-03-12", "0003-01-15", "0004-12-10"), "America/New_York")
  expect <- set_month(set_day(z, 1), 1)

  expect_identical(date_group(z, "year"), expect)
  expect_identical(date_group(z, "year", n = 3), expect[c(1, 1, 1, 4, 4)])

  w <- as.POSIXct("2019-01-01", "America/New_York")
  w <- add_hours(w, c(0, 1, 2, 4, 5))

  expect_identical(date_group(w, "hour", n = 2), w[c(1, 1, 3, 4, 4)])
})

test_that("can handle nonexistent times resulting from grouping", {
  x <- as.POSIXct("1970-04-26 03:30:00", "America/New_York")
  expect <- set_minute(x, 0)

  expect_snapshot_error(date_group(x, "hour", n = 2))
  expect_identical(date_group(x, "hour", n = 2, nonexistent = "roll-forward"), expect)
})

test_that("can't group by finer precisions", {
  x <- as.POSIXct("2019-01-01", "America/New_York")
  expect_snapshot_error(date_group(x, "nanosecond"))
})

test_that("can't group by non-year-month-day precisions", {
  x <- as.POSIXct("2019-01-01", "America/New_York")
  expect_snapshot_error(date_group(x, "quarter"))
})

# ------------------------------------------------------------------------------
# date_leap_year()

test_that("can detect leap years", {
  x <- c("2019-01-01", "2020-01-01", NA)
  x <- as.POSIXct(x, tz = "America/New_York")
  expect_identical(date_leap_year(x), c(FALSE, TRUE, NA))
})

# ------------------------------------------------------------------------------
# date_floor()

test_that("can floor by weeks", {
  x <- c("1970-01-01", "1970-01-07", "1970-01-08")
  x <- as.POSIXct(x, "America/New_York")
  expect_identical(date_floor(x, "week"), x[c(1, 1, 3)])
  expect_identical(date_floor(x, "week", n = 2), x[c(1, 1, 1)])
})

test_that("flooring can handle nonexistent times", {
  x <- as.POSIXct("1970-04-26 01:59:59", "America/New_York") + c(0, 1)
  expect <- as.POSIXct(c("1970-04-26 00:00:00", "1970-04-26 03:00:00"), "America/New_York")

  expect_snapshot_error(date_floor(x, "hour", n = 2))
  expect_identical(date_floor(x, "hour", n = 2, nonexistent = "roll-forward"), expect)
})

test_that("flooring can handle ambiguous times", {
  x <- as.POSIXct("1970-10-25 00:30:00", "America/New_York") + c(3600, 7200)
  expect <- as.POSIXct("1970-10-25 00:00:00", "America/New_York") + c(3600, 3600)

  expect_identical(date_floor(x, "hour", ambiguous = "earliest"), expect)
})

test_that("`origin` can be used", {
  origin <- as.POSIXct("1970-01-02", "America/New_York")
  x <- as.POSIXct(c("1970-01-01", "1970-01-02", "1970-01-03"), "America/New_York")
  expect <- as.POSIXct(c("1969-12-31", "1970-01-02", "1970-01-02"), "America/New_York")
  expect_identical(date_floor(x, "day", n = 2, origin = origin), expect)
})

test_that("`origin` is floored to the precision of `precision` with a potential warning before rounding", {
  origin <- as.POSIXct("1970-01-01 01:00:00", "America/New_York")
  x <- as.POSIXct(c("1970-01-01 00:00:00", "1970-01-02 00:00:00"), "America/New_York")

  expect_snapshot(date_floor(x, "day", origin = origin))
  expect_warning(date_floor(x, "day", origin = origin), class = "clock_warning_invalid_rounding_origin")
})

test_that("`origin` is validated", {
  zone <- "America/New_York"
  x <- as.POSIXct("2019-01-01", zone)
  expect_snapshot_error(date_floor(x, "day", origin = 1))
  expect_snapshot_error(date_floor(x, "day", origin = new_datetime(NA_real_, zone)))
  expect_snapshot_error(date_floor(x, "day", origin = new_datetime(Inf, zone)))
  expect_snapshot_error(date_floor(x, "day", origin = new_datetime(c(0, 1), zone)))
  expect_snapshot_error(date_floor(x, "day", origin = new_datetime(0, "")))
  expect_snapshot_error(date_floor(x, "day", origin = new_datetime(0, "America/Los_Angeles")))
})

# ------------------------------------------------------------------------------
# date_ceiling()

test_that("can ceiling", {
  x <- c("1970-01-01", "1970-01-07", "1970-01-08")
  x <- as.POSIXct(x, "America/New_York")
  expect_identical(date_ceiling(x, "week"), x[c(1, 3, 3)])
  expect_identical(date_ceiling(x, "week", n = 2), as.POSIXct(c("1970-01-01", "1970-01-15", "1970-01-15"), "America/New_York"))
})

# ------------------------------------------------------------------------------
# date_round()

test_that("can round", {
  x <- c("1970-01-01", "1970-01-03", "1970-01-05", "1970-01-08")
  x <- as.POSIXct(x, "America/New_York")
  expect_identical(date_round(x, "week"), x[c(1, 1, 4, 4)])
  expect_identical(date_round(x, "week", n = 2), as.POSIXct(c("1970-01-01", "1970-01-01", "1970-01-01", "1970-01-15"), "America/New_York"))
})

# ------------------------------------------------------------------------------
# date_weekday_factor()

test_that("can convert to a weekday factor", {
  x <- as.POSIXct("2019-01-01", tz = "America/New_York") + 0:6

  expect_identical(
    date_weekday_factor(x),
    weekday_factor(as_weekday(x)),
  )
  expect_identical(
    date_weekday_factor(x, labels = "fr", abbreviate = FALSE, encoding = "iso"),
    weekday_factor(as_weekday(x), labels = "fr", abbreviate = FALSE, encoding = "iso"),
  )
})

# ------------------------------------------------------------------------------
# date_month_factor()

test_that("can convert to a month factor", {
  x <- add_months(as.POSIXct("2019-01-01", tz = "America/New_York"), 0:11)

  expect_identical(
    date_month_factor(x, labels = "fr", abbreviate = TRUE),
    calendar_month_factor(as_year_month_day(x), labels = "fr", abbreviate = TRUE),
  )
})

# ------------------------------------------------------------------------------
# date_format()

test_that("default format is correct", {
  x <- date_time_parse("2019-01-01 00:00:00", "America/New_York")
  expect_snapshot(date_format(x))
})

test_that("can format date-times", {
  x <- as.POSIXct("2018-12-31 23:59:59", "America/New_York")
  formats <- test_all_formats()

  expect_snapshot(
    vapply(
      X = formats,
      FUN = function(format) date_format(x, format = format),
      FUN.VALUE = character(1)
    )
  )
  expect_snapshot(
    vapply(
      X = formats,
      FUN = function(format) date_format(x, format = format, locale = clock_locale("fr")),
      FUN.VALUE = character(1)
    )
  )
})

# ------------------------------------------------------------------------------
# date_zone()

test_that("can get the zone of a POSIXt", {
  ct <- as.POSIXct("2019-01-01", "America/New_York")
  lt <- as.POSIXlt(ct)

  expect_identical(date_zone(ct), "America/New_York")
  expect_identical(date_zone(lt), "America/New_York")
})

# ------------------------------------------------------------------------------
# date_set_zone()

test_that("can set the zone of a POSIXt", {
  ct <- as.POSIXct("2019-01-01", "America/New_York")
  lt <- as.POSIXlt(ct)

  expect <- as.POSIXct("2018-12-31 21:00:00", "America/Los_Angeles")

  expect_identical(date_set_zone(ct, "America/Los_Angeles"), expect)
  expect_identical(date_set_zone(lt, "America/Los_Angeles"), expect)
})

# ------------------------------------------------------------------------------
# date_time_parse()

test_that("can parse into a POSIXct", {
  expect_identical(
    date_time_parse("2019-12-31 23:59:59", "America/New_York"),
    as.POSIXct("2019-12-31 23:59:59", tz = "America/New_York")
  )
})

test_that("can resolve ambiguity and nonexistent times", {
  expect_snapshot_error(date_time_parse("1970-04-26 02:30:00", "America/New_York"))

  expect_identical(
    date_time_parse("1970-04-26 02:30:00", "America/New_York", nonexistent = "roll-forward"),
    date_time_parse("1970-04-26 03:00:00", "America/New_York")
  )

  expect_snapshot_error(date_time_parse("1970-10-25 01:30:00", "America/New_York"))

  expect_identical(
    date_time_parse("1970-10-25 01:30:00", "America/New_York", ambiguous = "earliest"),
    add_seconds(date_time_parse("1970-10-25 00:30:00", "America/New_York"), 3600)
  )
  expect_identical(
    date_time_parse("1970-10-25 01:30:00", "America/New_York", ambiguous = "latest"),
    add_seconds(date_time_parse("1970-10-25 00:30:00", "America/New_York"), 7200)
  )
})

test_that("failure to parse throws a warning", {
  expect_warning(date_time_parse("foo", "America/New_York"), class = "clock_warning_parse_failures")
  expect_snapshot(date_time_parse("foo", "America/New_York"))
})

# ------------------------------------------------------------------------------
# date_time_parse_complete()

test_that("can parse into a POSIXct", {
  expect_identical(
    date_time_parse_complete("2019-12-31T23:59:59-05:00[America/New_York]"),
    as.POSIXct("2019-12-31 23:59:59", tz = "America/New_York")
  )
})

test_that("ambiguity is resolved through the string", {
  expect_identical(
    date_time_parse_complete("1970-10-25T01:30:00-04:00[America/New_York]"),
    add_seconds(date_time_parse("1970-10-25 00:30:00", "America/New_York"), 3600)
  )
  expect_identical(
    date_time_parse_complete("1970-10-25T01:30:00-05:00[America/New_York]"),
    add_seconds(date_time_parse("1970-10-25 00:30:00", "America/New_York"), 7200)
  )
})

test_that("throws warning on failed parses", {
  expect_warning(date_time_parse_complete("foo"), class = "clock_warning_parse_failures")
  expect_snapshot(date_time_parse_complete("foo"))
})

# ------------------------------------------------------------------------------
# date_time_parse_abbrev()

test_that("can parse into a POSIXct using the abbrev", {
  expect_identical(
    date_time_parse_abbrev("2019-01-01 01:02:03 EST", "America/New_York"),
    as.POSIXct(as_naive_time(year_month_day(2019, 1, 1, 1, 2, 3)), "America/New_York")
  )
})

test_that("ambiguity is resolved through the string", {
  expect_identical(
    date_time_parse_abbrev("1970-10-25 01:30:00 EDT", "America/New_York"),
    add_seconds(date_time_parse("1970-10-25 00:30:00", "America/New_York"), 3600)
  )
  expect_identical(
    date_time_parse_abbrev("1970-10-25 01:30:00 EST", "America/New_York"),
    add_seconds(date_time_parse("1970-10-25 00:30:00", "America/New_York"), 7200)
  )
})

test_that("abbrev - throws warning on failed parses", {
  expect_warning(date_time_parse_abbrev("foo", "America/New_York"), class = "clock_warning_parse_failures")
  expect_snapshot(date_time_parse_abbrev("foo", "America/New_York"))
})

# ------------------------------------------------------------------------------
# date_time_parse_RFC_3339()

test_that("can parse default RFC 3339 format", {
  x <- "2019-01-01T00:00:01Z"
  expect_identical(
    date_time_parse_RFC_3339(x),
    date_time_build(2019, 1, 1, 0, 0, 1, zone = "UTC")
  )
})

test_that("`offset` is passed through", {
  x <- "2019-01-01T00:00:01-03:30"
  expect_identical(
    date_time_parse_RFC_3339(x, offset = "%Ez"),
    as_date_time(sys_time_parse_RFC_3339(x, offset = "%Ez"), zone = "UTC")
  )
})

test_that("`separator` is passed through", {
  x <- "2019-01-01 00:00:01Z"
  expect_identical(
    date_time_parse_RFC_3339(x, separator = " "),
    as_date_time(sys_time_parse_RFC_3339(x, separator = " "), zone = "UTC")
  )
})

# ------------------------------------------------------------------------------
# date_shift()

test_that("can shift date times", {
  x <- date_time_parse("2019-01-01 01:30:35", "America/New_York")
  x <- x + c(0, 1)

  monday <- weekday(clock_weekdays$monday)
  tuesday <- weekday(clock_weekdays$tuesday)

  expect_identical(
    date_shift(x, monday),
    add_days(x, 6)
  )
  expect_identical(
    date_shift(x, monday, which = "previous"),
    add_days(x, -1)
  )
  expect_identical(
    date_shift(x, tuesday, boundary = "advance"),
    add_days(x, 7)
  )
})

test_that("`ambiguous = x` retains the offset of `x` if applicable", {
  x <- date_time_parse("1970-10-25 01:30:00", "America/New_York", ambiguous = "earliest")
  x <- x + c(0, 3600)

  expect_snapshot_error(date_shift(x, as_weekday(x), ambiguous = "error"))
  expect_identical(date_shift(x, as_weekday(x)), x)
})

# ------------------------------------------------------------------------------
# date_time_build()

test_that("can build a POSIXct", {
  zone <- "America/New_York"

  expect_identical(date_time_build(2019, zone = zone), as.POSIXct("2019-01-01", tz = zone))
  expect_identical(date_time_build(2020, 2, 3, 4, 5, 6, zone = zone), as.POSIXct("2020-02-03 04:05:06", tz = zone))
})

test_that("`zone` is required", {
  expect_snapshot_error(date_time_build(2019))
})

test_that("can handle invalid dates", {
  zone <- "America/New_York"

  expect_snapshot_error(date_time_build(2019, 1:12, 31, zone = zone))

  expect_identical(
    date_time_build(2019, 1:12, 31, zone = zone, invalid = "previous-day"),
    date_time_build(2019, 1:12, "last", zone = zone)
  )
})

test_that("can handle nonexistent times", {
  zone <- "America/New_York"

  expect_snapshot_error(date_time_build(1970, 4, 26, 2, 30, zone = zone))

  expect_identical(
    date_time_build(1970, 4, 26, 2, 30, zone = zone, nonexistent = "roll-forward"),
    date_time_build(1970, 4, 26, 3, zone = zone)
  )
})

test_that("can handle ambiguous times", {
  zone <- "America/New_York"

  expect_snapshot_error(date_time_build(1970, 10, 25, 1, 30, zone = zone))

  expect_identical(
    date_time_build(1970, 10, 25, 1, 30, zone = zone, ambiguous = "earliest"),
    date_time_parse_complete("1970-10-25T01:30:00-04:00[America/New_York]")
  )
  expect_identical(
    date_time_build(1970, 10, 25, 1, 30, zone = zone, ambiguous = "latest"),
    date_time_parse_complete("1970-10-25T01:30:00-05:00[America/New_York]")
  )
})

# ------------------------------------------------------------------------------
# date_now()

test_that("can get the current date-time", {
  x <- date_now("America/New_York")
  expect_length(x, 1)
  expect_s3_class(x, "POSIXct")
})

# ------------------------------------------------------------------------------
# date_start()

test_that("can get the start", {
  zone <- "America/New_York"
  x <- date_time_build(2019, 2, 2, 2, 2, 2, zone = zone)
  expect_identical(date_start(x, "second"), x)
  expect_identical(date_start(x, "month"), date_time_build(2019, 2, 1, zone = zone))
  expect_identical(date_start(x, "year"), date_time_build(2019, 1, 1, zone = zone))
})

test_that("start: can't use invalid precisions", {
  expect_snapshot_error(date_start(date_time_build(2019, zone = "America/New_York"), "quarter"))
})

test_that("can resolve nonexistent start issues", {
  # In Asia/Beirut, DST gap from 2021-03-27 23:59:59 -> 2021-03-28 01:00:00
  zone <- "Asia/Beirut"
  x <- date_time_build(2021, 3, 28, 2, zone = zone)

  expect_snapshot_error(date_start(x, "day"), class = "clock_error_nonexistent_time")

  expect_identical(
    date_start(x, "day", nonexistent = "roll-forward"),
    date_time_build(2021, 3, 28, 1, zone = zone)
  )
})

test_that("can resolve ambiguous start issues", {
  # In Asia/Amman, DST fallback from 2021-10-29 00:59:59 -> 2021-10-29 00:00:00
  zone <- "Asia/Amman"
  x <- date_time_build(2021, 10, 29, 2, zone = zone)

  expect_snapshot_error(date_start(x, "day"), class = "clock_error_ambiguous_time")

  expect_identical(
    date_start(x, "day", ambiguous = "earliest"),
    date_time_parse_complete("2021-10-29T00:00:00+03:00[Asia/Amman]")
  )
  expect_identical(
    date_start(x, "day", ambiguous = "latest"),
    date_time_parse_complete("2021-10-29T00:00:00+02:00[Asia/Amman]")
  )
})

test_that("can automatically resolve ambiguous issues", {
  # In Asia/Amman, DST fallback from 2021-10-29 00:59:59 -> 2021-10-29 00:00:00
  zone <- "Asia/Amman"

  # Starts and ends the manipulation in the "earliest" hour
  x <- date_time_build(2021, 10, 29, 0, 20, zone = zone, ambiguous = "earliest")
  expect_identical(
    date_start(x, "day"),
    date_time_parse_complete("2021-10-29T00:00:00+03:00[Asia/Amman]")
  )

  # Starts and ends the manipulation in the "latest" hour
  x <- date_time_build(2021, 10, 29, 0, 20, zone = zone, ambiguous = "latest")
  expect_identical(
    date_start(x, "day"),
    date_time_parse_complete("2021-10-29T00:00:00+02:00[Asia/Amman]")
  )
})

# ------------------------------------------------------------------------------
# date_end()

test_that("can get the end", {
  zone <- "America/New_York"
  x <- date_time_build(2019:2020, 2, 2, 2, 2, 2, zone = zone)
  expect_identical(date_end(x, "day"), date_time_build(2019:2020, 2, 2, 23, 59, 59, zone = zone))
  expect_identical(date_end(x, "month"), date_time_build(2019:2020, 2, 28:29, 23, 59, 59, zone = zone))
})

test_that("end: can't use invalid precisions", {
  expect_snapshot_error(date_end(date_time_build(2019, zone = "America/New_York"), "quarter"))
})

# ------------------------------------------------------------------------------
# date_seq()

test_that("integer `by` means second precision", {
  expect_identical(
    date_seq(new_datetime(0), to = new_datetime(5), by = 1),
    date_seq(new_datetime(0), to = new_datetime(5), by = duration_seconds(1))
  )
})

test_that("calendar backed `by` works", {
  zone <- "America/New_York"

  expect_identical(
    date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2019, 6, 1, zone = zone), by = duration_months(2)),
    date_time_build(2019, c(1, 3, 5), 1, zone = zone)
  )
  expect_identical(
    date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2025, 1, 1, zone = zone), by = duration_years(2)),
    date_time_build(c(2019, 2021, 2023, 2025), 1, 1, zone = zone)
  )
})

test_that("naive-time backed `by` works", {
  zone <- "America/New_York"

  expect_identical(
    date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2019, 1, 6, zone = zone), by = duration_days(2)),
    date_time_build(2019, 1, c(1, 3, 5), zone = zone)
  )
})

test_that("sys-time backed `by` works", {
  zone <- "America/New_York"

  expect_identical(
    date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2019, 1, 1, 6, zone = zone), by = duration_hours(3)),
    date_time_build(2019, 1, 1, c(0, 3, 6), zone = zone)
  )
  expect_identical(
    date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2019, 1, 1, 0, 6, zone = zone), by = duration_minutes(3)),
    date_time_build(2019, 1, 1, 0, c(0, 3, 6), zone = zone)
  )
})

test_that("sub daily `by` uses sys-time around DST", {
  zone <- "America/New_York"
  from <- date_time_build(1970, 4, 26, 1, 30, zone = zone)

  expect_identical(
    date_seq(from, by = duration_hours(1), total_size = 3),
    date_time_build(1970, 4, 26, c(1, 3, 4), 30, zone = zone)
  )
})

test_that("daily `by` uses naive-time around DST gaps", {
  zone <- "America/New_York"
  from <- date_time_build(1970, 4, 25, 2, 30, zone = zone)

  expect_snapshot_error(date_seq(from, by = duration_days(1), total_size = 3))

  expect_identical(
    date_seq(from, by = duration_days(1), total_size = 3, nonexistent = "roll-forward"),
    date_time_build(1970, 4, c(25, 26, 27), c(2, 3, 2), c(30, 0, 30), zone = zone)
  )
})

test_that("daily `by` uses naive-time around DST fallbacks", {
  zone <- "America/New_York"
  from <- date_time_build(1970, 10, 24, 1, 30, zone = zone)

  expect_snapshot_error(date_seq(from, by = duration_days(1), total_size = 3))

  expect_identical(
    date_seq(from, by = duration_days(1), total_size = 3, ambiguous = "earliest"),
    date_time_build(1970, 10, c(24, 25, 26), 1, 30, zone = zone, ambiguous = "earliest")
  )
})

test_that("monthly / yearly `by` uses calendar -> naive-time around DST gaps", {
  zone <- "America/New_York"
  from <- date_time_build(1970, 3, 26, 2, 30, zone = zone)

  expect_snapshot_error(date_seq(from, by = duration_months(1), total_size = 3))

  expect_identical(
    date_seq(from, by = duration_months(1), total_size = 3, nonexistent = "roll-forward"),
    date_time_build(1970, c(3, 4, 5), 26, c(2, 3, 2), c(30, 0, 30), zone = zone)
  )
})

test_that("monthly / yearly `by` uses calendar -> naive-time around DST fallbacks", {
  zone <- "America/New_York"
  from <- date_time_build(1969, 10, 25, 1, 30, zone = zone)

  expect_snapshot_error(date_seq(from, by = duration_years(1), total_size = 3))

  expect_identical(
    date_seq(from, by = duration_years(1), total_size = 3, ambiguous = "earliest"),
    date_time_build(c(1969, 1970, 1971), 10, 25, 1, 30, zone = zone, ambiguous = "earliest")
  )
})

test_that("combining `by` with `total_size` works", {
  zone <- "America/New_York"

  expect_identical(
    date_seq(date_time_build(2019, 1, 1, zone = zone), by = 2, total_size = 3),
    date_time_build(2019, 1, 1, 0, 0, c(0, 2, 4), zone = zone)
  )
})

test_that("combining `to` with `total_size` works", {
  zone <- "America/New_York"

  expect_identical(
    date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2019, 1, 5, zone = zone), total_size = 3),
    date_time_build(2019, 1, c(1, 3, 5), zone = zone)
  )
})

test_that("can resolve invalid dates", {
  zone <- "America/New_York"

  from <- date_time_build(2019, 1, 31, zone = zone)
  to <- date_time_build(2019, 5, 31, zone = zone)

  expect_snapshot_error(date_seq(from, to = to, by = duration_months(1)))

  expect_identical(
    date_seq(from, to = to, by = duration_months(1), invalid = "previous-day"),
    date_time_build(2019, 1:5, "last", zone = zone)
  )
})

test_that("quarterly `by` works", {
  zone <- "America/New_York"

  expect_identical(
    date_seq(date_time_build(2019, 1, 2, zone = zone), by = duration_quarters(1), total_size = 3),
    date_seq(date_time_build(2019, 1, 2, zone = zone), by = duration_months(3), total_size = 3)
  )
})

test_that("weekly `by` works", {
  zone <- "America/New_York"

  expect_identical(
    date_seq(date_time_build(2019, 1, 2, zone = zone), by = duration_weeks(1), total_size = 3),
    date_seq(date_time_build(2019, 1, 2, zone = zone), by = duration_days(7), total_size = 3)
  )
})

test_that("components of `from` more precise than `by` are restored", {
  zone <- "America/New_York"

  expect_identical(
    date_seq(date_time_build(2019, 2, 3, 2, 2, 2, zone = zone), by = duration_minutes(1), total_size = 2),
    date_time_build(2019, 2, 3, 2, 2:3, 2, zone = zone)
  )
  expect_identical(
    date_seq(date_time_build(2019, 2, 3, 2, 2, 2, zone = zone), by = duration_hours(1), total_size = 2),
    date_time_build(2019, 2, 3, 2:3, 2, 2, zone = zone)
  )
  expect_identical(
    date_seq(date_time_build(2019, 2, 3, 2, 2, 2, zone = zone), by = duration_days(1), total_size = 2),
    date_time_build(2019, 2, 3:4, 2, 2, 2, zone = zone)
  )
  expect_identical(
    date_seq(date_time_build(2019, 2, 3, 2, 2, 2, zone = zone), by = duration_months(1), total_size = 2),
    date_time_build(2019, 2:3, 3, 2, 2, 2, zone = zone)
  )
  expect_identical(
    date_seq(date_time_build(2019, 2, 3, 2, 2, 2, zone = zone), by = duration_years(1), total_size = 2),
    date_time_build(2019:2020, 2, 3, 2, 2, 2, zone = zone)
  )
})

test_that("components of `to` more precise than `by` must match `from`", {
  zone <- "America/New_York"

  expect_snapshot_error(date_seq(date_time_build(2019, 1, 1, 2, 3, 20, zone = zone), to = date_time_build(2019, 2, 2, 1, 3, 5, zone = zone), by = duration_minutes(1)))

  expect_snapshot_error(date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2019, 2, 2, 2, zone = zone), by = duration_days(1)))

  expect_snapshot_error(date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2019, 2, 2, zone = zone), by = duration_months(1)))
  expect_snapshot_error(date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2019, 2, 1, 1, zone = zone), by = duration_months(1)))

  expect_snapshot_error(date_seq(date_time_build(2019, 1, 1, zone = zone), to = date_time_build(2019, 1, 2, zone = zone), by = duration_years(1)))
})

test_that("seq() with `from > to && by > 0` or `from < to && by > 0` results in length 0 output (#282)", {
  zone <- "America/New_York"

  expect_identical(
    date_seq(date_time_build(2019, 1, 2, second = 1, zone = zone), to = date_time_build(2019, 1, 2, zone = zone), by = 1),
    date_time_build(integer(), zone = zone)
  )

  expect_identical(
    date_seq(date_time_build(2019, zone = zone), to = date_time_build(2020, zone = zone), by = -1),
    date_time_build(integer(), zone = zone)
  )
})

test_that("`to` must have same time zone as `by`", {
  expect_snapshot_error(date_seq(date_time_build(1970, zone = "UTC"), to = date_time_build(1970, zone = "America/New_York"), by = 1))
})

test_that("validates integerish `by`", {
  expect_snapshot_error(date_seq(new_datetime(1), by = 1.5, total_size = 1))
})

test_that("validates `total_size` early", {
  expect_snapshot_error(date_seq(new_datetime(1), by = 1, total_size = 1.5))
  expect_snapshot_error(date_seq(new_datetime(1), by = 1, total_size = NA))
  expect_snapshot_error(date_seq(new_datetime(1), by = 1, total_size = -1))
})

test_that("`to` and `total_size` must not generate a non-fractional sequence", {
  expect_snapshot_error(date_seq(new_datetime(0), to = new_datetime(3), total_size = 3))
})

test_that("requires exactly two optional arguments", {
  expect_snapshot_error(date_seq(new_datetime(1), by = 1))
  expect_snapshot_error(date_seq(new_datetime(1), total_size = 1))
  expect_snapshot_error(date_seq(new_datetime(1), to = new_datetime(1)))
})

test_that("requires `to` to be POSIXt", {
  expect_snapshot_error(date_seq(new_datetime(1), to = 1, by = 1))
})

test_that("requires year, month, day, hour, minute, or second precision", {
  expect_snapshot_error(date_seq(new_datetime(1), to = new_datetime(2), by = duration_nanoseconds(1)))
})

test_that("checks empty dots", {
  expect_snapshot_error(date_seq(new_datetime(1), new_datetime(2)))
})

test_that("golden test: ensure that we never allow components of `to` to differ with `from` (#224)", {
  # DST gap from 01:59:59 -> 03:00:00 on the 26th
  from <- date_time_build(1970, 04, 25, 02, 30, 00, zone = "America/New_York")
  to <- date_time_build(1970, 04, 26, 03, 00, 00, zone = "America/New_York")

  expect_error(date_seq(from, to = to, by = duration_days(1), nonexistent = "shift-forward"))

  # Could theoretically generate this, where the second element is past `to`
  #> "1970-04-25 02:30:00 EST" "1970-04-26 03:30:00 EDT"
})

# ------------------------------------------------------------------------------
# date_count_between()

test_that("can compute precisions at year / quarter / month / week / day / hour / minute / second", {
  x <- date_time_build(2019, 1, 5, 5, zone = "UTC")

  y <- date_time_build(2025, 1, c(4, 6), zone = "UTC")
  expect_identical(date_count_between(x, y, "year"), c(5L, 6L))
  expect_identical(date_count_between(x, y, "quarter"), c(23L, 24L))
  expect_identical(date_count_between(x, y, "month"), c(71L, 72L))

  y <- date_time_build(2019, 1, c(25, 27), zone = "UTC")
  expect_identical(date_count_between(x, y, "week"), c(2L, 3L))
  expect_identical(date_count_between(x, y, "day"), c(19L, 21L))

  y <- date_time_build(2019, 1, 6, c(5, 6), c(59, 0), zone = "UTC")
  expect_identical(date_count_between(x, y, "hour"), c(24L, 25L))
  expect_identical(date_count_between(x, y, "minute"), c(1499L, 1500L))
  expect_identical(date_count_between(x, y, "second"), c(89940L, 90000L))
})

test_that("can use posixlt", {
  x <- as.POSIXlt(date_time_build(2019, 1, 5, 5, zone = "UTC"))
  y <- as.POSIXlt(date_time_build(2020, 1, 5, c(4, 5), zone = "UTC"))
  expect_identical(date_count_between(x, y, "year"), c(0L, 1L))
})

test_that("nonexistent times are handled correctly", {
  x <- date_time_build(1970, 4, 26, 1, 59, 59, zone = "America/New_York")
  y <- date_time_build(1970, 4, 26, 3, 00, 00, zone = "America/New_York")

  # sys-time for hour, minute, second
  expect_identical(date_count_between(x, y, "second"), 1L)
  expect_identical(date_count_between(x, y, "hour"), 0L)

  # naive-time for week and day
  z <- date_time_build(1970, 5, 3, 2, 00, 00, zone = "America/New_York")
  expect_identical(date_count_between(x, z, "day"), 7L)
  expect_identical(date_count_between(y, z, "day"), 6L)

  # calendar (naive) for year and month
  z <- date_time_build(1970, 5, 26, 2, 30, 00, zone = "America/New_York")
  expect_identical(date_count_between(x, z, "month"), 1L)
  expect_identical(date_count_between(y, z, "month"), 0L)
})

test_that("ambiguous times are handled correctly", {
  x <- date_time_build(1970, 10, 25, 1, 00, 01, zone = "America/New_York", ambiguous = "earliest")
  y <- date_time_build(1970, 10, 25, 1, 00, 01, zone = "America/New_York", ambiguous = "latest")

  # sys-time for hour, minute, second
  expect_identical(date_count_between(x, y, "second"), 3600L)
  expect_identical(date_count_between(x, y, "hour"), 1L)

  # naive-time for week and day
  z <- date_time_build(1970, 11, 01, 1, 00, c(00, 02), zone = "America/New_York")
  expect_identical(date_count_between(x, z, "week"), c(0L, 1L))
  expect_identical(date_count_between(y, z, "week"), c(0L, 1L))
  expect_identical(date_count_between(x, z, "day"), c(6L, 7L))
  expect_identical(date_count_between(y, z, "day"), c(6L, 7L))

  # calendar (naive) for year and month
  z <- date_time_build(1970, 11, 25, 1, 00, c(00, 02), zone = "America/New_York")
  expect_identical(date_count_between(x, z, "month"), c(0L, 1L))
  expect_identical(date_count_between(y, z, "month"), c(0L, 1L))
})

test_that("must use a valid POSIXt precision", {
  x <- date_time_build(2019, zone = "UTC")
  expect_snapshot((expect_error(date_count_between(x, x, "millisecond"))))
})

test_that("can't count between a POSIXt and a Date", {
  x <- date_time_build(2019, zone = "UTC")
  y <- date_build(2019)
  expect_snapshot((expect_error(date_count_between(x, y, "year"))))
})

# ------------------------------------------------------------------------------
# vec_arith()

test_that("<posixt> op <duration>", {
  zone <- "America/New_York"

  new_posixlt <- function(x, zone) {
    as.POSIXlt(new_datetime(x, zone))
  }

  expect_identical(vec_arith("+", new_datetime(0, zone), duration_years(1)), new_datetime(31536000, zone))
  expect_identical(vec_arith("+", new_posixlt(0, zone), duration_years(1)), new_datetime(31536000, zone))

  expect_identical(vec_arith("-", new_datetime(0, zone), duration_years(1)), new_datetime(-31536000, zone))
  expect_identical(vec_arith("-", new_posixlt(0, zone), duration_years(1)), new_datetime(-31536000, zone))

  expect_identical(vec_arith("+", new_datetime(0, zone), duration_seconds(1)), new_datetime(1, zone))
  expect_identical(vec_arith("+", new_posixlt(0, zone), duration_seconds(1)), new_datetime(1, zone))

  expect_snapshot_error(vec_arith("+", new_datetime(0, zone), duration_milliseconds(1)))
  expect_snapshot_error(vec_arith("+", new_posixlt(0, zone), duration_milliseconds(1)))

  expect_snapshot_error(vec_arith("*", new_datetime(0, zone), duration_years(1)))
  expect_snapshot_error(vec_arith("*", new_posixlt(0, zone), duration_years(1)))
})

test_that("<duration> op <posixt>", {
  zone <- "America/New_York"

  new_posixlt <- function(x, zone) {
    as.POSIXlt(new_datetime(x, zone))
  }

  expect_identical(vec_arith("+", duration_years(1), new_datetime(0, zone)), new_datetime(31536000, zone))
  expect_identical(vec_arith("+", duration_years(1), new_posixlt(0, zone)), new_datetime(31536000, zone))

  expect_identical(vec_arith("+", duration_seconds(1), new_datetime(0, zone)), new_datetime(1, zone))
  expect_identical(vec_arith("+", duration_seconds(1), new_posixlt(0, zone)), new_datetime(1, zone))

  expect_snapshot_error(vec_arith("-", duration_years(1), new_datetime(0, zone)))
  expect_snapshot_error(vec_arith("-", duration_years(1), new_posixlt(0, zone)))

  expect_snapshot_error(vec_arith("+", duration_milliseconds(1), new_datetime(0, zone)))
  expect_snapshot_error(vec_arith("+", duration_milliseconds(1), new_posixlt(0, zone)))

  expect_snapshot_error(vec_arith("*", duration_years(1), new_datetime(0, zone)))
  expect_snapshot_error(vec_arith("*", duration_years(1), new_posixlt(0, zone)))
})

# Generated by cpp11: do not edit by hand

new_duration_from_fields <- function(fields, precision_int, names) {
  .Call(`_clock_new_duration_from_fields`, fields, precision_int, names)
}

duration_restore <- function(x, to) {
  .Call(`_clock_duration_restore`, x, to)
}

format_duration_cpp <- function(fields, precision_int) {
  .Call(`_clock_format_duration_cpp`, fields, precision_int)
}

duration_helper_cpp <- function(n, precision_int) {
  .Call(`_clock_duration_helper_cpp`, n, precision_int)
}

duration_cast_cpp <- function(fields, precision_from, precision_to) {
  .Call(`_clock_duration_cast_cpp`, fields, precision_from, precision_to)
}

duration_plus_cpp <- function(x, y, precision_int) {
  .Call(`_clock_duration_plus_cpp`, x, y, precision_int)
}

duration_minus_cpp <- function(x, y, precision_int) {
  .Call(`_clock_duration_minus_cpp`, x, y, precision_int)
}

duration_modulus_cpp <- function(x, y, precision_int) {
  .Call(`_clock_duration_modulus_cpp`, x, y, precision_int)
}

duration_integer_divide_cpp <- function(x, y, precision_int) {
  .Call(`_clock_duration_integer_divide_cpp`, x, y, precision_int)
}

duration_scalar_multiply_cpp <- function(x, y, precision_int) {
  .Call(`_clock_duration_scalar_multiply_cpp`, x, y, precision_int)
}

duration_scalar_modulus_cpp <- function(x, y, precision_int) {
  .Call(`_clock_duration_scalar_modulus_cpp`, x, y, precision_int)
}

duration_scalar_divide_cpp <- function(x, y, precision_int) {
  .Call(`_clock_duration_scalar_divide_cpp`, x, y, precision_int)
}

duration_precision_common_cpp <- function(x_precision, y_precision) {
  .Call(`_clock_duration_precision_common_cpp`, x_precision, y_precision)
}

duration_has_common_precision_cpp <- function(x_precision, y_precision) {
  .Call(`_clock_duration_has_common_precision_cpp`, x_precision, y_precision)
}

duration_floor_cpp <- function(fields, precision_from, precision_to, n) {
  .Call(`_clock_duration_floor_cpp`, fields, precision_from, precision_to, n)
}

duration_ceiling_cpp <- function(fields, precision_from, precision_to, n) {
  .Call(`_clock_duration_ceiling_cpp`, fields, precision_from, precision_to, n)
}

duration_round_cpp <- function(fields, precision_from, precision_to, n) {
  .Call(`_clock_duration_round_cpp`, fields, precision_from, precision_to, n)
}

duration_unary_minus_cpp <- function(fields, precision_int) {
  .Call(`_clock_duration_unary_minus_cpp`, fields, precision_int)
}

duration_as_integer_cpp <- function(fields, precision_int) {
  .Call(`_clock_duration_as_integer_cpp`, fields, precision_int)
}

duration_as_double_cpp <- function(fields, precision_int) {
  .Call(`_clock_duration_as_double_cpp`, fields, precision_int)
}

duration_abs_cpp <- function(fields, precision_int) {
  .Call(`_clock_duration_abs_cpp`, fields, precision_int)
}

duration_sign_cpp <- function(fields, precision_int) {
  .Call(`_clock_duration_sign_cpp`, fields, precision_int)
}

duration_seq_by_lo_cpp <- function(from, precision_int, by, length_out) {
  .Call(`_clock_duration_seq_by_lo_cpp`, from, precision_int, by, length_out)
}

duration_seq_to_by_cpp <- function(from, precision_int, to, by) {
  .Call(`_clock_duration_seq_to_by_cpp`, from, precision_int, to, by)
}

duration_seq_to_lo_cpp <- function(from, precision_int, to, length_out) {
  .Call(`_clock_duration_seq_to_lo_cpp`, from, precision_int, to, length_out)
}

precision_to_string <- function(precision_int) {
  .Call(`_clock_precision_to_string`, precision_int)
}

clock_to_string <- function(clock_int) {
  .Call(`_clock_clock_to_string`, clock_int)
}

format_time_point_cpp <- function(fields, clock, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, decimal_mark) {
  .Call(`_clock_format_time_point_cpp`, fields, clock, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, decimal_mark)
}

format_zoned_time_cpp <- function(fields, zone, abbreviate_zone, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, decimal_mark) {
  .Call(`_clock_format_zoned_time_cpp`, fields, zone, abbreviate_zone, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, decimal_mark)
}

new_year_day_from_fields <- function(fields, precision_int, names) {
  .Call(`_clock_new_year_day_from_fields`, fields, precision_int, names)
}

year_day_restore <- function(x, to) {
  .Call(`_clock_year_day_restore`, x, to)
}

collect_year_day_fields <- function(fields, precision_int) {
  .Call(`_clock_collect_year_day_fields`, fields, precision_int)
}

format_year_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_format_year_day_cpp`, fields, precision_int)
}

invalid_detect_year_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_detect_year_day_cpp`, fields, precision_int)
}

invalid_any_year_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_any_year_day_cpp`, fields, precision_int)
}

invalid_count_year_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_count_year_day_cpp`, fields, precision_int)
}

invalid_resolve_year_day_cpp <- function(fields, precision_int, invalid_string) {
  .Call(`_clock_invalid_resolve_year_day_cpp`, fields, precision_int, invalid_string)
}

set_field_year_day_cpp <- function(fields, value, precision_fields, precision_value) {
  .Call(`_clock_set_field_year_day_cpp`, fields, value, precision_fields, precision_value)
}

set_field_year_day_last_cpp <- function(fields, precision_fields) {
  .Call(`_clock_set_field_year_day_last_cpp`, fields, precision_fields)
}

year_day_plus_duration_cpp <- function(fields, fields_n, precision_fields, precision_n) {
  .Call(`_clock_year_day_plus_duration_cpp`, fields, fields_n, precision_fields, precision_n)
}

as_sys_time_year_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_as_sys_time_year_day_cpp`, fields, precision_int)
}

as_year_day_from_sys_time_cpp <- function(fields, precision_int) {
  .Call(`_clock_as_year_day_from_sys_time_cpp`, fields, precision_int)
}

year_day_minus_year_day_cpp <- function(x, y, precision_int) {
  .Call(`_clock_year_day_minus_year_day_cpp`, x, y, precision_int)
}

new_year_month_day_from_fields <- function(fields, precision_int, names) {
  .Call(`_clock_new_year_month_day_from_fields`, fields, precision_int, names)
}

year_month_day_restore <- function(x, to) {
  .Call(`_clock_year_month_day_restore`, x, to)
}

collect_year_month_day_fields <- function(fields, precision_int) {
  .Call(`_clock_collect_year_month_day_fields`, fields, precision_int)
}

format_year_month_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_format_year_month_day_cpp`, fields, precision_int)
}

invalid_detect_year_month_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_detect_year_month_day_cpp`, fields, precision_int)
}

invalid_any_year_month_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_any_year_month_day_cpp`, fields, precision_int)
}

invalid_count_year_month_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_count_year_month_day_cpp`, fields, precision_int)
}

invalid_resolve_year_month_day_cpp <- function(fields, precision_int, invalid_string) {
  .Call(`_clock_invalid_resolve_year_month_day_cpp`, fields, precision_int, invalid_string)
}

set_field_year_month_day_cpp <- function(fields, value, precision_fields, precision_value) {
  .Call(`_clock_set_field_year_month_day_cpp`, fields, value, precision_fields, precision_value)
}

set_field_year_month_day_last_cpp <- function(fields, precision_fields) {
  .Call(`_clock_set_field_year_month_day_last_cpp`, fields, precision_fields)
}

year_month_day_plus_duration_cpp <- function(fields, fields_n, precision_fields, precision_n) {
  .Call(`_clock_year_month_day_plus_duration_cpp`, fields, fields_n, precision_fields, precision_n)
}

as_sys_time_year_month_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_as_sys_time_year_month_day_cpp`, fields, precision_int)
}

as_year_month_day_from_sys_time_cpp <- function(fields, precision_int) {
  .Call(`_clock_as_year_month_day_from_sys_time_cpp`, fields, precision_int)
}

year_month_day_minus_year_month_day_cpp <- function(x, y, precision_int) {
  .Call(`_clock_year_month_day_minus_year_month_day_cpp`, x, y, precision_int)
}

year_month_day_parse_cpp <- function(x, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, mark) {
  .Call(`_clock_year_month_day_parse_cpp`, x, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, mark)
}

gregorian_leap_year_cpp <- function(year) {
  .Call(`_clock_gregorian_leap_year_cpp`, year)
}

new_year_month_weekday_from_fields <- function(fields, precision_int, names) {
  .Call(`_clock_new_year_month_weekday_from_fields`, fields, precision_int, names)
}

year_month_weekday_restore <- function(x, to) {
  .Call(`_clock_year_month_weekday_restore`, x, to)
}

collect_year_month_weekday_fields <- function(fields, precision_int) {
  .Call(`_clock_collect_year_month_weekday_fields`, fields, precision_int)
}

format_year_month_weekday_cpp <- function(fields, precision_int) {
  .Call(`_clock_format_year_month_weekday_cpp`, fields, precision_int)
}

invalid_detect_year_month_weekday_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_detect_year_month_weekday_cpp`, fields, precision_int)
}

invalid_any_year_month_weekday_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_any_year_month_weekday_cpp`, fields, precision_int)
}

invalid_count_year_month_weekday_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_count_year_month_weekday_cpp`, fields, precision_int)
}

invalid_resolve_year_month_weekday_cpp <- function(fields, precision_int, invalid_string) {
  .Call(`_clock_invalid_resolve_year_month_weekday_cpp`, fields, precision_int, invalid_string)
}

set_field_year_month_weekday_cpp <- function(fields, value, precision_fields, component_string) {
  .Call(`_clock_set_field_year_month_weekday_cpp`, fields, value, precision_fields, component_string)
}

set_field_year_month_weekday_last_cpp <- function(fields, precision_fields) {
  .Call(`_clock_set_field_year_month_weekday_last_cpp`, fields, precision_fields)
}

year_month_weekday_plus_duration_cpp <- function(fields, fields_n, precision_fields, precision_n) {
  .Call(`_clock_year_month_weekday_plus_duration_cpp`, fields, fields_n, precision_fields, precision_n)
}

as_sys_time_year_month_weekday_cpp <- function(fields, precision_int) {
  .Call(`_clock_as_sys_time_year_month_weekday_cpp`, fields, precision_int)
}

as_year_month_weekday_from_sys_time_cpp <- function(fields, precision_int) {
  .Call(`_clock_as_year_month_weekday_from_sys_time_cpp`, fields, precision_int)
}

year_month_weekday_minus_year_month_weekday_cpp <- function(x, y, precision_int) {
  .Call(`_clock_year_month_weekday_minus_year_month_weekday_cpp`, x, y, precision_int)
}

new_iso_year_week_day_from_fields <- function(fields, precision_int, names) {
  .Call(`_clock_new_iso_year_week_day_from_fields`, fields, precision_int, names)
}

iso_year_week_day_restore <- function(x, to) {
  .Call(`_clock_iso_year_week_day_restore`, x, to)
}

collect_iso_year_week_day_fields <- function(fields, precision_int) {
  .Call(`_clock_collect_iso_year_week_day_fields`, fields, precision_int)
}

format_iso_year_week_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_format_iso_year_week_day_cpp`, fields, precision_int)
}

invalid_detect_iso_year_week_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_detect_iso_year_week_day_cpp`, fields, precision_int)
}

invalid_any_iso_year_week_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_any_iso_year_week_day_cpp`, fields, precision_int)
}

invalid_count_iso_year_week_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_invalid_count_iso_year_week_day_cpp`, fields, precision_int)
}

invalid_resolve_iso_year_week_day_cpp <- function(fields, precision_int, invalid_string) {
  .Call(`_clock_invalid_resolve_iso_year_week_day_cpp`, fields, precision_int, invalid_string)
}

set_field_iso_year_week_day_cpp <- function(fields, value, precision_fields, precision_value) {
  .Call(`_clock_set_field_iso_year_week_day_cpp`, fields, value, precision_fields, precision_value)
}

set_field_iso_year_week_day_last_cpp <- function(fields, precision_fields) {
  .Call(`_clock_set_field_iso_year_week_day_last_cpp`, fields, precision_fields)
}

iso_year_week_day_plus_duration_cpp <- function(fields, fields_n, precision_fields, precision_n) {
  .Call(`_clock_iso_year_week_day_plus_duration_cpp`, fields, fields_n, precision_fields, precision_n)
}

as_sys_time_iso_year_week_day_cpp <- function(fields, precision_int) {
  .Call(`_clock_as_sys_time_iso_year_week_day_cpp`, fields, precision_int)
}

as_iso_year_week_day_from_sys_time_cpp <- function(fields, precision_int) {
  .Call(`_clock_as_iso_year_week_day_from_sys_time_cpp`, fields, precision_int)
}

iso_year_week_day_minus_iso_year_week_day_cpp <- function(x, y, precision_int) {
  .Call(`_clock_iso_year_week_day_minus_iso_year_week_day_cpp`, x, y, precision_int)
}

naive_time_info_cpp <- function(fields, precision_int, zone) {
  .Call(`_clock_naive_time_info_cpp`, fields, precision_int, zone)
}

new_year_quarter_day_from_fields <- function(fields, precision_int, start, names) {
  .Call(`_clock_new_year_quarter_day_from_fields`, fields, precision_int, start, names)
}

year_quarter_day_restore <- function(x, to) {
  .Call(`_clock_year_quarter_day_restore`, x, to)
}

collect_year_quarter_day_fields <- function(fields, precision_int, start_int) {
  .Call(`_clock_collect_year_quarter_day_fields`, fields, precision_int, start_int)
}

format_year_quarter_day_cpp <- function(fields, precision_int, start_int) {
  .Call(`_clock_format_year_quarter_day_cpp`, fields, precision_int, start_int)
}

invalid_detect_year_quarter_day_cpp <- function(fields, precision_int, start_int) {
  .Call(`_clock_invalid_detect_year_quarter_day_cpp`, fields, precision_int, start_int)
}

invalid_any_year_quarter_day_cpp <- function(fields, precision_int, start_int) {
  .Call(`_clock_invalid_any_year_quarter_day_cpp`, fields, precision_int, start_int)
}

invalid_count_year_quarter_day_cpp <- function(fields, precision_int, start_int) {
  .Call(`_clock_invalid_count_year_quarter_day_cpp`, fields, precision_int, start_int)
}

invalid_resolve_year_quarter_day_cpp <- function(fields, precision_int, start_int, invalid_string) {
  .Call(`_clock_invalid_resolve_year_quarter_day_cpp`, fields, precision_int, start_int, invalid_string)
}

set_field_year_quarter_day_cpp <- function(fields, value, precision_fields, precision_value, start_int) {
  .Call(`_clock_set_field_year_quarter_day_cpp`, fields, value, precision_fields, precision_value, start_int)
}

set_field_year_quarter_day_last_cpp <- function(fields, precision_fields, start_int) {
  .Call(`_clock_set_field_year_quarter_day_last_cpp`, fields, precision_fields, start_int)
}

year_quarter_day_plus_duration_cpp <- function(fields, fields_n, precision_fields, precision_n, start_int) {
  .Call(`_clock_year_quarter_day_plus_duration_cpp`, fields, fields_n, precision_fields, precision_n, start_int)
}

as_sys_time_year_quarter_day_cpp <- function(fields, precision_int, start_int) {
  .Call(`_clock_as_sys_time_year_quarter_day_cpp`, fields, precision_int, start_int)
}

as_year_quarter_day_from_sys_time_cpp <- function(fields, precision_int, start_int) {
  .Call(`_clock_as_year_quarter_day_from_sys_time_cpp`, fields, precision_int, start_int)
}

year_quarter_day_minus_year_quarter_day_cpp <- function(x, y, precision_int, start_int) {
  .Call(`_clock_year_quarter_day_minus_year_quarter_day_cpp`, x, y, precision_int, start_int)
}

clock_rcrd_proxy <- function(x) {
  .Call(`_clock_clock_rcrd_proxy`, x)
}

clock_rcrd_names <- function(x) {
  .Call(`_clock_clock_rcrd_names`, x)
}

clock_rcrd_set_names <- function(x, names) {
  .Call(`_clock_clock_rcrd_set_names`, x, names)
}

sys_time_now_cpp <- function() {
  .Call(`_clock_sys_time_now_cpp`)
}

sys_time_info_cpp <- function(fields, precision_int, zone) {
  .Call(`_clock_sys_time_info_cpp`, fields, precision_int, zone)
}

new_time_point_from_fields <- function(fields, precision_int, clock_int, names) {
  .Call(`_clock_new_time_point_from_fields`, fields, precision_int, clock_int, names)
}

time_point_restore <- function(x, to) {
  .Call(`_clock_time_point_restore`, x, to)
}

time_point_parse_cpp <- function(x, format, precision_int, clock_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, mark) {
  .Call(`_clock_time_point_parse_cpp`, x, format, precision_int, clock_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, mark)
}

clock_init_utils <- function() {
  .Call(`_clock_clock_init_utils`)
}

weekday_add_days_cpp <- function(x, n) {
  .Call(`_clock_weekday_add_days_cpp`, x, n)
}

weekday_minus_weekday_cpp <- function(x, y) {
  .Call(`_clock_weekday_minus_weekday_cpp`, x, y)
}

weekday_from_time_point_cpp <- function(x) {
  .Call(`_clock_weekday_from_time_point_cpp`, x)
}

format_weekday_cpp <- function(x, labels) {
  .Call(`_clock_format_weekday_cpp`, x, labels)
}

zone_is_valid <- function(zone) {
  .Call(`_clock_zone_is_valid`, zone)
}

zone_current <- function() {
  .Call(`_clock_zone_current`)
}

new_zoned_time_from_fields <- function(fields, precision_int, zone, names) {
  .Call(`_clock_new_zoned_time_from_fields`, fields, precision_int, zone, names)
}

zoned_time_restore <- function(x, to) {
  .Call(`_clock_zoned_time_restore`, x, to)
}

get_naive_time_cpp <- function(fields, precision_int, zone) {
  .Call(`_clock_get_naive_time_cpp`, fields, precision_int, zone)
}

as_zoned_sys_time_from_naive_time_cpp <- function(fields, precision_int, zone, nonexistent_string, ambiguous_string) {
  .Call(`_clock_as_zoned_sys_time_from_naive_time_cpp`, fields, precision_int, zone, nonexistent_string, ambiguous_string)
}

as_zoned_sys_time_from_naive_time_with_reference_cpp <- function(fields, precision_int, zone, nonexistent_string, ambiguous_string, reference) {
  .Call(`_clock_as_zoned_sys_time_from_naive_time_with_reference_cpp`, fields, precision_int, zone, nonexistent_string, ambiguous_string, reference)
}

to_sys_duration_fields_from_sys_seconds_cpp <- function(seconds) {
  .Call(`_clock_to_sys_duration_fields_from_sys_seconds_cpp`, seconds)
}

to_sys_seconds_from_sys_duration_fields_cpp <- function(fields) {
  .Call(`_clock_to_sys_seconds_from_sys_duration_fields_cpp`, fields)
}

zoned_time_parse_complete_cpp <- function(x, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, mark) {
  .Call(`_clock_zoned_time_parse_complete_cpp`, x, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, mark)
}

zoned_time_parse_abbrev_cpp <- function(x, zone, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, mark) {
  .Call(`_clock_zoned_time_parse_abbrev_cpp`, x, zone, format, precision_int, month, month_abbrev, weekday, weekday_abbrev, am_pm, mark)
}

#' Get or set the time zone
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' - `date_zone()` is deprecated in favor of [date_time_zone()].
#'
#' - `date_set_zone()` is deprecated in favor of [date_time_set_zone()].
#'
#' @inheritParams date_time_zone
#'
#' @keywords internal
#' @name date-zone
NULL

#' @rdname date-zone
#' @export
date_zone <- function(x) {
  # - soft-deprecated: 0.7.0
  lifecycle::deprecate_soft(
    when = "0.7.0",
    what = "date_zone()",
    with = "date_time_zone()"
  )
  UseMethod("date_zone")
}

#' @export
date_zone.Date <- function(x) {
  abort("Can't get the zone of a 'Date'.")
}

#' @export
date_zone.POSIXt <- function(x) {
  posixt_tzone(x)
}

#' @rdname date-zone
#' @export
date_set_zone <- function(x, zone) {
  # - soft-deprecated: 0.7.0
  lifecycle::deprecate_soft(
    when = "0.7.0",
    what = "date_set_zone()",
    with = "date_time_set_zone()"
  )
  UseMethod("date_set_zone")
}

#' @export
date_set_zone.Date <- function(x, zone) {
  abort("Can't set the zone of a 'Date'.")
}

#' @export
date_set_zone.POSIXt <- function(x, zone) {
  x <- to_posixct(x)
  check_zone(zone)
  posixt_set_tzone(x, zone)
}

# ------------------------------------------------------------------------------

strict_validate_invalid <- function(invalid) {
  if (!is_null(invalid)) {
    return(invalid)
  }

  if (in_strict_mode()) {
    abort(
      paste0(
        "The global option, `clock.strict`, is currently set to `TRUE`. ",
        "In this mode, `invalid` must be set and cannot be left as `NULL`."
      )
    )
  }

  "error"
}

# ------------------------------------------------------------------------------

check_nonexistent_strict <- function(nonexistent, call = caller_env()) {
  if (!is_null(nonexistent)) {
    return(nonexistent)
  }

  if (in_strict_mode()) {
    message <- paste0(
      "The global option, `clock.strict`, is currently set to `TRUE`. ",
      "In this mode, `nonexistent` must be set and cannot be left as `NULL`."
    )
    abort(message, call = call)
  }

  "error"
}

# ------------------------------------------------------------------------------

check_ambiguous_strict <- function(ambiguous, call = caller_env()) {
  if (!is_null(ambiguous)) {
    return(ambiguous)
  }

  if (in_strict_mode()) {
    message <- paste0(
      "The global option, `clock.strict`, is currently set to `TRUE`. ",
      "In this mode, `ambiguous` must be set and cannot be left as `NULL`. ",
      "Additionally, `ambiguous` cannot be set to a zoned-time or POSIXct ",
      "unless it is paired with an ambiguous time resolution strategy, like: ",
      "`list(<zoned-time>, 'earliest')`."
    )
    abort(message, call = call)
  }

  "error"
}

# ------------------------------------------------------------------------------

in_strict_mode <- function() {
  strict <- getOption("clock.strict", default = FALSE)
  is_true(strict)
}

---
title: "Getting Started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(clock)
library(magrittr)
```

The goal of this vignette is to introduce you to clock's high-level API, which works directly on R's built-in date-time types, Date and POSIXct. For an overview of all of the functionality in the high-level API, check out the pkgdown reference section, [High Level API](https://clock.r-lib.org/reference/index.html#section-high-level-api). One thing you should immediately notice is that every function specific to R's date and date-time types are prefixed with `date_*()`. There are also additional functions for arithmetic (`add_*()`) and getting (`get_*()`) or setting (`set_*()`) components that are also used by other types in clock.

As you'll quickly see in this vignette, one of the main goals of clock is to guard you, the user, from unexpected issues caused by frustrating date manipulation concepts like invalid dates and daylight saving time. It does this by letting you know as soon as one of these issues happens, giving you the power to handle it explicitly with one of a number of different resolution strategies.

## Building

To create a vector of dates, you can use `date_build()`. This allows you to specify the components individually.

```{r}
date_build(2019, 2, 1:5)
```

If you happen to specify an _invalid date_, you'll get an error message:

```{r, error=TRUE}
date_build(2019, 1:12, 31)
```

One way to resolve this is by specifying an invalid date resolution strategy using the `invalid` argument. There are multiple options, but in this case we'll ask for the invalid dates to be set to the previous valid moment in time.

```{r}
date_build(2019, 1:12, 31, invalid = "previous")
```

To learn more about invalid dates, check out the documentation for `invalid_resolve()`.

If we were actually after the "last day of the month", an easier way to specify this would have been:

```{r}
date_build(2019, 1:12, "last")
```

You can also create date-times using `date_time_build()`, which generates a POSIXct. Note that you must supply a time zone!

```{r}
date_time_build(2019, 1:5, 1, 2, 30, zone = "America/New_York")
```

If you "build" a time that doesn't exist, you'll get an error. For example, on March 8th, 2020, there was a daylight saving time gap of 1 hour in the America/New_York time zone that took us from `01:59:59` directly to `03:00:00`, skipping the 2 o'clock hour entirely. Let's "accidentally" create a time in that gap:

```{r, error=TRUE}
date_time_build(2019:2021, 3, 8, 2, 30, zone = "America/New_York")
```

To resolve this issue, we can specify a nonexistent time resolution strategy through the `nonexistent` argument. There are a number of options, including rolling forward or backward to the next or previous valid moments in time:

```{r}
zone <- "America/New_York"

date_time_build(2019:2021, 3, 8, 2, 30, zone = zone, nonexistent = "roll-forward")
date_time_build(2019:2021, 3, 8, 2, 30, zone = zone, nonexistent = "roll-backward")
```

## Parsing

### Parsing dates

To parse dates, use `date_parse()`. Parsing dates requires a _format string_, a combination of _commands_ that specify where date components are in your string. By default, it assumes that you're working with dates in the form `"%Y-%m-%d"` (year-month-day).

```{r}
date_parse("2019-01-05")
```

You can change the format string using `format`:

```{r}
date_parse("January 5, 2020", format = "%B %d, %Y")
```

Various different locales are supported for parsing month and weekday names in different languages. To parse a French month:

```{r}
date_parse(
  "juillet 10, 2021", 
  format = "%B %d, %Y", 
  locale = clock_locale("fr")
)
```

You can learn about more locale options in the documentation for `clock_locale()`.

If you have heterogeneous dates, you can supply multiple format strings:

```{r}
x <- c("2020/1/5", "10-03-05", "2020/2/2")
formats <- c("%Y/%m/%d", "%y-%m-%d")

date_parse(x, format = formats)
```

### Parsing date-times

You have four options when parsing date-times:

- `date_time_parse()`: For strings like `"2020-01-01 01:02:03"` where there is neither a time zone offset nor a full (not abbreviated!) time zone name.

- `date_time_parse_complete()`: For strings like `"2020-01-01T01:02:03-05:00[America/New_York]"` where there is both a time zone offset and time zone name present in the string.

- `date_time_parse_abbrev()`: For strings like `"2020-01-01 01:02:03 EST"` where there is a time zone abbreviation in the string.

- `date_time_parse_RFC_3339()`: For strings like `"2020-01-01T01:02:03Z"` or `"2020-01-01T01:02:03-05:00"`, which are in RFC 3339 format and are intended to be interpreted as UTC.

#### date_time_parse()

`date_time_parse()` requires a `zone` argument, and will ignore any other zone information in the string (i.e. if you tried to specify `%z` and `%Z`). The default format string is `"%Y-%m-%d %H:%M:%S"`.

```{r}
date_time_parse("2020-01-01 01:02:03", "America/New_York")
```

If you happen to parse an invalid or ambiguous date-time, you'll get an error. For example, on November 1st, 2020, there were _two_ 1 o'clock hours in the America/New_York time zone due to a daylight saving time fallback. You can see that if we parse a time right before the fallback, and then shift it forward by 1 second, and then 1 hour and 1 second, respectively:

```{r}
before <- date_time_parse("2020-11-01 00:59:59", "America/New_York")

# First 1 o'clock
before + 1

# Second 1 o'clock
before + 1 + 3600
```

The following string doesn't include any information about which of these two 1 o'clocks it belongs to, so it is considered _ambiguous_. Ambiguous times will error when parsing:

```{r, error=TRUE}
date_time_parse("2020-11-01 01:30:00", "America/New_York")
```

To fix that, you can specify an ambiguous time resolution strategy with the `ambiguous` argument.

```{r}
zone <- "America/New_York"

date_time_parse("2020-11-01 01:30:00", zone, ambiguous = "earliest")
date_time_parse("2020-11-01 01:30:00", zone, ambiguous = "latest")
```

#### date_time_parse_complete()

`date_time_parse_complete()` doesn't have a `zone` argument, and doesn't require `ambiguous` or `nonexistent` arguments, since it assumes that the string you are providing is completely unambiguous. The only way this is possible is by having both a time zone offset, specified by `%z`, and a full time zone name, specified by `%Z`, in the string.

The following is an example of an "extended" RFC 3339 format used by Java 8's time library to specify complete date-time strings. This is something that `date_time_parse_complete()` can parse. The default format string follows this extended format, and is `"%Y-%m-%dT%H:%M:%S%z[%Z]"`.

```{r}
x <- "2020-01-01T01:02:03-05:00[America/New_York]"

date_time_parse_complete(x)
```

#### date_time_parse_abbrev()

`date_time_parse_abbrev()` is useful when your date-time strings contain a time zone abbreviation rather than a time zone offset or full time zone name.

```{r}
x <- "2020-01-01 01:02:03 EST"

date_time_parse_abbrev(x, "America/New_York")
```

The string is first parsed as a naive time without considering the abbreviation, and is then converted to a zoned-time using the supplied `zone`. If an ambiguous time is parsed, the abbreviation is used to resolve the ambiguity.

```{r}
x <- c(
  "1970-10-25 01:30:00 EDT",
  "1970-10-25 01:30:00 EST"
)

date_time_parse_abbrev(x, "America/New_York")
```

You might be wondering why you need to supply `zone` at all. Isn't the abbreviation enough? Unfortunately, multiple countries use the same time zone abbreviations, even though they have different time zones. This means that, in many cases, the abbreviation alone is ambiguous. For example, both India and Israel use `IST` for their standard times.

```{r}
x <- "1970-01-01 02:30:30 IST"

# IST = India Standard Time
date_time_parse_abbrev(x, "Asia/Kolkata")

# IST = Israel Standard Time
date_time_parse_abbrev(x, "Asia/Jerusalem")
```

#### date_time_parse_RFC_3339()

`date_time_parse_RFC_3339()` is useful when your date-time strings come from an API, which means they are likely in an ISO 8601 or RFC 3339 format, and should be interpreted as UTC.

The default format string parses the typical RFC 3339 format of `"%Y-%m-%dT%H:%M:%SZ"`.

```{r}
x <- "2020-01-01T01:02:03Z"

date_time_parse_RFC_3339(x)
```

If your date-time strings contain a numeric offset from UTC rather than a `"Z"`, then you'll need to set the `offset` argument to one of the following:

- `"%z"` if the offset is of the form `"-0500"`.
- `"%Ez"` if the offset is of the form `"-05:00"`.

```{r}
x <- "2020-01-01T01:02:03-0500"

date_time_parse_RFC_3339(x, offset = "%z")

x <- "2020-01-01T01:02:03-05:00"

date_time_parse_RFC_3339(x, offset = "%Ez")
```

## Grouping, rounding and shifting

When performing time-series related data analysis, you often need to summarize your series at a less precise precision. There are many different ways to do this, and the differences between them are subtle, but meaningful. clock offers three different sets of functions for summarization:

- `date_group()`

- `date_floor()`, `date_ceiling()`, and `date_round()`

- `date_shift()`

### Grouping

Grouping allows you to summarize a component of a date or date-time _within_ other components. An example of this is grouping by day of the month, which summarizes the day component _within_ the current year-month.

```{r}
x <- seq(date_build(2019, 1, 20), date_build(2019, 2, 5), by = 1)
x

# Grouping by 5 days of the current month
date_group(x, "day", n = 5)
```

The thing to note about grouping by day of the month is that at the end of each month, the groups restart. So this created groups for January of `[1, 5], [6, 10], [11, 15], [16, 20], [21, 25], [26, 30], [31]`.

You can also group by month or year:

```{r}
date_group(x, "month")
```

This also works with date-times, adding the ability to group by hour of the day, minute of the hour, and second of the minute.

```{r}
x <- seq(
  date_time_build(2019, 1, 1, 1, 55, zone = "UTC"),
  date_time_build(2019, 1, 1, 2, 15, zone = "UTC"),
  by = 120
)
x

date_group(x, "minute", n = 5)
```

### Rounding

While grouping is useful for summarizing _within_ a component, rounding is useful for summarizing _across_ components. It is great for summarizing by, say, a rolling set of 60 days.

Rounding operates on the underlying count that makes up your date or date-time. To see what I mean by this, try unclassing a date:

```{r}
unclass(date_build(2020, 1, 1))
```

This is a count of days since the _origin_ that R uses, 1970-01-01, which is considered day 0. If you were to floor by 60 days, this would bundle `[1970-01-01, 1970-03-02), [1970-03-02, 1970-05-01)`, and so on. Equivalently, it bundles counts of `[0, 60), [60, 120)`, etc.

```{r}
x <- seq(date_build(1970, 01, 01), date_build(1970, 05, 10), by = 20)

date_floor(x, "day", n = 60)
date_ceiling(x, "day", n = 60)
```

If you prefer a different origin, you can supply a Date `origin` to `date_floor()`, which determines what "day 0" is considered to be. This can be useful for grouping by multiple weeks if you want to control what is considered the start of the week. Since 1970-01-01 is a Thursday, flooring by 2 weeks would normally generate all Thursdays:

```{r}
as_weekday(date_floor(x, "week", n = 14))
```

To change this you can supply an `origin` on the weekday that you'd like to be considered the first day of the week.

```{r}
sunday <- date_build(1970, 01, 04)

date_floor(x, "week", n = 14, origin = sunday)

as_weekday(date_floor(x, "week", n = 14, origin = sunday))
```

If you only need to floor by 1 week, it is often easier to use `date_shift()`, as seen in the next section.

### Shifting

`date_shift()` allows you to target a weekday, and then shift a vector of dates forward or backward to the next instance of that target. It requires using one of the new types in clock, _weekday_, which is supplied as the target.

For example, to shift to the next Tuesday:

```{r}
x <- date_build(2020, 1, 1:2)

# Wednesday / Thursday
as_weekday(x)

# `clock_weekdays` is a helper that returns the code corresponding to
# the requested day of the week
clock_weekdays$tuesday

tuesday <- weekday(clock_weekdays$tuesday)
tuesday

date_shift(x, target = tuesday)
```

Shifting to the _previous_ day of the week is a nice way to floor by 1 week. It allows you to control the start of the week in a way that is slightly easier than using `date_floor(origin = )`.

```{r}
x <- seq(date_build(1970, 01, 01), date_build(1970, 01, "last"), by = 3)

date_shift(x, tuesday, which = "previous")
```

## Arithmetic

You can do arithmetic with dates and date-times using the family of `add_*()` functions. With dates, you can add years, months, and days. With date-times, you can additionally add hours, minutes, and seconds.

```{r}
x <- date_build(2020, 1, 1)

add_years(x, 1:5)
```

One of the neat parts about clock is that it requires you to be explicit about how you want to handle invalid dates when doing arithmetic. What is 1 month after January 31st? If you try and create this date, you'll get an error.

```{r, error=TRUE}
x <- date_build(2020, 1, 31)

add_months(x, 1)
```

clock gives you the power to handle this through the `invalid` option:

```{r}
# The previous valid moment in time
add_months(x, 1, invalid = "previous")

# The next valid moment in time
add_months(x, 1, invalid = "next")

# Overflow the days. There were 29 days in February, 2020, but we
# specified 31. So this overflows 2 days past day 29.
add_months(x, 1, invalid = "overflow")

# If you don't consider it to be a valid date
add_months(x, 1, invalid = "NA")
```

As a teaser, the low level library has a _calendar_ type named year-month-day that powers this operation. It actually gives you _more_ flexibility, allowing `"2020-02-31"` to exist in the wild:

```{r}
ymd <- as_year_month_day(x) + duration_months(1)
ymd
```

You can use `invalid_resolve(invalid =)` to resolve this like you did in `add_months()`, or you can let it hang around if you expect other operations to make it "valid" again. 

```{r}
# Adding 1 more month makes it valid again
ymd + duration_months(1)
```

When working with date-times, you can additionally add hours, minutes, and seconds.

```{r}
x <- date_time_build(2020, 1, 1, 2, 30, zone = "America/New_York")

x %>%
  add_days(1) %>%
  add_hours(2:5)
```

When adding units of time to a POSIXct, you have to be very careful with daylight saving time issues. clock tries to help you out by letting you know when you run into an issue:

```{r, error=TRUE}
x <- date_time_build(1970, 04, 25, 02, 30, 00, zone = "America/New_York")
x

# Daylight saving time gap on the 26th between 01:59:59 -> 03:00:00
x %>% add_days(1)
```

You can solve this using the `nonexistent` argument to control how these times should be handled.

```{r}
# Roll forward to the next valid moment in time
x %>% add_days(1, nonexistent = "roll-forward")

# Roll backward to the previous valid moment in time
x %>% add_days(1, nonexistent = "roll-backward")

# Shift forward by adding the size of the DST gap
# (this often keeps the time of day,
# but doesn't guaratee that relative ordering in `x` is maintained
# so I don't recommend it)
x %>% add_days(1, nonexistent = "shift-forward")

# Replace nonexistent times with an NA
x %>% add_days(1, nonexistent = "NA")
```

## Getting and setting

clock provides a family of getters and setters for working with dates and date-times. You can get and set the year, month, or day of a date.

```{r}
x <- date_build(2019, 5, 6)

get_year(x)
get_month(x)
get_day(x)

x %>%
  set_day(22) %>%
  set_month(10)
```

As you might expect by now, setting the date to an invalid date requires you to explicitly handle this:

```{r, error=TRUE}
x %>%
  set_day(31) %>%
  set_month(4)

x %>%
  set_day(31) %>%
  set_month(4, invalid = "previous")
```

You can additionally set the hour, minute, and second of a POSIXct.

```{r}
x <- date_time_build(2020, 1, 2, 3, zone = "America/New_York")
x

x %>%
  set_minute(5) %>%
  set_second(10)
```

As with other manipulations of POSIXct, you'll have to be aware of daylight saving time when setting components. You may need to supply the `nonexistent` or `ambiguous` arguments of the `set_*()` functions to handle these issues.

---
title: "Examples and Recipes"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Examples and Recipes}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(clock)
library(magrittr)
```

This vignette shows common examples and recipes that might be useful when learning about clock. Where possible, both the high and low level API are shown.

Many of these examples are adapted from the date C++ library's [Examples and Recipes](https://github.com/HowardHinnant/date/wiki/Examples-and-Recipes) page.

## The current local time

`zoned_time_now()` returns the current time in a particular time zone. It will display up to nanosecond precision, but the exact amount is OS dependent (on a Mac this displays microsecond level information at nanosecond resolution).

Using `""` as the time zone string will try and use whatever R thinks your local time zone is (i.e. from `Sys.timezone()`).

```{r, eval=FALSE}
zoned_time_now("")
#> <zoned_time<nanosecond><America/New_York (current)>[1]>
#> [1] "2021-02-10T15:54:29.875011000-05:00"
```

## The current time somewhere else

Pass a time zone name to `zoned_time_now()` to get the current time somewhere else.

```{r, eval=FALSE}
zoned_time_now("Asia/Shanghai")
#> <zoned_time<nanosecond><Asia/Shanghai>[1]>
#> [1] "2021-02-11T04:54:29.875011000+08:00"
```

## Set a meeting across time zones

Say you need to set a meeting with someone in Shanghai, but you live in New York. If you set a meeting for 9am, what time is that for them?

```{r}
my_time <- year_month_day(2019, 1, 30, 9) %>%
  as_naive_time() %>%
  as_zoned_time("America/New_York")

my_time

their_time <- zoned_time_set_zone(my_time, "Asia/Shanghai")

their_time
```

### High level API

```{r}
my_time <- as.POSIXct("2019-01-30 09:00:00", "America/New_York")

date_time_set_zone(my_time, "Asia/Shanghai")
```

## Force a specific time zone

Say your co-worker in Shanghai (from the last example) accidentally logged on at 9am _their time_. What time would this be for you?

The first step to solve this is to force `my_time` to have the same printed time, but use the Asia/Shanghai time zone. You can do this by going through naive-time:

```{r}
my_time <- year_month_day(2019, 1, 30, 9) %>%
  as_naive_time() %>%
  as_zoned_time("America/New_York")

my_time

# Drop the time zone information, retaining the printed time
my_time %>%
  as_naive_time()

# Add the correct time zone name back on,
# again retaining the printed time
their_9am <- my_time %>%
  as_naive_time() %>%
  as_zoned_time("Asia/Shanghai")

their_9am
```

Note that a conversion like this isn't always possible due to daylight saving time issues, in which case you might need to set the `nonexistent` and `ambiguous` arguments of `as_zoned_time()`.

What time would this have been for you in New York?

```{r}
zoned_time_set_zone(their_9am, "America/New_York")
```

### High level API

```{r}
my_time <- as.POSIXct("2019-01-30 09:00:00", "America/New_York")

my_time %>%
  as_naive_time() %>%
  as.POSIXct("Asia/Shanghai") %>%
  date_time_set_zone("America/New_York")
```

## Finding the next Monday (or Thursday)

Given a particular day precision naive-time, how can you compute the next Monday? This is very easily accomplished with `time_point_shift()`. It takes a time point vector and a "target" weekday, and shifts the time points to that target weekday.

```{r}
days <- as_naive_time(year_month_day(2019, c(1, 2), 1))

# A Tuesday and a Friday
as_weekday(days)

monday <- weekday(clock_weekdays$monday)

time_point_shift(days, monday)

as_weekday(time_point_shift(days, monday))
```

You can also shift to the previous instance of the target weekday:

```{r}
time_point_shift(days, monday, which = "previous")
```

If you happen to already be on the target weekday, the default behavior returns the input unchanged. However, you can also chose to advance to the next instance of the target.

```{r}
tuesday <- weekday(clock_weekdays$tuesday)

time_point_shift(days, tuesday)
time_point_shift(days, tuesday, boundary = "advance")
```

While `time_point_shift()` is built in to clock, it can be useful to discuss the arithmetic going on in the underlying weekday type which powers this function. To do so, we will build some parts of `time_point_shift()` from scratch.

The weekday type represents a single day of the week and implements _circular arithmetic_. Let's see the code for a simple version of `time_point_shift()` that just shifts to the next target weekday:

```{r}
next_weekday <- function(x, target) {
  x + (target - as_weekday(x))
}

next_weekday(days, monday)

as_weekday(next_weekday(days, monday))
```

Let's break down how `next_weekday()` works. The first step takes the difference between two weekday vectors. It does this using circular arithmetic. Once we get passed the 7th day of the week (whatever that may be), it wraps back around to the 1st day of the week. Implementing weekday arithmetic in this way means that the following nicely returns the number of days until the next Monday as a day based duration:

```{r}
monday - as_weekday(days)
```

Which can be added to our day precision `days` vector to get the date of the next Monday:

```{r}
days + (monday - as_weekday(days))
```

The current implementation will return the input if it is already on the target weekday. To use the `boundary = "advance"` behavior, you could implement `next_weekday()` as:

```{r}
next_weekday2 <- function(x, target) {
  x <- x + duration_days(1L)
  x + (target - as_weekday(x))
}

a_monday <- as_naive_time(year_month_day(2018, 12, 31))
as_weekday(a_monday)

next_weekday2(a_monday, monday)
```

### High level API

In the high level API, you can use `date_shift()`:

```{r}
monday <- weekday(clock_weekdays$monday)

x <- as.Date(c("2019-01-01", "2019-02-01"))

date_shift(x, monday)

# With a date-time
y <- as.POSIXct(
  c("2019-01-01 02:30:30", "2019-02-01 05:20:22"), 
  "America/New_York"
)

date_shift(y, monday)
```

Note that adding weekdays to a POSIXct could generate nonexistent or ambiguous times due to daylight saving time, which would have to be handled by supplying `nonexistent` and `ambiguous` arguments to `date_shift()`.

## Generate sequences of dates and date-times

clock implements S3 methods for the `seq()` generic function for the calendar and time point types it provides. The precision that you can generate sequences for depends on the type.

- year-month-day: Yearly or monthly sequences
- year-quarter-day: Yearly or quarterly sequences
- sys-time / naive-time: Weekly, Daily, Hourly, ..., Subsecond sequences

When generating sequences, the type and precision of `from` determine the result. For example:

```{r}
ym <- seq(year_month_day(2019, 1), by = 2, length.out = 10)
ym
```

```{r}
yq <- seq(year_quarter_day(2019, 1), by = 2, length.out = 10)
```

This allows you to generate sequences of year-months or year-quarters without having to worry about the day of the month/quarter becoming invalid. You can set the day of the results to get to a day precision calendar. For example, to get the last days of the month/quarter for this sequence:

```{r}
set_day(ym, "last")

set_day(yq, "last")
```

You won't be able to generate day precision sequences with calendars. Instead, you should use a time point.

```{r}
from <- as_naive_time(year_month_day(2019, 1, 1))
to <- as_naive_time(year_month_day(2019, 5, 15))

seq(from, to, by = 20)
```

If you use an integer `by` value, it is interpreted as a duration at the same precision as `from`. You can also use a duration object that can be cast to the same precision as `from`. For example, to generate a sequence spaced out by 90 minutes for these second precision end points:

```{r}
from <- as_naive_time(year_month_day(2019, 1, 1, 2, 30, 00))
to <- as_naive_time(year_month_day(2019, 1, 1, 12, 30, 00))

seq(from, to, by = duration_minutes(90))
```

### High level API

In the high level API, you can use `date_seq()` to generate sequences. This doesn't have all of the flexibility of the `seq()` methods above, but is still extremely useful and has the added benefit of switching between calendars, sys-times, and naive-times automatically for you.

If an integer `by` is supplied with a date `from`, it defaults to a daily sequence:

```{r}
date_seq(date_build(2019, 1), by = 2, total_size = 10)
```

You can generate a monthly sequence by supplying a month precision duration for `by`.

```{r}
date_seq(date_build(2019, 1), by = duration_months(2), total_size = 10)
```

If you supply `to`, be aware that all components of `to` that are more precise than the precision of `by` must match `from` exactly. For example, the day component of `from` and `to` doesn't match here, so the sequence isn't defined.

```{r, error=TRUE}
date_seq(
  date_build(2019, 1, 1),
  to = date_build(2019, 10, 2),
  by = duration_months(2)
)
```

`date_seq()` also catches invalid dates for you, forcing you to specify the `invalid` argument to specify how to handle them.

```{r, error=TRUE}
jan31 <- date_build(2019, 1, 31)
dec31 <- date_build(2019, 12, 31)

date_seq(jan31, to = dec31, by = duration_months(1))
```

By specifying `invalid = "previous"` here, we can generate month end values.

```{r}
date_seq(jan31, to = dec31, by = duration_months(1), invalid = "previous")
```

Compare this with the automatic "overflow" behavior of `seq()`, which is often a source of confusion.

```{r}
seq(jan31, to = dec31, by = "1 month")
```

## Grouping by months or quarters

When working on a data analysis, you might be required to summarize certain metrics at a monthly or quarterly level. With `calendar_group()`, you can easily summarize at the granular precision that you care about. Take this vector of day precision naive-times in 2019:

```{r}
from <- as_naive_time(year_month_day(2019, 1, 1))
to <- as_naive_time(year_month_day(2019, 12, 31))

x <- seq(from, to, by = duration_days(20))

x
```

To group by month, first convert to a year-month-day:

```{r}
ymd <- as_year_month_day(x)

head(ymd)

calendar_group(ymd, "month")
```

To group by quarter, convert to a year-quarter-day:

```{r}
yqd <- as_year_quarter_day(x)

head(yqd)

calendar_group(yqd, "quarter")
```

If you need to group by a multiple of months / quarters, you can do that too:

```{r}
calendar_group(ymd, "month", n = 2)

calendar_group(yqd, "quarter", n = 2)
```

Note that the returned calendar vector is at the precision we grouped by, not at the original precision with, say, the day of the month / quarter set to `1`.

Additionally, be aware that `calendar_group()` groups "within" the component that is one unit of precision larger than the `precision` you specify. So, when grouping by `"day"`, this groups by "day of the month", which can't cross the month or year boundary. If you need to bundle dates together by something like 60 days (i.e. crossing the month boundary), then you should use `time_point_floor()`.

### High level API

In the high level API, you can use `date_group()` to group Date vectors by one of their 3 components: year, month, or day. Since month precision dates can't be represented with Date vectors, `date_group()` sets the day of the month to 1.

```{r}
x <- seq(as.Date("2019-01-01"), as.Date("2019-12-31"), by = 20)

date_group(x, "month")
```

You won't be able to group by `"quarter"`, since this isn't one of the 3 components that the high level API lets you work with. Instead, this is a case where you should convert to a year-quarter-day, group on that type, then convert back to Date.

```{r}
x %>%
  as_year_quarter_day() %>%
  calendar_group("quarter") %>%
  set_day(1) %>%
  as.Date()
```

This is actually equivalent to `date_group(x, "month", n = 3)`. If your fiscal year starts in January, you can use that instead. However, if your fiscal year starts in a different month, say, June, you'll need to use the approach from above like so:

```{r}
x %>%
  as_year_quarter_day(start = clock_months$june) %>%
  calendar_group("quarter") %>%
  set_day(1) %>%
  as.Date()
```

## Flooring by days

While `calendar_group()` can group by "component", it isn't useful for bundling together sets of time points that can cross month/year boundaries, like "60 days" of data. For that, you are better off _flooring_ by rolling sets of 60 days.

```{r}
from <- as_naive_time(year_month_day(2019, 1, 1))
to <- as_naive_time(year_month_day(2019, 12, 31))

x <- seq(from, to, by = duration_days(20))
```

```{r}
time_point_floor(x, "day", n = 60)
```

Flooring operates on the underlying duration, which for day precision time points is a count of days since the _origin_, 1970-01-01.

```{r}
unclass(x[1])
```

The 60 day counter starts here, which means that any times between `[1970-01-01, 1970-03-02)` are all floored to 1970-01-01. At `1970-03-02`, the counter starts again.

If you would like to change this origin, you can provide a time point to start counting from with the `origin` argument. This is mostly useful if you are flooring by weeks and you want to change the day of the week that the count starts on. Since 1970-01-01 is a Thursday, flooring by 14 days defaults to returning all Thursdays.

```{r}
x <- seq(as_naive_time(year_month_day(2019, 1, 1)), by = 3, length.out = 10)
x

thursdays <- time_point_floor(x, "day", n = 14)
thursdays

as_weekday(thursdays)
```

You can use `origin` to change this to floor to Mondays.

```{r}
origin <- as_naive_time(year_month_day(2018, 12, 31))
as_weekday(origin)

mondays <- time_point_floor(x, "day", n = 14, origin = origin)
mondays

as_weekday(mondays)
```

### High level API

You can use `date_floor()` with Date and POSIXct types.

```{r}
x <- seq(as.Date("2019-01-01"), as.Date("2019-12-31"), by = 20)

date_floor(x, "day", n = 60)
```

The `origin` you provide should be another Date. For week precision flooring with Dates, you can specify `"week"` as the precision.

```{r}
x <- seq(as.Date("2019-01-01"), by = 3, length.out = 10)

origin <- as.Date("2018-12-31")

date_floor(x, "week", n = 2, origin = origin)
```

## Day of the year

To get the day of the year, convert to the year-day calendar type and extract the day with `get_day()`.

```{r}
x <- year_month_day(2019, clock_months$july, 4)

yd <- as_year_day(x)
yd

get_day(yd)
```

### High level API

```{r}
x <- as.Date("2019-07-04")

x %>%
  as_year_day() %>%
  get_day()
```

## Computing an age in years

To get the age of an individual in years, use `calendar_count_between()`.

```{r}
x <- year_month_day(1980, 12, 14:16)
today <- year_month_day(2005, 12, 15)

# Note that the month and day of the month are taken into account!
# (Time of day would also be taken into account if there was any.)
calendar_count_between(x, today, "year")
```

### High level API

You can use `date_count_between()` with Date and POSIXct types.

```{r}
x <- date_build(1980, 12, 14:16)
today <- date_build(2005, 12, 15)

date_count_between(x, today, "year")
```

## Computing number of weeks since the start of the year

`lubridate::week()` is a useful function that returns "the number of complete seven day periods that have occurred between the date and January 1st, plus one."

There is no direct equivalent to this, but it is possible to replicate with `calendar_start()` and `time_point_count_between()`.

```{r}
x <- year_month_day(2019, 11, 28)

# lubridate::week(as.Date(x))
# [1] 48

x_start <- calendar_start(x, "year")
x_start

time_point_count_between(
  as_naive_time(x_start),
  as_naive_time(x),
  "week"
) + 1L
```

You could also peek at the `lubridate::week()` implementation to see that this is just:

```{r}
doy <- get_day(as_year_day(x))
doy

(doy - 1L) %/% 7L + 1L
```

### High level API

This is actually a little easier in the high level API because you don't have to think about switching between types.

```{r}
x <- date_build(2019, 11, 28)

date_count_between(date_start(x, "year"), x, "week") + 1L
```

## Compute the number of months between two dates

How can we compute the number of months between these two dates?

```{r}
x <- year_month_day(2013, 10, 15)
y <- year_month_day(2016, 10, 13)
```

This is a bit of an ambiguous question because "month" isn't very well-defined, and there are various different interpretations we could take.

We might want to ignore the day component entirely, and just compute the number of months between `2013-10` and `2016-10`.

```{r}
calendar_narrow(y, "month") - calendar_narrow(x, "month")
```

Or we could include the day of the month, and say that `2013-10-15` to `2014-10-15` defines 1 month (i.e. you have to hit the same day of the month in the next month).

```{r}
calendar_count_between(x, y, "month")
```

With this you could also compute the number of days remaining between these two dates.

```{r}
x_close <- add_months(x, calendar_count_between(x, y, "month"))
x_close

x_close_st <- as_sys_time(x_close)
y_st <- as_sys_time(y)

time_point_count_between(x_close_st, y_st, "day")
```

Or we could compute the number of days between these two dates in units of seconds, and divide that by the average number of seconds in 1 proleptic Gregorian month.

```{r}
# Days between x and y
days <- as_sys_time(y) - as_sys_time(x)
days

# In units of seconds
days <- duration_cast(days, "second")
days <- as.numeric(days)
days

# Average number of seconds in 1 proleptic Gregorian month
avg_sec_in_month <- duration_cast(duration_months(1), "second")
avg_sec_in_month <- as.numeric(avg_sec_in_month)

days / avg_sec_in_month
```

### High level API

```{r}
x <- date_build(2013, 10, 15)
y <- date_build(2016, 10, 13)
```

To ignore the day of the month, first shift to the start of the month, then you can use `date_count_between()`.

```{r}
date_count_between(date_start(x, "month"), date_start(y, "month"), "month")
```

To utilize the day field, do the same as above but without calling `date_start()`.

```{r}
date_count_between(x, y, "month")
```

There is no high level equivalent to the average length of one proleptic Gregorian month example.

## Computing the ISO year or week

The ISO 8601 standard outlines an alternative calendar that is specified by the year, the week of the year, and the day of the week. It also specifies that the _start_ of the week is considered to be a Monday. This ends up meaning that the actual ISO year may be different from the Gregorian year, and is somewhat difficult to compute "by hand". Instead, you can use the `year_week_day()` calendar if you need to work with ISO week dates.

```{r}
x <- date_build(2019:2026)
y <- as_year_week_day(x, start = clock_weekdays$monday)

data.frame(x = x, y = y)
```

```{r}
get_year(y)
get_week(y)

# Last week in the ISO year
set_week(y, "last")
```

The year-week-day calendar is a fully supported calendar, meaning that all of the `calendar_*()` functions work on it:

```{r}
calendar_narrow(y, "week")
```

There is also an `iso_year_week_day()` calendar available, which is identical to `year_week_day(start = clock_weekdays$monday)`. That ISO calendar actually existed first, before we generalized it to any `start` weekday.

## Computing the Epidemiological year or week

Epidemiologists following the US CDC guidelines use a calendar that is similar to the ISO calendar, but defines the start of the week to be Sunday instead of Monday. `year_week_day()` supports this as well:

```{r}
x <- date_build(2019:2026)
iso <- as_year_week_day(x, start = clock_weekdays$monday)
epi <- as_year_week_day(x, start = clock_weekdays$sunday)

data.frame(x = x, iso = iso, epi = epi)
```

```{r}
get_year(epi)
get_week(epi)
```

## Converting a time zone abbreviation into a time zone name

It is possible that you might run into date-time strings of the form `"2020-10-25 01:30:00 IST"`, which contain a time zone _abbreviation_ rather than a full time zone name. Because time zone maintainers change the abbreviation they use throughout time, and because multiple time zones sometimes use the same abbreviation, it is generally impossible to parse strings of this form without more information. That said, if you know what time zone this abbreviation goes with, you can parse this time with `zoned_time_parse_abbrev()`, supplying the `zone`.

```{r}
x <- "2020-10-25 01:30:00 IST"

zoned_time_parse_abbrev(x, "Asia/Kolkata")
zoned_time_parse_abbrev(x, "Asia/Jerusalem")
```

If you _don't_ know what time zone this abbreviation goes with, then generally you are out of luck. However, there are low-level tools in this library that can help you generate a list of _possible_ zoned-times this could map to.

Assuming that `x` is a naive-time with its corresponding time zone abbreviation attached, the first thing to do is to parse this string as a naive-time.

```{r}
x <- naive_time_parse(x, format = "%Y-%m-%d %H:%M:%S IST")
x
```

Next, we'll develop a function that attempts to turn this naive-time into a zoned-time, iterating through all of the time zone names available in the time zone database. These time zone names are accessible through `tzdb_names()`. By using the low-level `naive_time_info()`, rather than `as_zoned_time()`, to lookup zone specific information, we'll also get back information about the UTC offset and time zone abbreviation that is currently in use. By matching this abbreviation against our input abbreviation, we can generate a list of zoned-times that use the abbreviation we care about at that particular instance in time.

```{r}
naive_find_by_abbrev <- function(x, abbrev) {
  if (!is_naive_time(x)) {
    abort("`x` must be a naive-time.")
  }
  if (length(x) != 1L) {
    abort("`x` must be length 1.")
  }
  if (!rlang::is_string(abbrev)) {
    abort("`abbrev` must be a single string.")
  }
  
  zones <- tzdb_names()
  info <- naive_time_info(x, zones)
  info$zones <- zones
  
  c(
    compute_uniques(x, info, abbrev),
    compute_ambiguous(x, info, abbrev)
  )
}

compute_uniques <- function(x, info, abbrev) {
  info <- info[info$type == "unique",]
  
  # If the abbreviation of the unique time matches the input `abbrev`,
  # then that candidate zone should be in the output
  matches <- info$first$abbreviation == abbrev
  zones <- info$zones[matches]
  
  lapply(zones, as_zoned_time, x = x)
}

compute_ambiguous <- function(x, info, abbrev) {
  info <- info[info$type == "ambiguous",]

  # Of the two possible times,
  # does the abbreviation of the earliest match the input `abbrev`?
  matches <- info$first$abbreviation == abbrev
  zones <- info$zones[matches]
  
  earliest <- lapply(zones, as_zoned_time, x = x, ambiguous = "earliest")
  
  # Of the two possible times,
  # does the abbreviation of the latest match the input `abbrev`?
  matches <- info$second$abbreviation == abbrev
  zones <- info$zones[matches]
  
  latest <- lapply(zones, as_zoned_time, x = x, ambiguous = "latest")
  
  c(earliest, latest)
}
```

```{r}
candidates <- naive_find_by_abbrev(x, "IST")
candidates
```

While it looks like we got 7 candidates, in reality we only have 3. Asia/Kolkata, Europe/Dublin, and Asia/Jerusalem are our 3 candidates. The others are aliases of those 3 that have been retired but are kept for backwards compatibility.

Looking at the code, there are two ways to add a candidate time zone name to the list.

If there is a unique mapping from `{naive-time, zone}` to `sys-time`,  then we check if the abbreviation that goes with that unique mapping matches our input abbreviation. If so, then we convert `x` to a zoned-time with that time zone.

If there is an ambiguous mapping from `{naive-time, zone}` to `sys-time`, which is due to a daylight saving fallback, then we check the abbreviation of both the _earliest_ and _latest_ possible times. If either matches, then we convert `x` to a zoned-time using that time zone and the information about which of the two ambiguous times were used.

This example is particularly interesting, since each of the 3 candidates came from a different path. The Asia/Kolkata one is unique, the Europe/Dublin one is ambiguous but the earliest was chosen, and the Asia/Jerusalem one is ambiguous but the latest was chosen:

```{r}
as_zoned_time(x, "Asia/Kolkata")
as_zoned_time(x, "Europe/Dublin", ambiguous = "earliest")
as_zoned_time(x, "Asia/Jerusalem", ambiguous = "latest")
```

## When is the next daylight saving time event?

Given a particular zoned-time, when will it next be affected by daylight saving time? For this, we can use a relatively low level helper, `zoned_time_info()`. It returns a data frame of information about the current daylight saving time transition points, along with information about the offset, the current time zone abbreviation, and whether or not daylight saving time is currently active or not.

```{r}
x <- zoned_time_parse_complete("2019-01-01T00:00:00-05:00[America/New_York]")

info <- zoned_time_info(x)

# Beginning of the current DST range
info$begin

# Beginning of the next DST range
info$end
```

So on 2018-11-04 at (the second) 1 o'clock hour, daylight saving time was turned off. On 2019-03-10 at 3 o'clock, daylight saving time will be considered on again. This is the next moment in time right after a daylight saving time gap of 1 hour, which you can see by subtracting 1 second (in sys-time):

```{r}
# Last moment in time in the current DST range
info$end %>%
  as_sys_time() %>%
  add_seconds(-1) %>%
  as_zoned_time(zoned_time_zone(x))
```

### High level API

`date_time_info()` exists in the high level API to do a similar thing. It is basically the same as `zoned_time_info()`, except the `begin` and `end` columns are returned as R POSIXct date-times rather than zoned-times, and the `offset` column is returned as an integer rather than as a clock duration (since we try not to expose high level API users to low level types).

```{r}
x <- date_time_parse("2019-01-01 00:00:00", zone = "America/New_York")

date_time_info(x)
```

# can't compare zoned-times with different zones

    Can't combine `..1` <zoned_time<second><America/New_York>> and `..2` <zoned_time<second><UTC>>.
    Zones can't differ.

# normal print method works

    Code
      x
    Output
      <zoned_time<second><America/New_York>[5]>
      [1] "2018-12-31T19:00:00-05:00" "2019-01-31T19:00:00-05:00"
      [3] "2019-02-28T19:00:00-05:00" "2019-03-31T20:00:00-04:00"
      [5] "2019-04-30T20:00:00-04:00"

# can limit with `max`

    Code
      print(x, max = 2)
    Output
      <zoned_time<second><America/New_York>[5]>
      [1] "2018-12-31T19:00:00-05:00" "2019-01-31T19:00:00-05:00"
      Reached `max` or `getOption('max.print')`. Omitted 3 values.

---

    Code
      print(x, max = 4)
    Output
      <zoned_time<second><America/New_York>[5]>
      [1] "2018-12-31T19:00:00-05:00" "2019-01-31T19:00:00-05:00"
      [3] "2019-02-28T19:00:00-05:00" "2019-03-31T20:00:00-04:00"
      Reached `max` or `getOption('max.print')`. Omitted 1 value.

---

    Code
      print(x, max = 5)
    Output
      <zoned_time<second><America/New_York>[5]>
      [1] "2018-12-31T19:00:00-05:00" "2019-01-31T19:00:00-05:00"
      [3] "2019-02-28T19:00:00-05:00" "2019-03-31T20:00:00-04:00"
      [5] "2019-04-30T20:00:00-04:00"

---

    Code
      print(x, max = 6)
    Output
      <zoned_time<second><America/New_York>[5]>
      [1] "2018-12-31T19:00:00-05:00" "2019-01-31T19:00:00-05:00"
      [3] "2019-02-28T19:00:00-05:00" "2019-03-31T20:00:00-04:00"
      [5] "2019-04-30T20:00:00-04:00"

# `max` defaults to `getOption('max.print')` but can be overridden

    Code
      x
    Output
      <zoned_time<second><America/New_York>[5]>
      [1] "2018-12-31T19:00:00-05:00" "2019-01-31T19:00:00-05:00"
      [3] "2019-02-28T19:00:00-05:00"
      Reached `max` or `getOption('max.print')`. Omitted 2 values.

---

    Code
      print(x, max = 4)
    Output
      <zoned_time<second><America/New_York>[5]>
      [1] "2018-12-31T19:00:00-05:00" "2019-01-31T19:00:00-05:00"
      [3] "2019-02-28T19:00:00-05:00" "2019-03-31T20:00:00-04:00"
      Reached `max` or `getOption('max.print')`. Omitted 1 value.

---

    Code
      print(x, max = 5)
    Output
      <zoned_time<second><America/New_York>[5]>
      [1] "2018-12-31T19:00:00-05:00" "2019-01-31T19:00:00-05:00"
      [3] "2019-02-28T19:00:00-05:00" "2019-03-31T20:00:00-04:00"
      [5] "2019-04-30T20:00:00-04:00"

# cannot parse nonexistent time

    Code
      zoned_time_parse_complete(x)
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      <zoned_time<second><America/New_York>[1]>
      [1] NA

# offset must align with unique offset

    Code
      zoned_time_parse_complete(x)
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      <zoned_time<second><America/New_York>[1]>
      [1] NA

# offset must align with one of two possible ambiguous offsets

    Code
      zoned_time_parse_complete(x)
    Condition
      Warning:
      Failed to parse 2 strings, beginning at location 1. Returning `NA` at the locations where there were parse failures.
    Output
      <zoned_time<second><America/New_York>[2]>
      [1] NA NA

# cannot have differing zone names

    All elements of `x` must have the same time zone name. Found different zone names of: 'America/New_York' and 'America/Los_Angeles'.

# zone name must be valid

    `%Z` must be used, and must result in a valid time zone name, not 'America/New_Yor'.

# abbreviation must match the one implied from naive + time zone name lookup

    Code
      zoned_time_parse_abbrev(x, "America/New_York")
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      <zoned_time<second><America/New_York>[1]>
      [1] NA

# %Z must be used

    `%Z` must be used and must result in a time zone abbreviation.

# zoned-times don't support arithmetic

    Zoned-times don't support `add_years()`.

---

    Zoned-times don't support `add_quarters()`.

---

    Zoned-times don't support `add_months()`.

---

    Zoned-times don't support `add_weeks()`.

---

    Zoned-times don't support `add_days()`.

---

    Zoned-times don't support `add_hours()`.

---

    Zoned-times don't support `add_minutes()`.

---

    Zoned-times don't support `add_seconds()`.

---

    Zoned-times don't support `add_milliseconds()`.

---

    Zoned-times don't support `add_microseconds()`.

---

    Zoned-times don't support `add_nanoseconds()`.

# precision: can only be called on zoned-times

    `x` must be a 'clock_zoned_time'.

# POSIXt time zones are standardized as expected

    Code
      posixt_tzone_standardize(character())
    Condition
      Warning in `posixt_tzone_standardize()`:
      POSIXt input had a corrupt time zone attribute of `character(0)`. Defaulting to the current zone by assuming the zone is `""`.
    Output
      [1] ""

---

    A POSIXt time zone should either be a character vector or `NULL`.

# validates value ranges

    `year` must be within the range of [-32767, 32767], not 50000.

---

    `quarter` must be within the range of [1, 4], not 5.

---

    `day` must be within the range of [1, 92], not 93.

---

    `hour` must be within the range of [0, 23], not 24.

---

    `minute` must be within the range of [0, 59], not 60.

---

    `second` must be within the range of [0, 59], not 60.

---

    `subsecond` must be within the range of [0, 999], not 1000.

---

    `subsecond` must be within the range of [0, 999999], not 1000000.

---

    `subsecond` must be within the range of [0, 999999999], not 1000000000.

# full ptype is correct

    [1] "year_quarter_day<January><year>"

---

    [1] "year_quarter_day<February><year>"

---

    [1] "year_quarter_day<January><day>"

---

    [1] "year_quarter_day<January><nanosecond>"

---

    [1] "year_quarter_day<January><day>"

# abbreviated ptype is correct

    [1] "yqd<Jan><year>"

---

    [1] "yqd<Feb><year>"

---

    [1] "yqd<Jan><day>"

---

    [1] "yqd<Jan><nano>"

---

    [1] "yqd<Jan><day>"

# default formats are correct

    Code
      format(year_quarter_day(2019))
    Output
      [1] "2019"

---

    Code
      format(year_quarter_day(2019, 1))
    Output
      [1] "2019-Q1"

---

    Code
      format(year_quarter_day(2019, 1, 1, 1))
    Output
      [1] "2019-Q1-01T01"

---

    Code
      format(year_quarter_day(2019, 1, 1, 1, 2, 3, 50, subsecond_precision = "microsecond"))
    Output
      [1] "2019-Q1-01T01:02:03.000050"

# can't compute a unsupported count precision

    Code
      (expect_error(calendar_count_between(x, x, "day")))
    Output
      <error/rlang_error>
      Error in `calendar_count_between_standardize_precision_n()`:
      ! `precision` must be one of: 'year', 'quarter'.

# only granular precisions are allowed

    `from` must be 'year' or 'quarter' precision.

# strict mode can be activated

    The global option, `clock.strict`, is currently set to `TRUE`. In this mode, `invalid` must be set and cannot be left as `NULL`.

# can't accidentally supply `zone` to reinterpret date-time in new zone

    `...` must be empty.
    x Problematic argument:
    * zone = "America/New_York"

# can resolve nonexistent midnight issues for Date -> POSIXct

    Nonexistent time due to daylight saving time at location 1.
    i Resolve nonexistent time issues by specifying the `nonexistent` argument.

# can resolve ambiguous midnight issues for Date -> POSIXct

    Ambiguous time due to daylight saving time at location 1.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# can handle nonexistent times resulting from grouping

    Nonexistent time due to daylight saving time at location 1.
    i Resolve nonexistent time issues by specifying the `nonexistent` argument.

# can't group by finer precisions

    Can't group at a precision (nanosecond) that is more precise than `x` (second).

# can't group by non-year-month-day precisions

    `precision` must be a valid precision for a 'year_month_day'.

# flooring can handle nonexistent times

    Nonexistent time due to daylight saving time at location 2.
    i Resolve nonexistent time issues by specifying the `nonexistent` argument.

# `origin` is floored to the precision of `precision` with a potential warning before rounding

    Code
      date_floor(x, "day", origin = origin)
    Condition
      Warning:
      `origin` has been floored from 'second' precision to 'day' precision to match `precision`. This floor has lost information.
    Output
      [1] "1970-01-01 EST" "1970-01-02 EST"

# `origin` is validated

    `origin` must be a 'POSIXt'.

---

    `origin` must not be `NA` or an infinite date.

---

    `origin` must not be `NA` or an infinite date.

---

    `origin` must have length 1.

---

    `origin` must have the same time zone as `x`.

---

    `origin` must have the same time zone as `x`.

# default format is correct

    Code
      date_format(x)
    Output
      [1] "2019-01-01T00:00:00-05:00[America/New_York]"

# can format date-times

    Code
      vapply(X = formats, FUN = function(format) date_format(x, format = format),
      FUN.VALUE = character(1))
    Output
                              C: %C                         y: %y 
                            "C: 20"                       "y: 18" 
                              Y: %Y                         b: %b 
                          "Y: 2018"                      "b: Dec" 
                              h: %h                         B: %B 
                           "h: Dec"                 "B: December" 
                              m: %m                         d: %d 
                            "m: 12"                       "d: 31" 
                              a: %a                         A: %A 
                           "a: Mon"                   "A: Monday" 
                              w: %w                         g: %g 
                             "w: 1"                       "g: 19" 
                              G: %G                         V: %V 
                          "G: 2019"                       "V: 01" 
                              u: %u                         U: %U 
                             "u: 1"                       "U: 52" 
                              W: %W                         j: %j 
                            "W: 53"                      "j: 365" 
                              D: %D                         x: %x 
                      "D: 12/31/18"                 "x: 12/31/18" 
                              F: %F                         H: %H 
                    "F: 2018-12-31"                       "H: 23" 
                              I: %I                         M: %M 
                            "I: 11"                       "M: 59" 
                              S: %S                         p: %p 
                            "S: 59"                       "p: PM" 
                              R: %R                         T: %T 
                         "R: 23:59"                 "T: 23:59:59" 
                              X: %X                         r: %r 
                      "X: 23:59:59"              "r: 11:59:59 PM" 
                              c: %c                         %: %% 
      "c: Mon Dec 31 23:59:59 2018"                        "%: %" 
                              z: %z                       Ez: %Ez 
                         "z: -0500"                  "Ez: -05:00" 
                              Z: %Z 
              "Z: America/New_York" 

---

    Code
      vapply(X = formats, FUN = function(format) date_format(x, format = format,
        locale = clock_locale("fr")), FUN.VALUE = character(1))
    Output
                                C: %C                           y: %y 
                              "C: 20"                         "y: 18" 
                                Y: %Y                           b: %b 
                            "Y: 2018"                       "b: déc." 
                                h: %h                           B: %B 
                            "h: déc."                   "B: décembre" 
                                m: %m                           d: %d 
                              "m: 12"                         "d: 31" 
                                a: %a                           A: %A 
                            "a: lun."                      "A: lundi" 
                                w: %w                           g: %g 
                               "w: 1"                         "g: 19" 
                                G: %G                           V: %V 
                            "G: 2019"                         "V: 01" 
                                u: %u                           U: %U 
                               "u: 1"                         "U: 52" 
                                W: %W                           j: %j 
                              "W: 53"                        "j: 365" 
                                D: %D                           x: %x 
                        "D: 12/31/18"                   "x: 12/31/18" 
                                F: %F                           H: %H 
                      "F: 2018-12-31"                         "H: 23" 
                                I: %I                           M: %M 
                              "I: 11"                         "M: 59" 
                                S: %S                           p: %p 
                              "S: 59"                         "p: PM" 
                                R: %R                           T: %T 
                           "R: 23:59"                   "T: 23:59:59" 
                                X: %X                           r: %r 
                        "X: 23:59:59"                "r: 11:59:59 PM" 
                                c: %c                           %: %% 
      "c: lun. déc. 31 23:59:59 2018"                          "%: %" 
                                z: %z                         Ez: %Ez 
                           "z: -0500"                    "Ez: -05:00" 
                                Z: %Z 
                "Z: America/New_York" 

# can resolve ambiguity and nonexistent times

    Nonexistent time due to daylight saving time at location 1.
    i Resolve nonexistent time issues by specifying the `nonexistent` argument.

---

    Ambiguous time due to daylight saving time at location 1.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# failure to parse throws a warning

    Code
      date_time_parse("foo", "America/New_York")
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      [1] NA

# throws warning on failed parses

    Code
      date_time_parse_complete("foo")
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      [1] NA

# abbrev - throws warning on failed parses

    Code
      date_time_parse_abbrev("foo", "America/New_York")
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      [1] NA

# `ambiguous = x` retains the offset of `x` if applicable

    Ambiguous time due to daylight saving time at location 1.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# `zone` is required

    `zone` is a required argument to `date_time_build()`.

# can handle invalid dates

    Invalid date found at location 2.
    i Resolve invalid date issues by specifying the `invalid` argument.

# can handle nonexistent times

    Nonexistent time due to daylight saving time at location 1.
    i Resolve nonexistent time issues by specifying the `nonexistent` argument.

# can handle ambiguous times

    Ambiguous time due to daylight saving time at location 1.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# start: can't use invalid precisions

    `precision` must be a valid precision for a 'year_month_day'.

# can resolve nonexistent start issues

    Nonexistent time due to daylight saving time at location 1.
    i Resolve nonexistent time issues by specifying the `nonexistent` argument.

# can resolve ambiguous start issues

    Ambiguous time due to daylight saving time at location 1.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# end: can't use invalid precisions

    `precision` must be a valid precision for a 'year_month_day'.

# daily `by` uses naive-time around DST gaps

    Nonexistent time due to daylight saving time at location 2.
    i Resolve nonexistent time issues by specifying the `nonexistent` argument.

# daily `by` uses naive-time around DST fallbacks

    Ambiguous time due to daylight saving time at location 2.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# monthly / yearly `by` uses calendar -> naive-time around DST gaps

    Nonexistent time due to daylight saving time at location 2.
    i Resolve nonexistent time issues by specifying the `nonexistent` argument.

# monthly / yearly `by` uses calendar -> naive-time around DST fallbacks

    Ambiguous time due to daylight saving time at location 2.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# can resolve invalid dates

    Invalid date found at location 2.
    i Resolve invalid date issues by specifying the `invalid` argument.

# components of `to` more precise than `by` must match `from`

    All components of `from` and `to` more precise than 'minute' must match.

---

    All components of `from` and `to` more precise than 'day' must match.

---

    All components of `from` and `to` more precise than 'month' must match.

---

    All components of `from` and `to` more precise than 'month' must match.

---

    All components of `from` and `to` more precise than 'year' must match.

# `to` must have same time zone as `by`

    `from` and `to` must have identical time zones.

# validates integerish `by`

    Can't convert `by` <double> to <integer>.

# validates `total_size` early

    Can't convert `total_size` <double> to <integer>.

---

    `total_size` can't be `NA`.

---

    `total_size` can't be negative.

# `to` and `total_size` must not generate a non-fractional sequence

    The supplied output size does not result in a non-fractional sequence between `from` and `to`.

# requires exactly two optional arguments

    Must specify exactly two of:
    - `to`
    - `by`
    - `total_size`

---

    Must specify exactly two of:
    - `to`
    - `by`
    - `total_size`

---

    Must specify exactly two of:
    - `to`
    - `by`
    - `total_size`

# requires `to` to be POSIXt

    If supplied, `to` must be a <POSIXct> or <POSIXlt>.

# requires year, month, day, hour, minute, or second precision

    `by` must have a precision of 'year', 'quarter', 'month', 'week', 'day', 'hour', 'minute', or 'second'.

# checks empty dots

    `...` must be empty.
    x Problematic argument:
    * ..1 = new_datetime(2)
    i Did you forget to name an argument?

# must use a valid POSIXt precision

    Code
      (expect_error(date_count_between(x, x, "millisecond")))
    Output
      <error/rlang_error>
      Error in `date_count_between_impl()`:
      ! `precision` must be one of: 'year', 'quarter', 'month', 'week', 'day', 'hour', 'minute', 'second'.

# can't count between a POSIXt and a Date

    Code
      (expect_error(date_count_between(x, y, "year")))
    Output
      <error/rlang_error>
      Error in `date_count_between()`:
      ! `end` must be a <POSIXt>.

# <posixt> op <duration>

    no applicable method for 'add_milliseconds' applied to an object of class "c('POSIXct', 'POSIXt')"

---

    no applicable method for 'add_milliseconds' applied to an object of class "c('POSIXlt', 'POSIXt')"

---

    <datetime<America/New_York>> * <duration<year>> is not permitted

---

    <POSIXlt<America/New_York>> * <duration<year>> is not permitted

# <duration> op <posixt>

    <duration<year>> - <datetime<America/New_York>> is not permitted
    Can't subtract a POSIXct/POSIXlt from a duration.

---

    <duration<year>> - <POSIXlt<America/New_York>> is not permitted
    Can't subtract a POSIXct/POSIXlt from a duration.

---

    no applicable method for 'add_milliseconds' applied to an object of class "c('POSIXct', 'POSIXt')"

---

    no applicable method for 'add_milliseconds' applied to an object of class "c('POSIXlt', 'POSIXt')"

---

    <duration<year>> * <datetime<America/New_York>> is not permitted

---

    <duration<year>> * <POSIXlt<America/New_York>> is not permitted

# validates value ranges

    `year` must be within the range of [-32767, 32767], not 50000.

---

    `week` must be within the range of [1, 53], not 54.

---

    `day` must be within the range of [1, 7], not 8.

---

    `hour` must be within the range of [0, 23], not 24.

---

    `minute` must be within the range of [0, 59], not 60.

---

    `second` must be within the range of [0, 59], not 60.

---

    `subsecond` must be within the range of [0, 999], not 1000.

---

    `subsecond` must be within the range of [0, 999999], not 1000000.

---

    `subsecond` must be within the range of [0, 999999999], not 1000000000.

# full ptype is correct

    [1] "iso_year_week_day<year>"

---

    [1] "iso_year_week_day<day>"

---

    [1] "iso_year_week_day<nanosecond>"

---

    [1] "iso_year_week_day<week>"

# abbreviated ptype is correct

    [1] "iso_ywd<year>"

---

    [1] "iso_ywd<day>"

---

    [1] "iso_ywd<nano>"

---

    [1] "iso_ywd<week>"

# default formats are correct

    Code
      format(iso_year_week_day(2019))
    Output
      [1] "2019"

---

    Code
      format(iso_year_week_day(2019, 1))
    Output
      [1] "2019-W01"

---

    Code
      format(iso_year_week_day(2019, 1, 1, 1))
    Output
      [1] "2019-W01-1T01"

---

    Code
      format(iso_year_week_day(2019, 1, 1, 1, 2, 3, 50, subsecond_precision = "microsecond"))
    Output
      [1] "2019-W01-1T01:02:03.000050"

# can't compute a unsupported count precision

    Code
      (expect_error(calendar_count_between(x, x, "week")))
    Output
      <error/rlang_error>
      Error in `calendar_count_between_standardize_precision_n()`:
      ! `precision` must be one of: 'year'.

# only year precision is allowed

    `from` must be 'year' precision.

# strict mode can be activated

    The global option, `clock.strict`, is currently set to `TRUE`. In this mode, `invalid` must be set and cannot be left as `NULL`.

# x must be naive

    `x` must be a naive-time.

# zone is vectorized and recycled against x

    Can't recycle `x` (size 4) to match `zone` (size 2).

# cannot parse invalid dates

    Code
      naive_time_parse(x, precision = "day")
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      <time_point<naive><day>[1]>
      [1] NA

# failure to parse throws a warning

    Code
      naive_time_parse("foo")
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      <time_point<naive><second>[1]>
      [1] NA

# default format is correct

    Code
      format(naive_seconds(0))
    Output
      [1] "1970-01-01T00:00:00"

# `%Z` generates format warnings (#204)

    Code
      format(x, format = "%Z")
    Condition
      Warning:
      Failed to format 1 string at location 1. Returning `NA` at that location.
    Output
      [1] NA

---

    Code
      format(c(x, x), format = "%Z")
    Condition
      Warning:
      Failed to format 2 strings, beginning at location 1. Returning `NA` at the locations where there were format failures.
    Output
      [1] NA NA

# `%z` generates format warnings (#204)

    Code
      format(x, format = "%z")
    Condition
      Warning:
      Failed to format 1 string at location 1. Returning `NA` at that location.
    Output
      [1] NA

---

    Code
      format(c(x, x), format = "%z")
    Condition
      Warning:
      Failed to format 2 strings, beginning at location 1. Returning `NA` at the locations where there were format failures.
    Output
      [1] NA NA

# can resolve ambiguous issues - character

    Ambiguous time due to daylight saving time at location 1.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# can resolve ambiguous issues - zoned-time

    Ambiguous time due to daylight saving time at location 2.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

---

    Ambiguous time due to daylight saving time at location 1.
    i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# can resolve nonexistent issues

    Nonexistent time due to daylight saving time at location 1.
    i Resolve nonexistent time issues by specifying the `nonexistent` argument.

# `ambiguous` is validated

    `ambiguous` must be a character vector, a zoned-time, a POSIXct, or a list.

---

    'foo' is not a recognized `ambiguous` option.

---

    `ambiguous` must have length 1, or 0.

---

    A zoned-time or POSIXct `ambiguous` must have the same zone as `zone`.

---

    A list `ambiguous` must have length 2.

---

    The first element of a list `ambiguous` must be a zoned-time or POSIXt.

---

    The second element of a list `ambiguous` must be a character vector, or `NULL`.

---

    'foo' is not a recognized `ambiguous` option.

# `nonexistent` is validated

    `nonexistent` must be a character vector, or `NULL`.

---

    'foo' is not a recognized `nonexistent` option.

---

    `nonexistent` must have length 1, or 0.

# zone is validated

    'foo' is not a known time zone.

---

    `zone` must be a single string.

---

    `zone` must be a single string.

# strict mode can be activated - nonexistent

    The global option, `clock.strict`, is currently set to `TRUE`. In this mode, `nonexistent` must be set and cannot be left as `NULL`.

# strict mode can be activated - ambiguous

    The global option, `clock.strict`, is currently set to `TRUE`. In this mode, `ambiguous` must be set and cannot be left as `NULL`. Additionally, `ambiguous` cannot be set to a zoned-time or POSIXct unless it is paired with an ambiguous time resolution strategy, like: `list(<zoned-time>, 'earliest')`.

---

    The global option, `clock.strict`, is currently set to `TRUE`. In this mode, `ambiguous` must be set and cannot be left as `NULL`. Additionally, `ambiguous` cannot be set to a zoned-time or POSIXct unless it is paired with an ambiguous time resolution strategy, like: `list(<zoned-time>, 'earliest')`.

---

    The global option, `clock.strict`, is currently set to `TRUE`. In this mode, `ambiguous` must be set and cannot be left as `NULL`. Additionally, `ambiguous` cannot be set to a zoned-time or POSIXct unless it is paired with an ambiguous time resolution strategy, like: `list(<zoned-time>, 'earliest')`.

# empty dots are checked

    `...` must be empty.
    x Problematic argument:
    * ..1 = "roll-forward"
    i Did you forget to name an argument?

# normal print method works

    Code
      x
    Output
      <year_month_day<month>[5]>
      [1] "2019-01" "2019-02" "2019-03" "2019-04" "2019-05"

# can limit with `max`

    Code
      print(x, max = 2)
    Output
      <year_month_day<month>[5]>
      [1] "2019-01" "2019-02"
      Reached `max` or `getOption('max.print')`. Omitted 3 values.

---

    Code
      print(x, max = 4)
    Output
      <year_month_day<month>[5]>
      [1] "2019-01" "2019-02" "2019-03" "2019-04"
      Reached `max` or `getOption('max.print')`. Omitted 1 value.

---

    Code
      print(x, max = 5)
    Output
      <year_month_day<month>[5]>
      [1] "2019-01" "2019-02" "2019-03" "2019-04" "2019-05"

---

    Code
      print(x, max = 6)
    Output
      <year_month_day<month>[5]>
      [1] "2019-01" "2019-02" "2019-03" "2019-04" "2019-05"

# `max` defaults to `getOption('max.print')` but can be overridden

    Code
      x
    Output
      <year_month_day<month>[5]>
      [1] "2019-01" "2019-02" "2019-03"
      Reached `max` or `getOption('max.print')`. Omitted 2 values.

---

    Code
      print(x, max = 4)
    Output
      <year_month_day<month>[5]>
      [1] "2019-01" "2019-02" "2019-03" "2019-04"
      Reached `max` or `getOption('max.print')`. Omitted 1 value.

---

    Code
      print(x, max = 5)
    Output
      <year_month_day<month>[5]>
      [1] "2019-01" "2019-02" "2019-03" "2019-04" "2019-05"

# `max` is validated

    `max` must be a positive number.

---

    `max` must be a single number, or `NULL`.

---

    `max` must be a single number, or `NULL`.

---

    Can't convert `max` <character> to <integer>.

# group: `precision` is validated

    `precision` not recognized.

# group: `precision` must be calendar specific

    `precision` must be a valid precision for a 'year_month_day'.

# group: `precision` can't be wider than `x`

    Can't group at a precision (second) that is more precise than `x` (day).

# group: can't group a subsecond precision `x` at another subsecond precision

    Can't group a subsecond precision `x` (nanosecond) by another subsecond precision (microsecond).

# narrow: `precision` is validated

    `precision` not recognized.

# narrow: `precision` must be calendar specific

    `precision` must be a valid precision for a 'year_month_day'.

# narrow: `precision` can't be wider than `x`

    Can't narrow to a precision (second) that is wider than `x` (day).

# narrow: can't narrow a subsecond precision `x` to another subsecond precision

    Can't narrow a subsecond precision `x` (nanosecond) to another subsecond precision (microsecond).

# widen: `precision` is validated

    `precision` not recognized.

# widen: `precision` must be calendar specific

    `precision` must be a valid precision for a 'year_month_day'.

# widen: `precision` can't be narrower than `x`

    Can't widen to a precision (month) that is narrower than `x` (day).

# widen: can't widen a subsecond precision `x` to another subsecond precision

    Can't widen a subsecond precision `x` (millisecond) to another subsecond precision (microsecond).

# start: `x` is validated

    no applicable method for 'calendar_start' applied to an object of class "c('double', 'numeric')"

# start: `precision` is validated

    `precision` not recognized.

---

    `precision` must be a string.

# start: errors on unsupported precision

    `precision` must be a valid precision for a 'year_month_day'.

# start: `precision` can't be more precise than `x`

    Can't compute the start of `x` (year) at a more precise precision (month).

# start: can't mix different subsecond precisions

    Can't compute the start of a subsecond precision `x` (microsecond) at another subsecond precision (millisecond).

# end: `x` is validated

    no applicable method for 'calendar_end' applied to an object of class "c('double', 'numeric')"

# end: `precision` is validated

    `precision` not recognized.

---

    `precision` must be a string.

# end: errors on unsupported precision

    `precision` must be a valid precision for a 'year_month_day'.

# end: `precision` can't be more precise than `x`

    Can't compute the end of `x` (year) at a more precise precision (month).

# end: can't mix different subsecond precisions

    Can't compute the end of a subsecond precision `x` (microsecond) at another subsecond precision (millisecond).

# `end` must be a calendar

    Code
      (expect_error(calendar_count_between(x, 1, "year")))
    Output
      <error/rlang_error>
      Error in `calendar_count_between()`:
      ! `end` must be a <clock_calendar>.

# can't count with a precision finer than the calendar precision

    Code
      (expect_error(calendar_count_between(x, x, "month")))
    Output
      <error/rlang_error>
      Error in `calendar_count_between()`:
      ! Precision of inputs must be at least as precise as `precision`.

# `n` is validated

    Code
      (expect_error(calendar_count_between(x, x, "year", n = NA_integer_)))
    Output
      <error/rlang_error>
      Error in `calendar_count_between()`:
      ! `n` must be a single positive integer.
    Code
      (expect_error(calendar_count_between(x, x, "year", n = -1)))
    Output
      <error/rlang_error>
      Error in `calendar_count_between()`:
      ! `n` must be a single positive integer.
    Code
      (expect_error(calendar_count_between(x, x, "year", n = 1.5)))
    Output
      <error/vctrs_error_cast_lossy>
      Error in `calendar_count_between()`:
      ! Can't convert from `n` <double> to <integer> due to loss of precision.
      * Locations: 1
    Code
      (expect_error(calendar_count_between(x, x, "year", n = "x")))
    Output
      <error/vctrs_error_incompatible_type>
      Error in `calendar_count_between()`:
      ! Can't convert `n` <character> to <integer>.
    Code
      (expect_error(calendar_count_between(x, x, "year", n = c(1L, 2L))))
    Output
      <error/rlang_error>
      Error in `calendar_count_between()`:
      ! `n` must be a single positive integer.

# precision: can only be called on calendars

    no applicable method for 'calendar_precision' applied to an object of class "c('clock_sys_time', 'clock_time_point', 'clock_rcrd', 'vctrs_rcrd', 'vctrs_vctr')"

# failure to parse throws a warning

    Code
      sys_time_parse("foo")
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      <time_point<sys><second>[1]>
      [1] NA

# `precision` must be at least second

    Code
      sys_time_parse_RFC_3339(x, precision = "day")
    Condition
      Error in `validate_RFC_3339_precision_string()`:
      ! `precision` must be at least 'second' precision.

# `separator` is validated

    Code
      sys_time_parse_RFC_3339(x, separator = 1)
    Condition
      Error in `sys_time_parse_RFC_3339()`:
      ! `separator` must be a string or character vector.

---

    Code
      sys_time_parse_RFC_3339(x, separator = "TT")
    Condition
      Error in `sys_time_parse_RFC_3339()`:
      ! `separator` must be one of "T", "t", or " ", not "TT".

# `offset` is validated

    Code
      sys_time_parse_RFC_3339(x, offset = 1)
    Condition
      Error in `sys_time_parse_RFC_3339()`:
      ! `offset` must be a string or character vector.

---

    Code
      sys_time_parse_RFC_3339(x, offset = "ZZ")
    Condition
      Error in `sys_time_parse_RFC_3339()`:
      ! `offset` must be one of "Z", "z", "%z", or "%Ez", not "ZZ".
      i Did you mean "%z"?

# sys-time-parse-RFC-3339: empty dots are checked

    Code
      sys_time_parse_RFC_3339(x, 1)
    Condition
      Error in `time_point_parse()`:
      ! `...` must be empty.
      x Problematic argument:
      * ..1 = 1
      i Did you forget to name an argument?

# default format is correct

    Code
      format(sys_seconds(0))
    Output
      [1] "1970-01-01T00:00:00"

# empty dots are checked

    `...` must be empty.
    x Problematic argument:
    * ..1 = 123
    i Did you forget to name an argument?

# normal print method works

    Code
      x
    Output
      <time_point<sys><day>[5]>
      [1] "2019-01-01" "2019-02-01" "2019-03-01" "2019-04-01" "2019-05-01"

# can limit with `max`

    Code
      print(x, max = 2)
    Output
      <time_point<sys><day>[5]>
      [1] "2019-01-01" "2019-02-01"
      Reached `max` or `getOption('max.print')`. Omitted 3 values.

---

    Code
      print(x, max = 4)
    Output
      <time_point<sys><day>[5]>
      [1] "2019-01-01" "2019-02-01" "2019-03-01" "2019-04-01"
      Reached `max` or `getOption('max.print')`. Omitted 1 value.

---

    Code
      print(x, max = 5)
    Output
      <time_point<sys><day>[5]>
      [1] "2019-01-01" "2019-02-01" "2019-03-01" "2019-04-01" "2019-05-01"

---

    Code
      print(x, max = 6)
    Output
      <time_point<sys><day>[5]>
      [1] "2019-01-01" "2019-02-01" "2019-03-01" "2019-04-01" "2019-05-01"

# `max` defaults to `getOption('max.print')` but can be overridden

    Code
      x
    Output
      <time_point<naive><day>[5]>
      [1] "2019-01-01" "2019-02-01" "2019-03-01"
      Reached `max` or `getOption('max.print')`. Omitted 2 values.

---

    Code
      print(x, max = 4)
    Output
      <time_point<naive><day>[5]>
      [1] "2019-01-01" "2019-02-01" "2019-03-01" "2019-04-01"
      Reached `max` or `getOption('max.print')`. Omitted 1 value.

---

    Code
      print(x, max = 5)
    Output
      <time_point<naive><day>[5]>
      [1] "2019-01-01" "2019-02-01" "2019-03-01" "2019-04-01" "2019-05-01"

# cannot floor to more precise precision

    Can't floor to a more precise precision.

# rounding with `origin` requires same clock

    Can't convert `origin` <time_point<sys><day>> to <time_point<naive><day>>.

# `origin` can be cast to a more precise `precision`, but not to a less precise one

    Can't convert `origin` <time_point<naive><millisecond>> to <time_point<naive><hour>>.
    Can't cast to a less precise precision.

# `origin` must be size 1

    `origin` must have length 1.

# `origin` must not be `NA`

    `origin` must not be `NA`.

# `origin` can't be Date or POSIXt

    Can't convert `origin` <date> to <time_point<naive><day>>.

---

    Can't convert `origin` <datetime<America/New_York>> to <time_point<naive><day>>.

# `target` is recycled to size of `x`

    Can't recycle `target` (size 2) to size 1.

# `x` is validated

    `x` must be a 'clock_time_point'.

# `target` is validated

    `target` must be a 'clock_weekday'.

# `which` is validated

    `which` must be either "next" or "previous".

---

    `which` must be either "next" or "previous".

---

    `which` must be either "next" or "previous".

# `boundary` is validated

    `boundary` must be either "keep" or "advance".

---

    `boundary` must be either "keep" or "advance".

---

    `boundary` must be either "keep" or "advance".

# OOB results return a warning and NA

    Code
      out <- time_point_count_between(sys_days(0), sys_days(1000), "nanosecond")
    Condition
      Warning:
      Conversion from duration to integer is outside the range of an integer. `NA` values have been introduced, beginning at location 1.

# both inputs must be time points

    Code
      (expect_error(time_point_count_between(sys_days(1), 1)))
    Output
      <error/rlang_error>
      Error in `time_point_count_between()`:
      ! `end` must be a <clock_time_point>.
    Code
      (expect_error(time_point_count_between(1, sys_days(1))))
    Output
      <error/rlang_error>
      Error in `time_point_count_between()`:
      ! `start` must be a <clock_time_point>.

# both inputs must be compatible

    Code
      (expect_error(time_point_count_between(x, y)))
    Output
      <error/vctrs_error_incompatible_type>
      Error in `time_point_count_between()`:
      ! Can't combine `start` <time_point<sys><day>> and `end` <time_point<naive><day>>.

# `n` is validated

    Code
      (expect_error(time_point_count_between(x, x, "day", n = NA_integer_)))
    Output
      <error/rlang_error>
      Error in `time_point_count_between()`:
      ! `n` must be a single positive integer.
    Code
      (expect_error(time_point_count_between(x, x, "day", n = -1)))
    Output
      <error/rlang_error>
      Error in `time_point_count_between()`:
      ! `n` must be a single positive integer.
    Code
      (expect_error(time_point_count_between(x, x, "day", n = 1.5)))
    Output
      <error/vctrs_error_cast_lossy>
      Error in `time_point_count_between()`:
      ! Can't convert from `n` <double> to <integer> due to loss of precision.
      * Locations: 1
    Code
      (expect_error(time_point_count_between(x, x, "day", n = "x")))
    Output
      <error/vctrs_error_incompatible_type>
      Error in `time_point_count_between()`:
      ! Can't convert `n` <character> to <integer>.
    Code
      (expect_error(time_point_count_between(x, x, "day", n = c(1L, 2L))))
    Output
      <error/rlang_error>
      Error in `time_point_count_between()`:
      ! `n` must be a single positive integer.

# `precision` must be a time point precision

    Code
      (expect_error(time_point_count_between(x, x, "year")))
    Output
      <error/rlang_error>
      Error in `time_point_count_between()`:
      ! `precision` must be at least 'week' precision.

# can't mix chronological time points and calendrical durations

    Can't convert `by` <duration<year>> to <duration<second>>.
    Can't cast between calendrical durations and chronological durations.

# can't mix clocks in seq()

    Can't convert `to` <time_point<naive><second>> to match type of `from` <time_point<sys><second>>.

# `to` is always cast to `from`

    Can't convert `to` <time_point<naive><second>> to match type of `from` <time_point<naive><day>>.
    Can't cast to a less precise precision.

# duration to add to a time-point must have at least week precision (#120)

    Can't combine `x` <duration<second>> and `y` <duration<year>>.
    Can't combine calendrical durations with chronological durations.

# precision: can only be called on time points

    `x` must be a 'clock_time_point'.

# can create a default locale object

    Code
      locale
    Output
      <clock_locale>
      Decimal Mark: .
      <clock_labels>
      Weekdays: Sunday (Sun), Monday (Mon), Tuesday (Tue), Wednesday (Wed), Thursday
                (Thu), Friday (Fri), Saturday (Sat)
      Months:   January (Jan), February (Feb), March (Mar), April (Apr), May (May),
                June (Jun), July (Jul), August (Aug), September (Sep),
                October (Oct), November (Nov), December (Dec)
      AM/PM:    AM/PM

# must use a valid clock-labels object

    `labels` must be a 'clock_labels' object.

# must use a valid decimal-mark

    `decimal_mark` must be either ',' or '.'.

# can change the decimal-mark

    Code
      x
    Output
      <clock_locale>
      Decimal Mark: ,
      <clock_labels>
      Weekdays: Sunday (Sun), Monday (Mon), Tuesday (Tue), Wednesday (Wed), Thursday
                (Thu), Friday (Fri), Saturday (Sat)
      Months:   January (Jan), February (Feb), March (Mar), April (Apr), May (May),
                June (Jun), July (Jul), August (Aug), September (Sep),
                October (Oct), November (Nov), December (Dec)
      AM/PM:    AM/PM

# can't round across common precision boundary

    Can't ceiling from a chronological precision (week) to a calendrical precision (month).

---

    Can't floor from a chronological precision (second) to a calendrical precision (year).

# seq() validates from

    `from` must have size 1, not size 2.

---

    `from` can't be `NA`.

# seq() validates length.out / along.with exclusiveness

    Can only specify one of `length.out` and `along.with`.

# seq() only takes two optional args

    Must specify exactly two of:
    - `to`
    - `by`
    - Either `length.out` or `along.with`

---

    Must specify exactly two of:
    - `to`
    - `by`
    - Either `length.out` or `along.with`

# seq() requires two optional args

    Must specify exactly two of:
    - `to`
    - `by`
    - Either `length.out` or `along.with`

---

    Must specify exactly two of:
    - `to`
    - `by`
    - Either `length.out` or `along.with`

---

    Must specify exactly two of:
    - `to`
    - `by`
    - Either `length.out` or `along.with`

---

    Must specify exactly two of:
    - `to`
    - `by`
    - Either `length.out` or `along.with`

# seq() validates `to`

    `to` must have size 1, not size 2.

---

    Can't convert `to` <double> to match type of `from` <duration<year>>.

---

    Can't convert `to` <duration<day>> to match type of `from` <duration<year>>.
    Can't cast between calendrical durations and chronological durations.

---

    `to` can't be `NA`.

# seq() validates `by`

    `by` must have size 1, not size 2.

---

    `by` can't be `NA`.

---

    `by` can't be `0`.

---

    `by` can't be `0`.

---

    Can't convert `by` <character> to <integer>.

# `by` must be castable to the type of `from`

    Can't convert `by` <duration<month>> to <duration<year>>.
    Can't cast to a less precise precision.

---

    Can't convert `by` <duration<day>> to <duration<year>>.
    Can't cast between calendrical durations and chronological durations.

---

    Can't convert `by` <duration<year>> to <duration<day>>.
    Can't cast between calendrical durations and chronological durations.

# seq() validates `length.out`

    `length.out` must have size 1, not size 2.

---

    `length.out` can't be `NA`.

---

    `length.out` can't be negative.

---

    Can't convert `length.out` <character> to <integer>.

# seq() validates dots

    `...` must be empty.
    x Problematic argument:
    * ..1 = 1
    i Did you forget to name an argument?

# seq() enforces non-fractional results

    The supplied output size does not result in a non-fractional sequence between `from` and `to`.

---

    The supplied output size does not result in a non-fractional sequence between `from` and `to`.

# `to` is always cast to `from`

    Can't convert `to` <duration<year>> to match type of `from` <duration<day>>.
    Can't cast between calendrical durations and chronological durations.

---

    Can't convert `to` <duration<month>> to match type of `from` <duration<year>>.
    Can't cast to a less precise precision.

# can't add chronological and calendrical durations

    Can't combine `x` <duration<year>> and `y` <duration<second>>.
    Can't combine calendrical durations with chronological durations.

---

    Can't combine `x` <duration<second>> and `y` <duration<year>>.
    Can't combine calendrical durations with chronological durations.

# can't convert calendrical duration to time point

    Can't combine `x` <duration<year>> and <duration<day>>.
    Can't combine calendrical durations with chronological durations.

---

    Can't combine `x` <duration<year>> and <duration<day>>.
    Can't combine calendrical durations with chronological durations.

# precision: can only be called on durations

    `x` must be a 'clock_duration'.

# `<duration> / <duration>` is not allowed

    Code
      (expect_error(duration_years(1) / duration_years(2)))
    Output
      <error/vctrs_error_incompatible_op>
      Error in `vec_arith()`:
      ! <duration<year>> / <duration<year>> is not permitted
      Durations only support integer division. Did you want `%/%`?

# `<duration> %/% <duration>` results in NA for OOB values

    Code
      out <- (numerator + one) %/% denominator
    Condition
      Warning:
      Conversion to integer is outside the range of an integer. `NA` values have been introduced, beginning at location 1.

---

    Code
      out <- (-numerator - one) %/% denominator
    Condition
      Warning:
      Conversion to integer is outside the range of an integer. `NA` values have been introduced, beginning at location 1.

# `<duration> %% <numeric>` casts the numeric to integer

    Code
      (expect_error(duration_hours(5) %% 2.5))
    Output
      <error/vctrs_error_cast_lossy>
      Error in `duration_scalar_arith()`:
      ! Can't convert from `y` <double> to <integer> due to loss of precision.
      * Locations: 1

# can make custom labels

    Code
      labels
    Output
      <clock_labels>
      Weekdays: Su, Mo, Tu, We, Th, Fr, Sa
      Months:   January, February, March, April, May, June, July, August, September,
                October, November, December
      AM/PM:    A/P

# input is validated

    `month` must be a character vector of length 12.

---

    `month` must be a character vector of length 12.

---

    `month_abbrev` must be a character vector of length 12.

---

    `month_abbrev` must be a character vector of length 12.

---

    `weekday` must be a character vector of length 7.

---

    `weekday` must be a character vector of length 7.

---

    `weekday_abbrev` must be a character vector of length 7.

---

    `weekday_abbrev` must be a character vector of length 7.

---

    `am_pm` must be a character vector of length 2.

---

    `am_pm` must be a character vector of length 2.

# can lookup a language

    Code
      labels
    Output
      <clock_labels>
      Weekdays: dimanche (dim.), lundi (lun.), mardi (mar.), mercredi (mer.), jeudi
                (jeu.), vendredi (ven.), samedi (sam.)
      Months:   janvier (janv.), février (févr.), mars (mars), avril (avr.), mai (mai),
                juin (juin), juillet (juil.), août (août), septembre (sept.),
                octobre (oct.), novembre (nov.), décembre (déc.)
      AM/PM:    AM/PM

# must be a valid language code

    `language` must be a character vector of length 1.

---

    Unknown language 'foo'.

# can list all the languages

    Code
      langs
    Output
        [1] "af"  "agq" "ak"  "am"  "ar"  "as"  "asa" "ast" "az"  "bas" "be"  "bem"
       [13] "bez" "bg"  "bm"  "bn"  "bo"  "br"  "brx" "bs"  "ca"  "ccp" "ce"  "cgg"
       [25] "chr" "ckb" "cs"  "cy"  "da"  "dav" "de"  "dje" "dsb" "dua" "dyo" "dz" 
       [37] "ebu" "ee"  "el"  "en"  "eo"  "es"  "et"  "eu"  "ewo" "fa"  "ff"  "fi" 
       [49] "fil" "fo"  "fr"  "fur" "fy"  "ga"  "gd"  "gl"  "gsw" "gu"  "guz" "gv" 
       [61] "ha"  "haw" "he"  "hi"  "hr"  "hsb" "hu"  "hy"  "id"  "ig"  "ii"  "is" 
       [73] "it"  "ja"  "jgo" "jmc" "ka"  "kab" "kam" "kde" "kea" "khq" "ki"  "kk" 
       [85] "kkj" "kl"  "kln" "km"  "kn"  "ko"  "kok" "ks"  "ksb" "ksf" "ksh" "kw" 
       [97] "ky"  "lag" "lb"  "lg"  "lkt" "ln"  "lo"  "lrc" "lt"  "lu"  "luo" "luy"
      [109] "lv"  "mas" "mer" "mfe" "mg"  "mgh" "mgo" "mk"  "ml"  "mn"  "mr"  "ms" 
      [121] "mt"  "mua" "my"  "mzn" "naq" "nb"  "nd"  "nds" "ne"  "nl"  "nmg" "nn" 
      [133] "nnh" "nus" "nyn" "om"  "or"  "os"  "pa"  "pl"  "ps"  "pt"  "qu"  "rm" 
      [145] "rn"  "ro"  "rof" "ru"  "rw"  "rwk" "sah" "saq" "sbp" "se"  "seh" "ses"
      [157] "sg"  "shi" "si"  "sk"  "sl"  "smn" "sn"  "so"  "sq"  "sr"  "sv"  "sw" 
      [169] "ta"  "te"  "teo" "tg"  "th"  "ti"  "to"  "tr"  "tt"  "twq" "tzm" "ug" 
      [181] "uk"  "ur"  "uz"  "vai" "vi"  "vun" "wae" "wo"  "xog" "yav" "yi"  "yo" 
      [193] "yue" "zgh" "zh"  "zu" 

# format - can format() a weekday

    Code
      format(weekday(1:7))
    Output
      [1] "Sun" "Mon" "Tue" "Wed" "Thu" "Fri" "Sat"

# format - can use full names

    Code
      format(weekday(1:7), abbreviate = FALSE)
    Output
      [1] "Sunday"    "Monday"    "Tuesday"   "Wednesday" "Thursday"  "Friday"   
      [7] "Saturday" 

# format - can use a different locale

    Code
      format(weekday(1:7), labels = "fr", abbreviate = FALSE)
    Output
      [1] "dimanche" "lundi"    "mardi"    "mercredi" "jeudi"    "vendredi" "samedi"  

# format - `labels` is validated

    `labels` must be a 'clock_labels' object.

# format - `abbreviate` is validated

    `abbreviate` must be `TRUE` or `FALSE`.

---

    `abbreviate` must be `TRUE` or `FALSE`.

---

    `abbreviate` must be `TRUE` or `FALSE`.

# validates `x`

    `x` must be a 'clock_weekday'.

# weekday_code - `encoding` is validated

    `encoding` must be one of "western" or "iso".

---

    `encoding` must be one of "western" or "iso".

---

    `encoding` must be one of "western" or "iso".

# `x` is validated

    `x` must be a 'clock_weekday' object.

# `labels` is validated

    `labels` must be a 'clock_labels' object.

# `encoding` is validated

    `encoding` must be one of "western" or "iso".

---

    `encoding` must be one of "western" or "iso".

---

    `encoding` must be one of "western" or "iso".

# `abbreviate` is validated

    `abbreviate` must be `TRUE` or `FALSE`.

---

    `abbreviate` must be `TRUE` or `FALSE`.

---

    `abbreviate` must be `TRUE` or `FALSE`.

# can't compare or order weekdays (#153)

    Can't compare or order values of the 'clock_weekday' type, as this type does not specify a 'first' day of the week.

---

    Can't compare or order values of the 'clock_weekday' type, as this type does not specify a 'first' day of the week.

---

    Can't compare or order values of the 'clock_weekday' type, as this type does not specify a 'first' day of the week.

---

    Can't compare or order values of the 'clock_weekday' type, as this type does not specify a 'first' day of the week.

# both day and index must be specified

    Code
      year_month_weekday(2020, 1, 1)
    Condition
      Error in `year_month_weekday()`:
      ! If either `day` or `index` is specified, both must be specified.

# validates value ranges

    Code
      year_month_weekday(50000)
    Condition
      Error in `year_month_weekday()`:
      ! `year` must be between [-32767, 32767].
      i Invalid results at locations: 1.

---

    Code
      year_month_weekday(2020, 13)
    Condition
      Error in `year_month_weekday()`:
      ! `month` must be between [1, 12].
      i Invalid results at locations: 1.

---

    Code
      year_month_weekday(2020, 1, 32, 1)
    Condition
      Error in `year_month_weekday()`:
      ! `day` must be between [1, 7].
      i Invalid results at locations: 1.

---

    Code
      year_month_weekday(2020, 1, 1, 6)
    Condition
      Error in `year_month_weekday()`:
      ! `index` must be between [1, 5].
      i Invalid results at locations: 1.

---

    Code
      year_month_weekday(2020, 1, 1, 1, 24)
    Condition
      Error in `year_month_weekday()`:
      ! `hour` must be between [0, 23].
      i Invalid results at locations: 1.

---

    Code
      year_month_weekday(2020, 1, 1, 1, 1, 60)
    Condition
      Error in `year_month_weekday()`:
      ! `minute` must be between [0, 59].
      i Invalid results at locations: 1.

---

    Code
      year_month_weekday(2020, 1, 1, 1, 1, 1, 60)
    Condition
      Error in `year_month_weekday()`:
      ! `second` must be between [0, 59].
      i Invalid results at locations: 1.

---

    Code
      year_month_weekday(2020, 1, 1, 1, 1, 1, 1, 1000, subsecond_precision = "millisecond")
    Condition
      Error in `year_month_weekday()`:
      ! `subsecond` must be between [0, 999].
      i Invalid results at locations: 1.

---

    Code
      year_month_weekday(2020, 1, 1, 1, 1, 1, 1, 1e+06, subsecond_precision = "microsecond")
    Condition
      Error in `year_month_weekday()`:
      ! `subsecond` must be between [0, 999999].
      i Invalid results at locations: 1.

---

    Code
      year_month_weekday(2020, 1, 1, 1, 1, 1, 1, 1e+09, subsecond_precision = "nanosecond")
    Condition
      Error in `year_month_weekday()`:
      ! `subsecond` must be between [0, 999999999].
      i Invalid results at locations: 1.

# cannot compare / sort with day precision or finer

    Code
      x > x
    Condition
      Error in `vec_proxy_compare()`:
      ! 'year_month_weekday' types with a precision of >= 'day' cannot be trivially compared or ordered. Convert to 'year_month_day' to compare using day-of-month values.

---

    Code
      vec_order(x)
    Condition
      Error in `vec_proxy_compare()`:
      ! 'year_month_weekday' types with a precision of >= 'day' cannot be trivially compared or ordered. Convert to 'year_month_day' to compare using day-of-month values.

# full ptype is correct

    [1] "year_month_weekday<year>"

---

    [1] "year_month_weekday<day>"

---

    [1] "year_month_weekday<nanosecond>"

---

    [1] "year_month_weekday<day>"

# abbreviated ptype is correct

    [1] "ymw<year>"

---

    [1] "ymw<day>"

---

    [1] "ymw<nano>"

---

    [1] "ymw<day>"

# setters recycling works both ways

    Code
      x <- year_month_weekday(1:2)
      y <- 1:3
      set_month(x, y)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't recycle `x` (size 2) to match `value` (size 3).

# setters require integer `value`

    Code
      set_year(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_month(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_day(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_index(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_hour(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_minute(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_second(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_millisecond(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_microsecond(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_nanosecond(x, 1.5)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

# setters check `value` range

    Code
      set_year(x, 1e+05)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [-32767, 32767].
      i Invalid results at locations: 1.

---

    Code
      set_month(x, 13)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [1, 12].
      i Invalid results at locations: 1.

---

    Code
      set_day(x, 8)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [1, 7].
      i Invalid results at locations: 1.

---

    Code
      set_index(x, 6)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [1, 5].
      i Invalid results at locations: 1.

---

    Code
      set_hour(x, 24)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [0, 23].
      i Invalid results at locations: 1.

---

    Code
      set_minute(x, 60)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [0, 59].
      i Invalid results at locations: 1.

---

    Code
      set_second(x, 60)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [0, 59].
      i Invalid results at locations: 1.

---

    Code
      set_millisecond(x, -1)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [0, 999].
      i Invalid results at locations: 1.

---

    Code
      set_microsecond(x, -1)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [0, 999999].
      i Invalid results at locations: 1.

---

    Code
      set_nanosecond(x, -1)
    Condition
      Error in `set_field_year_month_weekday()`:
      ! `value` must be between [0, 999999999].
      i Invalid results at locations: 1.

# setters require minimum precision

    Code
      set_day(year_month_weekday(year = 1), 1, index = 1)
    Condition
      Error in `set_day()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at least "month".
      i `x` has a precision of "year".

---

    Code
      set_hour(year_month_weekday(year = 1, month = 2), 1)
    Condition
      Error in `set_hour()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at least "day".
      i `x` has a precision of "month".

---

    Code
      set_minute(year_month_weekday(year = 1, month = 2, day = 3, index = 1), 1)
    Condition
      Error in `set_minute()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at least "hour".
      i `x` has a precision of "day".

---

    Code
      set_second(year_month_weekday(year = 1, month = 2, day = 3, index = 1, hour = 4),
      1)
    Condition
      Error in `set_second()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at least "minute".
      i `x` has a precision of "hour".

---

    Code
      set_millisecond(year_month_weekday(year = 1, month = 2, day = 3, index = 1,
        hour = 4, minute = 5), 1)
    Condition
      Error in `set_millisecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "millisecond".
      i `x` has a precision of "minute".

---

    Code
      set_microsecond(year_month_weekday(year = 1, month = 2, day = 3, index = 1,
        hour = 4, minute = 5), 1)
    Condition
      Error in `set_microsecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "microsecond".
      i `x` has a precision of "minute".

---

    Code
      set_nanosecond(year_month_weekday(year = 1, month = 2, day = 3, index = 1,
        hour = 4, minute = 5), 1)
    Condition
      Error in `set_nanosecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "nanosecond".
      i `x` has a precision of "minute".

# setters require correct subsecond precision

    Code
      set_millisecond(year_month_weekday(year = 1, month = 2, day = 3, index = 1,
        hour = 4, minute = 5, second = 6, subsecond = 7, subsecond_precision = "microsecond"),
      1)
    Condition
      Error in `set_millisecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "millisecond".
      i `x` has a precision of "microsecond".

---

    Code
      set_millisecond(year_month_weekday(year = 1, month = 2, day = 3, index = 1,
        hour = 4, minute = 5, second = 6, subsecond = 7, subsecond_precision = "nanosecond"),
      1)
    Condition
      Error in `set_millisecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "millisecond".
      i `x` has a precision of "nanosecond".

---

    Code
      set_microsecond(year_month_weekday(year = 1, month = 2, day = 3, index = 1,
        hour = 4, minute = 5, second = 6, subsecond = 7, subsecond_precision = "millisecond"),
      1)
    Condition
      Error in `set_microsecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "microsecond".
      i `x` has a precision of "millisecond".

---

    Code
      set_microsecond(year_month_weekday(year = 1, month = 2, day = 3, index = 1,
        hour = 4, minute = 5, second = 6, subsecond = 7, subsecond_precision = "nanosecond"),
      1)
    Condition
      Error in `set_microsecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "microsecond".
      i `x` has a precision of "nanosecond".

---

    Code
      set_nanosecond(year_month_weekday(year = 1, month = 2, day = 3, index = 1,
        hour = 4, minute = 5, second = 6, subsecond = 7, subsecond_precision = "millisecond"),
      1)
    Condition
      Error in `set_nanosecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "nanosecond".
      i `x` has a precision of "millisecond".

---

    Code
      set_nanosecond(year_month_weekday(year = 1, month = 2, day = 3, index = 1,
        hour = 4, minute = 5, second = 6, subsecond = 7, subsecond_precision = "microsecond"),
      1)
    Condition
      Error in `set_nanosecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "nanosecond".
      i `x` has a precision of "microsecond".

# default formats are correct

    Code
      format(year_month_weekday(2019))
    Output
      [1] "2019"

---

    Code
      format(year_month_weekday(2019, 1))
    Output
      [1] "2019-01"

---

    Code
      format(year_month_weekday(2019, 1, 1, 2, 1))
    Output
      [1] "2019-01-Sun[2]T01"

---

    Code
      format(year_month_weekday(2019, 1, 1, 2, 1, 2, 3, 50, subsecond_precision = "microsecond"))
    Output
      [1] "2019-01-Sun[2]T01:02:03.000050"

# can't compute start with a year_month_weekday at day precision or greater

    Code
      calendar_start(year_month_weekday(2019, 2, 2, 2), "day")
    Condition
      Error in `calendar_start()`:
      ! Computing the start of a 'year_month_weekday' with a precision equal to or more precise than 'day' is undefined.

---

    Code
      calendar_start(year_month_weekday(2019, 2, 2, 2), "month")
    Condition
      Error in `calendar_start()`:
      ! Computing the start of a 'year_month_weekday' with a precision equal to or more precise than 'day' is undefined.

# can't compute end with a year_month_weekday at day precision or greater

    Code
      calendar_end(year_month_weekday(2019, 2, 2, 2), "day")
    Condition
      Error in `calendar_end()`:
      ! Computing the end of a 'year_month_weekday' with a precision equal to or more precise than 'day' is undefined.

---

    Code
      calendar_end(year_month_weekday(2019, 2, 2, 2), "month")
    Condition
      Error in `calendar_end()`:
      ! Computing the end of a 'year_month_weekday' with a precision equal to or more precise than 'day' is undefined.

# can't compare a 'year_month_weekday' with day precision!

    Code
      (expect_error(calendar_count_between(x, x, "month")))
    Output
      <error/rlang_error>
      Error in `vec_proxy_compare()`:
      ! 'year_month_weekday' types with a precision of >= 'day' cannot be trivially compared or ordered. Convert to 'year_month_day' to compare using day-of-month values.

# only granular precisions are allowed

    Code
      seq(year_month_weekday(2019, 1, 1, 1), by = 1, length.out = 2)
    Condition
      Error in `seq.clock_year_month_day()`:
      ! `from` must be 'year' or 'month' precision.

# strict mode can be activated

    Code
      invalid_resolve(year_month_weekday(2019, 1, 1, 1))
    Condition
      Error in `strict_validate_invalid()`:
      ! The global option, `clock.strict`, is currently set to `TRUE`. In this mode, `invalid` must be set and cannot be left as `NULL`.

# throws known classed error

    Code
      invalid_resolve(year_month_weekday(2019, 1, 1, 5))
    Condition
      Error in `invalid_resolve()`:
      ! Invalid date found at location 1.
      i Resolve invalid date issues by specifying the `invalid` argument.

# errors on invalid precisions

    Code
      diff(year_month_weekday(2019, 1, 1, index = 2))
    Condition
      Error in `diff()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at most "month".
      i `x` has a precision of "day".

# minimums are right

    Code
      clock_minimum(clock_empty_year_month_weekday_year)
    Output
      <year_month_weekday<year>[1]>
      [1] "-32767"
    Code
      clock_minimum(clock_empty_year_month_weekday_month)
    Output
      <year_month_weekday<month>[1]>
      [1] "-32767-01"

---

    Code
      clock_minimum(year_month_weekday(1, 1, 1, 1))
    Condition
      Error in `clock_minimum()`:
      ! Can't compute the minimum of this <year_month_weekday>, as it is undefined.
      i The most precise allowed precision is "month".
      i `x` has precision "day".

# maximums are right

    Code
      clock_maximum(clock_empty_year_month_weekday_year)
    Output
      <year_month_weekday<year>[1]>
      [1] "32767"
    Code
      clock_maximum(clock_empty_year_month_weekday_month)
    Output
      <year_month_weekday<month>[1]>
      [1] "32767-12"

---

    Code
      clock_maximum(year_month_weekday(1, 1, 1, 1))
    Condition
      Error in `clock_maximum()`:
      ! Can't compute the maximum of this <year_month_weekday>, as it is undefined.
      i The most precise allowed precision is "month".
      i `x` has precision "day".

# min() / max() / range() aren't defined at or past day precision

    Code
      min(x)
    Condition
      Error in `vec_proxy_compare()`:
      ! 'year_month_weekday' types with a precision of >= 'day' cannot be trivially compared or ordered. Convert to 'year_month_day' to compare using day-of-month values.

---

    Code
      max(x)
    Condition
      Error in `vec_proxy_compare()`:
      ! 'year_month_weekday' types with a precision of >= 'day' cannot be trivially compared or ordered. Convert to 'year_month_day' to compare using day-of-month values.

---

    Code
      range(x)
    Condition
      Error in `vec_proxy_compare()`:
      ! 'year_month_weekday' types with a precision of >= 'day' cannot be trivially compared or ordered. Convert to 'year_month_day' to compare using day-of-month values.

---

    Code
      min(x)
    Condition
      Error in `clock_maximum()`:
      ! Can't compute the maximum of this <year_month_weekday>, as it is undefined.
      i The most precise allowed precision is "month".
      i `x` has precision "day".

---

    Code
      max(x)
    Condition
      Error in `clock_minimum()`:
      ! Can't compute the minimum of this <year_month_weekday>, as it is undefined.
      i The most precise allowed precision is "month".
      i `x` has precision "day".

---

    Code
      range(x)
    Condition
      Error in `clock_maximum()`:
      ! Can't compute the maximum of this <year_month_weekday>, as it is undefined.
      i The most precise allowed precision is "month".
      i `x` has precision "day".

# add_*() respect recycling rules

    Code
      add_years(year_month_weekday(1:2), 1:3)
    Condition
      Error in `add_years()`:
      ! Can't recycle `x` (size 2) to match `n` (size 3).

# invalid dates must be resolved when converting to a Date

    Code
      as.Date(year_month_day(2019, 2, 31))
    Condition
      Error in `as_sys_time()`:
      ! Can't convert `x` to another type because some dates are invalid.
      i The following locations are invalid: 1.
      i Resolve invalid dates with `invalid_resolve()`.

# can resolve nonexistent midnight issues

    Code
      (expect_error(as_zoned_time(x, zone), class = "clock_error_nonexistent_time"))
    Output
      <error/clock_error_nonexistent_time>
      Error in `as_zoned_time()`:
      ! Nonexistent time due to daylight saving time at location 1.
      i Resolve nonexistent time issues by specifying the `nonexistent` argument.

# can resolve ambiguous midnight issues

    Code
      (expect_error(as_zoned_time(x, zone), class = "clock_error_ambiguous_time"))
    Output
      <error/clock_error_ambiguous_time>
      Error in `as_zoned_time()`:
      ! Ambiguous time due to daylight saving time at location 1.
      i Resolve ambiguous time issues by specifying the `ambiguous` argument.

# can't group by finer precisions

    Code
      date_group(x, "hour")
    Condition
      Error in `calendar_group()`:
      ! Can't group at a precision ("hour") that is more precise than `x` ("day").

---

    Code
      date_group(x, "nanosecond")
    Condition
      Error in `calendar_group()`:
      ! Can't group at a precision ("nanosecond") that is more precise than `x` ("day").

# can't group by non-year-month-day precisions

    Code
      date_group(x, "quarter")
    Condition
      Error in `calendar_group()`:
      ! `precision` must be a valid precision for a <year_month_day>, not "quarter".

# can only floor by week/day

    Code
      date_floor(as.Date("2019-01-01"), "hour")
    Condition
      Error in `duration_rounder()`:
      ! Can't floor to a more precise precision.

---

    Code
      date_floor(as.Date("2019-01-01"), "month")
    Condition
      Error in `time_point_rounder()`:
      ! `precision` must be one of "day", "hour", "minute", "second", "millisecond", "microsecond", or "nanosecond", not "month".
      i Did you mean "minute"?

# `origin` is validated

    Code
      date_floor(x, "day", origin = 1)
    Condition
      Error in `date_floor()`:
      ! `origin` must be a <Date>, not the number 1.

---

    Code
      date_floor(x, "day", origin = new_date(NA_real_))
    Condition
      Error in `date_floor()`:
      ! `origin` can't contain missing values.
      i The following locations are missing: 1.

---

    Code
      date_floor(x, "day", origin = new_date(Inf))
    Condition
      Error in `date_floor()`:
      ! `origin` can't be an infinite date.

---

    Code
      date_floor(x, "day", origin = new_date(c(0, 1)))
    Condition
      Error in `date_floor()`:
      ! `origin` must have size 1, not size 2.

# can format dates

    Code
      vapply(X = formats, FUN = function(format) date_format(x, format = format),
      FUN.VALUE = character(1))
    Output
                              C: %C                         y: %y 
                            "C: 20"                       "y: 18" 
                              Y: %Y                         b: %b 
                          "Y: 2018"                      "b: Dec" 
                              h: %h                         B: %B 
                           "h: Dec"                 "B: December" 
                              m: %m                         d: %d 
                            "m: 12"                       "d: 31" 
                              a: %a                         A: %A 
                           "a: Mon"                   "A: Monday" 
                              w: %w                         g: %g 
                             "w: 1"                       "g: 19" 
                              G: %G                         V: %V 
                          "G: 2019"                       "V: 01" 
                              u: %u                         U: %U 
                             "u: 1"                       "U: 52" 
                              W: %W                         j: %j 
                            "W: 53"                      "j: 365" 
                              D: %D                         x: %x 
                      "D: 12/31/18"                 "x: 12/31/18" 
                              F: %F                         H: %H 
                    "F: 2018-12-31"                       "H: 00" 
                              I: %I                         M: %M 
                            "I: 12"                       "M: 00" 
                              S: %S                         p: %p 
                            "S: 00"                       "p: AM" 
                              R: %R                         T: %T 
                         "R: 00:00"                 "T: 00:00:00" 
                              X: %X                         r: %r 
                      "X: 00:00:00"              "r: 12:00:00 AM" 
                              c: %c                         %: %% 
      "c: Mon Dec 31 00:00:00 2018"                        "%: %" 

---

    Code
      vapply(X = formats, FUN = function(format) {
        date_format(x, format = format, locale = clock_locale("fr"))
      }, FUN.VALUE = character(1))
    Output
                                C: %C                           y: %y 
                              "C: 20"                         "y: 18" 
                                Y: %Y                           b: %b 
                            "Y: 2018"                       "b: déc." 
                                h: %h                           B: %B 
                            "h: déc."                   "B: décembre" 
                                m: %m                           d: %d 
                              "m: 12"                         "d: 31" 
                                a: %a                           A: %A 
                            "a: lun."                      "A: lundi" 
                                w: %w                           g: %g 
                               "w: 1"                         "g: 19" 
                                G: %G                           V: %V 
                            "G: 2019"                         "V: 01" 
                                u: %u                           U: %U 
                               "u: 1"                         "U: 52" 
                                W: %W                           j: %j 
                              "W: 53"                        "j: 365" 
                                D: %D                           x: %x 
                        "D: 12/31/18"                   "x: 12/31/18" 
                                F: %F                           H: %H 
                      "F: 2018-12-31"                         "H: 00" 
                                I: %I                           M: %M 
                              "I: 12"                         "M: 00" 
                                S: %S                           p: %p 
                              "S: 00"                         "p: AM" 
                                R: %R                           T: %T 
                           "R: 00:00"                   "T: 00:00:00" 
                                X: %X                           r: %r 
                        "X: 00:00:00"                "r: 12:00:00 AM" 
                                c: %c                           %: %% 
      "c: lun. déc. 31 00:00:00 2018"                          "%: %" 

# formatting Dates with `%z` or `%Z` returns NA with a warning

    Code
      date_format(x, format = "%z")
    Condition
      Warning:
      Failed to format 1 string at location 1. Returning `NA` at that location.
    Output
      [1] NA

---

    Code
      date_format(x, format = "%Z")
    Condition
      Warning:
      Failed to format 1 string at location 1. Returning `NA` at that location.
    Output
      [1] NA

# failure to parse throws a warning

    Code
      date_parse("foo")
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      [1] NA

# can handle invalid dates

    Code
      date_build(2019, 1:12, 31)
    Condition
      Error in `invalid_resolve()`:
      ! Invalid date found at location 2.
      i Resolve invalid date issues by specifying the `invalid` argument.

# start: can't use invalid precisions

    Code
      date_start(date_build(2019), "quarter")
    Condition
      Error in `calendar_start_end_checks()`:
      ! `precision` must be a valid precision for a <year_month_day>, not "quarter".

# end: can't use invalid precisions

    Code
      date_end(date_build(2019), "quarter")
    Condition
      Error in `calendar_start_end_checks()`:
      ! `precision` must be a valid precision for a <year_month_day>, not "quarter".

# can resolve invalid dates

    Code
      date_seq(from, to = to, by = duration_months(1))
    Condition
      Error in `invalid_resolve()`:
      ! Invalid date found at location 2.
      i Resolve invalid date issues by specifying the `invalid` argument.

# components of `to` more precise than `by` must match `from`

    Code
      date_seq(date_build(2019, 1, 1), to = date_build(2019, 2, 2), by = duration_months(
        1))
    Condition
      Error in `date_seq()`:
      ! All components of `from` and `to` more precise than "month" must match.
      i `from` is "2019-01-01".
      i `to` is "2019-02-02".

---

    Code
      date_seq(date_build(2019, 1, 1), to = date_build(2019, 3, 1), by = duration_years(
        1))
    Condition
      Error in `date_seq()`:
      ! All components of `from` and `to` more precise than "year" must match.
      i `from` is "2019-01-01".
      i `to` is "2019-03-01".

# validates integerish `by`

    Code
      date_seq(new_date(1), by = 1.5, total_size = 1)
    Condition
      Error in `date_seq()`:
      ! Can't convert from `by` <double> to <integer> due to loss of precision.
      * Locations: 1

# validates `total_size` early

    Code
      date_seq(new_date(1), by = 1, total_size = 1.5)
    Condition
      Error in `date_seq()`:
      ! Can't convert from `total_size` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      date_seq(new_date(1), by = 1, total_size = NA)
    Condition
      Error in `date_seq()`:
      ! `total_size` can't contain missing values.
      i The following locations are missing: 1.

---

    Code
      date_seq(new_date(1), by = 1, total_size = -1)
    Condition
      Error in `date_seq()`:
      ! `total_size` can't be negative.

# `to` and `total_size` must not generate a non-fractional sequence

    Code
      date_seq(new_date(0), to = new_date(3), total_size = 3)
    Condition
      Error:
      ! The supplied output size does not result in a non-fractional sequence between `from` and `to`.

# requires exactly two optional arguments

    Code
      date_seq(new_date(1), by = 1)
    Condition
      Error in `date_seq()`:
      ! Must specify exactly two of:
      * `to`
      * `by`
      * `total_size`

---

    Code
      date_seq(new_date(1), total_size = 1)
    Condition
      Error in `date_seq()`:
      ! Must specify exactly two of:
      * `to`
      * `by`
      * `total_size`

---

    Code
      date_seq(new_date(1), to = new_date(1))
    Condition
      Error in `date_seq()`:
      ! Must specify exactly two of:
      * `to`
      * `by`
      * `total_size`

# requires `to` to be Date

    Code
      date_seq(new_date(1), to = 1, by = 1)
    Condition
      Error in `date_seq()`:
      ! `to` must be a <Date>, not the number 1.

# requires year, month, or day precision

    Code
      date_seq(new_date(1), to = new_date(2), by = duration_nanoseconds(1))
    Condition
      Error in `date_seq()`:
      ! `by` must have a precision of "year", "quarter", "month", "week", or "day", not "nanosecond".

# checks empty dots

    Code
      date_seq(new_date(1), new_date(2))
    Condition
      Error in `date_seq()`:
      ! `...` must be empty.
      x Problematic argument:
      * ..1 = new_date(2)
      i Did you forget to name an argument?

# must use a valid Date precision

    Code
      (expect_error(date_count_between(x, x, "hour")))
    Output
      <error/rlang_error>
      Error in `date_count_between()`:
      ! `precision` must be "year", "quarter", "month", "week", or "day", not "hour".

# can't count between a Date and a POSIXt

    Code
      (expect_error(date_count_between(x, y, "year")))
    Output
      <error/rlang_error>
      Error in `date_count_between()`:
      ! `end` must be a <Date>, not a <POSIXct> object.

# <date> op <duration>

    Code
      vec_arith("+", new_date(0), duration_hours(1))
    Condition
      Error in `add_hours()`:
      ! Can't perform this operation on a <Date>.

---

    Code
      vec_arith("*", new_date(0), duration_years(1))
    Condition
      Error in `arith_date_and_duration()`:
      ! <date> * <duration<year>> is not permitted

# <duration> op <date>

    Code
      vec_arith("-", duration_years(1), new_date(0))
    Condition
      Error in `arith_duration_and_date()`:
      ! <duration<year>> - <date> is not permitted
      Can't subtract a Date from a duration.

---

    Code
      vec_arith("+", duration_hours(1), new_date(0))
    Condition
      Error in `add_hours()`:
      ! Can't perform this operation on a <Date>.

---

    Code
      vec_arith("*", duration_years(1), new_date(0))
    Condition
      Error in `arith_duration_and_date()`:
      ! <duration<year>> * <date> is not permitted

# `slide_index()` will error on calendrical arithmetic and invalid dates

    Code
      slider::slide_index(x, i, identity, .after = after)
    Condition
      Error in `invalid_resolve()`:
      ! Invalid date found at location 2.
      i Resolve invalid date issues by specifying the `invalid` argument.

# requires `subsecond_precision` as needed

    Code
      year_month_day(2019, 1, 1, 0, 0, 0, 1)
    Condition
      Error in `year_month_day()`:
      ! When `subsecond` is provided, `subsecond_precision` must also be specified.

# validates `subsecond_precision`

    Code
      year_month_day(2019, 1, 1, 0, 0, 0, 1, subsecond_precision = "second")
    Condition
      Error in `year_month_day()`:
      ! `subsecond_precision` must be one of "millisecond", "microsecond", or "nanosecond", not "second".
      i Did you mean "nanosecond"?

# validates value ranges

    Code
      year_month_day(50000)
    Condition
      Error in `year_month_day()`:
      ! `year` must be between [-32767, 32767].
      i Invalid results at locations: 1.

---

    Code
      year_month_day(2020, 13)
    Condition
      Error in `year_month_day()`:
      ! `month` must be between [1, 12].
      i Invalid results at locations: 1.

---

    Code
      year_month_day(2020, 1, 32)
    Condition
      Error in `year_month_day()`:
      ! `day` must be between [1, 31].
      i Invalid results at locations: 1.

---

    Code
      year_month_day(2020, 1, 1, 24)
    Condition
      Error in `year_month_day()`:
      ! `hour` must be between [0, 23].
      i Invalid results at locations: 1.

---

    Code
      year_month_day(2020, 1, 1, 1, 60)
    Condition
      Error in `year_month_day()`:
      ! `minute` must be between [0, 59].
      i Invalid results at locations: 1.

---

    Code
      year_month_day(2020, 1, 1, 1, 1, 60)
    Condition
      Error in `year_month_day()`:
      ! `second` must be between [0, 59].
      i Invalid results at locations: 1.

---

    Code
      year_month_day(2020, 1, 1, 1, 1, 1, 1000, subsecond_precision = "millisecond")
    Condition
      Error in `year_month_day()`:
      ! `subsecond` must be between [0, 999].
      i Invalid results at locations: 1.

---

    Code
      year_month_day(2020, 1, 1, 1, 1, 1, 1e+06, subsecond_precision = "microsecond")
    Condition
      Error in `year_month_day()`:
      ! `subsecond` must be between [0, 999999].
      i Invalid results at locations: 1.

---

    Code
      year_month_day(2020, 1, 1, 1, 1, 1, 1e+09, subsecond_precision = "nanosecond")
    Condition
      Error in `year_month_day()`:
      ! `subsecond` must be between [0, 999999999].
      i Invalid results at locations: 1.

# full ptype is correct

    [1] "year_month_day<year>"

---

    [1] "year_month_day<day>"

---

    [1] "year_month_day<nanosecond>"

---

    [1] "year_month_day<day>"

# abbreviated ptype is correct

    [1] "ymd<year>"

---

    [1] "ymd<day>"

---

    [1] "ymd<nano>"

---

    [1] "ymd<day>"

# subsecond precision getters require exact precisions

    Code
      get_millisecond(micro)
    Condition
      Error in `get_millisecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "millisecond".
      i `x` has a precision of "microsecond".

---

    Code
      get_microsecond(milli)
    Condition
      Error in `get_microsecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "microsecond".
      i `x` has a precision of "millisecond".

---

    Code
      get_nanosecond(micro)
    Condition
      Error in `get_nanosecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "nanosecond".
      i `x` has a precision of "microsecond".

# setters recycling works both ways

    Code
      x <- year_month_day(1:2)
      y <- 1:3
      set_month(x, y)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't recycle `x` (size 2) to match `value` (size 3).

# setters require integer `value`

    Code
      set_year(x, 1.5)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_month(x, 1.5)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_day(x, 1.5)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_hour(x, 1.5)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_minute(x, 1.5)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_second(x, 1.5)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_millisecond(x, 1.5)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_microsecond(x, 1.5)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

---

    Code
      set_nanosecond(x, 1.5)
    Condition
      Error in `set_field_year_month_day()`:
      ! Can't convert from `value` <double> to <integer> due to loss of precision.
      * Locations: 1

# setters check `value` range

    Code
      set_year(x, 1e+05)
    Condition
      Error in `set_field_year_month_day()`:
      ! `value` must be between [-32767, 32767].
      i Invalid results at locations: 1.

---

    Code
      set_month(x, 13)
    Condition
      Error in `set_field_year_month_day()`:
      ! `value` must be between [1, 12].
      i Invalid results at locations: 1.

---

    Code
      set_day(x, 32)
    Condition
      Error in `set_field_year_month_day()`:
      ! `value` must be between [1, 31].
      i Invalid results at locations: 1.

---

    Code
      set_hour(x, 24)
    Condition
      Error in `set_field_year_month_day()`:
      ! `value` must be between [0, 23].
      i Invalid results at locations: 1.

---

    Code
      set_minute(x, 60)
    Condition
      Error in `set_field_year_month_day()`:
      ! `value` must be between [0, 59].
      i Invalid results at locations: 1.

---

    Code
      set_second(x, 60)
    Condition
      Error in `set_field_year_month_day()`:
      ! `value` must be between [0, 59].
      i Invalid results at locations: 1.

---

    Code
      set_millisecond(x, -1)
    Condition
      Error in `set_field_year_month_day()`:
      ! `value` must be between [0, 999].
      i Invalid results at locations: 1.

---

    Code
      set_microsecond(x, -1)
    Condition
      Error in `set_field_year_month_day()`:
      ! `value` must be between [0, 999999].
      i Invalid results at locations: 1.

---

    Code
      set_nanosecond(x, -1)
    Condition
      Error in `set_field_year_month_day()`:
      ! `value` must be between [0, 999999999].
      i Invalid results at locations: 1.

# setters require minimum precision

    Code
      set_day(year_month_day(year = 1), 1)
    Condition
      Error in `set_day()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at least "month".
      i `x` has a precision of "year".

---

    Code
      set_hour(year_month_day(year = 1, month = 2), 1)
    Condition
      Error in `set_hour()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at least "day".
      i `x` has a precision of "month".

---

    Code
      set_minute(year_month_day(year = 1, month = 2, day = 3), 1)
    Condition
      Error in `set_minute()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at least "hour".
      i `x` has a precision of "day".

---

    Code
      set_second(year_month_day(year = 1, month = 2, day = 3, hour = 4), 1)
    Condition
      Error in `set_second()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at least "minute".
      i `x` has a precision of "hour".

---

    Code
      set_millisecond(year_month_day(year = 1, month = 2, day = 3, hour = 4, minute = 5),
      1)
    Condition
      Error in `set_millisecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "millisecond".
      i `x` has a precision of "minute".

---

    Code
      set_microsecond(year_month_day(year = 1, month = 2, day = 3, hour = 4, minute = 5),
      1)
    Condition
      Error in `set_microsecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "microsecond".
      i `x` has a precision of "minute".

---

    Code
      set_nanosecond(year_month_day(year = 1, month = 2, day = 3, hour = 4, minute = 5),
      1)
    Condition
      Error in `set_nanosecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "nanosecond".
      i `x` has a precision of "minute".

# setters require correct subsecond precision

    Code
      set_millisecond(year_month_day(year = 1, month = 2, day = 3, hour = 4, minute = 5,
        second = 6, subsecond = 7, subsecond_precision = "microsecond"), 1)
    Condition
      Error in `set_millisecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "millisecond".
      i `x` has a precision of "microsecond".

---

    Code
      set_millisecond(year_month_day(year = 1, month = 2, day = 3, hour = 4, minute = 5,
        second = 6, subsecond = 7, subsecond_precision = "nanosecond"), 1)
    Condition
      Error in `set_millisecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "millisecond".
      i `x` has a precision of "nanosecond".

---

    Code
      set_microsecond(year_month_day(year = 1, month = 2, day = 3, hour = 4, minute = 5,
        second = 6, subsecond = 7, subsecond_precision = "millisecond"), 1)
    Condition
      Error in `set_microsecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "microsecond".
      i `x` has a precision of "millisecond".

---

    Code
      set_microsecond(year_month_day(year = 1, month = 2, day = 3, hour = 4, minute = 5,
        second = 6, subsecond = 7, subsecond_precision = "nanosecond"), 1)
    Condition
      Error in `set_microsecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "microsecond".
      i `x` has a precision of "nanosecond".

---

    Code
      set_nanosecond(year_month_day(year = 1, month = 2, day = 3, hour = 4, minute = 5,
        second = 6, subsecond = 7, subsecond_precision = "millisecond"), 1)
    Condition
      Error in `set_nanosecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "nanosecond".
      i `x` has a precision of "millisecond".

---

    Code
      set_nanosecond(year_month_day(year = 1, month = 2, day = 3, hour = 4, minute = 5,
        second = 6, subsecond = 7, subsecond_precision = "microsecond"), 1)
    Condition
      Error in `set_nanosecond()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be "second" or "nanosecond".
      i `x` has a precision of "microsecond".

# invalid dates must be resolved when converting to another calendar

    Code
      as_year_quarter_day(year_month_day(2019, 2, 31))
    Condition
      Error in `as_sys_time()`:
      ! Can't convert `x` to another type because some dates are invalid.
      i The following locations are invalid: 1.
      i Resolve invalid dates with `invalid_resolve()`.

# invalid dates must be resolved when converting to a sys-time

    Code
      as_sys_time(year_month_day(2019, 2, 31))
    Condition
      Error in `as_sys_time()`:
      ! Can't convert `x` to another type because some dates are invalid.
      i The following locations are invalid: 1.
      i Resolve invalid dates with `invalid_resolve()`.

# invalid dates must be resolved when converting to a naive-time

    Code
      as_naive_time(year_month_day(2019, 2, 31))
    Condition
      Error in `as_sys_time()`:
      ! Can't convert `x` to another type because some dates are invalid.
      i The following locations are invalid: 1.
      i Resolve invalid dates with `invalid_resolve()`.

# default formats are correct

    Code
      format(year_month_day(2019))
    Output
      [1] "2019"

---

    Code
      format(year_month_day(2019, 1))
    Output
      [1] "2019-01"

---

    Code
      format(year_month_day(2019, 1, 1, 1))
    Output
      [1] "2019-01-01T01"

---

    Code
      format(year_month_day(2019, 1, 1, 1, 2, 3, 50, subsecond_precision = "microsecond"))
    Output
      [1] "2019-01-01T01:02:03.000050"

# failure to parse results in a warning

    Code
      year_month_day_parse("foo")
    Condition
      Warning:
      Failed to parse 1 string at location 1. Returning `NA` at that location.
    Output
      <year_month_day<day>[1]>
      [1] NA

# requires month precision

    Code
      calendar_month_factor(year_month_day(2019))
    Condition
      Error in `calendar_month_factor()`:
      ! `x` must have at least "month" precision.

# `labels` is validated

    Code
      calendar_month_factor(year_month_day(2019, 1), labels = 1)
    Condition
      Error in `calendar_month_factor()`:
      ! `labels` must be a <clock_labels>, not the number 1.

# `abbreviate` is validated

    Code
      calendar_month_factor(year_month_day(2019, 1), abbreviate = "foo")
    Condition
      Error in `calendar_month_factor()`:
      ! `abbreviate` must be `TRUE` or `FALSE`, not the string "foo".

---

    Code
      calendar_month_factor(year_month_day(2019, 1), abbreviate = 1)
    Condition
      Error in `calendar_month_factor()`:
      ! `abbreviate` must be `TRUE` or `FALSE`, not the number 1.

---

    Code
      calendar_month_factor(year_month_day(2019, 1), abbreviate = c(TRUE, FALSE))
    Condition
      Error in `calendar_month_factor()`:
      ! `abbreviate` must be `TRUE` or `FALSE`, not a logical vector.

# can't compute a unsupported count precision

    Code
      (expect_error(calendar_count_between(x, x, "day")))
    Output
      <error/rlang_error>
      Error in `calendar_count_between_standardize_precision_n()`:
      ! `precision` must be one of: 'year', 'quarter', 'month'.

# only granular precisions are allowed

    Code
      seq(year_month_day(2019, 1, 1), by = 1, length.out = 2)
    Condition
      Error in `seq()`:
      ! `from` must be 'year' or 'month' precision.

# strict mode can be activated

    Code
      invalid_resolve(year_month_day(2019, 1, 1))
    Condition
      Error in `strict_validate_invalid()`:
      ! The global option, `clock.strict`, is currently set to `TRUE`. In this mode, `invalid` must be set and cannot be left as `NULL`.

# throws known classed error

    Code
      invalid_resolve(year_month_day(2019, 2, 31))
    Condition
      Error in `invalid_resolve()`:
      ! Invalid date found at location 1.
      i Resolve invalid date issues by specifying the `invalid` argument.

# errors on invalid precisions

    Code
      diff(year_month_day(2019, 1, 2))
    Condition
      Error in `diff()`:
      ! Can't perform this operation because of the precision of `x`.
      i The precision of `x` must be at most "month".
      i `x` has a precision of "day".

# minimums are right

    Code
      clock_minimum(clock_empty_year_month_day_year)
    Output
      <year_month_day<year>[1]>
      [1] "-32767"
    Code
      clock_minimum(clock_empty_year_month_day_month)
    Output
      <year_month_day<month>[1]>
      [1] "-32767-01"
    Code
      clock_minimum(clock_empty_year_month_day_day)
    Output
      <year_month_day<day>[1]>
      [1] "-32767-01-01"
    Code
      clock_minimum(clock_empty_year_month_day_hour)
    Output
      <year_month_day<hour>[1]>
      [1] "-32767-01-01T00"
    Code
      clock_minimum(clock_empty_year_month_day_minute)
    Output
      <year_month_day<minute>[1]>
      [1] "-32767-01-01T00:00"
    Code
      clock_minimum(clock_empty_year_month_day_second)
    Output
      <year_month_day<second>[1]>
      [1] "-32767-01-01T00:00:00"
    Code
      clock_minimum(clock_empty_year_month_day_millisecond)
    Output
      <year_month_day<millisecond>[1]>
      [1] "-32767-01-01T00:00:00.000"
    Code
      clock_minimum(clock_empty_year_month_day_microsecond)
    Output
      <year_month_day<microsecond>[1]>
      [1] "-32767-01-01T00:00:00.000000"
    Code
      clock_minimum(clock_empty_year_month_day_nanosecond)
    Output
      <year_month_day<nanosecond>[1]>
      [1] "-32767-01-01T00:00:00.000000000"

# maximums are right

    Code
      clock_maximum(clock_empty_year_month_day_year)
    Output
      <year_month_day<year>[1]>
      [1] "32767"
    Code
      clock_maximum(clock_empty_year_month_day_month)
    Output
      <year_month_day<month>[1]>
      [1] "32767-12"
    Code
      clock_maximum(clock_empty_year_month_day_day)
    Output
      <year_month_day<day>[1]>
      [1] "32767-12-31"
    Code
      clock_maximum(clock_empty_year_month_day_hour)
    Output
      <year_month_day<hour>[1]>
      [1] "32767-12-31T23"
    Code
      clock_maximum(clock_empty_year_month_day_minute)
    Output
      <year_month_day<minute>[1]>
      [1] "32767-12-31T23:59"
    Code
      clock_maximum(clock_empty_year_month_day_second)
    Output
      <year_month_day<second>[1]>
      [1] "32767-12-31T23:59:59"
    Code
      clock_maximum(clock_empty_year_month_day_millisecond)
    Output
      <year_month_day<millisecond>[1]>
      [1] "32767-12-31T23:59:59.999"
    Code
      clock_maximum(clock_empty_year_month_day_microsecond)
    Output
      <year_month_day<microsecond>[1]>
      [1] "32767-12-31T23:59:59.999999"
    Code
      clock_maximum(clock_empty_year_month_day_nanosecond)
    Output
      <year_month_day<nanosecond>[1]>
      [1] "32767-12-31T23:59:59.999999999"

# add_*() respect recycling rules

    Code
      add_years(year_month_day(1:2), 1:3)
    Condition
      Error in `add_years()`:
      ! Can't recycle `x` (size 2) to match `n` (size 3).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-point.R
\name{time_point_count_between}
\alias{time_point_count_between}
\title{Counting: time point}
\usage{
time_point_count_between(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[clock_time_point]}

A pair of time points. These will be recycled to their common size.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"week"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
\code{time_point_count_between()} counts the number of \code{precision} units
between \code{start} and \code{end} (i.e., the number of days or hours). This count
corresponds to the \emph{whole number} of units, and will never return a
fractional value.

This is suitable for, say, computing the whole number of days between two
time points, accounting for the time of day.
}
\details{
Remember that \code{time_point_count_between()} returns an integer vector.
With extremely fine precisions, such as nanoseconds, the count can quickly
exceed the maximum value that is allowed in an integer. In this case, an
\code{NA} will be returned with a warning.
}
\section{Comparison Direction}{

The computed count has the property that if \code{start <= end}, then
\verb{start + <count> <= end}. Similarly, if \code{start >= end}, then
\verb{start + <count> >= end}. In other words, the comparison direction between
\code{start} and \code{end} will never change after adding the count to \code{start}. This
makes this function useful for repeated count computations at
increasingly fine precisions.
}

\examples{
x <- as_naive_time(year_month_day(2019, 2, 3))
y <- as_naive_time(year_month_day(2019, 2, 10))

# Whole number of days or hours between two time points
time_point_count_between(x, y, "day")
time_point_count_between(x, y, "hour")

# Whole number of 2-day units
time_point_count_between(x, y, "day", n = 2)

# Leap years are taken into account
x <- as_naive_time(year_month_day(c(2020, 2021), 2, 28))
y <- as_naive_time(year_month_day(c(2020, 2021), 3, 01))
time_point_count_between(x, y, "day")

# Time of day is taken into account.
# `2020-02-02T04 -> 2020-02-03T03` is not a whole day (because of the hour)
# `2020-02-02T04 -> 2020-02-03T05` is a whole day
x <- as_naive_time(year_month_day(2020, 2, 2, 4))
y <- as_naive_time(year_month_day(2020, 2, 3, c(3, 5)))
time_point_count_between(x, y, "day")
time_point_count_between(x, y, "hour")

# Can compute negative counts (using the same example from above)
time_point_count_between(y, x, "day")
time_point_count_between(y, x, "hour")

# Repeated computation at increasingly fine precisions
x <- as_naive_time(year_month_day(
  2020, 2, 2, 4, 5, 6, 200,
  subsecond_precision = "microsecond"
))
y <- as_naive_time(year_month_day(
  2020, 3, 1, 8, 9, 10, 100,
  subsecond_precision = "microsecond"
))

days <- time_point_count_between(x, y, "day")
x <- x + duration_days(days)

hours <- time_point_count_between(x, y, "hour")
x <- x + duration_hours(hours)

minutes <- time_point_count_between(x, y, "minute")
x <- x + duration_minutes(minutes)

seconds <- time_point_count_between(x, y, "second")
x <- x + duration_seconds(seconds)

microseconds <- time_point_count_between(x, y, "microsecond")
x <- x + duration_microseconds(microseconds)

data.frame(
  days = days,
  hours = hours,
  minutes = minutes,
  seconds = seconds,
  microseconds = microseconds
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sys-time.R
\name{as-zoned-time-sys-time}
\alias{as-zoned-time-sys-time}
\alias{as_zoned_time.clock_sys_time}
\title{Convert to a zoned-time from a sys-time}
\usage{
\method{as_zoned_time}{clock_sys_time}(x, zone, ...)
}
\arguments{
\item{x}{\verb{[clock_sys_time]}

A sys-time to convert to a zoned-time.}

\item{zone}{\verb{[character(1)]}

The zone to convert to.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A zoned-time vector.
}
\description{
This is a sys-time method for the \code{\link[=as_zoned_time]{as_zoned_time()}} generic.

Converting to a zoned-time from a sys-time retains the underlying duration,
but changes the printed time, depending on the \code{zone} that you choose.
Remember that sys-times are interpreted as UTC.

If you want to retain the printed time, try converting to a zoned-time
\link[=as-zoned-time-naive-time]{from a naive-time}, which is a time point
with a yet-to-be-determined time zone.
}
\examples{
x <- as_sys_time(year_month_day(2019, 02, 01, 02, 30, 00))
x

# Since sys-time is interpreted as UTC, converting to a zoned-time with
# a zone of UTC retains the printed time
x_utc <- as_zoned_time(x, "UTC")
x_utc

# Converting to a different zone results in a different printed time,
# which corresponds to the exact same point in time, just in a different
# part of the work
x_ny <- as_zoned_time(x, "America/New_York")
x_ny
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date-and-date-time-boundary}
\alias{date-and-date-time-boundary}
\alias{date_start}
\alias{date_end}
\title{Boundaries: date and date-time}
\usage{
date_start(x, precision, ...)

date_end(x, precision, ...)
}
\arguments{
\item{x}{\verb{[Date / POSIXct / POSIXlt]}

A date or date-time vector.}

\item{precision}{\verb{[character(1)]}

A precision. Allowed precisions are dependent on the input used.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} but with some components altered to be at the boundary value.
}
\description{
\itemize{
\item \code{date_start()} computes the date at the start of a particular
\code{precision}, such as the "start of the year".
\item \code{date_end()} computes the date at the end of a particular
\code{precision}, such as the "end of the month".
}

There are separate help pages for computing boundaries for dates and
date-times:
\itemize{
\item \link[=date-boundary]{dates (Date)}
\item \link[=posixt-boundary]{date-times (POSIXct/POSIXlt)}
}
}
\examples{
# See type specific documentation for more examples

x <- date_build(2019, 2:4)

date_end(x, "month")

x <- date_time_build(2019, 2:4, 3:5, 4, 5, zone = "America/New_York")

# Note that the hour, minute, and second components are also adjusted
date_end(x, "month")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date_seq}
\alias{date_seq}
\title{Sequences: date and date-time}
\usage{
date_seq(from, ..., to = NULL, by = NULL, total_size = NULL)
}
\arguments{
\item{from}{\verb{[Date(1) / POSIXct(1) / POSIXlt(1)]}

A date or date-time to start the sequence from.}

\item{...}{These dots are for future extensions and must be empty.}

\item{to}{\verb{[Date(1) / POSIXct(1) / POSIXlt(1) / NULL]}

A date or date-time to stop the sequence at.

\code{to} is only included in the result if the resulting sequence divides
the distance between \code{from} and \code{to} exactly.}

\item{by}{\verb{[integer(1) / clock_duration(1) / NULL]}

The unit to increment the sequence by.}

\item{total_size}{\verb{[positive integer(1) / NULL]}

The size of the resulting sequence.

If specified alongside \code{to}, this must generate a non-fractional sequence
between \code{from} and \code{to}.}
}
\value{
A date or date-time vector.
}
\description{
\code{date_seq()} generates a date (Date) or date-time (POSIXct/POSIXlt) sequence.

There are separate help pages for generating sequences for dates and
date-times:
\itemize{
\item \link[=date-sequence]{dates (Date)}
\item \link[=posixt-sequence]{date-times (POSIXct/POSIXlt)}
}
}
\examples{
# See method specific documentation for more examples

x <- as.Date("2019-01-01")
date_seq(x, by = duration_months(2), total_size = 20)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date_format}
\alias{date_format}
\title{Formatting: date and date-time}
\usage{
date_format(x, ...)
}
\arguments{
\item{x}{\verb{[Date / POSIXct / POSIXlt]}

A date or date-time vector.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A character vector of the formatted input.
}
\description{
\code{date_format()} formats a date (Date) or date-time (POSIXct/POSIXlt) using
a \code{format} string.

There are separate help pages for formatting dates and date-times:
\itemize{
\item \link[=date-formatting]{dates (Date)}
\item \link[=posixt-formatting]{date-times (POSIXct/POSIXlt)}
}
}
\examples{
# See method specific documentation for more examples

x <- as.Date("2019-01-01")
date_format(x, format = "year: \%Y, month: \%m, day: \%d")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date-count-between}
\alias{date-count-between}
\alias{date_count_between.Date}
\title{Counting: date}
\usage{
\method{date_count_between}{Date}(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[Date]}

A pair of date vectors. These will be recycled to their common
size.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"quarter"}
\item \code{"month"}
\item \code{"week"}
\item \code{"day"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
This is a Date method for the \code{\link[=date_count_between]{date_count_between()}} generic.

\code{date_count_between()} counts the number of \code{precision} units between
\code{start} and \code{end} (i.e., the number of years or months). This count
corresponds to the \emph{whole number} of units, and will never return a
fractional value.

This is suitable for, say, computing the whole number of years or months
between two dates, accounting for the day of the month.

\emph{Calendrical based counting:}

These precisions convert to a year-month-day calendar and count while in that
type.
\itemize{
\item \code{"year"}
\item \code{"quarter"}
\item \code{"month"}
}

\emph{Time point based counting:}

These precisions convert to a time point and count while in that type.
\itemize{
\item \code{"week"}
\item \code{"day"}
}

For dates, whether a calendar or time point is used is not all that
important, but is is fairly important for date-times.
}
\details{
\code{"quarter"} is equivalent to \code{"month"} precision with \code{n} set to \code{n * 3L}.
}
\section{Comparison Direction}{

The computed count has the property that if \code{start <= end}, then
\verb{start + <count> <= end}. Similarly, if \code{start >= end}, then
\verb{start + <count> >= end}. In other words, the comparison direction between
\code{start} and \code{end} will never change after adding the count to \code{start}. This
makes this function useful for repeated count computations at
increasingly fine precisions.
}

\examples{
start <- date_parse("2000-05-05")
end <- date_parse(c("2020-05-04", "2020-05-06"))

# Age in years
date_count_between(start, end, "year")

# Number of "whole" months between these dates. i.e.
# `2000-05-05 -> 2020-04-05` is 239 months
# `2000-05-05 -> 2020-05-05` is 240 months
# Since 2020-05-04 occurs before the 5th of that month,
# it gets a count of 239
date_count_between(start, end, "month")

# Number of "whole" quarters between (same as `"month"` with `n * 3`)
date_count_between(start, end, "quarter")
date_count_between(start, end, "month", n = 3)

# Number of days between
date_count_between(start, end, "day")

# Number of full 3 day periods between these two dates
date_count_between(start, end, "day", n = 3)

# Essentially the truncated value of this
date_count_between(start, end, "day") / 3

# ---------------------------------------------------------------------------

# Breakdown into full years, months, and days between
x <- start

years <- date_count_between(x, end, "year")
x <- add_years(x, years)

months <- date_count_between(x, end, "month")
x <- add_months(x, months)

days <- date_count_between(x, end, "day")
x <- add_days(x, days)

data.frame(
  start = start,
  end = end,
  years = years,
  months = months,
  days = days
)

# Note that when breaking down a date like that, you may need to
# set `invalid` during intermediate calculations
start <- date_build(2019, c(3, 3, 4), c(30, 31, 1))
end <- date_build(2019, 5, 05)

# These are 1 month apart (plus a few days)
months <- date_count_between(start, end, "month")

# But adding that 1 month to `start` results in an invalid date
try(add_months(start, months))

# You can choose various ways to resolve this
start_previous <- add_months(start, months, invalid = "previous")
start_next <- add_months(start, months, invalid = "next")

days_previous <- date_count_between(start_previous, end, "day")
days_next <- date_count_between(start_next, end, "day")

# Resulting in slightly different day values.
# No result is "perfect". Choosing "previous" or "next" both result
# in multiple `start` dates having the same month/day breakdown values.
data.frame(
  start = start,
  end = end,
  months = months,
  days_previous = days_previous,
  days_next = days_next
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{as-zoned-time-posixt}
\alias{as-zoned-time-posixt}
\alias{as_zoned_time.POSIXt}
\title{Convert to a zoned-time from a date-time}
\usage{
\method{as_zoned_time}{POSIXt}(x, ...)
}
\arguments{
\item{x}{\verb{[POSIXct / POSIXlt]}

A date-time.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A zoned-time.
}
\description{
This is a POSIXct/POSIXlt method for the \code{\link[=as_zoned_time]{as_zoned_time()}} generic.

Converting from one of R's native date-time classes (POSIXct or POSIXlt)
will retain the time zone of that object. There is no \code{zone} argument.
}
\examples{
x <- as.POSIXct("2019-01-01", tz = "America/New_York")
as_zoned_time(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quarterly-year-quarter-day.R
\name{is_year_quarter_day}
\alias{is_year_quarter_day}
\title{Is \code{x} a year-quarter-day?}
\usage{
is_year_quarter_day(x)
}
\arguments{
\item{x}{\verb{[object]}

An object.}
}
\value{
Returns \code{TRUE} if \code{x} inherits from \code{"clock_year_quarter_day"},
otherwise returns \code{FALSE}.
}
\description{
Check if \code{x} is a year-quarter-day.
}
\examples{
is_year_quarter_day(year_quarter_day(2019))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{year-month-weekday-widen}
\alias{year-month-weekday-widen}
\alias{calendar_widen.clock_year_month_weekday}
\title{Widen: year-month-weekday}
\usage{
\method{calendar_widen}{clock_year_month_weekday}(x, precision)
}
\arguments{
\item{x}{\verb{[clock_year_month_weekday]}

A year-month-weekday vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"month"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}
}
\value{
\code{x} widened to the supplied \code{precision}.
}
\description{
This is a year-month-weekday method for the \code{\link[=calendar_widen]{calendar_widen()}} generic. It
widens a year-month-weekday vector to the specified \code{precision}.
}
\details{
Widening a month precision year-month-weekday to day precision will set
the day and the index to \code{1}. This sets the weekday components to the
first Sunday of the month.
}
\examples{
# Month precision
x <- year_month_weekday(2019, 1)
x

# Widen to day precision
# Note that this sets both the day and index to 1,
# i.e. the first Sunday of the month.
calendar_widen(x, "day")

# Or second precision
sec <- calendar_widen(x, "second")
sec
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invalid.R
\name{clock-invalid}
\alias{clock-invalid}
\alias{invalid_detect}
\alias{invalid_any}
\alias{invalid_count}
\alias{invalid_remove}
\alias{invalid_resolve}
\title{Invalid calendar dates}
\usage{
invalid_detect(x)

invalid_any(x)

invalid_count(x)

invalid_remove(x)

invalid_resolve(x, ..., invalid = NULL)
}
\arguments{
\item{x}{\verb{[calendar]}

A calendar vector.}

\item{...}{These dots are for future extensions and must be empty.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}
}
\value{
\itemize{
\item \code{invalid_detect()}: Returns a logical vector detecting invalid dates.
\item \code{invalid_any()}: Returns \code{TRUE} if any invalid dates are detected.
\item \code{invalid_count()}: Returns a single integer containing the number of
invalid dates.
\item \code{invalid_remove()}: Returns \code{x} with invalid dates removed.
\item \code{invalid_resolve()}: Returns \code{x} with invalid dates resolved using the
\code{invalid} strategy.
}
}
\description{
This family of functions is for working with \emph{invalid} calendar dates.

Invalid dates represent dates made up of valid individual components, which
taken as a whole don't represent valid calendar dates. For example, for
\code{\link[=year_month_day]{year_month_day()}} the following component ranges are valid:
\verb{year: [-32767, 32767]}, \verb{month: [1, 12]}, \verb{day: [1, 31]}.
However, the date \code{2019-02-31} doesn't exist even though it is made up
of valid components. This is an example of an invalid date.

Invalid dates are allowed in clock, provided that they are eventually
resolved by using \code{invalid_resolve()} or by manually resolving them through
arithmetic or setter functions.
}
\details{
Invalid dates must be resolved before converting them to a time point.

It is recommended to use \code{"previous"} or \code{"next"} for resolving invalid
dates, as these ensure that \emph{relative ordering} among \code{x} is maintained.
This is a often a very important property to maintain when doing time series
data analysis. See the examples for more information.
}
\examples{
# Invalid date
x <- year_month_day(2019, 04, 30:31, c(3, 2), 30, 00)
x

invalid_detect(x)

# Previous valid moment in time
x_previous <- invalid_resolve(x, invalid = "previous")
x_previous

# Previous valid day, retaining time of day
x_previous_day <- invalid_resolve(x, invalid = "previous-day")
x_previous_day

# Note that `"previous"` retains the relative ordering in `x`
x[1] < x[2]
x_previous[1] < x_previous[2]

# But `"previous-day"` here does not!
x_previous_day[1] < x_previous_day[2]

# Remove invalid dates entirely
invalid_remove(x)

y <- year_quarter_day(2019, 1, 90:92)
y

# Overflow rolls forward by the number of days between `y` and the previous
# valid date
invalid_resolve(y, invalid = "overflow")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date_month_factor}
\alias{date_month_factor}
\title{Convert a date or date-time to an ordered factor of month names}
\usage{
date_month_factor(x, ..., labels = "en", abbreviate = FALSE)
}
\arguments{
\item{x}{\verb{[Date / POSIXct / POSIXlt]}

A date or date-time vector.}

\item{...}{These dots are for future extensions and must be empty.}

\item{labels}{\verb{[clock_labels / character(1)]}

Character representations of localized weekday names, month names, and
AM/PM names. Either the language code as string (passed on to
\code{\link[=clock_labels_lookup]{clock_labels_lookup()}}), or an object created by \code{\link[=clock_labels]{clock_labels()}}.}

\item{abbreviate}{\verb{[logical(1)]}

If \code{TRUE}, the abbreviated month names from \code{labels} will be used.

If \code{FALSE}, the full month names from \code{labels} will be used.}
}
\value{
An ordered factor representing the months.
}
\description{
\code{date_month_factor()} extracts the month values from a date or date-time and
converts them to an ordered factor of month names. This can be useful in
combination with ggplot2, or for modeling.
}
\examples{
x <- add_months(as.Date("2019-01-01"), 0:11)

date_month_factor(x)
date_month_factor(x, abbreviate = TRUE)
date_month_factor(x, labels = "fr")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{date_time_build}
\alias{date_time_build}
\title{Building: date-time}
\usage{
date_time_build(
  year,
  month = 1L,
  day = 1L,
  hour = 0L,
  minute = 0L,
  second = 0L,
  ...,
  zone,
  invalid = NULL,
  nonexistent = NULL,
  ambiguous = NULL
)
}
\arguments{
\item{year}{\verb{[integer]}

The year. Values \verb{[-32767, 32767]} are generally allowed.}

\item{month}{\verb{[integer]}

The month. Values \verb{[1, 12]} are allowed.}

\item{day}{\verb{[integer / "last"]}

The day of the month. Values \verb{[1, 31]} are allowed.

If \code{"last"}, then the last day of the month is returned.}

\item{hour}{\verb{[integer]}

The hour. Values \verb{[0, 23]} are allowed.}

\item{minute}{\verb{[integer]}

The minute. Values \verb{[0, 59]} are allowed.}

\item{second}{\verb{[integer]}

The second. Values \verb{[0, 59]} are allowed.}

\item{...}{These dots are for future extensions and must be empty.}

\item{zone}{\verb{[character(1)]}

A valid time zone name.

This argument is required, and must be specified by name.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}

\item{nonexistent}{\verb{[character / NULL]}

One of the following nonexistent time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"roll-forward"}: The next valid instant in time.
\item \code{"roll-backward"}: The previous valid instant in time.
\item \code{"shift-forward"}: Shift the nonexistent time forward by the size of
the daylight saving time gap.
\item \verb{"shift-backward}: Shift the nonexistent time backward by the size of
the daylight saving time gap.
\item \code{"NA"}: Replace nonexistent times with \code{NA}.
\item \code{"error"}: Error on nonexistent times.
}

Using either \code{"roll-forward"} or \code{"roll-backward"} is generally
recommended over shifting, as these two strategies maintain the
\emph{relative ordering} between elements of the input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{nonexistent} must be supplied
and cannot be \code{NULL}. This is a convenient way to make production code
robust to nonexistent times.}

\item{ambiguous}{\verb{[character / zoned_time / POSIXct / list(2) / NULL]}

One of the following ambiguous time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"earliest"}: Of the two possible times, choose the earliest one.
\item \code{"latest"}: Of the two possible times, choose the latest one.
\item \code{"NA"}: Replace ambiguous times with \code{NA}.
\item \code{"error"}: Error on ambiguous times.
}

Alternatively, \code{ambiguous} is allowed to be a zoned_time (or POSIXct) that
is either length 1, or the same length as the input. If an ambiguous time
is encountered, the zoned_time is consulted. If the zoned_time corresponds
to a naive_time that is also ambiguous \emph{and} uses the same daylight saving
time transition point as the original ambiguous time, then the offset of
the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
resolved by consulting the zoned_time, then this method falls back to
\code{NULL}.

Finally, \code{ambiguous} is allowed to be a list of size 2, where the first
element of the list is a zoned_time (as described above), and the second
element of the list is an ambiguous time resolution strategy to use when
the ambiguous time cannot be resolved by consulting the zoned_time.
Specifying a zoned_time on its own is identical to \verb{list(<zoned_time>, NULL)}.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{ambiguous} must be supplied and
cannot be \code{NULL}. Additionally, \code{ambiguous} cannot be specified as a
zoned_time on its own, as this implies \code{NULL} for ambiguous times that the
zoned_time cannot resolve. Instead, it must be specified as a list
alongside an ambiguous time resolution strategy as described above. This is
a convenient way to make production code robust to ambiguous times.}
}
\value{
A POSIXct.
}
\description{
\code{date_time_build()} builds a POSIXct from it's individual components.

To build a POSIXct, it is required that you specify the \code{zone}.
}
\details{
Components are recycled against each other.
}
\examples{
# The zone argument is required!
# clock always requires you to be explicit about your choice of `zone`.
try(date_time_build(2020))

date_time_build(2020, zone = "America/New_York")

# Nonexistent time due to daylight saving time gap from 01:59:59 -> 03:00:00
try(date_time_build(1970, 4, 26, 1:12, 30, zone = "America/New_York"))

# Resolve with a nonexistent time resolution strategy
date_time_build(
  1970, 4, 26, 1:12, 30,
  zone = "America/New_York",
  nonexistent = "roll-forward"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoned-time.R
\name{zoned_time_precision}
\alias{zoned_time_precision}
\title{Precision: zoned-time}
\usage{
zoned_time_precision(x)
}
\arguments{
\item{x}{\verb{[clock_zoned_time]}

A zoned-time.}
}
\value{
A single string holding the precision of the zoned-time.
}
\description{
\code{zoned_time_precision()} extracts the precision from a zoned-time. It
returns the precision as a single string.
}
\examples{
zoned_time_precision(zoned_time_now("America/New_York"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{calendar-boundary}
\alias{calendar-boundary}
\alias{calendar_start}
\alias{calendar_end}
\title{Boundaries: calendars}
\usage{
calendar_start(x, precision)

calendar_end(x, precision)
}
\arguments{
\item{x}{\verb{[calendar]}

A calendar vector.}

\item{precision}{\verb{[character(1)]}

A precision. Allowed precisions are dependent on the calendar used.}
}
\value{
\code{x} at the same precision, but with some components altered to be
at the boundary value.
}
\description{
\itemize{
\item \code{calendar_start()} computes the start of a calendar at a particular
\code{precision}, such as the "start of the quarter".
\item \code{calendar_end()} computes the end of a calendar at a particular
\code{precision}, such as the "end of the month".
}

For both \code{calendar_start()} and \code{calendar_end()}, the precision of \code{x} is
always retained.

Each calendar has its own help page describing the precisions that you
can compute a boundary at:
\itemize{
\item \link[=year-month-day-boundary]{year-month-day}
\item \link[=year-month-weekday-boundary]{year-month-weekday}
\item \link[=iso-year-week-day-boundary]{iso-year-week-day}
\item \link[=year-quarter-day-boundary]{year-quarter-day}
\item \link[=year-day-boundary]{year-day}
}
}
\examples{
# Hour precision
x <- year_month_day(2019, 2:4, 5, 6)
x

# Compute the start of the month
calendar_start(x, "month")

# Or the end of the month, notice that the hour value is adjusted as well
calendar_end(x, "month")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arithmetic.R
\name{clock-arithmetic}
\alias{clock-arithmetic}
\alias{add_years}
\alias{add_quarters}
\alias{add_months}
\alias{add_weeks}
\alias{add_days}
\alias{add_hours}
\alias{add_minutes}
\alias{add_seconds}
\alias{add_milliseconds}
\alias{add_microseconds}
\alias{add_nanoseconds}
\title{Clock arithmetic}
\usage{
add_years(x, n, ...)

add_quarters(x, n, ...)

add_months(x, n, ...)

add_weeks(x, n, ...)

add_days(x, n, ...)

add_hours(x, n, ...)

add_minutes(x, n, ...)

add_seconds(x, n, ...)

add_milliseconds(x, n, ...)

add_microseconds(x, n, ...)

add_nanoseconds(x, n, ...)
}
\arguments{
\item{x}{\verb{[object]}

An object.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} after performing the arithmetic.
}
\description{
This is the landing page for all clock arithmetic functions. There are
specific sub-pages describing how arithmetic works for different calendars
and time points, which is where you should look for more information.

Calendars are efficient at arithmetic with irregular units of time, such as
month, quarters, or years.
\itemize{
\item \link[=year-month-day-arithmetic]{year-month-day}
\item \link[=year-month-weekday-arithmetic]{year-month-weekday}
\item \link[=year-quarter-day-arithmetic]{year-quarter-day}
\item \link[=iso-year-week-day-arithmetic]{iso-year-week-day}
\item \link[=year-day-arithmetic]{year-day}
}

Time points, such as naive-times and sys-times, are efficient at arithmetic
with regular, well-defined units of time, such as days, hours, seconds,
or nanoseconds.
\itemize{
\item \link[=time-point-arithmetic]{time-point}
}

Durations can use any of these arithmetic functions, and return a new
duration with a precision corresponding to the common type of the
input and the function used.
\itemize{
\item \link[=duration-arithmetic]{duration}
}

Weekdays can perform day-based circular arithmetic.
\itemize{
\item \link[=weekday-arithmetic]{weekday}
}

There are also convenience methods for doing arithmetic directly on a
native R date or date-time type:
\itemize{
\item \link[=Date-arithmetic]{dates (Date)}
\item \link[=posixt-arithmetic]{date-times (POSIXct / POSIXlt)}
}
}
\details{
Months and years are considered "irregular" because some months have more
days then others (28, 29, 30, or 31), and some years have more days than
others (365 or 366).

Days are considered "regular" because they are defined as 86,400 seconds.
}
\examples{
# See each sub-page for more specific examples
x <- year_month_day(2019, 2, 1)
add_months(x, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-day.R
\name{year-day-getters}
\alias{year-day-getters}
\alias{get_year.clock_year_day}
\alias{get_day.clock_year_day}
\alias{get_hour.clock_year_day}
\alias{get_minute.clock_year_day}
\alias{get_second.clock_year_day}
\alias{get_millisecond.clock_year_day}
\alias{get_microsecond.clock_year_day}
\alias{get_nanosecond.clock_year_day}
\title{Getters: year-day}
\usage{
\method{get_year}{clock_year_day}(x)

\method{get_day}{clock_year_day}(x)

\method{get_hour}{clock_year_day}(x)

\method{get_minute}{clock_year_day}(x)

\method{get_second}{clock_year_day}(x)

\method{get_millisecond}{clock_year_day}(x)

\method{get_microsecond}{clock_year_day}(x)

\method{get_nanosecond}{clock_year_day}(x)
}
\arguments{
\item{x}{\verb{[clock_year_day]}

A year-day to get the component from.}
}
\value{
The component.
}
\description{
These are year-day methods for the \link[=clock-getters]{getter generics}.
\itemize{
\item \code{get_year()} returns the Gregorian year.
\item \code{get_day()} returns the day of the year.
\item There are sub-daily getters for extracting more precise components.
}
}
\examples{
x <- year_day(2019, 101:105, 1, 20, 30)

get_day(x)
get_second(x)

# Cannot extract more precise components
y <- year_day(2019, 1)
try(get_hour(y))

# Cannot extract components that don't exist for this calendar
try(get_quarter(x))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clock-locale.R
\name{clock_locale}
\alias{clock_locale}
\title{Create a clock locale}
\usage{
clock_locale(labels = "en", decimal_mark = ".")
}
\arguments{
\item{labels}{\verb{[clock_labels / character(1)]}

Character representations of localized weekday names, month names, and
AM/PM names. Either the language code as string (passed on to
\code{\link[=clock_labels_lookup]{clock_labels_lookup()}}), or an object created by \code{\link[=clock_labels]{clock_labels()}}.}

\item{decimal_mark}{\verb{[character(1)]}

Symbol used for the decimal place when formatting sub-second date-times.
Either \code{","} or \code{"."}.}
}
\value{
A \code{"clock_locale"} object.
}
\description{
A clock locale contains the information required to format and parse dates.
The defaults have been chosen to match US English. A clock locale object can
be provided to \code{format()} methods or parse functions (like
\code{\link[=year_month_day_parse]{year_month_day_parse()}}) to override the defaults.
}
\examples{
clock_locale()
clock_locale(labels = "fr")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{year-month-weekday-group}
\alias{year-month-weekday-group}
\alias{calendar_group.clock_year_month_weekday}
\title{Grouping: year-month-weekday}
\usage{
\method{calendar_group}{clock_year_month_weekday}(x, precision, ..., n = 1L)
}
\arguments{
\item{x}{\verb{[clock_year_month_weekday]}

A year-month-weekday vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"month"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
\code{x} grouped at the specified \code{precision}.
}
\description{
This is a year-month-weekday method for the \code{\link[=calendar_group]{calendar_group()}} generic.

Grouping for a year-month-weekday object can be done at any precision except
for \code{"day"}, as long as \code{x} is at least as precise as \code{precision}.
}
\details{
Grouping by \code{"day"} is undefined for a year-month-weekday because there are
two day fields, the weekday and the index, and there is no clear way to
define how to group by that.
}
\examples{
x <- year_month_weekday(2019, 1:12, clock_weekdays$sunday, 1, 00, 05, 05)
x

# Group by 3 months - drops more precise components!
calendar_group(x, "month", n = 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-year-week-day.R
\name{iso-year-week-day-getters}
\alias{iso-year-week-day-getters}
\alias{get_year.clock_iso_year_week_day}
\alias{get_week.clock_iso_year_week_day}
\alias{get_day.clock_iso_year_week_day}
\alias{get_hour.clock_iso_year_week_day}
\alias{get_minute.clock_iso_year_week_day}
\alias{get_second.clock_iso_year_week_day}
\alias{get_millisecond.clock_iso_year_week_day}
\alias{get_microsecond.clock_iso_year_week_day}
\alias{get_nanosecond.clock_iso_year_week_day}
\title{Getters: iso-year-week-day}
\usage{
\method{get_year}{clock_iso_year_week_day}(x)

\method{get_week}{clock_iso_year_week_day}(x)

\method{get_day}{clock_iso_year_week_day}(x)

\method{get_hour}{clock_iso_year_week_day}(x)

\method{get_minute}{clock_iso_year_week_day}(x)

\method{get_second}{clock_iso_year_week_day}(x)

\method{get_millisecond}{clock_iso_year_week_day}(x)

\method{get_microsecond}{clock_iso_year_week_day}(x)

\method{get_nanosecond}{clock_iso_year_week_day}(x)
}
\arguments{
\item{x}{\verb{[clock_iso_year_week_day]}

A iso-year-week-day to get the component from.}
}
\value{
The component.
}
\description{
These are iso-year-week-day methods for the
\link[=clock-getters]{getter generics}.
\itemize{
\item \code{get_year()} returns the ISO year. Note that this can differ from the
Gregorian year.
\item \code{get_week()} returns the ISO week of the current ISO year.
\item \code{get_day()} returns a value between 1-7 indicating the weekday of the
current ISO week, where 1 = Monday and 7 = Sunday, in line with the
ISO standard.
\item There are sub-daily getters for extracting more precise components.
}
}
\examples{
x <- iso_year_week_day(2019, 50:52, 1:3)
x

# Get the ISO week
get_week(x)

# Gets the weekday, 1 = Monday, 7 = Sunday
get_day(x)

# Note that the ISO year can differ from the Gregorian year
iso <- iso_year_week_day(2019, 1, 1)
ymd <- as_year_month_day(iso)

get_year(iso)
get_year(ymd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{year_month_weekday}
\alias{year_month_weekday}
\title{Calendar: year-month-weekday}
\usage{
year_month_weekday(
  year,
  month = NULL,
  day = NULL,
  index = NULL,
  hour = NULL,
  minute = NULL,
  second = NULL,
  subsecond = NULL,
  ...,
  subsecond_precision = NULL
)
}
\arguments{
\item{year}{\verb{[integer]}

The year. Values \verb{[-32767, 32767]} are generally allowed.}

\item{month}{\verb{[integer / NULL]}

The month. Values \verb{[1, 12]} are allowed.}

\item{day}{\verb{[integer / NULL]}

The weekday of the month. Values \verb{[1, 7]} are allowed, where \code{1} is
Sunday and \code{7} is Saturday.}

\item{index}{\verb{[integer / "last" / NULL]}

The index specifying that \code{day} is the n-th weekday of the month.
Values \verb{[1, 5]} are allowed.

If \code{"last"}, then the last instance of \code{day} in the current month
is returned.}

\item{hour}{\verb{[integer / NULL]}

The hour. Values \verb{[0, 23]} are allowed.}

\item{minute}{\verb{[integer / NULL]}

The minute. Values \verb{[0, 59]} are allowed.}

\item{second}{\verb{[integer / NULL]}

The second. Values \verb{[0, 59]} are allowed.}

\item{subsecond}{\verb{[integer / NULL]}

The subsecond. If specified, \code{subsecond_precision} must also be specified
to determine how to interpret the \code{subsecond}.

If using milliseconds, values \verb{[0, 999]} are allowed.

If using microseconds, values \verb{[0, 999999]} are allowed.

If using nanoseconds, values \verb{[0, 999999999]} are allowed.}

\item{...}{These dots are for future extensions and must be empty.}

\item{subsecond_precision}{\verb{[character(1) / NULL]}

The precision to interpret \code{subsecond} as. One of: \code{"millisecond"},
\code{"microsecond"}, or \code{"nanosecond"}.}
}
\value{
A year-month-weekday calendar vector.
}
\description{
\code{year_month_weekday()} constructs a calendar vector from the Gregorian
year, month, weekday, and index specifying that this is the n-th weekday
of the month.
}
\details{
Fields are recycled against each other.

Fields are collected in order until the first \code{NULL} field is located. No
fields after the first \code{NULL} field are used.
}
\examples{
# All Fridays in January, 2019
# Note that there was no 5th Friday in January
x <- year_month_weekday(
  2019,
  clock_months$january,
  clock_weekdays$friday,
  1:5
)
x

invalid_detect(x)

# Resolve this invalid date by using the previous valid date
invalid_resolve(x, invalid = "previous")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quarterly-year-quarter-day.R
\name{as_year_quarter_day}
\alias{as_year_quarter_day}
\title{Convert to year-quarter-day}
\usage{
as_year_quarter_day(x, ..., start = NULL)
}
\arguments{
\item{x}{\verb{[vector]}

A vector to convert to year-quarter-day.}

\item{...}{These dots are for future extensions and must be empty.}

\item{start}{\verb{[integer(1) / NULL]}

The month to start the fiscal year in. 1 = January and 12 = December.

If \code{NULL}:
\itemize{
\item If \code{x} is a year-quarter-day, it will be returned as is.
\item Otherwise, a \code{start} of January will be used.
}}
}
\value{
A year-quarter-day vector.
}
\description{
\code{as_year_quarter_day()} converts a vector to the year-quarter-day
calendar. Time points, Dates, POSIXct, and other calendars can all be
converted to year-quarter-day.
}
\examples{
# From Date
as_year_quarter_day(as.Date("2019-01-01"))
as_year_quarter_day(as.Date("2019-01-01"), start = 3)

# From POSIXct, which assumes that the naive time is what should be converted
as_year_quarter_day(as.POSIXct("2019-01-01 02:30:30", "America/New_York"))

# From other calendars
tuesday <- 3
as_year_quarter_day(year_month_weekday(2019, 2, tuesday, 2))

# Converting between `start`s
x <- year_quarter_day(2019, 01, 01, start = 2)
x

# Default keeps the same start
as_year_quarter_day(x)

# But you can change it
as_year_quarter_day(x, start = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{duration_precision}
\alias{duration_precision}
\title{Precision: duration}
\usage{
duration_precision(x)
}
\arguments{
\item{x}{\verb{[clock_duration]}

A duration.}
}
\value{
A single string holding the precision of the duration.
}
\description{
\code{duration_precision()} extracts the precision from a duration object. It
returns the precision as a single string.
}
\examples{
duration_precision(duration_seconds(1))
duration_precision(duration_nanoseconds(2))
duration_precision(duration_quarters(1:5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{Date-getters}
\alias{Date-getters}
\alias{get_year.Date}
\alias{get_month.Date}
\alias{get_day.Date}
\title{Getters: date}
\usage{
\method{get_year}{Date}(x)

\method{get_month}{Date}(x)

\method{get_day}{Date}(x)
}
\arguments{
\item{x}{\verb{[Date]}

A Date to get the component from.}
}
\value{
The component.
}
\description{
These are Date methods for the \link[=clock-getters]{getter generics}.
\itemize{
\item \code{get_year()} returns the Gregorian year.
\item \code{get_month()} returns the month of the year.
\item \code{get_day()} returns the day of the month.
}

For more advanced component extraction, convert to the calendar type
that you are interested in.
}
\examples{
x <- as.Date("2019-01-01") + 0:5
get_day(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-year-week-day.R
\name{iso-year-week-day-narrow}
\alias{iso-year-week-day-narrow}
\alias{calendar_narrow.clock_iso_year_week_day}
\title{Narrow: iso-year-week-day}
\usage{
\method{calendar_narrow}{clock_iso_year_week_day}(x, precision)
}
\arguments{
\item{x}{\verb{[clock_iso_year_week_day]}

A iso-year-week-day vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"week"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}
}
\value{
\code{x} narrowed to the supplied \code{precision}.
}
\description{
This is a iso-year-week-day method for the \code{\link[=calendar_narrow]{calendar_narrow()}} generic. It
narrows a iso-year-week-day vector to the specified \code{precision}.
}
\examples{
# Day precision
x <- iso_year_week_day(2019, 1, 5)
x

# Narrowed to week precision
calendar_narrow(x, "week")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sys-time.R
\name{sys_time_now}
\alias{sys_time_now}
\title{What is the current sys-time?}
\usage{
sys_time_now()
}
\value{
A sys-time of the current time in UTC.
}
\description{
\code{sys_time_now()} returns the current time in UTC.
}
\details{
The time is returned with a nanosecond precision, but the actual amount
of data returned is OS dependent. Usually, information at at least the
microsecond level is returned, with some platforms returning nanosecond
information.
}
\examples{
x <- sys_time_now()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date_leap_year}
\alias{date_leap_year}
\title{Is the year a leap year?}
\usage{
date_leap_year(x)
}
\arguments{
\item{x}{\verb{[Date / POSIXct / POSIXlt]}

A date or date-time to detect leap years in.}
}
\value{
A logical vector the same size as \code{x}. Returns \code{TRUE} if in a leap
year, \code{FALSE} if not in a leap year, and \code{NA} if \code{x} is \code{NA}.
}
\description{
\code{date_leap_year()} detects if the year is a leap year.
}
\examples{
x <- as.Date("2019-01-01")
x <- add_years(x, 0:5)
date_leap_year(x)

y <- as.POSIXct("2019-01-01", "America/New_York")
y <- add_years(y, 0:5)
date_leap_year(y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{posixt-setters}
\alias{posixt-setters}
\alias{set_year.POSIXt}
\alias{set_month.POSIXt}
\alias{set_day.POSIXt}
\alias{set_hour.POSIXt}
\alias{set_minute.POSIXt}
\alias{set_second.POSIXt}
\title{Setters: date-time}
\usage{
\method{set_year}{POSIXt}(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)

\method{set_month}{POSIXt}(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)

\method{set_day}{POSIXt}(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)

\method{set_hour}{POSIXt}(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)

\method{set_minute}{POSIXt}(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)

\method{set_second}{POSIXt}(x, value, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)
}
\arguments{
\item{x}{\verb{[POSIXct / POSIXlt]}

A date-time vector.}

\item{value}{\verb{[integer / "last"]}

The value to set the component to.

For \code{set_day()}, this can also be \code{"last"} to set the day to the
last day of the month.}

\item{...}{These dots are for future extensions and must be empty.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}

\item{nonexistent}{\verb{[character / NULL]}

One of the following nonexistent time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"roll-forward"}: The next valid instant in time.
\item \code{"roll-backward"}: The previous valid instant in time.
\item \code{"shift-forward"}: Shift the nonexistent time forward by the size of
the daylight saving time gap.
\item \verb{"shift-backward}: Shift the nonexistent time backward by the size of
the daylight saving time gap.
\item \code{"NA"}: Replace nonexistent times with \code{NA}.
\item \code{"error"}: Error on nonexistent times.
}

Using either \code{"roll-forward"} or \code{"roll-backward"} is generally
recommended over shifting, as these two strategies maintain the
\emph{relative ordering} between elements of the input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{nonexistent} must be supplied
and cannot be \code{NULL}. This is a convenient way to make production code
robust to nonexistent times.}

\item{ambiguous}{\verb{[character / zoned_time / POSIXct / list(2) / NULL]}

One of the following ambiguous time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"earliest"}: Of the two possible times, choose the earliest one.
\item \code{"latest"}: Of the two possible times, choose the latest one.
\item \code{"NA"}: Replace ambiguous times with \code{NA}.
\item \code{"error"}: Error on ambiguous times.
}

Alternatively, \code{ambiguous} is allowed to be a zoned_time (or POSIXct) that
is either length 1, or the same length as the input. If an ambiguous time
is encountered, the zoned_time is consulted. If the zoned_time corresponds
to a naive_time that is also ambiguous \emph{and} uses the same daylight saving
time transition point as the original ambiguous time, then the offset of
the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
resolved by consulting the zoned_time, then this method falls back to
\code{NULL}.

Finally, \code{ambiguous} is allowed to be a list of size 2, where the first
element of the list is a zoned_time (as described above), and the second
element of the list is an ambiguous time resolution strategy to use when
the ambiguous time cannot be resolved by consulting the zoned_time.
Specifying a zoned_time on its own is identical to \verb{list(<zoned_time>, NULL)}.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{ambiguous} must be supplied and
cannot be \code{NULL}. Additionally, \code{ambiguous} cannot be specified as a
zoned_time on its own, as this implies \code{NULL} for ambiguous times that the
zoned_time cannot resolve. Instead, it must be specified as a list
alongside an ambiguous time resolution strategy as described above. This is
a convenient way to make production code robust to ambiguous times.}
}
\value{
\code{x} with the component set.
}
\description{
These are POSIXct/POSIXlt methods for the \link[=clock-setters]{setter generics}.
\itemize{
\item \code{set_year()} sets the year.
\item \code{set_month()} sets the month of the year. Valid values are in the range
of \verb{[1, 12]}.
\item \code{set_day()} sets the day of the month. Valid values are in the range
of \verb{[1, 31]}.
\item There are sub-daily setters for setting more precise components, up to
a precision of seconds.
}
}
\examples{
x <- as.POSIXct("2019-02-01", tz = "America/New_York")

# Set the day
set_day(x, 12:14)

# Set to the "last" day of the month
set_day(x, "last")

# You cannot set a date-time to an invalid date like you can with
# a year-month-day. Instead, the default strategy is to error.
try(set_day(x, 31))
set_day(as_year_month_day(x), 31)

# You can resolve these issues while setting the day by specifying
# an invalid date resolution strategy with `invalid`
set_day(x, 31, invalid = "previous")

y <- as.POSIXct("2020-03-08 01:30:00", tz = "America/New_York")

# Nonexistent and ambiguous times must be resolved immediately when
# working with R's native date-time types. An error is thrown by default.
try(set_hour(y, 2))
set_hour(y, 2, nonexistent = "roll-forward")
set_hour(y, 2, nonexistent = "roll-backward")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoned-time.R
\name{zoned-zone}
\alias{zoned-zone}
\alias{zoned_time_zone}
\alias{zoned_time_set_zone}
\title{Get or set the time zone}
\usage{
zoned_time_zone(x)

zoned_time_set_zone(x, zone)
}
\arguments{
\item{x}{\verb{[zoned_time / Date / POSIXt]}

A zoned time to get or set the time zone of.}

\item{zone}{\verb{[character(1)]}

A valid time zone to switch to.}
}
\value{
\code{zoned_time_zone()} returns a string containing the time zone.

\code{zoned_time_set_zone()} returns \code{x} with an altered time zone attribute. The
underlying instant is \emph{not} changed.
}
\description{
\code{zoned_time_zone()} gets the time zone.

\code{zoned_time_set_zone()} sets the time zone \emph{without changing the
underlying instant}. This means that the result will represent the equivalent
time in the new time zone.
}
\examples{
x <- year_month_day(2019, 1, 1)
x <- as_zoned_time(as_naive_time(x), "America/New_York")
x

zoned_time_zone(x)

# Equivalent UTC time
zoned_time_set_zone(x, "UTC")

# To force a new time zone with the same printed time,
# convert to a naive time that has no implied time zone,
# then convert back to a zoned time in the new time zone.
nt <- as_naive_time(x)
nt
as_zoned_time(nt, "UTC")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{Date-setters}
\alias{Date-setters}
\alias{set_year.Date}
\alias{set_month.Date}
\alias{set_day.Date}
\title{Setters: date}
\usage{
\method{set_year}{Date}(x, value, ..., invalid = NULL)

\method{set_month}{Date}(x, value, ..., invalid = NULL)

\method{set_day}{Date}(x, value, ..., invalid = NULL)
}
\arguments{
\item{x}{\verb{[Date]}

A Date vector.}

\item{value}{\verb{[integer / "last"]}

The value to set the component to.

For \code{set_day()}, this can also be \code{"last"} to set the day to the
last day of the month.}

\item{...}{These dots are for future extensions and must be empty.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}
}
\value{
\code{x} with the component set.
}
\description{
These are Date methods for the \link[=clock-setters]{setter generics}.
\itemize{
\item \code{set_year()} sets the year.
\item \code{set_month()} sets the month of the year. Valid values are in the range
of \verb{[1, 12]}.
\item \code{set_day()} sets the day of the month. Valid values are in the range
of \verb{[1, 31]}.
}
}
\examples{
x <- as.Date("2019-02-01")

# Set the day
set_day(x, 12:14)

# Set to the "last" day of the month
set_day(x, "last")

# You cannot set a Date to an invalid day like you can with
# a year-month-day. Instead, the default strategy is to error.
try(set_day(x, 31))
set_day(as_year_month_day(x), 31)

# You can resolve these issues while setting the day by specifying
# an invalid date resolution strategy with `invalid`
set_day(x, 31, invalid = "previous")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{as_duration}
\alias{as_duration}
\title{Convert to a duration}
\usage{
as_duration(x)
}
\arguments{
\item{x}{\verb{[object]}

An object to convert to a duration.}
}
\value{
A duration with the same precision as \code{x}.
}
\description{
You generally convert to a duration from either a sys-time or a naive-time.
The precision of the input is retained in the returned duration.

To round an existing duration to another precision, see \code{\link[=duration_floor]{duration_floor()}}.
}
\examples{
x <- as_sys_time(year_month_day(2019, 01, 01))

# The number of days since 1970-01-01 UTC
as_duration(x)

x <- x + duration_seconds(1)
x

# The number of seconds since 1970-01-01 00:00:00 UTC
as_duration(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{date_time_info}
\alias{date_time_info}
\title{Info: date-time}
\usage{
date_time_info(x)
}
\arguments{
\item{x}{\verb{[POSIXct / POSIXlt]}

A date-time.}
}
\value{
A data frame of low level information.
}
\description{
\code{date_time_info()} retrieves a set of low-level information generally not
required for most date-time manipulations. It returns a data frame with the
same columns as \code{\link[=sys_time_info]{sys_time_info()}}, but the \code{begin} and \code{end} columns are
date-times with the same time zone as \code{x}, and the \code{offset} column is an
integer rather than a second based \link[=duration_seconds]{duration} column
since this is part of the high-level API.
}
\examples{
x <- date_time_build(
  2021, 03, 14, c(01, 03), c(59, 00), c(59, 00),
  zone = "America/New_York"
)

# x[1] is in EST, x[2] is in EDT
x

info <- date_time_info(x)
info

# `end` can be used to iterate through daylight saving time transitions
date_time_info(info$end)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weekday.R
\name{weekday_code}
\alias{weekday_code}
\title{Extract underlying weekday codes}
\usage{
weekday_code(x, ..., encoding = "western")
}
\arguments{
\item{x}{\verb{[weekday]}

A weekday vector.}

\item{...}{These dots are for future extensions and must be empty.}

\item{encoding}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"western"}: Encode weekdays assuming \code{1 == Sunday} and \code{7 == Saturday}.
\item \code{"iso"}: Encode weekdays assuming \code{1 == Monday} and \code{7 == Sunday}. This
is in line with the ISO standard.
}}
}
\value{
An integer vector of codes.
}
\description{
\code{weekday_code()} extracts out the integer code for the weekday.
}
\examples{
# Here we supply a western encoding to `weekday()`
x <- weekday(1:7)
x

# We can extract out the codes using different encodings
weekday_code(x, encoding = "western")
weekday_code(x, encoding = "iso")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clock-labels.R
\name{clock_labels}
\alias{clock_labels}
\alias{clock_labels_lookup}
\alias{clock_labels_languages}
\title{Create or retrieve date related labels}
\usage{
clock_labels(
  month,
  month_abbrev = month,
  weekday,
  weekday_abbrev = weekday,
  am_pm
)

clock_labels_lookup(language)

clock_labels_languages()
}
\arguments{
\item{month, month_abbrev}{\verb{[character(12)]}

Full and abbreviated month names. Starts with January.}

\item{weekday, weekday_abbrev}{\verb{[character(7)]}

Full and abbreviated weekday names. Starts with Sunday.}

\item{am_pm}{\verb{[character(2)]}

Names used for AM and PM.}

\item{language}{\verb{[character(1)]}

A BCP 47 locale, generally constructed from a two or three
digit language code. See \code{clock_labels_languages()} for a complete list of
available locales.}
}
\value{
A \code{"clock_labels"} object.
}
\description{
When parsing and formatting dates, you often need to know how weekdays of
the week and months are represented as text. These functions allow you
to either create your own labels, or look them up from a standard set of
language specific labels. The standard list is derived from ICU
(\url{https://unicode-org.github.io/icu/}) via the stringi package.
\itemize{
\item \code{clock_labels_lookup()} looks up a set of labels from a
given language code.
\item \code{clock_labels_languages()} lists the language codes that are accepted.
\item \code{clock_labels()} lets you create your own set of labels. Use this if the
currently supported languages don't meet your needs.
}
}
\examples{
clock_labels_lookup("en")
clock_labels_lookup("ko")
clock_labels_lookup("fr")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sys-time.R
\name{as_sys_time}
\alias{as_sys_time}
\title{Convert to a sys-time}
\usage{
as_sys_time(x)
}
\arguments{
\item{x}{\verb{[object]}

An object to convert to a sys-time.}
}
\value{
A sys-time vector.
}
\description{
\code{as_sys_time()} converts \code{x} to a sys-time.

You can convert to a sys-time from any calendar type, as long as it has
at least day precision. There also must not be any invalid dates. If invalid
dates exist, they must first be resolved with \code{\link[=invalid_resolve]{invalid_resolve()}}.

Converting to a sys-time from a naive-time retains the printed time,
but adds an assumption that the time should be interpreted in the UTC time
zone.

Converting to a sys-time from a zoned-time retains the underlying duration,
but the printed time is the equivalent UTC time to whatever the zoned-time's
zone happened to be.

Converting to a sys-time from a duration just wraps the duration in a
sys-time object, adding the assumption that the time should be interpreted
in the UTC time zone. The duration must have at least day precision.

There are convenience methods for converting to a sys-time from R's
native date and date-time types. Like converting from a zoned-time, these
retain the underlying duration, but will change the printed time if the
zone was not already UTC.
}
\examples{
x <- as.Date("2019-01-01")

# Dates are assumed to be naive, so the printed time is the same whether
# we convert it to sys-time or naive-time
as_sys_time(x)
as_naive_time(x)

y <- as.POSIXct("2019-01-01 01:00:00", tz = "America/New_York")

# The sys time displays the equivalent time in UTC (5 hours ahead of
# America/New_York at this point in the year)
as_sys_time(y)

ym <- year_month_day(2019, 02)

# A minimum of day precision is required
try(as_sys_time(ym))

ymd <- set_day(ym, 10)
as_sys_time(ymd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoned-time.R
\name{zoned_time_info}
\alias{zoned_time_info}
\title{Info: zoned-time}
\usage{
zoned_time_info(x)
}
\arguments{
\item{x}{\verb{[clock_zoned_time]}

A zoned-time.}
}
\value{
A data frame of low level information.
}
\description{
\code{zoned_time_info()} retrieves a set of low-level information generally not
required for most date-time manipulations. It returns a data frame with the
same columns as \code{\link[=sys_time_info]{sys_time_info()}}, but the \code{begin} and \code{end} columns are
zoned-times with the same time zone as \code{x}.
}
\examples{
x <- year_month_day(2021, 03, 14, c(01, 03), c(59, 00), c(59, 00))
x <- as_naive_time(x)
x <- as_zoned_time(x, "America/New_York")

# x[1] is in EST, x[2] is in EDT
x

info <- zoned_time_info(x)
info

# `end` can be used to iterate through daylight saving time transitions
zoned_time_info(info$end)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{calendar_narrow}
\alias{calendar_narrow}
\title{Narrow a calendar to a less precise precision}
\usage{
calendar_narrow(x, precision)
}
\arguments{
\item{x}{\verb{[calendar]}

A calendar vector.}

\item{precision}{\verb{[character(1)]}

A precision. Allowed precisions are dependent on the calendar used.}
}
\value{
\code{x} narrowed to the supplied \code{precision}.
}
\description{
\code{calendar_narrow()} narrows \code{x} to the specified \code{precision}. It does so
by dropping components that represent a precision that is finer than
\code{precision}.

Each calendar has its own help page describing the precisions that you
can narrow to:
\itemize{
\item \link[=year-month-day-narrow]{year-month-day}
\item \link[=year-month-weekday-narrow]{year-month-weekday}
\item \link[=iso-year-week-day-narrow]{iso-year-week-day}
\item \link[=year-quarter-day-narrow]{year-quarter-day}
\item \link[=year-day-narrow]{year-day}
}
}
\details{
A subsecond precision \code{x} cannot be narrowed to another subsecond precision.
You cannot narrow from, say, \code{"nanosecond"} to \code{"millisecond"} precision.
clock operates under the philosophy that once you have set the subsecond
precision of a calendar, it is "locked in" at that precision. If you
expected this to use integer division to divide the nanoseconds by 1e6 to
get to millisecond precision, you probably want to convert to a time point
first, and use \code{\link[=time_point_floor]{time_point_floor()}}.
}
\examples{
# Hour precision
x <- year_month_day(2019, 1, 3, 4)
x

# Narrowed to day precision
calendar_narrow(x, "day")

# Or month precision
calendar_narrow(x, "month")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{calendar_month_factor}
\alias{calendar_month_factor}
\title{Convert a calendar to an ordered factor of month names}
\usage{
calendar_month_factor(x, ..., labels = "en", abbreviate = FALSE)
}
\arguments{
\item{x}{\verb{[calendar]}

A calendar vector.}

\item{...}{These dots are for future extensions and must be empty.}

\item{labels}{\verb{[clock_labels / character(1)]}

Character representations of localized weekday names, month names, and
AM/PM names. Either the language code as string (passed on to
\code{\link[=clock_labels_lookup]{clock_labels_lookup()}}), or an object created by \code{\link[=clock_labels]{clock_labels()}}.}

\item{abbreviate}{\verb{[logical(1)]}

If \code{TRUE}, the abbreviated month names from \code{labels} will be used.

If \code{FALSE}, the full month names from \code{labels} will be used.}
}
\value{
An ordered factor representing the months.
}
\description{
\code{calendar_month_factor()} extracts the month values from a calendar and
converts them to an ordered factor of month names. This can be useful in
combination with ggplot2, or for modeling.

This function is only relevant for calendar types that use a month field,
i.e. \code{\link[=year_month_day]{year_month_day()}} and \code{\link[=year_month_weekday]{year_month_weekday()}}. The calendar type must
have at least month precision.
}
\examples{
x <- year_month_day(2019, 1:12)

calendar_month_factor(x)
calendar_month_factor(x, abbreviate = TRUE)
calendar_month_factor(x, labels = "fr")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{duration-helper}
\alias{duration-helper}
\alias{duration_years}
\alias{duration_quarters}
\alias{duration_months}
\alias{duration_weeks}
\alias{duration_days}
\alias{duration_hours}
\alias{duration_minutes}
\alias{duration_seconds}
\alias{duration_milliseconds}
\alias{duration_microseconds}
\alias{duration_nanoseconds}
\title{Construct a duration}
\usage{
duration_years(n = integer())

duration_quarters(n = integer())

duration_months(n = integer())

duration_weeks(n = integer())

duration_days(n = integer())

duration_hours(n = integer())

duration_minutes(n = integer())

duration_seconds(n = integer())

duration_milliseconds(n = integer())

duration_microseconds(n = integer())

duration_nanoseconds(n = integer())
}
\arguments{
\item{n}{\verb{[integer]}

The number of units of time to use when creating the duration.}
}
\value{
A duration of the specified precision.
}
\description{
These helpers construct durations of the specified precision. Durations
represent units of time.

Durations are separated into two categories:

\strong{Calendrical}
\itemize{
\item year
\item quarter
\item month
}

\strong{Chronological}
\itemize{
\item week
\item day
\item hour
\item minute
\item second
\item millisecond
\item microsecond
\item nanosecond
}

Calendrical durations are generally used when manipulating calendar types,
like year-month-day. Chronological durations are generally used when
working with time points, like sys-time or naive-time.
}
\section{Internal Representation}{


Durations are internally represented as an integer number of "ticks" along
with a ratio describing how it converts to a number of seconds. The
following duration ratios are used in clock:
\itemize{
\item \verb{1 year == 31556952 seconds}
\item \verb{1 quarter == 7889238 seconds}
\item \verb{1 month == 2629746 seconds}
\item \verb{1 week == 604800 seconds}
\item \verb{1 day == 86400 seconds}
\item \verb{1 hour == 3600 seconds}
\item \verb{1 minute == 60 seconds}
\item \verb{1 second == 1 second}
\item \verb{1 millisecond == 1 / 1000 seconds}
\item \verb{1 microsecond == 1 / 1000000 seconds}
\item \verb{1 nanosecond == 1 / 1000000000 seconds}
}

A duration of 1 year is defined to correspond to the
average length of a proleptic Gregorian year, i.e. 365.2425 days.

A duration of 1 month is defined as exactly 1/12 of a year.

A duration of 1 quarter is defined as exactly 1/4 of a year.

A duration of 1 week is defined as exactly 7 days.

These conversions come into play when doing operations like adding or
flooring durations. Generally, you add two calendrical durations together
to get a new calendrical duration, rather than adding a calendrical and
a chronological duration together. The one exception is \code{\link[=duration_cast]{duration_cast()}},
which can cast durations to any other precision, with a potential loss of
information.
}

\examples{
duration_years(1:5)
duration_nanoseconds(1:5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{posixt-arithmetic}
\alias{posixt-arithmetic}
\alias{add_years.POSIXt}
\alias{add_quarters.POSIXt}
\alias{add_months.POSIXt}
\alias{add_weeks.POSIXt}
\alias{add_days.POSIXt}
\alias{add_hours.POSIXt}
\alias{add_minutes.POSIXt}
\alias{add_seconds.POSIXt}
\title{Arithmetic: date-time}
\usage{
\method{add_years}{POSIXt}(x, n, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)

\method{add_quarters}{POSIXt}(x, n, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)

\method{add_months}{POSIXt}(x, n, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)

\method{add_weeks}{POSIXt}(x, n, ..., nonexistent = NULL, ambiguous = x)

\method{add_days}{POSIXt}(x, n, ..., nonexistent = NULL, ambiguous = x)

\method{add_hours}{POSIXt}(x, n, ...)

\method{add_minutes}{POSIXt}(x, n, ...)

\method{add_seconds}{POSIXt}(x, n, ...)
}
\arguments{
\item{x}{\verb{[POSIXct / POSIXlt]}

A date-time vector.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}

\item{nonexistent}{\verb{[character / NULL]}

One of the following nonexistent time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"roll-forward"}: The next valid instant in time.
\item \code{"roll-backward"}: The previous valid instant in time.
\item \code{"shift-forward"}: Shift the nonexistent time forward by the size of
the daylight saving time gap.
\item \verb{"shift-backward}: Shift the nonexistent time backward by the size of
the daylight saving time gap.
\item \code{"NA"}: Replace nonexistent times with \code{NA}.
\item \code{"error"}: Error on nonexistent times.
}

Using either \code{"roll-forward"} or \code{"roll-backward"} is generally
recommended over shifting, as these two strategies maintain the
\emph{relative ordering} between elements of the input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{nonexistent} must be supplied
and cannot be \code{NULL}. This is a convenient way to make production code
robust to nonexistent times.}

\item{ambiguous}{\verb{[character / zoned_time / POSIXct / list(2) / NULL]}

One of the following ambiguous time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"earliest"}: Of the two possible times, choose the earliest one.
\item \code{"latest"}: Of the two possible times, choose the latest one.
\item \code{"NA"}: Replace ambiguous times with \code{NA}.
\item \code{"error"}: Error on ambiguous times.
}

Alternatively, \code{ambiguous} is allowed to be a zoned_time (or POSIXct) that
is either length 1, or the same length as the input. If an ambiguous time
is encountered, the zoned_time is consulted. If the zoned_time corresponds
to a naive_time that is also ambiguous \emph{and} uses the same daylight saving
time transition point as the original ambiguous time, then the offset of
the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
resolved by consulting the zoned_time, then this method falls back to
\code{NULL}.

Finally, \code{ambiguous} is allowed to be a list of size 2, where the first
element of the list is a zoned_time (as described above), and the second
element of the list is an ambiguous time resolution strategy to use when
the ambiguous time cannot be resolved by consulting the zoned_time.
Specifying a zoned_time on its own is identical to \verb{list(<zoned_time>, NULL)}.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{ambiguous} must be supplied and
cannot be \code{NULL}. Additionally, \code{ambiguous} cannot be specified as a
zoned_time on its own, as this implies \code{NULL} for ambiguous times that the
zoned_time cannot resolve. Instead, it must be specified as a list
alongside an ambiguous time resolution strategy as described above. This is
a convenient way to make production code robust to ambiguous times.}
}
\value{
\code{x} after performing the arithmetic.
}
\description{
These are POSIXct/POSIXlt methods for the
\link[=clock-arithmetic]{arithmetic generics}.

Calendrical based arithmetic:

These functions convert to a naive-time, then to a year-month-day, perform
the arithmetic, then convert back to a date-time.
\itemize{
\item \code{add_years()}
\item \code{add_quarters()}
\item \code{add_months()}
}

Naive-time based arithmetic:

These functions convert to a naive-time, perform the arithmetic, then
convert back to a date-time.
\itemize{
\item \code{add_weeks()}
\item \code{add_days()}
}

Sys-time based arithmetic:

These functions convert to a sys-time, perform the arithmetic, then
convert back to a date-time.
\itemize{
\item \code{add_hours()}
\item \code{add_minutes()}
\item \code{add_seconds()}
}
}
\details{
Adding a single quarter with \code{add_quarters()} is equivalent to adding
3 months.

\code{x} and \code{n} are recycled against each other using
\link[vctrs:theory-faq-recycling]{tidyverse recycling rules}.

Calendrical based arithmetic has the potential to generate invalid dates
(like the 31st of February), nonexistent times (due to daylight saving
time gaps), and ambiguous times (due to daylight saving time fallbacks).

Naive-time based arithmetic will never generate an invalid date, but
may generate a nonexistent or ambiguous time (i.e. you added 1 day and
landed in a daylight saving time gap).

Sys-time based arithmetic operates in the UTC time zone, which means
that it will never generate any invalid dates or nonexistent / ambiguous
times.

The conversion from POSIXct/POSIXlt to the corresponding clock type uses
a "best guess" about whether you want to do the arithmetic using a naive-time
or a sys-time. For example, when adding months, you probably want to
retain the printed time when converting to a year-month-day to perform
the arithmetic, so the conversion goes through naive-time. However,
when adding smaller units like seconds, you probably want
\verb{"2020-03-08 01:59:59" + 1 second} in the America/New_York time zone to
return \code{"2020-03-08 03:00:00"}, taking into account the fact that there
was a daylight saving time gap. This requires doing the arithmetic in
sys-time, so that is what clock converts to. If you disagree with this
heuristic for any reason, you can take control and perform the conversions
yourself. For example, you could convert the previous example to a
naive-time instead of a sys-time manually with \code{\link[=as_naive_time]{as_naive_time()}}, add
1 second giving \code{"2020-03-08 02:00:00"}, then convert back to a
POSIXct/POSIXlt, dealing with the nonexistent time that gets created by
using the \code{nonexistent} argument of \code{as.POSIXct()}.
}
\section{Relative ordering}{


For the most part, adding time based units to date-times will retain the
relative ordering of the input. For example, if \code{x[1] < x[2]} before the
\verb{add_*()} call, then it is generally also true of the result. Using
\code{invalid = "previous" / "next"} and
\code{nonexistent = "roll-forward" / "roll-backward"} ensures that this holds
when invalid and nonexistent issues are encountered.

That said, with date-times there is an edge case related to ambiguous times
where the relative ordering could change. Consider these three date-times:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{x <- c(
  date_time_build(2012, 4, 1, 2, 30, zone = "Australia/Melbourne", ambiguous = "earliest"),
  date_time_build(2012, 4, 1, 2, 00, zone = "Australia/Melbourne", ambiguous = "latest"),
  date_time_build(2012, 4, 1, 2, 30, zone = "Australia/Melbourne", ambiguous = "latest")
)
x
#> [1] "2012-04-01 02:30:00 AEDT" "2012-04-01 02:00:00 AEST"
#> [3] "2012-04-01 02:30:00 AEST"
}\if{html}{\out{</div>}}

In this case, there was a daylight saving time fallback on \code{2012-04-01}
where the clocks went from \verb{02:59:59 AEDT -> 02:00:00 AEST}. So the times
above are precisely 30 minutes apart, and they are in increasing order.

If we add sys-time based units like hours, minutes, or seconds, then the
relative ordering of these date-times will be preserved. However, arithmetic
that goes through naive-time, like adding days or months, won't preserve
the ordering here:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{add_days(x, 1)
#> [1] "2012-04-02 02:30:00 AEST" "2012-04-02 02:00:00 AEST"
#> [3] "2012-04-02 02:30:00 AEST"
add_months(x, 1)
#> [1] "2012-05-01 02:30:00 AEST" "2012-05-01 02:00:00 AEST"
#> [3] "2012-05-01 02:30:00 AEST"
}\if{html}{\out{</div>}}

Note that the 1st and 3rd values of the result are the same, and the 1st
value is no longer before the 2nd value.

Adding larger units of time in naive-time generally does make more sense
than adding it in sys-time, but it does come with this one edge case to be
aware of when working with date-times (this does not affect dates). If this
has the potential to be an issue, consider only adding sys-time based units
(hours, minutes, and seconds) which can't have these issues.
}

\examples{
x <- as.POSIXct("2019-01-01", tz = "America/New_York")

add_years(x, 1:5)

y <- as.POSIXct("2019-01-31 00:30:00", tz = "America/New_York")

# Adding 1 month to `y` generates an invalid date. Unlike year-month-day
# types, R's native date-time types cannot handle invalid dates, so you must
# resolve them immediately. If you don't you get an error:
try(add_months(y, 1:2))
add_months(as_year_month_day(y), 1:2)

# Resolve invalid dates by specifying an invalid date resolution strategy
# with the `invalid` argument. Using `"previous"` here sets the date-time to
# the previous valid moment in time - i.e. the end of the month. The
# time is set to the last moment in the day to retain the relative ordering
# within your input. If you are okay with potentially losing this, and
# want to retain your time of day, you can use `"previous-day"` to set the
# date-time to the previous valid day, while keeping the time of day.
add_months(y, 1:2, invalid = "previous")
add_months(y, 1:2, invalid = "previous-day")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoned-time.R
\name{as_zoned_time}
\alias{as_zoned_time}
\title{Convert to a zoned-time}
\usage{
as_zoned_time(x, ...)
}
\arguments{
\item{x}{\verb{[object]}

An object to convert to a zoned-time.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A zoned-time vector.
}
\description{
\code{as_zoned_time()} converts \code{x} to a zoned-time. You generally convert
to a zoned time from either a sys-time or a naive time. Each are documented
on their own page:
\itemize{
\item \link[=as-zoned-time-sys-time]{sys-time}
\item \link[=as-zoned-time-naive-time]{naive-time}
}

There are also convenience methods for converting to a zoned time from
native R date and date-time types:
\itemize{
\item \link[=as-zoned-time-Date]{dates (Date)}
\item \link[=as-zoned-time-posixt]{date-times (POSIXct / POSIXlt)}
}
}
\examples{
x <- as.Date("2019-01-01")
as_zoned_time(x, "Europe/London")

y <- as_naive_time(year_month_day(2019, 2, 1))
as_zoned_time(y, zone = "America/New_York")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-point.R
\name{time_point_cast}
\alias{time_point_cast}
\title{Cast a time point between precisions}
\usage{
time_point_cast(x, precision)
}
\arguments{
\item{x}{\verb{[clock_sys_time / clock_naive_time]}

A sys-time or naive-time.}

\item{precision}{\verb{[character(1)]}

A time point precision. One of:
\itemize{
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}
}
\value{
\code{x} cast to the new \code{precision}.
}
\description{
Casting is one way to change a time point's precision.

Casting to a less precise precision will completely drop information that
is more precise than the precision that you are casting to. It does so
in a way that makes it round towards zero. When converting time points
to a less precise precision, you often want \code{\link[=time_point_floor]{time_point_floor()}} instead
of \code{time_point_cast()}, as that handles pre-1970 dates (which are
stored as negative durations) in a more intuitive manner.

Casting to a more precise precision is done through a multiplication by
a conversion factor between the current precision and the new precision.
}
\examples{
# Hour precision time points
# One is pre-1970, one is post-1970
x <- duration_hours(c(25, -25))
x <- as_naive_time(x)
x

# Casting rounds the underlying duration towards 0
cast <- time_point_cast(x, "day")
cast

# Flooring rounds the underlying duration towards negative infinity,
# which is often more intuitive for time points.
# Note that the cast ends up rounding the pre-1970 date up to the next
# day, while the post-1970 date is rounded down.
floor <- time_point_floor(x, "day")
floor

# Casting to a more precise precision, hour->millisecond
time_point_cast(x, "millisecond")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date-group}
\alias{date-group}
\alias{date_group.Date}
\title{Group date components}
\usage{
\method{date_group}{Date}(x, precision, ..., n = 1L, invalid = NULL)
}
\arguments{
\item{x}{\verb{[Date]}

A date vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"month"}
\item \code{"day"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}
}
\value{
\code{x}, grouped at \code{precision}.
}
\description{
This is a Date method for the \code{\link[=date_group]{date_group()}} generic.

\code{date_group()} groups by a single component of a Date, such as month
of the year, or day of the month.

If you need to group by more complex components, like ISO weeks, or quarters,
convert to a calendar type that contains the component you are interested
in grouping by.
}
\examples{
x <- as.Date("2019-01-01") + -3:5
x

# Group by 2 days of the current month.
# Note that this resets at the beginning of the month, creating day groups
# of [29, 30] [31] [01, 02] [03, 04].
date_group(x, "day", n = 2)

# Group by month
date_group(x, "month")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{duration-rounding}
\alias{duration-rounding}
\alias{duration_floor}
\alias{duration_ceiling}
\alias{duration_round}
\title{Duration rounding}
\usage{
duration_floor(x, precision, ..., n = 1L)

duration_ceiling(x, precision, ..., n = 1L)

duration_round(x, precision, ..., n = 1L)
}
\arguments{
\item{x}{\verb{[clock_duration]}

A duration.}

\item{precision}{\verb{[character(1)]}

A precision. One of:
\itemize{
\item \code{"year"}
\item \code{"quarter"}
\item \code{"month"}
\item \code{"week"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A positive integer specifying the multiple of \code{precision} to use.}
}
\value{
\code{x} rounded to the \code{precision}.
}
\description{
\itemize{
\item \code{duration_floor()} rounds a duration down to a multiple of the specified
\code{precision}.
\item \code{duration_ceiling()} rounds a duration up to a multiple of the specified
\code{precision}.
\item \code{duration_round()} rounds up or down depending on what is closer,
rounding up on ties.
}
}
\details{
You can floor calendrical durations to other calendrical durations, and
chronological durations to other chronological durations, but you can't
floor a chronological duration to a calendrical duration (such as flooring
from day to month). For more information, see the documentation on the
\link[=duration-helper]{duration helper} page.
}
\examples{
x <- duration_seconds(c(86399, 86401))

duration_floor(x, "day")
duration_ceiling(x, "day")

# Can't floor from a chronological duration (seconds)
# to a calendrical duration (months)
try(duration_floor(x, "month"))

# Every 2 days, using an origin of day 0
y <- duration_seconds(c(0, 86400, 86400 * 2, 86400 * 3))
duration_floor(y, "day", n = 2)

# Shifting the origin to be day 1
origin <- duration_days(1)
duration_floor(y - origin, "day", n = 2) + origin

# Rounding will round ties up
half_day <- 86400 / 2
half_day_durations <- duration_seconds(c(half_day - 1, half_day, half_day + 1))
duration_round(half_day_durations, "day")

# With larger units
x <- duration_months(c(0, 15, 24))
duration_floor(x, "year")
duration_floor(x, "quarter")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date_count_between}
\alias{date_count_between}
\title{Counting: date and date-time}
\usage{
date_count_between(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[Date / POSIXct / POSIXlt]}

A pair of date or date-time vectors. These will be recycled to their common
size.}

\item{precision}{\verb{[character(1)]}

A precision. Allowed precisions are dependent on the calendar used.}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
\code{date_count_between()} counts the number of \code{precision} units between
\code{start} and \code{end} (i.e., the number of years or months or hours). This count
corresponds to the \emph{whole number} of units, and will never return a
fractional value.

This is suitable for, say, computing the whole number of years or months
between two dates, accounting for the day and time of day.

There are separate help pages for counting for dates and date-times:
\itemize{
\item \link[=date-count-between]{dates (Date)}
\item \link[=posixt-count-between]{date-times (POSIXct/POSIXlt)}
}
}
\section{Comparison Direction}{

The computed count has the property that if \code{start <= end}, then
\verb{start + <count> <= end}. Similarly, if \code{start >= end}, then
\verb{start + <count> >= end}. In other words, the comparison direction between
\code{start} and \code{end} will never change after adding the count to \code{start}. This
makes this function useful for repeated count computations at
increasingly fine precisions.
}

\examples{
# See method specific documentation for more examples

start <- date_parse("2000-05-05")
end <- date_parse(c("2020-05-04", "2020-05-06"))

# Age in years
date_count_between(start, end, "year")

# Number of "whole" months between these dates
date_count_between(start, end, "month")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-day.R
\name{year-day-arithmetic}
\alias{year-day-arithmetic}
\alias{add_years.clock_year_day}
\title{Arithmetic: year-day}
\usage{
\method{add_years}{clock_year_day}(x, n, ...)
}
\arguments{
\item{x}{\verb{[clock_year_day]}

A year-day vector.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} after performing the arithmetic.
}
\description{
These are year-day methods for the
\link[=clock-arithmetic]{arithmetic generics}.
\itemize{
\item \code{add_years()}
}

Notably, \emph{you cannot add days to a year-day}. For day-based arithmetic,
first convert to a time point with \code{\link[=as_naive_time]{as_naive_time()}} or \code{\link[=as_sys_time]{as_sys_time()}}.
}
\details{
\code{x} and \code{n} are recycled against each other.
}
\examples{
x <- year_day(2019, 10)

add_years(x, 1:5)

# A valid day in a leap year
y <- year_day(2020, 366)
y

# Adding 1 year to `y` generates an invalid date
y_plus <- add_years(y, 1)
y_plus

# Invalid dates are fine, as long as they are eventually resolved
# by either manually resolving, or by calling `invalid_resolve()`

# Resolve by returning the previous / next valid moment in time
invalid_resolve(y_plus, invalid = "previous")
invalid_resolve(y_plus, invalid = "next")

# Manually resolve by setting to the last day of the year
invalid <- invalid_detect(y_plus)
y_plus[invalid] <- set_day(y_plus[invalid], "last")
y_plus
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{calendar_group}
\alias{calendar_group}
\title{Group calendar components}
\usage{
calendar_group(x, precision, ..., n = 1L)
}
\arguments{
\item{x}{\verb{[calendar]}

A calendar vector.}

\item{precision}{\verb{[character(1)]}

A precision. Allowed precisions are dependent on the calendar used.}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
\code{x} grouped at the specified \code{precision}.
}
\description{
\code{calendar_group()} groups at a multiple of the specified precision. Grouping
alters the value of a single component (i.e. the month component
if grouping by month). Components that are more precise than the precision
being grouped at are dropped altogether (i.e. the day component is dropped
if grouping by month).

Each calendar has its own help page describing the grouping process in more
detail:
\itemize{
\item \link[=year-month-day-group]{year-month-day}
\item \link[=year-month-weekday-group]{year-month-weekday}
\item \link[=iso-year-week-day-group]{iso-year-week-day}
\item \link[=year-quarter-day-group]{year-quarter-day}
\item \link[=year-day-group]{year-day}
}
}
\examples{
# See the calendar specific help pages for more examples
x <- year_month_day(2019, c(1, 1, 2, 2, 3, 3, 4, 4), 1:8)
x

# Group by two months
calendar_group(x, "month", n = 2)

# Group by two days of the month
calendar_group(x, "day", n = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{year-month-weekday-arithmetic}
\alias{year-month-weekday-arithmetic}
\alias{add_years.clock_year_month_weekday}
\alias{add_quarters.clock_year_month_weekday}
\alias{add_months.clock_year_month_weekday}
\title{Arithmetic: year-month-weekday}
\usage{
\method{add_years}{clock_year_month_weekday}(x, n, ...)

\method{add_quarters}{clock_year_month_weekday}(x, n, ...)

\method{add_months}{clock_year_month_weekday}(x, n, ...)
}
\arguments{
\item{x}{\verb{[clock_year_month_weekday]}

A year-month-weekday vector.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} after performing the arithmetic.
}
\description{
These are year-month-weekday methods for the
\link[=clock-arithmetic]{arithmetic generics}.
\itemize{
\item \code{add_years()}
\item \code{add_quarters()}
\item \code{add_months()}
}

Notably, \emph{you cannot add days to a year-month-weekday}. For day-based
arithmetic, first convert to a time point with \code{\link[=as_naive_time]{as_naive_time()}} or
\code{\link[=as_sys_time]{as_sys_time()}}.
}
\details{
Adding a single quarter with \code{add_quarters()} is equivalent to adding
3 months.

\code{x} and \code{n} are recycled against each other.
}
\examples{
# 2nd Friday in January, 2019
x <- year_month_weekday(2019, 1, clock_weekdays$friday, 2)
x

add_months(x, 1:5)

# These don't necessarily correspond to the same day of the month
as_year_month_day(add_months(x, 1:5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-point.R
\name{time_point_precision}
\alias{time_point_precision}
\title{Precision: time point}
\usage{
time_point_precision(x)
}
\arguments{
\item{x}{\verb{[clock_time_point]}

A time point.}
}
\value{
A single string holding the precision of the time point.
}
\description{
\code{time_point_precision()} extracts the precision from a time point, such
as a sys-time or naive-time. It returns the precision as a single string.
}
\examples{
time_point_precision(sys_time_now())
time_point_precision(as_naive_time(duration_days(1)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-year-week-day.R
\name{iso-year-week-day-boundary}
\alias{iso-year-week-day-boundary}
\alias{calendar_start.clock_iso_year_week_day}
\alias{calendar_end.clock_iso_year_week_day}
\title{Boundaries: iso-year-week-day}
\usage{
\method{calendar_start}{clock_iso_year_week_day}(x, precision)

\method{calendar_end}{clock_iso_year_week_day}(x, precision)
}
\arguments{
\item{x}{\verb{[clock_iso_year_week_day]}

A iso-year-week-day vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"week"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}
}
\value{
\code{x} at the same precision, but with some components altered to be
at the boundary value.
}
\description{
This is an iso-year-week-day method for the \code{\link[=calendar_start]{calendar_start()}} and
\code{\link[=calendar_end]{calendar_end()}} generics. They adjust components of a calendar to the
start or end of a specified \code{precision}.
}
\examples{
x <- iso_year_week_day(2019:2020, 5, 6, 10)
x

# Compute the last moment of the last iso week of the year
calendar_end(x, "year")

# Compare that to just setting the week to `"last"`,
# which doesn't affect the other components
set_week(x, "last")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-point.R
\name{time-point-arithmetic}
\alias{time-point-arithmetic}
\alias{add_weeks.clock_time_point}
\alias{add_days.clock_time_point}
\alias{add_hours.clock_time_point}
\alias{add_minutes.clock_time_point}
\alias{add_seconds.clock_time_point}
\alias{add_milliseconds.clock_time_point}
\alias{add_microseconds.clock_time_point}
\alias{add_nanoseconds.clock_time_point}
\title{Arithmetic: Time points}
\usage{
\method{add_weeks}{clock_time_point}(x, n, ...)

\method{add_days}{clock_time_point}(x, n, ...)

\method{add_hours}{clock_time_point}(x, n, ...)

\method{add_minutes}{clock_time_point}(x, n, ...)

\method{add_seconds}{clock_time_point}(x, n, ...)

\method{add_milliseconds}{clock_time_point}(x, n, ...)

\method{add_microseconds}{clock_time_point}(x, n, ...)

\method{add_nanoseconds}{clock_time_point}(x, n, ...)
}
\arguments{
\item{x}{\verb{[clock_sys_time / clock_naive_time]}

A time point vector.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} after performing the arithmetic.
}
\description{
These are naive-time and sys-time methods for the
\link[=clock-arithmetic]{arithmetic generics}.
\itemize{
\item \code{add_weeks()}
\item \code{add_days()}
\item \code{add_hours()}
\item \code{add_minutes()}
\item \code{add_seconds()}
\item \code{add_milliseconds()}
\item \code{add_microseconds()}
\item \code{add_nanoseconds()}
}

When working with zoned times, generally you convert to either sys-time
or naive-time, add the duration, then convert back to zoned time. Typically,
\emph{weeks and days} are added in \emph{naive-time}, and \emph{hours, minutes, seconds,
and subseconds} are added in \emph{sys-time}.

If you aren't using zoned times, arithmetic on sys-times and naive-time
is equivalent.

If you need to add larger irregular units of time, such as months, quarters,
or years, convert to a calendar type with a converter like
\code{\link[=as_year_month_day]{as_year_month_day()}}.
}
\details{
\code{x} and \code{n} are recycled against each other.
}
\examples{
library(magrittr)

# Say you started with this zoned time, and you want to add 1 day to it
x <- as_naive_time(year_month_day(1970, 04, 25, 02, 30, 00))
x <- as_zoned_time(x, "America/New_York")
x

# Note that there was a daylight saving time gap on 1970-04-26 where
# we jumped from 01:59:59 -> 03:00:00.

# You can choose to add 1 day in "system time", by first converting to
# sys-time (the equivalent UTC time), adding the day, then converting back to
# zoned time. If you sat still for exactly 86,400 seconds, this is the
# time that you would see after daylight saving time adjusted the clock
# (note that the hour field is shifted forward by the size of the gap)
as_sys_time(x)

x \%>\%
  as_sys_time() \%>\%
  add_days(1) \%>\%
  as_zoned_time(zoned_time_zone(x))

# Alternatively, you can add 1 day in "naive time". Naive time represents
# a clock time with a yet-to-be-specified time zone. It tries to maintain
# smaller units where possible, so adding 1 day would attempt to return
# "1970-04-26T02:30:00" in the America/New_York time zone, but...
as_naive_time(x)

try({
x \%>\%
  as_naive_time() \%>\%
  add_days(1) \%>\%
  as_zoned_time(zoned_time_zone(x))
})

# ...this time doesn't exist in that time zone! It is "nonexistent".
# You can resolve nonexistent times by setting the `nonexistent` argument
# when converting to zoned time. Let's roll forward to the next available
# moment in time.
x \%>\%
  as_naive_time() \%>\%
  add_days(1) \%>\%
  as_zoned_time(zoned_time_zone(x), nonexistent = "roll-forward")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quarterly-year-quarter-day.R
\name{year-quarter-day-count-between}
\alias{year-quarter-day-count-between}
\alias{calendar_count_between.clock_year_quarter_day}
\title{Counting: year-quarter-day}
\usage{
\method{calendar_count_between}{clock_year_quarter_day}(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[clock_year_quarter_day]}

A pair of year-quarter-day vectors. These will be recycled to their
common size.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"quarter"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
This is a year-quarter-day method for the \code{\link[=calendar_count_between]{calendar_count_between()}} generic.
It counts the number of \code{precision} units between \code{start} and \code{end} (i.e.,
the number of years or quarters).
}
\examples{
# Compute the number of whole quarters between two dates
x <- year_quarter_day(2020, 3, 91)
y <- year_quarter_day(2025, 4, c(90, 92))
calendar_count_between(x, y, "quarter")

# Note that this is not always the same as the number of whole 3 month
# periods between two dates
x <- as_year_month_day(x)
y <- as_year_month_day(y)
calendar_count_between(x, y, "month", n = 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/naive-time.R
\name{naive_time_info}
\alias{naive_time_info}
\title{Info: naive-time}
\usage{
naive_time_info(x, zone)
}
\arguments{
\item{x}{\verb{[clock_naive_time]}

A naive-time.}

\item{zone}{\verb{[character]}

A valid time zone name.

Unlike most functions in clock, in \code{naive_time_info()} \code{zone} is vectorized
and is recycled against \code{x}.}
}
\value{
A data frame of low level information.
}
\description{
\code{naive_time_info()} retrieves a set of low-level information generally not
required for most date-time manipulations. It is used implicitly
by \code{as_zoned_time()} when converting from a naive-time.

It returns a data frame with the following columns:
\itemize{
\item \code{type}: A character vector containing one of:
\itemize{
\item \code{"unique"}: The naive-time maps uniquely to a zoned-time that can be
created with \code{zone}.
\item \code{"nonexistent"}: The naive-time does not exist as a zoned-time that can
be created with \code{zone}.
\item \code{"ambiguous"}: The naive-time exists twice as a zoned-time that can be
created with \code{zone}.
}
\item \code{first}: A \code{\link[=sys_time_info]{sys_time_info()}} data frame.
\item \code{second}: A \code{\link[=sys_time_info]{sys_time_info()}} data frame.
}
\subsection{type == "unique"}{
\itemize{
\item \code{first} will be filled out with sys-info representing daylight saving time
information for that time point in \code{zone}.
\item \code{second} will contain only \code{NA} values, as there is no ambiguity to
represent information for.
}
}

\subsection{type == "nonexistent"}{
\itemize{
\item \code{first} will be filled out with the sys-info that ends just prior to \code{x}.
\item \code{second} will be filled out with the sys-info that begins just after \code{x}.
}
}

\subsection{type == "ambiguous"}{
\itemize{
\item \code{first} will be filled out with the sys-info that ends just after \code{x}.
\item \code{second} will be filled out with the sys-info that starts just before \code{x}.
}
}
}
\details{
If the tibble package is installed, it is recommended to convert the output
to a tibble with \code{as_tibble()}, as that will print the df-cols much nicer.
}
\examples{
library(vctrs)

x <- year_month_day(1970, 04, 26, 02, 30, 00)
x <- as_naive_time(x)

# Maps uniquely to a time in London
naive_time_info(x, "Europe/London")

# This naive-time never existed in New York!
# A DST gap jumped the time from 01:59:59 -> 03:00:00,
# skipping the 2 o'clock hour
zone <- "America/New_York"
info <- naive_time_info(x, zone)
info

# You can recreate various `nonexistent` strategies with this info
as_zoned_time(x, zone, nonexistent = "roll-forward")
as_zoned_time(info$first$end, zone)

as_zoned_time(x, zone, nonexistent = "roll-backward")
as_zoned_time(info$first$end - 1, zone)

as_zoned_time(x, zone, nonexistent = "shift-forward")
as_zoned_time(as_sys_time(x) - info$first$offset, zone)

as_zoned_time(x, zone, nonexistent = "shift-backward")
as_zoned_time(as_sys_time(x) - info$second$offset, zone)

# ---------------------------------------------------------------------------
# Normalizing to UTC

# Imagine you had the following printed times, and knowledge that they
# are to be interpreted as in the corresponding time zones
df <- data_frame(
  x = c("2020-01-05 02:30:00", "2020-06-03 12:20:05"),
  zone = c("America/Los_Angeles", "Europe/London")
)

# The times are assumed to be naive-times, i.e. if you lived in the `zone`
# at the moment the time was recorded, then you would have seen that time
# printed on the clock. Currently, these are strings. To convert them to
# a time based type, you'll have to acknowledge that R only lets you have
# 1 time zone in a vector of date-times at a time. So you'll need to
# normalize these naive-times. The easiest thing to normalize them to
# is UTC.
df$naive <- naive_time_parse(df$x)

# Get info about the naive times using a vector of zones
info <- naive_time_info(df$naive, df$zone)
info

# We'll assume that some system generated these naive-times with no
# chance of them ever being nonexistent or ambiguous. So now all we have
# to do is use the offset to convert the naive-time to a sys-time. The
# relationship used is:
# offset = naive_time - sys_time
df$sys <- as_sys_time(df$naive) - info$first$offset
df

# At this point, both times are in UTC. From here, you can convert them
# both to either America/Los_Angeles or Europe/London as required.
as_zoned_time(df$sys, "America/Los_Angeles")
as_zoned_time(df$sys, "Europe/London")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-day.R
\name{year-month-day-boundary}
\alias{year-month-day-boundary}
\alias{calendar_start.clock_year_month_day}
\alias{calendar_end.clock_year_month_day}
\title{Boundaries: year-month-day}
\usage{
\method{calendar_start}{clock_year_month_day}(x, precision)

\method{calendar_end}{clock_year_month_day}(x, precision)
}
\arguments{
\item{x}{\verb{[clock_year_month_day]}

A year-month-day vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"month"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}
}
\value{
\code{x} at the same precision, but with some components altered to be
at the boundary value.
}
\description{
This is a year-month-day method for the \code{\link[=calendar_start]{calendar_start()}} and
\code{\link[=calendar_end]{calendar_end()}} generics. They adjust components of a calendar to the
start or end of a specified \code{precision}.
}
\examples{
# Hour precision
x <- year_month_day(2019, 2:4, 5, 6)
x

# Compute the start of the month
calendar_start(x, "month")

# Or the end of the month, notice that the hour value is adjusted as well
calendar_end(x, "month")

# Compare that with just setting the day of the month to `"last"`, which
# doesn't adjust any other components
set_day(x, "last")

# You can't compute the start / end at a more precise precision than
# the input is at
try(calendar_start(x, "second"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{calendar-count-between}
\alias{calendar-count-between}
\alias{calendar_count_between}
\title{Counting: calendars}
\usage{
calendar_count_between(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[clock_calendar]}

A pair of calendar vectors. These will be recycled to their common size.}

\item{precision}{\verb{[character(1)]}

A precision. Allowed precisions are dependent on the calendar used.}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
\code{calendar_count_between()} counts the number of \code{precision} units between
\code{start} and \code{end} (i.e., the number of years or months). This count
corresponds to the \emph{whole number} of units, and will never return a
fractional value.

This is suitable for, say, computing the whole number of years or months
between two calendar dates, accounting for the day and time of day.

Each calendar has its own help page describing the precisions that you can
count at:
\itemize{
\item \link[=year-month-day-count-between]{year-month-day}
\item \link[=year-month-weekday-count-between]{year-month-weekday}
\item \link[=iso-year-week-day-count-between]{iso-year-week-day}
\item \link[=year-quarter-day-count-between]{year-quarter-day}
\item \link[=year-day-count-between]{year-day}
}
}
\section{Comparison Direction}{

The computed count has the property that if \code{start <= end}, then
\verb{start + <count> <= end}. Similarly, if \code{start >= end}, then
\verb{start + <count> >= end}. In other words, the comparison direction between
\code{start} and \code{end} will never change after adding the count to \code{start}. This
makes this function useful for repeated count computations at
increasingly fine precisions.
}

\examples{
# Number of whole years between these dates
x <- year_month_day(2000, 01, 05)
y <- year_month_day(2005, 01, 04:06)

# Note that `2000-01-05 -> 2005-01-04` is only 4 full years
calendar_count_between(x, y, "year")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{as_date}
\alias{as_date}
\alias{as_date.Date}
\alias{as_date.POSIXt}
\alias{as_date.clock_calendar}
\alias{as_date.clock_time_point}
\alias{as_date.clock_zoned_time}
\title{Convert to a date}
\usage{
as_date(x)

\method{as_date}{Date}(x)

\method{as_date}{POSIXt}(x)

\method{as_date}{clock_calendar}(x)

\method{as_date}{clock_time_point}(x)

\method{as_date}{clock_zoned_time}(x)
}
\arguments{
\item{x}{\verb{[vector]}

A vector.}
}
\value{
A date with the same length as \code{x}.
}
\description{
\code{as_date()} is a generic function that converts its input to a date (Date).

There are methods for converting date-times (POSIXct), calendars,
time points, and zoned-times to dates.

For converting to a date-time, see \code{\link[=as_date_time]{as_date_time()}}.
}
\details{
Note that clock always assumes that R's Date class is naive, so converting
a POSIXct to a Date will always retain the printed year, month, and day
value.

This is not a drop-in replacement for \code{as.Date()}, as it only converts a
limited set of types to Date. For parsing characters as dates, see
\code{\link[=date_parse]{date_parse()}}. For converting numerics to dates, see \code{\link[vctrs:new_date]{vctrs::new_date()}} or
continue to use \code{as.Date()}.
}
\examples{
x <- date_time_parse("2019-01-01 23:02:03", "America/New_York")

# R's `as.Date.POSIXct()` method defaults to changing the printed time
# to UTC before converting, which can result in odd conversions like this:
as.Date(x)

# `as_date()` will never change the printed time before converting
as_date(x)

# Can also convert from other clock types
as_date(year_month_day(2019, 2, 5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date_group}
\alias{date_group}
\title{Group date and date-time components}
\usage{
date_group(x, precision, ..., n = 1L)
}
\arguments{
\item{x}{\verb{[Date / POSIXct / POSIXlt]}

A date or date-time vector.}

\item{precision}{\verb{[character(1)]}

A precision. Allowed precisions are dependent on the input used.}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
\code{x}, grouped at \code{precision}.
}
\description{
\code{date_group()} groups by a single component of a date-time, such as month
of the year, or day of the month.

There are separate help pages for grouping dates and date-times:
\itemize{
\item \link[=date-group]{dates (Date)}
\item \link[=posixt-group]{date-times (POSIXct/POSIXlt)}
}
}
\examples{
# See type specific documentation for more examples
date_group(as.Date("2019-01-01") + 0:5, "day", n = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weekday.R
\name{weekday-arithmetic}
\alias{weekday-arithmetic}
\alias{add_days.clock_weekday}
\title{Arithmetic: weekday}
\usage{
\method{add_days}{clock_weekday}(x, n, ...)
}
\arguments{
\item{x}{\verb{[clock_weekday]}

A weekday vector.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} after performing the arithmetic.
}
\description{
These are weekday methods for the
\link[=clock-arithmetic]{arithmetic generics}.
\itemize{
\item \code{add_days()}
}

Also check out the examples on the \code{\link[=weekday]{weekday()}} page for more advanced
usage.
}
\details{
\code{x} and \code{n} are recycled against each other.
}
\examples{
saturday <- weekday(clock_weekdays$saturday)
saturday

add_days(saturday, 1)
add_days(saturday, 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{posixt-shifting}
\alias{posixt-shifting}
\alias{date_shift.POSIXt}
\title{Shifting: date and date-time}
\usage{
\method{date_shift}{POSIXt}(
  x,
  target,
  ...,
  which = "next",
  boundary = "keep",
  nonexistent = NULL,
  ambiguous = x
)
}
\arguments{
\item{x}{\verb{[POSIXct / POSIXlt]}

A date-time vector.}

\item{target}{\verb{[weekday]}

A weekday created from \code{\link[=weekday]{weekday()}} to target.

Generally this is length 1, but can also be the same length as \code{x}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{which}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"next"}: Shift to the next instance of the \code{target} weekday.
\item \verb{"previous}: Shift to the previous instance of the \code{target} weekday.
}}

\item{boundary}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"keep"}: If \code{x} is currently on the \code{target} weekday, return it.
\item \code{"advance"}: If \code{x} is currently on the \code{target} weekday, advance it
anyways.
}}

\item{nonexistent}{\verb{[character / NULL]}

One of the following nonexistent time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"roll-forward"}: The next valid instant in time.
\item \code{"roll-backward"}: The previous valid instant in time.
\item \code{"shift-forward"}: Shift the nonexistent time forward by the size of
the daylight saving time gap.
\item \verb{"shift-backward}: Shift the nonexistent time backward by the size of
the daylight saving time gap.
\item \code{"NA"}: Replace nonexistent times with \code{NA}.
\item \code{"error"}: Error on nonexistent times.
}

Using either \code{"roll-forward"} or \code{"roll-backward"} is generally
recommended over shifting, as these two strategies maintain the
\emph{relative ordering} between elements of the input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{nonexistent} must be supplied
and cannot be \code{NULL}. This is a convenient way to make production code
robust to nonexistent times.}

\item{ambiguous}{\verb{[character / zoned_time / POSIXct / list(2) / NULL]}

One of the following ambiguous time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"earliest"}: Of the two possible times, choose the earliest one.
\item \code{"latest"}: Of the two possible times, choose the latest one.
\item \code{"NA"}: Replace ambiguous times with \code{NA}.
\item \code{"error"}: Error on ambiguous times.
}

Alternatively, \code{ambiguous} is allowed to be a zoned_time (or POSIXct) that
is either length 1, or the same length as the input. If an ambiguous time
is encountered, the zoned_time is consulted. If the zoned_time corresponds
to a naive_time that is also ambiguous \emph{and} uses the same daylight saving
time transition point as the original ambiguous time, then the offset of
the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
resolved by consulting the zoned_time, then this method falls back to
\code{NULL}.

Finally, \code{ambiguous} is allowed to be a list of size 2, where the first
element of the list is a zoned_time (as described above), and the second
element of the list is an ambiguous time resolution strategy to use when
the ambiguous time cannot be resolved by consulting the zoned_time.
Specifying a zoned_time on its own is identical to \verb{list(<zoned_time>, NULL)}.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{ambiguous} must be supplied and
cannot be \code{NULL}. Additionally, \code{ambiguous} cannot be specified as a
zoned_time on its own, as this implies \code{NULL} for ambiguous times that the
zoned_time cannot resolve. Instead, it must be specified as a list
alongside an ambiguous time resolution strategy as described above. This is
a convenient way to make production code robust to ambiguous times.}
}
\value{
\code{x} shifted to the \code{target} weekday.
}
\description{
\code{date_shift()} shifts \code{x} to the \code{target} weekday. You can shift to the next
or previous weekday. If \code{x} is currently on the \code{target} weekday, you can
choose to leave it alone or advance it to the next instance of the \code{target}.

Shifting with date-times retains the time of day where possible. Be aware
that you can run into daylight saving time issues if you shift into a
daylight saving time gap or fallback period.
}
\examples{
tuesday <- weekday(clock_weekdays$tuesday)

x <- as.POSIXct("1970-04-22 02:30:00", "America/New_York")

# Shift to the next Tuesday
date_shift(x, tuesday)

# Be aware that you can run into daylight saving time issues!
# Here we shift directly into a daylight saving time gap
# from 01:59:59 -> 03:00:00
sunday <- weekday(clock_weekdays$sunday)
try(date_shift(x, sunday))

# You can resolve this with the `nonexistent` argument
date_shift(x, sunday, nonexistent = "roll-forward")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date-rounding}
\alias{date-rounding}
\alias{date_floor.Date}
\alias{date_ceiling.Date}
\alias{date_round.Date}
\title{Rounding: date}
\usage{
\method{date_floor}{Date}(x, precision, ..., n = 1L, origin = NULL)

\method{date_ceiling}{Date}(x, precision, ..., n = 1L, origin = NULL)

\method{date_round}{Date}(x, precision, ..., n = 1L, origin = NULL)
}
\arguments{
\item{x}{\verb{[Date]}

A date vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"week"}
\item \code{"day"}
}

\code{"week"} is an alias for \code{"day"} with \code{n * 7}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}

\item{origin}{\verb{[Date(1) / NULL]}

An origin to start counting from. The default \code{origin} is
1970-01-01.}
}
\value{
\code{x} rounded to the specified \code{precision}.
}
\description{
These are Date methods for the
\link[=date-and-date-time-rounding]{rounding generics}.
\itemize{
\item \code{date_floor()} rounds a date down to a multiple of
the specified \code{precision}.
\item \code{date_ceiling()} rounds a date up to a multiple of
the specified \code{precision}.
\item \code{date_round()} rounds up or down depending on what is closer,
rounding up on ties.
}

The only supported rounding \code{precision}s for Dates are \code{"day"} and \code{"week"}.
You can group by irregular periods such as \code{"month"} or \code{"year"} by using
\code{\link[=date_group]{date_group()}}.
}
\details{
When rounding by \code{"week"}, remember that the \code{origin} determines the "week
start". By default, 1970-01-01 is the implicit origin, which is a
Thursday. If you would like to round by weeks with a different week start,
just supply an origin on the weekday you are interested in.
}
\examples{
x <- as.Date("2019-03-31") + 0:5
x

# Flooring by 2 days, note that this is not tied to the current month,
# and instead counts from the specified `origin`, so groups can cross
# the month boundary
date_floor(x, "day", n = 2)

# Compare to `date_group()`, which groups by the day of the month
date_group(x, "day", n = 2)

y <- as.Date("2019-01-01") + 0:20
y

# Flooring by week uses an implicit `origin` of 1970-01-01, which
# is a Thursday
date_floor(y, "week")
as_weekday(date_floor(y, "week"))

# If you want to round by weeks with a different week start, supply an
# `origin` that falls on the weekday you care about. This uses a Monday.
origin <- as.Date("1970-01-05")
as_weekday(origin)

date_floor(y, "week", origin = origin)
as_weekday(date_floor(y, "week", origin = origin))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sys-time.R
\name{sys-parsing}
\alias{sys-parsing}
\alias{sys_time_parse}
\alias{sys_time_parse_RFC_3339}
\title{Parsing: sys-time}
\usage{
sys_time_parse(
  x,
  ...,
  format = NULL,
  precision = "second",
  locale = clock_locale()
)

sys_time_parse_RFC_3339(
  x,
  ...,
  separator = "T",
  offset = "Z",
  precision = "second"
)
}
\arguments{
\item{x}{\verb{[character]}

A character vector to parse.}

\item{...}{These dots are for future extensions and must be empty.}

\item{format}{\verb{[character / NULL]}

A format string. A combination of the following commands, or \code{NULL},
in which case a default format string is used.

A vector of multiple format strings can be supplied. They will be tried in
the order they are provided.

\strong{Year}
\itemize{
\item \verb{\%C}: The century as a decimal number. The modified command \verb{\%NC} where
\code{N} is a positive decimal integer specifies the maximum number of
characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%y}: The last two decimal digits of the year. If the century is not
otherwise specified (e.g. with \verb{\%C}), values in the range \verb{[69 - 99]} are
presumed to refer to the years \verb{[1969 - 1999]}, and values in the range
\verb{[00 - 68]} are presumed to refer to the years \verb{[2000 - 2068]}. The
modified command \verb{\%Ny}, where \code{N} is a positive decimal integer, specifies
the maximum number of characters to read. If not specified, the default is
\code{2}. Leading zeroes are permitted but not required.
\item \verb{\%Y}: The year as a decimal number. The modified command \verb{\%NY} where \code{N}
is a positive decimal integer specifies the maximum number of characters to
read. If not specified, the default is \code{4}. Leading zeroes are permitted
but not required.
}

\strong{Month}
\itemize{
\item \verb{\%b}, \verb{\%B}, \verb{\%h}: The \code{locale}'s full or abbreviated case-insensitive
month name.
\item \verb{\%m}: The month as a decimal number. January is \code{1}. The modified command
\verb{\%Nm} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
}

\strong{Day}
\itemize{
\item \verb{\%d}, \verb{\%e}: The day of the month as a decimal number. The modified
command \verb{\%Nd} where \code{N} is a positive decimal integer specifies the maximum
number of characters to read. If not specified, the default is \code{2}. Leading
zeroes are permitted but not required.
}

\strong{Day of the week}
\itemize{
\item \verb{\%a}, \verb{\%A}: The \code{locale}'s full or abbreviated case-insensitive weekday
name.
\item \verb{\%w}: The weekday as a decimal number (\code{0-6}), where Sunday is \code{0}. The
modified command \verb{\%Nw} where \code{N} is a positive decimal integer specifies
the maximum number of characters to read. If not specified, the default is
\code{1}. Leading zeroes are permitted but not required.
}

\strong{ISO 8601 week-based year}
\itemize{
\item \verb{\%g}: The last two decimal digits of the ISO week-based year. The
modified command \verb{\%Ng} where \code{N} is a positive decimal integer specifies
the maximum number of characters to read. If not specified, the default is
\code{2}. Leading zeroes are permitted but not required.
\item \verb{\%G}: The ISO week-based year as a decimal number. The modified command
\verb{\%NG} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{4}. Leading zeroes
are permitted but not required.
\item \verb{\%V}: The ISO week-based week number as a decimal number. The modified
command \verb{\%NV} where \code{N} is a positive decimal integer specifies the maximum
number of characters to read. If not specified, the default is \code{2}. Leading
zeroes are permitted but not required.
\item \verb{\%u}: The ISO weekday as a decimal number (\code{1-7}), where Monday is \code{1}.
The modified command \verb{\%Nu} where \code{N} is a positive decimal integer
specifies the maximum number of characters to read. If not specified, the
default is \code{1}. Leading zeroes are permitted but not required.
}

\strong{Week of the year}
\itemize{
\item \verb{\%U}: The week number of the year as a decimal number. The first Sunday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. The modified command \verb{\%NU} where \code{N} is a positive
decimal integer specifies the maximum number of characters to read. If not
specified, the default is \code{2}. Leading zeroes are permitted but not
required.
\item \verb{\%W}: The week number of the year as a decimal number. The first Monday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. The modified command \verb{\%NW} where \code{N} is a positive
decimal integer specifies the maximum number of characters to read. If not
specified, the default is \code{2}. Leading zeroes are permitted but not
required.
}

\strong{Day of the year}
\itemize{
\item \verb{\%j}: The day of the year as a decimal number. January 1 is \code{1}. The
modified command \verb{\%Nj} where \code{N} is a positive decimal integer specifies
the maximum number of characters to read. If not specified, the default is
\code{3}. Leading zeroes are permitted but not required.
}

\strong{Date}
\itemize{
\item \verb{\%D}, \verb{\%x}: Equivalent to \verb{\%m/\%d/\%y}.
\item \verb{\%F}: Equivalent to \verb{\%Y-\%m-\%d}. If modified with a width (like \verb{\%NF}),
the width is applied to only \verb{\%Y}.
}

\strong{Time of day}
\itemize{
\item \verb{\%H}: The hour (24-hour clock) as a decimal number. The modified command
\verb{\%NH} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%I}: The hour (12-hour clock) as a decimal number. The modified command
\verb{\%NI} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%M}: The minutes as a decimal number. The modified command \verb{\%NM} where
\code{N} is a positive decimal integer specifies the maximum number of
characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%S}: The seconds as a decimal number. Leading zeroes are permitted but
not required. If encountered, the \code{locale} determines the decimal point
character. Generally, the maximum number of characters to read is
determined by the precision that you are parsing at. For example, a
precision of \code{"second"} would read a maximum of 2 characters, while a
precision of \code{"millisecond"} would read a maximum of 6 (2 for the values
before the decimal point, 1 for the decimal point, and 3 for the values
after it). The modified command \verb{\%NS}, where \code{N} is a positive decimal
integer, can be used to exactly specify the maximum number of characters to
read. This is only useful if you happen to have seconds with more than 1
leading zero.
\item \verb{\%p}: The \code{locale}'s equivalent of the AM/PM designations associated with
a 12-hour clock. The command \verb{\%I} must precede \verb{\%p} in the format string.
\item \verb{\%R}: Equivalent to \verb{\%H:\%M}.
\item \verb{\%T}, \verb{\%X}: Equivalent to \verb{\%H:\%M:\%S}.
\item \verb{\%r}: Equivalent to \verb{\%I:\%M:\%S \%p}.
}

\strong{Time zone}
\itemize{
\item \verb{\%z}: The offset from UTC in the format \verb{[+|-]hh[mm]}. For example
\code{-0430} refers to 4 hours 30 minutes behind UTC. And \code{04} refers to 4 hours
ahead of UTC. The modified command \verb{\%Ez} parses a \code{:} between the hours and
minutes and leading zeroes on the hour field are optional:
\verb{[+|-]h[h][:mm]}. For example \code{-04:30} refers to 4 hours 30 minutes behind
UTC. And \code{4} refers to 4 hours ahead of UTC.
\item \verb{\%Z}: The full time zone name or the time zone abbreviation, depending on
the function being used. A single word is parsed. This word can only
contain characters that are alphanumeric, or one of \code{'_'}, \code{'/'}, \code{'-'} or
\code{'+'}.
}

\strong{Miscellaneous}
\itemize{
\item \verb{\%c}: A date and time representation. Equivalent to
\verb{\%a \%b \%d \%H:\%M:\%S \%Y}.
\item \code{\%\%}: A \verb{\%} character.
\item \verb{\%n}: Matches one white space character. \verb{\%n}, \verb{\%t}, and a space can be
combined to match a wide range of white-space patterns. For example \code{"\%n "}
matches one or more white space characters, and \code{"\%n\%t\%t"} matches one to
three white space characters.
\item \verb{\%t}: Matches zero or one white space characters.
}}

\item{precision}{\verb{[character(1)]}

A precision for the resulting time point. One of:
\itemize{
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}

Setting the \code{precision} determines how much information \verb{\%S} attempts
to parse.}

\item{locale}{\verb{[clock_locale]}

A locale object created from \code{\link[=clock_locale]{clock_locale()}}.}

\item{separator}{\verb{[character(1)]}

The separator between the date and time components of the string. One of:
\itemize{
\item \code{"T"}
\item \code{"t"}
\item \code{" "}
}}

\item{offset}{\verb{[character(1)]}

The format of the offset from UTC contained in the string. One of:
\itemize{
\item \code{"Z"}
\item \code{"z"}
\item \code{"\%z"} to parse a numeric offset of the form \code{"+0430"}
\item \code{"\%Ez"} to parse a numeric offset of the form \code{"+04:30"}
}}
}
\value{
A sys-time.
}
\description{
There are two parsers into a sys-time, \code{sys_time_parse()} and
\code{sys_time_parse_RFC_3339()}.
\subsection{sys_time_parse()}{

\code{sys_time_parse()} is useful when you have date-time strings like
\code{"2020-01-01T01:04:30"} that you know should be interpreted as UTC, or like
\code{"2020-01-01T01:04:30-04:00"} with a UTC offset but no zone name. If you find
yourself in the latter situation, then parsing this string as a sys-time
using the \verb{\%Ez} command to capture the offset is probably your best option.
If you know that this string should be interpreted in a specific time zone,
parse as a sys-time to get the UTC equivalent, then use \code{\link[=as_zoned_time]{as_zoned_time()}}.

The default options assume that \code{x} should be parsed at second precision,
using a \code{format} string of \code{"\%Y-\%m-\%dT\%H:\%M:\%S"}. This matches the default
result from calling \code{format()} on a sys-time.

\code{sys_time_parse()} is nearly equivalent to \code{\link[=naive_time_parse]{naive_time_parse()}}, except for
the fact that the \verb{\%z} command is actually used. Using \verb{\%z} assumes that the
rest of the date-time string should be interpreted as a naive-time, which is
then shifted by the UTC offset found in \verb{\%z}. The returned time can then be
validly interpreted as UTC.

\emph{\code{sys_time_parse()} ignores the \verb{\%Z} command.}
}

\subsection{sys_time_parse_RFC_3339()}{

\code{sys_time_parse_RFC_3339()} is a wrapper around \code{sys_time_parse()} that is
intended to parse the extremely common date-time format outlined by
\href{https://datatracker.ietf.org/doc/html/rfc3339}{RFC 3339}. This document
outlines a profile of the ISO 8601 format that is even more restrictive.

In particular, this function is intended to parse the following three
formats:

\if{html}{\out{<div class="sourceCode">}}\preformatted{2019-01-01T00:00:00Z
2019-01-01T00:00:00+0430
2019-01-01T00:00:00+04:30
}\if{html}{\out{</div>}}

This function defaults to parsing the first of these formats by using
a format string of \code{"\%Y-\%m-\%dT\%H:\%M:\%SZ"}.

If your date-time strings use offsets from UTC rather than \code{"Z"}, then set
\code{offset} to one of the following:
\itemize{
\item \code{"\%z"} if the offset is of the form \code{"+0430"}.
\item \code{"\%Ez"} if the offset is of the form \code{"+04:30"}.
}

The RFC 3339 standard allows for replacing the \code{"T"} with a \code{"t"} or a space
(\code{" "}). Set \code{separator} to adjust this as needed.

For this function, the \code{precision} must be at least \code{"second"}.
}
}
\details{
If your date-time strings contain a full time zone name and a UTC offset, use
\code{\link[=zoned_time_parse_complete]{zoned_time_parse_complete()}}. If they contain a time zone abbreviation, use
\code{\link[=zoned_time_parse_abbrev]{zoned_time_parse_abbrev()}}.

If your date-time strings don't contain an offset from UTC and you aren't
sure if they should be treated as UTC or not, you might consider using
\code{\link[=naive_time_parse]{naive_time_parse()}}, since the resulting naive-time doesn't come with an
assumption of a UTC time zone.
}
\section{Full Precision Parsing}{


It is highly recommended to parse all of the information in the date-time
string into a type at least as precise as the string. For example, if your
string has fractional seconds, but you only require seconds, specify a
sub-second \code{precision}, then round to seconds manually using whatever
convention is appropriate for your use case. Parsing such a string directly
into a second precision result is ambiguous and undefined, and is unlikely to
work as you might expect.
}

\examples{
sys_time_parse("2020-01-01T05:06:07")

# Day precision
sys_time_parse("2020-01-01", precision = "day")

# Nanosecond precision, but using a day based format
sys_time_parse("2020-01-01", format = "\%Y-\%m-\%d", precision = "nanosecond")

# Multiple format strings are allowed for heterogeneous times
sys_time_parse(
  c("2019-01-01", "2019/1/1"),
  format = c("\%Y/\%m/\%d", "\%Y-\%m-\%d"),
  precision = "day"
)

# The `\%z` command shifts the date-time by subtracting the UTC offset so
# that the returned sys-time can be interpreted as UTC
sys_time_parse(
  "2020-01-01 02:00:00 -0400",
  format = "\%Y-\%m-\%d \%H:\%M:\%S \%z"
)

# Remember that the `\%Z` command is ignored entirely!
sys_time_parse("2020-01-01 America/New_York", format = "\%Y-\%m-\%d \%Z")

# ---------------------------------------------------------------------------
# RFC 3339

# Typical UTC format
x <- "2019-01-01T00:01:02Z"
sys_time_parse_RFC_3339(x)

# With a UTC offset containing a `:`
x <- "2019-01-01T00:01:02+02:30"
sys_time_parse_RFC_3339(x, offset = "\%Ez")

# With a space between the date and time and no `:` in the offset
x <- "2019-01-01 00:01:02+0230"
sys_time_parse_RFC_3339(x, separator = " ", offset = "\%z")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/week-year-week-day.R
\name{year_week_day}
\alias{year_week_day}
\title{Calendar: year-week-day}
\usage{
year_week_day(
  year,
  week = NULL,
  day = NULL,
  hour = NULL,
  minute = NULL,
  second = NULL,
  subsecond = NULL,
  ...,
  start = NULL,
  subsecond_precision = NULL
)
}
\arguments{
\item{year}{\verb{[integer]}

The year. Values \verb{[-32767, 32767]} are generally allowed.}

\item{week}{\verb{[integer / "last" / NULL]}

The week. Values \verb{[1, 53]} are allowed.

If \code{"last"}, then the last week of the year is returned.}

\item{day}{\verb{[integer / NULL]}

The day of the week. Values \verb{[1, 7]} are allowed, with \verb{1 = start of week}
and \verb{7 = end of week}, in accordance with \code{start}.}

\item{hour}{\verb{[integer / NULL]}

The hour. Values \verb{[0, 23]} are allowed.}

\item{minute}{\verb{[integer / NULL]}

The minute. Values \verb{[0, 59]} are allowed.}

\item{second}{\verb{[integer / NULL]}

The second. Values \verb{[0, 59]} are allowed.}

\item{subsecond}{\verb{[integer / NULL]}

The subsecond. If specified, \code{subsecond_precision} must also be specified
to determine how to interpret the \code{subsecond}.

If using milliseconds, values \verb{[0, 999]} are allowed.

If using microseconds, values \verb{[0, 999999]} are allowed.

If using nanoseconds, values \verb{[0, 999999999]} are allowed.}

\item{...}{These dots are for future extensions and must be empty.}

\item{start}{\verb{[integer(1) / NULL]}

The day to consider the start of the week. 1 = Sunday and 7 = Saturday.

Use \link{clock_weekdays} for a readable way to specify the start.

If \code{NULL}, a \code{start} of Sunday will be used.}

\item{subsecond_precision}{\verb{[character(1) / NULL]}

The precision to interpret \code{subsecond} as. One of: \code{"millisecond"},
\code{"microsecond"}, or \code{"nanosecond"}.}
}
\value{
A year-week-day calendar vector.
}
\description{
\code{year_week_day()} constructs a calendar from the year, week number,
week day, and the \code{start} of the week.

Using \code{start = clock_weekdays$monday} represents the ISO week calendar and
is equivalent to using \code{\link[=iso_year_week_day]{iso_year_week_day()}}.

Using \code{start = clock_weekdays$sunday} is how Epidemiologists encode their
week-based data.
}
\details{
Fields are recycled against each other using
\link[vctrs:theory-faq-recycling]{tidyverse recycling rules}.

Fields are collected in order until the first \code{NULL} field is located. No
fields after the first \code{NULL} field are used.
}
\examples{
# Year-week
x <- year_week_day(2019:2025, "last")
x

# Start the week on Monday
y <- year_week_day(2019:2025, "last", start = clock_weekdays$monday)
y

# Last days of the year
as_year_month_day(set_day(x, 7))
as_year_month_day(set_day(y, 7))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sys-time.R
\name{sys_time_info}
\alias{sys_time_info}
\title{Info: sys-time}
\usage{
sys_time_info(x, zone)
}
\arguments{
\item{x}{\verb{[clock_sys_time]}

A sys-time.}

\item{zone}{\verb{[character]}

A valid time zone name.

Unlike most functions in clock, in \code{sys_time_info()} \code{zone} is vectorized
and is recycled against \code{x}.}
}
\value{
A data frame of low level information.
}
\description{
\code{sys_time_info()} retrieves a set of low-level information generally not
required for most date-time manipulations. It returns a data frame with the
following columns:
\itemize{
\item \code{begin}, \code{end}: Second precision sys-times specifying the range of the
current daylight saving time rule. The range is a half-open interval of
\verb{[begin, end)}.
\item \code{offset}: A second precision \code{duration} specifying the offset from UTC.
\item \code{dst}: A logical vector specifying if daylight saving time is currently
active.
\item \code{abbreviation}: The time zone abbreviation in use throughout this \code{begin}
to \code{end} range.
}
}
\details{
If there have never been any daylight saving time transitions, the minimum
supported year value is returned for \code{begin} (typically, a year value of
\code{-32767}).

If daylight saving time is no longer used in a time zone, the maximum
supported year value is returned for \code{end} (typically, a year value of
\code{32767}).

The \code{offset} is the bridge between sys-time and naive-time for the \code{zone}
being used. The relationship of the three values is:

\if{html}{\out{<div class="sourceCode">}}\preformatted{offset = naive_time - sys_time
}\if{html}{\out{</div>}}
}
\examples{
library(vctrs)

x <- year_month_day(2021, 03, 14, c(01, 03), c(59, 00), c(59, 00))
x <- as_naive_time(x)
x <- as_zoned_time(x, "America/New_York")

# x[1] is in EST, x[2] is in EDT
x

x_sys <- as_sys_time(x)

info <- sys_time_info(x_sys, zoned_time_zone(x))
info

# Convert `begin` and `end` to zoned-times to see the previous and
# next daylight saving time transitions
data_frame(
  x = x,
  begin = as_zoned_time(info$begin, zoned_time_zone(x)),
  end = as_zoned_time(info$end, zoned_time_zone(x))
)

# `end` can be used to iterate through daylight saving time transitions
# by repeatedly calling `sys_time_info()`
sys_time_info(info$end, zoned_time_zone(x))

# Multiple `zone`s can be supplied to look up daylight saving time
# information in different time zones
zones <- c("America/New_York", "America/Los_Angeles")

info2 <- sys_time_info(x_sys[1], zones)
info2

# The offset can be used to display the naive-time (i.e. the printed time)
# in both of those time zones
data_frame(
  zone = zones,
  naive_time = x_sys[1] + info2$offset
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weekday.R
\name{weekday_factor}
\alias{weekday_factor}
\title{Convert a weekday to an ordered factor}
\usage{
weekday_factor(x, ..., labels = "en", abbreviate = TRUE, encoding = "western")
}
\arguments{
\item{x}{\verb{[weekday]}

A weekday vector.}

\item{...}{These dots are for future extensions and must be empty.}

\item{labels}{\verb{[clock_labels / character(1)]}

Character representations of localized weekday names, month names, and
AM/PM names. Either the language code as string (passed on to
\code{\link[=clock_labels_lookup]{clock_labels_lookup()}}), or an object created by \code{\link[=clock_labels]{clock_labels()}}.}

\item{abbreviate}{\verb{[logical(1)]}

If \code{TRUE}, the abbreviated weekday names from \code{labels} will be used.

If \code{FALSE}, the full weekday names from \code{labels} will be used.}

\item{encoding}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"western"}: Encode the weekdays as an ordered factor with levels from
Sunday -> Saturday.
\item \code{"iso"}: Encode the weekdays as an ordered factor with levels from
Monday -> Sunday.
}}
}
\value{
An ordered factor representing the weekdays.
}
\description{
\code{weekday_factor()} converts a weekday object to an ordered factor. This
can be useful in combination with ggplot2, or for modeling.
}
\examples{
x <- weekday(1:7)

# Default to Sunday -> Saturday
weekday_factor(x)

# ISO encoding is Monday -> Sunday
weekday_factor(x, encoding = "iso")

# With full names
weekday_factor(x, abbreviate = FALSE)

# Or a different language
weekday_factor(x, labels = "fr")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setters.R
\name{clock-setters}
\alias{clock-setters}
\alias{set_year}
\alias{set_quarter}
\alias{set_month}
\alias{set_week}
\alias{set_day}
\alias{set_hour}
\alias{set_minute}
\alias{set_second}
\alias{set_millisecond}
\alias{set_microsecond}
\alias{set_nanosecond}
\alias{set_index}
\title{Calendar setters}
\usage{
set_year(x, value, ...)

set_quarter(x, value, ...)

set_month(x, value, ...)

set_week(x, value, ...)

set_day(x, value, ...)

set_hour(x, value, ...)

set_minute(x, value, ...)

set_second(x, value, ...)

set_millisecond(x, value, ...)

set_microsecond(x, value, ...)

set_nanosecond(x, value, ...)

set_index(x, value, ...)
}
\arguments{
\item{x}{\verb{[object]}

An object to set the component for.}

\item{value}{\verb{[integer]}

The value to set the component to.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} with the component set.
}
\description{
This family of functions sets fields in a calendar vector. Each
calendar has its own set of supported setters, which are documented on
their own help page:
\itemize{
\item \link[=year-month-day-setters]{year-month-day}
\item \link[=year-month-weekday-setters]{year-month-weekday}
\item \link[=iso-year-week-day-setters]{iso-year-week-day}
\item \link[=year-quarter-day-setters]{year-quarter-day}
\item \link[=year-day-setters]{year-day}
}

There are also convenience methods for setting certain components
directly on R's native date and date-time types.
\itemize{
\item \link[=Date-setters]{dates (Date)}
\item \link[=posixt-setters]{date-times (POSIXct / POSIXlt)}
}

Some general rules about setting components on calendar types:
\itemize{
\item You can only set components that are relevant to the calendar type that
you are working with. For example, you can't set the quarter of a
year-month-day type. You'd have to convert to year-quarter-day first.
\item You can set a component that is at the current precision, or one level
of precision more precise than the current precision. For example,
you can set the day field of a month precision year-month-day type,
but not the hour field.
\item Setting a component can result in an \emph{invalid date}, such as
\code{set_day(year_month_day(2019, 02), 31)}, as long as it is eventually
resolved either manually or with a strategy from \code{\link[=invalid_resolve]{invalid_resolve()}}.
\item With sub-second precisions, you can only set the component corresponding
to the precision that you are at. For example, you can set the nanoseconds
of the second while at nanosecond precision, but not milliseconds.
}
}
\details{
You cannot set components directly on a time point type, such as
sys-time or naive-time. Convert it to a calendar type first. Similarly,
a zoned-time must be converted to either a sys-time or naive-time, and
then to a calendar type, to be able to set components on it.
}
\examples{
x <- year_month_day(2019, 1:3)

# Set the day
set_day(x, 12:14)

# Set to the "last" day of the month
set_day(x, "last")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sys-time.R
\name{is_sys_time}
\alias{is_sys_time}
\title{Is \code{x} a sys-time?}
\usage{
is_sys_time(x)
}
\arguments{
\item{x}{\verb{[object]}

An object.}
}
\value{
\code{TRUE} if \code{x} inherits from \code{"clock_sys_time"}, otherwise \code{FALSE}.
}
\description{
This function determines if the input is a sys-time object.
}
\examples{
is_sys_time(1)
is_sys_time(as_sys_time(duration_days(1)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{year-month-weekday-getters}
\alias{year-month-weekday-getters}
\alias{get_year.clock_year_month_weekday}
\alias{get_month.clock_year_month_weekday}
\alias{get_day.clock_year_month_weekday}
\alias{get_index.clock_year_month_weekday}
\alias{get_hour.clock_year_month_weekday}
\alias{get_minute.clock_year_month_weekday}
\alias{get_second.clock_year_month_weekday}
\alias{get_millisecond.clock_year_month_weekday}
\alias{get_microsecond.clock_year_month_weekday}
\alias{get_nanosecond.clock_year_month_weekday}
\title{Getters: year-month-weekday}
\usage{
\method{get_year}{clock_year_month_weekday}(x)

\method{get_month}{clock_year_month_weekday}(x)

\method{get_day}{clock_year_month_weekday}(x)

\method{get_index}{clock_year_month_weekday}(x)

\method{get_hour}{clock_year_month_weekday}(x)

\method{get_minute}{clock_year_month_weekday}(x)

\method{get_second}{clock_year_month_weekday}(x)

\method{get_millisecond}{clock_year_month_weekday}(x)

\method{get_microsecond}{clock_year_month_weekday}(x)

\method{get_nanosecond}{clock_year_month_weekday}(x)
}
\arguments{
\item{x}{\verb{[clock_year_month_weekday]}

A year-month-weekday to get the component from.}
}
\value{
The component.
}
\description{
These are year-month-weekday methods for the
\link[=clock-getters]{getter generics}.
\itemize{
\item \code{get_year()} returns the Gregorian year.
\item \code{get_month()} returns the month of the year.
\item \code{get_day()} returns the day of the week encoded from 1-7, where 1 = Sunday
and 7 = Saturday.
\item \code{get_index()} returns a value from 1-5 indicating that the corresponding
weekday is the n-th instance of that weekday in the current month.
\item There are sub-daily getters for extracting more precise components.
}
}
\examples{
monday <- clock_weekdays$monday
thursday <- clock_weekdays$thursday

x <- year_month_weekday(2019, 1, monday:thursday, 1:4)
x

# Gets the weekday, 1 = Sunday, 7 = Saturday
get_day(x)

# Gets the index indicating which instance of that particular weekday
# it is in the current month (i.e. the "1st Monday of January, 2019")
get_index(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-point.R
\name{time-point-rounding}
\alias{time-point-rounding}
\alias{time_point_floor}
\alias{time_point_ceiling}
\alias{time_point_round}
\title{Time point rounding}
\usage{
time_point_floor(x, precision, ..., n = 1L, origin = NULL)

time_point_ceiling(x, precision, ..., n = 1L, origin = NULL)

time_point_round(x, precision, ..., n = 1L, origin = NULL)
}
\arguments{
\item{x}{\verb{[clock_sys_time / clock_naive_time]}

A sys-time or naive-time.}

\item{precision}{\verb{[character(1)]}

A time point precision. One of:
\itemize{
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A positive integer specifying the multiple of \code{precision} to use.}

\item{origin}{\verb{[clock_sys_time(1) / clock_naive_time(1) / NULL]}

An origin to begin counting from. Mostly useful when \code{n > 1} and you
want to control how the rounding groups are created.

If \code{x} is a sys-time, \code{origin} must be a sys-time.

If \code{x} is a naive-time, \code{origin} must be a naive-time.

The precision of \code{origin} must be equally precise as or less
precise than \code{precision}.

If \code{NULL}, a default origin of midnight on 1970-01-01 is used.}
}
\value{
\code{x} rounded to the new \code{precision}.
}
\description{
\itemize{
\item \code{time_point_floor()} rounds a sys-time or naive-time down to a multiple of
the specified \code{precision}.
\item \code{time_point_ceiling()} rounds a sys-time or naive-time up to a multiple of
the specified \code{precision}.
\item \code{time_point_round()} rounds up or down depending on what is closer,
rounding up on ties.
}

Rounding time points is mainly useful for rounding sub-daily time points
up to daily time points.

It can also be useful for flooring by a set number of days (like 20) with
respect to some origin. By default, the origin is 1970-01-01 00:00:00.

If you want to group by components, such as "day of the month", rather than
by "n days", see \code{\link[=calendar_group]{calendar_group()}}.
}
\section{Boundary Handling}{


To understand how flooring and ceiling work, you need to know how they
create their intervals for rounding.
\itemize{
\item \code{time_point_floor()} constructs intervals of \code{[lower, upper)} that
bound each element of \code{x}, then always chooses the \emph{left-hand side}.
\item \code{time_point_ceiling()} constructs intervals of \code{(lower, upper]} that
bound each element of \code{x}, then always chooses the \emph{right-hand side}.
}

As an easy example, consider 2020-01-02 00:00:05.

To floor this to the nearest day, the following interval is constructed,
and the left-hand side is returned at day precision:

\code{[2020-01-02 00:00:00, 2020-01-03 00:00:00)}

To ceiling this to the nearest day, the following interval
is constructed, and the right-hand side is returned at day precision:

\code{(2020-01-02 00:00:00, 2020-01-03 00:00:00]}

Here is another example, this time with a time point on a boundary,
2020-01-02 00:00:00.

To floor this to the nearest day, the following interval is constructed,
and the left-hand side is returned at day precision:

\code{[2020-01-02 00:00:00, 2020-01-03 00:00:00)}

To ceiling this to the nearest day, the following interval
is constructed, and the right-hand side is returned at day precision:

\code{(2020-01-01 00:00:00, 2020-01-02 00:00:00]}

Notice that, regardless of whether you are doing a floor or ceiling, if
the input falls on a boundary then it will be returned as is.
}

\examples{
library(magrittr)

x <- as_naive_time(year_month_day(2019, 01, 01))
x <- add_days(x, 0:40)
head(x)

# Floor by sets of 20 days
# The implicit origin to start the 20 day counter is 1970-01-01
time_point_floor(x, "day", n = 20)

# You can easily customize the origin by creating a duration out of the
# origin date of interest...
origin <- year_month_day(2019, 01, 01) \%>\%
  as_naive_time() \%>\%
  as_duration()

# Which you can subtract from, floor, and then add to your input
time_point_floor(x - origin, "day", n = 20) + origin

# For times on the boundary, floor and ceiling both return the input
# at the new precision. Notice how the first element is on the boundary,
# and the second is 1 second after the boundary.
y <- as_naive_time(year_month_day(2020, 01, 02, 00, 00, c(00, 01)))
time_point_floor(y, "day")
time_point_ceiling(y, "day")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-day.R
\name{year-month-day-setters}
\alias{year-month-day-setters}
\alias{set_year.clock_year_month_day}
\alias{set_month.clock_year_month_day}
\alias{set_day.clock_year_month_day}
\alias{set_hour.clock_year_month_day}
\alias{set_minute.clock_year_month_day}
\alias{set_second.clock_year_month_day}
\alias{set_millisecond.clock_year_month_day}
\alias{set_microsecond.clock_year_month_day}
\alias{set_nanosecond.clock_year_month_day}
\title{Setters: year-month-day}
\usage{
\method{set_year}{clock_year_month_day}(x, value, ...)

\method{set_month}{clock_year_month_day}(x, value, ...)

\method{set_day}{clock_year_month_day}(x, value, ...)

\method{set_hour}{clock_year_month_day}(x, value, ...)

\method{set_minute}{clock_year_month_day}(x, value, ...)

\method{set_second}{clock_year_month_day}(x, value, ...)

\method{set_millisecond}{clock_year_month_day}(x, value, ...)

\method{set_microsecond}{clock_year_month_day}(x, value, ...)

\method{set_nanosecond}{clock_year_month_day}(x, value, ...)
}
\arguments{
\item{x}{\verb{[clock_year_month_day]}

A year-month-day vector.}

\item{value}{\verb{[integer / "last"]}

The value to set the component to.

For \code{set_day()}, this can also be \code{"last"} to set the day to the
last day of the month.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} with the component set.
}
\description{
These are year-month-day methods for the
\link[=clock-setters]{setter generics}.
\itemize{
\item \code{set_year()} sets the Gregorian year.
\item \code{set_month()} sets the month of the year. Valid values are in the range
of \verb{[1, 12]}.
\item \code{set_day()} sets the day of the month. Valid values are in the range
of \verb{[1, 31]}.
\item There are sub-daily setters for setting more precise components.
}
}
\examples{
x <- year_month_day(2019, 1:3)

# Set the day
set_day(x, 12:14)

# Set to the "last" day of the month
set_day(x, "last")

# Set to an invalid day of the month
invalid <- set_day(x, 31)
invalid

# Then resolve the invalid day by choosing the next valid day
invalid_resolve(invalid, invalid = "next")

# Cannot set a component two levels more precise than where you currently are
try(set_hour(x, 5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{calendar_precision}
\alias{calendar_precision}
\title{Precision: calendar}
\usage{
calendar_precision(x)
}
\arguments{
\item{x}{\verb{[clock_calendar]}

A calendar.}
}
\value{
A single string holding the precision of the calendar.
}
\description{
\code{calendar_precision()} extracts the precision from a calendar object. It
returns the precision as a single string.
}
\examples{
calendar_precision(year_month_day(2019))
calendar_precision(year_month_day(2019, 1, 1))
calendar_precision(year_quarter_day(2019, 3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quarterly-year-quarter-day.R
\name{seq.clock_year_quarter_day}
\alias{seq.clock_year_quarter_day}
\title{Sequences: year-quarter-day}
\usage{
\method{seq}{clock_year_quarter_day}(from, to = NULL, by = NULL, length.out = NULL, along.with = NULL, ...)
}
\arguments{
\item{from}{\verb{[clock_year_quarter_day(1)]}

A \code{"year"} or \code{"quarter"} precision year-quarter-day to start the sequence
from.

\code{from} is always included in the result.}

\item{to}{\verb{[clock_year_quarter_day(1) / NULL]}

A \code{"year"} or \code{"quarter"} precision year-quarter-day to stop the sequence
at.

\code{to} is cast to the type of \code{from}.

\code{to} is only included in the result if the resulting sequence divides
the distance between \code{from} and \code{to} exactly.}

\item{by}{\verb{[integer(1) / clock_duration(1) / NULL]}

The unit to increment the sequence by.

If \code{by} is an integer, it is transformed into a duration with the
precision of \code{from}.

If \code{by} is a duration, it is cast to the type of \code{from}.}

\item{length.out}{\verb{[positive integer(1) / NULL]}

The length of the resulting sequence.

If specified, \code{along.with} must be \code{NULL}.}

\item{along.with}{\verb{[vector / NULL]}

A vector who's length determines the length of the resulting sequence.

Equivalent to \code{length.out = vec_size(along.with)}.

If specified, \code{length.out} must be \code{NULL}.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A sequence with the type of \code{from}.
}
\description{
This is a year-quarter-day method for the \code{\link[=seq]{seq()}} generic.

Sequences can only be generated for \code{"year"} and \code{"quarter"} precision
year-quarter-day vectors.

When calling \code{seq()}, exactly two of the following must be specified:
\itemize{
\item \code{to}
\item \code{by}
\item Either \code{length.out} or \code{along.with}
}
}
\examples{
# Quarterly sequence
x <- seq(year_quarter_day(2020, 1), year_quarter_day(2026, 3), by = 2)
x

# Which we can then set the day of the quarter of
set_day(x, "last")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-point.R
\name{seq.clock_time_point}
\alias{seq.clock_time_point}
\title{Sequences: time points}
\usage{
\method{seq}{clock_time_point}(from, to = NULL, by = NULL, length.out = NULL, along.with = NULL, ...)
}
\arguments{
\item{from}{\verb{[clock_sys_time(1) / clock_naive_time(1)]}

A time point to start the sequence from.

\code{from} is always included in the result.}

\item{to}{\verb{[clock_sys_time(1) / clock_naive_time(1) / NULL]}

A time point to stop the sequence at.

\code{to} is cast to the type of \code{from}.

\code{to} is only included in the result if the resulting sequence divides
the distance between \code{from} and \code{to} exactly.}

\item{by}{\verb{[integer(1) / clock_duration(1) / NULL]}

The unit to increment the sequence by.

If \code{by} is an integer, it is transformed into a duration with the
precision of \code{from}.

If \code{by} is a duration, it is cast to the type of \code{from}.}

\item{length.out}{\verb{[positive integer(1) / NULL]}

The length of the resulting sequence.

If specified, \code{along.with} must be \code{NULL}.}

\item{along.with}{\verb{[vector / NULL]}

A vector who's length determines the length of the resulting sequence.

Equivalent to \code{length.out = vec_size(along.with)}.

If specified, \code{length.out} must be \code{NULL}.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A sequence with the type of \code{from}.
}
\description{
This is a time point method for the \code{\link[=seq]{seq()}} generic. It works for sys-time
and naive-time vectors.

Sequences can be generated for all valid time point precisions (daily through
nanosecond).

When calling \code{seq()}, exactly two of the following must be specified:
\itemize{
\item \code{to}
\item \code{by}
\item Either \code{length.out} or \code{along.with}
}
}
\examples{
# Daily sequence
seq(
  as_naive_time(year_month_day(2019, 1, 1)),
  as_naive_time(year_month_day(2019, 2, 4)),
  by = 5
)

# Minutely sequence using minute precision naive-time
x <- as_naive_time(year_month_day(2019, 1, 2, 3, 3))
x

seq(x, by = 4, length.out = 10)

# You can use larger step sizes by using a duration-based `by`
seq(x, by = duration_days(1), length.out = 5)

# Nanosecond sequence
from <- as_naive_time(year_month_day(2019, 1, 1))
from <- time_point_cast(from, "nanosecond")
to <- from + 100
seq(from, to, by = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clock-deprecated.R
\name{date-zone}
\alias{date-zone}
\alias{date_zone}
\alias{date_set_zone}
\title{Get or set the time zone}
\usage{
date_zone(x)

date_set_zone(x, zone)
}
\arguments{
\item{x}{\verb{[POSIXct / POSIXlt]}

A date-time vector.}

\item{zone}{\verb{[character(1)]}

A valid time zone to switch to.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
\itemize{
\item \code{date_zone()} is deprecated in favor of \code{\link[=date_time_zone]{date_time_zone()}}.
\item \code{date_set_zone()} is deprecated in favor of \code{\link[=date_time_set_zone]{date_time_set_zone()}}.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{year-month-weekday-count-between}
\alias{year-month-weekday-count-between}
\alias{calendar_count_between.clock_year_month_weekday}
\title{Counting: year-month-weekday}
\usage{
\method{calendar_count_between}{clock_year_month_weekday}(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[clock_year_month_weekday]}

A pair of year-month-weekday vectors. These will be recycled to their
common size.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"quarter"}
\item \code{"month"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
This is a year-month-weekday method for the \code{\link[=calendar_count_between]{calendar_count_between()}}
generic. It counts the number of \code{precision} units between \code{start} and \code{end}
(i.e., the number of years or months).
}
\details{
Remember that year-month-weekday is not comparable when it is \code{"day"}
precision or finer, so this method is only defined for \code{"year"} and
\code{"month"} precision year-month-weekday objects.

\code{"quarter"} is equivalent to \code{"month"} precision with \code{n} set to \code{n * 3L}.
}
\examples{
# Compute the number of months between two dates
x <- year_month_weekday(2001, 2)
y <- year_month_weekday(2021, c(1, 3))

calendar_count_between(x, y, "month")

# Remember that day precision or finer year-month-weekday objects
# are not comparable, so this won't work
x <- year_month_weekday(2001, 2, 1, 1)
try(calendar_count_between(x, x, "month"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoned-time.R
\name{format.clock_zoned_time}
\alias{format.clock_zoned_time}
\title{Formatting: zoned-time}
\usage{
\method{format}{clock_zoned_time}(x, ..., format = NULL, locale = clock_locale(), abbreviate_zone = FALSE)
}
\arguments{
\item{x}{\verb{[clock_zoned_time]}

A zoned-time.}

\item{...}{\verb{[dots]}

Not used, but no error will be thrown if not empty to remain compatible
with usage of the \code{format()} generic.}

\item{format}{\verb{[character(1) / NULL]}

If \code{NULL}, a default format is used, which depends on the type of the
input.

Otherwise, a format string which is a combination of:

\strong{Year}
\itemize{
\item \verb{\%C}: The year divided by 100 using floored division. If the result
is a single decimal digit, it is prefixed with \code{0}.
\item \verb{\%y}: The last two decimal digits of the year. If the result is a single
digit it is prefixed by \code{0}.
\item \verb{\%Y}: The year as a decimal number. If the result is less than four
digits it is left-padded with \code{0} to four digits.
}

\strong{Month}
\itemize{
\item \verb{\%b}, \verb{\%h}: The \code{locale}'s abbreviated month name.
\item \verb{\%B}: The \code{locale}'s full month name.
\item \verb{\%m}: The month as a decimal number. January is \code{01}. If the result is a
single digit, it is prefixed with \code{0}.
}

\strong{Day}
\itemize{
\item \verb{\%d}: The day of month as a decimal number. If the result is a single
decimal digit, it is prefixed with \code{0}.
}

\strong{Day of the week}
\itemize{
\item \verb{\%a}: The \code{locale}'s abbreviated weekday name.
\item \verb{\%A}: The \code{locale}'s full weekday name.
\item \verb{\%w}: The weekday as a decimal number (\code{0-6}), where Sunday is \code{0}.
}

\strong{ISO 8601 week-based year}
\itemize{
\item \verb{\%g}: The last two decimal digits of the ISO week-based year. If the
result is a single digit it is prefixed by \code{0}.
\item \verb{\%G}: The ISO week-based year as a decimal number. If the result is less
than four digits it is left-padded with \code{0} to four digits.
\item \verb{\%V}: The ISO week-based week number as a decimal number. If the result
is a single digit, it is prefixed with \code{0}.
\item \verb{\%u}: The ISO weekday as a decimal number (\code{1-7}), where Monday is \code{1}.
}

\strong{Week of the year}
\itemize{
\item \verb{\%U}: The week number of the year as a decimal number. The first Sunday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. If the result is a single digit, it is prefixed with
\code{0}.
\item \verb{\%W}: The week number of the year as a decimal number. The first Monday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. If the result is a single digit, it is prefixed with
\code{0}.
}

\strong{Day of the year}
\itemize{
\item \verb{\%j}: The day of the year as a decimal number. January 1 is \code{001}. If the
result is less than three digits, it is left-padded with \code{0} to three
digits.
}

\strong{Date}
\itemize{
\item \verb{\%D}, \verb{\%x}: Equivalent to \verb{\%m/\%d/\%y}.
\item \verb{\%F}: Equivalent to \verb{\%Y-\%m-\%d}.
}

\strong{Time of day}
\itemize{
\item \verb{\%H}: The hour (24-hour clock) as a decimal number. If the result is a
single digit, it is prefixed with \code{0}.
\item \verb{\%I}: The hour (12-hour clock) as a decimal number. If the result is a
single digit, it is prefixed with \code{0}.
\item \verb{\%M}: The minute as a decimal number. If the result is a single digit, it
is prefixed with \code{0}.
\item \verb{\%S}: Seconds as a decimal number. Fractional seconds are printed at the
precision of the input. The character for the decimal point is localized
according to \code{locale}.
\item \verb{\%p}: The \code{locale}'s equivalent of the AM/PM designations associated with
a 12-hour clock.
\item \verb{\%R}: Equivalent to \verb{\%H:\%M}.
\item \verb{\%T}, \verb{\%X}: Equivalent to \verb{\%H:\%M:\%S}.
\item \verb{\%r}: Nearly equivalent to \verb{\%I:\%M:\%S \%p}, but seconds are always printed
at second precision.
}

\strong{Time zone}
\itemize{
\item \verb{\%z}: The offset from UTC in the ISO 8601 format. For example \code{-0430}
refers to 4 hours 30 minutes behind UTC. If the offset is zero, \code{+0000} is
used. The modified command \verb{\%Ez} inserts a \code{:} between the hour and
minutes, like \code{-04:30}.
\item \verb{\%Z}: The full time zone name. If \code{abbreviate_zone} is \code{TRUE}, the time
zone abbreviation.
}

\strong{Miscellaneous}
\itemize{
\item \verb{\%c}: A date and time representation. Similar to, but not exactly the
same as, \verb{\%a \%b \%d \%H:\%M:\%S \%Y}.
\item \code{\%\%}: A \verb{\%} character.
\item \verb{\%n}: A newline character.
\item \verb{\%t}: A horizontal-tab character.
}}

\item{locale}{\verb{[clock_locale]}

A locale object created from \code{\link[=clock_locale]{clock_locale()}}.}

\item{abbreviate_zone}{\verb{[logical(1)]}

If \code{TRUE}, \verb{\%Z} returns an abbreviated time zone name.

If \code{FALSE}, \verb{\%Z} returns the full time zone name.}
}
\value{
A character vector of the formatted input.
}
\description{
This is a zoned-time method for the \code{\link[=format]{format()}} generic.

This function allows you to format a zoned-time using a flexible \code{format}
string.

If \code{format} is \code{NULL}, a default format of \code{"\%Y-\%m-\%dT\%H:\%M:\%S\%Ez[\%Z]"} is
used. This matches the default format that \code{\link[=zoned_time_parse_complete]{zoned_time_parse_complete()}}
parses. Additionally, this format matches the de-facto standard extension to
RFC 3339 for creating completely unambiguous date-times.
}
\examples{
x <- year_month_day(2019, 1, 1)
x <- as_zoned_time(as_naive_time(x), "America/New_York")

format(x)
format(x, format = "\%B \%d, \%Y")
format(x, format = "\%B \%d, \%Y", locale = clock_locale("fr"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{duration-arithmetic}
\alias{duration-arithmetic}
\alias{add_years.clock_duration}
\alias{add_quarters.clock_duration}
\alias{add_months.clock_duration}
\alias{add_weeks.clock_duration}
\alias{add_days.clock_duration}
\alias{add_hours.clock_duration}
\alias{add_minutes.clock_duration}
\alias{add_seconds.clock_duration}
\alias{add_milliseconds.clock_duration}
\alias{add_microseconds.clock_duration}
\alias{add_nanoseconds.clock_duration}
\title{Arithmetic: duration}
\usage{
\method{add_years}{clock_duration}(x, n, ...)

\method{add_quarters}{clock_duration}(x, n, ...)

\method{add_months}{clock_duration}(x, n, ...)

\method{add_weeks}{clock_duration}(x, n, ...)

\method{add_days}{clock_duration}(x, n, ...)

\method{add_hours}{clock_duration}(x, n, ...)

\method{add_minutes}{clock_duration}(x, n, ...)

\method{add_seconds}{clock_duration}(x, n, ...)

\method{add_milliseconds}{clock_duration}(x, n, ...)

\method{add_microseconds}{clock_duration}(x, n, ...)

\method{add_nanoseconds}{clock_duration}(x, n, ...)
}
\arguments{
\item{x}{\verb{[clock_duration]}

A duration vector.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} after performing the arithmetic, possibly with a more precise
precision.
}
\description{
These are duration methods for the
\link[=clock-arithmetic]{arithmetic generics}.
\itemize{
\item \code{add_years()}
\item \code{add_quarters()}
\item \code{add_months()}
\item \code{add_weeks()}
\item \code{add_days()}
\item \code{add_hours()}
\item \code{add_minutes()}
\item \code{add_seconds()}
\item \code{add_milliseconds()}
\item \code{add_microseconds()}
\item \code{add_nanoseconds()}
}

When adding to a duration using one of these functions, a second duration
is created based on the function name and \code{n}. The two durations are then
added together, and the precision of the result is determined as the
\emph{more precise precision} of the two durations.
}
\details{
You can add calendrical durations to other calendrical durations,
and chronological durations to other chronological durations, but you can't
add a chronological duration to a calendrical duration
(such as adding days and months). For more information, see the
documentation on the \link[=duration-helper]{duration helper} page.

\code{x} and \code{n} are recycled against each other.
}
\examples{
x <- duration_seconds(5)

# Addition in the same precision
add_seconds(x, 1:10)

# Addition with days, defined as 86400 seconds
add_days(x, 1)

# Similarly, if you start with days and add seconds, you get the common
# precision of the two back, which is seconds
y <- duration_days(1)
add_seconds(y, 5)

# But you can't add a chronological duration (days) and
# a calendrical duration (months)
try(add_months(y, 1))

# You can add years to a duration of months, which adds
# an additional 12 months / year
z <- duration_months(5)
add_years(z, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quarterly-year-quarter-day.R
\name{year-quarter-day-getters}
\alias{year-quarter-day-getters}
\alias{get_year.clock_year_quarter_day}
\alias{get_quarter.clock_year_quarter_day}
\alias{get_day.clock_year_quarter_day}
\alias{get_hour.clock_year_quarter_day}
\alias{get_minute.clock_year_quarter_day}
\alias{get_second.clock_year_quarter_day}
\alias{get_millisecond.clock_year_quarter_day}
\alias{get_microsecond.clock_year_quarter_day}
\alias{get_nanosecond.clock_year_quarter_day}
\title{Getters: year-quarter-day}
\usage{
\method{get_year}{clock_year_quarter_day}(x)

\method{get_quarter}{clock_year_quarter_day}(x)

\method{get_day}{clock_year_quarter_day}(x)

\method{get_hour}{clock_year_quarter_day}(x)

\method{get_minute}{clock_year_quarter_day}(x)

\method{get_second}{clock_year_quarter_day}(x)

\method{get_millisecond}{clock_year_quarter_day}(x)

\method{get_microsecond}{clock_year_quarter_day}(x)

\method{get_nanosecond}{clock_year_quarter_day}(x)
}
\arguments{
\item{x}{\verb{[clock_year_quarter_day]}

A year-quarter-day to get the component from.}
}
\value{
The component.
}
\description{
These are year-quarter-day methods for the
\link[=clock-getters]{getter generics}.
\itemize{
\item \code{get_year()} returns the fiscal year. Note that this can differ from the
Gregorian year if \code{start != 1L}.
\item \code{get_quarter()} returns the fiscal quarter as a value between 1-4.
\item \code{get_day()} returns the day of the fiscal quarter as a value between 1-92.
\item There are sub-daily getters for extracting more precise components.
}
}
\examples{
x <- year_quarter_day(2020, 1:4)

get_quarter(x)

# Set and then get the last day of the quarter
x <- set_day(x, "last")
get_day(x)

# Start the fiscal year in November and choose the 50th day in
# each quarter of 2020
november <- 11
y <- year_quarter_day(2020, 1:4, 50, start = 11)
y

get_day(y)

# What does that map to in year-month-day?
as_year_month_day(y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-day.R
\name{year-month-day-group}
\alias{year-month-day-group}
\alias{calendar_group.clock_year_month_day}
\title{Grouping: year-month-day}
\usage{
\method{calendar_group}{clock_year_month_day}(x, precision, ..., n = 1L)
}
\arguments{
\item{x}{\verb{[clock_year_month_day]}

A year-month-day vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"month"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
\code{x} grouped at the specified \code{precision}.
}
\description{
This is a year-month-day method for the \code{\link[=calendar_group]{calendar_group()}} generic.

Grouping for a year-month-day object can be done at any precision, as
long as \code{x} is at least as precise as \code{precision}.
}
\examples{
steps <- duration_days(seq(0, 100, by = 5))
x <- year_month_day(2019, 1, 1)
x <- as_naive_time(x) + steps
x <- as_year_month_day(x)
x

# Group by a single month
calendar_group(x, "month")

# Or multiple months
calendar_group(x, "month", n = 2)

# Group 3 days of the month together
y <- year_month_day(2019, 1, 1:12)
calendar_group(y, "day", n = 3)

# Group by 5 nanosecond of the current second
z <- year_month_day(
  2019, 1, 2, 1, 5, 20, 1:20,
  subsecond_precision = "nanosecond"
)
calendar_group(z, "nanosecond", n = 5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{calendar_leap_year}
\alias{calendar_leap_year}
\title{Is the calendar year a leap year?}
\usage{
calendar_leap_year(x)
}
\arguments{
\item{x}{\verb{[calendar]}

A calendar type to detect leap years in.}
}
\value{
A logical vector the same size as \code{x}. Returns \code{TRUE} if in a leap
year, \code{FALSE} if not in a leap year, and \code{NA} if \code{x} is \code{NA}.
}
\description{
\code{calendar_leap_year()} detects if the calendar year is a leap year - i.e.
does it contain one or more extra components than other years?

A particular year is a leap year if:
\itemize{
\item \code{\link[=year_month_day]{year_month_day()}}: February has 29 days.
\item \code{\link[=year_month_weekday]{year_month_weekday()}}: February has a weekday that occurs 5 times.
\item \code{\link[=year_week_day]{year_week_day()}}: There are 53 weeks in the year, resulting in 371
days in the year.
\item \code{\link[=iso_year_week_day]{iso_year_week_day()}}: There are 53 weeks in the year, resulting in 371
days in the year.
\item \code{\link[=year_quarter_day]{year_quarter_day()}}: One of the quarters has 1 more day than normal (the
quarter with an extra day depends on the \code{start} used, but will always be
the same for a particular \code{start}). This aligns with Gregorian leap years
for all \code{start}s except February, in which case the leap year is always 1
year after the Gregorian leap year.
\item \code{\link[=year_day]{year_day()}}: There are 366 days in the year.
}
}
\examples{
x <- year_month_day(c(2019:2024, NA))
calendar_leap_year(x)

# For year-quarter-day, the leap year typically aligns with the Gregorian
# leap year, unless the `start` is February, in which case the leap year is
# always 1 year after the Gregorian leap year
x <- year_quarter_day(2020:2021, start = clock_months$january)
calendar_leap_year(x)

x <- year_quarter_day(2020:2021, start = clock_months$february)
calendar_leap_year(x)

# With a January start, 2020 has the extra day
get_day(year_quarter_day(2020, 1:4, "last", start = clock_months$january))
get_day(year_quarter_day(2021, 1:4, "last", start = clock_months$january))
get_day(year_quarter_day(2022, 1:4, "last", start = clock_months$january))

# With a February start, 2021 has the extra day
get_day(year_quarter_day(2020, 1:4, "last", start = clock_months$february))
get_day(year_quarter_day(2021, 1:4, "last", start = clock_months$february))
get_day(year_quarter_day(2022, 1:4, "last", start = clock_months$february))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quarterly-year-quarter-day.R
\name{year_quarter_day}
\alias{year_quarter_day}
\title{Calendar: year-quarter-day}
\usage{
year_quarter_day(
  year,
  quarter = NULL,
  day = NULL,
  hour = NULL,
  minute = NULL,
  second = NULL,
  subsecond = NULL,
  ...,
  start = NULL,
  subsecond_precision = NULL
)
}
\arguments{
\item{year}{\verb{[integer]}

The fiscal year. Values \verb{[-32767, 32767]} are generally allowed.}

\item{quarter}{\verb{[integer / NULL]}

The fiscal quarter. Values \verb{[1, 4]} are allowed.}

\item{day}{\verb{[integer / "last" / NULL]}

The day of the quarter. Values \verb{[1, 92]} are allowed.

If \code{"last"}, the last day of the quarter is returned.}

\item{hour}{\verb{[integer / NULL]}

The hour. Values \verb{[0, 23]} are allowed.}

\item{minute}{\verb{[integer / NULL]}

The minute. Values \verb{[0, 59]} are allowed.}

\item{second}{\verb{[integer / NULL]}

The second. Values \verb{[0, 59]} are allowed.}

\item{subsecond}{\verb{[integer / NULL]}

The subsecond. If specified, \code{subsecond_precision} must also be specified
to determine how to interpret the \code{subsecond}.

If using milliseconds, values \verb{[0, 999]} are allowed.

If using microseconds, values \verb{[0, 999999]} are allowed.

If using nanoseconds, values \verb{[0, 999999999]} are allowed.}

\item{...}{These dots are for future extensions and must be empty.}

\item{start}{\verb{[integer(1) / NULL]}

The month to start the fiscal year in. 1 = January and 12 = December.

If \code{NULL}, a \code{start} of January will be used.}

\item{subsecond_precision}{\verb{[character(1) / NULL]}

The precision to interpret \code{subsecond} as. One of: \code{"millisecond"},
\code{"microsecond"}, or \code{"nanosecond"}.}
}
\value{
A year-quarter-day calendar vector.
}
\description{
\code{year_quarter_day()} constructs a calendar from the fiscal year, fiscal
quarter, and day of the quarter, along with a value determining which
month the fiscal year \code{start}s in.
}
\details{
Fields are recycled against each other.

Fields are collected in order until the first \code{NULL} field is located. No
fields after the first \code{NULL} field are used.
}
\examples{
# Year-quarter type
x <- year_quarter_day(2019, 1:4)
x

add_quarters(x, 2)

# Set the day to the last day of the quarter
x <- set_day(x, "last")
x

# Start the fiscal year in June
june <- 6L
y <- year_quarter_day(2019, 1:4, "last", start = june)

# Compare the year-month-day values that result from having different
# fiscal year start months
as_year_month_day(x)
as_year_month_day(y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/naive-time.R
\name{as_naive_time}
\alias{as_naive_time}
\title{Convert to a naive-time}
\usage{
as_naive_time(x)
}
\arguments{
\item{x}{\verb{[object]}

An object to convert to a naive-time.}
}
\value{
A naive-time vector.
}
\description{
\code{as_naive_time()} converts \code{x} to a naive-time.

You can convert to a naive-time from any calendar type, as long as it has
at least day precision. There also must not be any invalid dates. If invalid
dates exist, they must first be resolved with \code{\link[=invalid_resolve]{invalid_resolve()}}.

Converting to a naive-time from a sys-time or zoned-time retains the printed
time, but drops the assumption that the time should be interpreted with any
specific time zone.

Converting to a naive-time from a duration just wraps the duration in a
naive-time object, there is no assumption about the time zone. The duration
must have at least day precision.

There are convenience methods for converting to a naive-time from R's
native date and date-time types. Like converting from a zoned-time, these
retain the printed time.
}
\examples{
x <- as.Date("2019-01-01")
as_naive_time(x)

ym <- year_month_day(2019, 02)

# A minimum of day precision is required
try(as_naive_time(ym))

ymd <- set_day(ym, 10)
as_naive_time(ymd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{date-time-parse}
\alias{date-time-parse}
\alias{date_time_parse}
\alias{date_time_parse_complete}
\alias{date_time_parse_abbrev}
\alias{date_time_parse_RFC_3339}
\title{Parsing: date-time}
\usage{
date_time_parse(
  x,
  zone,
  ...,
  format = NULL,
  locale = clock_locale(),
  nonexistent = NULL,
  ambiguous = NULL
)

date_time_parse_complete(x, ..., format = NULL, locale = clock_locale())

date_time_parse_abbrev(x, zone, ..., format = NULL, locale = clock_locale())

date_time_parse_RFC_3339(x, ..., separator = "T", offset = "Z")
}
\arguments{
\item{x}{\verb{[character]}

A character vector to parse.}

\item{zone}{\verb{[character(1)]}

A full time zone name.}

\item{...}{These dots are for future extensions and must be empty.}

\item{format}{\verb{[character / NULL]}

A format string. A combination of the following commands, or \code{NULL},
in which case a default format string is used.

A vector of multiple format strings can be supplied. They will be tried in
the order they are provided.

\strong{Year}
\itemize{
\item \verb{\%C}: The century as a decimal number. The modified command \verb{\%NC} where
\code{N} is a positive decimal integer specifies the maximum number of
characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%y}: The last two decimal digits of the year. If the century is not
otherwise specified (e.g. with \verb{\%C}), values in the range \verb{[69 - 99]} are
presumed to refer to the years \verb{[1969 - 1999]}, and values in the range
\verb{[00 - 68]} are presumed to refer to the years \verb{[2000 - 2068]}. The
modified command \verb{\%Ny}, where \code{N} is a positive decimal integer, specifies
the maximum number of characters to read. If not specified, the default is
\code{2}. Leading zeroes are permitted but not required.
\item \verb{\%Y}: The year as a decimal number. The modified command \verb{\%NY} where \code{N}
is a positive decimal integer specifies the maximum number of characters to
read. If not specified, the default is \code{4}. Leading zeroes are permitted
but not required.
}

\strong{Month}
\itemize{
\item \verb{\%b}, \verb{\%B}, \verb{\%h}: The \code{locale}'s full or abbreviated case-insensitive
month name.
\item \verb{\%m}: The month as a decimal number. January is \code{1}. The modified command
\verb{\%Nm} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
}

\strong{Day}
\itemize{
\item \verb{\%d}, \verb{\%e}: The day of the month as a decimal number. The modified
command \verb{\%Nd} where \code{N} is a positive decimal integer specifies the maximum
number of characters to read. If not specified, the default is \code{2}. Leading
zeroes are permitted but not required.
}

\strong{Day of the week}
\itemize{
\item \verb{\%a}, \verb{\%A}: The \code{locale}'s full or abbreviated case-insensitive weekday
name.
\item \verb{\%w}: The weekday as a decimal number (\code{0-6}), where Sunday is \code{0}. The
modified command \verb{\%Nw} where \code{N} is a positive decimal integer specifies
the maximum number of characters to read. If not specified, the default is
\code{1}. Leading zeroes are permitted but not required.
}

\strong{ISO 8601 week-based year}
\itemize{
\item \verb{\%g}: The last two decimal digits of the ISO week-based year. The
modified command \verb{\%Ng} where \code{N} is a positive decimal integer specifies
the maximum number of characters to read. If not specified, the default is
\code{2}. Leading zeroes are permitted but not required.
\item \verb{\%G}: The ISO week-based year as a decimal number. The modified command
\verb{\%NG} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{4}. Leading zeroes
are permitted but not required.
\item \verb{\%V}: The ISO week-based week number as a decimal number. The modified
command \verb{\%NV} where \code{N} is a positive decimal integer specifies the maximum
number of characters to read. If not specified, the default is \code{2}. Leading
zeroes are permitted but not required.
\item \verb{\%u}: The ISO weekday as a decimal number (\code{1-7}), where Monday is \code{1}.
The modified command \verb{\%Nu} where \code{N} is a positive decimal integer
specifies the maximum number of characters to read. If not specified, the
default is \code{1}. Leading zeroes are permitted but not required.
}

\strong{Week of the year}
\itemize{
\item \verb{\%U}: The week number of the year as a decimal number. The first Sunday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. The modified command \verb{\%NU} where \code{N} is a positive
decimal integer specifies the maximum number of characters to read. If not
specified, the default is \code{2}. Leading zeroes are permitted but not
required.
\item \verb{\%W}: The week number of the year as a decimal number. The first Monday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. The modified command \verb{\%NW} where \code{N} is a positive
decimal integer specifies the maximum number of characters to read. If not
specified, the default is \code{2}. Leading zeroes are permitted but not
required.
}

\strong{Day of the year}
\itemize{
\item \verb{\%j}: The day of the year as a decimal number. January 1 is \code{1}. The
modified command \verb{\%Nj} where \code{N} is a positive decimal integer specifies
the maximum number of characters to read. If not specified, the default is
\code{3}. Leading zeroes are permitted but not required.
}

\strong{Date}
\itemize{
\item \verb{\%D}, \verb{\%x}: Equivalent to \verb{\%m/\%d/\%y}.
\item \verb{\%F}: Equivalent to \verb{\%Y-\%m-\%d}. If modified with a width (like \verb{\%NF}),
the width is applied to only \verb{\%Y}.
}

\strong{Time of day}
\itemize{
\item \verb{\%H}: The hour (24-hour clock) as a decimal number. The modified command
\verb{\%NH} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%I}: The hour (12-hour clock) as a decimal number. The modified command
\verb{\%NI} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%M}: The minutes as a decimal number. The modified command \verb{\%NM} where
\code{N} is a positive decimal integer specifies the maximum number of
characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%S}: The seconds as a decimal number. Leading zeroes are permitted but
not required. If encountered, the \code{locale} determines the decimal point
character. Generally, the maximum number of characters to read is
determined by the precision that you are parsing at. For example, a
precision of \code{"second"} would read a maximum of 2 characters, while a
precision of \code{"millisecond"} would read a maximum of 6 (2 for the values
before the decimal point, 1 for the decimal point, and 3 for the values
after it). The modified command \verb{\%NS}, where \code{N} is a positive decimal
integer, can be used to exactly specify the maximum number of characters to
read. This is only useful if you happen to have seconds with more than 1
leading zero.
\item \verb{\%p}: The \code{locale}'s equivalent of the AM/PM designations associated with
a 12-hour clock. The command \verb{\%I} must precede \verb{\%p} in the format string.
\item \verb{\%R}: Equivalent to \verb{\%H:\%M}.
\item \verb{\%T}, \verb{\%X}: Equivalent to \verb{\%H:\%M:\%S}.
\item \verb{\%r}: Equivalent to \verb{\%I:\%M:\%S \%p}.
}

\strong{Time zone}
\itemize{
\item \verb{\%z}: The offset from UTC in the format \verb{[+|-]hh[mm]}. For example
\code{-0430} refers to 4 hours 30 minutes behind UTC. And \code{04} refers to 4 hours
ahead of UTC. The modified command \verb{\%Ez} parses a \code{:} between the hours and
minutes and leading zeroes on the hour field are optional:
\verb{[+|-]h[h][:mm]}. For example \code{-04:30} refers to 4 hours 30 minutes behind
UTC. And \code{4} refers to 4 hours ahead of UTC.
\item \verb{\%Z}: The full time zone name or the time zone abbreviation, depending on
the function being used. A single word is parsed. This word can only
contain characters that are alphanumeric, or one of \code{'_'}, \code{'/'}, \code{'-'} or
\code{'+'}.
}

\strong{Miscellaneous}
\itemize{
\item \verb{\%c}: A date and time representation. Equivalent to
\verb{\%a \%b \%d \%H:\%M:\%S \%Y}.
\item \code{\%\%}: A \verb{\%} character.
\item \verb{\%n}: Matches one white space character. \verb{\%n}, \verb{\%t}, and a space can be
combined to match a wide range of white-space patterns. For example \code{"\%n "}
matches one or more white space characters, and \code{"\%n\%t\%t"} matches one to
three white space characters.
\item \verb{\%t}: Matches zero or one white space characters.
}}

\item{locale}{\verb{[clock_locale]}

A locale object created from \code{\link[=clock_locale]{clock_locale()}}.}

\item{nonexistent}{\verb{[character / NULL]}

One of the following nonexistent time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"roll-forward"}: The next valid instant in time.
\item \code{"roll-backward"}: The previous valid instant in time.
\item \code{"shift-forward"}: Shift the nonexistent time forward by the size of
the daylight saving time gap.
\item \verb{"shift-backward}: Shift the nonexistent time backward by the size of
the daylight saving time gap.
\item \code{"NA"}: Replace nonexistent times with \code{NA}.
\item \code{"error"}: Error on nonexistent times.
}

Using either \code{"roll-forward"} or \code{"roll-backward"} is generally
recommended over shifting, as these two strategies maintain the
\emph{relative ordering} between elements of the input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{nonexistent} must be supplied
and cannot be \code{NULL}. This is a convenient way to make production code
robust to nonexistent times.}

\item{ambiguous}{\verb{[character / zoned_time / POSIXct / list(2) / NULL]}

One of the following ambiguous time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"earliest"}: Of the two possible times, choose the earliest one.
\item \code{"latest"}: Of the two possible times, choose the latest one.
\item \code{"NA"}: Replace ambiguous times with \code{NA}.
\item \code{"error"}: Error on ambiguous times.
}

Alternatively, \code{ambiguous} is allowed to be a zoned_time (or POSIXct) that
is either length 1, or the same length as the input. If an ambiguous time
is encountered, the zoned_time is consulted. If the zoned_time corresponds
to a naive_time that is also ambiguous \emph{and} uses the same daylight saving
time transition point as the original ambiguous time, then the offset of
the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
resolved by consulting the zoned_time, then this method falls back to
\code{NULL}.

Finally, \code{ambiguous} is allowed to be a list of size 2, where the first
element of the list is a zoned_time (as described above), and the second
element of the list is an ambiguous time resolution strategy to use when
the ambiguous time cannot be resolved by consulting the zoned_time.
Specifying a zoned_time on its own is identical to \verb{list(<zoned_time>, NULL)}.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{ambiguous} must be supplied and
cannot be \code{NULL}. Additionally, \code{ambiguous} cannot be specified as a
zoned_time on its own, as this implies \code{NULL} for ambiguous times that the
zoned_time cannot resolve. Instead, it must be specified as a list
alongside an ambiguous time resolution strategy as described above. This is
a convenient way to make production code robust to ambiguous times.}

\item{separator}{\verb{[character(1)]}

The separator between the date and time components of the string. One of:
\itemize{
\item \code{"T"}
\item \code{"t"}
\item \code{" "}
}}

\item{offset}{\verb{[character(1)]}

The format of the offset from UTC contained in the string. One of:
\itemize{
\item \code{"Z"}
\item \code{"z"}
\item \code{"\%z"} to parse a numeric offset of the form \code{"+0430"}
\item \code{"\%Ez"} to parse a numeric offset of the form \code{"+04:30"}
}}
}
\value{
A POSIXct.
}
\description{
There are four parsers for parsing strings into POSIXct date-times,
\code{date_time_parse()}, \code{date_time_parse_complete()},
\code{date_time_parse_abbrev()}, and \code{date_time_parse_RFC_3339()}.
\subsection{date_time_parse()}{

\code{date_time_parse()} is useful for strings like \code{"2019-01-01 00:00:00"}, where
the UTC offset and full time zone name are not present in the string. The
string is first parsed as a naive-time without any time zone assumptions, and
is then converted to a POSIXct with the supplied \code{zone}.

Because converting from naive-time to POSIXct may result in nonexistent or
ambiguous times due to daylight saving time, these must be resolved
explicitly with the \code{nonexistent} and \code{ambiguous} arguments.

\code{date_time_parse()} completely ignores the \verb{\%z} and \verb{\%Z} commands. The only
time zone specific information that is used is the \code{zone}.

The default \code{format} used is \code{"\%Y-\%m-\%d \%H:\%M:\%S"}. This matches the default
result from calling \code{format()} on a POSIXct date-time.
}

\subsection{date_time_parse_complete()}{

\code{date_time_parse_complete()} is a parser for \emph{complete} date-time strings,
like \code{"2019-01-01T00:00:00-05:00[America/New_York]"}. A complete date-time
string has both the time zone offset and full time zone name in the string,
which is the only way for the string itself to contain all of the information
required to unambiguously construct a zoned-time. Because of this,
\code{date_time_parse_complete()} requires both the \verb{\%z} and \verb{\%Z} commands to be
supplied in the \code{format} string.

The default \code{format} used is \code{"\%Y-\%m-\%dT\%H:\%M:\%S\%Ez[\%Z]"}. This matches the
default result from calling \code{date_format()} on a POSIXct date-time.
Additionally, this format matches the de-facto standard extension to RFC 3339
for creating completely unambiguous date-times.
}

\subsection{date_time_parse_abbrev()}{

\code{date_time_parse_abbrev()} is a parser for date-time strings containing only
a time zone abbreviation, like \code{"2019-01-01 00:00:00 EST"}. The time zone
abbreviation is not enough to identify the full time zone name that the
date-time belongs to, so the full time zone name must be supplied as the
\code{zone} argument. However, the time zone abbreviation can help with resolving
ambiguity around daylight saving time fallbacks.

For \code{date_time_parse_abbrev()}, \verb{\%Z} must be supplied and is interpreted as
the time zone abbreviation rather than the full time zone name.

If used, the \verb{\%z} command must parse correctly, but its value will be
completely ignored.

The default \code{format} used is \code{"\%Y-\%m-\%d \%H:\%M:\%S \%Z"}. This matches the
default result from calling \code{print()} or \code{format(usetz = TRUE)} on a POSIXct
date-time.
}

\subsection{date_time_parse_RFC_3339()}{

\code{date_time_parse_RFC_3339()} is a parser for date-time strings in the
extremely common date-time format outlined by \href{https://datatracker.ietf.org/doc/html/rfc3339}{RFC 3339}. This document outlines
a profile of the ISO 8601 format that is even more restrictive, but
corresponds to the most common formats that are likely to be used in
internet protocols (i.e. through APIs).

In particular, this function is intended to parse the following three
formats:

\if{html}{\out{<div class="sourceCode">}}\preformatted{2019-01-01T00:00:00Z
2019-01-01T00:00:00+0430
2019-01-01T00:00:00+04:30
}\if{html}{\out{</div>}}

This function defaults to parsing the first of these formats by using
a format string of \code{"\%Y-\%m-\%dT\%H:\%M:\%SZ"}.

If your date-time strings use offsets from UTC rather than \code{"Z"}, then set
\code{offset} to one of the following:
\itemize{
\item \code{"\%z"} if the offset is of the form \code{"+0430"}.
\item \code{"\%Ez"} if the offset is of the form \code{"+04:30"}.
}

The RFC 3339 standard allows for replacing the \code{"T"} with a \code{"t"} or a space
(\code{" "}). Set \code{separator} to adjust this as needed.

The date-times returned by this function will always be in the UTC time zone.
}
}
\details{
If \code{date_time_parse_complete()} is given input that is length zero, all
\code{NA}s, or completely fails to parse, then no time zone will be able to be
determined. In that case, the result will use \code{"UTC"}.

If you have strings with sub-second components, then these date-time parsers
are not appropriate for you. Remember that clock treats POSIXct as a second
precision type, so parsing a string with fractional seconds directly into a
POSIXct is ambiguous and undefined. Instead, fully parse the string,
including its fractional seconds, into a clock type that can handle it, such
as a naive-time with \code{\link[=naive_time_parse]{naive_time_parse()}}, then round to seconds with
whatever rounding convention is appropriate for your use case, such as
\code{\link[=time_point_floor]{time_point_floor()}}, and finally convert that to POSIXct with
\code{\link[=as_date_time]{as_date_time()}}. This gives you complete control over how the fractional
seconds are handled when converting to POSIXct.
}
\examples{
# Parse with a known `zone`, even though that information isn't in the string
date_time_parse("2020-01-01 05:06:07", "America/New_York")

# Same time as above, except this is a completely unambiguous parse that
# doesn't require a `zone` argument, because the zone name and offset are
# both present in the string
date_time_parse_complete("2020-01-01T05:06:07-05:00[America/New_York]")

# Only day components
date_time_parse("2020-01-01", "America/New_York", format = "\%Y-\%m-\%d")

# `date_time_parse()` may have issues with ambiguous times due to daylight
# saving time fallbacks. For example, there were two 1'oclock hours here:
x <- date_time_parse("1970-10-25 00:59:59", "America/New_York")

# First (earliest) 1'oclock hour
add_seconds(x, 1)
# Second (latest) 1'oclock hour
add_seconds(x, 3601)

# If you try to parse this ambiguous time directly, you'll get an error:
ambiguous_time <- "1970-10-25 01:00:00"
try(date_time_parse(ambiguous_time, "America/New_York"))

# Resolve it by specifying whether you'd like to use the
# `earliest` or `latest` of the two possible times
date_time_parse(ambiguous_time, "America/New_York", ambiguous = "earliest")
date_time_parse(ambiguous_time, "America/New_York", ambiguous = "latest")

# `date_time_parse_complete()` doesn't have these issues, as it requires
# that the offset and zone name are both in the string, which resolves
# the ambiguity
complete_times <- c(
  "1970-10-25T01:00:00-04:00[America/New_York]",
  "1970-10-25T01:00:00-05:00[America/New_York]"
)
date_time_parse_complete(complete_times)

# `date_time_parse_abbrev()` also doesn't have these issues, since it
# uses the time zone abbreviation name to resolve the ambiguity
abbrev_times <- c(
  "1970-10-25 01:00:00 EDT",
  "1970-10-25 01:00:00 EST"
)
date_time_parse_abbrev(abbrev_times, "America/New_York")

# ---------------------------------------------------------------------------
# RFC 3339

# Typical UTC format
x <- "2019-01-01T00:01:02Z"
date_time_parse_RFC_3339(x)

# With a UTC offset containing a `:`
x <- "2019-01-01T00:01:02+02:30"
date_time_parse_RFC_3339(x, offset = "\%Ez")

# With a space between the date and time and no `:` in the offset
x <- "2019-01-01 00:01:02+0230"
date_time_parse_RFC_3339(x, separator = " ", offset = "\%z")

# ---------------------------------------------------------------------------
# Sub-second components

# If you have a string with sub-second components, but only require up to
# seconds, first parse them into a clock type that can handle sub-seconds to
# fully capture that information, then round using whatever convention is
# required for your use case before converting to a date-time.
x <- c("2019-01-01T00:00:01.1", "2019-01-01T00:00:01.78")

x <- naive_time_parse(x, precision = "millisecond")
x

time_point_floor(x, "second")
time_point_round(x, "second")

as_date_time(time_point_round(x, "second"), "America/New_York")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{is_year_month_weekday}
\alias{is_year_month_weekday}
\title{Is \code{x} a year-month-weekday?}
\usage{
is_year_month_weekday(x)
}
\arguments{
\item{x}{\verb{[object]}

An object.}
}
\value{
Returns \code{TRUE} if \code{x} inherits from \code{"clock_year_month_weekday"},
otherwise returns \code{FALSE}.
}
\description{
Check if \code{x} is a year-month-weekday.
}
\examples{
is_year_month_weekday(year_month_weekday(2019))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{year-month-weekday-setters}
\alias{year-month-weekday-setters}
\alias{set_year.clock_year_month_weekday}
\alias{set_month.clock_year_month_weekday}
\alias{set_day.clock_year_month_weekday}
\alias{set_index.clock_year_month_weekday}
\alias{set_hour.clock_year_month_weekday}
\alias{set_minute.clock_year_month_weekday}
\alias{set_second.clock_year_month_weekday}
\alias{set_millisecond.clock_year_month_weekday}
\alias{set_microsecond.clock_year_month_weekday}
\alias{set_nanosecond.clock_year_month_weekday}
\title{Setters: year-month-weekday}
\usage{
\method{set_year}{clock_year_month_weekday}(x, value, ...)

\method{set_month}{clock_year_month_weekday}(x, value, ...)

\method{set_day}{clock_year_month_weekday}(x, value, ..., index = NULL)

\method{set_index}{clock_year_month_weekday}(x, value, ...)

\method{set_hour}{clock_year_month_weekday}(x, value, ...)

\method{set_minute}{clock_year_month_weekday}(x, value, ...)

\method{set_second}{clock_year_month_weekday}(x, value, ...)

\method{set_millisecond}{clock_year_month_weekday}(x, value, ...)

\method{set_microsecond}{clock_year_month_weekday}(x, value, ...)

\method{set_nanosecond}{clock_year_month_weekday}(x, value, ...)
}
\arguments{
\item{x}{\verb{[clock_year_month_weekday]}

A year-month-weekday vector.}

\item{value}{\verb{[integer / "last"]}

The value to set the component to.

For \code{set_index()}, this can also be \code{"last"} to adjust to the last
instance of the corresponding weekday in that month.}

\item{...}{These dots are for future extensions and must be empty.}

\item{index}{\verb{[NULL / integer / "last"]}

This argument is only used with \code{set_day()}, and allows you to set the
index while also setting the weekday.

If \code{x} is a month precision year-month-weekday, \code{index} is required to
be set, as you must specify the weekday and the index simultaneously to
promote from month to day precision.}
}
\value{
\code{x} with the component set.
}
\description{
These are year-month-weekday methods for the
\link[=clock-setters]{setter generics}.
\itemize{
\item \code{set_year()} sets the Gregorian year.
\item \code{set_month()} sets the month of the year. Valid values are in the range
of \verb{[1, 12]}.
\item \code{set_day()} sets the day of the week. Valid values are in the range of
\verb{[1, 7]}, with 1 = Sunday, and 7 = Saturday.
\item \code{set_index()} sets the index indicating that the corresponding
weekday is the n-th instance of that weekday in the current month. Valid
values are in the range of \verb{[1, 5]}.
\item There are sub-daily setters for setting more precise components.
}
}
\examples{
x <- year_month_weekday(2019, 1:3)

set_year(x, 2020:2022)

# Setting the weekday on a month precision year-month-weekday requires
# also setting the `index` to fully specify the day information
x <- set_day(x, clock_weekdays$sunday, index = 1)
x

# Once you have at least day precision, you can set the weekday and
# the index separately
set_day(x, clock_weekdays$monday)
set_index(x, 3)

# Set to the "last" instance of the corresponding weekday in this month
# (Note that some months have 4 Sundays, and others have 5)
set_index(x, "last")

# Set to an invalid index
# January and February of 2019 don't have 5 Sundays!
invalid <- set_index(x, 5)
invalid

# Resolve the invalid dates by choosing the previous/next valid moment
invalid_resolve(invalid, invalid = "previous")
invalid_resolve(invalid, invalid = "next")

# You can also "overflow" the index. This keeps the weekday, but resets
# the index to 1 and increments the month value by 1.
invalid_resolve(invalid, invalid = "overflow")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date-shifting}
\alias{date-shifting}
\alias{date_shift.Date}
\title{Shifting: date}
\usage{
\method{date_shift}{Date}(x, target, ..., which = "next", boundary = "keep")
}
\arguments{
\item{x}{\verb{[Date]}

A date vector.}

\item{target}{\verb{[weekday]}

A weekday created from \code{\link[=weekday]{weekday()}} to target.

Generally this is length 1, but can also be the same length as \code{x}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{which}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"next"}: Shift to the next instance of the \code{target} weekday.
\item \verb{"previous}: Shift to the previous instance of the \code{target} weekday.
}}

\item{boundary}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"keep"}: If \code{x} is currently on the \code{target} weekday, return it.
\item \code{"advance"}: If \code{x} is currently on the \code{target} weekday, advance it
anyways.
}}
}
\value{
\code{x} shifted to the \code{target} weekday.
}
\description{
\code{date_shift()} shifts \code{x} to the \code{target} weekday. You can shift to the next
or previous weekday. If \code{x} is currently on the \code{target} weekday, you can
choose to leave it alone or advance it to the next instance of the \code{target}.

Weekday shifting is one of the easiest ways to floor by week while
controlling what is considered the first day of the week. You can also
accomplish this with the \code{origin} argument of \code{\link[=date_floor]{date_floor()}}, but this is
slightly easier.
}
\examples{
x <- as.Date("2019-01-01") + 0:1

# A Tuesday and Wednesday
as_weekday(x)

monday <- weekday(clock_weekdays$monday)

# Shift to the next Monday
date_shift(x, monday)

# Shift to the previous Monday
# This is an easy way to "floor by week" with a target weekday in mind
date_shift(x, monday, which = "previous")

# What about Tuesday?
tuesday <- weekday(clock_weekdays$tuesday)

# Notice that the day that was currently on a Tuesday was not shifted
date_shift(x, tuesday)

# You can force it to `"advance"`
date_shift(x, tuesday, boundary = "advance")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{calendar_widen}
\alias{calendar_widen}
\title{Widen a calendar to a more precise precision}
\usage{
calendar_widen(x, precision)
}
\arguments{
\item{x}{\verb{[calendar]}

A calendar vector.}

\item{precision}{\verb{[character(1)]}

A precision. Allowed precisions are dependent on the calendar used.}
}
\value{
\code{x} widened to the supplied \code{precision}.
}
\description{
\code{calendar_widen()} widens \code{x} to the specified \code{precision}. It does so
by setting new components to their smallest value.

Each calendar has its own help page describing the precisions that you
can widen to:
\itemize{
\item \link[=year-month-day-widen]{year-month-day}
\item \link[=year-month-weekday-widen]{year-month-weekday}
\item \link[=iso-year-week-day-widen]{iso-year-week-day}
\item \link[=year-quarter-day-widen]{year-quarter-day}
\item \link[=year-day-widen]{year-day}
}
}
\details{
A subsecond precision \code{x} cannot be widened. You cannot widen from, say,
\code{"millisecond"} to \code{"nanosecond"} precision. clock operates under the
philosophy that once you have set the subsecond precision of a calendar,
it is "locked in" at that precision. If you expected this to multiply
the milliseconds by 1e6 to get to nanosecond precision, you probably
want to convert to a time point first, and use \code{\link[=time_point_cast]{time_point_cast()}}.

Generally, clock treats calendars at a specific precision as a \emph{range} of
values. For example, a month precision year-month-day is treated as a range
over \verb{[yyyy-mm-01, yyyy-mm-last]}, with no assumption about the day of the
month. However, occasionally it is useful to quickly widen a calendar,
assuming that you want the beginning of this range to be used for each
component. This is where \code{calendar_widen()} can come in handy.
}
\examples{
# Month precision
x <- year_month_day(2019, 1)
x

# Widen to day precision
calendar_widen(x, "day")

# Or second precision
calendar_widen(x, "second")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date_build}
\alias{date_build}
\title{Building: date}
\usage{
date_build(year, month = 1L, day = 1L, ..., invalid = NULL)
}
\arguments{
\item{year}{\verb{[integer]}

The year. Values \verb{[-32767, 32767]} are generally allowed.}

\item{month}{\verb{[integer]}

The month. Values \verb{[1, 12]} are allowed.}

\item{day}{\verb{[integer / "last"]}

The day of the month. Values \verb{[1, 31]} are allowed.

If \code{"last"}, then the last day of the month is returned.}

\item{...}{These dots are for future extensions and must be empty.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}
}
\value{
A Date.
}
\description{
\code{date_build()} builds a Date from it's individual components.
}
\details{
Components are recycled against each other.
}
\examples{
date_build(2019)
date_build(2019, 1:3)

# Generating invalid dates will trigger an error
try(date_build(2019, 1:12, 31))

# You can resolve this with `invalid`
date_build(2019, 1:12, 31, invalid = "previous")

# But this particular case (the last day of the month) is better
# specified as:
date_build(2019, 1:12, "last")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-day.R,
%   R/gregorian-year-month-day.R, R/gregorian-year-month-weekday.R,
%   R/iso-year-week-day.R, R/naive-time.R, R/quarterly-year-quarter-day.R,
%   R/sys-time.R, R/weekday.R
\name{vec_arith.clock_year_day}
\alias{vec_arith.clock_year_day}
\alias{clock-arith}
\alias{vec_arith.clock_year_month_day}
\alias{vec_arith.clock_year_month_weekday}
\alias{vec_arith.clock_iso_year_week_day}
\alias{vec_arith.clock_naive_time}
\alias{vec_arith.clock_year_quarter_day}
\alias{vec_arith.clock_sys_time}
\alias{vec_arith.clock_weekday}
\title{Support for vctrs arithmetic}
\usage{
\method{vec_arith}{clock_year_day}(op, x, y, ...)

\method{vec_arith}{clock_year_month_day}(op, x, y, ...)

\method{vec_arith}{clock_year_month_weekday}(op, x, y, ...)

\method{vec_arith}{clock_iso_year_week_day}(op, x, y, ...)

\method{vec_arith}{clock_naive_time}(op, x, y, ...)

\method{vec_arith}{clock_year_quarter_day}(op, x, y, ...)

\method{vec_arith}{clock_sys_time}(op, x, y, ...)

\method{vec_arith}{clock_weekday}(op, x, y, ...)
}
\arguments{
\item{op}{An arithmetic operator as a string}

\item{x, y}{A pair of vectors. For \code{!}, unary \code{+} and unary \code{-}, \code{y} will be
a sentinel object of class \code{MISSING}, as created by \code{MISSING()}.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
The result of the arithmetic operation.
}
\description{
Support for vctrs arithmetic
}
\examples{
vctrs::vec_arith("+", year_month_day(2019), 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quarterly-year-quarter-day.R
\name{year-quarter-day-boundary}
\alias{year-quarter-day-boundary}
\alias{calendar_start.clock_year_quarter_day}
\alias{calendar_end.clock_year_quarter_day}
\title{Boundaries: year-quarter-day}
\usage{
\method{calendar_start}{clock_year_quarter_day}(x, precision)

\method{calendar_end}{clock_year_quarter_day}(x, precision)
}
\arguments{
\item{x}{\verb{[clock_year_quarter_day]}

A year-quarter-day vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"quarter"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}
}
\value{
\code{x} at the same precision, but with some components altered to be
at the boundary value.
}
\description{
This is a year-quarter-day method for the \code{\link[=calendar_start]{calendar_start()}} and
\code{\link[=calendar_end]{calendar_end()}} generics. They adjust components of a calendar to the
start or end of a specified \code{precision}.
}
\examples{
x <- year_quarter_day(2019:2020, 2:3, 5, 6, 7, 8, start = clock_months$march)
x

# Compute the last moment of the fiscal quarter
calendar_end(x, "quarter")

# Compare that to just setting the day to `"last"`,
# which doesn't affect the other components
set_day(x, "last")

# Compute the start of the fiscal year
calendar_start(x, "year")

as_date(calendar_start(x, "year"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoned-time.R
\name{zoned_time_now}
\alias{zoned_time_now}
\title{What is the current zoned-time?}
\usage{
zoned_time_now(zone)
}
\arguments{
\item{zone}{\verb{[character(1)]}

A time zone to get the current time for.}
}
\value{
A zoned-time of the current time.
}
\description{
\code{zoned_time_now()} returns the current time in the corresponding \code{zone}. It
is a wrapper around \code{\link[=sys_time_now]{sys_time_now()}} that attaches the time zone.
}
\details{
The time is returned with a nanosecond precision, but the actual amount
of data returned is OS dependent. Usually, information at at least the
microsecond level is returned, with some platforms returning nanosecond
information.
}
\examples{
x <- zoned_time_now("America/New_York")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-year-week-day.R
\name{iso-year-week-day-count-between}
\alias{iso-year-week-day-count-between}
\alias{calendar_count_between.clock_iso_year_week_day}
\title{Counting: iso-year-week-day}
\usage{
\method{calendar_count_between}{clock_iso_year_week_day}(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[clock_iso_year_week_day]}

A pair of iso-year-week-day vectors. These will be recycled to their
common size.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
This is an iso-year-week-day method for the \code{\link[=calendar_count_between]{calendar_count_between()}}
generic. It counts the number of \code{precision} units between \code{start} and \code{end}
(i.e., the number of ISO years).
}
\examples{
# Compute the number of whole ISO years between two dates
x <- iso_year_week_day(2001, 1, 2)
y <- iso_year_week_day(2021, 1, c(1, 3))
calendar_count_between(x, y, "year")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date_spanning_seq}
\alias{date_spanning_seq}
\title{Spanning sequence: date and date-time}
\usage{
date_spanning_seq(x)
}
\arguments{
\item{x}{\verb{[Date / POSIXct / POSIXlt]}

A date or date-time vector.}
}
\value{
A sequence along \verb{[min(x), max(x)]}.
}
\description{
\code{date_spanning_seq()} generates a regular sequence along the span of
\code{x}, i.e. along \verb{[min(x), max(x)]}. For dates, this generates a day precision
sequence, and for date-times it generates a second precision sequence.
}
\details{
Missing and infinite values are automatically removed before the sequence is
generated.

For date-times, sys-time based sequences are generated, consistent with
\code{\link[=posixt-sequence]{date_seq()}} when using a second precision \code{by} value.

If you need more precise sequence generation, call \code{\link[=range]{range()}} and \code{\link[=date_seq]{date_seq()}}
directly.
}
\examples{
x <- date_build(2020, c(1, 2, 1), c(10, 5, 12))
date_spanning_seq(x)

# Missing and infinite dates are removed before the sequence is generated
x <- c(x, NA, Inf, -Inf)
x

date_spanning_seq(x)

# For date-times, sequences are generated at second precision
x <- date_time_build(
  2020, 1, 2, 3, c(5, 4, 5), c(10, 48, 12),
  zone = "America/New_York"
)
x

date_spanning_seq(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-year-week-day.R
\name{is_iso_year_week_day}
\alias{is_iso_year_week_day}
\title{Is \code{x} a iso-year-week-day?}
\usage{
is_iso_year_week_day(x)
}
\arguments{
\item{x}{\verb{[object]}

An object.}
}
\value{
Returns \code{TRUE} if \code{x} inherits from \code{"clock_iso_year_week_day"},
otherwise returns \code{FALSE}.
}
\description{
Check if \code{x} is a iso-year-week-day.
}
\examples{
is_iso_year_week_day(iso_year_week_day(2019))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{Date-arithmetic}
\alias{Date-arithmetic}
\alias{add_years.Date}
\alias{add_quarters.Date}
\alias{add_months.Date}
\alias{add_weeks.Date}
\alias{add_days.Date}
\title{Arithmetic: date}
\usage{
\method{add_years}{Date}(x, n, ..., invalid = NULL)

\method{add_quarters}{Date}(x, n, ..., invalid = NULL)

\method{add_months}{Date}(x, n, ..., invalid = NULL)

\method{add_weeks}{Date}(x, n, ...)

\method{add_days}{Date}(x, n, ...)
}
\arguments{
\item{x}{\verb{[Date]}

A Date vector.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}
}
\value{
\code{x} after performing the arithmetic.
}
\description{
These are Date methods for the
\link[=clock-arithmetic]{arithmetic generics}.

Calendrical based arithmetic:

These functions convert to a year-month-day calendar, perform
the arithmetic, then convert back to a Date.
\itemize{
\item \code{add_years()}
\item \code{add_quarters()}
\item \code{add_months()}
}

Time point based arithmetic:

These functions convert to a time point, perform the arithmetic, then
convert back to a Date.
\itemize{
\item \code{add_weeks()}
\item \code{add_days()}
}
}
\details{
Adding a single quarter with \code{add_quarters()} is equivalent to adding
3 months.

\code{x} and \code{n} are recycled against each other.

Only calendrical based arithmetic has the potential to generate invalid
dates. Time point based arithmetic, like adding days, will always generate
a valid date.
}
\examples{
x <- as.Date("2019-01-01")

add_years(x, 1:5)

y <- as.Date("2019-01-31")

# Adding 1 month to `y` generates an invalid date. Unlike year-month-day
# types, R's native Date type cannot handle invalid dates, so you must
# resolve them immediately. If you don't you get an error:
try(add_months(y, 1:2))
add_months(as_year_month_day(y), 1:2)

# Resolve invalid dates by specifying an invalid date resolution strategy
# with the `invalid` argument. Using `"previous"` here sets the date to
# the previous valid date - i.e. the end of the month.
add_months(y, 1:2, invalid = "previous")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/naive-time.R
\name{as-zoned-time-naive-time}
\alias{as-zoned-time-naive-time}
\alias{as_zoned_time.clock_naive_time}
\title{Convert to a zoned-time from a naive-time}
\usage{
\method{as_zoned_time}{clock_naive_time}(x, zone, ..., nonexistent = NULL, ambiguous = NULL)
}
\arguments{
\item{x}{\verb{[clock_naive_time]}

A naive-time to convert to a zoned-time.}

\item{zone}{\verb{[character(1)]}

The zone to convert to.}

\item{...}{These dots are for future extensions and must be empty.}

\item{nonexistent}{\verb{[character / NULL]}

One of the following nonexistent time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"roll-forward"}: The next valid instant in time.
\item \code{"roll-backward"}: The previous valid instant in time.
\item \code{"shift-forward"}: Shift the nonexistent time forward by the size of
the daylight saving time gap.
\item \verb{"shift-backward}: Shift the nonexistent time backward by the size of
the daylight saving time gap.
\item \code{"NA"}: Replace nonexistent times with \code{NA}.
\item \code{"error"}: Error on nonexistent times.
}

Using either \code{"roll-forward"} or \code{"roll-backward"} is generally
recommended over shifting, as these two strategies maintain the
\emph{relative ordering} between elements of the input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{nonexistent} must be supplied
and cannot be \code{NULL}. This is a convenient way to make production code
robust to nonexistent times.}

\item{ambiguous}{\verb{[character / zoned_time / POSIXct / list(2) / NULL]}

One of the following ambiguous time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"earliest"}: Of the two possible times, choose the earliest one.
\item \code{"latest"}: Of the two possible times, choose the latest one.
\item \code{"NA"}: Replace ambiguous times with \code{NA}.
\item \code{"error"}: Error on ambiguous times.
}

Alternatively, \code{ambiguous} is allowed to be a zoned_time (or POSIXct) that
is either length 1, or the same length as the input. If an ambiguous time
is encountered, the zoned_time is consulted. If the zoned_time corresponds
to a naive_time that is also ambiguous \emph{and} uses the same daylight saving
time transition point as the original ambiguous time, then the offset of
the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
resolved by consulting the zoned_time, then this method falls back to
\code{NULL}.

Finally, \code{ambiguous} is allowed to be a list of size 2, where the first
element of the list is a zoned_time (as described above), and the second
element of the list is an ambiguous time resolution strategy to use when
the ambiguous time cannot be resolved by consulting the zoned_time.
Specifying a zoned_time on its own is identical to \verb{list(<zoned_time>, NULL)}.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{ambiguous} must be supplied and
cannot be \code{NULL}. Additionally, \code{ambiguous} cannot be specified as a
zoned_time on its own, as this implies \code{NULL} for ambiguous times that the
zoned_time cannot resolve. Instead, it must be specified as a list
alongside an ambiguous time resolution strategy as described above. This is
a convenient way to make production code robust to ambiguous times.}
}
\value{
A zoned-time vector.
}
\description{
This is a naive-time method for the \code{\link[=as_zoned_time]{as_zoned_time()}} generic.

Converting to a zoned-time from a naive-time retains the printed time,
but changes the underlying duration, depending on the \code{zone} that you choose.

Naive-times are time points with a yet-to-be-determined time zone. By
converting them to a zoned-time, all you are doing is specifying that
time zone while attempting to keep all other printed information the
same (if possible).

If you want to retain the underlying duration, try converting to a zoned-time
\link[=as-zoned-time-sys-time]{from a sys-time}, which is a time point
interpreted as having a UTC time zone.
}
\section{Daylight Saving Time}{


Converting from a naive-time to a zoned-time is not always possible due to
daylight saving time issues. There are two types of these issues:

\emph{Nonexistent} times are the result of daylight saving time "gaps".
For example, in the America/New_York time zone, there was a daylight
saving time gap 1 second after \code{"2020-03-08 01:59:59"}, where the clocks
changed from \code{01:59:59 -> 03:00:00}, completely skipping the 2 o'clock hour.
This means that if you had a naive time of \code{"2020-03-08 02:30:00"}, you
couldn't convert that straight into a zoned-time with this time zone. To
resolve these issues, the \code{nonexistent} argument can be used to specify
one of many nonexistent time resolution strategies.

\emph{Ambiguous} times are the result of daylight saving time "fallbacks".
For example, in the America/New_York time zone, there was a daylight
saving time fallback 1 second after \code{"2020-11-01 01:59:59 EDT"}, at which
point the clocks "fell backwards" by 1 hour, resulting in a printed time of
\code{"2020-11-01 01:00:00 EST"} (note the EDT->EST shift). This resulted in two
1 o'clock hours for this day, so if you had a naive time of
\code{"2020-11-01 01:30:00"}, you wouldn't be able to convert that directly
into a zoned-time with this time zone, as there is no way for clock to know
which of the two ambiguous times you wanted. To resolve these issues,
the \code{ambiguous} argument can be used to specify one of many ambiguous
time resolution strategies.
}

\examples{
library(magrittr)

x <- as_naive_time(year_month_day(2019, 1, 1))

# Converting a naive-time to a zoned-time generally retains the
# printed time, while changing the underlying duration.
as_zoned_time(x, "America/New_York")
as_zoned_time(x, "America/Los_Angeles")

# ---------------------------------------------------------------------------
# Nonexistent time:

new_york <- "America/New_York"

# There was a daylight saving gap in the America/New_York time zone on
# 2020-03-08 01:59:59 -> 03:00:00, which means that one of these
# naive-times don't exist in that time zone. By default, attempting to
# convert it to a zoned time will result in an error.
nonexistent_time <- year_month_day(2020, 03, 08, c(02, 03), c(45, 30), 00)
nonexistent_time <- as_naive_time(nonexistent_time)
try(as_zoned_time(nonexistent_time, new_york))

# Resolve this by specifying a nonexistent time resolution strategy
as_zoned_time(nonexistent_time, new_york, nonexistent = "roll-forward")
as_zoned_time(nonexistent_time, new_york, nonexistent = "roll-backward")

# Note that rolling backwards will choose the last possible moment in
# time at the current precision of the input
nonexistent_nanotime <- time_point_cast(nonexistent_time, "nanosecond")
nonexistent_nanotime
as_zoned_time(nonexistent_nanotime, new_york, nonexistent = "roll-backward")

# A word of caution - Shifting does not guarantee that the relative ordering
# of the input is maintained
shifted <- as_zoned_time(
  nonexistent_time,
  new_york,
  nonexistent = "shift-forward"
)
shifted

# 02:45:00 < 03:30:00
nonexistent_time[1] < nonexistent_time[2]
# 03:45:00 > 03:30:00 (relative ordering is lost)
shifted[1] < shifted[2]

# ---------------------------------------------------------------------------
# Ambiguous time:

new_york <- "America/New_York"

# There was a daylight saving time fallback in the America/New_York time
# zone on 2020-11-01 01:59:59 EDT -> 2020-11-01 01:00:00 EST, resulting
# in two 1 o'clock hours. This means that the following naive time is
# ambiguous since we don't know which of the two 1 o'clocks it belongs to.
# By default, attempting to convert it to a zoned time will result in an
# error.
ambiguous_time <- year_month_day(2020, 11, 01, 01, 30, 00)
ambiguous_time <- as_naive_time(ambiguous_time)
try(as_zoned_time(ambiguous_time, new_york))

# Resolve this by specifying an ambiguous time resolution strategy
earliest <- as_zoned_time(ambiguous_time, new_york, ambiguous = "earliest")
latest <- as_zoned_time(ambiguous_time, new_york, ambiguous = "latest")
na <- as_zoned_time(ambiguous_time, new_york, ambiguous = "NA")
earliest
latest
na

# Now assume that you were given the following zoned-times, i.e.,
# you didn't build them from scratch so you already know their otherwise
# ambiguous offsets
x <- c(earliest, latest)
x

# To set the seconds to 5 in both, you might try:
x_naive <- x \%>\%
  as_naive_time() \%>\%
  as_year_month_day() \%>\%
  set_second(5) \%>\%
  as_naive_time()

x_naive

# But this fails because you've "lost" the information about which
# offsets these ambiguous times started in
try(as_zoned_time(x_naive, zoned_time_zone(x)))

# To get around this, you can use that information by specifying
# `ambiguous = x`, which will use the offset from `x` to resolve the
# ambiguity in `x_naive` as long as `x` is also an ambiguous time with the
# same daylight saving time transition point as `x_naive` (i.e. here
# everything has a transition point of `"2020-11-01 01:00:00 EST"`).
as_zoned_time(x_naive, zoned_time_zone(x), ambiguous = x)

# Say you added one more time to `x` that would not be considered ambiguous
# in naive-time
x <- c(x, as_zoned_time(as_sys_time(latest) + 3600, zoned_time_zone(latest)))
x

# Imagine you want to floor this vector to a multiple of 2 hours, with
# an origin of 1am that day. You can do this by subtracting the origin,
# flooring, then adding it back
origin <- year_month_day(2019, 11, 01, 01, 00, 00) \%>\%
  as_naive_time() \%>\%
  as_duration()

x_naive <- x \%>\%
  as_naive_time() \%>\%
  add_seconds(-origin) \%>\%
  time_point_floor("hour", n = 2) \%>\%
  add_seconds(origin)

x_naive

# You again have ambiguous naive-time points, so you might try using
# `ambiguous = x`. It looks like this took care of the first two problems,
# but we have an issue at location 3.
try(as_zoned_time(x_naive, zoned_time_zone(x), ambiguous = x))

# When we floored from 02:30:00 -> 01:00:00, we went from being
# unambiguous -> ambiguous. In clock, this is something you must handle
# explicitly, and cannot be handled by using information from `x`. You can
# handle this while still retaining the behavior for the other two
# time points that were ambiguous before and after the floor by passing a
# list containing `x` and an ambiguous time resolution strategy to use
# when information from `x` can't resolve ambiguities:
as_zoned_time(x_naive, zoned_time_zone(x), ambiguous = list(x, "latest"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{year-month-weekday-boundary}
\alias{year-month-weekday-boundary}
\alias{calendar_start.clock_year_month_weekday}
\alias{calendar_end.clock_year_month_weekday}
\title{Boundaries: year-month-weekday}
\usage{
\method{calendar_start}{clock_year_month_weekday}(x, precision)

\method{calendar_end}{clock_year_month_weekday}(x, precision)
}
\arguments{
\item{x}{\verb{[clock_year_month_weekday]}

A year-month-weekday vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"month"}
}}
}
\value{
\code{x} at the same precision, but with some components altered to be
at the boundary value.
}
\description{
This is a year-month-weekday method for the \code{\link[=calendar_start]{calendar_start()}} and
\code{\link[=calendar_end]{calendar_end()}} generics. They adjust components of a calendar to the
start or end of a specified \code{precision}.

This method is restricted to only \code{"year"} and \code{"month"} \code{precision}s, and
\code{x} can't be more precise than month precision. Computing the "start" of
a day precision year-month-weekday object isn't defined because
a year-month-weekday with \verb{day = 1, index = 1} doesn't necessarily occur
earlier (chronologically) than \verb{day = 2, index = 1}. Because of these
restrictions, this method isn't particularly useful, but is included for
completeness.
}
\examples{
# Month precision
x <- year_month_weekday(2019, 1:5)
x

# Compute the last month of the year
calendar_end(x, "year")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-day.R
\name{year_day}
\alias{year_day}
\title{Calendar: year-day}
\usage{
year_day(
  year,
  day = NULL,
  hour = NULL,
  minute = NULL,
  second = NULL,
  subsecond = NULL,
  ...,
  subsecond_precision = NULL
)
}
\arguments{
\item{year}{\verb{[integer]}

The year. Values \verb{[-32767, 32767]} are generally allowed.}

\item{day}{\verb{[integer / NULL]}

The day of the year. Values \verb{[1, 366]} are allowed.}

\item{hour}{\verb{[integer / NULL]}

The hour. Values \verb{[0, 23]} are allowed.}

\item{minute}{\verb{[integer / NULL]}

The minute. Values \verb{[0, 59]} are allowed.}

\item{second}{\verb{[integer / NULL]}

The second. Values \verb{[0, 59]} are allowed.}

\item{subsecond}{\verb{[integer / NULL]}

The subsecond. If specified, \code{subsecond_precision} must also be specified
to determine how to interpret the \code{subsecond}.

If using milliseconds, values \verb{[0, 999]} are allowed.

If using microseconds, values \verb{[0, 999999]} are allowed.

If using nanoseconds, values \verb{[0, 999999999]} are allowed.}

\item{...}{These dots are for future extensions and must be empty.}

\item{subsecond_precision}{\verb{[character(1) / NULL]}

The precision to interpret \code{subsecond} as. One of: \code{"millisecond"},
\code{"microsecond"}, or \code{"nanosecond"}.}
}
\value{
A year-day calendar vector.
}
\description{
\code{year_day()} constructs a calendar vector from the Gregorian
year and day of the year.
}
\details{
Fields are recycled against each other.

Fields are collected in order until the first \code{NULL} field is located. No
fields after the first \code{NULL} field are used.
}
\examples{
# Just the year
x <- year_day(2019:2025)
x

year_day(2020, 1:10)

# Last day of the year, accounting for leap years
year_day(2019:2021, "last")

# Precision can go all the way out to nanosecond
year_day(2019, 100, 2, 40, 45, 200, subsecond_precision = "nanosecond")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-day.R
\name{year-day-widen}
\alias{year-day-widen}
\alias{calendar_widen.clock_year_day}
\title{Widen: year-day}
\usage{
\method{calendar_widen}{clock_year_day}(x, precision)
}
\arguments{
\item{x}{\verb{[clock_year_day]}

A year-day vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}
}
\value{
\code{x} widened to the supplied \code{precision}.
}
\description{
This is a year-day method for the \code{\link[=calendar_widen]{calendar_widen()}} generic. It
widens a year-day vector to the specified \code{precision}.
}
\examples{
# Year precision
x <- year_day(2019)
x

# Widen to day precision
calendar_widen(x, "day")

# Or second precision
sec <- calendar_widen(x, "second")
sec

# Second precision can be widened to subsecond precision
milli <- calendar_widen(sec, "millisecond")
micro <- calendar_widen(sec, "microsecond")
milli
micro

# But once you have "locked in" a subsecond precision, it can't
# be widened again
try(calendar_widen(milli, "microsecond"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-year-week-day.R
\name{iso_year_week_day}
\alias{iso_year_week_day}
\title{Calendar: iso-year-week-day}
\usage{
iso_year_week_day(
  year,
  week = NULL,
  day = NULL,
  hour = NULL,
  minute = NULL,
  second = NULL,
  subsecond = NULL,
  ...,
  subsecond_precision = NULL
)
}
\arguments{
\item{year}{\verb{[integer]}

The ISO year. Values \verb{[-32767, 32767]} are generally allowed.}

\item{week}{\verb{[integer / "last" / NULL]}

The ISO week. Values \verb{[1, 53]} are allowed.

If \code{"last"}, then the last week of the ISO year is returned.}

\item{day}{\verb{[integer / NULL]}

The day of the week. Values \verb{[1, 7]} are allowed, with 1 = Monday
and 7 = Sunday, in accordance with the ISO specifications.}

\item{hour}{\verb{[integer / NULL]}

The hour. Values \verb{[0, 23]} are allowed.}

\item{minute}{\verb{[integer / NULL]}

The minute. Values \verb{[0, 59]} are allowed.}

\item{second}{\verb{[integer / NULL]}

The second. Values \verb{[0, 59]} are allowed.}

\item{subsecond}{\verb{[integer / NULL]}

The subsecond. If specified, \code{subsecond_precision} must also be specified
to determine how to interpret the \code{subsecond}.

If using milliseconds, values \verb{[0, 999]} are allowed.

If using microseconds, values \verb{[0, 999999]} are allowed.

If using nanoseconds, values \verb{[0, 999999999]} are allowed.}

\item{...}{These dots are for future extensions and must be empty.}

\item{subsecond_precision}{\verb{[character(1) / NULL]}

The precision to interpret \code{subsecond} as. One of: \code{"millisecond"},
\code{"microsecond"}, or \code{"nanosecond"}.}
}
\value{
A iso-year-week-day calendar vector.
}
\description{
\code{iso_year_week_day()} constructs a calendar from the ISO year, week number,
and week day.
}
\details{
Fields are recycled against each other.

Fields are collected in order until the first \code{NULL} field is located. No
fields after the first \code{NULL} field are used.
}
\examples{
# Year-week
x <- iso_year_week_day(2019:2025, 1)
x

# 2nd day of the first ISO week in multiple years
iso_days <- set_day(x, clock_iso_weekdays$tuesday)
iso_days

# What year-month-day is this?
as_year_month_day(iso_days)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-point.R
\name{time_point_spanning_seq}
\alias{time_point_spanning_seq}
\title{Spanning sequence: time points}
\usage{
time_point_spanning_seq(x)
}
\arguments{
\item{x}{\verb{[clock_sys_time / clock_naive_time]}

A time point vector.}
}
\value{
A sequence along \verb{[min(x), max(x)]}.
}
\description{
\code{time_point_spanning_seq()} generates a regular sequence along the span of
\code{x}, i.e. along \verb{[min(x), max(x)]}. The sequence is generated at the
precision of \code{x}.
}
\details{
Missing values are automatically removed before the sequence is generated.

If you need more precise sequence generation, call \code{\link[=range]{range()}} and \code{\link[=seq]{seq()}}
directly.
}
\examples{
x <- as_naive_time(year_month_day(2019, c(1, 2, 1, 2), c(15, 4, 12, 2)))
x

time_point_spanning_seq(x)

# The sequence is generated at the precision of `x`
x <- as_naive_time(c(
  year_month_day(2019, 1, 1, 5),
  year_month_day(2019, 1, 2, 10),
  year_month_day(2019, 1, 1, 3)
))
time_point_spanning_seq(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{posixt-getters}
\alias{posixt-getters}
\alias{get_year.POSIXt}
\alias{get_month.POSIXt}
\alias{get_day.POSIXt}
\alias{get_hour.POSIXt}
\alias{get_minute.POSIXt}
\alias{get_second.POSIXt}
\title{Getters: date-time}
\usage{
\method{get_year}{POSIXt}(x)

\method{get_month}{POSIXt}(x)

\method{get_day}{POSIXt}(x)

\method{get_hour}{POSIXt}(x)

\method{get_minute}{POSIXt}(x)

\method{get_second}{POSIXt}(x)
}
\arguments{
\item{x}{\verb{[POSIXct / POSIXlt]}

A date-time type to get the component from.}
}
\value{
The component.
}
\description{
These are POSIXct/POSIXlt methods for the \link[=clock-getters]{getter generics}.
\itemize{
\item \code{get_year()} returns the Gregorian year.
\item \code{get_month()} returns the month of the year.
\item \code{get_day()} returns the day of the month.
\item There are sub-daily getters for extracting more precise components, up to
a precision of seconds.
}

For more advanced component extraction, convert to the calendar type
that you are interested in.
}
\examples{
x <- as.POSIXct("2019-01-01", tz = "America/New_York")

x <- add_days(x, 0:5)
x <- set_second(x, 10:15)

get_day(x)
get_second(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-year-week-day.R
\name{iso-year-week-day-arithmetic}
\alias{iso-year-week-day-arithmetic}
\alias{add_years.clock_iso_year_week_day}
\title{Arithmetic: iso-year-week-day}
\usage{
\method{add_years}{clock_iso_year_week_day}(x, n, ...)
}
\arguments{
\item{x}{\verb{[clock_iso_year_week_day]}

A iso-year-week-day vector.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} after performing the arithmetic.
}
\description{
These are iso-year-week-day methods for the
\link[=clock-arithmetic]{arithmetic generics}.
\itemize{
\item \code{add_years()}
}

You cannot add weeks or days to an iso-year-week-day calendar. Adding
days is much more efficiently done by converting to a time point first
by using \code{\link[=as_naive_time]{as_naive_time()}} or \code{\link[=as_sys_time]{as_sys_time()}}. Adding weeks is equally
as efficient as adding 7 days. Additionally, adding weeks to an invalid
iso-year-week object containing \code{iso_year_week_day(2019, 53)} would be
undefined, as the 53rd ISO week of 2019 doesn't exist to begin with.
}
\details{
\code{x} and \code{n} are recycled against each other.
}
\examples{
x <- iso_year_week_day(2019, 1, 1)
add_years(x, 1:2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date-and-date-time-rounding}
\alias{date-and-date-time-rounding}
\alias{date_floor}
\alias{date_ceiling}
\alias{date_round}
\title{Date and date-time rounding}
\usage{
date_floor(x, precision, ..., n = 1L, origin = NULL)

date_ceiling(x, precision, ..., n = 1L, origin = NULL)

date_round(x, precision, ..., n = 1L, origin = NULL)
}
\arguments{
\item{x}{\verb{[Date / POSIXct / POSIXlt]}

A date or date-time vector.}

\item{precision}{\verb{[character(1)]}

A precision. Allowed precisions are dependent on the input used.}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}

\item{origin}{\verb{[Date(1) / POSIXct(1) / POSIXlt(1) / NULL]}

An origin to start counting from. The default \code{origin} is
midnight on 1970-01-01 in the time zone of \code{x}.}
}
\value{
\code{x} rounded to the specified \code{precision}.
}
\description{
\itemize{
\item \code{date_floor()} rounds a date or date-time down to a multiple of
the specified \code{precision}.
\item \code{date_ceiling()} rounds a date or date-time up to a multiple of
the specified \code{precision}.
\item \code{date_round()} rounds up or down depending on what is closer,
rounding up on ties.
}

There are separate help pages for rounding dates and date-times:
\itemize{
\item \link[=date-rounding]{dates (Date)}
\item \link[=posixt-rounding]{date-times (POSIXct/POSIXlt)}
}

These functions round the underlying duration itself, relative to an
\code{origin}. For example, rounding to 15 hours will construct groups of
15 hours, starting from \code{origin}, which defaults to a naive time of
1970-01-01 00:00:00.

If you want to group by components, such as "day of the month", see
\code{\link[=date_group]{date_group()}}.
}
\examples{
# See the type specific documentation for more examples

x <- as.Date("2019-03-31") + 0:5
x

# Flooring by 2 days, note that this is not tied to the current month,
# and instead counts from the specified `origin`.
date_floor(x, "day", n = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{duration_spanning_seq}
\alias{duration_spanning_seq}
\title{Spanning sequence: duration}
\usage{
duration_spanning_seq(x)
}
\arguments{
\item{x}{\verb{[clock_duration]}

A duration vector.}
}
\value{
A sequence along \verb{[min(x), max(x)]}.
}
\description{
\code{duration_spanning_seq()} generates a regular sequence along the span of \code{x},
i.e. along \verb{[min(x), max(x)]}.
}
\details{
Missing values are automatically removed before the sequence is generated.

If you need more precise sequence generation, call \code{\link[=range]{range()}} and \code{\link[=seq]{seq()}}
directly.
}
\examples{
x <- duration_days(c(1, 5, 2))
duration_spanning_seq(x)

# Missing values are removed before the sequence is created
x <- vctrs::vec_c(NA, x, NA)
duration_spanning_seq(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{posixt-count-between}
\alias{posixt-count-between}
\alias{date_count_between.POSIXt}
\title{Counting: date-times}
\usage{
\method{date_count_between}{POSIXt}(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[POSIXct / POSIXlt]}

A pair of date-time vectors. These will be recycled to their common
size.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"quarter"}
\item \code{"month"}
\item \code{"week"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
This is a POSIXct/POSIXlt method for the \code{\link[=date_count_between]{date_count_between()}} generic.

\code{date_count_between()} counts the number of \code{precision} units between
\code{start} and \code{end} (i.e., the number of years or months). This count
corresponds to the \emph{whole number} of units, and will never return a
fractional value.

This is suitable for, say, computing the whole number of years or months
between two dates, accounting for the day of the month and the time of day.

Internally, the date-time is converted to one of the following three clock
types, and the counting is done directly on that type. The choice of type is
based on the most common interpretation of each precision, but is ultimately
a heuristic. See the examples for more information.

\emph{Calendrical based counting:}

These precisions convert to a year-month-day calendar and count while in that
type.
\itemize{
\item \code{"year"}
\item \code{"quarter"}
\item \code{"month"}
}

\emph{Naive-time based counting:}

These precisions convert to a naive-time and count while in that type.
\itemize{
\item \code{"week"}
\item \code{"day"}
}

\emph{Sys-time based counting:}

These precisions convert to a sys-time and count while in that type.
\itemize{
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
}
}
\details{
\code{"quarter"} is equivalent to \code{"month"} precision with \code{n} set to \code{n * 3L}.
}
\section{Comparison Direction}{

The computed count has the property that if \code{start <= end}, then
\verb{start + <count> <= end}. Similarly, if \code{start >= end}, then
\verb{start + <count> >= end}. In other words, the comparison direction between
\code{start} and \code{end} will never change after adding the count to \code{start}. This
makes this function useful for repeated count computations at
increasingly fine precisions.
}

\examples{
start <- date_time_parse("2000-05-05 02:00:00", zone = "America/New_York")
end <- date_time_parse(
  c("2020-05-05 01:00:00", "2020-05-05 03:00:00"),
  zone = "America/New_York"
)

# Age in years
date_count_between(start, end, "year")

# Number of "whole" months between these dates. i.e.
# `2000-05-05 02:00:00 -> 2020-04-05 02:00:00` is 239 months
# `2000-05-05 02:00:00 -> 2020-05-05 02:00:00` is 240 months
# Since `2020-05-05 01:00:00` occurs before the 2nd hour,
# it gets a count of 239
date_count_between(start, end, "month")

# Number of seconds between
date_count_between(start, end, "second")

# ---------------------------------------------------------------------------
# Naive-time VS Sys-time interpretation

# The difference between whether `start` and `end` are converted to a
# naive-time vs a sys-time comes into play when dealing with daylight
# savings.

# Here are two times around a 1 hour DST gap where clocks jumped from
# 01:59:59 -> 03:00:00
x <- date_time_build(1970, 4, 26, 1, 50, 00, zone = "America/New_York")
y <- date_time_build(1970, 4, 26, 3, 00, 00, zone = "America/New_York")

# When treated like sys-times, these are considered to be 10 minutes apart,
# which is the amount of time that would have elapsed if you were watching
# a clock as it changed between these two times.
date_count_between(x, y, "minute")

# Lets add a 3rd date that is ~1 day ahead of these
z <- date_time_build(1970, 4, 27, 1, 55, 00, zone = "America/New_York")

# When treated like naive-times, `z` is considered to be at least 1 day ahead
# of `x`, because `01:55:00` is after `01:50:00`. This is probably what you
# expected.
date_count_between(x, z, "day")

# If these were interpreted like sys-times, then `z` would not be considered
# to be 1 day ahead. That would look something like this:
date_count_between(x, z, "second")
trunc(date_count_between(x, z, "second") / 86400)

# This is because there have only been 83,100 elapsed seconds since `x`,
# which isn't a full day's worth (86,400 seconds). But we'd generally
# consider `z` to be 1 day ahead of `x` (and ignore the DST gap), so that is
# how it is implemented.

# You can override this by converting directly to sys-time, then using
# `time_point_count_between()`
x_st <- as_sys_time(x)
x_st

z_st <- as_sys_time(z)
z_st

time_point_count_between(x_st, z_st, "day")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date_parse}
\alias{date_parse}
\title{Parsing: date}
\usage{
date_parse(x, ..., format = NULL, locale = clock_locale())
}
\arguments{
\item{x}{\verb{[character]}

A character vector to parse.}

\item{...}{These dots are for future extensions and must be empty.}

\item{format}{\verb{[character / NULL]}

A format string. A combination of the following commands, or \code{NULL},
in which case a default format string is used.

A vector of multiple format strings can be supplied. They will be tried in
the order they are provided.

\strong{Year}
\itemize{
\item \verb{\%C}: The century as a decimal number. The modified command \verb{\%NC} where
\code{N} is a positive decimal integer specifies the maximum number of
characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%y}: The last two decimal digits of the year. If the century is not
otherwise specified (e.g. with \verb{\%C}), values in the range \verb{[69 - 99]} are
presumed to refer to the years \verb{[1969 - 1999]}, and values in the range
\verb{[00 - 68]} are presumed to refer to the years \verb{[2000 - 2068]}. The
modified command \verb{\%Ny}, where \code{N} is a positive decimal integer, specifies
the maximum number of characters to read. If not specified, the default is
\code{2}. Leading zeroes are permitted but not required.
\item \verb{\%Y}: The year as a decimal number. The modified command \verb{\%NY} where \code{N}
is a positive decimal integer specifies the maximum number of characters to
read. If not specified, the default is \code{4}. Leading zeroes are permitted
but not required.
}

\strong{Month}
\itemize{
\item \verb{\%b}, \verb{\%B}, \verb{\%h}: The \code{locale}'s full or abbreviated case-insensitive
month name.
\item \verb{\%m}: The month as a decimal number. January is \code{1}. The modified command
\verb{\%Nm} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
}

\strong{Day}
\itemize{
\item \verb{\%d}, \verb{\%e}: The day of the month as a decimal number. The modified
command \verb{\%Nd} where \code{N} is a positive decimal integer specifies the maximum
number of characters to read. If not specified, the default is \code{2}. Leading
zeroes are permitted but not required.
}

\strong{Day of the week}
\itemize{
\item \verb{\%a}, \verb{\%A}: The \code{locale}'s full or abbreviated case-insensitive weekday
name.
\item \verb{\%w}: The weekday as a decimal number (\code{0-6}), where Sunday is \code{0}. The
modified command \verb{\%Nw} where \code{N} is a positive decimal integer specifies
the maximum number of characters to read. If not specified, the default is
\code{1}. Leading zeroes are permitted but not required.
}

\strong{ISO 8601 week-based year}
\itemize{
\item \verb{\%g}: The last two decimal digits of the ISO week-based year. The
modified command \verb{\%Ng} where \code{N} is a positive decimal integer specifies
the maximum number of characters to read. If not specified, the default is
\code{2}. Leading zeroes are permitted but not required.
\item \verb{\%G}: The ISO week-based year as a decimal number. The modified command
\verb{\%NG} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{4}. Leading zeroes
are permitted but not required.
\item \verb{\%V}: The ISO week-based week number as a decimal number. The modified
command \verb{\%NV} where \code{N} is a positive decimal integer specifies the maximum
number of characters to read. If not specified, the default is \code{2}. Leading
zeroes are permitted but not required.
\item \verb{\%u}: The ISO weekday as a decimal number (\code{1-7}), where Monday is \code{1}.
The modified command \verb{\%Nu} where \code{N} is a positive decimal integer
specifies the maximum number of characters to read. If not specified, the
default is \code{1}. Leading zeroes are permitted but not required.
}

\strong{Week of the year}
\itemize{
\item \verb{\%U}: The week number of the year as a decimal number. The first Sunday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. The modified command \verb{\%NU} where \code{N} is a positive
decimal integer specifies the maximum number of characters to read. If not
specified, the default is \code{2}. Leading zeroes are permitted but not
required.
\item \verb{\%W}: The week number of the year as a decimal number. The first Monday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. The modified command \verb{\%NW} where \code{N} is a positive
decimal integer specifies the maximum number of characters to read. If not
specified, the default is \code{2}. Leading zeroes are permitted but not
required.
}

\strong{Day of the year}
\itemize{
\item \verb{\%j}: The day of the year as a decimal number. January 1 is \code{1}. The
modified command \verb{\%Nj} where \code{N} is a positive decimal integer specifies
the maximum number of characters to read. If not specified, the default is
\code{3}. Leading zeroes are permitted but not required.
}

\strong{Date}
\itemize{
\item \verb{\%D}, \verb{\%x}: Equivalent to \verb{\%m/\%d/\%y}.
\item \verb{\%F}: Equivalent to \verb{\%Y-\%m-\%d}. If modified with a width (like \verb{\%NF}),
the width is applied to only \verb{\%Y}.
}

\strong{Time of day}
\itemize{
\item \verb{\%H}: The hour (24-hour clock) as a decimal number. The modified command
\verb{\%NH} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%I}: The hour (12-hour clock) as a decimal number. The modified command
\verb{\%NI} where \code{N} is a positive decimal integer specifies the maximum number
of characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%M}: The minutes as a decimal number. The modified command \verb{\%NM} where
\code{N} is a positive decimal integer specifies the maximum number of
characters to read. If not specified, the default is \code{2}. Leading zeroes
are permitted but not required.
\item \verb{\%S}: The seconds as a decimal number. Leading zeroes are permitted but
not required. If encountered, the \code{locale} determines the decimal point
character. Generally, the maximum number of characters to read is
determined by the precision that you are parsing at. For example, a
precision of \code{"second"} would read a maximum of 2 characters, while a
precision of \code{"millisecond"} would read a maximum of 6 (2 for the values
before the decimal point, 1 for the decimal point, and 3 for the values
after it). The modified command \verb{\%NS}, where \code{N} is a positive decimal
integer, can be used to exactly specify the maximum number of characters to
read. This is only useful if you happen to have seconds with more than 1
leading zero.
\item \verb{\%p}: The \code{locale}'s equivalent of the AM/PM designations associated with
a 12-hour clock. The command \verb{\%I} must precede \verb{\%p} in the format string.
\item \verb{\%R}: Equivalent to \verb{\%H:\%M}.
\item \verb{\%T}, \verb{\%X}: Equivalent to \verb{\%H:\%M:\%S}.
\item \verb{\%r}: Equivalent to \verb{\%I:\%M:\%S \%p}.
}

\strong{Time zone}
\itemize{
\item \verb{\%z}: The offset from UTC in the format \verb{[+|-]hh[mm]}. For example
\code{-0430} refers to 4 hours 30 minutes behind UTC. And \code{04} refers to 4 hours
ahead of UTC. The modified command \verb{\%Ez} parses a \code{:} between the hours and
minutes and leading zeroes on the hour field are optional:
\verb{[+|-]h[h][:mm]}. For example \code{-04:30} refers to 4 hours 30 minutes behind
UTC. And \code{4} refers to 4 hours ahead of UTC.
\item \verb{\%Z}: The full time zone name or the time zone abbreviation, depending on
the function being used. A single word is parsed. This word can only
contain characters that are alphanumeric, or one of \code{'_'}, \code{'/'}, \code{'-'} or
\code{'+'}.
}

\strong{Miscellaneous}
\itemize{
\item \verb{\%c}: A date and time representation. Equivalent to
\verb{\%a \%b \%d \%H:\%M:\%S \%Y}.
\item \code{\%\%}: A \verb{\%} character.
\item \verb{\%n}: Matches one white space character. \verb{\%n}, \verb{\%t}, and a space can be
combined to match a wide range of white-space patterns. For example \code{"\%n "}
matches one or more white space characters, and \code{"\%n\%t\%t"} matches one to
three white space characters.
\item \verb{\%t}: Matches zero or one white space characters.
}}

\item{locale}{\verb{[clock_locale]}

A locale object created from \code{\link[=clock_locale]{clock_locale()}}.}
}
\value{
A Date.
}
\description{
\code{date_parse()} parses strings into a Date.

The default \code{format} used is \code{"\%Y-\%m-\%d"}. This matches the default
result from calling \code{print()} or \code{format()} on a Date.
}
\details{
\emph{\code{date_parse()} ignores both the \verb{\%z} and \verb{\%Z} commands,} as clock treats
Date as a \emph{naive} type, with a yet-to-be-specified time zone.

Parsing strings with sub-daily components, such as hours, minutes, or
seconds, should generally be done with \code{\link[=date_time_parse]{date_time_parse()}}. If you only
need the date components from a string with sub-daily components, choose
one of the following:
\itemize{
\item If the date components are at the front of the string, and you don't want
the time components to affect the date in any way, you can use
\code{\link[=date_parse]{date_parse()}} to parse only the date components. For example,
\code{date_parse("2019-01-05 00:01:02", format = "\%Y-\%m-\%d")} will parse
through \code{05} and then stop.
\item If you want the time components to influence the date, then parse the full
string with \code{\link[=date_time_parse]{date_time_parse()}}, round to day precision with a
rounding function like \code{\link[=date_round]{date_round()}}, and cast to date with \code{\link[=as_date]{as_date()}}.
}

Attempting to directly parse all components of a sub-daily string into a
Date is ambiguous and undefined, and is unlikely to work as you might expect.
For example, \code{date_parse("2019-01-05 00:01:02", format = "\%Y-\%m-\%d \%H:\%M:\%S")} is not officially supported, even if it works in
some cases.
}
\examples{
date_parse("2020-01-01")

date_parse(
  "January 5, 2020",
  format = "\%B \%d, \%Y"
)

# With a different locale
date_parse(
  "janvier 5, 2020",
  format = "\%B \%d, \%Y",
  locale = clock_locale("fr")
)

# A neat feature of `date_parse()` is the ability to parse
# the ISO year-week-day format
date_parse("2020-W01-2", format = "\%G-W\%V-\%u")

# ---------------------------------------------------------------------------
# Sub-daily components

# If you have a string with sub-daily components, but only require the date,
# first parse them as date-times to fully parse the sub-daily components,
# then round using whatever convention is required for your use case before
# converting to date.
x <- c("2019-01-01 11", "2019-01-01 12")

x <- date_time_parse(x, zone = "UTC", format = "\%Y-\%m-\%d \%H")
x

date_floor(x, "day")
date_round(x, "day")

as_date(date_round(x, "day"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getters.R
\name{clock-getters}
\alias{clock-getters}
\alias{get_year}
\alias{get_quarter}
\alias{get_month}
\alias{get_week}
\alias{get_day}
\alias{get_hour}
\alias{get_minute}
\alias{get_second}
\alias{get_millisecond}
\alias{get_microsecond}
\alias{get_nanosecond}
\alias{get_index}
\title{Calendar getters}
\usage{
get_year(x)

get_quarter(x)

get_month(x)

get_week(x)

get_day(x)

get_hour(x)

get_minute(x)

get_second(x)

get_millisecond(x)

get_microsecond(x)

get_nanosecond(x)

get_index(x)
}
\arguments{
\item{x}{\verb{[object]}

An object to get the component from.}
}
\value{
The component.
}
\description{
This family of functions extract fields from a calendar vector. Each
calendar has its own set of supported getters, which are documented on
their own help page:
\itemize{
\item \link[=year-month-day-getters]{year-month-day}
\item \link[=year-month-weekday-getters]{year-month-weekday}
\item \link[=iso-year-week-day-getters]{iso-year-week-day}
\item \link[=year-quarter-day-getters]{year-quarter-day}
\item \link[=year-day-getters]{year-day}
}

There are also convenience methods for extracting certain components
directly from R's native date and date-time types.
\itemize{
\item \link[=Date-getters]{dates (Date)}
\item \link[=posixt-getters]{date-times (POSIXct / POSIXlt)}
}
}
\details{
You cannot extract components directly from a time point type, such as
sys-time or naive-time. Convert it to a calendar type first. Similarly,
a zoned-time must be converted to either a sys-time or naive-time, and
then to a calendar type, to be able to extract components from it.
}
\examples{
x <- year_month_day(2019, 1:3, 5:7, 1, 20, 30)
get_month(x)
get_day(x)
get_second(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weekday.R
\name{as_weekday}
\alias{as_weekday}
\title{Convert to a weekday}
\usage{
as_weekday(x)
}
\arguments{
\item{x}{\verb{[object]}

An object to convert to a weekday. Usually a sys-time or naive-time.}
}
\value{
A weekday.
}
\description{
\code{as_weekday()} converts to a weekday type. This is normally useful for
converting to a weekday from a sys-time or naive-time. You can use this
function along with the \emph{circular arithmetic} that weekday implements to
easily get to the "next Monday" or "previous Sunday".
}
\examples{
x <- as_naive_time(year_month_day(2019, 01, 05))

# This is a Saturday!
as_weekday(x)

# See the examples in `?weekday` for more usage.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{duration_cast}
\alias{duration_cast}
\title{Cast a duration between precisions}
\usage{
duration_cast(x, precision)
}
\arguments{
\item{x}{\verb{[clock_duration]}

A duration.}

\item{precision}{\verb{[character(1)]}

A precision. One of:
\itemize{
\item \code{"year"}
\item \code{"quarter"}
\item \code{"month"}
\item \code{"week"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}
}
\value{
\code{x} cast to the new \code{precision}.
}
\description{
Casting is one way to change a duration's precision.

Casting to a less precise precision will completely drop information that
is more precise than the precision that you are casting to. It does so
in a way that makes it round towards zero.

Casting to a more precise precision is done through a multiplication by
a conversion factor between the current precision and the new precision.
}
\details{
When you want to change to a less precise precision, you often want
\code{\link[=duration_floor]{duration_floor()}} instead of \code{duration_cast()}, as that rounds towards
negative infinity, which is generally the desired behavior when working with
time points (especially ones pre-1970, which are stored as negative
durations).
}
\examples{
x <- duration_seconds(c(86401, -86401))

# Casting rounds towards 0
cast <- duration_cast(x, "day")
cast

# Flooring rounds towards negative infinity
floor <- duration_floor(x, "day")
floor

# Flooring is generally more useful when working with time points,
# note that the cast ends up rounding the pre-1970 date up to the next
# day, while the post-1970 date is rounded down.
as_sys_time(x)
as_sys_time(cast)
as_sys_time(floor)

# Casting to a more precise precision
duration_cast(x, "millisecond")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clock-codes.R
\docType{data}
\name{clock-codes}
\alias{clock-codes}
\alias{clock_months}
\alias{clock_weekdays}
\alias{clock_iso_weekdays}
\title{Integer codes}
\format{
\itemize{
\item \code{clock_months}: An environment containing month codes.
}

\itemize{
\item \code{clock_weekdays}: An environment containing weekday codes.
}

\itemize{
\item \code{clock_iso_weekdays}: An environment containing ISO weekday codes.
}
}
\usage{
clock_months

clock_weekdays

clock_iso_weekdays
}
\description{
Objects with useful mappings from month names and weekday names
to integer codes.
\subsection{Month codes (\code{clock_months})}{
\itemize{
\item \code{january == 1}
\item \code{february == 2}
\item \code{march == 3}
\item \code{april == 4}
\item \code{may == 5}
\item \code{june == 6}
\item \code{july == 7}
\item \code{august == 8}
\item \code{september == 9}
\item \code{october == 10}
\item \code{november == 11}
\item \code{december == 12}
}
}

\subsection{Weekday codes (\code{clock_weekdays})}{
\itemize{
\item \code{sunday == 1}
\item \code{monday == 2}
\item \code{tuesday == 3}
\item \code{wednesday == 4}
\item \code{thursday == 5}
\item \code{friday == 6}
\item \code{saturday == 7}
}
}

\subsection{ISO weekday codes (\code{clock_iso_weekdays})}{
\itemize{
\item \code{monday == 1}
\item \code{tuesday == 2}
\item \code{wednesday == 3}
\item \code{thursday == 4}
\item \code{friday == 5}
\item \code{saturday == 6}
\item \code{sunday == 7}
}
}
}
\examples{
weekday(clock_weekdays$wednesday)

year_month_weekday(2019, clock_months$april, clock_weekdays$monday, 1:4)

iso_year_week_day(2020, 52, clock_iso_weekdays$thursday)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-year-week-day.R
\name{iso-year-week-day-setters}
\alias{iso-year-week-day-setters}
\alias{set_year.clock_iso_year_week_day}
\alias{set_week.clock_iso_year_week_day}
\alias{set_day.clock_iso_year_week_day}
\alias{set_hour.clock_iso_year_week_day}
\alias{set_minute.clock_iso_year_week_day}
\alias{set_second.clock_iso_year_week_day}
\alias{set_millisecond.clock_iso_year_week_day}
\alias{set_microsecond.clock_iso_year_week_day}
\alias{set_nanosecond.clock_iso_year_week_day}
\title{Setters: iso-year-week-day}
\usage{
\method{set_year}{clock_iso_year_week_day}(x, value, ...)

\method{set_week}{clock_iso_year_week_day}(x, value, ...)

\method{set_day}{clock_iso_year_week_day}(x, value, ...)

\method{set_hour}{clock_iso_year_week_day}(x, value, ...)

\method{set_minute}{clock_iso_year_week_day}(x, value, ...)

\method{set_second}{clock_iso_year_week_day}(x, value, ...)

\method{set_millisecond}{clock_iso_year_week_day}(x, value, ...)

\method{set_microsecond}{clock_iso_year_week_day}(x, value, ...)

\method{set_nanosecond}{clock_iso_year_week_day}(x, value, ...)
}
\arguments{
\item{x}{\verb{[clock_iso_year_week_day]}

A iso-year-week-day vector.}

\item{value}{\verb{[integer / "last"]}

The value to set the component to.

For \code{set_week()}, this can also be \code{"last"} to adjust to the last
week of the current ISO year.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} with the component set.
}
\description{
These are iso-year-week-day methods for the
\link[=clock-setters]{setter generics}.
\itemize{
\item \code{set_year()} sets the ISO year.
\item \code{set_week()} sets the ISO week of the year. Valid values are in the range
of \verb{[1, 53]}.
\item \code{set_day()} sets the day of the week. Valid values are in the range of
\verb{[1, 7]}, with 1 = Monday, and 7 = Sunday.
\item There are sub-daily setters for setting more precise components.
}
}
\examples{
# Year precision vector
x <- iso_year_week_day(2019:2023)

# Promote to week precision by setting the week
# (Note that some ISO weeks have 52 weeks, and others have 53)
x <- set_week(x, "last")
x

# Set to an invalid week
invalid <- set_week(x, 53)
invalid

# Here are the invalid ones (they only have 52 weeks)
invalid[invalid_detect(invalid)]

# Resolve the invalid dates by choosing the previous/next valid moment
invalid_resolve(invalid, invalid = "previous")
invalid_resolve(invalid, invalid = "next")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-year-week-day.R
\name{iso-year-week-day-group}
\alias{iso-year-week-day-group}
\alias{calendar_group.clock_iso_year_week_day}
\title{Grouping: iso-year-week-day}
\usage{
\method{calendar_group}{clock_iso_year_week_day}(x, precision, ..., n = 1L)
}
\arguments{
\item{x}{\verb{[clock_iso_year_week_day]}

A iso-year-week-day vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"week"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
\code{x} grouped at the specified \code{precision}.
}
\description{
This is a iso-year-week-day method for the \code{\link[=calendar_group]{calendar_group()}} generic.

Grouping for a iso-year-week-day object can be done at any precision, as
long as \code{x} is at least as precise as \code{precision}.
}
\examples{
x <- iso_year_week_day(2019, 1:52)

# Group by 3 ISO weeks
calendar_group(x, "week", n = 3)

y <- iso_year_week_day(2000:2020, 1, 1)

# Group by 2 ISO years
calendar_group(y, "year", n = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quarterly-year-quarter-day.R
\name{year-quarter-day-arithmetic}
\alias{year-quarter-day-arithmetic}
\alias{add_years.clock_year_quarter_day}
\alias{add_quarters.clock_year_quarter_day}
\title{Arithmetic: year-quarter-day}
\usage{
\method{add_years}{clock_year_quarter_day}(x, n, ...)

\method{add_quarters}{clock_year_quarter_day}(x, n, ...)
}
\arguments{
\item{x}{\verb{[clock_year_quarter_day]}

A year-quarter-day vector.}

\item{n}{\verb{[integer / clock_duration]}

An integer vector to be converted to a duration, or a duration
corresponding to the arithmetic function being used. This corresponds
to the number of duration units to add. \code{n} may be negative to subtract
units of duration.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} after performing the arithmetic.
}
\description{
These are year-quarter-day methods for the
\link[=clock-arithmetic]{arithmetic generics}.
\itemize{
\item \code{add_years()}
\item \code{add_quarters()}
}

Notably, \emph{you cannot add days to a year-quarter-day}. For day-based
arithmetic, first convert to a time point with \code{\link[=as_naive_time]{as_naive_time()}} or
\code{\link[=as_sys_time]{as_sys_time()}}.
}
\details{
\code{x} and \code{n} are recycled against each other.
}
\examples{
x <- year_quarter_day(2019, 1:3)
x

add_quarters(x, 2)

# Make the fiscal year start in March
y <- year_quarter_day(2019, 1:2, 1, start = 3)
y

add_quarters(y, 1)

# What year-month-day does this correspond to?
# Note that the fiscal year doesn't necessarily align with the Gregorian
# year!
as_year_month_day(add_quarters(y, 1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date-formatting}
\alias{date-formatting}
\alias{date_format.Date}
\title{Formatting: date}
\usage{
\method{date_format}{Date}(x, ..., format = NULL, locale = clock_locale())
}
\arguments{
\item{x}{\verb{[Date]}

A date vector.}

\item{...}{These dots are for future extensions and must be empty.}

\item{format}{\verb{[character(1) / NULL]}

If \code{NULL}, a default format is used, which depends on the type of the
input.

Otherwise, a format string which is a combination of:

\strong{Year}
\itemize{
\item \verb{\%C}: The year divided by 100 using floored division. If the result
is a single decimal digit, it is prefixed with \code{0}.
\item \verb{\%y}: The last two decimal digits of the year. If the result is a single
digit it is prefixed by \code{0}.
\item \verb{\%Y}: The year as a decimal number. If the result is less than four
digits it is left-padded with \code{0} to four digits.
}

\strong{Month}
\itemize{
\item \verb{\%b}, \verb{\%h}: The \code{locale}'s abbreviated month name.
\item \verb{\%B}: The \code{locale}'s full month name.
\item \verb{\%m}: The month as a decimal number. January is \code{01}. If the result is a
single digit, it is prefixed with \code{0}.
}

\strong{Day}
\itemize{
\item \verb{\%d}: The day of month as a decimal number. If the result is a single
decimal digit, it is prefixed with \code{0}.
}

\strong{Day of the week}
\itemize{
\item \verb{\%a}: The \code{locale}'s abbreviated weekday name.
\item \verb{\%A}: The \code{locale}'s full weekday name.
\item \verb{\%w}: The weekday as a decimal number (\code{0-6}), where Sunday is \code{0}.
}

\strong{ISO 8601 week-based year}
\itemize{
\item \verb{\%g}: The last two decimal digits of the ISO week-based year. If the
result is a single digit it is prefixed by \code{0}.
\item \verb{\%G}: The ISO week-based year as a decimal number. If the result is less
than four digits it is left-padded with \code{0} to four digits.
\item \verb{\%V}: The ISO week-based week number as a decimal number. If the result
is a single digit, it is prefixed with \code{0}.
\item \verb{\%u}: The ISO weekday as a decimal number (\code{1-7}), where Monday is \code{1}.
}

\strong{Week of the year}
\itemize{
\item \verb{\%U}: The week number of the year as a decimal number. The first Sunday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. If the result is a single digit, it is prefixed with
\code{0}.
\item \verb{\%W}: The week number of the year as a decimal number. The first Monday
of the year is the first day of week \code{01}. Days of the same year prior to
that are in week \code{00}. If the result is a single digit, it is prefixed with
\code{0}.
}

\strong{Day of the year}
\itemize{
\item \verb{\%j}: The day of the year as a decimal number. January 1 is \code{001}. If the
result is less than three digits, it is left-padded with \code{0} to three
digits.
}

\strong{Date}
\itemize{
\item \verb{\%D}, \verb{\%x}: Equivalent to \verb{\%m/\%d/\%y}.
\item \verb{\%F}: Equivalent to \verb{\%Y-\%m-\%d}.
}

\strong{Time of day}
\itemize{
\item \verb{\%H}: The hour (24-hour clock) as a decimal number. If the result is a
single digit, it is prefixed with \code{0}.
\item \verb{\%I}: The hour (12-hour clock) as a decimal number. If the result is a
single digit, it is prefixed with \code{0}.
\item \verb{\%M}: The minute as a decimal number. If the result is a single digit, it
is prefixed with \code{0}.
\item \verb{\%S}: Seconds as a decimal number. Fractional seconds are printed at the
precision of the input. The character for the decimal point is localized
according to \code{locale}.
\item \verb{\%p}: The \code{locale}'s equivalent of the AM/PM designations associated with
a 12-hour clock.
\item \verb{\%R}: Equivalent to \verb{\%H:\%M}.
\item \verb{\%T}, \verb{\%X}: Equivalent to \verb{\%H:\%M:\%S}.
\item \verb{\%r}: Nearly equivalent to \verb{\%I:\%M:\%S \%p}, but seconds are always printed
at second precision.
}

\strong{Time zone}
\itemize{
\item \verb{\%z}: The offset from UTC in the ISO 8601 format. For example \code{-0430}
refers to 4 hours 30 minutes behind UTC. If the offset is zero, \code{+0000} is
used. The modified command \verb{\%Ez} inserts a \code{:} between the hour and
minutes, like \code{-04:30}.
\item \verb{\%Z}: The full time zone name. If \code{abbreviate_zone} is \code{TRUE}, the time
zone abbreviation.
}

\strong{Miscellaneous}
\itemize{
\item \verb{\%c}: A date and time representation. Similar to, but not exactly the
same as, \verb{\%a \%b \%d \%H:\%M:\%S \%Y}.
\item \code{\%\%}: A \verb{\%} character.
\item \verb{\%n}: A newline character.
\item \verb{\%t}: A horizontal-tab character.
}}

\item{locale}{\verb{[clock_locale]}

A locale object created from \code{\link[=clock_locale]{clock_locale()}}.}
}
\value{
A character vector of the formatted input.
}
\description{
This is a Date method for the \code{\link[=date_format]{date_format()}} generic.

\code{date_format()} formats a date (Date) using a \code{format} string.

If \code{format} is \code{NULL}, a default format of \code{"\%Y-\%m-\%d"} is used.
}
\details{
Because a Date is considered to be a \emph{naive} type in clock, meaning that
it currently has no implied time zone, using the \verb{\%z} or \verb{\%Z} format commands
is not allowed and will result in \code{NA}.
}
\examples{
x <- as.Date("2019-01-01")

# Default
date_format(x)

date_format(x, format = "year: \%Y, month: \%m, day: \%d")

# With different locales
date_format(x, format = "\%A, \%B \%d, \%Y")
date_format(x, format = "\%A, \%B \%d, \%Y", locale = clock_locale("fr"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/naive-time.R
\name{is_naive_time}
\alias{is_naive_time}
\title{Is \code{x} a naive-time?}
\usage{
is_naive_time(x)
}
\arguments{
\item{x}{\verb{[object]}

An object.}
}
\value{
\code{TRUE} if \code{x} inherits from \code{"clock_naive_time"}, otherwise \code{FALSE}.
}
\description{
This function determines if the input is a naive-time object.
}
\examples{
is_naive_time(1)
is_naive_time(as_naive_time(duration_days(1)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{posixt-boundary}
\alias{posixt-boundary}
\alias{date_start.POSIXt}
\alias{date_end.POSIXt}
\title{Boundaries: date-time}
\usage{
\method{date_start}{POSIXt}(
  x,
  precision,
  ...,
  invalid = NULL,
  nonexistent = NULL,
  ambiguous = x
)

\method{date_end}{POSIXt}(x, precision, ..., invalid = NULL, nonexistent = NULL, ambiguous = x)
}
\arguments{
\item{x}{\verb{[POSIXct / POSIXlt]}

A date-time vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"month"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}

\item{nonexistent}{\verb{[character / NULL]}

One of the following nonexistent time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"roll-forward"}: The next valid instant in time.
\item \code{"roll-backward"}: The previous valid instant in time.
\item \code{"shift-forward"}: Shift the nonexistent time forward by the size of
the daylight saving time gap.
\item \verb{"shift-backward}: Shift the nonexistent time backward by the size of
the daylight saving time gap.
\item \code{"NA"}: Replace nonexistent times with \code{NA}.
\item \code{"error"}: Error on nonexistent times.
}

Using either \code{"roll-forward"} or \code{"roll-backward"} is generally
recommended over shifting, as these two strategies maintain the
\emph{relative ordering} between elements of the input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{nonexistent} must be supplied
and cannot be \code{NULL}. This is a convenient way to make production code
robust to nonexistent times.}

\item{ambiguous}{\verb{[character / zoned_time / POSIXct / list(2) / NULL]}

One of the following ambiguous time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"earliest"}: Of the two possible times, choose the earliest one.
\item \code{"latest"}: Of the two possible times, choose the latest one.
\item \code{"NA"}: Replace ambiguous times with \code{NA}.
\item \code{"error"}: Error on ambiguous times.
}

Alternatively, \code{ambiguous} is allowed to be a zoned_time (or POSIXct) that
is either length 1, or the same length as the input. If an ambiguous time
is encountered, the zoned_time is consulted. If the zoned_time corresponds
to a naive_time that is also ambiguous \emph{and} uses the same daylight saving
time transition point as the original ambiguous time, then the offset of
the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
resolved by consulting the zoned_time, then this method falls back to
\code{NULL}.

Finally, \code{ambiguous} is allowed to be a list of size 2, where the first
element of the list is a zoned_time (as described above), and the second
element of the list is an ambiguous time resolution strategy to use when
the ambiguous time cannot be resolved by consulting the zoned_time.
Specifying a zoned_time on its own is identical to \verb{list(<zoned_time>, NULL)}.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{ambiguous} must be supplied and
cannot be \code{NULL}. Additionally, \code{ambiguous} cannot be specified as a
zoned_time on its own, as this implies \code{NULL} for ambiguous times that the
zoned_time cannot resolve. Instead, it must be specified as a list
alongside an ambiguous time resolution strategy as described above. This is
a convenient way to make production code robust to ambiguous times.}
}
\value{
\code{x} but with some components altered to be at the boundary value.
}
\description{
This is a POSIXct/POSIXlt method for the \code{\link[=date_start]{date_start()}} and \code{\link[=date_end]{date_end()}}
generics.
}
\examples{
x <- date_time_build(2019:2021, 2:4, 3:5, 4, 5, 6, zone = "America/New_York")
x

# Last moment of the month
date_end(x, "month")

# Notice that this is different from just setting the day to `"last"`
set_day(x, "last")

# Last moment of the year
date_end(x, "year")

# First moment of the hour
date_start(x, "hour")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-weekday.R
\name{as_year_month_weekday}
\alias{as_year_month_weekday}
\title{Convert to year-month-weekday}
\usage{
as_year_month_weekday(x)
}
\arguments{
\item{x}{\verb{[vector]}

A vector to convert to year-month-weekday.}
}
\value{
A year-month-weekday vector.
}
\description{
\code{as_year_month_weekday()} converts a vector to the year-month-weekday
calendar. Time points, Dates, POSIXct, and other calendars can all be
converted to year-month-weekday.
}
\examples{
# From Date
as_year_month_weekday(as.Date("2019-01-01"))

# From POSIXct, which assumes that the naive time is what should be converted
as_year_month_weekday(as.POSIXct("2019-01-01 02:30:30", "America/New_York"))

# From other calendars
as_year_month_weekday(year_quarter_day(2019, quarter = 2, day = 50))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{seq.clock_duration}
\alias{seq.clock_duration}
\title{Sequences: duration}
\usage{
\method{seq}{clock_duration}(from, to = NULL, by = NULL, length.out = NULL, along.with = NULL, ...)
}
\arguments{
\item{from}{\verb{[clock_duration(1)]}

A duration to start the sequence from.}

\item{to}{\verb{[clock_duration(1) / NULL]}

A duration to stop the sequence at.

\code{to} is cast to the type of \code{from}.

\code{to} is only included in the result if the resulting sequence divides
the distance between \code{from} and \code{to} exactly.}

\item{by}{\verb{[integer(1) / clock_duration(1) / NULL]}

The unit to increment the sequence by.

If \code{by} is an integer, it is transformed into a duration with the
precision of \code{from}.

If \code{by} is a duration, it is cast to the type of \code{from}.}

\item{length.out}{\verb{[positive integer(1) / NULL]}

The length of the resulting sequence.

If specified, \code{along.with} must be \code{NULL}.}

\item{along.with}{\verb{[vector / NULL]}

A vector who's length determines the length of the resulting sequence.

Equivalent to \code{length.out = vec_size(along.with)}.

If specified, \code{length.out} must be \code{NULL}.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A sequence with the type of \code{from}.
}
\description{
This is a duration method for the \code{\link[=seq]{seq()}} generic.

Using \code{seq()} on duration objects always retains the type of \code{from}.

When calling \code{seq()}, exactly two of the following must be specified:
\itemize{
\item \code{to}
\item \code{by}
\item Either \code{length.out} or \code{along.with}
}
}
\details{
If \code{from > to} and \code{by > 0}, then the result will be length 0. This matches
the behavior of \code{\link[rlang:seq2]{rlang::seq2()}}, and results in nicer theoretical
properties when compared with throwing an error. Similarly, if \code{from < to}
and \code{by < 0}, then the result will also be length 0.
}
\examples{
seq(duration_days(0), duration_days(100), by = 5)

# Using a duration `by`. Note that `by` is cast to the type of `from`.
seq(duration_days(0), duration_days(100), by = duration_weeks(1))

# `to` is cast from 5 years to 60 months
# `by` is cast from 1 quarter to 4 months
seq(duration_months(0), duration_years(5), by = duration_quarters(1))

seq(duration_days(20), by = 2, length.out = 5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weekday.R
\name{is_weekday}
\alias{is_weekday}
\title{Is \code{x} a weekday?}
\usage{
is_weekday(x)
}
\arguments{
\item{x}{\verb{[object]}

An object.}
}
\value{
\code{TRUE} if \code{x} inherits from \code{"clock_weekday"}, otherwise \code{FALSE}.
}
\description{
This function determines if the input is a weekday object.
}
\examples{
is_weekday(1)
is_weekday(weekday(1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-day.R
\name{year-day-count-between}
\alias{year-day-count-between}
\alias{calendar_count_between.clock_year_day}
\title{Counting: year-day}
\usage{
\method{calendar_count_between}{clock_year_day}(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[clock_year_day]}

A pair of year-day vectors. These will be recycled to their
common size.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
This is a year-day method for the \code{\link[=calendar_count_between]{calendar_count_between()}} generic.
It counts the number of \code{precision} units between \code{start} and \code{end}
(i.e., the number of years).
}
\examples{
# Compute an individual's age in years
x <- year_day(2001, 100)
y <- year_day(2021, c(99, 101))

calendar_count_between(x, y, "year")

# Or in a whole number multiple of years
calendar_count_between(x, y, "year", n = 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-day.R
\name{year-month-day-count-between}
\alias{year-month-day-count-between}
\alias{calendar_count_between.clock_year_month_day}
\title{Counting: year-month-day}
\usage{
\method{calendar_count_between}{clock_year_month_day}(start, end, precision, ..., n = 1L)
}
\arguments{
\item{start, end}{\verb{[clock_year_month_day]}

A pair of year-month-day vectors. These will be recycled to their
common size.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"quarter"}
\item \code{"month"}
}}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}
}
\value{
An integer representing the number of \code{precision} units between
\code{start} and \code{end}.
}
\description{
This is a year-month-day method for the \code{\link[=calendar_count_between]{calendar_count_between()}} generic.
It counts the number of \code{precision} units between \code{start} and \code{end}
(i.e., the number of years or months).
}
\details{
\code{"quarter"} is equivalent to \code{"month"} precision with \code{n} set to \code{n * 3L}.
}
\examples{
# Compute an individual's age in years
x <- year_month_day(2001, 2, 4)
today <- year_month_day(2021, 11, 30)
calendar_count_between(x, today, "year")

# Compute the number of months between two dates, taking
# into account the day of the month and time of day
x <- year_month_day(2000, 4, 2, 5)
y <- year_month_day(2000, 7, c(1, 2, 2), c(3, 4, 6))
calendar_count_between(x, y, "month")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-month-day.R
\name{year-month-day-narrow}
\alias{year-month-day-narrow}
\alias{calendar_narrow.clock_year_month_day}
\title{Narrow: year-month-day}
\usage{
\method{calendar_narrow}{clock_year_month_day}(x, precision)
}
\arguments{
\item{x}{\verb{[clock_year_month_day]}

A year-month-day vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"year"}
\item \code{"month"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
\item \code{"millisecond"}
\item \code{"microsecond"}
\item \code{"nanosecond"}
}}
}
\value{
\code{x} narrowed to the supplied \code{precision}.
}
\description{
This is a year-month-day method for the \code{\link[=calendar_narrow]{calendar_narrow()}} generic. It
narrows a year-month-day vector to the specified \code{precision}.
}
\examples{
# Hour precision
x <- year_month_day(2019, 1, 3, 4)
x

# Narrowed to day precision
calendar_narrow(x, "day")

# Or month precision
calendar_narrow(x, "month")

# Subsecond precision can be narrowed to second precision
milli <- calendar_widen(x, "millisecond")
micro <- calendar_widen(x, "microsecond")
milli
micro

calendar_narrow(milli, "second")
calendar_narrow(micro, "second")

# But once you have "locked in" a subsecond precision, it can't be
# narrowed to another subsecond precision
try(calendar_narrow(micro, "millisecond"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{posixt-rounding}
\alias{posixt-rounding}
\alias{date_floor.POSIXt}
\alias{date_ceiling.POSIXt}
\alias{date_round.POSIXt}
\title{Rounding: date-time}
\usage{
\method{date_floor}{POSIXt}(
  x,
  precision,
  ...,
  n = 1L,
  origin = NULL,
  nonexistent = NULL,
  ambiguous = x
)

\method{date_ceiling}{POSIXt}(
  x,
  precision,
  ...,
  n = 1L,
  origin = NULL,
  nonexistent = NULL,
  ambiguous = x
)

\method{date_round}{POSIXt}(
  x,
  precision,
  ...,
  n = 1L,
  origin = NULL,
  nonexistent = NULL,
  ambiguous = x
)
}
\arguments{
\item{x}{\verb{[POSIXct / POSIXlt]}

A date-time vector.}

\item{precision}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"week"}
\item \code{"day"}
\item \code{"hour"}
\item \code{"minute"}
\item \code{"second"}
}

\code{"week"} is an alias for \code{"day"} with \code{n * 7}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{n}{\verb{[positive integer(1)]}

A single positive integer specifying a multiple of \code{precision} to use.}

\item{origin}{\verb{[POSIXct(1) / POSIXlt(1) / NULL]}

An origin to start counting from.

\code{origin} must have exactly the same time zone as \code{x}.

\code{origin} will be floored to \code{precision}. If information is lost when
flooring, a warning will be thrown.

If \code{NULL}, defaults to midnight on 1970-01-01 in the time zone of \code{x}.}

\item{nonexistent}{\verb{[character / NULL]}

One of the following nonexistent time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"roll-forward"}: The next valid instant in time.
\item \code{"roll-backward"}: The previous valid instant in time.
\item \code{"shift-forward"}: Shift the nonexistent time forward by the size of
the daylight saving time gap.
\item \verb{"shift-backward}: Shift the nonexistent time backward by the size of
the daylight saving time gap.
\item \code{"NA"}: Replace nonexistent times with \code{NA}.
\item \code{"error"}: Error on nonexistent times.
}

Using either \code{"roll-forward"} or \code{"roll-backward"} is generally
recommended over shifting, as these two strategies maintain the
\emph{relative ordering} between elements of the input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{nonexistent} must be supplied
and cannot be \code{NULL}. This is a convenient way to make production code
robust to nonexistent times.}

\item{ambiguous}{\verb{[character / zoned_time / POSIXct / list(2) / NULL]}

One of the following ambiguous time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"earliest"}: Of the two possible times, choose the earliest one.
\item \code{"latest"}: Of the two possible times, choose the latest one.
\item \code{"NA"}: Replace ambiguous times with \code{NA}.
\item \code{"error"}: Error on ambiguous times.
}

Alternatively, \code{ambiguous} is allowed to be a zoned_time (or POSIXct) that
is either length 1, or the same length as the input. If an ambiguous time
is encountered, the zoned_time is consulted. If the zoned_time corresponds
to a naive_time that is also ambiguous \emph{and} uses the same daylight saving
time transition point as the original ambiguous time, then the offset of
the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
resolved by consulting the zoned_time, then this method falls back to
\code{NULL}.

Finally, \code{ambiguous} is allowed to be a list of size 2, where the first
element of the list is a zoned_time (as described above), and the second
element of the list is an ambiguous time resolution strategy to use when
the ambiguous time cannot be resolved by consulting the zoned_time.
Specifying a zoned_time on its own is identical to \verb{list(<zoned_time>, NULL)}.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{ambiguous} must be supplied and
cannot be \code{NULL}. Additionally, \code{ambiguous} cannot be specified as a
zoned_time on its own, as this implies \code{NULL} for ambiguous times that the
zoned_time cannot resolve. Instead, it must be specified as a list
alongside an ambiguous time resolution strategy as described above. This is
a convenient way to make production code robust to ambiguous times.}
}
\value{
\code{x} rounded to the specified \code{precision}.
}
\description{
These are POSIXct/POSIXlt methods for the
\link[=date-and-date-time-rounding]{rounding generics}.
\itemize{
\item \code{date_floor()} rounds a date-time down to a multiple of
the specified \code{precision}.
\item \code{date_ceiling()} rounds a date-time up to a multiple of
the specified \code{precision}.
\item \code{date_round()} rounds up or down depending on what is closer,
rounding up on ties.
}

You can group by irregular periods such as \code{"month"} or \code{"year"} by using
\code{\link[=date_group]{date_group()}}.
}
\details{
When rounding by \code{"week"}, remember that the \code{origin} determines the "week
start". By default, 1970-01-01 is the implicit origin, which is a
Thursday. If you would like to round by weeks with a different week start,
just supply an origin on the weekday you are interested in.
}
\examples{
x <- as.POSIXct("2019-03-31", "America/New_York")
x <- add_days(x, 0:5)

# Flooring by 2 days, note that this is not tied to the current month,
# and instead counts from the specified `origin`, so groups can cross
# the month boundary
date_floor(x, "day", n = 2)

# Compare to `date_group()`, which groups by the day of the month
date_group(x, "day", n = 2)

# Note that daylight saving time gaps can throw off rounding
x <- as.POSIXct("1970-04-26 01:59:59", "America/New_York") + c(0, 1)
x

# Rounding is done in naive-time, which means that rounding by 2 hours
# will attempt to generate a time of 1970-04-26 02:00:00, which doesn't
# exist in this time zone
try(date_floor(x, "hour", n = 2))

# You can handle this by specifying a nonexistent time resolution strategy
date_floor(x, "hour", n = 2, nonexistent = "roll-forward")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gregorian-year-day.R
\name{as_year_day}
\alias{as_year_day}
\title{Convert to year-day}
\usage{
as_year_day(x, ...)
}
\arguments{
\item{x}{\verb{[vector]}

A vector to convert to year-day.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A year-day vector.
}
\description{
\code{as_year_day()} converts a vector to the year-day calendar.
Time points, Dates, POSIXct, and other calendars can all be converted to
year-day.
}
\examples{
# From Date
as_year_day(as.Date("2019-05-01"))

# From POSIXct, which assumes that the naive time is what should be converted
as_year_day(as.POSIXct("2019-05-01 02:30:30", "America/New_York"))

# From other calendars
as_year_day(year_quarter_day(2019, quarter = 2, day = 50))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R, R/posixt.R
\name{date-today}
\alias{date-today}
\alias{date_today}
\alias{date_now}
\title{Current date and date-time}
\usage{
date_today(zone)

date_now(zone)
}
\arguments{
\item{zone}{\verb{[character(1)]}

A time zone to get the current time for.}
}
\value{
\itemize{
\item \code{date_today()} a single Date.
\item \code{date_now()} a single POSIXct.
}
}
\description{
\itemize{
\item \code{date_today()} returns the current date in the specified \code{zone} as a Date.
\item \code{date_now()} returns the current date-time in the specified \code{zone} as a
POSIXct.
}
}
\details{
clock assumes that Date is a \emph{naive} type, like naive-time. This means that
\code{date_today()} first looks up the current date-time in the specified \code{zone},
then converts that to a Date, retaining the printed time while dropping any
information about that time zone.
}
\examples{
# Current date in the local time zone
date_today("")

# Current date in a specified time zone
date_today("Europe/London")

# Current date-time in that same time zone
date_now("Europe/London")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/week-year-week-day.R
\name{year-week-day-getters}
\alias{year-week-day-getters}
\alias{get_year.clock_year_week_day}
\alias{get_week.clock_year_week_day}
\alias{get_day.clock_year_week_day}
\alias{get_hour.clock_year_week_day}
\alias{get_minute.clock_year_week_day}
\alias{get_second.clock_year_week_day}
\alias{get_millisecond.clock_year_week_day}
\alias{get_microsecond.clock_year_week_day}
\alias{get_nanosecond.clock_year_week_day}
\title{Getters: year-week-day}
\usage{
\method{get_year}{clock_year_week_day}(x)

\method{get_week}{clock_year_week_day}(x)

\method{get_day}{clock_year_week_day}(x)

\method{get_hour}{clock_year_week_day}(x)

\method{get_minute}{clock_year_week_day}(x)

\method{get_second}{clock_year_week_day}(x)

\method{get_millisecond}{clock_year_week_day}(x)

\method{get_microsecond}{clock_year_week_day}(x)

\method{get_nanosecond}{clock_year_week_day}(x)
}
\arguments{
\item{x}{\verb{[clock_year_week_day]}

A year-week-day to get the component from.}
}
\value{
The component.
}
\description{
These are year-week-day methods for the
\link[=clock-getters]{getter generics}.
\itemize{
\item \code{get_year()} returns the year. Note that this can differ from the
Gregorian year.
\item \code{get_week()} returns the week of the current year.
\item \code{get_day()} returns a value between 1-7 indicating the weekday of the
current week, where \verb{1 = start of week} and \verb{7 = end of week}, in line with
the chosen \code{start}.
\item There are sub-daily getters for extracting more precise components.
}
}
\examples{
x <- year_week_day(2019, 50:52, 1:3)
x

# Get the week
get_week(x)

# Gets the weekday
get_day(x)

# Note that the year can differ from the Gregorian year
iso <- year_week_day(2019, 1, 1, start = clock_weekdays$monday)
ymd <- as_year_month_day(iso)

get_year(iso)
get_year(ymd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clock-package.R
\docType{package}
\name{clock-package}
\alias{clock}
\alias{clock-package}
\title{clock: Date-Time Types and Tools}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides a comprehensive library for date-time manipulations using a new family of orthogonal date-time classes (durations, time points, zoned-times, and calendars) that partition responsibilities so that the complexities of time zones are only considered when they are really needed. Capabilities include: date-time parsing, formatting, arithmetic, extraction and updating of components, and rounding.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://clock.r-lib.org}
  \item \url{https://github.com/r-lib/clock}
  \item Report bugs at \url{https://github.com/r-lib/clock/issues}
}

}
\author{
\strong{Maintainer}: Davis Vaughan \email{davis@rstudio.com}

Other contributors:
\itemize{
  \item RStudio [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date-and-date-time-shifting}
\alias{date-and-date-time-shifting}
\alias{date_shift}
\title{Shifting: date and date-time}
\usage{
date_shift(x, target, ..., which = "next", boundary = "keep")
}
\arguments{
\item{x}{\verb{[Date / POSIXct / POSIXlt]}

A date or date-time vector.}

\item{target}{\verb{[weekday]}

A weekday created from \code{\link[=weekday]{weekday()}} to target.

Generally this is length 1, but can also be the same length as \code{x}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{which}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"next"}: Shift to the next instance of the \code{target} weekday.
\item \verb{"previous}: Shift to the previous instance of the \code{target} weekday.
}}

\item{boundary}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"keep"}: If \code{x} is currently on the \code{target} weekday, return it.
\item \code{"advance"}: If \code{x} is currently on the \code{target} weekday, advance it
anyways.
}}
}
\value{
\code{x} shifted to the \code{target} weekday.
}
\description{
\code{date_shift()} shifts \code{x} to the \code{target} weekday. You can shift to the next
or previous weekday. If \code{x} is currently on the \code{target} weekday, you can
choose to leave it alone or advance it to the next instance of the \code{target}.

There are separate help pages for shifting dates and date-times:
\itemize{
\item \link[=date-shifting]{dates (Date)}
\item \link[=posixt-shifting]{date-times (POSIXct/POSIXlt)}
}
}
\examples{
# See the type specific documentation for more examples

x <- as.Date("2019-01-01") + 0:1

# A Tuesday and Wednesday
as_weekday(x)

monday <- weekday(clock_weekdays$monday)

# Shift to the next Monday
date_shift(x, monday)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quarterly-year-quarter-day.R
\name{year-quarter-day-setters}
\alias{year-quarter-day-setters}
\alias{set_year.clock_year_quarter_day}
\alias{set_quarter.clock_year_quarter_day}
\alias{set_day.clock_year_quarter_day}
\alias{set_hour.clock_year_quarter_day}
\alias{set_minute.clock_year_quarter_day}
\alias{set_second.clock_year_quarter_day}
\alias{set_millisecond.clock_year_quarter_day}
\alias{set_microsecond.clock_year_quarter_day}
\alias{set_nanosecond.clock_year_quarter_day}
\title{Setters: year-quarter-day}
\usage{
\method{set_year}{clock_year_quarter_day}(x, value, ...)

\method{set_quarter}{clock_year_quarter_day}(x, value, ...)

\method{set_day}{clock_year_quarter_day}(x, value, ...)

\method{set_hour}{clock_year_quarter_day}(x, value, ...)

\method{set_minute}{clock_year_quarter_day}(x, value, ...)

\method{set_second}{clock_year_quarter_day}(x, value, ...)

\method{set_millisecond}{clock_year_quarter_day}(x, value, ...)

\method{set_microsecond}{clock_year_quarter_day}(x, value, ...)

\method{set_nanosecond}{clock_year_quarter_day}(x, value, ...)
}
\arguments{
\item{x}{\verb{[clock_year_quarter_day]}

A year-quarter-day vector.}

\item{value}{\verb{[integer / "last"]}

The value to set the component to.

For \code{set_day()}, this can also be \code{"last"} to adjust to the last
day of the current fiscal quarter.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
\code{x} with the component set.
}
\description{
These are year-quarter-day methods for the
\link[=clock-setters]{setter generics}.
\itemize{
\item \code{set_year()} sets the fiscal year.
\item \code{set_quarter()} sets the fiscal quarter of the year.
Valid values are in the range of \verb{[1, 4]}.
\item \code{set_day()} sets the day of the fiscal quarter.
Valid values are in the range of \verb{[1, 92]}.
\item There are sub-daily setters for setting more precise components.
}
}
\examples{
library(magrittr)

# Quarter precision vector
x <- year_quarter_day(2019, 1:4)
x

# Promote to day precision by setting the day
x <- set_day(x, 1)
x

# Or set to the last day of the quarter
x <- set_day(x, "last")
x

# What year-month-day is this?
as_year_month_day(x)

# Set to an invalid day of the quarter
# (not all quarters have 92 days)
invalid <- set_day(x, 92)
invalid

# Here are the invalid ones
invalid[invalid_detect(invalid)]

# Resolve the invalid dates by choosing the previous/next valid moment
invalid_resolve(invalid, invalid = "previous")
invalid_resolve(invalid, invalid = "next")

# Or resolve by "overflowing" by the number of days that you have
# gone past the last valid day
invalid_resolve(invalid, invalid = "overflow")

# This is similar to
days <- get_day(invalid) - 1L
invalid \%>\%
  set_day(1) \%>\%
  as_naive_time() \%>\%
  add_days(days) \%>\%
  as_year_quarter_day()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weekday.R
\name{weekday}
\alias{weekday}
\title{Construct a weekday vector}
\usage{
weekday(code = integer(), ..., encoding = "western")
}
\arguments{
\item{code}{\verb{[integer]}

Integer codes between \verb{[1, 7]} representing days of the week. The
interpretation of these values depends on \code{encoding}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{encoding}{\verb{[character(1)]}

One of:
\itemize{
\item \code{"western"}: Encode weekdays assuming \code{1 == Sunday} and \code{7 == Saturday}.
\item \code{"iso"}: Encode weekdays assuming \code{1 == Monday} and \code{7 == Sunday}. This
is in line with the ISO standard.
}}
}
\value{
A weekday vector.
}
\description{
A \code{weekday} is a simple type that represents a day of the week.

The most interesting thing about the weekday type is that it implements
\emph{circular arithmetic}, which makes determining the "next Monday" or
"previous Tuesday" from a sys-time or naive-time easy to compute.
See the examples.
}
\examples{
x <- as_naive_time(year_month_day(2019, 01, 05))

# This is a Saturday!
as_weekday(x)

# Adjust to the next Wednesday
wednesday <- weekday(clock_weekdays$wednesday)

# This returns the number of days until the next Wednesday using
# circular arithmetic
# "Wednesday - Saturday = 4 days until next Wednesday"
wednesday - as_weekday(x)

# Advance to the next Wednesday
x_next_wednesday <- x + (wednesday - as_weekday(x))

as_weekday(x_next_wednesday)

# What about the previous Tuesday?
tuesday <- weekday(clock_weekdays$tuesday)
x - (as_weekday(x) - tuesday)

# What about the next Saturday?
# With an additional condition that if today is a Saturday,
# then advance to the next one.
saturday <- weekday(clock_weekdays$saturday)
x + 1L + (saturday - as_weekday(x + 1L))

# You can supply an ISO coding for `code` as well, where 1 == Monday.
weekday(1:7, encoding = "western")
weekday(1:7, encoding = "iso")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{date-sequence}
\alias{date-sequence}
\alias{date_seq.Date}
\title{Sequences: date}
\usage{
\method{date_seq}{Date}(from, ..., to = NULL, by = NULL, total_size = NULL, invalid = NULL)
}
\arguments{
\item{from}{\verb{[Date(1)]}

A date to start the sequence from.}

\item{...}{These dots are for future extensions and must be empty.}

\item{to}{\verb{[Date(1) / NULL]}

A date to stop the sequence at.

\code{to} is only included in the result if the resulting sequence divides
the distance between \code{from} and \code{to} exactly.

If \code{to} is supplied along with \code{by}, all components of \code{to} more precise
than the precision of \code{by} must match \code{from} exactly. For example, if \code{by = duration_months(1)}, the day component of \code{to} must match the day component
of \code{from}. This ensures that the generated sequence is, at a minimum, a
weakly monotonic sequence of dates.}

\item{by}{\verb{[integer(1) / clock_duration(1) / NULL]}

The unit to increment the sequence by.

If \code{by} is an integer, it is equivalent to \code{duration_days(by)}.

If \code{by} is a duration, it is allowed to have a precision of:
\itemize{
\item year
\item quarter
\item month
\item week
\item day
}}

\item{total_size}{\verb{[positive integer(1) / NULL]}

The size of the resulting sequence.

If specified alongside \code{to}, this must generate a non-fractional sequence
between \code{from} and \code{to}.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}
}
\value{
A date vector.
}
\description{
This is a Date method for the \code{\link[=date_seq]{date_seq()}} generic.

\code{date_seq()} generates a date (Date) sequence.

When calling \code{date_seq()}, exactly two of the following must be specified:
\itemize{
\item \code{to}
\item \code{by}
\item \code{total_size}
}
}
\examples{
from <- date_build(2019, 1)
to <- date_build(2019, 4)

# Defaults to daily sequence
date_seq(from, to = to, by = 7)

# Use durations to change to monthly or yearly sequences
date_seq(from, to = to, by = duration_months(1))
date_seq(from, by = duration_years(-2), total_size = 3)

# Note that components of `to` more precise than the precision of `by`
# must match `from` exactly. For example, this is not well defined:
from <- date_build(2019, 5, 2)
to <- date_build(2025, 7, 5)
try(date_seq(from, to = to, by = duration_years(1)))

# The month and day components of `to` must match `from`
to <- date_build(2025, 5, 2)
date_seq(from, to = to, by = duration_years(1))

# ---------------------------------------------------------------------------

# Invalid dates must be resolved with the `invalid` argument
from <- date_build(2019, 1, 31)
to <- date_build(2019, 12, 31)

try(date_seq(from, to = to, by = duration_months(1)))
date_seq(from, to = to, by = duration_months(1), invalid = "previous")

# Compare this to the base R result, which is often a source of confusion
seq(from, to = to, by = "1 month")

# This is equivalent to the overflow invalid resolution strategy
date_seq(from, to = to, by = duration_months(1), invalid = "overflow")

# ---------------------------------------------------------------------------

# Usage of `to` and `total_size` must generate a non-fractional sequence
# between `from` and `to`
from <- date_build(2019, 1, 1)
to <- date_build(2019, 1, 4)

# These are fine
date_seq(from, to = to, total_size = 2)
date_seq(from, to = to, total_size = 4)

# But this is not!
try(date_seq(from, to = to, total_size = 3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posixt.R
\name{posixt-sequence}
\alias{posixt-sequence}
\alias{date_seq.POSIXt}
\title{Sequences: date-time}
\usage{
\method{date_seq}{POSIXt}(
  from,
  ...,
  to = NULL,
  by = NULL,
  total_size = NULL,
  invalid = NULL,
  nonexistent = NULL,
  ambiguous = NULL
)
}
\arguments{
\item{from}{\verb{[POSIXct(1) / POSIXlt(1)]}

A date-time to start the sequence from.}

\item{...}{These dots are for future extensions and must be empty.}

\item{to}{\verb{[POSIXct(1) / POSIXlt(1) / NULL]}

A date-time to stop the sequence at.

\code{to} is only included in the result if the resulting sequence divides
the distance between \code{from} and \code{to} exactly.

If \code{to} is supplied along with \code{by}, all components of \code{to} more precise
than the precision of \code{by} must match \code{from} exactly. For example, if \code{by = duration_months(1)}, the day, hour, minute, and second components of \code{to}
must match the corresponding components of \code{from}. This ensures that the
generated sequence is, at a minimum, a weakly monotonic sequence of
date-times.

The time zone of \code{to} must match the time zone of \code{from} exactly.}

\item{by}{\verb{[integer(1) / clock_duration(1) / NULL]}

The unit to increment the sequence by.

If \code{by} is an integer, it is equivalent to \code{duration_seconds(by)}.

If \code{by} is a duration, it is allowed to have a precision of:
\itemize{
\item year
\item quarter
\item month
\item week
\item day
\item hour
\item minute
\item second
}}

\item{total_size}{\verb{[positive integer(1) / NULL]}

The size of the resulting sequence.

If specified alongside \code{to}, this must generate a non-fractional sequence
between \code{from} and \code{to}.}

\item{invalid}{\verb{[character(1) / NULL]}

One of the following invalid date resolution strategies:
\itemize{
\item \code{"previous"}: The previous valid instant in time.
\item \code{"previous-day"}: The previous valid day in time, keeping the time of
day.
\item \code{"next"}: The next valid instant in time.
\item \code{"next-day"}: The next valid day in time, keeping the time of day.
\item \code{"overflow"}: Overflow by the number of days that the input is invalid
by. Time of day is dropped.
\item \code{"overflow-day"}: Overflow by the number of days that the input is
invalid by. Time of day is kept.
\item \code{"NA"}: Replace invalid dates with \code{NA}.
\item \code{"error"}: Error on invalid dates.
}

Using either \code{"previous"} or \code{"next"} is generally recommended, as these
two strategies maintain the \emph{relative ordering} between elements of the
input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{invalid} must be supplied and
cannot be \code{NULL}. This is a convenient way to make production code robust
to invalid dates.}

\item{nonexistent}{\verb{[character / NULL]}

One of the following nonexistent time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"roll-forward"}: The next valid instant in time.
\item \code{"roll-backward"}: The previous valid instant in time.
\item \code{"shift-forward"}: Shift the nonexistent time forward by the size of
the daylight saving time gap.
\item \verb{"shift-backward}: Shift the nonexistent time backward by the size of
the daylight saving time gap.
\item \code{"NA"}: Replace nonexistent times with \code{NA}.
\item \code{"error"}: Error on nonexistent times.
}

Using either \code{"roll-forward"} or \code{"roll-backward"} is generally
recommended over shifting, as these two strategies maintain the
\emph{relative ordering} between elements of the input.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{nonexistent} must be supplied
and cannot be \code{NULL}. This is a convenient way to make production code
robust to nonexistent times.}

\item{ambiguous}{\verb{[character / zoned_time / POSIXct / list(2) / NULL]}

One of the following ambiguous time resolution strategies, allowed to be
either length 1, or the same length as the input:
\itemize{
\item \code{"earliest"}: Of the two possible times, choose the earliest one.
\item \code{"latest"}: Of the two possible times, choose the latest one.
\item \code{"NA"}: Replace ambiguous times with \code{NA}.
\item \code{"error"}: Error on ambiguous times.
}

Alternatively, \code{ambiguous} is allowed to be a zoned_time (or POSIXct) that
is either length 1, or the same length as the input. If an ambiguous time
is encountered, the zoned_time is consulted. If the zoned_time corresponds
to a naive_time that is also ambiguous \emph{and} uses the same daylight saving
time transition point as the original ambiguous time, then the offset of
the zoned_time is used to resolve the ambiguity. If the ambiguity cannot be
resolved by consulting the zoned_time, then this method falls back to
\code{NULL}.

Finally, \code{ambiguous} is allowed to be a list of size 2, where the first
element of the list is a zoned_time (as described above), and the second
element of the list is an ambiguous time resolution strategy to use when
the ambiguous time cannot be resolved by consulting the zoned_time.
Specifying a zoned_time on its own is identical to \verb{list(<zoned_time>, NULL)}.

If \code{NULL}, defaults to \code{"error"}.

If \code{getOption("clock.strict")} is \code{TRUE}, \code{ambiguous} must be supplied and
cannot be \code{NULL}. Additionally, \code{ambiguous} cannot be specified as a
zoned_time on its own, as this implies \code{NULL} for ambiguous times that the
zoned_time cannot resolve. Instead, it must be specified as a list
alongside an ambiguous time resolution strategy as described above. This is
a convenient way to make production code robust to ambiguous times.}
}
\value{
A date-time vector.
}
\description{
This is a POSIXct method for the \code{\link[=date_seq]{date_seq()}} generic.

\code{date_seq()} generates a date-time (POSIXct) sequence.

When calling \code{date_seq()}, exactly two of the following must be specified:
\itemize{
\item \code{to}
\item \code{by}
\item \code{total_size}
}
}
\section{Sequence Generation}{


Different methods are used to generate the sequences, depending on the
precision implied by \code{by}. They are intended to generate the most intuitive
sequences, especially around daylight saving time gaps and fallbacks.

See the examples for more details.
\subsection{Calendrical based sequences:}{

These convert to a naive-time, then to a year-month-day, generate
the sequence, then convert back to a date-time.
\itemize{
\item \code{by = duration_years()}
\item \code{by = duration_quarters()}
\item \code{by = duration_months()}
}
}

\subsection{Naive-time based sequences:}{

These convert to a naive-time, generate the sequence, then
convert back to a date-time.
\itemize{
\item \code{by = duration_weeks()}
\item \code{by = duration_days()}
}
}

\subsection{Sys-time based sequences:}{

These convert to a sys-time, generate the sequence, then
convert back to a date-time.
\itemize{
\item \code{by = duration_hours()}
\item \code{by = duration_minutes()}
\item \code{by = duration_seconds()}
}
}
}

\examples{
zone <- "America/New_York"

from <- date_time_build(2019, 1, zone = zone)
to <- date_time_build(2019, 1, second = 50, zone = zone)

# Defaults to second precision sequence
date_seq(from, to = to, by = 7)

to <- date_time_build(2019, 1, 5, zone = zone)

# Use durations to change to alternative precisions
date_seq(from, to = to, by = duration_days(1))
date_seq(from, to = to, by = duration_hours(10))
date_seq(from, by = duration_minutes(-2), total_size = 3)

# Note that components of `to` more precise than the precision of `by`
# must match `from` exactly. For example, this is not well defined:
from <- date_time_build(2019, 1, 1, 0, 1, 30, zone = zone)
to <- date_time_build(2019, 1, 1, 5, 2, 20, zone = zone)
try(date_seq(from, to = to, by = duration_hours(1)))

# The minute and second components of `to` must match `from`
to <- date_time_build(2019, 1, 1, 5, 1, 30, zone = zone)
date_seq(from, to = to, by = duration_hours(1))

# ---------------------------------------------------------------------------

# Invalid dates must be resolved with the `invalid` argument
from <- date_time_build(2019, 1, 31, zone = zone)
to <- date_time_build(2019, 12, 31, zone = zone)

try(date_seq(from, to = to, by = duration_months(1)))
date_seq(from, to = to, by = duration_months(1), invalid = "previous-day")

# Compare this to the base R result, which is often a source of confusion
seq(from, to = to, by = "1 month")

# This is equivalent to the overflow invalid resolution strategy
date_seq(from, to = to, by = duration_months(1), invalid = "overflow")

# ---------------------------------------------------------------------------

# This date-time is 2 days before a daylight saving time gap that occurred
# on 2021-03-14 between 01:59:59 -> 03:00:00
from <- as.POSIXct("2021-03-12 02:30:00", "America/New_York")

# So creating a daily sequence lands us in that daylight saving time gap,
# creating a nonexistent time
try(date_seq(from, by = duration_days(1), total_size = 5))

# Resolve the nonexistent time with `nonexistent`. Note that this importantly
# allows times after the gap to retain the `02:30:00` time.
date_seq(from, by = duration_days(1), total_size = 5, nonexistent = "roll-forward")

# Compare this to the base R behavior, where the hour is adjusted from 2->3
# as you cross the daylight saving time gap, and is never restored. This is
# equivalent to always using sys-time (rather than naive-time, like clock
# uses for daily sequences).
seq(from, by = "1 day", length.out = 5)

# You can replicate this behavior by generating a second precision sequence
# of 86,400 seconds. Seconds always add in sys-time.
date_seq(from, by = duration_seconds(86400), total_size = 5)

# ---------------------------------------------------------------------------

# Usage of `to` and `total_size` must generate a non-fractional sequence
# between `from` and `to`
from <- date_time_build(2019, 1, 1, 0, 0, 0, zone = "America/New_York")
to <- date_time_build(2019, 1, 1, 0, 0, 3, zone = "America/New_York")

# These are fine
date_seq(from, to = to, total_size = 2)
date_seq(from, to = to, total_size = 4)

# But this is not!
try(date_seq(from, to = to, total_size = 3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calendar.R
\name{calendar_spanning_seq}
\alias{calendar_spanning_seq}
\title{Spanning sequence: calendars}
\usage{
calendar_spanning_seq(x)
}
\arguments{
\item{x}{\verb{[clock_calendar]}

A calendar vector.}
}
\value{
A sequence along \verb{[min(x), max(x)]}.
}
\description{
\code{calendar_spanning_seq()} generates a regular sequence along the span of
\code{x}, i.e. along \verb{[min(x), max(x)]}. The sequence is generated at the
precision of \code{x}.

Importantly, sequences can only be generated if the underlying \code{\link[=seq]{seq()}} method
for the calendar in question supports a \code{from} and \code{to} value at the same
precision as \code{x}. For example, you can't compute a day precision spanning
sequence for a \code{\link[=year_month_day]{year_month_day()}} calendar (you can only compute a year
and month one). To create a day precision sequence, you'd have to convert to
a time-point first. See the individual \code{\link[=seq]{seq()}} method documentation to learn
what precisions are allowed.
}
\details{
Missing values are automatically removed before the sequence is generated.

If you need more precise sequence generation, call \code{\link[=range]{range()}} and \code{\link[=seq]{seq()}}
directly.
}
\examples{
x <- year_month_day(c(2019, 2022, 2020), c(2, 5, 3))
x

# Month precision spanning sequence
calendar_spanning_seq(x)

# Quarter precision:
x <- year_quarter_day(c(2005, 2006, 2003), c(4, 2, 3))
calendar_spanning_seq(x)

# Can't generate sequences if `seq()` doesn't allow the precision
x <- year_month_day(2019, c(1, 2, 1), c(20, 3, 25))
try(calendar_spanning_seq(x))

# Generally this means you need to convert to a time point and use
# `time_point_spanning_seq()` instead
time_point_spanning_seq(as_sys_time(x))
}
